// $Id: ReducedModel.hh,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2002):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################


#ifndef _OCS_REDUCED_MODEL_HH_
#define  _OCS_REDUCED_MODEL_HH_

#include <iostream.h>
#include <complex.h>

#include "ocs.hh"


#include "ReduStartup.hh"

/*
#include "ocsHistogram.hh"
#include "ocsRanNumGen.hh"
#include "ocsReduOptSignal.hh"
#include "ocsReduOptFiber.hh"
#include "ocsReduOptAmplifier.hh"
#include "ocsReduPolDepAttenuator.hh"
#include "ocsReduPolarizTransformer.hh"

#include "ocsOptSignal.hh"
#include "ocsOptFilter.hh"
#include "ocsElecFilter.hh"
#include "ocsPhotodetector.hh"
#include "ocsElecSignalStat.hh"
*/

//! This class contains the application methods we used to 
//! validate the reduced model and the importance sampling
//! method for outage probability due to PMD, PDL and PDG in WDM systems.
//! See the paper by Ivan Lima et al. in OFC 2002, paper TuI7 for
//! a discussion of the importance sampling method when the PDG is zero.

class ReducedModel
{
   public:

//! Constructor

/*!

The constructor initializes the following class parameters
by reading in their values from the Startup.in file:

-# ReducedModel::OpticalBandwidth 
-# ReducedModel::ElectricalBandwidth 
-# ReducedModel::PolDepAttenuation 
-# ReducedModel::qtLoops
-# ReducedModel::IsStraightLine
-# ReducedModel::LoopPolarizControl
-# ReducedModel::ChannelArrayIndex
-# ReducedModel::InitialSample
-# ReducedModel::qtStatSamples
-# ReducedModel::qtMapSpans
-# ReducedModel::BandwidthOSA
-# ReducedModel::PolarizNoiseBandwidth
-# ReducedModel::BiasingParam

The rest of the constructor is fairly easy to understand with the aid
of other parts of this documentation.

 */

      ReducedModel(MyStartup *oMyStartup);

  //! Destructor

      ~ReducedModel();      

      void SweepReduPDL(double MinPDL, double MaxPDL,
                        int qtSweepPDL, double PowerPenaltyMargin);
      void GetReduRelativeQfactor_pdf(void);
      void GetReduQfactor_pdf(void);     
      void GetReduQfactorWithPeriodicity_pdf(void);                                    
      void GetReduOutageProb(double PowerPenaltyMargin,
                             int MinNumChannels, 
                             int MaxNumChannels,
                             int StepNumChannels);       
      void GetSignalFreeNoiseEvolution(void);    
      

      double GetQ_FactorSemiAnalytical(int ChannelArrayIndex);  
      double GetESNR_SemiAnalytical(int ChannelArrayIndex);
                            
   private:
      void TransmitReduOptSystem(double *PropagatedLength);
      void TransmitReduOptSystem1PDL_Loop(double *PropagatedLength);
      double TransmitReduOptSystem1PDL_LoopWithMonitoring
                             (double *PropagatedLength);      

      RanNumGen *oRanPMD;
      ReduOptSignal *oReduOptSignal;
      ReduPolDepAttenuator *oReduPolDepAttenuator;
      ReduOptAmplifier *oReduOptAmplifier;
      Histogram *oHistQfactor;
      Histogram *oHistOptSNR;
      Histogram *oHistElecTimeDomainSNR;
      ReduOptFiber *oReduOptFiber[100]; //## Up to 100 fibers per map      

      //## Objects required to compute the enhancement factor and
      //#  the Q-factor using the pulse format and receiver 
      //#  characteristics
      RanNumGen *oRanBitString;
      OptSignal  *oOptSignal;
      OptFilter  *oOptFilter;
      ElecFilter *oElecFilter;
      Photodetector *oPhotodetector;
      ElecSignalStat *oElecSignalStat;
      //## Parameters required to compute the Q-factor as a function
      //#  of the OSNR
      double EnhancementFactor;
      double ExtinctionRatio;
      double NormNumberNoiseModes;
      double NormKcoeff1;
      double NormKcoeff0;

  //! Used to initialize ReduOptSignal::SignalNoiseBandwidth

      double BandwidthOSA;      

  //! Used to initialize ReduOptSignal::PolarizationNoiseBandwidth

      double PolarizNoiseBandwidth;


  //! The bandwidth of the optical filter in the reciever in Hz. 
                          
  /*! Typically 60 GHz for a 10 Gb/s signal.
      But see the recent papers of Winzer (in JLT?)
      for a more thorough study of optimal filter
      bandwidths for different pulse formats.

      \sa ElectricalBandwidth, ReduOptSignal::OptFilterBandwidth
  */

      double OpticalBandwidth;

  //! The bandwidth of the Electrical filter in the receiver in Hz. 
                          
    /*!  Typically 8.6 GHz for a 10 Gb/s signal.
      But see the recent papers of Winzer (in JLT?).

      \sa OpticalBandwidth, ReduOptSignal::ElecFilterBandwidth
  */
     
      double ElectricalBandwidth;                

  //! The polarization dependent attenuation in dB
  /*!
      \sa void ReduPolDepAttenuator::SetPolDepAttenuation(double, double)
   */
                 
      double PolDepAttenuation;

  //! The number of Monte Carlo experiments (either standard or with 
  //! importnce smpling).

      int qtStatSamples;

  //! Flag that takes value one when the system is a straight line system
  //! as opposed to a recirculating loop.

  /*!
      In a recirculating loop the random realization of the PMD that
      the signal experiences repeats periodically each round trip of the 
      loop. In combination with PDL and amplifier gain saturation this
      effect results in a very different Delta Q distribution for a loop
      than for  straight line system in which the PMD rotations are truely
      random. See Yu Sun et al. in OFC 2002, paper ThI4 for more details.
   */

      int IsStraightLine;                                   

  //! The number of maps in one loop
  /*!
      In the reduced model we do not have dispersion maps. However
      we still have maps that determine  the distance between amplifiers. \n
      The parameter qtMapSpans determines how many maps there are in a loop.
      

      \sa ReducedModel::qtLoops

   */
      int qtMapSpans;

 //! The number of loops in the entire system.
  /*!
      This parameter together with qtMpSpans determines the total length
      of the system. 
      
      \sa ReducedModel::qtLoops, ReducedModel:IsStraightLine

   */

      int qtLoops;
      
  //! Flag to indicate whether we have a polarization controller in the loop.

      int LoopPolarizControl;

  //! Used to help restart a simulation

      int InitialSample;

  //! Used to help restart a simulation

      int CurrentSample;

  //! The index of the channel that we are monitoring. 
  /*! Values are in range 0 <= ChannelArrayIndex <= NumChannels - 1.
      
      In some methods in this class we compute the Delta Q distribution and
      outage probability for a particular channel. 

  */

      int ChannelArrayIndex;

 
        //## For the simplified loop randomization

   //! For loops. You won't use it!

  /*!
      \sa ReduPolarizTransformer
   */


        ReduPolarizTransformer *oReduPolarizController;

//! For loops. You won't use it.

  /*!
      \sa ReduPolarizTransformer
   */



        RanNumGen *oRanReduPolarizController;
      
      //## Importance Sampling



  //! The biasing parameter used in importance sampling

      double BiasingParam;

  //! The likelihood ratio used in importance sampling

      double LikelihoodRatio;
 
 
  //! Used in importance sampling for outage due to PMD, PDL and PDG

  /*!
      \sa ReduPolarizTransformer
   */

 
      ReduPolarizTransformer *oReduPolarizTransformer;

  //! Used in importance sampling for outage due to PMD, PDL and PDG

  /*!
      \sa ReduPolarizTransformer
   */

      RanNumGen *oRanImportSampPolarizTransform;      
};

#endif /* _OCS_REDUCED_MODEL_HH_ */





