//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################


// ###############################################
// 
// The applications in this file were written by John Zweck
// and are not to be modified by anyone else for now.
// 
// ###############################################

#include <iostream.h>
#include <complex.h>

#include "ocs.hh"
#include "StartupLinearn.hh"

#include <time.h>
#include <sys/time.h>
#include <sys/times.h>


extern ofstream *LogFile;

void Tests(StartupZweck * Startup);
void Soliton(StartupZweck * Startup);
void TycoNoiseFree(StartupZweck * Startup);

// ####################################################

// ## Application Wrapper

void MyApplication(int argc, char *argv[])  
{

  StartupZweck Startup("Startup.in");

  switch(Startup.GetSimType())
   {
    
     case 1: 
       Tests(&Startup);
       break;

     case 2:
       Soliton(&Startup);
       break;  

     case 3:
       TycoNoiseFree(&Startup);
       break;

    default:
      cerr << "MyApplication: Invalid choice of application in SimType" 
           << endl;
      exit(1);
          
   } // end switch

} // end MyAppn


// ####################################################

void Tests(StartupZweck * Startup)
{

  time_t StartTime,EndTime;
  struct tms my_tms;

  StartTime = time(NULL);
  cout << "Start Time: " << ctime(&StartTime) << endl;

  LogFileSeparator();

  *LogFile <<  endl << "Convention Tests" << endl;

  string InDir = Startup->GetInputFilesDir();
  string Job = Startup->GetJobPrefix();

  RanNumGen RNGSignal(InDir + "RanNumGenSig.in");
  RanNumGen RNGAmps(InDir + "RanNumGenAmps.in");
  RanNumGen RNGFibers(InDir + "RanNumGenFibers.in");

   // ## Initiating the optical pulses

  OptSignal * Signal;

  Signal = new OptSignal("Signal.in",&RNGSignal);

  OptDataWriter DataWriter(InDir + "DataWriter.in",InDir,Job,Signal);

  cout << "tAveragePower [mW] = " << 1e3*Signal->GetAveragePower() << endl;
  cout << "fAveragePower [mW] = " << 1e3*Signal->GetAveragePowerFreqDomain() << endl;

  Signal->WriteLocalFrequencyAndChirp(Job+"LF.dat");
  DataWriter.WriteInitialData();

  OptFiberLocalError Fiber("Fiber.in",Signal,&RNGSignal,Job);

  double PropagatedLength = 0.0;
  int ZStepNum = 0;

  /***

  // Uncomment for a first-order soliton

  Fiber.SetFstOrDispFiber(-0.1*1e-27);  // -0.1 ps^2/km
  Fiber.SetSndOrDispFiber(0*1e-39);     // 0 ps^3/km
  Fiber.SetGammaFiber(2.2*1e-3);        // 2.2 W^{-1}km^{-1}


  // Uncomment for pure third order dispersion

  Fiber.SetFstOrDispFiber(0);  // -0.1 ps^2/km
  Fiber.SetSndOrDispFiber(1e-39);     // 0 ps^3/km
  Fiber.SetGammaFiber(0);        // 2.2 W^{-1}km^{-1}

  ****/

  Fiber.Propagate(&PropagatedLength,&ZStepNum);

  DataWriter.WriteEndTransmission(0,0,PropagatedLength);

  OptFilter* TunableFilter;
  TunableFilter = new OptFilter(InDir + "TunableOptFilter.in",
				Signal->GetTypeSimulation());

  OptDemuxer *oOptDemuxer;
  oOptDemuxer = new OptDemuxer(Signal, TunableFilter);
  OptSignal * SCSignal;
  SCSignal = new OptSignal(oOptDemuxer->ExtractChannel(0));


  Photodetector oPhotodetector("Photodetector.in",SCSignal,&RNGSignal);
  oPhotodetector.DetectOptSignal();
  oPhotodetector.WriteFileTimePower(Job + "B4.dat");

  //oPhotodetector.TimeShiftSignal(25e-12);
  //oPhotodetector.WriteFileTimePower(Job + "After.dat");

  ElecFilter * oElecFilter;
  oElecFilter = new ElecFilter(InDir + "ElecFilter.in",&oPhotodetector);

  ElecSignalStat *oElecSignalStat;
  oElecSignalStat = new  ElecSignalStat(SCSignal,&oPhotodetector);
			

  Histogram2D*RonEyeDiagram;
  RonEyeDiagram = new  Histogram2D(InDir + "Histogram2D.in", SCSignal);

  //  oElecFilter->FilterElecSignal();     
  oElecSignalStat->GetSingleStringElecSignalStatConvoluting();
  oElecSignalStat->WriteFileEyeDiagram(Job+"Eye.dat"); 

  RonEyeDiagram->UpdateEyeDiagram(SCSignal,&oPhotodetector);
  RonEyeDiagram->WriteEyeDiagramHistogram(Job+"RonEye.dat");

  EndTime = time(NULL);
  double TotalTime = double(EndTime - StartTime);

  times(&my_tms);

  double CPUtime = (double(my_tms.tms_utime))/(double(sysconf(_SC_CLK_TCK)));

  cout << "Total Run Time [hrs] = " << TotalTime/3600.0 << endl
       << "Total CPU Time [hrs] = " << CPUtime/3600.0 << endl
       << "Total CPU Time [secs] = " << CPUtime << " = " 
       << 100.0*CPUtime/TotalTime << "%" << endl
       << "End Time: " << ctime(&EndTime) << endl;

   cout << "THE END" << endl;
   *LogFile << "THE END" << endl;


} // ## end Tests

// #################################################

void Soliton(StartupZweck * Startup)
{

  LogFileSeparator();

  *LogFile <<  endl << "Soliton" << endl;

  string InDir = Startup->GetInputFilesDir();
  string Job = Startup->GetJobPrefix();

  RanNumGen RNGSignal(InDir + "RanNumGenSig.in");
  RanNumGen RNGFiber(InDir + "RanNumGenFiber.in");

   // ## Initiating the optical pulses

  OptSignal Signal("Signal.in",&RNGSignal);

  OptDataWriter DataWriter(InDir + "DataWriter.in",InDir,Job,&Signal);

  DataWriter.WriteInitialData();
  Signal.WriteComplexFields(Job+"tComplexIn.dat",0);

  OptFiberLocalError Fiber("Fiber.in",&Signal,&RNGFiber,Job);

  double PropagatedLength = 0.0;
  int ZStepNum = 0;

  // Uncomment for a first-order soliton

  Fiber.SetFstOrDispFiber(-0.1*1e-27);  // -0.1 ps^2/km
  Fiber.SetSndOrDispFiber(0*1e-39);     // 0 ps^3/km
  Fiber.SetGammaFiber(2.2*1e-3);        // 2.2 W^{-1}km^{-1}

  /***

  // Uncomment for pure third order dispersion

  Fiber.SetFstOrDispFiber(0);  // -0.1 ps^2/km
  Fiber.SetSndOrDispFiber(1e-39);     // 0 ps^3/km
  Fiber.SetGammaFiber(0);        // 2.2 W^{-1}km^{-1}

  ***/

  Fiber.Propagate(&PropagatedLength,&ZStepNum);
  Signal.WriteComplexFields(Job+"tComplexOut.dat",0);

  DataWriter.WriteEndTransmission(0,0,PropagatedLength);

  cout << "PropagatedLength [km] = " << 1e-3*PropagatedLength << endl;
  cout << "Number of Steps = " << ZStepNum << endl;
  cout << "THE END" << endl;

} // ## end Soliton

// #################################################

void TycoNoiseFree(StartupZweck * Startup)
{

  time_t StartTime,EndTime;
  struct tms my_tms;

  StartTime = time(NULL);
  cout << "Start Time: " << ctime(&StartTime) << endl;

  LogFileSeparator();

  *LogFile <<  endl << "TycoNoiseFreee" << endl;

  string InDir = Startup->GetInputFilesDir();
  string Job = Startup->GetJobPrefix();

  RanNumGen RNGSignal(InDir + "RanNumGenSig.in");
  RanNumGen RNGAmps(InDir + "RanNumGenAmps.in");
  RanNumGen RNGFibers(InDir + "RanNumGenFibers.in");

   // ## Initiating the optical pulses

  OptSignal * Signal;

  Signal = new OptSignal("Signal.in",&RNGSignal);

  //OptFilter TransmitterOptFilter("OptFilterTx.in",Signal);
    // Careful re this when do WDM!!

  //TransmitterOptFilter.FilterOptSignal();

  OptDataWriter DataWriter(InDir + "DataWriter.in",InDir,Job,Signal);

  DataWriter.WriteInitialData();
 
   // ## Declare and initialize the fibers 

    OptFiberLocalError PreDCF(InDir + "PreDCF.in",Signal,&RNGFibers,Job);
    OptFiberLocalError PostDCF(InDir + "PostDCF.in",Signal,&RNGFibers,Job);
    OptFiberLocalError DSF35(InDir + "DSF35.in",Signal,&RNGFibers,Job);
    OptFiberLocalError DSF45(InDir + "DSF45.in",Signal,&RNGFibers,Job);
    OptFiberLocalError SMF(InDir + "SMF.in",Signal,&RNGFibers,Job);

   
    int NumFibers = 5;

    OptFiberLocalError * FiberPtr[NumFibers];

    FiberPtr[0] = &PreDCF;
    FiberPtr[1] = &PostDCF;
    FiberPtr[2] = &DSF45;
    FiberPtr[3] = &DSF35;
    FiberPtr[4] = &SMF;

    OptAmplifier Amp(InDir + "Amp.in",Signal,&RNGAmps);
    OptAmplifier PAmp(InDir + "PAmp.in",Signal,&RNGAmps);

    int NumAmps = 2;

    OptAmplifier * AmpPtr[NumAmps];

    AmpPtr[0] =  &Amp;
    AmpPtr[1] =  &PAmp;

  if(Startup->GetDebugLevelAmps())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetDebugLevel(Startup->GetDebugLevelAmps());
     }
   
  for(int i=0;i<NumAmps;i++)
    AmpPtr[i]->SetTypeAmplifier(typeAmplifier(Startup->GetTypeAmp()));
     
   if(Startup->GetAmpsNoiseOffFlag())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetTypeAmplifierNoise(NOISE_OFF);
     }

   if(Startup->GetBeta3ZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetSndOrDispFiber(0);
     }

   if(Startup->GetNonLinearityZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetGammaFiber(0);
     }

   if(Startup->GetFiberLossZeroFlag()) 
        for(int ii = 0; ii< NumFibers;ii++)
            FiberPtr[ii]->SetAttenuation(0);

    for(int ii = 0; ii< NumFibers;ii++)
         FiberPtr[ii]->SetRelativeErrorGoal(Startup->GetRelativeErrorGoal());

      for(int ii = 0; ii< NumFibers;ii++)
	{
	  FiberPtr[ii]->SetTypeStepSizes( Startup->GetTypeStepSizes());

          if(Startup->GetTypeStepSizes() == 3)
	    FiberPtr[ii]->SetWalkOffParameter(Startup->GetWalkOffParameter());

           FiberPtr[ii]->SetOutputStepSizesFlag(
                            Startup->GetOutputStepSizesFlag());
	}


       // #### Propagate signal through fiber ####

       double PropagatedLength = 0;
       int ZStepNum = 0;



      PreDCF.Propagate(&PropagatedLength,&ZStepNum);
      
      PAmp.AmplifyOptSignal();

      int kk = 0;

      DataWriter.WriteEndMap(0,kk,PropagatedLength);

    for(kk = 0; kk < Startup->GetQtMapSpans(); kk++) 
     {
       SMF.Propagate(&PropagatedLength,&ZStepNum);
       DSF35.Propagate(&PropagatedLength,&ZStepNum);

       Amp.AmplifyOptSignal();

       DSF45.Propagate(&PropagatedLength,&ZStepNum);
       Amp.AmplifyOptSignal();
       DSF45.Propagate(&PropagatedLength,&ZStepNum);
       Amp.AmplifyOptSignal();
     
      
       DSF35.Propagate(&PropagatedLength,&ZStepNum);
       SMF.Propagate(&PropagatedLength,&ZStepNum);
       Amp.AmplifyOptSignal();

       DataWriter.WriteEndMap(0,kk,PropagatedLength);

      if(kk % 10 == 0 && kk != 0)
	cout << "Completed " << kk << " periods of dispersion map " << endl;

    }// ## end for kk-loop

    PostDCF.Propagate(&PropagatedLength,&ZStepNum);
    PAmp.AmplifyOptSignal();

    // ############### End of Propagation #################

  DataWriter.WriteEndTransmission(0,kk,PropagatedLength);
 

        cout <<"Total PropagatedLength [km] = "<<PropagatedLength*1e-3<< endl;
        cout <<"Number Steps in Z taken by solver = " << ZStepNum << endl;


  OptFilter* TunableFilter;
  TunableFilter = new OptFilter(InDir + "TunableOptFilter.in",
				Signal->GetTypeSimulation());

  OptDemuxer *oOptDemuxer;
  oOptDemuxer = new OptDemuxer(Signal, TunableFilter);
  OptSignal * SCSignal;
  SCSignal = new OptSignal(oOptDemuxer->ExtractChannel(0));


  Photodetector oPhotodetector("Photodetector.in",SCSignal,&RNGSignal);
  oPhotodetector.DetectOptSignal();
  
  ElecFilter * oElecFilter;
  oElecFilter = new ElecFilter(InDir + "ElecFilter.in",&oPhotodetector);

  ElecSignalStat *oElecSignalStat;
  oElecSignalStat = new  ElecSignalStat(SCSignal,&oPhotodetector);
			
  Histogram2D*RonEyeDiagram;
  RonEyeDiagram = new  Histogram2D(InDir + "Histogram2D.in", SCSignal);

  oElecFilter->FilterElecSignal();     

  oPhotodetector.WriteFileTimePower(Job+"tPD.dat");

  oElecSignalStat->GetSingleStringElecSignalStatConvolutingNEW();
  oElecSignalStat->WriteFileEyeDiagram(Job+"Eye.dat"); 

  RonEyeDiagram->UpdateEyeDiagram(SCSignal,&oPhotodetector);
  RonEyeDiagram->WriteEyeDiagramHistogram(Job+"RonEye.dat");

  cout << "MinAmpMargin = " << oElecSignalStat->GetAmplitudeMargin() << endl;
  cout << "Q factor = " << oElecSignalStat->GetQ_Factor() << endl;

  EndTime = time(NULL);
  double TotalTime = double(EndTime - StartTime);

  times(&my_tms);

  double CPUtime = (double(my_tms.tms_utime))/(double(sysconf(_SC_CLK_TCK)));

  cout << "Total Run Time [hrs] = " << TotalTime/3600.0 << endl
       << "Total CPU Time [hrs] = " << CPUtime/3600.0 << endl
       << "Total CPU Time [secs] = " << CPUtime << " = " 
       << 100.0*CPUtime/TotalTime << "%" << endl
       << "End Time: " << ctime(&EndTime) << endl;

   cout << "THE END" << endl;
   *LogFile << "THE END" << endl;
   cout << "PHYSICS" << endl;

} // ## end TycoNoiseFree

// ############################################################






