//###################################################################
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################


// ###############################################
// 
// The applications in this file were written by John Zweck
// and are not to be modified by anyone else for now.
// 
// ###############################################

#include <iostream.h>
#include <complex.h>

#include "ocs.hh"
#include "StartupLinearn.hh"

#include <time.h>
#include <sys/time.h>
#include <sys/times.h>


extern ofstream *LogFile;

void TycoNoiseFree(StartupZweck * Startup);
void TycoNoiseCovariance(StartupZweck * Startup);
void OlegNoiseCovariance(StartupZweck * Startup);

void MarcoNoiseCovariance(StartupZweck * Startup);

void TestNoiseCovariance(StartupZweck * Startup);

void TestChiSqRxArbPol(StartupZweck * Startup);

void TestNoiseCovarianceAssignmentOperator(StartupZweck * Startup);

void TestChiSquareReceiver(StartupZweck * Startup);

// ####################################################

// ## Application Wrapper

void MyApplication(int argc, char *argv[])  
{

  StartupZweck Startup("Startup.in");

  switch(Startup.GetSimType())
   {
    
     case 1: 
       TycoNoiseFree(&Startup);
       break;

     case 2: 
       TycoNoiseCovariance(&Startup);
       break;

     case 3: 
       OlegNoiseCovariance(&Startup);
       break;

     case 4: 
       MarcoNoiseCovariance(&Startup);
       break;

     case 5: 
       TestNoiseCovariance(&Startup);
       break;

     case 6: 
       TestChiSqRxArbPol(&Startup);
       break;

     case 7:
       TestNoiseCovarianceAssignmentOperator(&Startup);
       break;

      case 8:
	TestChiSquareReceiver(&Startup);
        break;

    default:
      cerr << "MyApplication: Invalid choice of application in SimType" 
           << endl;
      exit(1);
          
   } // end switch

} // end MyAppn


// ####################################################

void TycoNoiseFree(StartupZweck * Startup)
{

  time_t StartTime,EndTime;
  struct tms my_tms;

  StartTime = time(NULL);
  cout << "Start Time: " << ctime(&StartTime) << endl;

  LogFileSeparator();

  *LogFile <<  endl << "TycoNoiseFreee" << endl;

  string InDir = Startup->GetInputFilesDir();
  string Job = Startup->GetJobPrefix();

  RanNumGen RNGSignal(InDir + "RanNumGenSig.in");
  RanNumGen RNGAmps(InDir + "RanNumGenAmps.in");
  RanNumGen RNGFibers(InDir + "RanNumGenFibers.in");

   // ## Initiating the optical pulses

  OptSignal * Signal;

  Signal = new OptSignal("Signal.in",&RNGSignal);

  //OptFilter TransmitterOptFilter("OptFilterTx.in",Signal);
    // Careful re this when do WDM!!

  //TransmitterOptFilter.FilterOptSignal();

  OptDataWriter DataWriter(InDir + "DataWriter.in",InDir,Job,Signal);

  DataWriter.WriteInitialData();
 
   // ## Declare and initialize the fibers 

    OptFiberLocalError PreDCF(InDir + "PreDCF.in",Signal,&RNGFibers,Job);
    OptFiberLocalError PostDCF(InDir + "PostDCF.in",Signal,&RNGFibers,Job);
    OptFiberLocalError DSF35(InDir + "DSF35.in",Signal,&RNGFibers,Job);
    OptFiberLocalError DSF45(InDir + "DSF45.in",Signal,&RNGFibers,Job);
    OptFiberLocalError SMF(InDir + "SMF.in",Signal,&RNGFibers,Job);

   
    int NumFibers = 5;

    OptFiberLocalError * FiberPtr[NumFibers];

    FiberPtr[0] = &PreDCF;
    FiberPtr[1] = &PostDCF;
    FiberPtr[2] = &DSF45;
    FiberPtr[3] = &DSF35;
    FiberPtr[4] = &SMF;

    OptAmplifier Amp(InDir + "Amp.in",Signal,&RNGAmps);
    OptAmplifier PAmp(InDir + "PAmp.in",Signal,&RNGAmps);

    int NumAmps = 2;

    OptAmplifier * AmpPtr[NumAmps];

    AmpPtr[0] =  &Amp;
    AmpPtr[1] =  &PAmp;

  if(Startup->GetDebugLevelAmps())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetDebugLevel(Startup->GetDebugLevelAmps());
     }
   
  for(int i=0;i<NumAmps;i++)
    AmpPtr[i]->SetTypeAmplifier(typeAmplifier(Startup->GetTypeAmp()));
     
   if(Startup->GetAmpsNoiseOffFlag())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetTypeAmplifierNoise(NOISE_OFF);
     }

   if(Startup->GetBeta3ZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetSndOrDispFiber(0);
     }

   if(Startup->GetNonLinearityZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetGammaFiber(0);
     }

   if(Startup->GetFiberLossZeroFlag()) 
        for(int ii = 0; ii< NumFibers;ii++)
            FiberPtr[ii]->SetAttenuationdB(0);

    for(int ii = 0; ii< NumFibers;ii++)
         FiberPtr[ii]->SetRelativeErrorGoal(Startup->GetRelativeErrorGoal());

      for(int ii = 0; ii< NumFibers;ii++)
	{
	  FiberPtr[ii]->SetTypeStepSizes( Startup->GetTypeStepSizes());

          if(Startup->GetTypeStepSizes() == 3)
	    FiberPtr[ii]->SetWalkOffParameter(Startup->GetWalkOffParameter());

           FiberPtr[ii]->SetOutputStepSizesFlag(
                            Startup->GetOutputStepSizesFlag());
	}


       // #### Propagate signal through fiber ####

       double PropagatedLength = 0;
       int ZStepNum = 0;



      PreDCF.Propagate(&PropagatedLength,&ZStepNum);
      
      PAmp.AmplifyOptSignal();

      int kk = 0;

      DataWriter.WriteEndMap(0,kk,PropagatedLength);

    for(kk = 0; kk < Startup->GetQtMapSpans(); kk++) 
     {
       SMF.Propagate(&PropagatedLength,&ZStepNum);
       DSF35.Propagate(&PropagatedLength,&ZStepNum);

       Amp.AmplifyOptSignal();

       DSF45.Propagate(&PropagatedLength,&ZStepNum);
       Amp.AmplifyOptSignal();
       DSF45.Propagate(&PropagatedLength,&ZStepNum);
       Amp.AmplifyOptSignal();
     
      
       DSF35.Propagate(&PropagatedLength,&ZStepNum);
       SMF.Propagate(&PropagatedLength,&ZStepNum);
       Amp.AmplifyOptSignal();

       DataWriter.WriteEndMap(0,kk,PropagatedLength);

      if(kk % 10 == 0 && kk != 0)
	cout << "Completed " << kk << " periods of dispersion map " << endl;

    }// ## end for kk-loop

    PostDCF.Propagate(&PropagatedLength,&ZStepNum);
    PAmp.AmplifyOptSignal();

    // ############### End of Propagation #################

  DataWriter.WriteEndTransmission(0,kk,PropagatedLength);

  Photodetector * oPhotodetector;
  ElecFilter * oElecFilter;

  oPhotodetector = new Photodetector("Photodetector.in",Signal,&RNGSignal);
  oElecFilter = new ElecFilter("ElecFilter.in",oPhotodetector);

  oPhotodetector->DetectOptSignal();
  oElecFilter->FilterElecSignal();

  oPhotodetector->WriteFileTimePower(Job  + "tElecFiltNoiseFreeSignal.dat");
  oPhotodetector->WriteFileFreqPower(Job  + "fElecFiltNoiseFreeSignal.dat",0);


        cout <<"Total PropagatedLength [km] = "<<PropagatedLength*1e-3<< endl;
        cout <<"Number Steps in Z taken by solver = " << ZStepNum << endl;


  EndTime = time(NULL);
  double TotalTime = double(EndTime - StartTime);

  times(&my_tms);

  double CPUtime = (double(my_tms.tms_utime))/(double(sysconf(_SC_CLK_TCK)));

  cout << "Total Run Time [hrs] = " << TotalTime/3600.0 << endl
       << "Total CPU Time [hrs] = " << CPUtime/3600.0 << endl
       << "Total CPU Time [secs] = " << CPUtime << " = " 
       << 100.0*CPUtime/TotalTime << "%" << endl
       << "End Time: " << ctime(&EndTime) << endl;

   cout << "THE END" << endl;
   *LogFile << "THE END" << endl;


} // ## end TycoNoiseFree

// #################################################

void TycoNoiseCovariance(StartupZweck * Startup)
{

  time_t StartTime,EndTime;
  struct tms my_tms;

  StartTime = time(NULL);
  cout << "Start Time: " << ctime(&StartTime) << endl;

  LogFileSeparator();

  *LogFile <<  endl << "TycoNoiseCovarince" << endl;

  string InDir = Startup->GetInputFilesDir();
  string Job = Startup->GetJobPrefix();

  RanNumGen RNGSignal(InDir + "RanNumGenSig.in");
  RanNumGen RNGAmps(InDir + "RanNumGenAmps.in");
  RanNumGen RNGFibers(InDir + "RanNumGenFibers.in");

  // ## Initiating the optical pulses

  OptSignal * Signal;

  Signal = new OptSignal("Signal.in",&RNGSignal);

 
  OptDataWriter DataWriter(InDir + "DataWriter.in",InDir,Job,Signal);

  DataWriter.WriteInitialData();
 
   // ## Declare and initialize the fibers 

  OptFiberLocalError * PreDCF;
  OptFiberLocalError * PostDCF;
  OptFiberLocalError * DSF35;
  OptFiberLocalError * DSF45;
  OptFiberLocalError * SMF;


    PreDCF = new OptFiberLocalError(InDir + "PreDCF.in",Signal,&RNGFibers,Job);

    //  PostDCF = new OptFiberLocalError(InDir + "PostDCF.in" ...

    PostDCF = new OptFiberLocalError(InDir + "PostDCF.in",Signal,&RNGFibers,
				     Job);

    DSF35 = new OptFiberLocalError(InDir + "DSF35.in",Signal,&RNGFibers,Job);
    DSF45 = new OptFiberLocalError(InDir + "DSF45.in",Signal,&RNGFibers,Job);
    SMF = new OptFiberLocalError(InDir + "SMF.in",Signal,&RNGFibers,Job);

   
    int NumFibers = 5;

    OptFiberLocalError * FiberPtr[NumFibers];

    FiberPtr[0] = PreDCF;
    FiberPtr[1] = PostDCF;
    FiberPtr[2] = DSF45;
    FiberPtr[3] = DSF35;
    FiberPtr[4] = SMF;


    OptAmplifier * Amp, *PAmp;

    Amp =  new OptAmplifier(InDir + "Amp.in",Signal,&RNGAmps);
    PAmp =  new OptAmplifier(InDir + "PAmp.in",Signal,&RNGAmps);

    int NumAmps = 2;

    OptAmplifier * AmpPtr[NumAmps];

    AmpPtr[0] =  Amp;
    AmpPtr[1] =  PAmp;

  if(Startup->GetDebugLevelAmps())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetDebugLevel(Startup->GetDebugLevelAmps());
     }
   
  for(int i=0;i<NumAmps;i++)
    AmpPtr[i]->SetTypeAmplifier(typeAmplifier(Startup->GetTypeAmp()));
     
   if(Startup->GetAmpsNoiseOffFlag())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetTypeAmplifierNoise(NOISE_OFF);
     }

   if(Startup->GetBeta3ZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetSndOrDispFiber(0);
     }

   if(Startup->GetNonLinearityZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetGammaFiber(0);
     }

   if(Startup->GetFiberLossZeroFlag()) 
        for(int ii = 0; ii< NumFibers;ii++)
            FiberPtr[ii]->SetAttenuationdB(0);

    for(int ii = 0; ii< NumFibers;ii++)
         FiberPtr[ii]->SetRelativeErrorGoal(Startup->GetRelativeErrorGoal());

      for(int ii = 0; ii< NumFibers;ii++)
	{
	  FiberPtr[ii]->SetTypeStepSizes( Startup->GetTypeStepSizes());

          if(Startup->GetTypeStepSizes() == 3)
	    FiberPtr[ii]->SetWalkOffParameter(Startup->GetWalkOffParameter());

           FiberPtr[ii]->SetOutputStepSizesFlag(
                            Startup->GetOutputStepSizesFlag());
	}



  NoiseCovariance * NoiseCov;

  NoiseCov = new NoiseCovariance(InDir + "NoiseCovariance.in",Job,Signal);


  ChiSquareReceiver *  ChiSqRx;
 
  ChiSqRx = new ChiSquareReceiver("ChiSquareReceiver.in",NoiseCov);


/****

   typeReceiver TypeReceiver = SEMI_ANALYTICAL;

  OptSignal SignalAtEndTransmission = *Signal;


   ReceiverSemiAnalytical Receiver(TypeReceiver,
				  &SignalAtEndTransmission,
				  &RNGAmps,
				  0,
				  InDir, 
				  Job);
****/

  // #### Propagate signal through fiber ####

    cout << "Input Av Power [mW] = " 
	 << 1e3*Signal->GetAveragePower() << endl;
 
    cout << "Input Av Power [dBm] = " 
	 << Linear2dBm(Signal->GetAveragePower()) << endl;

  // JZ June 14, 2006. Put no loss in PreDCF and comment out PAmp


  PreDCF->SetAttenuationdB(0);
  PreDCF->SetGammaFiber(0);

  NoiseCov->Propagate(PreDCF);
  
  //  NoiseCov->Amplify(PAmp);

  DataWriter.WriteWithinMap(0,0,NoiseCov->GetPropagatedLength());

  int kk = 0;

  if(Startup->GetJitterSeparationFrequency() > 0)
    {
     NoiseCov->TurnOffJitterSeparation();
     cout << "Before entering maps for loop TurnOffJitterSeparation" <<endl;
    }

  for(kk = 1; kk <= Startup->GetQtMapSpans(); kk++) 
   {

      if(Startup->GetJitterSeparationFrequency() > 0 &&
         kk % Startup->GetJitterSeparationFrequency() == 0)
        {
          NoiseCov->TurnOnJitterSeparation();
          cout << "Starting Map " << kk << ": TurnOnJitterSeparation" << endl; 
        }

     NoiseCov->Propagate(SMF);

      if(Startup->GetJitterSeparationFrequency() > 0 &&
         kk % Startup->GetJitterSeparationFrequency() == 0)
        {
          NoiseCov->TurnOffJitterSeparation();
          cout << "Starting Map " << kk << ": TurnOffJitterSeparation" << endl;
        }

     DataWriter.WriteWithinMap(0,kk,NoiseCov->GetPropagatedLength());
     NoiseCov->Propagate(DSF35);
     NoiseCov->Amplify(Amp);

     DataWriter.WriteWithinMap(0,kk,NoiseCov->GetPropagatedLength());
      

     NoiseCov->Propagate(DSF45);
     NoiseCov->Amplify(Amp);

     DataWriter.WriteWithinMap(0,kk,NoiseCov->GetPropagatedLength());
      


     NoiseCov->Propagate(DSF45);
     NoiseCov->Amplify(Amp);

     DataWriter.WriteWithinMap(0,kk,NoiseCov->GetPropagatedLength());
      

     
     NoiseCov->Propagate(DSF35);
     DataWriter.WriteWithinMap(0,kk,NoiseCov->GetPropagatedLength());
     NoiseCov->Propagate(SMF);
     NoiseCov->Amplify(Amp);

     DataWriter.WriteEndMap(0,kk,NoiseCov->GetPropagatedLength());

      if(kk % 5 == 0 && kk != 0)
	cout << "Completed " << kk << " periods of dispersion map " << endl;

    }// ## end for kk-loop

    // JZ June 14, 2006. Put no loss in PostDCF and comment out PAmp

    NoiseCov->TurnOnJitterSeparation();
    cout << "Before very last fiber: TurnOnJitterSeparation" <<endl;


    PostDCF->SetAttenuationdB(0);
    PreDCF->SetGammaFiber(0);

    NoiseCov->Propagate(PostDCF);
    //NoiseCov->Amplify(PAmp);

    DataWriter.WriteEndTransmission(0,0,NoiseCov->GetPropagatedLength());

    string Suffix = "";

    //NoiseCov->WriteFileNoiseFreeSignalAndCovarianceMatrix(Suffix);
 
    cout << "Output Av Power [dBm] = " 
	 << Linear2dBm(Signal->GetAveragePower()) << endl;

    if(Startup->GetWriteComplexFieldsFlag())
     Signal->WriteFileComplexFieldsTimeAndFreq(Job + "tfComplexFields.dat");



    // ############### End of Propagation #################

  //SignalAtEndTransmission = *Signal;

  ChiSqRx->ReceiveSignal(Suffix);

  ChiSqRx->ComputeGaussianPDFsAndBER();

  //Receiver.ReceiveSignal(34);
  
  double B_OSA=25e9;

  cout << "OSNR [dB] = " << ChiSqRx->GetOSNR_dB(B_OSA)
       << endl;
  cout << "BER = " << ChiSqRx->GetBitErrorRatio()
       << endl;
  cout << "Decision Threshold [mV] = " << ChiSqRx->GetDecisionThresholdmV()
       << endl;

  string BERFileName = Job + "BER.dat";

  ofstream BERFile(BERFileName.c_str(),ios::out);

  if (!BERFile)
   {
    cerr << "Error: Output file "<< BERFileName
         << " could not be opened."<< endl;
    exit(1);
   }


  BERFile << "%(1) OSNR [dB] (2) BER (3) Decision Threshold [mV]" << endl
	  << ChiSqRx->GetOSNR_dB(B_OSA) << "\t"
          << ChiSqRx->GetBitErrorRatio()<< "\t"
          << ChiSqRx->GetDecisionThresholdmV();

  BERFile.close();

  ChiSqRx->~ChiSquareReceiver();
  NoiseCov->~NoiseCovariance();
 
    // ################ End of Receiver ###################

  EndTime = time(NULL);
  double TotalTime = double(EndTime - StartTime);

  times(&my_tms);

  double CPUtime = (double(my_tms.tms_utime))/(double(sysconf(_SC_CLK_TCK)));

  cout << "Total Run Time [hrs] = " << TotalTime/3600.0 << endl
       << "Total CPU Time [hrs] = " << CPUtime/3600.0 << endl
       << "Total CPU Time [secs] = " << CPUtime << " = " 
       << 100.0*CPUtime/TotalTime << "%" << endl
       << "End Time: " << ctime(&EndTime) << endl;

   cout << "THE END" << endl;
   *LogFile << "THE END" << endl;


} // ## end TycoNoiseCovariance

// #############################################

void OlegNoiseCovariance(StartupZweck * Startup)
{

  time_t StartTime,EndTime;
  struct tms my_tms;

  StartTime = time(NULL);
  cout << "Start Time: " << ctime(&StartTime) << endl;

  LogFileSeparator();

  *LogFile <<  endl << "TycoNoiseFreee" << endl;

  string InDir = Startup->GetInputFilesDir();
  string Job = Startup->GetJobPrefix();

  RanNumGen RNGSignal(InDir + "RanNumGenSig.in");
  RanNumGen RNGAmps(InDir + "RanNumGenAmps.in");
  RanNumGen RNGFibers(InDir + "RanNumGenFibers.in");

  // ## Initiating the optical pulses

  OptSignal * Signal;

  Signal = new OptSignal("Signal.in",&RNGSignal);

 
  OptDataWriter DataWriter(InDir + "DataWriter.in",InDir,Job,Signal);

  DataWriter.WriteInitialData();

  OptFiberLocalError * PreDCF;
  OptFiberLocalError * PostDCF;
  OptFiberLocalError * Dplus;
  OptFiberLocalError * Dminus;

  PreDCF = new OptFiberLocalError(InDir + "PreDCF.in",Signal,&RNGFibers,Job);
  PostDCF = new OptFiberLocalError(InDir + "PostDCF.in",Signal,&RNGFibers,Job);
  Dplus = new OptFiberLocalError(InDir + "Dplus.in",Signal,&RNGFibers,Job);
  Dminus = new OptFiberLocalError(InDir + "Dminus.in",Signal,&RNGFibers,Job);

  int NumFibers = 4;

  OptFiberLocalError * FiberPtr[NumFibers];

  FiberPtr[0] = PreDCF;
  FiberPtr[1] = PostDCF;
  FiberPtr[2] = Dplus;
  FiberPtr[3] = Dminus;

  OptAmplifier * Amp, *PreAmp, *PostAmp;

  Amp =  new OptAmplifier(InDir + "Amp.in",Signal,&RNGAmps);
  PreAmp =  new OptAmplifier(InDir + "PreAmp.in",Signal,&RNGAmps);
  PostAmp =  new OptAmplifier(InDir + "PostAmp.in",Signal,&RNGAmps);

  int NumAmps = 3;

  OptAmplifier * AmpPtr[NumAmps];

  AmpPtr[0] =  Amp;
  AmpPtr[1] =  PreAmp;
  AmpPtr[2] =  PostAmp;

 if(Startup->GetDebugLevelAmps())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetDebugLevel(Startup->GetDebugLevelAmps());
     }
   
  for(int i=0;i<NumAmps;i++)
    AmpPtr[i]->SetTypeAmplifier(typeAmplifier(Startup->GetTypeAmp()));
     
   if(Startup->GetAmpsNoiseOffFlag())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetTypeAmplifierNoise(NOISE_OFF);
     }

   if(Startup->GetBeta3ZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetSndOrDispFiber(0);
     }

   if(Startup->GetNonLinearityZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetGammaFiber(0);
     }

   if(Startup->GetFiberLossZeroFlag()) 
        for(int ii = 0; ii< NumFibers;ii++)
            FiberPtr[ii]->SetAttenuationdB(0);

   for(int ii = 0; ii< NumFibers;ii++)
         FiberPtr[ii]->SetRelativeErrorGoal(Startup->GetRelativeErrorGoal());

   for(int ii = 0; ii< NumFibers;ii++)
	{
	  FiberPtr[ii]->SetTypeStepSizes( Startup->GetTypeStepSizes());

          if(Startup->GetTypeStepSizes() == 3)
	    FiberPtr[ii]->SetWalkOffParameter(Startup->GetWalkOffParameter());

           FiberPtr[ii]->SetOutputStepSizesFlag(
                            Startup->GetOutputStepSizesFlag());
	}


  NoiseCovariance * NoiseCov;

  NoiseCov = new NoiseCovariance(InDir + "NoiseCovariance.in",Job,Signal);

  ChiSquareReceiver *  ChiSqRx;
 
  ChiSqRx = new ChiSquareReceiver("ChiSquareReceiver.in",NoiseCov);

 // #### Propagate signal through fiber ####

    cout << "Input Av Power [mW] = " 
	 << 1e3*Signal->GetAveragePower() << endl;
 
    cout << "Input Av Power [dBm] = " 
	 << Linear2dBm(Signal->GetAveragePower()) << endl;

  PreDCF->SetAttenuationdB(0);
  PreDCF->SetGammaFiber(0);

  NoiseCov->Propagate(PreDCF);
  //NoiseCov->Amplify(PreAmp);

  DataWriter.WriteWithinMap(0,0,NoiseCov->GetPropagatedLength());

  int kk = 0;

  for(kk = 0; kk < Startup->GetQtMapSpans(); kk++) 
   {
     NoiseCov->Propagate(Dplus);
     DataWriter.WriteWithinMap(0,kk,NoiseCov->GetPropagatedLength());
     NoiseCov->Propagate(Dminus);
     DataWriter.WriteWithinMap(0,kk,NoiseCov->GetPropagatedLength());
     NoiseCov->Amplify(Amp);
     DataWriter.WriteEndMap(0,kk,NoiseCov->GetPropagatedLength());
      

      if(kk % 10 == 0 && kk != 0)
	cout << "Completed " << kk << " periods of dispersion map " << endl;

    }// ## end for kk-loop


   PostDCF->SetAttenuationdB(0);
   PostDCF->SetGammaFiber(0);

   NoiseCov->Propagate(PostDCF);
   //NoiseCov->Amplify(PostAmp);

   DataWriter.WriteEndTransmission(0,0,NoiseCov->GetPropagatedLength());

   string Suffix = "";

   NoiseCov->WriteFileNoiseFreeSignalAndCovarianceMatrix(Suffix);
 
   cout << "Output Av Power [dBm] = " 
	<< Linear2dBm(Signal->GetAveragePower()) << endl;


    // ############### End of Propagation #################

 ChiSqRx->ReceiveSignal(Suffix);
 ChiSqRx->ComputeGaussianPDFsAndBER();

 double B_OSA=25e9;

  cout << "OSNR [dB] = " << ChiSqRx->GetOSNR_dB(B_OSA)
       << endl;
  cout << "BER = " << ChiSqRx->GetBitErrorRatio()
       << endl;
  cout << "Decision Threshold [mV] = " << ChiSqRx->GetDecisionThresholdmV()
       << endl;

  string BERFileName = Job + "BER.dat";

  ofstream BERFile(BERFileName.c_str(),ios::out);

  if (!BERFile)
   {
    cerr << "Error: Output file "<< BERFileName
         << " could not be opened."<< endl;
    exit(1);
   }


  BERFile << "%(1) OSNR [dB] (2) BER (3) Decision Threshold [mV]" << endl
	  << ChiSqRx->GetOSNR_dB(B_OSA) << "\t"
          << ChiSqRx->GetBitErrorRatio()<< "\t"
          << ChiSqRx->GetDecisionThresholdmV();

  BERFile.close();

  // ChiSqRx->~ChiSquareReceiver();
  // NoiseCov->~NoiseCovariance();
 
    // ################ End of Receiver ###################

  EndTime = time(NULL);
  double TotalTime = double(EndTime - StartTime);

  times(&my_tms);

  double CPUtime = (double(my_tms.tms_utime))/(double(sysconf(_SC_CLK_TCK)));

  cout << "Total Run Time [hrs] = " << TotalTime/3600.0 << endl
       << "Total CPU Time [hrs] = " << CPUtime/3600.0 << endl
       << "Total CPU Time [secs] = " << CPUtime << " = " 
       << 100.0*CPUtime/TotalTime << "%" << endl
       << "End Time: " << ctime(&EndTime) << endl;

   cout << "THE END" << endl;
   *LogFile << "THE END" << endl;

} // ## end OlegNoiseCovariance

// ###########################################

void MarcoNoiseCovariance(StartupZweck * Startup)
{

  time_t StartTime,EndTime;
  struct tms my_tms;

  StartTime = time(NULL);
  cout << "Start Time: " << ctime(&StartTime) << endl;

  LogFileSeparator();

  *LogFile <<  endl << "TycoNoiseCovarince" << endl;

  string InDir = Startup->GetInputFilesDir();
  string Job = Startup->GetJobPrefix();

  RanNumGen RNGSignal(InDir + "RanNumGenSig.in");
  RanNumGen RNGAmps(InDir + "RanNumGenAmps.in");
  RanNumGen RNGFibers(InDir + "RanNumGenFibers.in");

  // ## Initiating the optical pulses

  OptSignal * Signal;

  Signal = new OptSignal("Signal.in",&RNGSignal);

 
  OptDataWriter DataWriter(InDir + "DataWriter.in",InDir,Job,Signal);

  DataWriter.WriteInitialData();
 
   // ## Declare and initialize the fibers 

  OptFiberLocalError * Fiber;

  Fiber = new OptFiberLocalError(InDir + "Fiber.in",Signal,&RNGFibers,Job);

  Fiber->SetGammaFiber(2e-3);
   
    int NumFibers = 1;

    OptFiberLocalError * FiberPtr[NumFibers];

    FiberPtr[0] = Fiber;

    LumpedLoss * Loss;
    Loss = new LumpedLoss(InDir + "Loss.in",Signal);

    OptAmplifier * Amp;

    Amp =  new OptAmplifier(InDir + "Amp.in",Signal,&RNGAmps);
    

    int NumAmps = 1;

    OptAmplifier * AmpPtr[NumAmps];

    AmpPtr[0] =  Amp;
    
  if(Startup->GetDebugLevelAmps())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetDebugLevel(Startup->GetDebugLevelAmps());
     }
   
  for(int i=0;i<NumAmps;i++)
    AmpPtr[i]->SetTypeAmplifier(typeAmplifier(Startup->GetTypeAmp()));
     
   if(Startup->GetAmpsNoiseOffFlag())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetTypeAmplifierNoise(NOISE_OFF);
     }

   if(Startup->GetBeta3ZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetSndOrDispFiber(0);
     }

   if(Startup->GetNonLinearityZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetGammaFiber(0);
     }

   if(Startup->GetFiberLossZeroFlag()) 
        for(int ii = 0; ii< NumFibers;ii++)
            FiberPtr[ii]->SetAttenuationdB(0);

    for(int ii = 0; ii< NumFibers;ii++)
         FiberPtr[ii]->SetRelativeErrorGoal(Startup->GetRelativeErrorGoal());

      for(int ii = 0; ii< NumFibers;ii++)
	{
	  FiberPtr[ii]->SetTypeStepSizes( Startup->GetTypeStepSizes());

          if(Startup->GetTypeStepSizes() == 3)
	    FiberPtr[ii]->SetWalkOffParameter(Startup->GetWalkOffParameter());

           FiberPtr[ii]->SetOutputStepSizesFlag(
                            Startup->GetOutputStepSizesFlag());
	}


  NoiseCovariance * NoiseCov;

  NoiseCov = new NoiseCovariance(InDir + "NoiseCovariance.in",Job,Signal);

  ChiSquareReceiver *  ChiSqRx;
 
  ChiSqRx = new ChiSquareReceiver("ChiSquareReceiver.in",NoiseCov);

  // #### Propagate signal through fiber ####

    cout << "Input Av Power [mW] = " 
	 << 1e3*Signal->GetAveragePower() << endl;
 
    cout << "Input Av Power [dBm] = " 
	 << Linear2dBm(Signal->GetAveragePower()) << endl;

    NoiseCov->Attenuate(Loss);
    NoiseCov->Amplify(Amp);

    cout << "N_ASE = " << 
    Signal->GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics() << endl;

    //NoiseCov->InitializeCovarianceMatrix(8e-15);
    NoiseCov->Propagate(Fiber);

    // If this line is uncommented we compute CovarianceMatrix using the
   //  the analytic formula I derived based on Carena's paper.

    NoiseCov->PropagateAnalyticCWSignalNoLossNoDispersion(Fiber,Amp);

    DataWriter.WriteEndTransmission(0,0,NoiseCov->GetPropagatedLength());

    string Suffix = "";

    NoiseCov->WriteFileNoiseFreeSignalAndCovarianceMatrix(Suffix);
 
    cout << "Output Av Power [dBm] = " 
	 << Linear2dBm(Signal->GetAveragePower()) << endl;

    if(Startup->GetWriteComplexFieldsFlag())
     Signal->WriteFileComplexFieldsTimeAndFreq(Job + "tfComplexFields.dat");



    // ############### End of Propagation #################

  ChiSqRx->ReceiveSignal(Suffix);
 
  double NF = ChiSqRx->GetNormalizationFactor();
 
  double B_OSA=12.478e9; // 0.1 nm

  cout << "OSNR [dB] = " << ChiSqRx->GetOSNR_dB(B_OSA)
       << endl;
  cout << "BER = " << ChiSqRx->GetBitErrorRatio()
       << endl;
  cout << "Decision Threshold [mV] = " << ChiSqRx->GetDecisionThresholdmV()
       << endl;

  string BERFileName = Job + "BER.dat";

  ofstream BERFile(BERFileName.c_str(),ios::out);

  if (!BERFile)
   {
    cerr << "Error: Output file "<< BERFileName
         << " could not be opened."<< endl;
    exit(1);
   }


  BERFile << "%(1) OSNR [dB] (2) BER (3) Decision Threshold [mV]" << endl
	  << ChiSqRx->GetOSNR_dB(B_OSA) << "\t"
          << ChiSqRx->GetBitErrorRatio()<< "\t"
          << ChiSqRx->GetDecisionThresholdmV();

  BERFile.close();

  // As a double check also compute characteristic function of pdf
  // analytically 

  Suffix = "AnalChar";

 

  ChiSqRx->ReceiveSignalCWNoLossNoDisp(Suffix,
                        Signal->GetOptPowerChann(),Fiber->GetFiberLength(), 
                        Fiber->GetGamma(), 
      Signal->GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics(),NF);

  ChiSqRx->~ChiSquareReceiver();
  NoiseCov->~NoiseCovariance();
 
    // ################ End of Receiver ###################

  EndTime = time(NULL);
  double TotalTime = double(EndTime - StartTime);

  times(&my_tms);

  double CPUtime = (double(my_tms.tms_utime))/(double(sysconf(_SC_CLK_TCK)));

  cout << "Total Run Time [hrs] = " << TotalTime/3600.0 << endl
       << "Total CPU Time [hrs] = " << CPUtime/3600.0 << endl
       << "Total CPU Time [secs] = " << CPUtime << " = " 
       << 100.0*CPUtime/TotalTime << "%" << endl
       << "End Time: " << ctime(&EndTime) << endl;

   cout << "THE END" << endl;
   *LogFile << "THE END" << endl;

} // ## end MarcoNoiseCovariance

// #############################################


void TestNoiseCovariance(StartupZweck * Startup)
{

  time_t StartTime,EndTime;
  struct tms my_tms;

  StartTime = time(NULL);
  cout << "Start Time: " << ctime(&StartTime) << endl;

  LogFileSeparator();

  *LogFile <<  endl << "TycoNoiseFreee" << endl;

  string InDir = Startup->GetInputFilesDir();
  string Job = Startup->GetJobPrefix();

  RanNumGen RNGSignal(InDir + "RanNumGenSig.in");
  RanNumGen RNGAmps(InDir + "RanNumGenAmps.in");
  RanNumGen RNGFibers(InDir + "RanNumGenFibers.in");

   // ## Initiating the optical pulses

  OptSignal * Signal;
  OptFiberLocalError * Fiber1, *Fiber2; 
  OptAmplifier * Amp;

  Signal = new OptSignal("Signal.in",&RNGSignal);
  Fiber1 = new OptFiberLocalError(InDir + "Fiber1.in",Signal,&RNGFibers,Job);
  Fiber2 = new OptFiberLocalError(InDir + "Fiber2.in",Signal,&RNGFibers,Job);

  Amp = new OptAmplifier(InDir + "Amp.in",Signal,&RNGAmps);

  OptDataWriter DataWriter(InDir + "DataWriter.in",InDir,Job,Signal);

  DataWriter.WriteInitialData();

  NoiseCovariance * NoiseCov;

  NoiseCov = new NoiseCovariance(InDir + "NoiseCovariance.in",Job,Signal);

  for(int Map = 1; Map <= Startup->GetQtMapSpans();Map++)
    {

     NoiseCov->Propagate(Fiber1);
     NoiseCov->Propagate(Fiber2);

     NoiseCov->Amplify(Amp);

     cout << "Map = " << Map << endl;

    } 

  string Suffix = "";

  NoiseCov->WriteFileNoiseFreeSignalAndCovarianceMatrix(Suffix);

  /*****

  // To check if NF signal is correct (it is)

  double PropagatedLength = 0.0;
  int ZStepNum = 0;

  Signal->RegenerateSignalString();
  Fiber1->Propagate(&PropagatedLength,&ZStepNum);
  Fiber2->Propagate(&PropagatedLength,&ZStepNum);
  Amp->SetTypeAmplifierNoise(NOISE_OFF);
  Amp->AmplifyOptSignal();

  DataWriter.WriteEndTransmission(0,0,0);
  *********/

  cout << "THE END" << endl;
  *LogFile << "THE END" << endl;


} // ## end TestNoiseCovariance


// #####################################################

void TestChiSqRxArbPol(StartupZweck * Startup)
{

  // Creates noise free signal with X and Y polarization states
  // for input to noise_john.m which is then used to compute
  // pdf of current by taking convolution of pdf_X and pdf_Y,
  // assuming these two pdfs are independent.

  LogFileSeparator();

  *LogFile <<  endl << "TestChiSqRxArbPol" << endl;

  string InDir = Startup->GetInputFilesDir();
  string Job = Startup->GetJobPrefix();

  RanNumGen RNGSignal(InDir + "RanNumGenSig.in");
 
   // ## Initiating the optical pulses

  OptSignal * Signal;

  Signal = new OptSignal("Signal.in",&RNGSignal);

  // Add OptFiber object here!!!!

   string FileName = Job + "X_fNoiseFreeSignal.dat";
   ofstream File(FileName.c_str(),ios::out);  

   if (!File) 
    {
      cerr << "Error: " << FileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   File 
     << "% (1) Freq [GHz] (2) Power [W] (3) Real Part (4) Imag Part"
     << endl;

   //  If you use fftshift here then DC term is at first index of output
   //  rather than being in the center.

   int CovMatrixFirstIndex = 0;
   int CovMatrixLastIndex = Signal->GetQtPoints()-1;


   Signal->fft->fftshift(&(Signal->sfftM.fOptSignalX[CovMatrixFirstIndex]));

   for(int jj=CovMatrixFirstIndex;jj<=CovMatrixLastIndex;jj++)  
     {
       
       File 
	 << 1e-9*jj*Signal->fft->delta_f << "\t"
	 << sq(abs(Signal->sfftM.fOptSignalX[jj])) << "\t"
	 << real(Signal->sfftM.fOptSignalX[jj]) << "\t"
	 << imag(Signal->sfftM.fOptSignalX[jj]) << endl;
     }

   Signal->fft->ifftshift(&(Signal->sfftM.fOptSignalX[CovMatrixFirstIndex]));

   File.close();

   FileName = Job + "Y_fNoiseFreeSignal.dat";
   File.open(FileName.c_str(),ios::out);  

   if (!File) 
    {
      cerr << "Error: " << FileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   File 
     << "% (1) Freq [GHz] (2) Power [W] (3) Real Part (4) Imag Part"
     << endl;

   //  If you use fftshift here then DC term is at first index of output
   //  rather than being in the center.


   Signal->fft->fftshift(&(Signal->sfftM.fOptSignalY[CovMatrixFirstIndex]));


   for(int jj=CovMatrixFirstIndex;jj<=CovMatrixLastIndex;jj++)  
     {
       
       File 
	 << 1e-9*jj*Signal->fft->delta_f << "\t"
	 << sq(abs(Signal->sfftM.fOptSignalY[jj])) << "\t"
	 << real(Signal->sfftM.fOptSignalY[jj]) << "\t"
	 << imag(Signal->sfftM.fOptSignalY[jj]) << endl;
     }

   Signal->fft->ifftshift(&(Signal->sfftM.fOptSignalY[CovMatrixFirstIndex]));
 
   File.close();

} // ## end TestChiSqRxArbPol

// ####################################################################

void TestNoiseCovarianceAssignmentOperator(StartupZweck * Startup)
{


  LogFileSeparator();
 
  *LogFile <<  "TestNoiseCovarianceiAssignmentOperator" << endl;
 
  string InDir = Startup->GetInputFilesDir();
  string Job = Startup->GetJobPrefix();
 
  RanNumGen RNGSignal(InDir + "RanNumGenSig.in");
  RanNumGen RNGAmps(InDir + "RanNumGenAmps.in");
  RanNumGen RNGFibers(InDir + "RanNumGenFibers.in");
 
   // ## Initiating the optical pulses
 
  OptSignal * Signal, *CopyOfSignal;
  OptFiberLocalError * Fiber1, *Fiber2,* CopyOfFiber1, *CopyOfFiber2;
  OptAmplifier * Amp, *CopyOfAmp;
 
  Signal = new OptSignal("Signal.in",&RNGSignal);
  CopyOfSignal = new OptSignal("Signal.in",&RNGSignal);

  Fiber1 = new OptFiberLocalError(InDir + "Fiber1.in",Signal,&RNGFibers,Job);
  Fiber2 = new OptFiberLocalError(InDir + "Fiber2.in",Signal,&RNGFibers,Job);
 
  Amp = new OptAmplifier(InDir + "Amp.in",Signal,&RNGAmps);
 
  CopyOfFiber1 = new OptFiberLocalError(InDir + "Fiber1.in",CopyOfSignal,&RNGFibers,Job);
  CopyOfFiber2 = new OptFiberLocalError(InDir + "Fiber2.in",CopyOfSignal,&RNGFibers,Job);

  CopyOfAmp = new OptAmplifier(InDir + "Amp.in",CopyOfSignal,&RNGAmps);


  OptDataWriter DataWriter(InDir + "DataWriter.in",InDir,Job,Signal);

  DataWriter.WriteInitialData();

  NoiseCovariance * NoiseCov;

  NoiseCov = new NoiseCovariance(InDir + "NoiseCovariance.in",Job,Signal);

  NoiseCovariance * CopyOfNoiseCov;

  CopyOfNoiseCov = new NoiseCovariance(InDir + "NoiseCovariance.in",Job,CopyOfSignal);


  int NumOrigMaps = 2;  // Startup->GetQtMapSpans();

  NumOrigMaps = Startup->GetQtMapSpans();

  for(int Map = 1; Map <= NumOrigMaps;Map++)
    {

     NoiseCov->Propagate(Fiber1);
     NoiseCov->Propagate(Fiber2);

     NoiseCov->Amplify(Amp);

     cout << "Map = " << Map << endl;

    }

   string Suffix = "Orig";

   NoiseCov->WriteFileNoiseFreeSignalAndCovarianceMatrix(Suffix);

  *CopyOfNoiseCov = *NoiseCov;

  for(int Map = 1; Map <= Startup->GetQtMapSpans()-NumOrigMaps;Map++) 
    { 
 
     CopyOfNoiseCov->Propagate(CopyOfFiber1);
     CopyOfNoiseCov->Propagate(CopyOfFiber2);

     CopyOfNoiseCov->Amplify(CopyOfAmp);

     cout << "Map = " << Map << endl;

    }
  


  Suffix = "Copy";

  CopyOfNoiseCov->WriteFileNoiseFreeSignalAndCovarianceMatrix(Suffix);
  

  /*****

  // To check if NF signal is correct (it is)

  double PropagatedLength = 0.0;
  int ZStepNum = 0;

  Signal->RegenerateSignalString();
  Fiber1->Propagate(&PropagatedLength,&ZStepNum);
  Fiber2->Propagate(&PropagatedLength,&ZStepNum);
  Amp->SetTypeAmplifierNoise(NOISE_OFF);
  Amp->AmplifyOptSignal();

  DataWriter.WriteEndTransmission(0,0,0);
  *********/

  cout << "THE END" << endl;
  *LogFile << "THE END" << endl;


} // ## end TestNoiseCovariance


//#################################################

void TestChiSquareReceiver(StartupZweck * Startup)
{

  time_t StartTime,EndTime;
  struct tms my_tms;

  StartTime = time(NULL);
  cout << "Start Time: " << ctime(&StartTime) << endl;

  LogFileSeparator();

  *LogFile <<  endl << "TycoNoiseFreee" << endl;

  string InDir = Startup->GetInputFilesDir();
  string Job = Startup->GetJobPrefix();

  RanNumGen RNGSignal(InDir + "RanNumGenSig.in");
  RanNumGen RNGAmps(InDir + "RanNumGenAmps.in");
  
   // ## Initiating the optical pulses

  OptSignal * Signal;
  OptAmplifier * Amp;
  LumpedLoss * Loss;

  Signal = new OptSignal("Signal.in",&RNGSignal);
  
  Amp = new OptAmplifier(InDir + "Amp.in",Signal,&RNGAmps);
  Loss = new LumpedLoss(InDir + "Loss.in",Signal);

  OptDataWriter DataWriter(InDir + "DataWriter.in",InDir,Job,Signal);

  DataWriter.WriteInitialData();

  NoiseCovariance * NoiseCov;

  NoiseCov = new NoiseCovariance(InDir + "NoiseCovariance.in",Job,Signal);

  int NumAmpsPerMap = 4; // Tyco system

  for(int jj=0;jj<NumAmpsPerMap*Startup->GetQtMapSpans();jj++)
  {
    NoiseCov->Attenuate(Loss);
    NoiseCov->Amplify(Amp);
  }

  string Suffix = "";

  //NoiseCov->WriteFileNoiseFreeSignalAndCovarianceMatrix(Suffix);

  ChiSquareReceiver *  ChiSqRx;
 
  ChiSqRx = new ChiSquareReceiver("ChiSquareReceiver.in",NoiseCov);

  ChiSqRx->ReceiveSignal(Suffix);

  ChiSqRx->ComputeGaussianPDFsAndBER();

  //Receiver.ReceiveSignal(34);

  double B_OSA=25e9;

  cout << "OSNR [dB] = " << ChiSqRx->GetOSNR_dB(B_OSA)
       << endl;
  cout << "BER = " << ChiSqRx->GetBitErrorRatio()
       << endl;
  cout << "Decision Threshold [mV] = " << ChiSqRx->GetDecisionThresholdmV()
       << endl;

  string BERFileName = Job + "BER.dat";

  ofstream BERFile(BERFileName.c_str(),ios::out);

  if (!BERFile)
   {
    cerr << "Error: Output file "<< BERFileName
         << " could not be opened."<< endl;
    exit(1);
   }


  BERFile << "%(1) OSNR [dB] (2) BER (3) Decision Threshold [mV]" << endl
          << ChiSqRx->GetOSNR_dB(B_OSA) << "\t"
          << ChiSqRx->GetBitErrorRatio()<< "\t"
          << ChiSqRx->GetDecisionThresholdmV();

  BERFile.close();


/*
  for(int jj=1;jj<3;jj++)
    {
      ChiSqRx->SetOptFilterBandwidth(jj*40e9);
      ChiSqRx->SetElecFilterBandwidth(jj*4e9);
      ChiSqRx->SetOSNR(15);


      ChiSqRx->ReceiveSignal(Suffix);
  
  cout << "BER = " << ChiSqRx->GetBitErrorRatio() << endl;
  cout << "Decision Threshold [mV] = " << ChiSqRx->GetDecisionThresholdmV()
       << endl;

    }
*/

  ChiSqRx->~ChiSquareReceiver();
  NoiseCov->~NoiseCovariance();
 
  cout << "THE END" << endl;
  *LogFile << "THE END" << endl;


} // ## end TestChiSquareReceiver








