// $Id: ReducedModel.cc,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

// #####################################################################
//
//    ReducedModel.cc
//
//    Simulation Module of the ocs system
//    Optical Communication Systems Simulator
//
//    Copyright:
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
// #####################################################################


#include "ReducedModel.hh"

extern ofstream *LogFile;

//## Class of applications using the reduced model


//## Sweep the Reduced Model by the PDL

//! This function is used to compute Delta Q as a function of the PDL
/*! The function is called from void MyApplication(int argc, char *argv[])
    in ocsAppLima.cc, and is given by SimType = 32 in StartupLima.in. \n
    The function constructs a new ReducedModel object and then calls
    ReducedModel::SweepReduPDL()
*/


void SweepReduPDL(MyStartup *oMyStartup)
{
   cout << endl << "-> Starting SweepReduPDL()" << endl;
   ReducedModel *oReducedModel = new ReducedModel(oMyStartup);
   double MinPDL = oMyStartup->GetDoubleParam(3);
   double MaxPDL = oMyStartup->GetDoubleParam(4);    
   double PowerPenaltyMargin = oMyStartup->GetDoubleParam(5);        
   cout << "PowerPenaltyMargin = " << PowerPenaltyMargin << endl;
   int qtSweepPDL = oMyStartup->GetIntParam(3);     
   oReducedModel->SweepReduPDL(MinPDL,MaxPDL,qtSweepPDL,PowerPenaltyMargin);
   delete oReducedModel;
}


//## Get the distribution of the relative Q factor for 
//#  the reduced model

//! Computes the  Delta Q distribution for a given value
//! of the PDL using Monte Carlo experiments with importance sampling
//! or simply using standard importance sampling. 

/*!  The function is called from void MyApplication(int argc, char *argv[])
    in ocsAppLima.cc, and is given by SimType = 31 in StartupLima.in. \n
    The function constructs a new ReducedModel object and then calls
    ReducedModel::GetReduRelativeQfactor_pdf()

   When the method is used for importance sampling the method computes
   results for one value of the biasing parameter. The function 
   MergeSelectMaxRelNumHits_pdfs in ReduExampleApplications.cc
   is used to merge the distrubutions for the different bias values into
   a single distribution from which outage probabilities can then be 
   computed.

*/

void GetReduRelativeQfactor_pdf(MyStartup *oMyStartup)
{
   cout << endl << "-> Starting GetReduRelativeQfactor_pdf()" << endl;
   ReducedModel *oReducedModel = new ReducedModel(oMyStartup);    
   oReducedModel->GetReduRelativeQfactor_pdf();
   delete oReducedModel;
}


//## Get the distribution of the Q factor for the reduced model
//#  in order to compare with Yu Sun's results.

//! See ReducedModel::GetReduQfactor_pdf()

void GetReduQfactor_pdf(MyStartup *oMyStartup)
{
   cout << endl << "-> Starting GetReduQfactor_pdf()" << endl;
   ReducedModel *oReducedModel = new ReducedModel(oMyStartup);    
   oReducedModel->GetReduQfactor_pdf();
   delete oReducedModel;
}


//! Used to study polarization effects in periodic loops

/*!
  Since you are unlikely to use this method I am not documenting it much.
 
  Get the distribution of the Q factor for the reduced model
  in order to compare with Yu Sun's results.
  This function will compute three distributions for 
  the Q-factor. Two of this distributions will be 
  conditioned to the level of periodicity of the system.
*/

void GetReduQfactorWithPeriodicity_pdf(MyStartup *oMyStartup)
{
   cout << endl << "-> Starting GetReduQfactor_pdf()" << endl;
   ReducedModel *oReducedModel = new ReducedModel(oMyStartup);    
   oReducedModel->GetReduQfactorWithPeriodicity_pdf();
   delete oReducedModel;
}


//## Get the distribution of the Q factor for the reduced model

//! Computes outage probabilities as a function of the number of channels

/*! The function is called from void MyApplication(int argc, char *argv[])
    in ocsAppLima.cc, and is given by SimType = 33 in StartupLima.in. \n
    The function constructs a new ReducedModel object and then calls
    ReducedModel::GetReduOutageProb()

    NOTE: This method does not use importance sampling, but rather uses
    Gaussian extrapolation as in [Wang and Menyuk, JLT 2001].

*/

void GetReduOutageProb(MyStartup *oMyStartup)
{
   cout << endl << "-> Starting GetReduOutageProb()" << endl;

   ReducedModel *oReducedModel = new ReducedModel(oMyStartup);    

   int MinNumChannels  = oMyStartup->GetIntParam(3);
   int MaxNumChannels  = oMyStartup->GetIntParam(4);
   int StepNumChannels = oMyStartup->GetIntParam(5);  

   double PowerPenaltyMargin = oMyStartup->GetDoubleParam(5);     

   cout << "PowerPenaltyMargin = " << PowerPenaltyMargin << endl;

   if (StepNumChannels == 0) 
    {
      StepNumChannels = 2;      
    }
      
   oReducedModel->GetReduOutageProb(PowerPenaltyMargin,
                                    MinNumChannels, 
                                    MaxNumChannels, 
                                    StepNumChannels);

   delete oReducedModel;
}


// #############################################################

//## Get the evolution of the noise without signal
//#  It was written exclusively for the SAIC report

//! For debugging only

void GetSignalFreeNoiseEvolution(MyStartup *oMyStartup) 
{          
   cout << endl << "-> Starting GetSignalFreeNoiseEvolution()" << endl;
   ReducedModel *oReducedModel = new ReducedModel(oMyStartup);            
   oReducedModel->GetSignalFreeNoiseEvolution();          
   delete oReducedModel;
}


// ############################################################


//## Constructor

ReducedModel::ReducedModel(MyStartup *oMyStartup)
{
   cout << "An object of class ReducedModel has been created." 
        << endl;    
 
   LogFileSeparator();   
   *LogFile << "Constructing ReducedModel" << endl << endl;

 
//   OpticalBandwidth    = oMyStartup->GetDoubleParam(1);
//   *LogFile << "OpticalBandwidth = " << OpticalBandwidth << endl;
//   ElectricalBandwidth = oMyStartup->GetDoubleParam(2); 
//   *LogFile << "ElectricalBandwidth = " << ElectricalBandwidth << endl;
   
   PolDepAttenuation   = oMyStartup->GetDoubleParam(3);                            
   *LogFile << "PolDepAttenuation = " << PolDepAttenuation << endl;
   qtLoops             = oMyStartup->GetIntParam(1);   
      
   *LogFile << "qtLoops = " << qtLoops  << endl;      
   IsStraightLine      = oMyStartup->GetIntParam(2);      
   *LogFile << "IsStraightLine = " << IsStraightLine  << endl;      
  
   LoopPolarizControl   = oMyStartup->GetIntParam(4);
   *LogFile <<"LoopPolarizControl = "<<LoopPolarizControl<<endl; 

   ChannelArrayIndex = oMyStartup->GetIntParam(5);
   *LogFile <<"ChannelArrayIndex = "<<ChannelArrayIndex<<endl; 
   cout <<"ChannelArrayIndex = "<<ChannelArrayIndex<<endl;    

   InitialSample   = oMyStartup->GetIntParam(9);
   *LogFile <<"InitialSample = "<< InitialSample <<endl;

   qtStatSamples       = oMyStartup->GetQtStatSamples();                
                
   *LogFile << "qtStatSamples = " << qtStatSamples  << endl;   
   qtMapSpans          = oMyStartup->GetQtMapSpans();
   *LogFile << "qtMapSpans = " << qtMapSpans  << endl;      
   
   //BandwidthOSA        = oMyStartup->GetDoubleParam(6);
   //*LogFile << "BandwidthOSA = " << BandwidthOSA  << endl;
   PolarizNoiseBandwidth = oMyStartup->GetDoubleParam(7);
   *LogFile << "PolarizNoiseBandwidth = " 
           <<  PolarizNoiseBandwidth  << endl;         
   BiasingParam = oMyStartup->GetDoubleParam(8);
   *LogFile << "BiasingParam = " << BiasingParam << endl;      
   cout << endl;
   cout << "qtStatSamples = " << qtStatSamples << endl;
   cout << "qtMapSpans (fibers) = " << qtMapSpans << endl;
   if ( IsStraightLine) {
      cout << "qt Sets of Maps = " << qtLoops << endl;      
   }   
   else {
      cout << "qtLoops = " << qtLoops << endl;      
   }                  
   cout << endl;
           
   oRanPMD = new RanNumGen("RanPMD.in");   
      
//   if (BiasingParam != 0) {
//      oRanImportSampPolarizTransform 
//            = new RanNumGen("RanImportSampPolarizTransform.in");   
//   }
   

   //## Initialization of the full model objects that are used
   //#  to compute the Q-factor in the post-processing.
   oRanBitString = new RanNumGen("RanBitString.in");   
   oOptSignal = new OptSignal("OptSignal.in", oRanBitString);
   oOptSignal->ReduceToSingleChannel(oOptSignal->GetChannelArrayIndex(0));
   oOptSignal->RegenerateSignalString();
   if (    oOptSignal->GetInStokes(0,1) != 1
        || oOptSignal->GetInStokes(0,2) != 0
	|| oOptSignal->GetInStokes(0,3) != 0) {   
      LogFileSeparator();
      *LogFile << "Error: You should set the Stokes parameters" << endl
           << "of the 1st channel to be (1,0,0)" << endl;
      *LogFile << "It has to do with the computation of the Q-factor" << endl
               << "for the reduced model." << endl;	   	
      cout << "Error!!! See ocs.log for details." << endl;	   
      exit(1);  
   }   
   
   oOptFilter = new OptFilter("OptFilter.in", oOptSignal);  
   oPhotodetector = new Photodetector("Photodetector.in", oOptSignal);   
   oElecFilter = new ElecFilter("ElecFilter.in", oPhotodetector);
   oElecSignalStat = new ElecSignalStat("ElecSignalStat.in",
                                        oOptSignal,  
					oPhotodetector);   
   oElecSignalStat->SetOptElecFiltersForSemiAnalyticalStatistics(
                           oOptFilter, oElecFilter);
   //## End of full model objects for post-processing   
   

   //## The signal noise bandwidth is equal to the bandwidth of the
   //#  optical filter, but the polarized part of the noise comprizes
   //#  the whole channel spacing
   
   BandwidthOSA = oOptFilter->GetFreqFWHM_OptFilter();
   *LogFile << "BandwidthOSA is being set as the bandwidth of the optical filter"
        << endl;
   *LogFile << "BandwidthOSA = " << BandwidthOSA << endl;	
   oReduOptSignal = new ReduOptSignal("OptSignal.in", 
                                     BandwidthOSA);   

   EnhancementFactor = oElecSignalStat->GetEnhancementFactor(BandwidthOSA);
   ExtinctionRatio = oElecSignalStat->GetCurrentExtinctionRatio_SemiAnalytical();
   NormNumberNoiseModes = oElecSignalStat->GetNormNumberNoiseModes_SemiAnalytical();
   NormKcoeff1 = oElecSignalStat->GetNormKcoeff1_SemiAnalytical();
   NormKcoeff0 = oElecSignalStat->GetNormKcoeff0_SemiAnalytical();

   cout << "EnhancementFactor = " << EnhancementFactor << endl;
   cout << "ExtinctionRatio = " << ExtinctionRatio << endl;
   cout << "NormNumberNoiseModes = " << NormNumberNoiseModes << endl;
   cout << "NormKcoeff1 = " << NormKcoeff1 << endl;
   cout << "NormKcoeff0 = " << NormKcoeff0 << endl;

   if (BiasingParam != 0) 
    {
      oRanImportSampPolarizTransform 
            = new RanNumGen("RanImportSampPolarizTransform.in");   
      oReduPolarizTransformer 
            = new ReduPolarizTransformer(RX_RZ_RY, oReduOptSignal);      
      oReduPolarizTransformer->InitImportSampPolarizTransformation(
            BiasingParam,oRanImportSampPolarizTransform);            
   }



   oReduPolDepAttenuator = new ReduPolDepAttenuator(oReduOptSignal);

   oReduPolDepAttenuator->SetPolDepAttenuation(0., 
                                        PolDepAttenuation);        
   
   oReduOptAmplifier = new ReduOptAmplifier("OptAmplifier.in"
                                            ,oReduOptSignal);                
   
   //cout << "FixedOutputPower = " 
   //     <<oReduOptAmplifier->GetOutputPowerLinear()<< endl; 
                                           
   oReduOptAmplifier->SetFixedOutputPowerFromCurrentOptSignal();

   //cout << "FixedOutputPower = " 
   //     <<oReduOptAmplifier->GetOutputPowerLinear()<< endl; 
                      
   
   oHistQfactor = new Histogram("Histogram.in");
   oHistOptSNR  = new Histogram("HistOptSNR.in"); 
   oHistElecTimeDomainSNR = new Histogram("HistElecTimeDomainSNR.in");      


   oReduOptSignal->SetOptElecFilterBandwidth(
                oOptFilter->GetFreqFWHM_OptFilter(),   // ## Optical BW
                oElecFilter->Get_f_3dB_ElecFilter() );// ## Electrical BW   
   
  // oReduOptSignal->SetOptElecFilterBandwidth(
  //              OpticalBandwidth,   // ## Optical BW
  //              ElectricalBandwidth );// ## Electrical BW   


   //## Prevent from choosing excessively small BandwidthOSA   
   //if (BandwidthOSA < .01*oReduOptSignal->GetChannSpacing()) {
   //   cout << "Error: The BandwidthOSA is too small." 
   //        << endl;   
   //   exit(1);
   //}   
   //else if (BandwidthOSA > oReduOptSignal->GetChannSpacing()) {
   //   cout << "Error: The BandwidthOSA is too large." 
   //        << endl;   
   //   exit(1);   
   //}
   

   //## This feature exists to produce loops with
   //#  arbitrary fiber types

   if (qtMapSpans > 100) 
   {
      cout << "qtMapSpan = " << qtMapSpans 
           << " > 100. Check the startup file."<< endl;
      exit(1);      
   }

   for (int jj = 0; jj < qtMapSpans; jj++) 
    {
      oReduOptFiber[jj] = new ReduOptFiber("OptFiber.in"
                              ,oReduOptSignal
                              ,oRanPMD);   
   }                    
   
   //## Get one single set or random fibers

   if ( LoopPolarizControl && ! IsStraightLine) 
   {
      for (int jj = 0; jj < qtMapSpans; jj++) 
      {  
         oReduOptFiber[jj]->GetFiberSampleCompleteScatt();               
      }   
   }  
    
   oReduPolarizController 
            = new ReduPolarizTransformer(RX_RY_RX, oReduOptSignal);            
   
   //## test lines
   //oReduOptFiber[0]->GetFiberSampleCompleteScatt();             
   //cout << "DGD = "
   //<< oReduOptFiber[0]->GetDGD(oReduOptSignal->GetCenterFreq()) << endl;
                  
   //exit(1);   
   
   //## The user can now set the parameters of the 
   //#  fibers that are different     

} // ## end Constructor

// ###################################################################

//## Destructor
ReducedModel::~ReducedModel()
{
   delete oRanPMD;
   delete oReduOptSignal;
   delete oReduPolDepAttenuator;
   delete oReduOptAmplifier;
   delete oHistQfactor;
   delete oHistOptSNR;
   delete oHistElecTimeDomainSNR;
   delete oReduPolarizController;  
   if (BiasingParam != 0) {   
      delete oRanImportSampPolarizTransform;   
      delete oReduPolarizTransformer;
   }
   //## For the computation of the Q-factor
   delete oRanBitString;
   delete oOptSignal;
   delete oOptFilter;
   delete oElecFilter;
   delete oPhotodetector;
   delete oElecSignalStat;      
}



//##############################################################
//## Get the Q-factor from a semi-analytical method

//! Get the Q-factor from a semi-analytical method.

/*! This method of computing Q-factors for the reduced model
is yet to be published. It is based on the work by 
P. Winzer et al. JLT 2001.
*/
//## By: Ivan Lima (2/5/20)
double ReducedModel::GetQ_FactorSemiAnalytical(int ChannelArrayIndex)
{
      double AverageSignalPower = oReduOptSignal->GetSignal(ChannelArrayIndex); 
      //double NoiseSpectralDensity = oReduOptSignal->GetNoiseInsideChannel(
      //                                    ChannelArrayIndex)
      //				   /BandwidthOSA;	  
//                                   /oOptFilter->GetFreqFWHM_OptFilter();

      //double CoPolarizedNoiseSpectralDensity = NoiseSpectralDensity/2.;
      //double OrthogPolarizedSpectralNoiseDenzity = NoiseSpectralDensity/2.;
      double CoPolarizedNoiseSpectralDensity = 
                   oReduOptSignal->GetCoPolarizedNoiseInsideChannel(
                                   ChannelArrayIndex)/BandwidthOSA;	      
      double OrthogPolarizedNoiseSpectralDensity = 
                   oReduOptSignal->GetOrthogPolarizedNoiseInsideChannel(
                                   ChannelArrayIndex)/BandwidthOSA;

      //cout << " CoPolarizedNoiseSpectralDensity = " 
      //     <<  CoPolarizedNoiseSpectralDensity 
      //     << " OrthogPolarizedNoiseSpectralDensity = " 
      //     <<  OrthogPolarizedNoiseSpectralDensity << endl;
				   
      //## I should separate the paralel and the perpendicular components
      //#  of the noise (with respect to the signal)				   
      //## This step is unecessary unless you are using the method
      //#  from oElecSignalStat object to get the Q-factor.
      //oOptSignal->RegenerateSignalWithGivenAveragePower(AverageSignalPower);

      double OSNR = AverageSignalPower
                   /(BandwidthOSA
		     *( CoPolarizedNoiseSpectralDensity 
		       +OrthogPolarizedNoiseSpectralDensity) );
      double ESNR = EnhancementFactor*OSNR;		       
      double gammaParalel = sqrt(2.)*CoPolarizedNoiseSpectralDensity
                          /sqrt( sq(CoPolarizedNoiseSpectralDensity)
			        +sq(OrthogPolarizedNoiseSpectralDensity) );
      double gamma = ( CoPolarizedNoiseSpectralDensity
                     +OrthogPolarizedNoiseSpectralDensity)/sqrt(2.)
                   /sqrt( sq(CoPolarizedNoiseSpectralDensity)
 		         +sq(OrthogPolarizedNoiseSpectralDensity) );
      //cout << "gammaParalel = " << gammaParalel << endl;
      //cout << "gammaOrthog  = " << gammaOrthog << endl;
            
      double Q_factor =  (1. - ExtinctionRatio)*ESNR*sqrt(NormNumberNoiseModes)
                        *gamma
            /( sqrt(1. + NormKcoeff1*gamma*gammaParalel*ESNR) 
              +sqrt(1. + NormKcoeff0*gamma*gammaParalel*ExtinctionRatio*ESNR)); 
		       
//      double Q_factor = oElecSignalStat->GetQ_FactorSemiAnalyticalWorsePattern(
//                                    CoPolarizedNoiseSpectralDensity,
//                                    OrthogPolarizedNoiseSpectralDensity);      
      	
      return Q_factor;			    
}



//##############################################################
//## Get the ESNR from the OSNR using a semi-analytical method

//! Get the ESNR from the OSNR using a semi-analytical method.

/*! This method of computing the ESNR from the OSNR 
is yet to be published. 
*/
//## By: Ivan Lima (2/5/20)
double ReducedModel::GetESNR_SemiAnalytical(int ChannelArrayIndex)
{
      double AverageSignalPower = oReduOptSignal->GetSignal(ChannelArrayIndex); 
				   
      //## I should separate the paralel and the perpendicular components
      //#  of the noise (with respect to the signal)				   
      //oOptSignal->RegenerateSignalWithGivenAveragePower(AverageSignalPower);
      /*
      cout << "AverageSignalPower  = " << AverageSignalPower << endl;
      cout << "AverageSignalPower* = " << oOptSignal->GetAveragePower() << endl;      
      //oOptSignal->WriteFileTimePower("tOutOptSignal.dat");   
      //oOptSignal->WriteFileFreqPower("fOutOptSignal.dat");     
      cout << "$#$ EnhancementFactor = " 
           << oElecSignalStat->GetEnhancementFactor(
                    oOptFilter->GetFreqFWHM_OptFilter() ) << endl;
      exit(1);		    
      */
      //return oElecSignalStat->GetEnhancementFactor(
      //                          BandwidthOSA )
      //           *oReduOptSignal->GetOptSNR(ChannelArrayIndex);      			       
      return  EnhancementFactor
             *oReduOptSignal->GetOptSNR(ChannelArrayIndex);     
}



// ##############################################################

//## Transmits the signal through the optical system

//! Transmits the reduced signal through the optical transmission system.

/*! Allows for importance sampling to be used as detailed in Lima et al.
    OFC 2002 TuI7.

    Also allows user to model loop systems as in Yu Sun et al. OFC 2002 ThI4.
    But you probably won't be doing that yourself.
*/

void ReducedModel::TransmitReduOptSystem(double *PropagatedLength)
{
   LikelihoodRatio = 1.0;

   double psi,theta,phi;      

   if ( ! IsStraightLine && ! LoopPolarizControl) 
   { //## It does only once
   
     for (int jj = 0; jj < qtMapSpans; jj++) 
      {  
         oReduOptFiber[jj]->GetFiberSampleCompleteScatt();               
      }   
     
   }  
 
   if (! IsStraightLine && LoopPolarizControl) 
    {
      psi   = 2*pi*oRanPMD->GetRanNum();
      theta = acos( 2*(oRanPMD->GetRanNum() - 0.5) );
      phi   = 2*pi*oRanPMD->GetRanNum();     
   }   


   for (int kk = 0; kk < qtLoops; kk++) 
    {
      if ( IsStraightLine) 
       { //## Each loop is independent
         for (int jj = 0; jj < qtMapSpans; jj++) 
          {
            oReduOptFiber[jj]->GetFiberSampleCompleteScatt();             
            //cout << "DGD = " 
	    //<<oReduOptFiber[jj]->GetDGD(oReduOptSignal->GetCenterFreq()) 
	    //<< endl;            
         }
      }   
   
      if (! IsStraightLine && LoopPolarizControl) 
      {
         oReduPolarizController->TransformPolarization(psi,theta,phi);      
      }

      for (int jj = 0; jj < qtMapSpans; jj++) 
       {            
         //## Do not propagate for the samples previously
         //#  obtained.
         //cout << CurrentSample <<"  " << InitialSample << endl;
       
         if (CurrentSample >= InitialSample) 
         {      
            oReduOptFiber[jj]->PropagateFiberCoupledNLS
                              (PropagatedLength); 
         }
 
         //## Ding Wang has one PDL per OptAmplifier                  
         //# as it is in a real system.
         
         //## Pre-amplification. Check if you commented the pos-amplification
         if (CurrentSample >= InitialSample) {         
            //oReduOptSignal->DepolarizeReduNoise();   
            oReduOptAmplifier->AmplifyOptSignal();     
         }            
         
         if (BiasingParam != 0) 
          {
          if (oReduPolDepAttenuator->GetLinearPolDepAttenuationPowerX() != 1)  
            {
               cout << "Error in ReducedModel::TransmitReduOptSystem():" 
               << endl 
               <<"There should be no PolDepAttenuationX in the PDL element. "
               << endl;
               exit(1);     
            }

            //## Attention: The importance sampling is not coded to 
            //#  work in a loop due to the periodicity.
         
            oReduPolarizTransformer->GetImportSampPolarizTransformation(
                  ChannelArrayIndex, &LikelihoodRatio);

// Linear2dB(oReduPolDepAttenuator->GetLinearPolDepAttenuationPowerY()),
//      oReduOptAmplifier->GetPolDepGain(), &LikelihoodRatio);         
        
    if (CurrentSample >= InitialSample) 
       {                                    
         oReduPolarizTransformer->TransformPolarization();                                         
       }                                                  
	  } // ## end if BiasingParam != 0


         if (CurrentSample >= InitialSample) {             
            oReduPolDepAttenuator->PolDepAttenuate(); 
         }   

         if (BiasingParam != 0) {
            if (CurrentSample >= InitialSample) {                
               oReduPolarizTransformer->UndoLastPolarizTransformation();         
            }   
         }      
   
         //## Pos-amplification. Check if you commented the pre-amplification
         //if (CurrentSample >= InitialSample) 
         // {          
         //   //oReduOptSignal->DepolarizeReduNoise();  
         //   oReduOptAmplifier->AmplifyOptSignal();               
         //}   
      }         

   }

} //## End of ReducedModel::TransmitReduOptSystem()

// ##################################################################




//## Transmits the signal through the optical system
//#  Only one PDL per loop, as Yu Sun did in her
//#  model. It assumes that the loop is stable 
//#  during the experiments. Thus, the PDL per 
//#  loop should keep constant throughout the 
//#  experiment. 

//! You probably won't use this method.

void ReducedModel::TransmitReduOptSystem1PDL_Loop(double *PropagatedLength)
{
   LikelihoodRatio = 1.; 
   double psi, theta, phi;

//   double psi[2],theta[2],phi[2];
      
   //   psi   = 2*pi*oRanPMD->GetRanNum();
   //   theta = acos( 2*(oRanPMD->GetRanNum() - 0.5) );
   //   phi   = 2*pi*oRanPMD->GetRanNum();      
   //      oReduPolarizController->TransformPolarization(
   //            psi,theta,phi);
   
   if ( ! IsStraightLine && ! LoopPolarizControl) { 
      //## It does only once per loop     
      for (int jj = 0; jj < qtMapSpans; jj++) {  
         oReduOptFiber[jj]->GetFiberSampleCompleteScatt();               
      }     
   }   
   if (! IsStraightLine && LoopPolarizControl) {
      psi   = 2*pi*oRanPMD->GetRanNum();
      theta = acos( 2*(oRanPMD->GetRanNum() - 0.5) );
      phi   = 2*pi*oRanPMD->GetRanNum();     

//      psi[1]   = 2*pi*oRanPMD->GetRanNum();
//      theta[1] = acos( 2*(oRanPMD->GetRanNum() - 0.5) );
//      phi[1]   = 2*pi*oRanPMD->GetRanNum();     

      //oReduOptFiber[0]->GetFiberSampleCompleteScatt(); //## Just the 1st fiber      
      //oReduOptFiber[3]->GetFiberSampleCompleteScatt();
         //for (int jj = 0; jj < qtMapSpans; jj++) {
         //   oReduOptFiber[jj]->GetFiberSampleCompleteScatt();             
         //}            
   }      
   
   
   for (int kk = 0; kk < qtLoops; kk++) {
      if ( IsStraightLine) { //## Each loop is independent
         for (int jj = 0; jj < qtMapSpans; jj++) {
            oReduOptFiber[jj]->GetFiberSampleCompleteScatt();             
         }
      }                 
      if (! IsStraightLine && LoopPolarizControl) {
         //## Transformation before the PDL element
         oReduPolarizController->TransformPolarization(
               psi,theta,phi);          
//               psi[0],theta[0],phi[0]);         
      }  
      for (int jj = 0; jj < qtMapSpans; jj++) {
         ////oReduOptFiber[jj]->SetScatteringMatrix( 0, 0., 0., 0.); //######## Test line      
         oReduOptFiber[jj]->PropagateFiberCoupledNLS
                           (PropagatedLength);       
         oReduOptAmplifier->AmplifyOptSignal();        
      }         
      
      // ## Yu Sun has one PDL per Loop                  
      if (BiasingParam != 0) {
          //## Attention: The importance sampling is not coded to 
          //#  work in a loop due to the periodicity.
         oReduPolarizTransformer->GetImportSampPolarizTransformation(
               ChannelArrayIndex, &LikelihoodRatio);         
      }      
      //cout << oReduOptSignal->StokesSignal[0][0] << " ";
      //cout << oReduOptSignal->StokesSignal[1][0] << " ";      
      //cout << oReduOptSignal->StokesSignal[2][0] << " ";      
      //cout << oReduOptSignal->StokesSignal[3][0] << endl;            
      oReduPolDepAttenuator->PolDepAttenuate();       
      if (BiasingParam != 0) {
         oReduPolarizTransformer->UndoLastPolarizTransformation();         
      }              
//      if (! IsStraightLine && LoopPolarizControl) {
//         //## Transformation after the PDL element
//         oReduPolarizController->TransformPolarization(
//               psi[1],theta[1],phi[1]);         
//      }              
   }
} //## End of ReducedModel::TransmitReduOptSystem1PDL_Loop(.)

// ################################################################

//## Transmits the signal through the optical system
//#  Only one PDL per loop, as Yu Sun did in her
//#  model. It assumes that the loop is stable 
//#  during the experiments. Thus, the PDL per 
//#  loop should keep constant throughout the 
//#  experiment. 
//## This method returns a parameter that describes the
//#  level of periodicity of the loop (if it is a loop)
//#  for the channel number stored in ChannelArrayIndex.
//#  It will return a number between 0.5 and 1 if the 
//#  loop is periodic.
double ReducedModel::TransmitReduOptSystem1PDL_LoopWithMonitoring(double *PropagatedLength)
{
   int ChannelArrayIndex = 0;
   double Periodicity = 0.;
   double PolarizLast[3], PolarizThis[3];

   LikelihoodRatio = 1.; 
   double psi, theta, phi;   
//   double psi[2],theta[2],phi[2];
   
   //   psi   = 2*pi*oRanPMD->GetRanNum();
   //   theta = acos( 2*(oRanPMD->GetRanNum() - 0.5) );
   //   phi   = 2*pi*oRanPMD->GetRanNum();      
   //      oReduPolarizController->TransformPolarization(
   //            psi,theta,phi);
   
   if ( ! IsStraightLine && ! LoopPolarizControl) { 
      //## It does only once per loop     
      for (int jj = 0; jj < qtMapSpans; jj++) {  
         oReduOptFiber[jj]->GetFiberSampleCompleteScatt();               
      }     
   }   
   if (! IsStraightLine && LoopPolarizControl) {
      psi   = 2*pi*oRanPMD->GetRanNum();
      theta = acos( 2*(oRanPMD->GetRanNum() - 0.5) );
      phi   = 2*pi*oRanPMD->GetRanNum();     

//      psi[1]   = 2*pi*oRanPMD->GetRanNum();
//      theta[1] = acos( 2*(oRanPMD->GetRanNum() - 0.5) );
//      phi[1]   = 2*pi*oRanPMD->GetRanNum();     

      //oReduOptFiber[0]->GetFiberSampleCompleteScatt(); //## Just the 1st fiber      
      //oReduOptFiber[3]->GetFiberSampleCompleteScatt();
         //for (int jj = 0; jj < qtMapSpans; jj++) {
         //   oReduOptFiber[jj]->GetFiberSampleCompleteScatt();             
         //}            
   }      
   
   //for (int ii = 0; ii < 3; ii++) {
   //   PolarizLast[ii] = oReduOptSignal->StokesSignal[ii+1][ChannelArrayIndex];
   //}
   
   for (int kk = 0; kk < qtLoops; kk++) {
      for (int ii = 0; ii < 3; ii++) {
         PolarizThis[ii] = oReduOptSignal->StokesSignal[ii+1][ChannelArrayIndex];
      }   
      if (kk > 0) {
         Periodicity += ScalarProduct(PolarizThis,PolarizLast)
                        /( GetVectorLength(PolarizThis,3)
                          *GetVectorLength(PolarizLast,3) );
      }                    
      for (int ii = 0; ii < 3; ii++) {
         PolarizLast[ii] = PolarizThis[ii];
      }                          
      
      if ( IsStraightLine) { //## Each loop is independent
         for (int jj = 0; jj < qtMapSpans; jj++) {
            oReduOptFiber[jj]->GetFiberSampleCompleteScatt();             
         }
      }                 
      if (! IsStraightLine && LoopPolarizControl) {
         //## Transformation before the PDL element
         oReduPolarizController->TransformPolarization(
               psi,theta,phi);             
//               psi[0],theta[0],phi[0]);         
      }  
      for (int jj = 0; jj < qtMapSpans; jj++) {
         ////oReduOptFiber[jj]->SetScatteringMatrix( 0, 0., 0., 0.); //######## Test line      
         oReduOptFiber[jj]->PropagateFiberCoupledNLS
                           (PropagatedLength);       

///         if (jj == 1 || qtMapSpans == 1) {      
            //## Yu Sun has one PDL per Loop,  
            //#  but the correct is one PDL per amplifier                 
            if (BiasingParam != 0) {
                //## Attention: The importance sampling is not coded to 
                //#  work in a loop due to the periodicity.
               oReduPolarizTransformer->GetImportSampPolarizTransformation(
                     ChannelArrayIndex, &LikelihoodRatio);         
            }               
            //cout << oReduOptSignal->StokesSignal[0][0] << " ";
            //cout << oReduOptSignal->StokesSignal[1][0] << " ";      
            //cout << oReduOptSignal->StokesSignal[2][0] << " ";      
            //cout << oReduOptSignal->StokesSignal[3][0] << endl;            
            oReduPolDepAttenuator->PolDepAttenuate();       
            if (BiasingParam != 0) {
               oReduPolarizTransformer->UndoLastPolarizTransformation();         
            }                     
///         }                           
         oReduOptAmplifier->AmplifyOptSignal();        
      }         
     
//      if (! IsStraightLine && LoopPolarizControl) {
//         //## Transformation after the PDL element
//         oReduPolarizController->TransformPolarization(
//               psi[1],theta[1],phi[1]);         
//      }              
   }
   return Periodicity/(qtLoops - 1); //## The last loop is not included

} //## End of ReducedModel::TransmitReduOptSystem1PDL_LoopWithMonitoring(.)



//#### Sweep the PDL to compare with Ding's results
//#    The loop is redundant, but it was done to simplify
//#    the comparison with the full model.
//## By: Ivan Lima



//! This method sweeps over a range of PDL values 
//! and computes the statistics of the Delta Q distribution
//! for each value of PDL considered in the sweep. 

/*!

The method ws used to obtain the results given in Figs. 1-22
of the August 2001 Report to SAIC [SAIC August],
 and in Figs. 1 of OFC 2002 paper TuI7 [OFC 2002].

In particular it's main use has been to perform the reduced model simulations
in our validation of  the Reduced Model
by comparison to full time domain simulations.

\param MinPDL: The minimum value of PDL in the sweep 
\param MinPDL: The minimum value of PDL in the sweep 
\param qtSweepPDL: The number of PDL values taken in the range
                   [MinPDL, MaxPDL].
\param PowerPenaltyMargin:  NOT USED in current version!!
 
The method writes out the following data files. All files have two
columns. The first column is the PDL values included in the sweep.
The second column is described below.

-# PDLvMeanQfactor.dat: The mean of the Delta Q distribution
-# PDLvStdDevQfactor.dat: The standard deviation of the Delta Q distribution
-# OptSNR_Mean.dat: The mean of the optical signal-to-noise ratio (OSNR) 
                    distribution
-# OptSNR_StdDev.dat: The standard deviation of the Delta OSNR distribution
-# ElecTimeDomainSNR_Mean.dat: The mean of the Delta electrical time 
                               domain SNR (ESNR) pdf
-# ElecTimeDomainSNR_StdDev.dat: The standard deviation of the Delta ESNR

In addition the initial Stokes parameters of the signal are written to 
ReduOptSignalTx.dat using ReduOptSignal::WriteFileOptSignal() and the
Stokes parameters for the output signal in the last Monte Carlo experiment
are written to the file ReduOptSignal.dat

The algorithm used in this method 
to compute the distributions is carefully described in [SAIC August].

Additional comments:

-# To reset the random number generator (RNG) for the fiber for each PDL value
we write the initial state of the RNG to the file RanPMD.tmp.
-# The Delta Q, Delta OSNR and Delta ESNR are obtained by computing 
   Q, OSNR anbd ESNR relative to their back-to-back values. The back-to-back
   values are obtained by transmitting through the system with PMD but without
   PDL or PDG.
-# All channels contribute to the distributions of Q, OSNR and ESNR.

   
 */


void ReducedModel::SweepReduPDL(double MinPDL, double MaxPDL,
                                int qtSweepPDL, double PowerPenaltyMargin)
  // ## SimType == 32
{

   //oReduOptSignal->SetPeakPower_AveragePower( 
  // 2./( 30e-12*oReduOptSignal->GetBitRateChann()/2*
   //                sqrt(pi/log(2)) ) );        
   //cout << "PeakPower_AveragePower = " 
   //     <<  oReduOptSignal->GetPeakPower_AveragePower() << endl;
   
   string FileName;
   
//   FileName = "OutageQfactor.dat";
//   ofstream OutageQfactor(FileName.c_str(), ios::out);
//   if (! OutageQfactor) {
//      cerr << "Error: "<<FileName<<" could not be created."<<endl;
//      exit(1);
//   }      

   FileName = "PDLvMeanQfactor.dat"; 
   ofstream PDLvMeanQfactor(FileName.c_str(), ios::out);
   if (! PDLvMeanQfactor) {
      cerr << "Error: "<< FileName<<" could not be created."<<endl;
      exit(1);
   }   
   
   FileName = "PDLvStdDevQfactor.dat";
   ofstream PDLvStdDevQfactor(FileName.c_str(), ios::out);
   if (! PDLvStdDevQfactor) {
      cerr << "Error: "<< FileName<<" could not be created."<<endl;
      exit(1);
   } 

   FileName = "OptSNR_Mean.dat";  
   ofstream OptSNR_Mean(FileName.c_str(), ios::out);
   if (! OptSNR_Mean) {
      cerr << "Error: "<< FileName<<" could not be created."<<endl;
      exit(1);
   }      

   FileName = "OptSNR_StdDev.dat";
   ofstream OptSNR_StdDev(FileName.c_str(), ios::out);
   if (! OptSNR_StdDev) {
      cerr << "Error: "<<"OptSNR_StdDev.dat"<<" could not be created."<<endl;
      exit(1);
   }         

   FileName = "ElecTimeDomainSNR_Mean.dat";
   ofstream ElecTimeDomainSNR_Mean(FileName.c_str(), ios::out);
   if (! ElecTimeDomainSNR_Mean) {
      cerr << "Error: "<< FileName <<" could not be created."<<endl;
      exit(1);
   }      

   FileName = "ElecTimeDomainSNR_StdDev.dat";
   ofstream ElecTimeDomainSNR_StdDev(FileName.c_str(), ios::out);
   if (! ElecTimeDomainSNR_StdDev) {
      cerr << "Error: "<< FileName <<" could not be created."<<endl;
      exit(1);
   }            
     
//   ofstream OutFileDOP("DegOfPolariz.dat",ios::out);
//   if (! OutFileDOP) {
//      cerr << "Error: "<<"DegOfPolariz.dat"<<"could not be created."<<endl;
//      exit(1);
//   }

   double PolDepGainOptAmplifOrig = oReduOptAmplifier->GetPolDepGain();
           
   double PropagatedLength;
   double SweepPDL;
   double StepPDL;

   if (qtSweepPDL > 1 && MinPDL != MaxPDL) 
    {
      StepPDL = (MaxPDL - MinPDL)/(qtSweepPDL-1);  
    }
   else 
    {
      StepPDL = (MaxPDL - MinPDL)/qtSweepPDL;  
    }    

   //## Save the input Stokes of the signal

   oReduOptSignal->RegenerateReduOptSignal();
   oReduOptSignal->WriteFileOptSignal("ReduOptSignalTx.dat");

   int NumChannels = oReduOptSignal->GetNumChannels();
   double *Back2BackQ_Factor = new double[NumChannels];   
   double *Back2BackOptSNR = new double[NumChannels];
   double *Back2BackElecTimeDomainSNR = new double[NumChannels];
   
   for (int jj = 0; jj < qtSweepPDL; jj++) 
    {
      SweepPDL = jj*StepPDL + MinPDL;                           

      // ## Sets the PDL 
    
      oReduPolDepAttenuator->SetPolDepAttenuation(0.,SweepPDL);                
    
      //## Forces the ensamble of fibers to be
      //#  the same for each case. But I won't do it any 
      //#  more, because it doesn't make evident the 
      //#  statistical error for small n, specially for the 
      //#  full model.      
    
      oRanPMD->ResetRanNumGen();           
      oHistQfactor->ClearHistogram();
      oHistOptSNR->ClearHistogram();      
    
      for (int ii = 0; ii <  qtStatSamples; ii++) 
        {
         PropagatedLength = 0;
         oReduOptAmplifier->SetPolDepGain(0.);         
         oReduPolDepAttenuator->SetPolDepAttenuation(0.,0.);       
         oRanPMD->WriteRanSeqParameters("RanPMD.tmp");         
         oReduOptSignal->RegenerateReduOptSignal();
         TransmitReduOptSystem(&PropagatedLength);
         	 
         for (int kk = 0; kk < NumChannels; kk++) 
          {  	      
            Back2BackQ_Factor[kk] = GetQ_FactorSemiAnalytical(kk); 	    
            //Back2BackQ_Factor[kk] = oReduOptSignal->GetQ_Factor(kk); 
            Back2BackOptSNR[kk] = oReduOptSignal->GetOptSNR(kk);            
            Back2BackElecTimeDomainSNR[kk] = GetESNR_SemiAnalytical(kk);
            //Back2BackElecTimeDomainSNR[kk] = 
            //                     oReduOptSignal->GetElecTimeDomainSNR(kk);

            if (qtStatSamples <= 20) 
             {               
               cout << kk << " B2B_OptSNR = " 
                    << Back2BackOptSNR[kk]
                    << "  B2B_ElecSNR = " 
                    << Back2BackElecTimeDomainSNR[kk]
		    << "  B2B_Q_Factor = "
		    << Back2BackQ_Factor[kk] << endl;            
            }        
         }   
 
         //## Get the case with PDL and PDG

         oReduOptAmplifier->SetPolDepGain(PolDepGainOptAmplifOrig); 
              
         oReduPolDepAttenuator->SetPolDepAttenuation(0.,SweepPDL); 
                  
         oRanPMD->ReadRanSeqParameters("RanPMD.tmp");         
         oReduOptSignal->RegenerateReduOptSignal();  
                           
         TransmitReduOptSystem(&PropagatedLength);               

         for (int kk = 0; kk < NumChannels; kk++) 
          {                     
           oHistQfactor->AddSample(Linear2dB( 
	          GetQ_FactorSemiAnalytical(kk)
                 /Back2BackQ_Factor[kk]));	   	   
           //oHistQfactor->AddSample(Linear2dB( oReduOptSignal->GetQ_Factor(kk)
           //                                     /Back2BackQ_Factor[kk]));
           oHistOptSNR->AddSample( Linear2dB( 
                            oReduOptSignal->GetOptSNR(kk) ) );
           //                  /Back2BackOptSNR[kk] )  );              

            oHistElecTimeDomainSNR->AddSample( Linear2dB(
               GetESNR_SemiAnalytical(kk) ) );
            //   /Back2BackElecTimeDomainSNR[kk]) );   
	                                       
            //oHistElecTimeDomainSNR->AddSample( Linear2dB(
            //   oReduOptSignal->GetElecTimeDomainSNR(kk)
            //   /Back2BackElecTimeDomainSNR[kk]) );                
              
            if (qtStatSamples <= 20) {                      
               cout << kk << " GetSNR_OSA = " << oReduOptSignal->GetOptSNR(kk)
                    << "  GetSNR_ElecTimeDomain = "
		    << GetESNR_SemiAnalytical(kk)
		    << "  GetQ_Factor = " << GetQ_FactorSemiAnalytical(kk)
		    << endl;
                    //<< oReduOptSignal->GetElecTimeDomainSNR(kk)            
                    //<< endl;    
            }                 
                    
         }      
         
         //## Lines for debugging purposes

         if (qtStatSamples <= 20) {                
            cout <<  "SignalChann(0) = " << oReduOptSignal->GetSignal(0)
                 << "  NoiseChann(0) = " 
                 << oReduOptSignal->GetNoiseInsideChannel(0) //*25./124 
                 << "  OptSNR = " << oReduOptSignal->GetOptSNR(0) //*1./(25./124)
                 << endl;                   
            cout << "  -> DGD 1st fiber = " 
                 << oReduOptFiber[0]->GetDGD(oReduOptSignal->GetCenterFreq()) 
                 << endl;                   
         }        
         
         //cout << SweepPDL <<" "<<ii<<" OptSNR(0) = "
         //     << oReduOptSignal->GetSNR_Chann(0) << endl;         
         //        oHistOptSNR->AddSample( Linear2dB( 1/(BandwidthOSA/
         //                                 oReduOptSignal->GetChannSpacing())
         //                          *oReduOptSignal->GetSNR_Chann(0)         
         //   /Back2BackOptSNR )  )
         //cout << ii << "  "<< oReduOptSignal.GetQ_Factor(0) << endl;

	} // ## end for qtSamples loop

          //      if (SweepPDL != 0 ) {
         //cout << "Time to get the outage prob:" << endl;
         //cout << "oHistQfactor->GetMean() = "
      //<< oHistQfactor->GetMean() << endl;
         //cout << "oHistQfactor->GetStdDev() = "
      //<< oHistQfactor->GetStdDev() << endl;
      //         OutageQfactor << SweepPDL << " "
//                       << oReduOptSignal->GetOutageProbGaussian(
//                                       oHistQfactor->GetMean(),
//                                       oHistQfactor->GetStdDev(),
//                                       -PowerPenaltyMargin) << endl;      
//      }                                   


      PDLvMeanQfactor   << SweepPDL << " " 
                        << oHistQfactor->GetMean()<<endl;        
      PDLvStdDevQfactor << SweepPDL << " " 
                        << oHistQfactor->GetStdDev()<<endl;

      OptSNR_Mean   << SweepPDL << " " 
                    << oHistOptSNR->GetMean()<<endl;  
      OptSNR_StdDev << SweepPDL << " " 
                    << oHistOptSNR->GetStdDev()<<endl;                                           

      ElecTimeDomainSNR_Mean << SweepPDL << " " 
                    << oHistElecTimeDomainSNR->GetMean()<<endl;  
      ElecTimeDomainSNR_StdDev << SweepPDL << " " 
                    << oHistElecTimeDomainSNR->GetStdDev()<<endl;                 
      cout << "SweepPDL = " << SweepPDL 
           << "  MeanOptSNR = " << oHistOptSNR->GetMean()   
           << "  StdDevOptSNR = " << oHistOptSNR->GetStdDev() << endl;                                   
      cout << "  ->  MeanElecSNR = " << oHistElecTimeDomainSNR->GetMean()     
           << "  StdDevlElecSNR = " << oHistElecTimeDomainSNR->GetStdDev() << endl;

      //cout << "SweepPDL = " << SweepPDL 
      //     << "  <Q> = " << oHistQfactor->GetMean()            
      //     << "  <Delta Q> = " << oHistQfactor->GetMean()   
      //     << "  sigma = " << oHistQfactor->GetStdDev() << endl;           

    } // ## end for qtSweeps loop

   cout << "Total distance : " << PropagatedLength << endl;
   oReduOptSignal->WriteFileOptSignal("ReduOptSignal.dat");

   //double QfactorMean = oHistQfactor.GetMean();
   cout << endl <<"For PDL = " << SweepPDL << endl;
   cout <<"MeanQfactor   = "<<oHistQfactor->GetMean()<<endl;
   cout <<"StdDevQfactor = "<<oHistQfactor->GetStdDev()<<endl;   
   cout <<"MeanOptSNR    = " <<oHistOptSNR->GetMean()<<endl;
   cout <<"StdDevOptSNR  = " <<oHistOptSNR->GetStdDev()<<endl;         
   cout <<"qtStatSamples = " << oHistQfactor->GetQtSamples() 
        << endl;                                  

///   OutageQfactor.close();        
   PDLvMeanQfactor.close();
   PDLvStdDevQfactor.close();
   OptSNR_Mean.close();
   OptSNR_StdDev.close();            
   ElecTimeDomainSNR_Mean.close();   
   ElecTimeDomainSNR_StdDev.close();          
   delete [] Back2BackQ_Factor;
   delete [] Back2BackOptSNR;   
   delete [] Back2BackElecTimeDomainSNR;   

} //## End of ReducedModel::SweepReduPDL()

// ####################################################################


//#### Gets the pdf of the  Qfactor/QfactorBack2Back from the 
//#    reduced model.
//#    Results.
//## By: Ivan Lima



//! This method is used to compute the Delta Q distribution for a given value
//! of the PDL using Monte Carlo experiments with importance sampling
//! or simply using standard importance sampling. 

/*!

We compute the Delta Q
distribution for the channel indexed by ReducedModel::ChannelArrayIndex

When the method is used for importance sampling the method computes
results for one value of the biasing parameter. The function 
void MergeSelectMaxRelNumHits_pdfs(int Convert2dB_Scale) in
ReduExampleApplications.cc
is used to merge the distrubutions for the different bias values into
a single distribution from which outage probabilities can then be 
computed.

The method outputs the following data files:

-# Qfactor_pdf.dat: The probability density function (pdf) of the 
                    Delta Q factor by calling the function 
                    ocsHistogram::WriteFile_pdf()
-# Gaussian_pdf.dat: The pdf of the Gaussian distribution with the same mean
                    and standard deviation as the Delta Q distribution.
                    The data is written by calling the function 
                    ocsHistogram::WriteFileGaussian_pdf()

Data is written to these files every 1000 Monte Carlo experiments.

In addition the initial Stokes parameters of the signal are written to 
ReduOptSignalTx.dat using ReduOptSignal::WriteFileOptSignal() and the
Stokes parameters for the output signal in the last Monte Carlo experiment
are written to the file ReduOptSignal.dat

The Delta Q distribution is computed as in ReducedModel::SweepReduPDL().

If a long simulation performed using this method is stopped before it is
completed the user can restart the simulation from where it finished.
This is what the variables InitialSample and CurrentSample are for.
Only do this if you know what you're doing. 

*/


void ReducedModel::GetReduRelativeQfactor_pdf(void)
{   

//   ofstream OutFileDOP("DegOfPolariz.dat",ios::out);
//   if (! OutFileDOP) {
//      cerr << "Error: "<<"DegOfPolariz.dat"<<"could not be created."<<endl;
//      exit(1);
//   }

   oHistQfactor->ClearHistogram();
   double PropagatedLength;

   PropagatedLength = 0;
   oReduOptSignal->RegenerateReduOptSignal();      

   //## Save the input Stokes of the signal

   oReduOptSignal->WriteFileOptSignal("ReduOptSignalTx.dat");

   //## I should also remove the PDG when computing the 
   //#  back to back.

   double PolDepGainOptAmplifOrig = oReduOptAmplifier->GetPolDepGain(); 
             
   oReduOptAmplifier->SetPolDepGain(0.);      
   
   oReduPolDepAttenuator->SetPolDepAttenuation(0.,0.);                      

   CurrentSample = InitialSample; //## To allow the propagation

   TransmitReduOptSystem(&PropagatedLength);
   oReduOptAmplifier->SetPolDepGain(PolDepGainOptAmplifOrig);
   oReduPolDepAttenuator->SetPolDepAttenuation(0.,PolDepAttenuation);
                   
   //coReduOptSignal->GetQ_Factor(0);
   double QfactorBack2Back = GetQ_FactorSemiAnalytical(ChannelArrayIndex); 
	    
   cout << "Back to Back Q-factor = "<< QfactorBack2Back << endl;
   
   for (int ii = 0; ii <  qtStatSamples; ii++) 
    {
      CurrentSample = ii;
      PropagatedLength = 0;

      if (CurrentSample >= InitialSample) 
       {    
         oReduOptSignal->RegenerateReduOptSignal();     
       }   

      TransmitReduOptSystem(&PropagatedLength);

      //## Do not propagate for the samples previously
      //#  obtained.

      if (InitialSample != 0) 
       {
         cout << CurrentSample << endl;
       }
   
      //cout << "CurrentSample = " << CurrentSample 
      //     << "  Initial Sample = " << InitialSample << endl;            

      if (CurrentSample >= InitialSample) 
       {     
      
      //## Only the result of the 0-th channel is being used

      //oHistQfactor->AddImportantSample(
      //                oReduOptSignal->GetQ_Factor(ChannelArrayIndex)
      //                                 /QfactorBack2Back,LikelihoodRatio);

      oHistQfactor->AddImportantSample(
                      GetQ_FactorSemiAnalytical(ChannelArrayIndex)
                                       /QfactorBack2Back,LikelihoodRatio);				                                             
      //cout << ii << " Q_Factor = " 
      //     << GetQ_FactorSemiAnalytical(ChannelArrayIndex)
      //   << " Q_FactorB2B = " << QfactorBack2Back << endl;


      if (ii == 0)
         oReduOptSignal->WriteFileOptSignal("ReduOptSignal.dat");      

      // ## Creates periodic output file with the pdf     
      // #  After every 1e3 samples for preview    

      if ( (ii+1)/1e3 == (int) ( (ii+1)/1e3) ) {

         cout << " qtSamples = " << ii+1 << " ->  "
              <<"QfactorMean = "<<oHistQfactor->GetMean()
              <<"  Range: from " 
              << oHistQfactor->GetMinSample()
              <<" to " << oHistQfactor->GetMaxSample()  <<endl;           

         oHistQfactor->Obtain_pdf();              
         oHistQfactor->WriteFile_pdf("Qfactor_pdf.dat");       
         oHistQfactor->WriteFileGaussian_pdf("Gaussian_pdf.dat",1000);                  
      } 
      
      } //## End of if (CurrentSample >= ...
   }

   cout << "Total distance : " << PropagatedLength << endl;

   //double QfactorMean = oHistQfactor.GetMean();

   cout <<"QfactorMean = "<<oHistQfactor->GetMean()<<endl;
   cout <<"qtStatSamples = " << oHistQfactor->GetQtSamples() 
                             << endl;
   oHistQfactor->Obtain_pdf();

   // ## Creates output files with the pdf                             

   oHistQfactor->WriteFile_pdf("Qfactor_pdf.dat");                              
   oHistQfactor->WriteFileGaussian_pdf("Gaussian_pdf.dat",1000);

   cout << " Percentage inside the range Qfactor = "
        << oHistQfactor->GetMinRange()
        << " to " << oHistQfactor->GetMaxRange()<<" = " 
        << oHistQfactor->GetQtSamplesInRange()/ 
           (double) oHistQfactor->GetQtSamples()*100. 
        << "%." << endl;         

} // ## End of ReducedModel::GetReduRelQ_Factor_pdf()

// #################################################################

// #### Gets the pdf of the Q-factor based on the reduced model
// #    This function was written to compare with Yu Sun 
// #    Results.
// ## By: Ivan Lima

//! Used to obtain the Stokes parameters and the Q factor
//! for a single channel loop system. 

/*!

   This code was written to double check Yu Sun's results
   detailed in Yu Sun et al. OFC 2002 ThI4. It's structure is similar
   to ReducedModel::GetReduRelQ_Factor_pdf().

   \sa ReducedModel::TransmitReduOptSystem1PDL_Loop(&PropagatedLength);


 */


void ReducedModel::GetReduQfactor_pdf(void) 
{   
   //ofstream OutFileDOP("DegOfPolariz.dat",ios::out);
   //if (! OutFileDOP) {
   //   cerr << "Error: "<<"DegOfPolariz.dat"<<"could not be created."<<endl;
   //   exit(1);
   //}

   string OutFile = "ChannStokesAndQ.dat";

   ofstream OutFileChannStokesAndQ(OutFile.c_str(),ios::out);

   if (! OutFileChannStokesAndQ) 
    {
      cerr << "Error: "<< OutFile<<"could not be created."<<endl;
      exit(1);
    }
   
   oHistQfactor->ClearHistogram();
   double PropagatedLength;

   for (int ii = 0; ii <  qtStatSamples; ii++) 
    {
      PropagatedLength = 0;
      oReduOptSignal->RegenerateReduOptSignal();
      //TransmitReduOptSystem(&PropagatedLength);      
      TransmitReduOptSystem1PDL_Loop(&PropagatedLength);

      OutFileChannStokesAndQ 
                 << oReduOptSignal->GetStokesSignal(1,ChannelArrayIndex) 
            <<" "<< oReduOptSignal->GetStokesSignal(2,ChannelArrayIndex) 
            <<" "<< oReduOptSignal->GetStokesSignal(3,ChannelArrayIndex) 
            <<" "<< GetQ_FactorSemiAnalytical(ChannelArrayIndex)
   	   << endl;

      //## Only the result of ChannelArrayIndex is being used
    
//     oHistQfactor->AddImportantSample(
//                   oReduOptSignal->GetQ_Factor(ChannelArrayIndex),
//                                               LikelihoodRatio);

      oHistQfactor->AddImportantSample(
                   GetQ_FactorSemiAnalytical(ChannelArrayIndex),
                                               LikelihoodRatio);

      if (ii == 0)
         oReduOptSignal->WriteFileOptSignal("ReduOptSignal.dat");      

      // ## Creates periodic output file with the pdf     
      // #  After every 100 samples for preview    

      if ( (ii+1)/1e2 == (int) ( (ii+1)/1e2) ) 
       {

         cout << " qtSamples = " << ii+1 << " ->  "
              <<"QfactorMean = "<<oHistQfactor->GetMean()
              <<"  Range: from " 
              << oHistQfactor->GetMinSample()
              <<" to " << oHistQfactor->GetMaxSample()  <<endl;           
         oHistQfactor->Obtain_pdf();              
         oHistQfactor->WriteFile_pdf("Qfactor_pdf.dat");       
         oHistQfactor->WriteFileGaussian_pdf("Gaussian_pdf.dat",1000);                  
      } 
   }

   cout << "Total distance : " << PropagatedLength << endl;

   //double QfactorMean = oHistQfactor.GetMean();

   cout <<"QfactorMean = "<<oHistQfactor->GetMean()<<endl;
   cout <<"qtStatSamples = " << oHistQfactor->GetQtSamples() 
                             << endl;
   oHistQfactor->Obtain_pdf();
   // ## Creates output files with the pdf                             

   oHistQfactor->WriteFile_pdf("Qfactor_pdf.dat");                              
   oHistQfactor->WriteFileGaussian_pdf("Gaussian_pdf.dat",1000);

   cout << " Percentage inside the range Qfactor = "
        << oHistQfactor->GetMinRange()
        << " to " << oHistQfactor->GetMaxRange()<<" = " 
        << oHistQfactor->GetQtSamplesInRange()/ 
           (double) oHistQfactor->GetQtSamples()*100. 
        << "%." << endl;         

   OutFileChannStokesAndQ.close();        

} // ## End of ReducedModel::GetQ_Factor_pdf()


//#########################################################
//#### Gets the pdf of the Q-factor based on the reduced model
//#    This function was written to compare with Yu Sun 
//#    Results.
//#    This method generates three pdfs: one with all 
//#    the Q-factor samples, one with the
//#    Q-factor samples of a quasi-periodic loop, and the 
//#    other with the Q-factor samples of the non-periodic
//#    loop realizations. 
//## By: Ivan Lima


//! For loop systems. You probably won't use this method.

void ReducedModel::GetReduQfactorWithPeriodicity_pdf(void) 
{   
   double Periodicity;    
   Histogram *oHistQfactorPeriodic = new Histogram("Histogram.in");
   Histogram *oHistQfactorNonPeriodic = new Histogram("Histogram.in");   
   
   //ofstream OutFileDOP("DegOfPolariz.dat",ios::out);
   //if (! OutFileDOP) {
   //   cerr << "Error: "<<"DegOfPolariz.dat"<<"could not be created."<<endl;
   //   exit(1);
   //}
   string OutFile = "ChannStokesAndQ.dat";
   ofstream OutFileChannStokesAndQ(OutFile.c_str(),ios::out);
   if (! OutFileChannStokesAndQ) {
      cerr << "Error: "<< OutFile<<"could not be created."<<endl;
      exit(1);
   }   
   oHistQfactor->ClearHistogram();
   double PropagatedLength;
   for (int ii = 0; ii <  qtStatSamples; ii++) {
      PropagatedLength = 0;
      oReduOptSignal->RegenerateReduOptSignal();
      //TransmitReduOptSystem(&PropagatedLength);      
      Periodicity = TransmitReduOptSystem1PDL_LoopWithMonitoring(&PropagatedLength);
      
      OutFileChannStokesAndQ 
            <<      oReduOptSignal->GetStokesSignal(1,ChannelArrayIndex) 
            <<" "<< oReduOptSignal->GetStokesSignal(2,ChannelArrayIndex) 
            <<" "<< oReduOptSignal->GetStokesSignal(3,ChannelArrayIndex) 
            <<" "<< GetQ_FactorSemiAnalytical(ChannelArrayIndex)
   	   << endl;
	   
      //## Only the result of ChannelArrayIndex is being used
      oHistQfactor->AddImportantSample(
                   GetQ_FactorSemiAnalytical(ChannelArrayIndex),
                                               LikelihoodRatio);
      
/*
      OutFileChannStokesAndQ << oReduOptSignal->GetStokesSignal(1,0) 
                        <<" "<< oReduOptSignal->GetStokesSignal(2,0) 
                        <<" "<< oReduOptSignal->GetStokesSignal(3,0) 
                        <<" "<< oReduOptSignal->GetQ_Factor(0)
                        << endl; 

      //## Only the result of the 0-th channel is being used
      oHistQfactor->AddImportantSample(oReduOptSignal->GetQ_Factor(0),
                              LikelihoodRatio);
*/			      
      cout << ii <<"  "<< Periodicity << endl;
      if (Periodicity > 0.5) {
         oHistQfactorPeriodic->AddImportantSample(
	                   GetQ_FactorSemiAnalytical(ChannelArrayIndex),
                           LikelihoodRatio);                              
      }                           
      else {
         oHistQfactorNonPeriodic->AddImportantSample(
	                   GetQ_FactorSemiAnalytical(ChannelArrayIndex),
                           LikelihoodRatio);      
      }
                              
      if (ii == 0)
         oReduOptSignal->WriteFileOptSignal("ReduOptSignal.dat");      
      // ## Creates periodic output file with the pdf     
      // #  After every 100 samples for preview    
      if ( (ii+1)/1e2 == (int) ( (ii+1)/1e2) ) {

         cout << " qtSamples = " << ii+1 << " ->  "
              <<"QfactorMean = "<<oHistQfactor->GetMean()
              <<"  Range: from " 
              << oHistQfactor->GetMinSample()
              <<" to " << oHistQfactor->GetMaxSample()  <<endl;           
                
         oHistQfactor->Obtain_pdf();              
         oHistQfactor->WriteFile_pdf("Qfactor_pdf.dat");       
         oHistQfactor->WriteFileGaussian_pdf("Gaussian_pdf.dat",1000);                  

         oHistQfactorPeriodic->WriteFile_pdf("QfactorPeriodic_pdf.dat");                
         oHistQfactorNonPeriodic->WriteFile_pdf("QfactorNonPeriodic_pdf.dat");                
      } 
   }
   cout << "Total distance : " << PropagatedLength << endl;

   //double QfactorMean = oHistQfactor.GetMean();
   cout <<"QfactorMean = "<<oHistQfactor->GetMean()<<endl;
   cout <<"qtStatSamples = " << oHistQfactor->GetQtSamples() 
                             << endl;
   cout << " % SamplesNonPeriodic = " 
        << 100*oHistQfactorNonPeriodic->GetQtSamples()
              /oHistQfactor->GetQtSamples()
        << "   % SamplesPeriodic = "       
        << 100*oHistQfactorPeriodic->GetQtSamples()
              /oHistQfactor->GetQtSamples() << endl;                                  
   oHistQfactor->Obtain_pdf();
   // ## Creates output files with the pdf                             
   oHistQfactor->WriteFile_pdf("Qfactor_pdf.dat");                              
   oHistQfactor->WriteFileGaussian_pdf("Gaussian_pdf.dat",1000);

   oHistQfactorPeriodic->WriteFile_pdf("QfactorPeriodic_pdf.dat");                
   oHistQfactorNonPeriodic->WriteFile_pdf("QfactorNonPeriodic_pdf.dat");                

   cout << " Percentage inside the range Qfactor = "
        << oHistQfactor->GetMinRange()
        << " to " << oHistQfactor->GetMaxRange()<<" = " 
        << oHistQfactor->GetQtSamplesInRange()/ 
           (double) oHistQfactor->GetQtSamples()*100. 
        << "%." << endl;         
   OutFileChannStokesAndQ.close();        
   
   delete oHistQfactorPeriodic;
   delete oHistQfactorNonPeriodic;
   
} // ## End of ReducedModel::GetReduQfactorWithPeriodicity_pdf()



// #### Gets the outage probability to compare with Ding's results
// ##   This function is still using numerical extrapolation
// #   This function should be replaced by the importance 
// #   sampling. Thus, it is not even going to be tested.
// ## By: Ivan Lima


//! See void GetReduOutageProb(MyStartup *oMyStartup)

void ReducedModel::GetReduOutageProb(double PowerPenaltyMargin,
                                     int MinNumChannels, 
                                     int MaxNumChannels,
                                     int StepNumChannels) 
{
   ofstream OutageQfactor("OutageQfactor.dat",ios::out);
   if (! OutageQfactor) {
      cerr << "Error: "<<"OutageQfactor.dat"<<" could not be created."<<endl;
      exit(1);
   }
   
   //ofstream OutFileDOP("DegOfPolariz.dat",ios::out);
   //if (! OutFileDOP) {
   //   cerr << "Error: "<<"DegOfPolariz.dat"<<"could not be created."<<endl;
   //   exit(1);
   //}
           
   double PropagatedLength;
   int SweepNumChannels;
   double MeanQfactor0;

   double PolDepGainOptAmplifOrig = oReduOptAmplifier->GetPolDepGain();

   int qtStatSamplesTmp = 1;
   for ( int SweepNumChannels = MinNumChannels;
        SweepNumChannels <=  MaxNumChannels;
        SweepNumChannels += StepNumChannels) {
      oRanPMD->ResetRanNumGen();          
      oReduOptSignal->SetNumChannels(SweepNumChannels);   
      for (int jj = 0; jj < qtMapSpans; jj++) {      
         oReduOptFiber[jj]->SyncNumChannels(); 
      }                
      oReduOptAmplifier->SyncNumChannels();         
      oReduPolDepAttenuator->SyncNumChannels();         
                         
      oReduPolDepAttenuator->SetPolDepAttenuation(0.,0.);    
      oReduOptAmplifier->SetPolDepGain(0.);      

      //## Propagate once without any PDL and PDG
      //## to get the PDL/PDG free Q-factor
      //#  MeanQfactor0
      PropagatedLength = 0;
      oReduOptSignal->RegenerateReduOptSignal();
      TransmitReduOptSystem(&PropagatedLength);
      //MeanQfactor0 = oReduOptSignal->GetQ_Factor(0);
      MeanQfactor0 = GetQ_FactorSemiAnalytical(ChannelArrayIndex);

      //## Now starts the actual averaging
      oReduPolDepAttenuator->SetPolDepAttenuation(0.,PolDepAttenuation);       
      oReduOptAmplifier->SetPolDepGain(PolDepGainOptAmplifOrig);      
      oHistQfactor->ClearHistogram();
      for (int ii = 0; ii <  qtStatSamples; ii++) {
         PropagatedLength = 0;
         oReduOptSignal->RegenerateReduOptSignal();
         TransmitReduOptSystem(&PropagatedLength);
         //## It should be in linear scale
         //oHistQfactor->AddSample(  oReduOptSignal->GetQ_Factor(0)
         //                                    /MeanQfactor0);
         oHistQfactor->AddSample(  GetQ_FactorSemiAnalytical(ChannelArrayIndex)
                                             /MeanQfactor0);

      }                      
      //OutageQfactor << SweepNumChannels << " "
      //              << 1.- (double) oHistQfactor->GetQtSamplesInRange()
      //                   /oHistQfactor->GetQtSamples() << endl;
      //cout << "GetMean() = " << oHistQfactor->GetMean() << endl;
      //cout << "GetStdDev() = " << oHistQfactor->GetStdDev() << endl;      
      OutageQfactor << SweepNumChannels << " "
                    << oReduOptSignal->GetOutageProbGaussian(
                                    oHistQfactor->GetMean(),
                                    oHistQfactor->GetStdDev(),
                                    -PowerPenaltyMargin) << endl;
     cout  << "2.5dB -> " << SweepNumChannels << " "
                    << 1.- (double) oHistQfactor->GetQtSamplesInRange()
                         /oHistQfactor->GetQtSamples() << endl;
     cout << SweepNumChannels << " "
                    << oReduOptSignal->GetOutageProbGaussian(
                                    oHistQfactor->GetMean(),
                                    oHistQfactor->GetStdDev(),
                                    -PowerPenaltyMargin) << endl;                         
      cout << oHistQfactor->GetMinSample() << " "
           << oHistQfactor->GetMaxSample() << " "                         
           << oHistQfactor->GetQtSamples() << " "   
           << oHistQfactor->GetQtSamplesInRange() << endl;   
      oHistQfactor->WriteFile_pdf("Qfactor_pdf.dat");                             
      //if (SweepPDL == MinPDL) {
      //   MeanQfactor0 = oHistQfactor.GetMean();
      //   PDLvMeanQfactor   << SweepPDL <<" "<< 0 << endl;         
      //   PDLvStdDevQfactor << SweepPDL <<" "<< 0 << endl;                             
      // }
      //else {
      //   PDLvMeanQfactor   << SweepPDL << " " 
      //                     << oHistQfactor.GetMean()<<endl;        
      //   PDLvStdDevQfactor << SweepPDL << " " 
      //                     << oHistQfactor.GetStdDev()<<endl;
      //   cout << "SweepPDL = " << SweepPDL 
      //        << "  <Q> = " << oHistQfactor.GetMean()            
      //        << "  <Delta Q> = " << oHistQfactor.GetMean()   
      //        << "  sigma = " << oHistQfactor.GetStdDev() << endl;
      //}        
   }
   cout << "Total distance : " << PropagatedLength << endl;
   oReduOptSignal->WriteFileOptSignal("ReduOptSignal.dat");

   //double QfactorMean = oHistQfactor->GetMean();
   //cout << endl <<"For PDL = " << SweepPDL << endl;
   //cout <<"MeanQfactor   = "<<oHistQfactor.GetMean()<<endl;
   //cout <<"StdDevQfactor = "<<oHistQfactor.GetStdDev()<<endl;   
   //cout <<"qtStatSamples = " << oHistQfactor.GetQtSamples() 
   //                          << endl;                                  
} // ## End of ReducedModel::GetReduOutageProb(.)


// ##########################################################

//## Get the evolution of the noise without signal
//#  It was written exclusivelly for the SAIC report


//! FOR DEBUGGING

void ReducedModel::GetSignalFreeNoiseEvolution(void) 
{   
   double NumChannels = oReduOptSignal->GetNumChannels();
   //if (NumChannels != 1) {
   //   cout << "This module was designed for a single channel"
   //        << "System."<< endl;
   //   exit(1);
   //}
   string OutFile = "NoiseDOP.dat";
   ofstream OutFileNoiseDOP(OutFile.c_str(),ios::out);
   if (! OutFileNoiseDOP) {
      cerr << "Error: "<< OutFile<<"could not be created."<<endl;
      exit(1);
   }   

   oReduOptSignal->RegenerateReduOptSignal();   
   double *PropagatedLength = new double;
   *PropagatedLength = 0;   
   oRanPMD->WriteRanSeqParameters("RanPMD.tmp");     
   TransmitReduOptSystem1PDL_Loop(PropagatedLength);           
   cout << "Q-factor = " << GetQ_FactorSemiAnalytical(ChannelArrayIndex)
        << endl;

   oRanPMD->ReadRanSeqParameters("RanPMD.tmp");  
   qtLoops *= 2;
   *PropagatedLength = 0;   
   oReduOptSignal->RegenerateReduOptSignal();  
   for (int ii = 0; ii < NumChannels; ii++) {
      for (int jj = 0; jj < 4; jj++) {
         oReduOptSignal->StokesSignal[jj][ii] *= 0;
      }
      oReduOptSignal->StokesSignal[0][ii] = 1e-7;        
   }  
   
   //## Imported from 
   //#  void ReducedModel::TransmitReduOptSystem1PDL_Loop(double *PropagatedLength)             
   LikelihoodRatio = 1.; 
   double psi,theta,phi;
   if ( ! IsStraightLine && ! LoopPolarizControl) { //## It does only once      
      for (int jj = 0; jj < qtMapSpans; jj++) {  
         oReduOptFiber[jj]->GetFiberSampleCompleteScatt();               
      }     
   }   
   if (! IsStraightLine && LoopPolarizControl) {
      psi   = 2*pi*oRanPMD->GetRanNum();
      theta = acos( 2*(oRanPMD->GetRanNum() - 0.5) );
      phi   = 2*pi*oRanPMD->GetRanNum();     
   }      
   for (int kk = 0; kk < qtLoops; kk++) {
      if ( IsStraightLine) { //## Each loop is independent
         for (int jj = 0; jj < qtMapSpans; jj++) {
            oReduOptFiber[jj]->GetFiberSampleCompleteScatt();             
         }
      }                 
      if (! IsStraightLine && LoopPolarizControl) {
         oReduPolarizController->TransformPolarization(
               psi,theta,phi);
      }      
      for (int jj = 0; jj < qtMapSpans; jj++) {
         oReduOptFiber[jj]->PropagateFiberCoupledNLS
                           (PropagatedLength);                                     
         oReduOptAmplifier->AmplifyOptSignal();          
         /*       
         cout << "StokesSignal[0][0] = "<< oReduOptSignal->StokesSignal[0][0] << endl;
         cout << "StokesSignal[1][0] = "<< oReduOptSignal->StokesSignal[1][0] << endl;
         cout << "StokesNoise[0][0] = "<< oReduOptSignal->StokesNoise[0][0] << endl;
         cout << "StokesNoise[1][0] = "<< oReduOptSignal->StokesNoise[1][0] << endl;
         exit(1);
         */ 
      }   
      // ## Yu Sun has one PDL per Loop                  
      if (BiasingParam != 0) {
          //## Attention: The importance sampling is not coded to 
          //#  work in a loop due to the periodicity.
         oReduPolarizTransformer->GetImportSampPolarizTransformation(
               ChannelArrayIndex, &LikelihoodRatio);         
      }      
      //cout << oReduOptSignal->StokesSignal[0][0] << " ";
      //cout << oReduOptSignal->StokesSignal[1][0] << " ";      
      //cout << oReduOptSignal->StokesSignal[2][0] << " ";      
      //cout << oReduOptSignal->StokesSignal[3][0] << endl;            
      oReduPolDepAttenuator->PolDepAttenuate();       
      if (BiasingParam != 0) {
         oReduPolarizTransformer->UndoLastPolarizTransformation();         
      }            
      //## Since the switcher partially polarizes the noise,
      //#  we will start from 20% of noise DOP in the channel 0.
      //#  In this situation, starting with no signal,
      //#  a saturable amplifier would be advisable.

      double PolarizedPower;                                        
      if (kk == 0 ) {
         for (int ii = 0; ii < NumChannels; ii++) {
            PolarizedPower = 
               sqrt( sq(oReduOptSignal->StokesNoise[1][ii])
                    +sq(oReduOptSignal->StokesNoise[2][ii])
                    +sq(oReduOptSignal->StokesNoise[3][ii]) );               
            for (int jj = 0; jj < 4; jj++) {
               oReduOptSignal->StokesNoise[jj][ii] =          
                      .3 * oReduOptSignal->StokesNoise[0][ii]
                      * oReduOptSignal->StokesNoise[jj][ii]
                       /PolarizedPower;
            }           
         }
      }                  
      //## End of a loop
      OutFileNoiseDOP << *PropagatedLength << " "
                      << oReduOptSignal->GetDegreeOfPolarization() << endl;
//            << sqrt( sq(oReduOptSignal->GetStokesNoiseInsideChannel(1,0) ) 
//              +sq(oReduOptSignal->GetStokesNoiseInsideChannel(2,0) ) 
//              +sq(oReduOptSignal->GetStokesNoiseInsideChannel(3,0) ) )
//            << endl; 
      
   }       
   OutFileNoiseDOP.close();
}


//######################################################

//! FOR DEBUGGING

void TestPDLandPDG(MyStartup *oMyStartup)
{
   RanNumGen oRanPMD("RanPMD.in");
   ReduOptSignal oReduOptSignal("OptSignal.in",25e9);
   double OpticalBandwidth = oMyStartup->GetDoubleParam(1);
   double ElectricalBandwidth 
                         = oMyStartup->GetDoubleParam(2); 
   oReduOptSignal.SetOptElecFilterBandwidth
       (OpticalBandwidth   // ## Optical BW
        ,ElectricalBandwidth );// ## Electrical BW
                                              
   ReduPolDepAttenuator oReduPolDepAttenuator
                             (&oReduOptSignal);                                                                       
   double PolDepAttenuation = oMyStartup->GetDoubleParam(3);                          
   oReduPolDepAttenuator.SetPolDepAttenuation
                   (0., PolDepAttenuation);     
   ReduOptAmplifier  oReduOptAmplifier("OptAmplifier.in"
                                      ,&oReduOptSignal);                
//                                      ,OpticalBandwidth);   

   ReduOptFiber oReduOptFiber("OptFiber.in"
                              ,&oReduOptSignal
                              ,&oRanPMD); 
                              
   oReduOptSignal.RegenerateReduOptSignal();                     
   
   double PropagatedLength = 0;

   oReduOptFiber.GetFiberSampleCompleteScatt();             
   //oReduOptFiber.PropagateFiberCoupledNLS
   //                           (&PropagatedLength);   
   oReduPolDepAttenuator.PolDepAttenuate();                              
   //oReduOptAmplifier.AmplifyOptSignal();

   cout << "Total distance : " << PropagatedLength << endl;
   oReduOptSignal.WriteFileOptSignal("ReduOptSignal.dat");
       

} // ## End of TestPDLandPDG

// ##############################################################
