//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################


// ###############################################
// 
// The applications in this file were written by John Zweck
// and are not to be modified by anyone else for now.
// 
// ###############################################

#include <iostream.h>
#include <complex.h>

#include "ocs.hh"
#include "StartupTyco.hh"
#include "ocsOptDataWriter.hh"
#include "MachZehnder.hh"


extern ofstream *LogFile;

void Tyco10Basic(StartupZweck * MyStartup);

// ####################################################

// ## Application Wrapper

void MyApplication(int argc, char *argv[])  
{

  StartupZweck MyStartup("Startup.in");

  switch(MyStartup.GetSimType())
   {
    
     case 1: 
       Tyco10Basic(&MyStartup);
       break;

    default:
      cerr << "MyApplication: Invalid choice of application in SimType" 
           << endl;
      exit(1);
          
   } // end switch

} // end MyAppn


// ####################################################

void Tyco10Basic(StartupZweck * MyStartup)
{

  LogFileSeparator();

  *LogFile <<  endl << "Tyco10Basic" << endl;

  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();

  RanNumGen RNG_BS1(InDir + "RanNumGenBitString1.in");
  RanNumGen RNG_BS2(InDir + "RanNumGenBitString2.in");
  RanNumGen RNG(InDir + "RanNumGen.in");

   // ## Initiating the optical pulses

  OptSignal * Signal;
  OptSignal * Signal1;
  OptSignal * Signal2;
  MachZehnder * MZ1;
  MachZehnder * MZ2;

  if(MyStartup->GetHybridSignalFlag())
    {
      Signal1 = new OptSignal(InDir + "Signal1.in",&RNG_BS1);
 
      if(Signal1->GetTypePulse() == NRZ)
      {
        MZ1  = new MachZehnder(InDir + "MZ.in",Signal1);
        MZ1->MZModulator();
      }

      Signal2 = new OptSignal(InDir + "Signal2.in",&RNG_BS2);

     if(Signal2->GetTypePulse() == NRZ)
      {
        MZ2 = new MachZehnder(InDir + "MZ.in",Signal2);
        MZ2->MZModulator();
      }

     Signal = new OptSignal (*Signal1,*Signal2,&RNG);  
                             // I think RNG this is OK
    }
  else
    {
      Signal = new OptSignal(InDir + "Signal.in",&RNG_BS1);

     if(Signal->GetTypePulse() == NRZ)
      {
        MZ1 = new MachZehnder(InDir + "MZ.in",Signal);
        MZ1->MZModulator();
      }
    }

   // #########################

  // To get initial OSNR correct we fudge things by preamplifying
  // to get the correct SNR and then applying some lumped loss to get
  // max signal power correct

  OptAmplifier TransmitterAmp(InDir + "TransmitterAmp.in",Signal,&RNG);
  
  if(MyStartup->GetAmpsNoiseOffFlag())
     {
       TransmitterAmp.SetTypeAmplifierNoise(NOISE_OFF);
     }


  TransmitterAmp.AmplifyOptSignal();
  TransmitterAmp.SetTypeAmplifier(typeAmplifier(MyStartup->GetTypeAmp()));


  LumpedLoss TransmitterLoss(InDir + "TransmitterLoss.in",Signal);
  TransmitterLoss.ApplyLumpedLoss();



   OptDataWriter DataWriter(InDir + "DataWriter.in",Job,Signal);

   DataWriter.WriteInitialData();

 // ## Output  pulse width using rms measure

   cout << "Initial Average FWHM in time [ps] = " 
	<< 1e12*Signal->ComputeAverageSigmaMarks()
        << endl;

   // FIX FACTOR ABOVE FOR CRZ


   cout << "Initial Average Spread in Frequency [GHz] = "
        << 1e-9*sqrt(Signal->ComputeSigmaSqFrequency(
                     MyStartup->GetKeepDCFlag()))
        << endl;



   // ## Declare and initialize the fibers 

    OptFiberLocalError PDCF(InDir + "PDCF.in",Signal,&RNG,Job);
    OptFiberLocalError DSF45(InDir + "DSF45.in",Signal,&RNG,Job);
    OptFiberLocalError DSF25(InDir + "DSF25.in",Signal,&RNG,Job);
    OptFiberLocalError SMF(InDir + "SMF.in",Signal,&RNG,Job);

   
    int NumFibers = 4;

    OptFiberLocalError * FiberPtr[NumFibers];

    FiberPtr[0] = &PDCF;
    FiberPtr[1] = &DSF45;
    FiberPtr[2] = &DSF25;
    FiberPtr[3] = &SMF;

    OptAmplifier Amp(InDir + "Amp.in",Signal,&RNG);
    OptAmplifier PAmp(InDir + "PAmp.in",Signal,&RNG);

    int NumAmps = 2;

    OptAmplifier * AmpPtr[NumAmps];

    AmpPtr[0] =  &Amp;
    AmpPtr[1] =  &PAmp;

  if(MyStartup->GetDebugLevelAmps())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetDebugLevel(MyStartup->GetDebugLevelAmps());
     }


   
  for(int i=0;i<NumAmps;i++)
    AmpPtr[i]->SetTypeAmplifier(typeAmplifier(MyStartup->GetTypeAmp()));
     


   if(MyStartup->GetAmpsNoiseOffFlag())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetTypeAmplifierNoise(NOISE_OFF);
     }


   if(MyStartup->GetBeta3ZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetSndOrDispFiber(0);
     }

   if(MyStartup->GetNonLinearityZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetGammaFiber(0);
     }

   if(MyStartup->GetFiberLossZeroFlag()) 
        for(int ii = 0; ii< NumFibers;ii++)
            FiberPtr[ii]->SetAttenuation(0);

    for(int ii = 0; ii< NumFibers;ii++)
         FiberPtr[ii]->SetRelativeErrorGoal(MyStartup->GetRelativeErrorGoal());

      for(int ii = 0; ii< NumFibers;ii++)
	{
          FiberPtr[ii]->SetWritePulseWidthFlag(
                        MyStartup->GetWritePulseWidthFlag());
	  
          FiberPtr[ii]->SetWriteAveragePowerdBmFlag(
                        MyStartup->GetWriteAveragePowerdBmFlag());

           FiberPtr[ii]->SetOutputStepSizesFlag(
                            MyStartup->GetOutputStepSizesFlag());
	}



  // ########## Initialize OSAs ###########################

    int OSAAppendFlag = 0;

    OptSpectrumAnalyzer OSAInitial(InDir + "OSA.in",Job,OSAAppendFlag, 
                                   Signal);
    OptSpectrumAnalyzer OSAFinal(InDir + "OSA.in",Job,OSAAppendFlag, 
                                   Signal);

  // ##### Back to Back OSA ################

    OSAInitial.Update(0);
    OSAInitial.Output(0);

    // exit(1);

 // ############ Initialize Receiver susbsystems  #####
 

   ReceiverSubsystem Receiver(Signal,&RNG,InDir,Job,
                              MyStartup->GetDemuxChannelIndex());

   Receiver.ReceiveSignalBackToBack(Signal);

 
    // #### Propagate signal through fiber ####


   for(int expt_num=0;expt_num<MyStartup->GetQtStatSamples();expt_num++)
      {

         // #### Propagate through loop ####
 
       double PropagatedLength = 0;
       int ZStepNum = 0;


      PDCF.Propagate(&PropagatedLength,&ZStepNum);

      int kk = 0;

      DataWriter.WriteEndMap(expt_num,kk,PropagatedLength);

    for (kk = 0; kk < MyStartup->GetQtMapSpans(); kk++) 
     {

       for(int jj = 0; jj < 3; jj++)
	 {

          DSF45.Propagate(&PropagatedLength,&ZStepNum);
          DataWriter.WriteWithinMap(expt_num,kk,PropagatedLength);
          Amp.AmplifyOptSignal();

         DataWriter.WriteWithinMap(expt_num,kk,PropagatedLength);
         
	 } // end for jj inner loop

       DSF25.Propagate(&PropagatedLength,&ZStepNum);
       DataWriter.WriteWithinMap(expt_num,kk,PropagatedLength);

       SMF.Propagate(&PropagatedLength,&ZStepNum);
       DataWriter.WriteWithinMap(expt_num,kk,PropagatedLength);
       Amp.AmplifyOptSignal();
       DataWriter.WriteEndMap(expt_num,kk,PropagatedLength);

      if(expt_num == 0 && kk % 10 == 0 && kk != 0)
	cout << "Completed " << kk << " periods of dispersion map " << endl;

    }// ## end for kk-loop

  DataWriter.WriteEndMap(expt_num,kk,PropagatedLength);

  PDCF.Propagate(&PropagatedLength,&ZStepNum);
  DataWriter.WriteEndMap(expt_num,kk,PropagatedLength);
  PAmp.AmplifyOptSignal();

    // ############### End of Propagation #################

  DataWriter.WriteEndTransmission(expt_num,kk,PropagatedLength);
 
  // ##### Final OSA and SNR ########

  if(MyStartup->GetWriteSNRandOSAFlag())
      OSAFinal.Update(PropagatedLength);

 if(MyStartup->GetWriteSNRandOSAFlag()  
    &&(expt_num == 0 || expt_num % MyStartup->GetFreqOutputSNR() == 0) )
    {
     OSAFinal.Output(expt_num+1);
    }


 // ######## Enter Receiver subsystem and output Eye and Q inforamtion ###

  int UpdateIvansEDFlag = 0;
  int OutputPD_DataFlag = 0;

  //if(expt_num == 0)
  //   OutputPD_DataFlag = 1;

  if(expt_num <= MyStartup->GetNumExptsForIvansED())
     UpdateIvansEDFlag = 1;

  
  Receiver.ReceiveSignal(Signal,UpdateIvansEDFlag,
                             OutputPD_DataFlag,0,0,PropagatedLength);
    
   if(expt_num % MyStartup->GetFreqOutputReceiverStats() == 0)
    {
         int StatsAppendFlag;

         if(expt_num==0)
           StatsAppendFlag = 0;
         else
           StatsAppendFlag = 1;

          Receiver.PostProcessReceiverStats(
                                     expt_num+1,0,StatsAppendFlag);
         
        
    } // ## end if

  if(expt_num==0)
    {

       // ####### Output  pulse width using rms measure

  
     cout << "Final Average FWHM = " 
	<< 1e12*Signal->ComputeAverageSigmaMarks()
        << endl; // FIX FOR CRZ

     cout << "Final Average Spread in Frequency [GHz] = "
        << 1e-9*sqrt(Signal->ComputeSigmaSqFrequency(
                                MyStartup->GetKeepDCFlag()))
        << endl;

        cout <<"Total PropagatedLength [km] = "<<PropagatedLength*1e-3<< endl;
        cout <<"Number Steps in Z taken by solver = " << ZStepNum << endl;

    }// end if expt_num==0

   

    
   // ######### Prepare for next expt ##############

    for(int i=0;i<NumFibers;i++)
    {
     FiberPtr[i]->ResetVolatileVariables();
     FiberPtr[i]->SetWriteAveragePowerdBmFlag(0);
    }

    if(MyStartup->GetHybridSignalFlag())
    {
     if(MyStartup->GetRandomizeBitStringsForMC())
      {
        Signal1->GetNewRandomSignalString();
        Signal2->GetNewRandomSignalString();
      }

     Signal1->RegenerateSignalString();
      if(Signal1->GetTypePulse() == NRZ)
       MZ1->MZModulator();

     Signal2->RegenerateSignalString();
      if(Signal2->GetTypePulse() == NRZ)
       MZ2->MZModulator();

     AddSignals(Signal1,Signal2,Signal);
    }
  else
   {
    if(MyStartup->GetRandomizeBitStringsForMC())
      {
        Signal->GetNewRandomSignalString();
      }

     Signal->RegenerateSignalString();
      if(Signal->GetTypePulse() == NRZ)
       MZ1->MZModulator();

   }

    TransmitterAmp.AmplifyOptSignal();
    TransmitterLoss.ApplyLumpedLoss();

   
    cout << "Finished Expt " << expt_num << endl;

      } // ## end for expt loop

   cout << "THE END" << endl;
   *LogFile << "THE END" << endl;



} // ## end Tyco10Basic

// #############################################







