// $Id: ocsExampleApplication.cc,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

// ##########################################
//
// This file contains  sample applications to show
// how the code can be used. 
//
// ##########################################

#include <iostream.h>
#include <complex.h>
#include "ocs.hh"

#include "StartupExample.hh"

// ## Sample Applications

// ## These applications illustrate how to use the ocs library.
// ## They do not attempt to be realistic simulations of 
// ## fiber optic communaction systems. 

 

void SingleChannelScalarApplication(StartupExample * MyStartup);
void WDM_Application(StartupExample * MyStartup);

// ## I suggest that Ivan contribute a simple single channel vector
// ## application in which the propagation is done using the coarse-step
// ## Manakov PMD equation (JZ) 

extern ofstream *LogFile;

// #### The optical pulses must be local globals 



// #######################################################

// ## Application Wrapper

void MyApplication(int argc, char *argv[]) 
{

 StartupExample MyStartup("StartupExample.in",argc,argv);


  switch(MyStartup.GetSimType())
   {
    
    case 1:
      SingleChannelScalarApplication(&MyStartup);
      break;

    case 2:
      WDM_Application(&MyStartup);
      break;

    default:
      cerr << "Invalid choice of application in SimType" << endl;
      exit(1);
          
   } // end switch

} // ## end MyApplication

// ########################################################

void SingleChannelScalarApplication(StartupExample * MyStartup)
{

  // Written by John Zweck, Dec 14, 2000

  // InDir is the directory in which the input files are kept.
  // Job is a string of the form <jobdirectorypath>/<job_id>
  // where <jobdirectorypath> is the directory in which the output is to be
  // sent and <job_id> is the name of the particular job. 

  // The code needs to be run from the directory in which the input files
  // reside.

  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();

  LogFileSeparator();
  *LogFile << "SingleChannel_Application" << endl;

  // ## Declare RanNumGen and OptSignal objects 

  RanNumGen RNG(InDir + "RanNumGen.in");
  OptSignal SC_Signal(InDir + "OptSignal.in", &RNG);

  // ## Write the initial optical signal power in time and 
  // ## in frequency to a file 

  int OmitDCFlag = 1;

  SC_Signal.WriteFileTimePower(Job + "tInOptPower.dat");
  SC_Signal.WriteFileFreqPower(Job + "fInOptPower.dat",OmitDCFlag);

  // Declare the components of a receiver subsystem

  OptFilter oOptFilter(InDir + "OptFilter.in",&SC_Signal);
  Photodetector  oPhotodetector(InDir + "Photodetector.in",&SC_Signal); 
  ElecFilter oElecFilter(InDir + "ElecFilter.in",&oPhotodetector);
  ElecSignalStat oElecSignalStat(&SC_Signal,&oPhotodetector);   

  // Pass the back-to-back optical signal through 
  // the receiver 

   oOptFilter.FilterOptSignal();
   oPhotodetector.DetectOptSignal();   
   oElecFilter.FilterElecSignal();
   oElecSignalStat.GetSingleStringElecSignalStat();

   // Write the electrical signal power in time and in frequency  to a file

   oPhotodetector.WriteFileTimePower(Job + "tInFilteredPower.dat");   
   oPhotodetector.WriteFileFreqPower(Job + "fInFilteredPower.dat",OmitDCFlag);

   // ## Prepare to  propagate  the signal over the system    
   // Since the Optical Filter applied above changes the data stored in the
   // arrays we need to re-initialize it

   SC_Signal.RegenerateSignalString();  

   OptFiber MyFiber(InDir + "OptFiber.in", &SC_Signal,&RNG);

   double PropagatedLength = 0.0;

   // The propagation below is scalar - only the X-polarozation is tracked.
   // The Y-polarization is set to zero.

   // Because of this we need to call a scalar optical amplifier,
   // to ensure that ASE noise is put into the X-polarization only.

   MyFiber.PropagateFiberScalarNLS(&PropagatedLength);
              
   OptAmplifier oAmp(InDir + "OptAmplifier.in" ,&SC_Signal,&RNG);
   oAmp.AmplifyOptSignal();   
                                
   // Receive the signal

   oOptFilter.FilterOptSignal();
   oPhotodetector.DetectOptSignal();   
   oElecFilter.FilterElecSignal();
   oElecSignalStat.GetSingleStringElecSignalStat();

   oPhotodetector.WriteFileTimePower(Job + "tOutFilteredPower.dat");   
   oPhotodetector.WriteFileFreqPower(Job + "fOutFilteredPower.dat",OmitDCFlag);


} // ## end SingleChannelScalarApplication


// #######################################################

void WDM_Application(StartupExample * MyStartup)
{

 // Written by John Zweck, Dec 14, 2000

  LogFileSeparator();
  *LogFile << "WDM_Application" << endl;

  // InDir is the directory in which the input files are kept.
  // Job is a string of the form <jobdirectorypath>/<job_id>
  // where <jobdirectorypath> is the directory in which the output is to be
  // sent and <job_id> is the name of the particular job. 

  // The code needs to be run from the directory in which the input files
  // reside.

  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();

  RanNumGen RNG(InDir + "RanNumGen.in");
  OptSignal WDMSignal(InDir + "OptSignal.in",&RNG);

  int OmitDCFlag = 0;

  WDMSignal.WriteFileTimePower(Job + "tInOptPower.dat");
  WDMSignal.WriteFileFreqPower(Job + "fInOptPower.dat",OmitDCFlag);
  
  OptDataWriter DataWriter(InDir +"DataWriter.in" ,InDir,Job,&WDMSignal);

  OptFiber Fiber(InDir + "OptFiber.in",&WDMSignal,&RNG);
  OptAmplifier Amp(InDir + "OptAmplifier.in",&WDMSignal,&RNG);

// ############ Initialize Receiver susbsystems  #####
 

  ReceiverSubsystem * Receiver = new 
                      ReceiverSubsystem[MyStartup->GetNumChannsDemux()];

  for(int dd = 0; dd < MyStartup->GetNumChannsDemux(); dd++)
    {

      Receiver[dd].ReceiverSubsystemInit(&WDMSignal,&RNG,InDir,Job, 
                  MyStartup->GetDemuxChannsArray(dd),&DataWriter);

    }


  // ################ Initialize Back to Back Pulses ##############

  
 for(int dd = 0; dd < MyStartup->GetNumChannsDemux(); dd++)
    {
      Receiver[dd].ReceiveSignalBackToBack(&WDMSignal);
    }


  // ################ Propagate through loop ############


  LogFileSeparator();
  *LogFile << "Propagating through loop" << endl;

 double PropagatedLength; 

 for(int expt_index = 0; expt_index < MyStartup->GetQtStatSamples();
      expt_index++)
   {

    PropagatedLength = 0.0;

    for (int kk = 0; kk < MyStartup->GetQtMapSpans(); kk++) 
     {

       // Scalar propagation as in  SingleChannelScalarApplication

         Fiber.PropagateFiberScalarNLS(&PropagatedLength);
         Amp.AmplifyOptSignal();

     } // end kk-loop

    if(expt_index == 0)
        {

       cout <<"Total PropagatedLength [km] = "<<PropagatedLength*1e-3<< endl;
 
       WDMSignal.WriteFileTimePower(Job + "tOutOP.dat");
       WDMSignal.WriteFileFreqPower(Job + "fOutOP.dat",OmitDCFlag);
      }

     int UpdateIvansEDFlag = 0;
     int OutputPD_DataFlag = 0;

     if(expt_index == 0)
        OutputPD_DataFlag = 1;

     if(expt_index <= 5) // Better to read the "5" in from your own startup.in
         UpdateIvansEDFlag = 1;

     for(int dd = 0; dd < MyStartup->GetNumChannsDemux(); dd++)
      {
        Receiver[dd].ReceiveSignal(&WDMSignal,UpdateIvansEDFlag,
                                OutputPD_DataFlag,0,0,0.0);
      }


    // Regenerate a (random) initial signal for next expt.
    
    WDMSignal.GetNewRandomSignalString();
    WDMSignal.RegenerateSignalString();



   } // end for expt_index loop


  for(int dd = 0; dd < MyStartup->GetNumChannsDemux(); dd++)
    {
      Receiver[dd].PostProcessReceiverStats(0,MyStartup->GetQtMapSpans(),0);
    }

  LogFileSeparator();

   *LogFile << "THE END!!" << endl;
   cout <<   "THE END!!" << endl;

} // ## end WDM_Application

// ############################################
