// $Id:

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2006):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

/*! Written by John Zweck, June 2006
    Based on code written by Ron Holzloehner, 2003

    Class for Multicanonical Monte Carlo (MMC) simulations

Before MMC::EvaluateSample() runs, Stage = 0. 
The first sample is always accepted and Stage is set to 1.
In Stage 1, the simulation runs NumSamplesStage1 to decouple
from the initial state. The point is to avoid oversampling
the pdf where it is large. NumSamplesStage1 should be chosen
to be larger than DecouplingTime. The histogram H[]
is not incremented in this stage, so Stage 1 samples are thrown away.

Stage 2 is the production stage. NumSamplesStage2 are run and the
histogram H[] is filled. After Stage 2 is complete MMC::Iterate()
must be run. NumSamplesStage2 should be several times DecouplingTime

*/

#ifndef _MMC_OCS_HH_
#define _MMC_OCS_HH_

#ifndef _OCS_SYSTEM_HDRS_
#define _OCS_SYSTEM_HDRS_

#include <iostream.h>
#include <string.h>
#include <math.h> 
#include <fstream.h>
#include <complex.h>
#include <stdlib.h>
#include <stdio.h>
#include <strstream>

#include <string>
  using namespace std;

#endif /* _OCS_SYSTEM_HDRS_ */

#include "ocsConst.hh"
#include "ocsTools.hh"
#include "ocsReadWriteTools.hh"
#include "ocsRanNumGen.hh"

class MMC
{

public:

  MMC(unsigned DimStateSpaceIn, RanNumGen *RNGIn,
      string JobNameIn, string InFileName,bool RestartFlagIn);

  ~MMC();


  //! Chooses next trial sample in Metropolis random walk
  //! Decision whether to accept or reject will be made
  //! based on the entire sample vector z, not on its individual
  //! components.

/*! This method sets class variables z_try, LogRhoTry and RhoRatio.

    StanDevRealImag should be StanDev/sqrt(2.0).

*/

  void GaussianPerturbationSingleDecision(double StanDevRealImag);

  //! Chooses next trial sample in Metropolis random walk
  //! Decision whether to accept or reject will be made
  //! based on the individual compenents of the sample vector

/*! This method sets class variables z_try, and sets RhoRatio=1
    so that the sample will be correctly accepted or rejected in
    MMC::AcceptSample() as described in the paper of Holzloehner
    and Menyuk.

    StanDevRealImag should be StanDev/sqrt(2.0).

*/

  void GaussianPerturbationIndependentDecisions(double StanDevRealImag);


  //! 

  /*!
     Stage 0: If V_try is in range, AcceptSample() and set Stage = 1
              Else throw sample out, stay in Stage 0, and print out warning
              to LogFile. 
            
     Stage 1: Run NumSamplesStage1 to decouple from initial state and thereby
              avoid oversampling where pdf is large. NumSamplesStage1 should
              be chosen to be larger than DecouplingTime. The histogram 
              H[] is not incremented, so these samples are all thrown away.
              After NumSamplesStage1 are run, Stage 2 begins.

     Stage 2: NumSamplesStage2 are run and the histogram H[] is filled.
              NumSamplesStage2 should be several times DecouplingTime

  */

  void EvaluateSample(void);

  void Iterate(void);

private:

  void AllocateMemory(void);
  void InitializeSimulation(void);
  void InitializeIteration(void);
  
  void Gaussian_pdf(unsigned Dim,double StanDevRealImag,double *vec);

  void AcceptSample(void);

  //! Adapt Epsilon to keep AcceptanceRatio near TargetAcceptanceRatio

  void AdaptEpsilon(void);

  // JZ FIX: What if doing pdf of ones. Do we need to go right to left????

  void ModifiedUpdate(void);

  void InitOutFiles(void);
  void InitOutFilesRestart(void);

  void InitAlgorithmStatsFile(void);
  void InitVoltagePDFsFile(void);
  void InitNumHitsHistogramsFile(void);

  void WriteOutFiles(void);

  void WriteRestartFile(void);
  void ReadRestartFile(void);

  void WriteAlgorithmStatsFile(void);
  void WriteVoltagePDFsFile(void);
  void WriteNumHitsHistogramsFile(void);

  //! Returns the mean of array in[] of length len

  double Mean(unsigned int *in, size_t len);

  double Mean(double *in, size_t len);

  //! Returns the variance of array in[] of length len

  double Var(unsigned int *in, size_t length, double mean_in=NAN);

  double Var(double *in, size_t length, double mean_in=NAN);

public:

 double V_min,V_max;
 double *z_try;
 double V_try;
 bool AtEndOfStage2;
 double CurrentMaxRelError;
 double LowestPDFValue;

  //! The number of the current iteration

  int IterationNumber;

private:

  RanNumGen * RNG;
  string Job;
  bool RestartFlag;
  string CrashedJob;
  char DebugLevel;

  ofstream *MMCAlgorithmStatsFile;
  ofstream *VoltagePDFsOutFile;
  ofstream *NumHitsHistogramsOutFile;

  // DimStateSpace is total number of components of the real vector z

  unsigned DimStateSpace;
  int Stage;

  unsigned NumBins;
  int NumSamplesStage1, NumSamplesStage2;
  int InitialNumSamplesStage2;
  double NumSamplesIncrementFactor;
  double InitialEpsilon, Epsilon, TargetAcceptanceRatio;

  //! The number of the current sample in all stages of the current iteration
 
  int SampleNumber;

  //! The number of the current sample in stage 2 of the current iteration

  int SampleNumberStage2;

  //! The number of the current sample counting in all stages of all iterations

  int SampleNumberTotal;

  //! Frequency with which do a large (standard Monte Carlo) step rather than a
  //! step in a Metropolis walk. 

  int BigStepFrequency;

  bool DoBigStep;
  bool DoStdMCFlag;

  int NumAcceptedSamplesStage2;
  bool SampleAccepted; 

  double AcceptanceRatio;

  int NumSamplesOutOfBounds;

  double LogRho, LogRhoTry, RhoRatio;

  //! The index of the bin in the histogram of V that V_try falls into

  int BinIndex,BinIndexTry;

  double *z;
  double V;

  //! Array to store Histogram of number of hits in each bin of
  //! Voltage histogram. H[k] corresponds to  H^j_k in Holzloehner 
  //! and Menyuk, where j = IterationNumber 

  unsigned *H;

  //! Probability vector of size NumBins.
  //! P[k] corresponds to  P^j_k in Holzloehner and Menyuk, where
  //! j = IterationNumber

  double *P;

  //! Probability vector of size NumBins.
  //! P_old[k] corresponds to  P^{j-1}_k in Holzloehner and Menyuk

  double *P_old;

  double *g_sum, *g_hat;
  double *diffusion_stat;
 
  double Diffusion;

 
  //! Estimate of the number of steps required to decouple from current state

  double DecouplingTime;

  double nonzero_bin_fraction;
  double  rel_std_H;
  int NumPoorAcceptanceRatioLocal;
};

#endif // _MMC_OCS_HH_
