// $Id: wichmann_hill.hh,v 1.1 2007/10/24 14:39:21 zweck Exp $
//
//  wichmann_hill.hh
//  Implements the modified Wichmann-Hill random number generator
//  Copyright Ronald Holzloehner, 2001
//  Begin: 8 Dec 2001
//  Version 1.0
//
//  Create uniformly distributed random variables in [0,1)
//
//  See VPI's 'systemdesiger.pdf' manual for details
//  (Part of the TransmissionMaker documentation)
//
//  VPI claims the modified Wichmann-Hill generator has a period
//  of ca. 10^{18} !
//  According to L'Ecuyer, it should be safe to extract
//  sqrt(10^{18}) = 10^{9} high-quality random numbers

#ifndef __WICHMANN_HILL_H__
#define __WICHMANN_HILL_H__

#include <math.h>
#include <unistd.h>
#include <stdlib.h>

class Wichmann_Hill
{
  public:
    static unsigned int num_istances;
    unsigned int x, y, z;    // the variables
    unsigned int x0, y0, z0; // the seeds

    // seed the generator with given seed
    Wichmann_Hill(unsigned int x0_in, unsigned int y0_in, unsigned int z0_in);

    // use random seed
    Wichmann_Hill(void);

    // use random seed and outputs the seed used
    Wichmann_Hill(unsigned* x0_used, unsigned* y0_used,unsigned* z0_used);
    
    ~Wichmann_Hill();
    
    void SetSeeds(unsigned int x_seed, unsigned int y_seed, unsigned int
    z_seed);

    void Wichmann_Hill::check_uint(void);
    void test(unsigned int N); // test function, N: number of runs (e.g. 1e7)
    
    // the generator itself
    inline double Wichmann_Hill::rand(void)
    {
      // see the file systemdesigner.pdf for these numbers, p. 2-12
      const unsigned int a = 1284;
      const unsigned int b = 1281;
      const unsigned int c = 1276;
      const unsigned int p = 1664431;
      const unsigned int q = 1664417;
      const unsigned int r = 1664387;

      x = (a*x) % p; // x = ax mod p
      y = (b*y) % q; // y = by mod q
      z = (c*z) % r; // z = cz mod r
      // result: (x/p + y/q + z/r) mod 1.0
      return fmod(double(x)/double(p) + double(y)/double(q)
	+ double(z)/double(r), 1.0);
    }
};

#endif // __WICHMANN_HILL_H__

