// $Id:
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2006):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

/*! Written by John Zweck, June 2006
    Based on code written by Ron Holzloehner, 2003

    Class for Multicanonical Monte Carlo (MMC) simulations

*/

#include "mmc_ocs.hh"

extern ofstream *LogFile;

MMC::MMC(unsigned DimStateSpaceIn, RanNumGen *RNGIn, 
	 string JobNameIn, string InFileName, bool RestartFlagIn)
{
 
   DimStateSpace =  DimStateSpaceIn;
   RNG = RNGIn;
   Job = JobNameIn;
      
   ifstream InFile(InFileName.c_str(),ios::in);

   if (!InFile)
     {
       cerr << "Error: Input file "<< InFileName
	    << " could not be opened."<< endl;
       exit(1);
     }

  V_min = ReadDouble("MinimumVoltage"," ",1,-1,NO_BOUNDS,MANDATORY,&InFile);
  V_max = ReadDouble("MaximumVoltage"," ",1,-1,NO_BOUNDS,MANDATORY,&InFile);

  if(V_max <= V_min)
    {
      cerr <<"MaximumVoltage <= MinimumVoltage. Aborting." <<endl;
      exit(1);
    }


  NumBins = unsigned(ReadInt("NumBins"," ",0,-1,LOWER_ONLY,MANDATORY,&InFile));


  NumSamplesStage1 = ReadInt("NumSamplesStage1"," ",0,-1,LOWER_ONLY,
                              MANDATORY,1000,&InFile);


  InitialNumSamplesStage2 =  ReadInt("InitialNumSamplesStage2"," ",
				       0,-1,LOWER_ONLY,MANDATORY,1000,&InFile);


  NumSamplesIncrementFactor = ReadDouble("NumSamplesIncrementFactor"," ",1,-1,
				LOWER_ONLY,MANDATORY,1, &InFile);


  InitialEpsilon = ReadDouble("InitialEpsilon"," ",0,-1,LOWER_ONLY,
                               MANDATORY,1, &InFile);

 
  TargetAcceptanceRatio= ReadDouble("TargetAcceptanceRatio"," ",0,1,
				     LOWER_AND_UPPER,MANDATORY,0.5,&InFile);

  BigStepFrequency = ReadInt("BigStepFrequency"," ",
                              0,-1,LOWER_ONLY,MANDATORY,2*InitialNumSamplesStage2,&InFile);

  DebugLevel = char(ReadInt("DebugLevel"," ",0,4,LOWER_AND_UPPER,
                                  OPTIONAL_NO_WARNING,1,&InFile));


  DoStdMCFlag = bool(ReadInt("DoStdMCFlag"," ",
                              0,1,LOWER_AND_UPPER,OPTIONAL_WARNING,0,&InFile));


  if(DoStdMCFlag)
    {
     BigStepFrequency = 1;
     InitialNumSamplesStage2  = ReadInt("NumStdMCSamples"," ",
                                         0,-1,LOWER_ONLY,MANDATORY,1,&InFile);
    }


  RestartFlag = RestartFlagIn;

  if(RestartFlag)
    CrashedJob = ReadString("CrashedJob",&InFile) + "_";


  InFile.close();

  AllocateMemory();
  InitializeSimulation();

} // ## end Constructor

// ##################################################################

void MMC::AllocateMemory(void)
{

  z = new double[DimStateSpace];
  z_try = new double[DimStateSpace];
  P = new double[NumBins];
  P_old = new double[NumBins];
  g_sum = new double[NumBins];
  g_hat = new double[NumBins];
  H = new unsigned[NumBins];

} // ## end AllocateMemory

// ##############################################################

MMC::~MMC(void)
{

  MMCAlgorithmStatsFile->close();
  VoltagePDFsOutFile->close();
  NumHitsHistogramsOutFile->close();

  delete [] z;
  delete [] z_try;
  delete [] P;
  delete [] P_old;
  delete [] g_sum;
  delete [] g_hat;
  delete [] H;

} // ## end Destructor

// ########################################################


void MMC::InitializeSimulation(void)
{

  Epsilon = InitialEpsilon;
  NumSamplesStage2 = InitialNumSamplesStage2;
  IterationNumber = 0;
  SampleNumberTotal=0;

 for(int kk=0;kk<NumBins;kk++)
   g_hat[kk]=0;

 for(int kk=0;kk<NumBins;kk++)
   P[kk]=1.0/double(NumBins);

 for(int kk=0;kk<NumBins;kk++)
   g_sum[kk]=0.0;

 if(RestartFlag)
    ReadRestartFile();

 InitializeIteration();

 if(RestartFlag)
  InitOutFilesRestart();
 else
  InitOutFiles();

} // ## end InitializeSimulation

// ########################################################

void MMC::InitializeIteration(void)
{


 SampleNumber = 0;
 AtEndOfStage2 = false;
 SampleNumberStage2 = 0;
 NumAcceptedSamplesStage2 = 0;
 Stage = 0;
 SampleAccepted = false;
 
 LogRhoTry = 0.0;

 Diffusion = 0.0;
 DecouplingTime = 0.0;

 for(int kk=0;kk<NumBins;kk++)
   H[kk]=0;

   if(IterationNumber > 0)
  NumSamplesStage2 = int(floor(NumSamplesIncrementFactor*
                                      NumSamplesStage2));

  NumPoorAcceptanceRatioLocal = 0;
  
  NumSamplesOutOfBounds = 0;

} // ## end InitializeIteration

// ########################################################

void MMC::GaussianPerturbationSingleDecision(double StanDevRealImag)
{

  register double *zr     = z;
  register double *zr_try = z_try;

  register const double *const zr_try_end = zr_try + DimStateSpace;

  LogRhoTry = 0.0;

  DoBigStep = false;

  /* Every BigStepFrequency steps we select a standard Monte Carlo
     noise sample. However, we MUST accept/reject this sample as
     for the Metropolis walk. (I tested this by running only big
     steps. Without the accept/reject the pdf is wrong after 1st
     iteration, as the theory predicts. With the accept/reject the
     pdf is correct, though it does not of course explore the tails
     as much as we would like. By allowing big steps we explore
     the entire space more efficiently, and the decoupling time
     goes down. 
  */

  if(SampleNumber > 0 && !(SampleNumber%BigStepFrequency))
     DoBigStep=true;

  if(SampleNumber==0 || DoBigStep)
    {
 
    // Choose sample from Gaussian pdf with same variance 
    // as in a standard Monte Carlo simulation 

    Gaussian_pdf(DimStateSpace,StanDevRealImag,zr_try);

    for(; zr_try < zr_try_end; zr_try++)
      LogRhoTry += sq(*zr_try);

    }
 else
    {
      // Choose next step in Metropolis random walk using a small perturbation
      // of the current sample.
      // So we choose sample from Gaussian pdf with standard deviation
      // given by Epsilon*StanDevRealImag

     Gaussian_pdf(DimStateSpace,Epsilon*StanDevRealImag,zr_try);

     for(; zr_try < zr_try_end; zr++, zr_try++)
      {
       *zr_try = (*zr_try) + (*zr);
       LogRhoTry += sq(*zr_try);
      }
    
    }

  LogRhoTry *= -0.5/sq(StanDevRealImag);

  if(SampleNumber==0) 
   LogRho =  LogRhoTry;

  RhoRatio = exp(LogRhoTry - LogRho);

} // ## end GaussianPerturbationSingleDecision

// ########################################################

void MMC::GaussianPerturbationIndependentDecisions(double StanDevRealImag)
{

  register double *zr     = z;
  register double *zr_try = z_try;
  register const double *const zr_try_end = zr_try + DimStateSpace;
  register double  SumSqs_z_try = 0.0;

  DoBigStep = false;

  if(SampleNumber > 0 && !(SampleNumber%BigStepFrequency))
     DoBigStep=true;

  if(DoBigStep)
     cout << "DoBigStep" << endl;

  if(SampleNumber==0)
    {
   
    // Choose sample from Gaussian pdf with same variance 
    // as in a standard Monte Carlo simulation 

    Gaussian_pdf(DimStateSpace,StanDevRealImag,zr_try);

    RhoRatio = 1.0; // Irrelevant, so might as well set to 1.0

    }
 else
    {
      // Choose next step in Metropolis random walk using a small perturbation
      // of the current sample.
      // So we choose sample from Gaussian pdf with standard deviation
      // given by Epsilon*StanDevRealImag

      register unsigned NumAcceptedLocally = 0; 

     // As for SingleDecision method we must accept or reject the big steps too

     if(DoBigStep)
       Gaussian_pdf(DimStateSpace,StanDevRealImag,zr_try);
     else
       Gaussian_pdf(DimStateSpace,Epsilon*StanDevRealImag,zr_try);

     for(; zr_try < zr_try_end; zr++, zr_try++)
      {
        
	// Accept or reject each component of z_try with probability
	// given by min( Prob(z_try)/Prob(z), 1)

       register double val_try;

       if(DoBigStep)
        val_try = *zr_try;
       else 
        val_try = *zr + *zr_try;

       register double DiffSqs = sq(*zr) - sq(val_try);

       if(DiffSqs >= 0.0)
        { 
         *zr_try = val_try; 
          NumAcceptedLocally++;
        }
       else 
	 if(exp(0.5*DiffSqs/sq(StanDevRealImag)) > RNG->GetRanNum())
         {
          *zr_try = val_try; 
	  NumAcceptedLocally++;
         }
        else // reject
        *zr_try = *zr; // keep old value

        SumSqs_z_try += sq(*zr_try);

      }// for(zr_try < zr_try_end)

     if(NumAcceptedLocally > 0)
       RhoRatio=1.0;
     else
       RhoRatio=0.0;  
 
     //cout << RhoRatio << endl;
 
     const double AcceptanceRatioLocal = 
       double(NumAcceptedLocally)/double(DimStateSpace);

    // Collect statistics to verify that the second moments
    // agrees with StanDevRealImag.
    // this value should be close to 1...

    const double RelStdDev
      = (sqrt(SumSqs_z_try/double(DimStateSpace)))/StanDevRealImag;

    //...and usually not differ from 1 by more than

    const double ExpectedRelErr = 1.0 / sqrt(double(DimStateSpace));

    if(AcceptanceRatioLocal < 0.05 || AcceptanceRatioLocal > 0.95)
       NumPoorAcceptanceRatioLocal++;

    if(DebugLevel > 2 && 
       (AcceptanceRatioLocal < 0.05 || AcceptanceRatioLocal > 0.95))
    {
      *LogFile << "MMC::GaussianPerturbationIndependentDecisions: " << endl
	   << "RelativeExpectedError (Should be < 1.0): "
	   << abs(RelStdDev-1.0)/ExpectedRelErr
	   << ", PercentAccepted: "
	   << 100.0*AcceptanceRatioLocal << "%" << endl;
    }


  } // end else

} // ## end GaussianPerturbationIndependentDecisions

// ########################################################

void MMC::Gaussian_pdf(unsigned Dim,double StanDevRealImag,double *vec)
{

  if(Dim%2)
    {
      cerr << "MMC::Gaussian_pdf: Dim must be even. ABORT" << endl;
      exit(1);
    }

  double *p_vec = &(vec[0]);
  double *const p_vec_end = &(vec[Dim]);
  double v1, v2, SSs;

  for( ;p_vec < p_vec_end; )
  {
    do{
        v1 = 2.0*RNG->GetRanNum() - 1.0;
        v2 = 2.0*RNG->GetRanNum() - 1.0;
        SSs = sq(v1) + sq(v2);
      }while(SSs >= 1.0 || SSs == 0.0);

    double a =  StanDevRealImag*sqrt(-2.0 * log(SSs)/SSs);
    *p_vec = a * v1;
    p_vec++;
    *p_vec = a * v2;
    p_vec++;
  }

} // ## end Gaussian_pdf

// ##########################################################

void MMC::EvaluateSample(void)
{
  
  SampleAccepted = false;

  bool  VoltageIsInRange = (V_try >= V_min && V_try < V_max);

  if(!VoltageIsInRange)
    {

      if(DebugLevel > 1)
       {
        *LogFile << "MMC::EvaluateSample: Voltage is not in range: ";
        *LogFile << "IterationNumber = " << IterationNumber << endl;
        *LogFile << "V_try = " << V_try << endl;
       }

      NumSamplesOutOfBounds++;

      // Throw away sample and start again. Do not count this sample at all.

      return; 
    }
 
  SampleNumber++; 
  SampleNumberTotal++;

  if(Stage==2)
    SampleNumberStage2++;
 
  BinIndexTry = int(floor(NumBins*(V_try - V_min)/(V_max-V_min)));

  if(BinIndexTry < 0 || BinIndexTry >= NumBins)
     {
      cerr << "MMC::EvaluateSample(): BinIndexTry is out of range! ABORTING." 
	   << endl;
      exit(1);
     }
 
  if(SampleNumber==1)
  {
    Stage = 1;
    AcceptSample();
    return; // SampleAccepted = true here!
  }

  if(SampleNumber==NumSamplesStage1 + 1)
    {
      Stage = 2;
      SampleNumberStage2 = 1;

      if(DebugLevel > 2)
	cout << "MMC:EvaluateSample(): Entering Stage 2 of Iteration "
	     << IterationNumber << endl
             << "SampleNumber = " << SampleNumber
	     << " SampleNumberTotal = " << SampleNumberTotal << endl;

    }



    if(DoStdMCFlag)
     AcceptSample();
    else
     {

     // Work out whether to accept sample or not: 

     double P_ratio = P[BinIndex]/P[BinIndexTry];

     double pi_ab = min(P_ratio,1.0)*min(RhoRatio,1.0);

     if(pi_ab > RNG->GetRanNum())
      AcceptSample();

     }

  if(Stage==2)
    {
      H[BinIndex]++;

      if(SampleNumberStage2 == NumSamplesStage2)
	{
         AtEndOfStage2 = true;

         if(DebugLevel > 2)
           cout << "MMC:EvaluateSample(): Ending Iteration "
                << IterationNumber << endl
                << "SampleNumber = " << SampleNumber
                << " SampleNumberStage2 = " << SampleNumberStage2
                << " SampleNumberTotal = " << SampleNumberTotal << endl;

	}
    }

} // ## end EvaluateSample

// #########################################################

void MMC::AcceptSample(void)
{

  SampleAccepted = true;

  BinIndex = BinIndexTry;

  if(Stage==2)
    {
      NumAcceptedSamplesStage2++;
      Diffusion += sq(V-V_try);
    }

  V = V_try;

  for(int ii=0;ii<DimStateSpace;ii++)
    z[ii] = z_try[ii];

  LogRho = LogRhoTry; // Only required for GaussianPerturbationSingleDecision

} // ## end AcceptSample

// #########################################################

void MMC::Iterate(void)
{


  IterationNumber++;

  int TotalNumHits = 0;

  for(int kk=0;kk<NumBins;kk++)
    TotalNumHits += H[kk];

  if(TotalNumHits!= NumSamplesStage2)
    {

      cerr << "TotalNumHits = " << TotalNumHits 
           << " NumSamplesStage2 = " << NumSamplesStage2 << endl;
      cerr << "MMC::Iterate: TotalNumHits!= NumSamplesStage2. ABORT!" << endl;
      exit(1);
    }

  AcceptanceRatio = double(NumAcceptedSamplesStage2)/double(NumSamplesStage2);
  Diffusion = sqrt(Diffusion/double(NumSamplesStage2));

  // The factor of sqrt(6) comes from fact that mean square distance between
  // two randomly chosen points x,y in [0,1) is
  // int_0^1 int_0^1 (y-x)^2 dy dx = 1/6.

  // So we are asking for number of steps to diffuse over distance
  // (V_max-V_min)/sqrt(6.0)

  DecouplingTime = sq( (V_max-V_min)/(Diffusion*sqrt(6.0)) );

  // rel_std_H should be as small as possible

  double mean_H = Mean(H, NumBins);
  rel_std_H = sqrt(Var(H, NumBins, mean_H)) / mean_H;
  
  nonzero_bin_fraction = 0.0;

  for(int kk=0; kk < NumBins; kk++)
    if(H[kk] > 0)
    nonzero_bin_fraction += 1.0;
 
  nonzero_bin_fraction /= double(NumBins);
 
  if(DebugLevel)
    {

    *LogFile << "MMC::Iterate(): After iteration " 
	     << IterationNumber << ":" << endl 
	     << " SampleNumberStage2 = " << SampleNumberStage2 
	     << " AcceptanceRatio = " << 100*AcceptanceRatio << "%" 
	     << " Diffusion = " << Diffusion << endl
	     << " DecouplingTime = " << DecouplingTime
	     << " rel_std_H = " << rel_std_H
	     << " nonzero_bin_fraction = " << 100*nonzero_bin_fraction << "%"
	     << endl;
    }

  ModifiedUpdate();

  // P^{j+1} is valid beyond this point

  LowestPDFValue=P[0];

  for(int kk=1;kk<NumBins;kk++)
    if(LowestPDFValue > P[kk])
      LowestPDFValue = P[kk];

  CurrentMaxRelError = abs((P[0] - P_old[0])/P[0]);

  for(int kk=1;kk<NumBins;kk++)
    {
      double RelErrorBin = abs((P[kk] - P_old[kk])/P[kk]);

      if(CurrentMaxRelError < RelErrorBin)
         CurrentMaxRelError = RelErrorBin;
    }

  WriteOutFiles();

  AdaptEpsilon();
 
  InitializeIteration(); 

  WriteRestartFile();
 
} // ## end Iterate

// ########################################################

// inner (smaller) deviation...
#define EPSILON_LIM_1	0.05
// ...and the small correction factor
#define EPSILON_FAC_1	1.1
// out (larger) deviation...
#define EPSILON_LIM_2	0.2
// ...and the large correction factor
#define EPSILON_FAC_2	1.3

void MMC::AdaptEpsilon(void)
{
  double Epsilon_save = Epsilon;

  if(AcceptanceRatio > TargetAcceptanceRatio+EPSILON_LIM_2)
    Epsilon *= EPSILON_FAC_2;
  else if(AcceptanceRatio > TargetAcceptanceRatio+EPSILON_LIM_1)
    Epsilon *= EPSILON_FAC_1;
  else if(AcceptanceRatio < TargetAcceptanceRatio-EPSILON_LIM_2)
    Epsilon /= EPSILON_FAC_2;
  else if(AcceptanceRatio < TargetAcceptanceRatio-EPSILON_LIM_1)
    Epsilon /= EPSILON_FAC_1;
  
  if(DebugLevel)
    *LogFile  << "MMC::AdaptEpsilon:" << endl 
              << " Old Epsilon = " << Epsilon_save 
	      << " New Epsilon = " << Epsilon << endl
	      << " TargetAcceptanceRatio, AcceptanceRatio = "
	      << TargetAcceptanceRatio << " " << AcceptanceRatio << endl;

}

// #######################################################

// JZ FIX: What if doing pdf of ones. Do we need to go right to left????

void MMC::ModifiedUpdate(void)
{

  // compute P^{j+1}. First save P^{j}

  memcpy(P_old, P, sizeof(double)*NumBins); 

  P[0]=1;

  double P_Sum = P[0]; // keep old value for now

  for(int k=0; k < NumBins-1; k++)
  {
    const double H_k = double(H[k]);
    const double H_kp1 = double(H[k+1]);

    const bool H_nonzero = (H_k > 0.0 && H_kp1 > 0.0)? true:false;

    const double g_j_k = (H_nonzero)? H_k*H_kp1/(H_k+H_kp1) : 0.0; 

    g_sum[k] += g_j_k;  

    g_hat[k] = (H_nonzero)? g_j_k/g_sum[k] : 0.0; 

    P[k+1] = (P[k]*P_old[k+1]/P_old[k])*pow(H_kp1/H_k, g_hat[k]); 
      
    // Limit the ratio between neighboring P's
    // If this warning comes on, DeltaV is too large

    P_Sum += P[k+1]; // k=0 already taken care of above
   
  } // end for k
  
  
  // P_Sum > 1 as P[0]=1

  for(int k=0;k<NumBins;k++)
    P[k] /= P_Sum;

} // ## end ModifiedUpdate

// ####################################################

void MMC::InitOutFiles(void)
{

  InitAlgorithmStatsFile();
  InitVoltagePDFsFile();
  InitNumHitsHistogramsFile();

} // ## end InitOutFiles

// ##################################################

void MMC::InitOutFilesRestart(void)
{

 string MMCAlgorithmStatsOutFileName = Job + "MMCAlgorithmStats.dat";

 MMCAlgorithmStatsFile = new ofstream(MMCAlgorithmStatsOutFileName.c_str(),
                                      ios::app);


 *MMCAlgorithmStatsFile << endl << "%RESTARTING" << endl << endl;

 MMCAlgorithmStatsFile->flush();  

 string VoltagePDFsOutFileName = Job + "VoltagePDFs.dat";

 VoltagePDFsOutFile = new ofstream(VoltagePDFsOutFileName.c_str(),
                                   ios::app);

 VoltagePDFsOutFile->setf(ios::scientific);

 *VoltagePDFsOutFile << endl << "%RESTARTING" << endl << endl;

 VoltagePDFsOutFile->flush(); 

 string NumHitsHistogramsOutFileName = Job + "NumHitsHistograms.dat";

 NumHitsHistogramsOutFile = new ofstream(NumHitsHistogramsOutFileName.c_str(),
                                         ios::app);

 *NumHitsHistogramsOutFile << endl << "%RESTARTING" << endl << endl;

 NumHitsHistogramsOutFile->flush();

} // ## end InitOutFilesRestart


// ###################################################

void MMC::InitAlgorithmStatsFile(void)
{

 string MMCAlgorithmStatsOutFileName = Job + "MMCAlgorithmStats.dat";

 MMCAlgorithmStatsFile = new ofstream(MMCAlgorithmStatsOutFileName.c_str(),
				  ios::out);

 *MMCAlgorithmStatsFile
   << "% (1) IterationNum (2) Total NumSamples (3) NumSamplesStage1"
   << " (4) NumSamplesStage2 (5) DecouplingTime (6) AcceptanceRatio" 
   << endl
   << "% (7) Epsilon (8) RelStdDev_H (9) NonZeroBinFraction"
   << " (10) log10(lowest pdf value) (11) CurrentMaxRelError"
   << " (12) V_min (13) V_max (14) NumBins"
   << endl
   << "% (15) NumSamplesOutOfBounds_Itn (16) NumPoorARL_Itn"
   << endl;

  MMCAlgorithmStatsFile->flush();

}

// ##################################################

void MMC::InitVoltagePDFsFile(void)
{

 string VoltagePDFsOutFileName = Job + "VoltagePDFs.dat";

 VoltagePDFsOutFile = new ofstream(VoltagePDFsOutFileName.c_str(),
				   ios::out);
 
 VoltagePDFsOutFile->setf(ios::scientific);

 *VoltagePDFsOutFile << "% j-th row is P^j, starting from j=0" << endl;

 *VoltagePDFsOutFile << "% Next Row: P^0:"  << endl;

   for(int kk=0;kk<NumBins;kk++)
     *VoltagePDFsOutFile << P[kk] << "\t";

 *VoltagePDFsOutFile << endl;

 VoltagePDFsOutFile->flush();

}

// #####################################################


void MMC::InitNumHitsHistogramsFile(void)
{

 string NumHitsHistogramsOutFileName = Job + "NumHitsHistograms.dat";

 NumHitsHistogramsOutFile = new ofstream(NumHitsHistogramsOutFileName.c_str(),
					 ios::out);
 
 *NumHitsHistogramsOutFile << "% j-th row is H^j, starting from j=0" << endl;

 *NumHitsHistogramsOutFile << "% Next Row: H^0:"  << endl;

   for(int kk=0;kk<NumBins;kk++)
     *NumHitsHistogramsOutFile << H[kk] << "\t";

 *NumHitsHistogramsOutFile << endl;

 NumHitsHistogramsOutFile->flush();

}

// #####################################################

void MMC::WriteOutFiles(void)
{

  WriteAlgorithmStatsFile();
  WriteVoltagePDFsFile();
  WriteNumHitsHistogramsFile();

} // ## end WriteOutFiles

// ####################################################

void MMC::WriteAlgorithmStatsFile(void)
{
  
 *MMCAlgorithmStatsFile
   <<  IterationNumber << "\t" 
   <<  SampleNumberTotal << "\t" 
   <<  NumSamplesStage1 << "\t"
   <<  NumSamplesStage2 << "\t"
   <<  DecouplingTime  << "\t" 
   <<  AcceptanceRatio << "\t" 
   <<  Epsilon  << "\t" 
   <<  rel_std_H  << "\t" 
   <<  nonzero_bin_fraction << "\t" 
   <<  log10(LowestPDFValue) << "\t"  
   <<  CurrentMaxRelError << "\t" 
   <<  V_min << "\t" 
   <<  V_max << "\t" 
   <<  NumBins << "\t"
   <<  NumSamplesOutOfBounds << "\t"
   <<  NumPoorAcceptanceRatioLocal << endl;

  MMCAlgorithmStatsFile->flush();

}

// #####################################################

void MMC::WriteVoltagePDFsFile(void)
{

  *VoltagePDFsOutFile << "% Next Row: P^"  << IterationNumber << ":" << endl;

   for(int kk=0;kk<NumBins;kk++)
     *VoltagePDFsOutFile << P[kk] << "\t";

 *VoltagePDFsOutFile << endl;

 VoltagePDFsOutFile->flush();

}
// #####################################################

void MMC::WriteNumHitsHistogramsFile(void)
{

  *NumHitsHistogramsOutFile << "% Next Row: H^"  
			    << IterationNumber << ":" << endl;

   for(int kk=0;kk<NumBins;kk++)
     *NumHitsHistogramsOutFile << H[kk] << "\t";

 *NumHitsHistogramsOutFile << endl;

 NumHitsHistogramsOutFile->flush();

}

// ###################################################

void MMC::WriteRestartFile(void)
{

 string RestartFileName = Job + "Restart.dat";

 ofstream * RestartFile = new ofstream(RestartFileName.c_str(),ios::out);

 RestartFile->setf(ios::scientific);

 *RestartFile << IterationNumber << " " 
              << Epsilon << " " 
	      << SampleNumberTotal << " "
              << NumSamplesStage2 << endl;

 for(int kk=0;kk<NumBins;kk++)
  *RestartFile << P[kk] << "\t";

 *RestartFile << endl;

  for(int kk=0;kk<NumBins;kk++)
  *RestartFile << g_sum[kk] << "\t";

 *RestartFile << endl;


 RestartFile->flush();

 RestartFile->close();

 RNG->WriteRanSeqParameters(Job + "RestartRNG.dat");

} // end MMC::WriteRestartFile

// #################################################

void MMC::ReadRestartFile(void)
{

 string RestartFileName = CrashedJob + "Restart.dat";

 ifstream * RestartFile = new ifstream(RestartFileName.c_str(),ios::in);

 RestartFile->setf(ios::scientific);
  
 *RestartFile >> IterationNumber 
              >> Epsilon 
              >> SampleNumberTotal 
              >> NumSamplesStage2;

 for(int kk=0;kk<NumBins;kk++)
  *RestartFile >> P[kk];

  for(int kk=0;kk<NumBins;kk++)
  *RestartFile >> g_sum[kk];

 RestartFile->close();

 *LogFile << "MMC::ReadRestartFile" << endl
          << "IterationNumber = " << IterationNumber
	  << " Epsilon = " << Epsilon 
          << " SampleNumberTotal = " << SampleNumberTotal
          << " NumSamplesStage2 = " << NumSamplesStage2 << endl;

 LogFile->flush();


 RNG->ReadRanSeqParameters(CrashedJob + "RestartRNG.dat");

} // end MMC::ReadRestartFile

// ####################################################

double MMC::Mean(unsigned int *in, size_t len)
{
  const  unsigned int *const p_end = in+len;

  register const unsigned int *p = in;
  register double mean_l = 0.0;

  for(; p<p_end; p++)
    mean_l +=(double) *p;

  mean_l /= double(len);

  return mean_l;

}

// ####################################################

double MMC::Mean(double *in, size_t len)
{
  const  double *const p_end = in+len;

  register const double *p = in;
  register double mean_l = 0.0;

  for(; p<p_end; p++)
    mean_l += *p;

  mean_l /= double(len);

  return mean_l;

}

// #######################################################

//! Returns the variance of array in[] of length len

/*! The computation of \sum (x_i - <x>)^2 is better than
    <x^2> - <x>^2 since roundoff errors are smaller
*/

double MMC::Var(unsigned int *in, size_t len, double mean_in)
{

  register double mean_l = (isnan(mean_in))? Mean(in, len) : mean_in;

  const unsigned int *const p_end = in+len;
  register const unsigned int *p = in;
  register double var_l = 0.0;

  for(; p<p_end; p++)
    var_l += sq(*p - mean_l);
  
// yes, this is really divided by length-1, not length  (unbiased mean)

  var_l /= double(len-1);

  return var_l;

}

// ############################################################

double MMC::Var(double *in, size_t len, double mean_in)
{

  register double mean_l = (isnan(mean_in))? Mean(in, len) : mean_in;

  const double *const p_end = in+len;
  register const double *p = in;
  register double var_l = 0.0;

  for(; p<p_end; p++)
    var_l += sq(*p - mean_l);

// yes, this is really divided by length-1, not length  (unbiased mean)

  var_l /= double(len-1);

  return var_l;

}

