// $Id

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2004):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

// Written by John Zweck, May 2004

#ifndef _CHI_SQUARE_RECEIVER_HH_
#define _CHI_SQUARE_RECEIVER_HH_

#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_roots.h>

#include "NoiseCovariance.hh"
#include "ocsOptDemuxer.hh"
#include "ocsPhotodetector.hh"
#include "ocsElecFilter.hh"
#include "ocsHistogram2D.hh"
#include "ocsElecSignalStat.hh"

//! Structure required to pass information to GSL root finding algorithm

struct ChiSquareReceiverParameters
{
  double ActiveCurrentValue;
  int CovMatrixDim;
  double * LambdaVec;
  double * LambdaQSqVec;
};


cplx ChiSquareReceiverPsiCplx(cplx s, void * InParameters);
double ChiSquareReceiverPsi(double s, void * InParameters);
double ChiSquareReceiverDerivativePsi(double s, void * InParameters);
double ChiSquareReceiverSecondDerivativePsi(double s, void * InParameters);


//! Enumeration type that specifies sort of statistics of electrical pdf
//! to be generated by the class
/*!
    Possible values:

    - PDFS = 1, i.e., pdfs in each bit at clock recovery 
                      together with average pdfs of marks and spaces
    - EYE = 2, i.e., a contour eye diagram 

*/

enum typeOutput { PDFS = 1, EYE = 2};

//! Enumeration type that specifies how to calculate the integral
//! to obtain the value of the pdf at a given value of the current
//! from the characteristic function
/*!
    Automatically set by code. 
  
    Possible values:

    - FOURIER_TRANSFORM = 1 For now we don't seem to need this option
                            So not implemented yet
    - STEEPEST_DESCENT = 2
*/

enum typeIntegrationMethod { FOURIER_TRANSFORM = 1, STEEPEST_DESCENT = 2};

// For most compliers Fortran77 subroutines like myfun are recognized
// as myfun_ Therefore:

#define F77NAME(x) x##_

extern "C" void F77NAME(dsyev)(char*,char*,int*,double*,int*,double*,double*,
			       int*,int*);
// The last argument is an int which is the length of the
// character array name.

extern "C" int F77NAME(ilaenv)(int*,char*,char*,int*,int*,int*,int*,int);

//! Given a noise free signal, covariance matrix and receive filters
//! this class computes the PDFs of the current in each bit and the 
//! bit-error ratio

/*!

  This code was written and documented by John Zweck, June 2004
  Please report all bugs to zweck@math.umbc.edu

  The code is based on the receiver model in Ronald Holzloehner's thesis.
  However, rather than computing the inverse Fourier transform of the
  characteristic function of the pdf in each bit using a DFT, we use the
  method of steepest descent along the same lines as in the paper of
  Forestieri that is referenced in Ron's thesis. Actually Forestieri
  explains how to compute the BER, which is a different problem from
  that of computing the PDFs. I adapted Forestieri's method
  to compute the PDFs. I have handwritten notes explaining the theory
  available upon request. 

  Currently, the noise-free signal and the covariance matrix must be
  of SCALAR rather than VECTOR type, i.e. we don't have polarization
  effects yet. (That's the next step.) 

  In addition, if you are doing WDM simulations I recommend using
  EvenlySpacedChannels = 1 in the Signal.in file. The receiver will
  receive the channel at the center frequency. In general it will receive 
  the channel with ChannelPhysicalIndex = 0, and it assumed that this channel
  is at the center frequency. 

  To use the class you must also use an instance of the NoiseCovariance
  class, to produce a noise-free signal and a noise covariance matrix
  with either AWGN or coloured noise.

  The following code fragments should be used in your application code:

  -NoiseCovariance * NoiseCov;

  -NoiseCov = new NoiseCovariance(InDir + "NoiseCovariance.in",Job,Signal);

  -ChiSquareReceiver *  ChiSqRx;
 
  -ChiSqRx = new ChiSquareReceiver("ChiSquareReceiver.in",NoiseCov);

  -Calls to propagate noise covariance matrix and/or noise-free signal
  through system go here

  -string Suffix = "";

  -NoiseCov->WriteFileNoiseFreeSignalAndCovarianceMatrix(Suffix);

  -ChiSqRx->ReceiveSignal(Suffix);

  -cout << "OSNR [dB] = " << ChiSqRx->GetOSNR_dB(B_OSA)
        << endl;
  -cout << "BER = " << ChiSqRx->GetBitErrorRatio() << endl;
  -cout << "Decision Threshold [mV] = " << ChiSqRx->GetDecisionThresholdmV()
        << endl;

  -ChiSqRx->~ChiSquareReceiver();
  -NoiseCov->~NoiseCovariance();
  
  You can use the receiver in a for loop, say with different realizations
  of polarization effects, or different amounts of noise, or different
  receiver filters etc. Just make the changes you need to the system
  at the start of the for loop, change the Suffix string if you like
  and then call ChiSqRx->ReceiveSignal(Suffix).

  Here is an explanation of the input parameter file ChiSquareReceiver.in
  together with a discussion of the capabilities of the class:

  TypeOutput: 

    - Possible values:

    - PDFS = 1, i.e., pdfs in each bit at clock recovery 
                      together with average pdfs of marks and spaces
                      and bit-error ratio. 
    - EYE = 2, i.e., computes a contour eye diagram 


  ReadInSignalAndCovMatrixFlag:

    - Flag to indicate whether or not to read in the ReducedNoiseFreeSignal
      and CovarianceMatrix from files that were previously written by
      NoiseCovariance::WriteFileReducedNoiseFreeSignalAndCovarianceMatrix()
    - This is useful when it takes a long time to propagate through
      the transmission line, and you want to do pass the same noise-free 
      signal and covariance matrix through the receiver several times
      for debugging or parameter studies.
    - Also, the NoiseCovariance.in and Signal.in files must be
      identical to the job in which the signal and cov matrix were
      written out!!
    - If this flag is set to 1 then you must correctly set   
      SignalAndCovMatrixDir and SignalAndCovMatrixJob, eg.
     -$SignalAndCovMatrixDir  /home/zweck/OCS/noise/expts/May04/
     -$SignalAndCovMatrixJob  T1

  SetOSNRdB:

     - Suppose you have a noise-free signal obtained either by
       using the NoiseCovariance class to propagate through a system
       or by reading in a signal as explained above. Also suppose you
       want AWGN noise with a given OSNR. 
     - If so: set SetOSNRdB to the
       desired OSNR in dB. Here we define 
     - Can be used in conjunction with ReadInSignalAndCovMatrixFlag
     - Here we define OSNR = Linear2dB(SignalPower/NoisePower) where
     - SignalPower = Signal->GetAveragePower()/NumChannels
     - NoisePower = 
       Signal->GetAccumulatedTotalNoiseSpectralDensitySemiAnalytical()*B_OSA
     - The value of B_OSA is set in in input parameter file, in Hz, and is
       the power equivalent spectral bandwidth of the OSA. See Ivan Lima's
       JLT paper or his thesis on receiver modeling.

  NumBitsToPlot:   

    - The number of bits that contribute to the pdfs or eyes.
    - If this input parameter is set to 0 (default) then NumBitsToPlot is
    initialized to Signal->GetStringLength();

  
   NumBitsBetweenSlices:  

    -The number of bits between slices for TypeOutput == PDF

  NumSlicesPerBit:

    -The number of slices per bit for TypeOutput == EYE

  MinCurrent:

    - The minimum value of the current for the PDFs or EYE. Negative
      values are allowed

  MaxCurrent: 
    
    - The maximum value of the current for the PDFs or EYE

  NumPointsCurrent:

    -  Number of points for discretization of current 
       in the discrete pdfs and eye

  ReadInClockRecoveryTimeFlag:

   - If value = 1 then reads in ClockRecoveryTime from input parameter
     file. Otherwise, class computes ClockRecoveryTime automatically
     in method ChiSquareReceiver::ReceiveNoiseFreeSignalAndRecoverClock()
  
  DebugLevel:
   
   - Higher means more output. Possible values 0,1,2,3
   - Default is 0, (no debugging output)

  Finally, what tests are in the code to ensure it is working correctly.

  - The first part of the code computes the KL modes. This involves doing
    a couple of eigendecompositions of matrices of the form A = U^T D U.
    We check that A - U^T D U and U^T U - I have 
    Relative Inf Norm < 1e-10. If not then we write a WARNING to cout.
  - We also check that (4.6.8) of Ron's thesis 
    with q_k = 0 holds up to a relative error of 1e-6
  - The second part of the code computes the integral of the characteristic
    function, i.e. the integral (4.6.13) of Ron's thesis (in case of zero
    timing jitter \tau for now) using the method of steepest descent. 
  - If the method fails for a particular value of the time and 
    current (see below) 
    then we output an error message to cout and set the value of the PDF
    in for that time and current value  to be 0. If we need to later
    we will implement the FOURIER_TRANSFORM method for computing the pdf
    in the cases of failure. 
  - Here is how steepest descents can fail. 
  - First for a given current, time value the code numerically finds
    the zero of a particular function using brent algorithm, which is 
    closely related to the bisection method. It must start with an interval
    for which the signs of the function are different at the two endpoints
    of the interval. If not we get a ZERO CROSSING error message to cout.
    For positive values of the current this should never happen, according 
    to my theory. 
  - Second, there are stringent conditions for success in the root finding
    algorithm, which finds a root by narrowing the interval in which it is 
    contained. First we insist that the width of the interval in which a root
    is gaurenteed to lie is small, second we insist that the residual 
    value of the function at the final iterate is very close to zero. 
    If we don't meet
    these tests then an error message is written to cout, and the PDF value
    is set to 0. See  ChiSquareReceiver::FindRootInterval for details.
  - The value of the pdf is obtained by evaluating another function
    at the root we have found numerically. We test to see that for the
    final interval, the relative error in the value of this function
    evaluated at the left and right endpoints of the final interval
    found by the root finding algorithm is very small. If it fails
    this test  an error message is written to cout, and the PDF value
    is set to 0. See  ChiSquareReceiver::FindRootInterval for details.

  Future improvements:

   - Validate against MMC (soon to be done)
   - Extend to two polarization states
   - Add in timing jitter (4.6.16) of Ron's thesis
   - If required add the fourier transform method to deal with cases
     that steepest descent fails

*/

class ChiSquareReceiver
{

 public:

  //! Constructor. 
  /*!
      - InFileName is name of input parameter file 
      - oNoiseCovariance2 is NoiseCovariance class that must
         be passed to this class, which contains data for
         NoiseFreeSignal (aka Signal) and Covariance Matrix of noise.
         There are methods in NoiseCovariance class that allow one to
         easily obtain a diagonal noise covariance matrix from
         an OSNR value, noise spectral density, or AWGN noise propagation
         model.
  */

  ChiSquareReceiver(string InFileName, NoiseCovariance * oNoiseCovariance2);

  ~ChiSquareReceiver(void);

  //! Call this method each time you want to receive the signal.
  /*!
     - For now code only works to receive channel that is at center freq
       (at least in case have evenly spaced freqs)
     - OutFileNamePrefix2 will be prepended to all output files generated
        by a call to this method.

     You can call this method multiple times during the course of a 
     simulation with different data in oNoiseCovariance2
  */

  void ReceiveSignal(string OutFileNamePrefix2);
 
  void ComputeGaussianPDFsAndBER(void);
 
  double GetOSNR_dB(double B_OSA);
  double GetBitErrorRatio(void) {return BitErrorRatio;};
  double GetDecisionThresholdmV(void) 
         {return 1e3*DecisionThreshold/NormalizationFactor;};

  //! Sets the required OSNR for AWGN noise.
  //! Call before calling ChiSquareReceiver::ReceiveSignal
  //! Can be used in for loops. 

  /*! Resets OSNRdB to Value. Must have set  SetOSNRdB to be nonzero
      in input file, so B_OSA is read in */


  void SetOSNR(double Value);

  //! Sets the electrical filter f_3dB bandwidth. Useful in for loops
  //! Call before calling ChiSquareReceiver::ReceiveSignal

  void SetElecFilterBandwidth(double ValueHz);

  //! Sets the optical filter FWHM bandwidth. Useful in for loops
  //! Call before calling ChiSquareReceiver::ReceiveSignal

  void SetOptFilterBandwidth(double ValueHz);

  //! This method uses an analytic formula for the characteristic
  //! function of the pdf of the current in the case of a CW signal
  //! propagated through a fiber with no loss and no dispersion.
  //! It is included in the code for debugging purposes.
  //! The formula was derived by J.Z. based on work of Hui and others. 

  void ReceiveSignalCWNoLossNoDisp(string OutFileNamePrefix2,
                        double P_0, double z, double gamma, double N_ASE, double NF);
 

  //! For debugging purposes only

  double GetNormalizationFactor() {return NormalizationFactor;};
  void SetNormalizationFactor(double Value) {NormalizationFactor = Value;};

private:
  void ReceiveNoiseFreeSignalAndRecoverClock(void);
  void PrepareToComputeKLModes(void);

  void GetReducedNoiseFreeSigAndDiagonalizeCovMatrix(void);
  void ComputeTimeIndependentPartOfFilterMatrix(void);
  
  void ComputeKLModes(int TimeIndex);

  void ComputePDFAtTimeValue(int TimeIndex);
  void ComputePDFAtTimeValueFourierTransform(int TimeIndex);  
  double ComputePDFValueSteepestDescent(int TimeIndex,int CurrentIndex);
  double EvaluateGaussianApproximation(double SaddlePoint);

  void WriteFilePDFs(void);

  //! A one-bit eye is written to a file so that an image of the output
  //! matrix looks like the eye. Signal::BitLength, MinCurrent, 
  //! and MaxCurrent are hidden in columns 1,2,3 of the top row.

  void WriteFileEye(void);

 //! Computes the average PDFs of the ZEROs and ONEs and stores
  //! results in arrays ChiSquareReceiver::AveragePDFZeros and 
  //! ChiSquareReceiver::AveragePDFOnes

  /*! This method identifies which time slices are ZEROs and which
   are ONEs by choosing an optimal bit shift which maximizes the 
   correlation between the BitString 
   vector and the vector containing the electrical current at each
   time in the vector TimeSlices. This has only been debugged when
   the channel is at the center frequency (in which case there is
   not much to do!) 
  */

  void ComputeAveragePDFs(void);

  //! Computes the decision threshold that minimizes the bit-error ratio
  //! together with the corresponding BER using the average PDFs of the
  //! marks and spaces. 
  
  void ComputeMinimumBERAndDecisionThreshold(void);

  //! Computes the mean and standard deviation of the current
  //! in each bit.

  void ComputeMomentsInEachBit(void);

  //! Computes EyeMatrix, which stores eye for one bit. Use matlab
  //! to get a double eye.

  void ComputeEye(void);

  //! Interface to LAPACK subroutine DSYEV which computes the
  //! eigenvalues and vectors of a real symmetric Dim x Dim matrix, Matrix

  /*! We assume that Matrix is symmetric

      Matrix = EigVector*Diag(EigValue)*EigVector^T

      The columns of EigVector are the eigenvectors of Matrix
      The eigenvalues are in ascending order

      So in Matlab's notation: [EigVector,EigValue] = eig(Matrix)
  */

  void SymmetricEig(double ** Matrix, double * EigValue,
		    double ** EigVector,int Dim);

  //! Called once to allocate memory for work array for SymmetricEig

  void PrepareSymmetricEig(void);

  //! If SymmetricEig computes the decomposition U^T A U = Diag, U^T U = I
  //! then this method checks how well these two equations hold
  //! using InfNormMatrix. If InfNorm(U^T A U - Diag) relative to
  //! InfNorm(A) or InfNorm(U^T U - Id) relative to InfNorm(A)
  //! exceeds 1e-10 then eigen decomposition fails test.
  //! Returns 1 if passes test, 0 if fails test.

  bool TestEigDecomposition(double ** Matrix, double * EigValue,
		            double ** EigVector,int Dim);

  //! The k-th ReducedFrequency in Hz. Goes from -ReducedFreqWindow/2
  //! to ReducedFreqWindow/2 - DeltaFreq

  inline double ReducedFrequency(int k) 
                {return (CovMatrixFirstIndex + k)*DeltaFreq;};

  void WriteFileSquareMatrix(double **A, int dim, string OutFileName, 
	  	             double NormalizationFactor);

  void WriteFileVector(double *V, int dim, string OutFileName, 
	  	   double NormalizationFactor);

  void WriteFilePairVectors(double * x, double *y, 
                            int dim,string OutFileName, 
			    double xNormalizationFactor, 
			    double yNormalizationFactor);

  void WriteFileComplexMatrix(cplx **A, int dim,
			      string RealOutFileName, 
			      string ImagOutFileName, 
			      double NormalizationFactor);

  //! The columns of the output file are the pdfs at the different TimeSlices
  //! So the output matrix is NumPointsCurrent x NumTimeSlices

  void WriteFilePDFTimeSlices(string OutFileName);

  void WriteFileGSLFunction(double CurrentValue,double s_min,double s_max,
			    int NumPts,string OutFileName);

  void InitializeGSL_FunctionToDerivativePsi(void);
  double FindRootInterval(double &LowerBound,double &UpperBound);


 private:

  string Job;
  string OutFileNamePrefix;

  typeIntegrationMethod TypeIntegrationMethod;

  //! Flag to indicate whether or not to read in the ReducedNoiseFreeSignal
  //! and CovarianceMatrix from files that were previously written by
  //! NoiseCovariance::WriteFileReducedNoiseFreeSignalAndCovarianceMatrix()

  bool ReadInSignalAndCovMatrixFlag;

 
  //! Directory name for the files containing ReducedNoiseFreeSignal
  //! and CovarianceMatrix \sa ReadInSignalAndCovMatrixFlag

  string SignalAndCovMatrixDir;

  //! Job name prefix for the files containing ReducedNoiseFreeSignal
  //! and CovarianceMatrix \sa ReadInSignalAndCovMatrixFlag

  string SignalAndCovMatrixJob;

  //! Input parameter. Default value = 0. If value is nonzero then
  //! resets CovarianceMatrix to be the multiple of the identity matrix
  //! that with the NoiseFreeSignal gives the OSNR whose value in dB is
  //! SetOSNRdB. Only works when NoiseCovariance::TypeNoiseMethod == AWGN.

  double SetOSNRdB;

  //! Noise equivalent bandwidth of the optical spectrum analyzer used
  //! to set the OSNR for AWGN noise \sa SetOSNRdB

  double B_OSA;

  //! Higher means more output. Possible values 0,1,2,3
  //! Default is 0, (no debugging output)

  int DebugLevel;

  int DemuxChannArrayIndex;
  NoiseCovariance * oNoiseCovariance;  
  OptSignal * Signal;
  
  fftMatrices ReceiverReducedNoiseFreeSignal;
  double * ReducedNoiseFreeSignalRealVector;

  RanNumGen * RNG;

  OptSignal * DemuxedNoiseFreeSignal;
  OptFilter * TunableOptFilterForNoiseFreeSignal;
  OptDemuxer * OptDemuxerForNoiseFreeSignal;
  Photodetector * PhotodetectorForNoiseFreeSignal;
  ElecFilter * ElecFilterForNoiseFreeSignal;
  ElecSignalStat * ElecSignalStatForNoiseFreeSignal;

  OptSignal * OptFilterTransferFunction;
  Photodetector * ElecFilterTransferFunction;

  OptFilter * TunableOptFilterForTransferFunction;
  ElecFilter * ElecFilterForTransferFunction;

  double DeltaFreq; // equals ReducedDeltaFreq
  double TimeWindow; // equals ReducedTimeWindow
  double ReducedDeltaTime;
  double ReducedFreqWindow;

  //! Real covariance matrix of size 
  //! 2*NoiseCovariance::CovMatrixDim x 2*NoiseCovariance::CovMatrixDim
  /*!
      This is the script K in Ron's papers/thesis
  */

  double **CovarianceMatrix;

  typeOutput TypeOutput;

  //! The number of bits that contribute to the pdfs or eyes
  //! If this input parameter is set to 0 (default) then NumBitsToPlot is
  //! initialized to Signal->GetStringLength();

  int NumBitsToPlot;
  
  //! The number of bits between slices for TypeOutput == PDF

  int NumBitsBetweenSlices;

  //! The number of slices per bit for TypeOutput == EYE

  int NumSlicesPerBit;

  //! The time interval between the times of the pdfs or the time 
  //! slices of the eye

  /*! Units = seconds

     Default Values:

      - If TypeOutput = PDFS the default value is Signal->GetBitPeriod()
      - If TypeOutput = EYE the default value is Signal->GetDeltaTime()

  */

  double TimeBetweenSlices;


  //! If value = 1 then read in ClockRecoveryTime from input parameter
  //! file. Otherwise, class computes ClockRecoveryTime automatically
  //! in method ChiSquareReceiver::ReceiveNoiseFreeSignalAndRecoverClock()

  bool ReadInClockRecoveryTimeFlag;

  //! ClockRecoveryTime \sa ReadInClockRecoveryTimeFlag

  double ClockRecoveryTime;

  //! The ClockCurrent from ElecSignalStat::GetClockCurrent for
  //! the noise free signal

  double ClockRecoveryCurrent;

  int NumTimeSlices;

  //! Array of size NumTimeSlices to hold the times of the time slices.

  double * TimeSlices;

  //! Minimum value of current in the discrete pdfs and eye (Volts)

  double MinCurrent;
 
  //! Maximum value of current in the discrete pdfs and eye (Volts)

  double MaxCurrent;

  //! Number of points for discretization of current 
  //! in the discrete pdfs and eye

  int NumPointsCurrent;

  double DeltaCurrent;

  //! Array of size NumPointsCurrent
  //! to store the discrete current values used in pdfs and eye

  double * CurrentVector;

  //! Pointer to the active (current) element of the array CurrentVector
  /*! Used by friend functions
   */

  double * ActiveCurrentValuePtr;

  int CovMatrixDim;
  
  int CovMatrixFirstIndex;
  int CovMatrixLastIndex;

  double NormalizationFactor;

  double LowerPole;
  double UpperPole;

  double DecisionThreshold;
  double BitErrorRatio;

  int LengthSymmetricEigWork;

  double * SymmetricEigWork;

  //! Array to store eigenvalues of CovarianceMatrix

  double * CovarianceEvalues;

  //! Real matrix of eigenvectors of CovarianceMatrix with size 
  //! 2*CovMatrixDim x 2*CovMatrixDim 
  
  double ** CovarianceEvectors;

   //! Actual array used to store eigenvectors of CovarianceMatrix
  /*! \sa double ** NoiseCovariance::CovarianceEvectors
   */

  double * CovarianceEvectorsStorageVec;
  
  //! Complex matrix of size CovMatrixDim x CovMatrixDim.

  cplx ** FilterMatrixTimeIndependent;

  //! Actual array used to store FilterMatrixTimeIndependent

  cplx * FilterMatrixTimeIndependentStorageVec;

    //! Real matrix of size 2*CovMatrixDim x 2*CovMatrixDim.
    /*! This matrix is {\cal W} (script W) defined in (4.6.5b)
        of Ron Holzloehner's thesis
    */

  double ** FilterMatrix;

  //! Actual array used to store FilterMatrix

  double * FilterMatrixStorageVec;

  //! The matrix H in (A.3.1g) of Ron Holzloehner's thesis

  double ** HMatrix;

  //! Actual array used to store HMatrix

  double * HMatrixStorageVec;

  //! The matrix V in (A.3.1f) of Ron Holzloehner's thesis

  double ** VMatrix;

  //! Actual array used to store VMatrix

  double * VMatrixStorageVec;

  //! Temporary storage used in ComputeKLModes()

  double ** WorkMatrix;

  //! Actual array used to store TempMatrix

  double * WorkMatrixStorageVec;

    //! Temporary storage used in TestEigDecomposition

  double ** TestEigMatrix;

  //! Actual array used to store TestEigMatrix

  double * TestEigMatrixStorageVec;


  //! The vector Lambda of (A.3.1b) in Ron's thesis

  double * LambdaVec;

  //! The vector whose kk entry is Lambda[kk]QSq[kk] 
  
  double * LambdaQSqVec;

  //! WorkVec has size 2*CovMatrixDim

  double * WorkVec;

  gsl_function * GSL_Function;
  int RootFinderMaxIterations;

  double MinNegLambda;

  struct ChiSquareReceiverParameters * Parameters;

  // PDFTimeSlices[tt][vv], so tt-th row of PDFTimeSlices is the pdfs of
  // the tt-th TimeSlice. We write out its transpose though

  double ** PDFTimeSlices;
 
  //! Actual array used to store PDFTimeSlices

  double * PDFTimeSlicesStorageVec;

  //! Array to store average PDF of the zeros.

  double * AveragePDFZeros;

  //! Array to store average PDF of the ones.

  double * AveragePDFOnes;

  //! Used in ComputeAveragePDFs

  double * PowerInSlice;

  //! Used in ComputeAveragePDFs

  int * BitString;

  //! Matrix to store the eye diagram, EyeMatrix[tt][vv] is of size
  //! NumSlicesPerBit x NumPointsCurrent

  /*! To help with plotting, we hide Signal::BitLength, MinCurrent,
      and MaxCurrent in this matrix
      \sa ChiSquareReceiver::WriteFileEye
  */

  double ** EyeMatrix;

  //! Actual array used to store EyeMatrix

  double * EyeMatrixStorageVec;

  //! Array to store BER as function of Decision threshold

  double * BER;

  double * MomentsIntegrand;

  double * MeanCurrent;

  double * StandardDeviationCurrent;

};

#endif
