// $Id:

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2006):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

// Written by John Zweck, Jan 2006


#ifndef _OCS_GAUSSIAN_RECEIVER_HH_
#define _OCS_GAUSSIAN_RECEIVER_HH_

#include "ocsOptSignal.hh"
#include "ocsOptFilter.hh"
#include "ocsOptDemuxer.hh"
#include "ocsPhotodetector.hh"
#include "ocsElecFilter.hh"
#include "ocsElecSignalStat.hh"
#include "ocsOptDataWriter.hh"
#include "ChiSquareReceiver.hh"
#include "ocsOptFiber.hh"

/*

From ChiSquareReceiver.hh:

Enumeration type that specifies sort of statistics of electrical pdf
to be generated by the class

    Possible values:

    - PDFS = 1, i.e., pdfs in each bit at clock recovery 
                      together with average pdfs of marks and spaces
    - EYE = 2, i.e., a contour eye diagram 
          
enum typeOutput { PDFS = 1, EYE = 2};

*/

//! Enumeration type that specifies form of the Stokes parameters of the noise
/*!
    Possible values:
    
    - UNPOLARIZED = 1, Noise statistics determined by TotalNoiseSpectralDensity
                       which is obtained using 
              OptSignal::GetAccumulatedTotalNoiseSpectralDensitySemiAnalytical
              Signal can be SCALAR or VECTOR

    - PARTIALLY_POLARIZED = 2, Noise statistics determined by 
                       Stokes parameters of noise, i.e. by 
                       NoiseSpectralDensity, DOP of noise, and 
                       Stokes vector of polarized part of noise.
                       Noise is assumed to be wide sense stationary
                       as in Ivan Lima's JLT paper. In particular,
                       noise is not affected by PMD in any way,
                       but it can be affected by PDL. TypeSignal is assumed
                       to be VECTOR and  TypePMDonSignal is either
                       NO_PMD or FIRST_ORDER_PMD.

    - PARTIALLY_POLARIZED_WITH_DGD = 3, Noise statistics determined by 
                       Stokes parameters of noise, i.e. by 
                       NoiseSpectralDensity, DOP of noise, and 
                       Stokes vector of polarized part of noise.
                       We assume here that PARTIALLY_POLARIZED noise
                       and VECTOR signal with FIRST_ORDER_PMD
                       have both passed through same 
                       fiber with only first-order PMD. 
                       [NOT IMPLEMENTED YET]
                        
*/



enum typeGaussianNoise { UNPOLARIZED = 1, 
			 PARTIALLY_POLARIZED = 2, 
			 PARTIALLY_POLARIZED_WITH_DGD = 3};

//! Enumeration type that specifies type of PMD acting on signal
/*!
    Possible values:
      -NO_PMD = 1, Can have TypeSignal = SCALAR or VECTOR. 
      -FIRST_ORDER_PMD = 2, In this case TypeSignal = VECTOR must hold.
       
*/                    

enum typePMDonSignal { NO_PMD = 1, FIRST_ORDER_PMD = 2 };


//! Enumeration type that specifies how clock is recovered
/*!
    Possible values:
    -AUTOMATIC = 1, Uses appropriate method given in ElecSignalStat class
                    acting on noise-free signal entering receiver.
    -MANUALLY_SET = 2, Clock time is input by user.
*/

enum typeClockRecoveryGaussianReceiver { AUTOMATIC = 1, MANUALLY_SET = 2 };

//! Given a noise free signal, appropriate noise statistics and receive filters
//! this class computes the Gaussian approximation of the PDF of the current 
//! in each bit and the corresponding bit-error ratio

class GaussianReceiver
{

public:

  GaussianReceiver(OptSignal * WDMSignal2,
		   RanNumGen *RNG2,
		   int DemuxChannArrayIndex2,
		   string InDir, 
		   string Job2);

  ~GaussianReceiver();

  string WriteTypeOutput(void);
  string WriteTypeGaussianNoise(void);
  string WriteTypePMDonSignal(void);
  string WriteTypeClockRecovery(void);

  void ReceiveSignal(string FileID);

  //! StokesParametersPlusPlus and StokesParametersPlusMinus are 4x1 vectors
  //! for the first-order PMD fiber. 
  //! c_plus2 and c_minus2 are the cplx splitting factors.

    void SetFirstOrderPMDParameters(
		double * StokesParametersPlusPlus,
		cplx * StokesParametersPlusMinus, cplx c_plus2, cplx c_minus2);

  //! The OptFiber PMF must be associated to the OptSignal WDMSignal.
  //! You must have called PMF->GetFiberSampleCompleteScatt() before
  //! calling this method. The implementation of the theory given in this method
  //! assumes that WDMSignal has only been generated using a signal constructor
  //! or OptSignal::RegenerateSignalString() and is to be propagated through
  //! the PMF. In particular, the signal has not been sent through a 
  //! PolarizTransformer. Instead, you can use OptSignal::SetInJones() and
  //! OptSignal::RegenerateSignalString()

  void ComputeFirstOrderPMDParameters(OptFiber * PMF);

//! Receives a signal that is the sum of (1) a noise-free signal that has
//! gone through first-order PMD fiber (ie PMF) and (2) Partially polarized
//! noise. To be used when know input parameters, say from an experiment.
//! See Jan 5th 2006 "How to validate receiver formula" by J. Zweck

/*! UnitStokesPlusPlus2, 
    UnitAverageStokesSignal and UnitStokesVectorPolarizedPartNoise
    are 3x1 and have length one.
*/

void ReceiveSignalDGD_NoisePartiallyPolarized(
     double DGD2, 
     double ClockRecoveryTime2, 
     double *UnitStokesPlusPlus2, 
     double *UnitAverageStokesSignal,
     double DOP_Signal,
     double TotalNoiseSpectralDensity,
     double DOP_Noise, 
     double * UnitStokesVectorPolarizedPartNoise,
     string OutFileID);

private:

  // METHODS:

void RecoverClockAndSetTimeSlices(void);
void SetTimeSlices(void);
void ReceiveSignalUnpolarizedNoise(string FileID);
void ReceiveSignalNoPMD_NoisePartiallyPolarized(string FileID);

//! Receives a signal that is the sum of (1) a noise-free signal that has
//! gone through first-order PMD fiber (ie PMF) and (2) Partially polarized
//! noise.

/*! For each different fiber realization and input polarization state, you
    must call GaussianReceiver::ComputeFirstOrderPMDParameters. You must also
    propagate the signal through an OptFiber object
    before calling this method. 
    The clock will be recovered based on the signal after the PMF unless you
    use TypeClockRecovery = MANUALLY_SET.
*/

void ReceiveSignalDGD_NoisePartiallyPolarized(string FileID);
void ComputeMinimumBERAndDecisionThreshold(string FileID);
void WriteFilePairVectors(double * x, double *y, int dim, string OutFileName);
void WriteFilePDFTimeSlices(string OutFileName);
void ComputeAveragePDFs(string FileID);

  // DATA:

  typeOutput TypeOutput;  
  // enumtype  typeOutput is defined in ChiSquareReceiver.hh
  typeGaussianNoise TypeGaussianNoise;
  typePMDonSignal  TypePMDonSignal;
  typeClockRecoveryGaussianReceiver TypeClockRecovery;

  int DebugLevel;

  double ClockRecoveryTime;

  OptSignal * WDMSignal;
  OptSignal * SingleChannelSignal;
  OptSignal * WDMSignalTmp;

  OptFilter * TunableOptFilter;
  OptDemuxer *oOptDemuxer;

  Photodetector * oPhotodetector;
  ElecFilter * oElecFilter;

  ElecSignalStat * oElecSignalStat;

  int DemuxChannArrayIndex;
  string Job;

  int ObjectInitialized;

  int NumTimeSlices;
  int NumSlicesPerBit;
  double TimeBetweenSlices;
 
  int NumBitsToPlot;

  double * TimeSlices;
  int * TimeSliceIndices;

  double * VarianceCurrent;
  double * MeanCurrent;
  double MeanNoise;

  double  MinCurrent,MaxCurrent,DeltaCurrent;
  int NumPointsCurrent;
  double * CurrentVector;

  double * BER;
  double * AveragePDFZeros;
  double * AveragePDFOnes;
  double ** PDFTimeSlices;
  double * PDFTimeSlicesStorageVec;
  int * BitString;


  // Parameters for signal with FirstOrderPMD

  bool FirstOrderPMDParametersSet;
  double DGD;
  double * StokesPlusPlus;
  cplx * StokesPlusMinus;
  cplx c_plus;
  cplx c_minus;

};

#endif
