#ifndef _GENERAL_REAL_MATRIX_H
#define _GENERAL_REAL_MATRIX_H

#ifndef _OCS_SYSTEM_HDRS_
#define _OCS_SYSTEM_HDRS_

#include <iostream.h>
#include <string.h>
#include <math.h> 

#include <fstream.h>
#include <complex.h>
#include <stdlib.h>
#include <stdio.h>
#include <strstream>

#include <string>
  using namespace std;


#endif /* _OCS_SYSTEM_HDRS_ */

#include "ocsConst.hh"

// For most compliers Fortran77 subroutines like myfun are recognized
// as myfun_ Therefore:

#define F77NAME(x) x##_

extern "C" void F77NAME(dgetrf) (int*, int*, double*,int*,int*,int*);

extern "C" void F77NAME(dgetrs)(char*,int*,int*,double*,int*,int*,double*,
				int*,int*);
		
extern "C" void F77NAME(dsyev)(char*,char*,int*,double*,int*,double*,double*,
                               int*,int*);

// The last argument is an int which is the length of the
// character array name.

extern "C" int F77NAME(ilaenv)(int*,char*,char*,int*,int*,int*,int*,int);

class GeneralRealMatrix
{

public:

  GeneralRealMatrix(int NumRows2, int NumCol2);      
  GeneralRealMatrix(const GeneralRealMatrix & Copy);       

  ~GeneralRealMatrix(void);

  void WriteFileMatrix(string OutFileName);

  //! Gets the (i,j) element of the matrix, 
  //! where 1 <= i <= NumRows, 1<=j <= NumCols

  /*!
     The matrix is stored in the array Data in Column Major order
     A(i,j) = Data[ (i-1) + (j-1) * NumRows ],
     i.e. elements on the same column of A are stored contiguously in Data.
     This is the convention used in Fortran, and hence in BLAS and Lapack
  */

  inline double Get(int i, int j) const
   {
     BoundsCheck(i,j);
     return Data[ (i-1) + (j-1) * NumRows ];
   };

  //! Gets the (i,j) element of the matrix to Val, 
  //! where 1 <= i <= NumRows, 1<=j <= NumCols

  inline void Set(int i, int j, double Val)
   {
    BoundsCheck(i,j);
    Data[ (i-1) + (j-1) * NumRows ] = Val;
   };

  //! Accesses memory location of (i,j)-element of matrix
  /*! Typical use: Element(i,j) += Val
   */

  inline double & Element(int i, int j)
  {
    BoundsCheck(i,j);
    return Data[ (i-1) + (j-1) * NumRows ];
  }


  inline int GetNumRows(void) const { return NumRows; };
  inline int GetNumCols(void) const { return NumCols; };

  //! Sets all elements of the matrix to Val

  void SetAll(double Val);

  //! Sets this matrix to be zero

  void Zero(void);

  //! Matrix multiplication. Sets this matrix to be the product C = A*B

  void Multiply(const GeneralRealMatrix & A, const GeneralRealMatrix & B);

  //! Sets this matrix to be the product C = Transpose(A)*B

  void MultiplyTrans(const GeneralRealMatrix & A, const GeneralRealMatrix & B);

  //! Produce image, Y = A * X, where A is this.

  void Image(const GeneralRealMatrix & X, GeneralRealMatrix & Y) const;

  //! For square matricies, sets this = 0.5*(this + Transpose(this))

  void Symmetrize(void);
 
  //! Sets Atrans to be the transpose of this matrix

  void Transpose(GeneralRealMatrix & Atrans) const;

  //! Solves A*X=B for X, where A = this matrix is assumed to be square
  //! WARNING: This method has not been tested!!!

  int Solve( GeneralRealMatrix& X, const GeneralRealMatrix& B) const;

  //! Finds the eigenvalues and eigenvectors of this matrix,
  //! which is symmetric. 

  void SymmetricEig(GeneralRealMatrix & EigValue,
		    GeneralRealMatrix & EigVector) const;
 
  double* Data;

private:  

  inline void BoundsCheck(int i, int j) const
   {
     if( i > NumRows || i < 1)
       {  
         cerr << "GeneralRealMatrix::BoundsCheck: Row index out of bounds."
              << endl
	      << "ABORT" << endl;
         exit(1);
       }

     if( j > NumCols || j < 1)
       {  
        cerr << "GeneralRealMatrix::BoundsCheck: Col index out of bounds."
              << endl
	      << "ABORT" << endl;
         exit(1);
       }
   }


  int NumRows;
  int NumCols;

};

#endif // ## _GENERAL_REAL_MATRIX_H
