#ifndef _MATRIX_TOOLS_H_
#define _MATRIX_TOOLS_H_

#include <fstream.h>
#include "ocsTools.hh"

// Written by Ron Holzloehner. All Matrices are assumed to be square.


//! Matrix Assignment LHS_Matrix = RHS_Matrix

void CopyMatrix(double ** LHS_Matrix, double ** RHS_Matrix, int dim);

//! Computes the matrix product R = A * B or R = A * B^T depending
//! on whether the flag Transpose_B is 0 or 1.

void DenseMatrixProduct(double **A, double **B, double **R,
                        bool Transpose_B, bool R_is_symmetric, int dim);


//! Computes the matrix product R = A*B*A^T or R = A^T*B*A depending
//! on whether the flag Transpose_A is 0 or 1, respectively

void DenseMatrixSimilarityProduct(double **A, double **B, double **R,
				  double ** Work, bool Transpose_A, 
				  bool R_is_symmetric,int dim);

//! Multiply all elements of the dim x dim double matrix A by the double scale.

void DenseMatrixScale(double **A, int dim, double scale);

//!  Copies the lower half of the matrix A to the upper half, so that the
//!  resulting matrix is symmetric.
 
void CopyLowerToUpperHalf(double **A, int dim);

//! Replaces A with 0.5*(A+A') where A' is the transpose of the real matrix A

void SymmetrizeMatrix(double **A, int dim);

//! If input[i][j] = A_ij then output[i][j]= A_ji

void TransposeMatrix(double **A, int dim);

//! Sets all entries of Matrix A to 0.

void ZeroMatrix(double **A, int dim);

//! Filters the matrix A with filter

void DenseMatrixFilter(double **A, int dim, double *filter);

//! Writes the Matrix scaling*A of size dim x dim to the file F  

void WriteMatrix(double **A, int dim, ofstream *F, double scaler);

//! Reads in the Matrix scaling*A of size dim x dim from the file F  
//! that was previously written using WriteMatrix

void ReadMatrix(double **A, int dim, ifstream *F, double scaling);

//! Computes the matrix-vector product Out = A*In where In and Out are vectors
//! and A is square of size dim x dim

void DenseMatrixVectorProduct(double * Out, double **A, double * In, int dim);


//! Computes the matrix-vector product Out = In*A where In and Out are vectors
//! and A is square of size dim x dim

void DenseVectorMatrixProduct(double *Out, double *In, double **A, int dim);

//! Adds the double add to the principal diagonal of Matrix A.

void AddToPrincipalDiag(double **A, int dim, double add);

//! Adds the array ScalarMultiple*V, where V is a vector 
//! to the principal diagonal of Matrix A.

void AddVectorToPrincipalDiag(double **A, double *V, 
			      double ScalarMultiple,int dim);

//! Returns the double
//! (aa,bb)/(bb,bb) for aa, bb complex vectors of length len
//! where (aa,bb) := Re[conj(aa) * bb] = Re(aa)*Re(bb) + Im(aa)*Im(bb)

double ScalarProductNormalized(cplx *aa, cplx *bb, int len);

//! Returns the angle between complex vectors aa and bb of length len
//! using the inner product 
//! (aa,bb) := Re[conj(aa) * bb] = Re(aa)*Re(bb) + Im(aa)*Im(bb)

double VectorAngle(cplx *aa, cplx *bb, int len);

//! Performs Gram-Schmidt orthogonalization of num vectors of length len
//! where vec[i][] is the i-th vector 
//! i.e., vec is a cplx array of dim num x len
//! alpha[] must be a double array of length (num^2-num)/2
//! Can start orthogonalization proceedure with the k1-th vector
//! instead of the 1-st vector if you like. (k1 is an optional argument).
//! Uses the inner product 
//! (aa,bb) := Re[conj(aa) * bb] = Re(aa)*Re(bb) + Im(aa)*Im(bb)


void SchmidtOrthogonalizer(cplx **vec, int num, int len,
                           double *alpha);

void SchmidtOrthogonalizer(cplx **vec, int num, int len, 
                           double *alpha, int k1);


//! The infinity norm of a square matrix A is the largest row sum
//! of the matrix whose entries are the absolute values of A 


double InfNormMatrix(double **A,int dim);

//! The one norm of a square matrix A is the largest column sum
//! of the matrix whose entries are the absolute values of A 

double OneNormMatrix(double **A,int dim);

//! Returns the standard inner product of two real vectors

double InnerProduct(double * V, double *W, int Length);

#endif // _MATRIX_TOOLS_H_






