// $Id: NoiseCovariance.hh,v 1.4 2007/10/24 14:29:19 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2002):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _NOISE_COVARIANCE_
#define _NOISE_COVARIANCE_


#include "ocsOptSignal.hh"
#include "ocsOptFiberLocalError.hh"
#include "ocsOptAmplifier.hh"
#include "ocsLumpedLoss.hh"
#include "MatrixTools.hh"

//! Enumeration type used to determine method for computing 
//! Covariance Matrix.

/*! Possible values 
    - AWGN = 0
    - MC_WITHOUT_K = 1
    - MC_WITH_K = 2
    - ODE = 3

    /sa AlsoRemoveZerosJitter

*/

enum typeNoiseMethod { AWGN = 0, MC_WITHOUT_K=1,
                       MC_WITH_K=2, ODE=3};

//! Enumeration type used to determine method for separating out jitter

/*! Possible values 
    - NONE = 0
    - PHASE_ONLY = 1
    - PHASE_AND_TIMING = 2
    
*/

enum typeJitterSeparation { NONE = 0, PHASE_ONLY = 1, PHASE_AND_TIMING = 2};


//! The NoiseCovariance class is used in the computation of 
//! accurate eye diagrams
//! and bit error rates in the presence of fiber nonlinearity, chromatic
//! dispersion and amplifier ASE noise using the noise linearization 
//! methods of Ronald Holzloehner. 

/*! The class computes the noise-free signal
    and the noise covariance matrix using either the Monte Carlo or the ODE
    method. Details of Ron's method can be found in his papers in JLT and PTL
    in 2002. The class is based on Ron's code and was written for OCS by JZ. 
*/

class NoiseCovariance
{

  public:

  // ######### Public Methods ##################

  //! Constructor

  NoiseCovariance(string InFileName, string Job2,OptSignal * Signal2);

  //! Assignment operator. Assumes LHS NoiseCovariance  object has
  //! already been constructed.

  NoiseCovariance & operator=(NoiseCovariance & RHS_NoiseCovariance);

  //! Resets parameters to values should have at beginning of system

  void ResetVolatileParameters(void);
   
  //! Destructor

  ~NoiseCovariance(void);

  //! Propagates through the fiber *Fiber2

   void Propagate(OptFiberLocalError * Fiber2);

  //! Amplifies Signal using amplifier *Amplifier2

   void Amplify(OptAmplifier * Amplifier2);

  //! Attenuates the signal and covariance matrix
  //! using LumpedLoss object oLumpedLoss

   void Attenuate(LumpedLoss * oLumpedLoss);

 //! Used to change when in application code jitter separation occurs

  void TurnOffJitterSeparation(void);

 //! Used to change when in application code jitter separation occurs

  void TurnOnJitterSeparation(void);

  string WriteTypeNoiseMethod(void);

  string WriteTypeJitterSeparation(void);

  void WriteFileNoiseFreeSignalAndCovarianceMatrix(
			 string OutFileNamesSuffix);

  //! Reads in files previously written out by 
  //! NoiseCovariance::WriteFileNoiseFreeSignalAndCovarianceMatrix()

  void ReadFileNoiseFreeSignalAndCovarianceMatrix(
                          string InFileNamesPrefix);
  
  //! No real need to use this method anymore.

  void WriteFileAWGNNoiseFreeSignalAndNoiseSpectralDensity(
				   string OutFileNameSuffix);

  void WriteFileTimeField(string OutFileName,fftMatrices Field);
  void WriteFileFreqField(string OutFileName,fftMatrices Field);

  //! This method is used by the ChiSquareReceiver class to set a pointer
  //! to the ReducedNoiseFreeSignal. 

  /*! Access as 

      ReceiverReducedNoiseFreeSignal.tOptSignalX[jj] 
      ReceiverReducedNoiseFreeSignal.fOptSignalX[jj] 

      where jj=0...CovMatrixDim-1
  */

  void GetReducedNoiseFreeSignal(fftMatrices & ReceiverReducedNoiseFreeSignal);

  //! This method is used  to compute the
  //! Covariance matrix from using 
  //! Signal->GetTotalAccumulatedNoiseSpectralDensitySemiAnalytical()

  void ComputeCovarianceMatrixAWGN(void);

  //! Sets CovarianceMatrix to be the multiple of the identity matrix
  //! that with the NoiseFreeSignal gives the OSNR whose value in dB is
  //! OSNRdB. Only works when NoiseCovariance::TypeNoiseMethod == AWGN.
  //! For WDM signals divides total average power by number of channels
  //! to calculate the signal power

  void SetCovarianceMatrixAWGN(double OSNRdB, double B_OSA);

  //! Sets the CovarianceMatrix to be diagonal matrix with specified
  //! TotalNoiseSpectralDensity. (For SCALAR signals only foe now.)

  void InitializeCovarianceMatrix(double TotalNoiseSpectralDensity);

//! Computes covariance matrix in case of CW signal for an amplifier
//! followed by a  single fiber with no loss and no dispersion.
//! The power of the CW signal is specified by Signal->GetOptPowerChann()
//! Based on a special case of a formula of Carena et al.
//! in PTL Vol 9 Num 4 April 1997, p535.

   void PropagateAnalyticCWSignalNoLossNoDispersion(
        OptFiberLocalError * Fiber2,OptAmplifier * Amplifier2);

  //! This method is used by the ChiSquareReceiver class to set a pointer
  //! to the CovarianceMatrix

  void GetCovarianceMatrix(double ** ReceiverCovarianceMatrix,
			   bool ReadInSignalAndCovMatrixFlag,
			   double SetOSNRdB, double B_OSA);

  int GetCovMatrixDim(void) {return CovMatrixDim;};
  int GetCovMatrixFirstIndex(void) {return CovMatrixFirstIndex;};
  int GetCovMatrixLastIndex(void) {return CovMatrixLastIndex;};
  
  string GetJobName(void) {return Job;};

  double GetPropagatedLength(void) {return PropagatedLength;};

  private: 

  // ########### Private Methods ################

  //! Allocates and initializes memory for vectors and matrices
  //! Called by NoiseCovariance::NoiseCovariance

  void AllocateMemory(void);


  //! Computes the Propagator Matrix Psi(z) and the Noise Free Signal u_0

  void Propagator(void);

  //! Computes a column of the propagator matrix
  /*!
      Removes phase and timing jitter
      Called by NoiseCovariance::Propagator
  */


  void ComputeColumnPropagator(int PerturbedFreqIndex,double Delta, 
			  bool RealNotImag,double TotalEnergy);


  //! Propagates a pertrubation of the noise free signal through the fiber.

  void Perturb(int PerturbedFreqIndex,double Delta, bool RealNotImag);


  //! Propagates a pertrubation of the noise free signal through the fiber.
  /*!
     Called by NoiseCovariance::Propagator

     Comments by Walter. [JZ: I'm not sure if this realy works yet]

     Computes the perturbation of any channel and therefore allows for the  
     computation of the CovarianceMatrix  for any channel
 
     ChannelIndex has to be between 0 and NumberOfChannels-1;
     ChannelIndex allows to compute the Covariance Matrix of a channel
     different from the central one. In other words it allows to compute
     each block matrix on the diagonal of an ideal complete big CovarianceMatrix
     for a WDM simulation

   */
  
  void Perturb(int PerturbedFreqIndex,double Delta, 
	       bool RealNotImag, int ChannelIndex);

  //! Computes K_new = Psi*K_old*(Psi^T)

  void UpdateCovMatrixODE(void);  

  //! ArtificialDispersionCompensation for removal of phase and timing jitter

  /*!
      When DisentangleFlag = true this method disentangles the pulses
      in the signal stored in SignalData by linearly compressing them
      with dispersion equal to the negative of 
      NoiseCovariance::AccumulatedFirstOrderDispersion.
      When DisentangleFlag = false the method re-entangles the pulses
      by apply dispersion of NoiseCovariance::AccumulatedFirstOrderDispersion.
  */

  void ApplyArtificialDispersionCompensation(fftMatrices SignalData, 
					     bool DisentangleFlag);


  //! Computes the Artificial Dispersion Compensation Operator used in 
  //! NoiseCovariance::ApplyArtificialDispersionCompensation

  void ComputeArtificialDispersionCompensationOperator(void);

  void AmplifyCovarianceMatrix(double LinearGain);

  void LumpedLossCovarianceMatrix(double LossLinear);

  void ComputeReducedSignal(fftMatrices Signal2,fftMatrices ReducedSignal2);

  //! Extracts all the channels but the central one out of the WDM signal
  //! in order to be able to separate the phase and/or timing jitter
  //! from that channel

  void ApplyArtificialCentralChannelDemux(fftMatrices& PerturbationEnd);

  //! Inserts the channels extracted by ApplyArtificialCentralChannelDemux
  //! into the SingleChannel PerturbationEnd signal from which the
  //! phase and/or timing jitter has been extracted.

  void RestoreChannels(fftMatrices& PerturbationEnd);

  // ########### Public Data ####################

public:
  OptSignal * Signal;

  // ########### Private Data ###################

private:

  typeNoiseMethod TypeNoiseMethod;
  typeJitterSeparation TypeJitterSeparation;
  typeJitterSeparation TypeJitterSeparationInput;
  bool AlsoRemoveZerosJitter;

  string Job;
  OptFiberLocalError * Fiber;
  OptAmplifier * Amplifier;

  double PropagatedLength;
  double PropagatedLengthSaved;
  int ZStepNum;
  int ZStepNumSaved;

  //! The accumulated first order dispersion at the center frequency.
  //! Used for artificial dispersion compensation.
  //! Units: ?? 

  double AccumulatedFirstOrderDispersion;

  //! The accumulated second order dispersion at the center frequency.
  //! Used for artificial dispersion compensation.
  //! Units: ?? 

  double AccumulatedSndOrderDispersion;

  //! The perturbation amount is a fraction DeltaFrac of the average power 

  double DeltaFrac;

  //! The signal data structure that actually propagates through the system.

  fftMatrices DataToPropagate;

  typeSimulation TypeSimulation;

  //! This is same as OptSignal::qtPoints

  int SignalVecDim;

  //! Pointer to the fft class to take fft's of 
  //! size NoiseCovariance::SignalVecDim

  cfftw * fft;


  //! Number of rows (or columns) of CovarianceMatrix 
  /*!
      Ron's N. Must be even
  */

  int CovMatrixDim; 

  //! CovMatrixFirstIndex = -CovMatrixDim/2
  /*!
      Ron's N1
  */

  int CovMatrixFirstIndex;

  //! CovMatrixLastIndex = CovMatrixDim/2 - 1
  /*!
      Ron's N2
  */

  int CovMatrixLastIndex;


  //! Pointer to the fft class to take fft's of 
  //! size NoiseCovariance::CovMatrixDim

  /*!
     ReducedTimeWindow = TimeWindow
     ReducedDeltaTime = TimeWindow/CovMatrixDim
     ReducedDeltaFreq = DeltaFreq
     ReducedFreqWindow = DeltaFreq*CovMatrixDim
  */


  cfftw * Reduced_fft;

  //! Real covariance matrix of size 
  //! 2*NoiseCovariance::CovMatrixDim x 2*NoiseCovariance::CovMatrixDim
  /*!
      This is the script K in Ron's papers/thesis
  */

  double **CovarianceMatrix;

  //! Real propagator matrix of size 
  //! 2*NoiseCovariance::CovMatrixDim x 2*NoiseCovariance::CovMatrixDim
  /*!
      This is the \Psi in Ron's papers/thesis
  */
 
  double **PropagatorMatrix;

  //! Real matrix of size 
  //! 2*NoiseCovariance::CovMatrixDim x 2*NoiseCovariance::CovMatrixDim 
  //! used in  UpdateCovMatrixODE()
  /*!
      This is the matrix C in Ron's code
  */

  double **TempProductMatrix;

  //! Real vector of length 4*sq(NoiseCovariance::CovMatrixDim)
  //! where the elements of NoiseCovariance::CovarianceMatrix are
  //! actually stored

  double * CovarianceStorageVec;

  //! Real vector of length 4*sq(NoiseCovariance::CovMatrixDim)
  //! where the elements of NoiseCovariance::PropagatorMatrix are
  //! actually stored

  double * PropagatorStorageVec;

  //! Real vector of length 4*sq(NoiseCovariance::CovMatrixDim)
  //! where the elements of NoiseCovariance::TempProductMatrix are
  //! actually stored

  double * TempProductStorageVec;


  //! Used in Propagator method to store the noise free signal 
  //! at start of a fiber

  fftMatrices NoiseFreeSignalStart;

  //! Used in Propagator method

  fftMatrices NoiseFreeSignalEnd;

  //! Used in Propagator method:. jc times NoiseFreeSignalEnd
  //! Used to handle phase jitter

  fftMatrices jcNoiseFreeSignalEnd;

  //! Used in Propagator method to store: jc frequency NoiseFreeSignalEnd
  //! Used to handle timing jitter

  fftMatrices jcOmegaNoiseFreeSignalEnd;

  //! Used in NoiseCovariance::ComputeColumnPropagator

  fftMatrices PerturbationEnd;

  //! Used in NoiseCovariance::ComputeColumnPropagator

  /*! 
    Reduced signals have size CovMatrixDim. 
    The reduced signal in frequency is obtained
    from the full length signal by extracting the central
    CovMatrixDim frequencies.
    If required the reduced signal in time should be computed
    from the reduced signal in frequency using the cfftw object
    NoiseCovariance::Reduced_fft

  */

  fftMatrices ReducedPerturbationEnd;

  cplx * ArtificialDispersionCompensationOperator;

 /*! 
    Reduced signals have size CovMatrixDim. 
    The reduced signal in frequency is obtained
    from the full length noise free signal by extracting the central
    CovMatrixDim frequencies.
    The reduced signal in time should be computed
    from the reduced signal in frequency using the cfftw object
    NoiseCovariance::Reduced_fft. 
    The ReducedNoiseFreeSignal is computed in 
    NoiseCovariance::ComputeReducedSignal

    Indices for the reduced signal start at CovMatrixFirstIndex and end at
    CovMatrixLastIndex. The actual data is stored in the arrays 
    fftMatrices ReducedNoiseFreeSignalStorage whose indices go from
    0 to CovMatrixDim - 1.

  */

  fftMatrices ReducedNoiseFreeSignal;

  //! \sa fftMatrices NoiseCovariance::ReducedNoiseFreeSignal

  fftMatrices ReducedNoiseFreeSignalStorage;
 
 
  // added by Walter for WDM systems
  
  // used for the version of perturb() that allows to compute covariance
  // matrixes for each channel  
  int ChannelOfCovarianceMatrix;
  
  // data for WDM
  int NumberOfChannels;
    
  // save PerturbationEnd of WDM signal to be able to reconstruct WDM signal
  // after the jitter separation
  fftMatrices PerturbationEnd_saved;
   
};

#endif // _NOISE_COVARIANCE_






