// $Id:

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2009)
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _SATURABLE_ABSORBER_HH_
#define _SATURABLE_ABSORBER_HH_

#include "ocsOptSignal.hh"
#include "ocsOptFiberLocalError.hh"

// Written by John Zweck August 2009.

//! An enumeration type for the type of saturable absorber
/*! 

   \arg NO_SAT_ABS = 0

     \arg QUADRATIC_NO_LOSS_ANALYTIC=1

         Fast saturable absorber with a quadratic nonlinearity
 
          du/dz = \delta*(|u|^2)*u. 

        Solved analytically. See OptFiberLocalError class


    \arg QUADRATIC_NO_LOSS=2

         Fast saturable absorber with a quadratic nonlinearity
 
          du/dz = \delta*(|u|^2)*u. 

          Solved numerically using Heun's method
          
          We use parameter name:

         FastSaturableAbsorptionNonlinearCoefficient = \delta


    \arg QUADRATIC_WITH_LOSS_ANALYTIC = 3 

         Uses analytical formula for exact solution
         of fast saturable absorber with a 
         quadratic nonlinearity and loss

    \arg QUADRATIC_WITH_LOSS = 4

          Fast saturable absorber with a quadratic nonlinearity
          and loss, ie, 
          du/dz = \delta*(|u|^2 - \epsilon)*u. 
          Solved numerically using Heun's method
          
          We use parameter names:

         FastSaturableAbsorptionNonlinearCoefficient = \delta
         FastSaturableAbsorptionAttenuationCoefficient = \epsilon

   \arg RATIONAL_QUADRATIC=5

     Fast saturable absorber along lines of Ablowitz/Kaertner/Chen-Menyuk
     models:

     du/dz = -\delta*\epsilon/(1 + |u|^2/\epsilon)*u

     Solved numerically using Heun's method.

     The parameters were chosen so that if use 1+x \approx 1-x
     approximation, then the equation agrees with QUADRATIC_WITH_LOSS.

     We use parameter names:

     FastSaturableAbsorptionNonlinearCoefficient = \delta
     FastSaturableAbsorptionAttenuationCoefficient = \epsilon



    \arg COUPLED_ODE = 6

      Slow saturable absorber based on Haus JQE 11 (9) Sept 1975, pp 736-746.
      Also see book of W. Koechner and M. Bass.
 
      du/dz = -l(z,t)u
      dl/dt = - (l-l_U)/tau - l|u(z,t)|^2/(tau*P_{sat})

      where l=l(z,t) is saturable loss in m^{-1}
            l_U is unsaturated loss in m^{-1}
            tau is relaxation time in seconds
            P_{sat} is saturation power in Watts.

      For each z, we solve for l using initial condition that l(t=0)=l_U
      where we assume t=0 is such that u(z,0) = 0 holds, ie pulse is not
      centered near t=0. We run ODE solver for t forward in time.

      This model is based on a three-level atomic system with a fast
      relaxation of the upper level. We assume that saturable absorber
      relaxes completely between pulses, ie, when only one pulse in laser
      at a time we are assuming that \tau << Round trip time in laser.
      

*/

enum typeSaturableAbsorption {NO_SAT_ABS=0,
                              QUADRATIC_NO_LOSS_ANALYTIC=1,
	                      QUADRATIC_NO_LOSS=2,
                              QUADRATIC_WITH_LOSS_ANALYTIC=3,
                              QUADRATIC_WITH_LOSS=4,
                              RATIONAL_QUADRATIC=5,
                              COUPLED_ODE=6};


class SaturableAbsorber
{
   public:

   SaturableAbsorber(string InFileName, OptSignal *oOptSignal2);

   void Apply(double  * PropagatedLength);

   string WriteTypeSaturableAbsorption(void);
   string WriteTypeStepSizes(void);

   private:
   
   void ApplyQuadraticNoLossAnalytic(void);
   void ApplyQuadraticWithLossAnalytic(void);

   void ApplyHeunsMethod(void);

   void ApplyHeunsMethodConstantSteps(void);
   void ApplyHeunsMethodLocalError(void);

   void ApplyHeunsSingleStep(double StepSizeZ);
   void ComputeSAFunction(cplx * U_XIn, cplx * U_YIn, double * SAFOut);

   double RungaKutta4Step(cplx* U_XIn, cplx * U_YIn,
                          double x_in, int time_index_in);

   double FastSaturableAbsorptionFunction(cplx u);
   double CoupledODESatAbsFunction(double Power,double loss);

   void UpdateTime(void);
   void UpdateFreq(void);
   void ComputeDeltaZMax(void);
   double ComputeRelativeError(void);
   void CopyFftMatricesTimeData(fftMatrices * OrigData,
                                fftMatrices * CopyOfData);


   typeSimulation TypeSimulation;
   OptSignal * oOptSignal;
   cfftw *fft;
   fftMatrices sfftM;
   int qtPoints;

   int DebugFlag;

   //! \sa typeSaturableAbsorption 
   typeSaturableAbsorption TypeSaturableAbsorption;

   //! Length of SaturableAbsorber in meters

   double SaturableAbsorberLength;

   /*! For TypeSaturableAbsorption = QUADRATIC_ANALYTIC or QUADRATIC 
      this is \delta in (Wm)^{-1}
   */

   double FastSaturableAbsorptionNonlinearCoefficient;

   /*! For TypeSaturableAbsorption = QUADRATIC_ANALYTIC or QUADRATIC 
       this is \epsilon in Watts and represents the maximum instantaneous
       power of electric field that one can have without singularities
       occuring for any \delta and SaturableAbsorberLength.
   */

   double FastSaturableAbsorptionAttenuationCoefficient;

   //! Unsaturated Loss parameter l_U in m^{-1} for Coupled ODE model 

   double CoupledODEUnsaturatedNepperLoss;

   //! Relaxation Time parameter \tau in secs for Coupled ODE model

   double CoupledODERelaxationTime;

   //! Saturation Power parameter P_{sat} in Watts for Coupled ODE model

   double CoupledODESaturationPower;

   //! \sa enum typeStepSizes

   typeStepSizes TypeStepSizes;

   //! NumSteps for ODE solver
   int NumSteps;
   
   double RelativeErrorGoal;
   double DeltaZMaxInfile;
   double DeltaZMax;
   double DeltaZInitial;
   bool InitializedDeltaZInitialFlag;

   fftMatrices sfftM_Start;
   fftMatrices sfftM_Coarse;
   cplx * LaserWorkX;
   cplx * LaserWorkY;

   double * SAF1;
   double * SAF2;


};



#endif /* _SATURABLE_ABSORBER_HH_ */
