

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_ELEC_SIGNAL_STAT_HH_
#define _OCS_ELEC_SIGNAL_STAT_HH_

#include "ocsBitString.hh"
#include "ocsOptSignal.hh"
#include "ocsPhotodetector.hh"
#include "ocsOptFilter.hh"
#include "ocsElecFilter.hh"
#include "ocsTools.hh"


/*! Enumeration type for ClockRecovery algorithm.

    Possible values:
    -USER_SET = 0 (Set by user in input parameter file using input parameter
                   ClockRecoveryTime. Must use appropriate constructor.)
    -FREQ_DOUBLING =  1 (Recommended and default value for NRZ)
    -RZ_SIGNAL_FREQ = 2 (Recommended and default value for RZ)
    -NRZ_HALF_SIGNAL_FREQ = 3 (Works for noise free NRZ)
    -BEST_Q_FACTOR = 4 (Works for RZ and NRZ (not physical))
    -DIFFERENTIATE_SIGNAL = 5 (Works for NRZ 
                               (and RZ with narrow electric filter))
    -CONTINUOUS_EXOR = 6 (Works for NRZ(and RZ with narrow electric filter))
    -BINARY_EXOR = 7 (Works for NRZ(and RZ with narrow electric filter))
    -BEST_MIN_EYE_OPENING = 8 (Works for RZ and NRZ (PMD w/o noise)
                               (not physical))

*/

enum typeClockRecovery { USER_SET = 0,
                         FREQ_DOUBLING = 1,
                         RZ_SIGNAL_FREQ = 2,
                         NRZ_HALF_SIGNAL_FREQ = 3, 
                         BEST_Q_FACTOR = 4,
                         DIFFERENTIATE_SIGNAL = 5,
                         CONTINUOUS_EXOR = 6,
                         BINARY_EXOR = 7,
                         BEST_MIN_EYE_OPENING = 8 };


//! The ElecSignalStat class recovers the clock and 
//! computes the mean and standard deviation of the received 
//! electrically filtered current in each bit at the clock 
//! recovery time.

/*!

The class can be used in several different modes: For noise-free simulations,
with noise in conjunction with Monte Carlo simulations, or with noise
using the Gaussian statistics method for computing the first two moments 
in each bit given by the method of Winzer/Lima. Rather than using this
class directly it is better to use one of the receiver classes:
GaussianReceiver, ChiSquareReceiver, or ReceiverSemianalytical

The class was primarily written by Ivan Lima with additional code by
Aurenice Lima, Brian Marks and John Zweck. The class was substantially
edited and reorganized by John Zweck in Jan 2006.

The GaussianStatistics methods implement the equations in the paper

"A Receiver Model for Optical Fiber Communication Systems With Arbitrarily 
Polarized Noise,"  by 

I. T. Lima, Jr., A. O. Lima, Y. Sun, H. Jiao, J. Zweck, C. R. Menyuk, 
and G. M. Carter, in

J. Lightwave Technol. 23 (3), pp. 1478--1490, 2005.

Equation numbers given in documentation refer to equation numbers
in that paper, eg JLT05_22

*/

class ElecSignalStat
{
   public: 

      //! Constructor without input file

      ElecSignalStat(OptSignal * oOptSignal2, 
                     Photodetector *oPhotodetector2);


      //!Constructor that reads parameters from an input file
      /*! Input parameters are 
	    - SetTypeClockRecoveryFlag 
	        -- TypeClockRecovery (values 0 to 8)
	    - ComputePatternStatsFlag
                -- PatternLength
	    - UseSmoothingToRecoverClockFlag (Out of date)
	    - GaussianSmoothingFWHM (Out of date)
      */

      ElecSignalStat(string InFileName,OptSignal *oOptSignal2,
		     Photodetector *oPhotodetector2);

      ~ElecSignalStat(void);

  //! Use after constructor if want to take abs(Current) rather than
  //! real(Current) when computing statistics. (Mostly for backward
  //! compatibility
  
      void SetUseAbsCurrentFlag(bool Value);
 
  //! Returns the type of the clock recovery method

      string WriteTypeClockRecovery(void);

  //! Resets variables to correctly calculate statistics
  //! Use when doing parameter studies before change parameters

      void ClearElecSignalStat(void);

  void ClearGaussianPDFsMonteCarlo(void); // Added by Hua on Jan. 24, 2006

  //! Computes the intensity of received current in each bit
  //! and the mean and standard deviation of the received current
  //! in each bit at clock recovery time. Used in either noise-free
  //! simulations or with noise in Monte Carlo simulations.

      void ComputePerformanceStatistics(void);

  //! Computes intensity of received current in each bit
  //! at clock recovery time and updates variables that
  //! store mean and standard deviation of current in each bit.
  //! Used in Monte Carlo simulations with noise. 
  //! Called by ComputePerformanceStatistics().

      void UpdatePerformanceStatistics(void);

  //! A version of ComputePerformanceStatistics() used in conjunction with
  //! an ElecEqualizer object. The BitIntensity array must have been
  //! previously set by the  ElecEqualizer object.
    
  void ComputeEqualizedPerformanceStatistics(void);   
 
  //! A version of UpdatePerformanceStatistics() used in conjunction with
  //! an ElecEqualizer object. The BitIntensity array must have been
  //! previously set by the  ElecEqualizer object.

  void  UpdateEqualizedPerformanceStatistics(void);
  
  cplx *GetSignalTimeDomain(void) {return sfftPM.tPowerSignal;};
   
  int GetBitFromString(int ii){return oOptSignal->oBitString->GetBit(0,ii);};

 //! Returns power difference between smallest one and largest zero.
 //! Useful when studying PMD.
 //! Only to be used in noise-free simulations! 

   double GetMinAmplitudeMargin();      
      
  //! Writes an eye diagram to a file. 
  //! Useful with noise-free simulations.    

      void WriteFileEyeDiagram(string outFile);  
 
  //! Writes an eye diagram to a file.  
  //! Useful with Monte Carlo simulations with noise.

     void WriteFileEyeDiagramFixedCLK(string outFile);  
      
  //! Sets and returns ClockRecoveryTime 
  //! Call after ComputePerformanceStatistics() 

      double GetClockRecoveryTime(void)
            {return ClockRecoveryTime=GetTargetTime();};

  //! See ElecSignalStat::ComputePerformanceStatistics()

      double GetClockRecoveryTimeBeforeTimeShift(void)
           { return ClockRecoveryTimeBeforeTimeShift;};

  //! Use in conjunction with WriteFileEyeDiagram() to generate
  //! eye diagram that can plot in Matlab  

      double GetClockTimeForMonteCarloEye(void);

      double GetClockCurrent(void);

  //! For use in conjunction with Monte Carlo simulations with noise.
  //! Must set DoGaussianPDFsMonteCarloFlag = 1 and other required parameters
  //! in ElecSignalStat infile and call constructor:
  //! ElecSignalStat::ElecSignalStat(string InFileName,OptSignal *oOptSignal2,
  //!                             Photodetector *oPhotodetector2)
  //! 
  //! First call UpdatePerformanceStatistics() many times within
  //! a loop.
  //! Outputs mean and stddev of current in each bit, computes
  //! Gaussian approximation to pdf in each bit, computes
  //! average pdfs of zeros and ones, based on the individual Gaussian
  //! approximations, and finally computes BER from these average pdfs.
  //! Outputs all the above statistics to  files with prefix FileID.
  //! Note FileID should be prefixed by JobDir and JobName.

  void OutputGaussianStatisticsFromMonteCarlo(string FileID);
  void SetTimeSlices(void);
  void ComputeAndOutputMomentsFromMonteCarlo(string FileID);
  void ComputeAveragePDFs(void);
  void WriteFilePDFTimeSlicesAndAveragePDFs(string FileID);
  void ComputeMinimumBERAndDecisionThreshold(string FileID);
  void WriteFilePairVectors(double * x, double *y, int dim, 
					  string OutFileName);

  //! Mean of current in bit with index BitIndex at clock recovery time.
  //! Must first call UpdatePerformanceStatistics(). For Monte Carlo.

  double GetPowerMeanBit(int BitIndex);
         

 
  //! Standard deviation of current in bit with index BitIndex 
  //! at clock recovery time.
  //! Must first call 
  //! UpdatePerformanceStatistics(). For Monte Carlo.

  double GetPowerStdDevBit(int BitIndex);
        

  //! Number of zeros. First call ComputePerformanceStatistics() or 
  //! UpdatePerformanceStatistics()

     double GetQtBit0(void) {return qtBit0;};

  //! Number of ones. First call ComputePerformanceStatistics() or 
  //! UpdatePerformanceStatistics()

     double GetQtBit1(void) {return qtBit1;};

  //! Maximum current in zeros. First call ComputePerformanceStatistics() or 
  //! UpdatePerformanceStatistics()

     double GetBit0_PowerMax(void) {return Bit0_PowerMax;};

  //! Minimum current in zeros. First call ComputePerformanceStatistics() or 
  //! UpdatePerformanceStatistics()


     double GetBit1_PowerMin(void) {return Bit1_PowerMin;};
      
  //! Mean of current in zeros at clock recovery time
  //! Must first call ComputePerformanceStatistics() or 
  //! UpdatePerformanceStatistics()

      double GetBit0_PowerMean(void) 
         {return Bit0_PowerSum/qtBit0;}; 

  //! Mean of current in ones at clock recovery time
  //! Must first call ComputePerformanceStatistics() or 
  //! UpdatePerformanceStatistics()

      double GetBit1_PowerMean(void) 
         {return Bit1_PowerSum/qtBit1;}; 

  //! Standard deviation of current in zeros at clock recovery time
  //! Must first call ComputePerformanceStatistics() or 
  //! UpdatePerformanceStatistics()

      double GetBit0_PowerStdDev(void)
         {return sqrt( (Bit0_PowerSquareSum
                      -qtBit0*sq(GetBit0_PowerMean()))
                      /(qtBit0 - 1.) );};

  //! Standard deviation of current in ones at clock recovery time
  //! Must first call ComputePerformanceStatistics() or 
  //! UpdatePerformanceStatistics()


      double GetBit1_PowerStdDev(void)
         {return sqrt( (Bit1_PowerSquareSum
                      -qtBit1*sq(GetBit1_PowerMean()))
                      /(qtBit1 - 1.) );};

      double GetElecTimeDomainSNR(void) 
         {return (GetBit1_PowerMean() - GetBit0_PowerMean() )
                 /GetBit0_PowerMean();};

//! Returns the Q-factor computed from the mean and the standard
//! deviation of the marks and spaces
//! By Ivan Lima

  double GetQ_Factor(void);
		
//! Returns the bit error rate computed from the mean and the standard
//! deviation of the marks and spaces
//! By Ivan Lima

  double GetBitErrorRate(void) {return 
                exp(-sq(GetQ_Factor())/2)/(GetQ_Factor()*sqrt(2*pi));};

//! Returns the bit error rate computed from an input Q-factor
//! By Ivan Lima

  double GetBitErrorRate(double Q_Factor) {return 
                exp(-sq(Q_Factor)/2)/(Q_Factor*sqrt(2*pi));};

  double GetAmplitudeMargin(void) {return AmplitudeMargin;};
    
  double GetMeanPower(void) {return MeanPower;};
            
  int GetBitStringShift(void) {return BitStringShift;};


  //! Statistics obtained using isolated zeros and ones

      double GetQtIsolatedBit0(void) {return qtIsolatedBit0;};
      double GetQtIsolatedBit1(void) {return qtIsolatedBit1;};
      double GetIsolatedBit0_PowerMax(void) {return IsolatedBit0_PowerMax;};
      double GetIsolatedBit1_PowerMin(void) {return IsolatedBit1_PowerMin;};

      double GetIsolatedBit0_PowerMean(void) 
         {return IsolatedBit0_PowerSum/qtIsolatedBit0;}; 

      double GetIsolatedBit1_PowerMean(void) 
         {return IsolatedBit1_PowerSum/qtIsolatedBit1;}; 

      double GetIsolatedBit0_PowerStdDev(void)
         {return sqrt( (IsolatedBit0_PowerSquareSum
                      -qtIsolatedBit0*sq(GetIsolatedBit0_PowerMean()))
                      /(qtIsolatedBit0 - 1.) );};

      double GetIsolatedBit1_PowerStdDev(void)
         {return sqrt( (IsolatedBit1_PowerSquareSum
                      -qtIsolatedBit1*sq(GetIsolatedBit1_PowerMean()))
                      /(qtIsolatedBit1 - 1.) );};

      double GetElecTimeDomainSNR_IsolatedBits(void) 
         {return (GetIsolatedBit1_PowerMean() - GetIsolatedBit0_PowerMean() )
                 /GetIsolatedBit0_PowerMean();};

      double GetQ_FactorFromIsolatedBits(void); 
             
					   
//! Constructs auxilliary optical and the electrical filters that will
//! be used to compute the Gaussian statistics. 


void SetOptElecFiltersForGaussianStatistics(
        OptFilter *oOptFilter,ElecFilter *oElecFilter);

//! Use this method if you want to use the same ElecSignalStat object
//! to compute the GaussianStatistics for oOptSignal
//! at different times during your simulation (JZ, April 2003).
//!
//! Call each time system parameters change    

 void ResetOptElecFiltersForGaussianStatistics(void);

//! To be used in conjunction with
//! ResetOptElecFiltersForGaussianStatistics  (JZ, April 2004)

void ResetOptElecFilterBandwidthsForGaussianStatistics(
			   double FreqFWHM_OptFilter2, 
                           double f3dB_ElecFilter2);

void CheckGaussianStatisticsInitialization(string CallingMethod);


//! Returns mean of noise <i_n> given by JLT05_22

double GetMeanNoiseGaussianStatistics(double TotalNoiseSpectralDensity);

//! Returns noise-noise beating variance given by JLT05_24

  double GetNoiseNoiseVarianceGaussianStatistics(
		 double TotalNoiseSpectralDensity,
		 double DOP_Noise);

 //! Returns I_{ASE-ASE} given by JLT05_26 using formula in JLT05_53
 //! This integral represents the factor in the noise-noise beating
 //! variance that is independent of the noise statistics, and only
 //! depends on the optical and electrical filters.

 double GetIntegralForNoiseNoiseVarianceGaussianStatistics(void);


 //! Returns the signal-noise beating variance in case noise is 
 //! unpolarized for an arbitrary signal. 

 double GetSignalNoiseVarianceForUnpolarizedNoiseGaussianStatistics(
          double Time,
          double TotalNoiseSpectralDensity);

//! Returns the signal-noise beating variance at time Time 
//! for a polarized VECTOR signal in case of (possibly) partially 
//! polarized Gaussian white noise, using JLT05_32.

 double GetSignalNoiseVarianceForPolarizedSignalGaussianStatistics(
          double Time,
	  double * UnitStokesSignal,
	  double * UnitStokesNoise,
          double TotalNoiseSpectralDensity,
	  double DOP_Noise);

 //! Returns I_{S-ASE}(t) given by JLT05_29 using formula in JLT05_54
 //! where e_(\tau) is replaced by OptSignalA shifted right by TimeShiftA
 //! and e_(\tau') replaced by OptSignalB shifted right by TimeShiftB
 //! and Time is the time t.

 double GetIntegralForSignalNoiseVariance(OptSignal *OptSignalA,
					  OptSignal *OptSignalB, 
					  double TimeShiftA, double TimeShiftB, double Time);


      double GetEnhancementFactor(double OSA_NoiseBandwidth);
      double GetNormalizedEnhancementFactor(void);
       
      double GetESNR_Mark_GaussianStatistics(double NoiseSpectralDensity);
      
      double GetESNR_GaussianStatistics(double NoiseSpectralDensity);
      double GetNormKcoeff1_GaussianStatistics(void);
      double GetNormKcoeff0_GaussianStatistics(void);      
      double GetNormNumberNoiseModes_GaussianStatistics(void);
      double GetCurrentExtinctionRatio_GaussianStatistics(void);
      double GetQ_FactorFromESNR_GaussianStatistics( double ESNR);
      double GetQ_FactorFromESNR_GaussianStatistics( double ESNR,
                       double NoiseSpectralDensityX,
	   	       double NoiseSpectralDensityY);


      double GetQ_FactorGaussianStatisticsWorstPattern(
			double NoiseSpectralDensity);
      double GetQ_FactorGaussianStatisticsWorstPattern(
		   double NoiseSpectralDensityX,double NoiseSpectralDensityY);

      double GetQ_FactorGaussianStatisticsFirstPattern(
		 double NoiseSpectralDensityX,double NoiseSpectralDensityY);

      //! Added by Jonathan
      double GetQ_FactorGaussianStatisticsFromBER(double NoiseSpectralDensity);
      double GetQ_FactorGaussianStatisticsFromBER(double NoiseSpectralDensityX,
                                              double NoiseSpectralDensityY);
	  
      double GetBER_GaussianStatisticsFromDecisionThreshold(
				 double DecisionCurrent);
      double GetBER_GaussianStatistics(void);
      double MinimizeSqQ_FactorVersusBER(double Q_FactorTmp);
  
      double brentHu(double ax, double bx, double cx, double (*f)(double), 
		     double tol,double *xmin);

      double testsq(double ax);

	  
      double GetMeanASE_ASE(double NoiseSpectralDensity);
      double GetMeanASE_ASE(double NoiseSpectralDensityX,
                            double NoiseSpectralDensityY);

      double GetStdDevASE_ASE(double NoiseSpectralDensity);	      
      double GetStdDevASE_ASE(double NoiseSpectralDensityX,
                              double NoiseSpectralDensityY);
      				
      double GetStdDevSignal_ASE_BitSlot(int BitSlot,
                                     double NoiseSpectralDensity);
      double GetStdDevSignal_ASE_BitSlot(int BitSlot,	     
                                     double NoiseSpectralDensityX,
             		             double NoiseSpectralDensityY);
      double GetStdDevSignal_ASE_TimeIndex(int TargetIndexTime2,
                                 double NoiseSpectralDensityX,
         		         double NoiseSpectralDensityY);
				 
    
      double GetStdDevWorstBit1_GaussianStatistics(double NoiseSpectralDensity);
      double GetStdDevWorstBit1_GaussianStatistics(
                             double NoiseSpectralDensityX,
			     double NoiseSpectralDensityY);

      double GetStdDevWorstBit0_GaussianStatistics(double NoiseSpectralDensity);
      double GetStdDevWorstBit0_GaussianStatistics(
                             double NoiseSpectralDensityX,
			     double NoiseSpectralDensityY);

      double GetMinCurrentBit1(void);
      double GetMaxCurrentBit0(void);			     

      double GetMinCurrentBit1(int *BitSlotMinCurrentBit1);
      double GetMaxCurrentBit0(int *BitSlotMaxCurrentBit0);
      

      double GetCurrentFirstBit1(int *BitSlotCurrentFirstBit1);
      double GetCurrentFirstBit0(int *BitSlotCurrentFirstBit0);     

      double GetCurrentBitFromRecoveredSignal(int BitSlotCurrentBit);      
      double GetCurrentBit(int BitSlotCurrentBit);    
      
				  
      double GetMeanCurrentIsolatedBit1(void);
      double GetMeanCurrentIsolatedBit0(void);

 int GetQtPoints(void) {return qtPoints;};
 double GetDeltaTime(void) {return DeltaTime;};
 int GetBit(int BitSlot) {return oOptSignal->oBitString->GetBit(0,BitSlot);}; 
 double GetBitLength(void) {return BitLength;};
 int GetStringLength(void) {return StringLength;};

                
  // ###### METHODS FOR COMPUTING PATTERN STATS #####################

  // ######## Written by John Zweck, June 23rd 2002 #################

  //! Initializes parameters and memory if user is computing pattern stats
  /*! Called by the constructor      
      ElecSignalStat(string, OptSignal,Photodetector)
  */

  void InitPatternStats(void);

  //! Called by ElecSignalStat::InitPatternStats

  void AllocateMemoryForPatternStats(void);

  //! Called by ElecSignalStat::~ElecSignalStat()

  void ReleaseMemoryForPatternStats(void);

  //! Sets all pattern stat variables to zero.
  /*!
     Called by ElecSignalStat::ClearElecSignalStat(void) and 
     ElecSignalStat::InitPatternStats(void)
  */

  void ClearPatternStats(void);

//! Writes all bit patterns of length PatternLength to a file
//! together with their PatternIndicies which are the decimal
//! integers corresponding to each of the bit patterns

  void EnumeratePatterns(ofstream * OutFile);

 //! Converts the integer PatternIndex into binary format
 //! returning coefficient of pow(2,BitIndex) in this binary expansion


  bool GetPattern(int PatternIndex,int BitIndex);

  //! Returns the center bit of the pattern with index PaternIndex

  bool GetCenterBitPattern(int PatternIndex);

  //! Returns the PatternIndex correpsonding to the bit string given by
  //! LocalPatternArray

  int GetPatternIndex(bool * LocalPatternArray, int LengthOfPattern);

  //! Call this function once per Monte Carlo run for each bit.
  /*! 
      Called by ElecSignalStat::UpdateMoments

      Updates values of 
      - NumOccurencesPattern
      - VoltageSumPattern
      - VoltageSumSquaresPattern
      - MaximumVoltagePattern
      - MinimumVoltagePattern

  */

  void UpdatePatternStats(int BitIndex,double PowerElecSignal);


  //! Tests for equality of two bit patterns

  bool PatternEquality(bool * PatternArray1, bool * PatternArray2, 
                       int LengthOfPattern);


  //! Computes the means and standard deviations for the center bit
  //! in each pattern from the arrays VoltageSumPattern and 
  //! VoltageSumSquaresPattern
  /*! 
      Called by ElecSignalStat::WriteFilePatternStats(string)
   */

  void PreparePatternStatsForOutput(void);

  //! Writes the pattern stats to a file

  void WriteFilePatternStats(string OutFileName);

  // ## PUBLIC DATA

   OptSignal *oOptSignal;
   Photodetector *oPhotodetector;
   double *BitIntensity; 


  //########## PRIVATE  #############################

 private:

  // ## PRIVATE METHODS 

 void InitElecSignalStat(OptSignal * oOptSignal2, 
                     Photodetector *oPhotodetector2);
      

  //! A version of ElecSignalStat::ComputePerformanceStatistics
  //! that does not involve time-shifting the signal. 
  //! A private function that should ONLY be used by semianalytical methods

  void ComputePerformanceStatisticsNoTimeShift(void);

//! Internal function called by UpdatePerformanceStatistics

 void UpdateMoments(int BitIndex,double PowerElecSignal);

 void ComputeTargetFreqIndex(void);

//! Internal function that returns integer index of the clock recovery time

 int GetTargetIndexTime(void);


//! Gets the double-precision target time associated with the clock recovery.  

 double GetTargetTime(void);

 double ComputePhaseTargetFreq(void);
 int GetTargetIndexTimeBestQ_Factor(void);
  int GetTargetIndexTimeBestMinEyeOpening(void);

 //! Internal Function 
 void UpdateClock(void);

  //## PRIVATE DATA 

 int ObjectInitialized;
 bool UseAbsCurrentFlag;

 cfftw *fft;
 fftPowerMatrices sfftPM;   
 typePulse TypePulse;
 int qtPoints;
 int qtPointsLast;        
 double  DeltaTime, TimeWindow;
 double DeltaFreq;
 int StringLength;
 double BitLength;                    
 double BitRateChann;

 typeClockRecovery TypeClockRecovery;

 int UseSmoothingToRecoverClockFlag;
 double GaussianSmoothingFWHM;

 int BitStringShift; 
  
 double InputClockRecoveryTime;
 double ClockRecoveryTime;
 double ClockRecoveryTimeBeforeTimeShift;

 int TargetFreqIndex;
 int FirstTargetIndexTime;
 int SumRelTargetIndexTime;

 int NumCallsToUpdateClock; // For UpdateClock()
 double SumTargetIndexTimesForUpdateClock;
           
 int qtRunsEyeDiagram;
 int qtRunsEyeDiagramFixedCLK;
      
 double Q_Factor;
 double BitErrorRate;      
 double AmplitudeMargin;
 double MeanPower;      
 
 int qtBit0, qtBit1;      
 double Bit0_PowerSum,Bit1_PowerSum;
 double Bit0_PowerSquareSum, Bit1_PowerSquareSum; 
 double Bit0_PowerMax, Bit1_PowerMin;

 int qtIsolatedBit0, qtIsolatedBit1;      
 double IsolatedBit0_PowerSum,IsolatedBit1_PowerSum;
 double IsolatedBit0_PowerSquareSum, IsolatedBit1_PowerSquareSum; 
 double IsolatedBit0_PowerMax, IsolatedBit1_PowerMin;

 // Variables for the semi-analytical method

 bool InitializedGaussianStatisticsFlag;
 double *VariancesCurrentSignalNoiseBeating; 
      
 OptSignal *oOptSignal1; 
 OptSignal *oOptSignal2; 
 OptSignal *oOptSignal3;
 OptSignal *oOptSignalX;

 OptFilter *oOrigOptFilter;
 OptFilter *oOptFilter2;
 Photodetector *oPhotodetector2;
 ElecFilter *oElecFilter2;

 ElecSignalStat *oElecSignalStat2; 
	
 // add by Jonathan Hu
      
 OptSignal *oOptSignal4;
 OptFilter *oOptFilter4;
 Photodetector *oPhotodetector4;
 ElecFilter *oElecFilter4;
 ElecSignalStat *oElecSignalStat4;  
 double BER_GaussianStatistics;
 double Q_FactorPre;
 double NoiseSpectralDensityX_PD;
 double NoiseSpectralDensityY_PD;

  // Variables for DoGaussianPDFsMonteCarlo

  bool DoGaussianPDFsMonteCarloFlag;
  int NumTimeSlices;
  double * TimeSlices;
  int * TimeSliceIndices;
  int * BitString;

  double * StdDevCurrentMonteCarlo;
  double * MeanCurrentMonteCarlo;

  double * PowerSumBit;
  double * PowerSquareSumBit;
  int * qtBit;

  double  MinCurrent,MaxCurrent,DeltaCurrent;
  int NumPointsCurrent;
  double * CurrentVector;

  double * BERGaussianPDFs;
  double * AverageGaussianPDFZeros;
  double * AverageGaussianPDFOnes;
  double ** GaussianPDFTimeSlices;
  double * GaussianPDFTimeSlicesStorageVec;


	   // ####### VARIABLES FOR PATTERN STATS ###################

  //! Flag is 1 if user wants to compute stats for patterns. 
  //! Initialized from input parameter file by constructor

  bool ComputePatternStatsFlag;

  //! Lenth of the patterns for which stats are to be computed. Must be odd
  //! Initialized from input parameter file by constructor

  int PatternLength;

  //! The number of patterns which is pow(2,PatternLength)

  int NumPatterns;

  //! A bool array to store a pattern

  bool * PatternArray;

  //! Array of the means of the patterns, indexed by the PaternIndicies.
  //! The mean of a pattern is the mean of the voltage in the center bit of the
  //! given pattern


  double * MeanPattern;

   //! Array of the standard devaitions of the patterns, 
  //! indexed by the PaternIndicies.
  //! The std dev of a pattern is the std dev of the voltage 
  //! in the center bit of the given pattern

  double * StdDevPattern;

  //! Stores the total number of occurneces of each pattern.

  int * NumOccurencesPattern;

  //! Array to store the sums of the voltages for each pattern

  double * VoltageSumPattern;      


  //! Array to store the sum of the squares of the voltages for each pattern

  double * VoltageSumSquaresPattern;      

  //! Array to store the maximum voltage for each pattern
 
  double * MaximumVoltagePattern;

  //! Array to store the minimum voltage for each pattern
 
  double * MinimumVoltagePattern;
       
};

#endif /*  _OCS_ELEC_SIGNAL_STAT_HH_ */
