// $Id: ocsHistogram2D.hh,v 1.3 2008/08/08 02:20:44 zweck Exp $


//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_HISTOGRAM_2D_HH_
#define  _OCS_HISTOGRAM_2D_HH_

#include "ocsOptSignal.hh"
#include "ocsPhotodetector.hh"

/*! \brief Class to construct histogram of a pair of random variables

    For use in computing two-dimensional histograms, particularly 
    in Monte Carlo sampling. The main applications the code was written
    for include construction of eye diagrams in optical communications
    systems and quantification of statistics related to random polarization 
    effects using importance sampling. However the class is generic enough
    to be used in other situations. 

    For simple applications call as follows: 

    \arg Histogram2D MyHist("Histogram2D.in");
    \arg for i=1 to N, Generate samples x_i, y_i 
         and call MyHist.AddSample(x_i,y_i)
    \arg MyHist.WriteFile_pdf2D("TwoDimPDF.dat");

    For applications to eyediagrams call as 

    \arg Histogram2D MyHist("Histogram2D.in",Signal);
    \arg MyHist.UpdateEyeDiagram(Signal,oPhotodetector);
    \arg MyHist.WriteEyeDiagramHistogram("EyeDiagram.dat");

    The two random variables are refered to as x and y throughout the code.

    If we think of x and y as being on a 2D coordinate system as usual
    then the bins are indexed 
    by a pair of integers that describes the position of the lower left
    corner of each bin, with the indices starting at (0,0). 

    So x-index goes from 0 to NumBinsX - 1.

    Samples out of range are discarded and their number is tracked. 

    For importance sampling applications the pdfs for each bias choice
    are first written to files using WriteFile_pdf2D in one application. 
    Then in a second aplication they are read back in and merged together
    usually using the balanaced heuristic method by calling a method like
    MultipleHeuristicBalanceImport. Finally the merged pdf is written out
    once more using WriteFile_pdf2D 

*/

class Histogram2D
{

public:

  //! Constructor

  /*! \sa x_min, x_max, y_min, y_max 
      If you update this constructor update other two as well as
       copy constructor and assignement operator*/

  Histogram2D(string InFileName,float x_min2,float x_max2,
                                float y_min2, float y_max2);

  //! Internal method called by constructors

  void InitializeParameters(void);

  //! Constructor

  Histogram2D(string InFileName);

  // Note: The following method should really belong to a derived 
  // EyeDiagram class of the Histogram2D class.

  //! Constructor to be used when making eye diagram 

  Histogram2D(string InFileName,OptSignal *oOptSignal);

  //! Copy constructor

  Histogram2D(Histogram2D & OrigHist);

  //! Assignment operator

  Histogram2D & operator=(Histogram2D & RHSHist);

  //! Destructor

  ~Histogram2D();

  /*! \brief
  Sets all dynamic variables to their initial state. For use when
  want to output multiple independent histograms during course of a simulation
 */

  void Clear();

  //! \sa x_max

  void SetMaxX(double x_max2); 
 
  //! \sa y_max

  void SetMaxY(double y_max2); 
 
  void SetMinMaxXY(double x_min2, double x_max2,
		   double y_min2, double y_max2);
  
 float GetMaxX(void) {return x_max;};
   
 float GetMaxY(void) {return y_max;};

 float GetMinX(void) {return x_min;};
 float GetMinY(void) {return y_min;};

 float GetDeltaX(void) {return delta_x;};
 float GetDeltaY(void) {return delta_y;};
 int GetNumBinsX(void) {return NumBinsX;};
 int GetNumBinsY(void) {return NumBinsX;};

  //! \sa SamplesMaxX

  float GetSamplesMaxX(void) {return SamplesMaxX;};

    //! \sa SamplesMaxY

  float GetSamplesMaxY(void) {return SamplesMaxY;}; 

  //! Internal method

  void SetBin(int x_index, int y_index, double value);
  
  double GetBin(int x_index, int y_index);
  double GetBinSquare(int x_index, int y_index);  
  int GetIndex(int x_index, int y_index)
     { return   y_index + NumBinsY*x_index;};   

  //! Internal method

  double Get_pdfConfidence(int x_index, int y_index);

  //! Computes width of confidence interval for each bin and stores
  //! result in 2D array pdfConfidence

  void ObtainConfidence(void);
  
  int GetQtSamplesBin(int x_index, int y_index);

  void IncrementBin(int x_index, int y_index, double LikelihoodRatio);

  //! Returns total number of samples \sa NumSamplesTotal

  int GetQtSamples(void) {return NumSamplesTotal;}; 

  //! Method to add a sample in a standard Monte Carlo simulation

  void AddSample(double x, double y);

  //! Method to add sample in a Monte Carlo simulation with importance sampling

  void AddSample(double x, double y, double LikelihoodRatio);  


  //! Writes out pdf computed by the histogram to a file. 
  
  /*!
      Each line of the files corresponds to a different bin.
      On each line we have the following entries 

      -# x-value of center of bin
      -# y-value of center of bin
      -# pdf value in that bin (ie the mean)
      -# standard deviation in that bin, computed from the 
         pdf of the confidence
      -# standard deviation relative to the mean
  */


  void WriteFile_pdf2D(string OutFileName);

  /*! \brief Returns mean of X calculated from samples that are in range
      \warning Not correct for importance sampling!
   */

  double GetMeanX(void);

  /*! \brief Returns mean of Y calculated from samples that are in range
      \warning Not correct for importance sampling!
   */

  double GetMeanY(void);


  //! For applications to eye diagrams in Monte Carlo simulations

  /*! We use real part of current when making eye diagram */

  void UpdateEyeDiagram(OptSignal *oOptSignal,Photodetector *oPhotodetector);


  /*! 
   Outputs 2D histogram for eye diagram applications
  */

  void WriteEyeDiagramHistogram(string OutFileName);


  void MultipleSelectMaxRelNumHitsImport(string InFileName,
                                         int Convert2dB_Scale);					 
  void GetLevelCurvesMultipleSelectMaxRelNumHitsImport(string InFileName,
                                               int Convert2dB_Scale,
    					       int CosnvertParam2dB_Scale);

  //! Use this function for merging 2D histograms named in InFileName

  void MultipleHeuristicBalanceImport(string InFileName,
                                       int Convert2dB_Scale);

  /*! Use this function for merging 2D histograms named in InFileName
     Computes the 2D probability density function 
     f(X=x,Y=y| z_min < Z < z_max), the joint probability density 
     for X and Y given that z_min < Z < z_max. It is assumed
     that samples of the random variables X,Y,Z have been obtained from an 
     importance sampling simulation, and that the names of the files containing
     the samples are stored in the file InFileName. 
   */

  void MergeMultipleConditionalHeuristicBalance(string InFileName,
                                                double z_min, double z_max,
                                                int Convert2dB_Scale);

  /*! Use this function for merging 2D histograms named in InFileName 
     Computes the 2D probability density function 
     f(X=x,Y=y| z_min < Z < z_max, w_min < W < w_max), 
     Samples of the random variables X,Y,Z,W have been obtained from an 
     importance sampling simulation, and that the names of the files containing
     the samples are stored in the file InFileName. 
   */

  void MergeMultipleConditionalHeuristicBalance(string InFileName,
                                                double z_min, double z_max,
						double w_min, double w_max,
                                                int Convert2dB_Scale);


   void GetLevelCurvesMultipleHeuristicBalanceImport(string InFileName,
                                                  int Convert2dB_Scale,
						  int ConvertParam2dB_Scale);
						  
   void GetLevelCurvesMultipleHeuristicBalanceImportVarianceWeighted(
						  string InFileName,
                                                  int Convert2dB_Scale,
						  int ConvertParam2dB_Scale);


   //! For applications where want to manipulate PDF rather than histogram
   //! first call this method to compute PDF from histogram. 

   /*! PDF data is stored in separate 2D array double * PDF
       All integrals involving PDF data done using (iterated) Simpson's Rule
   */

   void ComputePDFfromHistogram(void);

   //! Normalize so integrates to 1. Useful after numerical
   //! operations are performed on PDF

   void NormalizePDF(void);

   //! \sa Histogram2D::ComputePDFfromHistogram()

   void SetPDF(int x_index, int y_index, double value);

   //! \sa Histogram2D::ComputePDFfromHistogram()

   double GetPDF(int x_index, int y_index);

   void ComputeXMarginalPDF(void);
   void ComputeYMarginalPDF(void);

   double ComputeIntegralPDF(void);
   double ComputeMeanX_PDF(void);
   double ComputeMeanY_PDF(void);

   //! Returns the maximum value of the PDF obtained using 
   //! \sa Histogram2D::ComputePDFfromHistogram()

   double GetMaxPDF(void);
 
   //! Sets values of PDF that are below the Threshold to zero
   //! and renormalizes PDF.

   double SetPDFToZeroBelowThreshold(double Threshold);

   //! Analytic formula for Gaussian PDF. For debugging purposes

   void GaussianPDF(double mean_x, double mean_y, double std);

   //! Writes out data in array PDF \sa Histogram2D::ComputePDFfromHistogram()

   void WritePDF(string OutFileName);

   double BilinearInterpolationPDF(double x, double y);

   //! Rotates PDF about (CenterX,CenterY) throuhg angle Angle
   //! and stores result in OutData.

   /*! OutData should be the 1D data array double * PDF from another 
       Histogram2D object. BilinearInterpolationPDF() is used to get a value
       at ecah rotated point */


   void RotatePDF(double CenterX, double CenterY, double Angle, 
                  double *OutData);

public:

  //! The 2D PDF array

  double * PDF;

private:

  // Parameters describing physical domain of pdf

  //! Minimum x value for 2D Histogram of x and y

  float x_min;

  //! Maximum x value for 2D Histogram of x and y


  float x_max;

  //! Minimum y value for 2D Histogram of x and y

  float y_min;

  //! Maximum y value for 2D Histogram of x and y

  float y_max;
 
  float x_window_size,y_window_size;

  //! Number of bins in X variable

  int NumBinsX;


 //! Number of bins in Y variable

  int  NumBinsY;

  float delta_x, delta_y;
  
  //! The 2D Histogram array

  double * Bin;
  int *qtSamplesBin;
  
  double * BinSquare;
  double * pdfConfidence;
  bool PreviouslyComputed_pdfConfidenceFlag;

  // Data to manipulate PDFs

  double * XData;
  double * YData;
  double * XMarginal;
  double * YMarginal;


  // Parameters relating to statistics of pdf

  int NumSamplesTotal;
  int NumSamplesInRange;
  int NumSamplesWithOutOfRangeX;
  int NumSamplesWithOutOfRangeY;

  // Used to determine size of physical grid for later runs

  //! Minimum X value of all samples (even if lie outside Histogram)

  float SamplesMinX; 

  //! Maximum X value of all samples (even if lie outside Histogram)

  float SamplesMaxX;

  //! Minimum Y value of all samples (even if lie outside Histogram)

  float SamplesMinY; 

  //! Maximum Y value of all samples (even if lie outside Histogram)

  float SamplesMaxY;

  //! Used by merge algorithms in importance sampling applications
  
  string HistInFileName; 

  double MeanX;
  double MeanY;

};

#endif /*  _OCS_HISTOGRAM_2D_HH_ */

