// $Id: ocsOptDemuxer.hh,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_OPT_DEMUXER_HH_
#define _OCS_OPT_DEMUXER_HH_


#include "ocsOptFilter.hh"
#include "ocsOptSignal.hh"


// ##### Created by John Zweck 19th Oct, 2000 ########

// An object of type class OptDemux takes an optical 
//  WDM Signal of type OptSignal and a tunable filter of type OptFilter
// and for extracts channels one at at time from the WDM Signal
// which can later be passed through a photodetector and analyzed.

// The private data consists of pointers to a WDM_Signal and a Tunable
// Filter which have been initialized in the application program. 
// It is important to note the the methods in OptDemuxer do not
// alter the WDM_Signal!! However they do alter the CenterFreq of the
// Tunable filter, which is course is why it's called "tunable"!

// The ExtractChannel method optically filters the WDM_Signal and returns 
// a reference to an OptSignal containing a single channel and with 
// CenterFreq given by the CenterFreq of the given channel.
// A typical application would use the OptDemuxer class as follows.
// Here we assume that WDM_Signal has been instantiated.

// Note that the same OptDemuxer object can be used to demux the
// WDM_Signal many times during the course of a simulation.
// The constructor is called once and then each time you want to 
// extract a channel you call ExtractChannel.
 
// FIX:

// At the moment the single channel signal extracted by ExtractChannel
// the correct NumChannels (=1) and the correct length (StringLength)
// and entries in its BitString. 
// However the Frequency, InStokes and InJones arrays have NOT yet been
// changed to be correct for the single channel signal.
// We can do this later if we need to using code similar in style to 
// BitString::ReduceToSingleChannelBitString

// (JZ) Dec 4th 2000:
// A receiver subsystem class has been added to the library.
// The optDemuxer is called automatically by receiver subsystem objects.
// To see how to call the OptDemuxer the simplest thing to do is
// to look in ocsReceiverSubsystem, or even better in ocsReceiverSemiAnalytical




//###########################################################
//## If you use the second constructor, in which both 
//#  OptSignal are specified, you can extract the 
//#  channel using "void ExtractChannel(int Channel)".
//#  In that case the memory for the single channel
//#  signal is allocated only once, but will always 
//#  contain the channel that was selected in 
//#  the method ExtractChannel(.). The single channel
//#  that will be passed in the constructor does not
//#  need to be allocated. It will be created as a
//#  hard copy of the WDM signal before the single 
//#  channel is extracted.
//#  I strongly recommend to use the ReceiverSubsystem 
//#  class instead of the OptDemuxer. The advantage of
//#  the ReceiverSubsystem is that it contains all 
//#  the electrical domain modules, besides containing 
//#  a demuxer object which is used to single channel
//#  extraction.
//#  By: Ivan Lima.
//##########################################################

class OptDemuxer
{

public:

  OptDemuxer(OptSignal * WDM_Signal2, OptFilter * TunableFilter2);

  ~OptDemuxer();

 //## Added by Ivan Lima (6/6/01)
 OptDemuxer(OptSignal* WDM_Signal2, OptFilter* TunableFilter2,
                       OptSignal* SingleChannelSignal2);


  OptSignal& ExtractChannel(int ChannelArrayIndex);

  //! Use this method to demux rather than using 
  //! OptSignal& ExtractChannel(int ChannelArrayIndex). JZApril 2003
  //! Based on void ExtractSingleChannel(int ChannelArrayIndex);

  void DemuxChannel(int ChannelArrayIndex);

  //! Does all of DemuxChannel expcept the actual filtering
  //! Used in ReceiverSemiAnalytical. JZApril 2003

  void PrepareToDemuxChannel(int ChannelArrayIndex);


  //! Ivan's code. Very similar to DemuxChannel.

  void ExtractSingleChannel(int ChannelArrayIndex);

  //! Written by JZ (Aug 2001) to demux about a given
  //! freq rather than a ch number.
  //! The BitString for the demuxed signal is given by SCBitString 

  OptSignal & ExtractChannelCenterFreq(double CenterFreqOptFilter, 
                                       BitString *SCBitString,
                                       bool DoFilterFlag);
 

  bool GetDoAdditionalFiltering(void) {return  DoAdditionalFilteringFlag;};
  

private:

  OptSignal * WDM_Signal;
  OptFilter * TunableFilter;

  OptSignal *SingleChannelSignal;

  int ChannelArrayIndex;

  bool DoAdditionalFilteringFlag;

};

#endif /* _OCS_OPT_DEMUXER_HH_ */
