// $Id: ocsOptFiberLocalError.hh,v 1.8 2008/08/08 02:21:34 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_OPT_FIBER_LOCAL_ERROR_HH_
#define _OCS_OPT_FIBER_LOCAL_ERROR_HH_

#include "ocsOptSignal.hh"
#include "ocsRamanAmp.hh"


//! The type of Raman Amplification
/*! An enumeration type
    \arg NO_RAMAN = 0 (Default)
    \arg SIMPLE_RAMAN = 1 (Done within OptFiberLocalError)  
    \arg FULL_RAMAN = 2 (Raman gain precomputed by class RamanAmp)
*/

enum typeRamanAmplification { NO_RAMAN = 0, 
                      	      SIMPLE_RAMAN = 1, 
                              FULL_RAMAN = 2 }; 


//! The type of loss or gain in the fiber
/*! An enumeration type

	\arg LOSSY_FIBER = 0 (Default) 
	\arg FIBER_AMPLIFIER = 1

    The OptFiberLocalError class allows one to model both
    fibers with loss (which for example are used in 
    telecommunications applications) and fiber
    amplifiers. 

    The fiber amplifier model is a simplified
    one which simply assumes the propagating signal undergoes
    a given (possibly saturable and frequency-dependent)
    gain per unit length
    along the fiber. For the two gain models we have implemented 
   
    \sa enum typeFiberAmplifierGain.

    When OptFiberLocalError::TypeFiberLossGain = FIBER_AMPLIFIER,
    noise may also be added prior to each step along the fiber.
    
    \sa  enum typeFiberAmplifierNoise for more details on this noise.

    For now OptFiberLocalError::TypeFiberLossGain = FIBER_AMPLIFIER 
    has only been verified when OptSignal::TypeSolver = SCALAR_NLS. 
*/

enum typeFiberLossGain {LOSSY_FIBER = 0,
                        FIBER_AMPLIFIER = 1};

/*! \brief The type of noise used when 
    OptFiberLocalError::TypeFiberLossGain = FIBER_AMPLIFIER

     An enumeration type
	\arg FIBER_AMP_NOISE_OFF = 0 (default value)
	\arg FIBER_AMP_NOISE_GAUSSIAN_WHITE = 2

    \see enum typeFiberLossGain for more details.

    We use the same noise types as in the OptAmplifier class.

    We need the step sizes along the fiber to be chosen small enough
    so that adding noise at start of each step gives a result
    that is close to what would occur if noise was added continuosly.
    For this to be the case, we would need that if we added noise
    with step size delta_z/2 we would get approximately the same noise
    power as if we used step size delta_z.  
    The noise power added in each step of size delta_z is 
    proportional to exp(delta_z*g)-1, where g is the gain coefficient
    in units of inverse length. So for the output noise power to
    be independent of step size we would  the approximation 
    exp(delta_z*g)-1 ~ delta_z*g to hold. So we need delta_z << 1/g.

    \sa OptFiberLocalError::FiberAmplifierNoiseMaxStepSizeFactor.

    When OptFiberLocalError::TypeFiberAmplifierGain == 
    FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN
    the gain g in the previous paragragh is the unsaturated electric field
    gain, OptFiberLocalError::FiberAmplifierUnsaturatedNepperGain/2.    

    \sa OptFiberLocalError::FiberAmplifierComputeNoiseAmplitudeFactor

*/

enum typeFiberAmplifierNoise {FIBER_AMP_NOISE_OFF = 0,
                              FIBER_AMP_NOISE_GAUSSIAN_WHITE = 2};


/*! The type of gain  used when 
    OptFiberLocalError::TypeFiberLossGain = FIBER_AMPLIFIER

     An enumeration type
	\arg FIBER_AMP_UNSATURATED_FLAT_GAIN = 0 (default value)
	\arg FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN = 1

    \see enum typeFiberLossGain for more details.

    When OptFiberLocalError::TypeFiberAmplifierGain == 
    FIBER_AMP_UNSATURATED_FLAT_GAIN the gain coefficient
    is the only parameter. We set the gain
    by choosing a negative value for the input parameter 
    OptFiberLocalError::AttenuationFiber in dB/m, or 
    NepperAttenuationFiber in m^(-1) or a value of TotalLinearAttenuationFiber
    that is less than 1. 

    When OptFiberLocalError::TypeFiberAmplifierGain == 
    FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN
    we replace the constant gain term in the NLS by a term of the form 

    u_z = -0.5*l + 
           0.5*g(z)*(u+ (-1)^{n+1}*(u^{(2n)}/
                    (FiberAmplifierSuperGaussFiltGainCoefficient^{2n}))) 

    where l is a loss coefficient and g(z) is the gain coefficient
    both in units of inverse length  and 
    OptFiberLocalError::FiberAmplifierSuperGaussFiltGainCoefficient 
    is in units of inverse time. Here the integer 
    n=FiberAmplifierSuperGaussFiltOrder is the order of the super Gaussian
    filter. n=1 corresponds to standard Gaussian filter, and is the default
    value. 

    \sa enum typeFiberLossGain

    We include the loss term to enable modeling of a laser using Haus's
    Master equation in which all effects are assumed to act continuously
    throughout the laser cavity. 
   (See for example "Noise of mode-locked lasers" by Haus and Mecozzi, 
    IEEE J. Quantum Electronics, 29 (3), March 1993, p983-996.)

    The loss parameter is set using the input parameter 
    AttenuationFiber, and is given in dB/m
    or using the input parameter 
    NepperAttenuationFiber in m^(-1) (these two must be
    non-negative) or equivalently a value greater than one for
    OptFiberLocalError::TotalLinearAttenuationFiber.
   
    The saturable power gain function is 
    
    g(z) = OptFiberLocalError::FiberAmplifierUnsaturatedNepperGainPerLength/
            (1+ OptSignal::GetTotalEnergy() /
               OptFiberLocalError::FiberAmplifierSaturationEnergy),

    where OptSignal::GetTotalEnergy() is the total energy 
    at point z along the fiber.
    
    Here FiberAmplifierUnsaturatedNepperGainPerLength
           = log(dB2Linear(FiberAmplifierUnsaturatedGaindBPerLength))
           = FiberAmplifierUnsaturatedGaindBPerUnitLength*log(10)/10

    We input one of three equivalent parameters:
    FiberAmplifierUnsaturatedGaindBPerLength (dB/m),
    FiberAmplifierUnsaturatedNepperGainPerLength, (m^(-1)) or
    FiberAmplifierUnsaturatedTotalLinearGain. 
    We also input FiberAmplifierSaturationEnergy (J)
    
    FiberAmplifierUnsaturatedTotalLinearGain = 
    exp(FiberAmplifierUnsaturatedNepperGainPerLength*FiberLength).

    (If we are in the unsaturated regime
    FiberAmplifierUnsaturatedGain plays the same role as -AttenuationFiber
    in the unsaturated fiber amplifier.)

    We use energy rather than average power to compute the gain since energy
    is independent of the time and frequency windows. 
    For communications simulations in which there is a fixed bit rate and
    half the bits are marks the average power will be independent of the 
    time window, but for laser systems in which a single pulse is being 
    simulated the average power will inversely proportional to the time
    window. 

    We implement this additional term by adding an imaginary term 
    to the chromatic gain coefficient in the method 
    OptFiberLocalError::ComputeFreqDomainOperator() 

    For a step along the fiber of size delta_z we use Heun's Method
    (aka Modified Euler Method or 2nd-order Runga-Kutta) to update 
    the field u in the frequency domain. We use a split-step algorithm
    in the frequency domain (1/2 step chromatic dispersion and loss,
    1 step saturable gain, 1/2 step chromatic dispersion and loss)
    to update the Fourier transform of the field u. The chromatic dispersion 
    and loss update is exact, the saturable absorption step is locally
    third-order accurate. Since the update in time (Kerr nonlinearity
    and fast saturable absorption) is also locally
    third-order accurate, the split-step algorithm gaurantees that the
    entire step is third-order accurate, and hence with constant step size
    the global error is
    second-order, ie, proportional to the square of the step size.
    We have verified this for the Haus-Mecozzi chirped soliton.

    To ensure the step sizes are small enough, in addition to any other
    step size selection criteria,
    we choose the step size small enough so that 
     Power(z+delta_z)/Power(z) < FiberAmplifierGainSaturationMaxGainPerStep

    This will hold if 

    delta_z < log(FiberAmplifierGainSaturationMaxGainPerStep)*2/
               FiberAmplifierUnsaturatedNepperGainPerLength. 

    \sa double FiberAmplifierGainSaturationMaxGainPerStep. 

    Note that we do not allow  TypeStepSizes=LOCAL_ERROR_3RD
    when TypeFiberAmplifierGain = FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN,
    since this would require the saturable absorption step to be
    locally 4-th order accurate, which we have not (yet) implemented.

*/

enum typeFiberAmplifierGain {FIBER_AMP_UNSATURATED_FLAT_GAIN = 0,
		             FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN = 1};	


//! An enumeration type for the functional form of the fast saturable absoprtion
/*! 
    \arg QUADRATIC_ANALYTIC = 0 (use analytical formula for exact solution
                                 of fast saturable absorber with a 
                                 quadratic nonlinearity)
    \arg QUADRATIC = 1 (fast saturable absorber with a quadratic nonlinearity
                        ie, du/dz = \delta |u|^2 u. In each step of our
                        nested split-step scheme, we use Heun's method
                        to solve this equation numerically.)

*/

enum typeFastSaturableAbsorption {QUADRATIC_ANALYTIC=0,QUADRATIC=1};


//! An enumeration type for the step size selection criterion
/*! See Oleg Sinkin's JLT paper for details

    \arg LOCAL_ERROR_3RD = 1 (3rd order local error method)
    \arg LOCAL_ERROR_2ND = 2 (2nd order local error method)
    \arg WALK_OFF = 3 
    \arg CONSTANT = 4

    The local error options are explained in documentation for the
    class OptFiberLocalError.
    The method is controlled by two input parameters, 
    OptFiberLocalError::RelativeErrorGoal and OptFiberLocalError::DeltaZMax.

    \sa OptFiberLocalError::ComputeDeltaZMax()

    For CONSTANT step sizes the user specifies the step size by setting the
    parameter OptFiberLocalError::NumStepsBetweenScatterings. 

    In simulations without PMD, the step size is then set to be 
    OptFiberLocalError::DeltaZ = 
    OptFiberLocalError::LengthFiber/NumStepsBetweenScatterings.

    In simulations with PMD, the step size is then set to be 
    OptFiberLocalError::DeltaZ = 
    OptFiberLocalError::ScatteringStepSize/NumStepsBetweenScatterings,
    where ScatteringStepSize is an input parameter that specifies the
    distance between random rotations on the Poincar'e sphere, i.e. the
    lengths of the sections in the coarse-step method for solving the
    Manakov PDE.

    For the WALK_OFF method the user specifies 
    OptFiberLocalError::WalkOffParameter in the input parameter file.

    OptFiberLocalError::DeltaZ = 
    OptFiberLocalError::WalkOffParameter / abs( D_1 lambda_1 - D_2 lambda_2 )

    where D_1 and D_2 are the dispersions in ps/nm-km at wavelengths
    lambda_1 and lambda_2.

    We choose lambda_1,2 as follows:

    WDM Case: lambda_1,2 are the central wavelengths of the edge channels

    Single Channel Case: lambda_1,2 correspond to frequencies

                          freq_1,2 =  CenterFreq +- 
                                       RootMeanSquare spectral width of signal.

    Of course for the last step we may need to choose a smaller step size
    so that the overall propagation distance is correct.

*/

enum typeStepSizes { LOCAL_ERROR_3RD = 1, 
		     LOCAL_ERROR_2ND = 2, 
		     WALK_OFF = 3,
		     CONSTANT = 4 };

//! An unnamed enumeration type used in 
//! OptFiberLocalError::ApplyFreqDomainOperator

enum { APPLY_ONCE = 0, APPLY_TWICE = 1 };

/*! 

\brief The OptFiberLocalError class solves the scalar and vector generalized
       nonlinear Schroedinger equation (NLSE) and the Manakov-PMD equation
       for propagation of an optical signal in optical fibers. 
       It also solves the
       scalar Haus master equation (HME) for propagation in lasers. 
       Users are encouraged
       to first read the document  
<a href = "http://www.umbc.edu/photonics/home/members/jzweck/PhoSSiL/OFLE.pdf">
On the use and validation of the OptFiberLocalError class in PhoSSiL</a>

  This class is used to propagate an optical signal through a fiber
  with first- and second-order chromatic dispersion, Kerr nonlinearity, 
  fiber loss, fiber amplification (with optional saturable gain,
  parabolic gain shape and noise), fast saturable absorption,
  polarization-mode dispersion (PMD), and Raman amplification.

  The class allows for SCALAR or VECTOR signals, although not all 
  physical effects are verified/implemented for VECTOR signals.

  \warning The implementation of PMD in this class is quite rudimentary
  and has not been carefully tested. Users are instead encouraged to use
  the class OptFiber for modeling of PMD, which can be done there
   in conjunction with  first- and second-order chromatic dispersion, 
   Kerr nonlinearity, and fiber loss.

  \sa enum      typeSolver
  \sa enum  	typeFiberLossGain
  \sa enum  	typeFiberAmplifierGain
  \sa enum  	typeFiberAmplifierNoise 
  \sa enum  	typeRamanAmplification
  \sa RamanAmp
  \sa OptFiber

  This class was written by John Zweck, based on the following
references:

-# O.V. Sinkin, R. Holzloehner, J. Zweck, and C.R. Menyuk, 
 "Optimization of the split-step Fourier method in modeling
 optical fiber communications systems", 
 Journal of Lightwave Technology, vol. 21, no. 1, pp. 61-68, 2003.
-# C.R. Menyuk, "Application of Multiple Length-Scale 
Methods to the Study of Optical Fiber Transmission," 
J. Engin. Math. vol. 36, pp. 113-136, 1999. 
-# H.A. Haus, J.G. Fujimoto, E.P. Ippen,
"Structures for additive pulse mode locking", J. Opt. Soc. Am., vol. 8,
no. 10, pp 2068-2076, 1991.
-# H.A. Haus and A. Mecozzi, "Noise of mode-locked lasers",
IEEE J. Qunatum Elec., vol. 29, no. 3, p 983-996, 1993.

\section IPF Example Input Parameter Files

<a href = "http://www.umbc.edu/photonics/home/members/jzweck/PhoSSiL/infiles/SMF.in">SMF.in</a>

<a href = "http://www.umbc.edu/photonics/home/members/jzweck/PhoSSiL/infiles/HausMasterEquationFiber.in">HausMasterEquationFiber.in</a>

\section USE Example Application Code

The files linked below are an application code with input parameter files
for a simple application of the OptFiberLocalError class. This application
solves the Haus Master Equation using an initial chirped sech pulse.
In the references above Haus gives an analytical formula for the solution
which we have coded up in the matlab files below. The agreement with PhoSSiL
is excellent. The matlab code should first be used to determine the
parameters of the input chirped sech pulse for a given choice of fiber
parameters.

<a href = "http://www.umbc.edu/photonics/home/members/jzweck/PhoSSiL/ExampleApps/src/SechLaser.cc">SechLaser.cc</a>

<a href = "http://www.umbc.edu/photonics/home/members/jzweck/PhoSSiL/ExampleApps/src/StartupSechLaser.cc">StartupSechLaser.cc</a>

<a href = "http://www.umbc.edu/photonics/home/members/jzweck/PhoSSiL/ExampleApps/include/StartupSechLaser.hh">StartupSechLaser.hh</a>

<a href = "http://www.umbc.edu/photonics/home/members/jzweck/PhoSSiL/ExampleApps/infiles/SechLaserInfiles.tar.gz">Input parameter files [.tar.gz]</a>

<a href = "http://www.umbc.edu/photonics/home/members/jzweck/PhoSSiL/ExampleApps/matlab/SechLaserMatlabFiles.tar.gz">Matlab Files [.tar.gz]</a>

The files linked below show how to model a realistic optical fiber 
communications system, loosely based on a WDM system of Tyco's from about 1998.

The file LinearnTyco.cc contains several functions. The simplest is
the first, TycoNoiseFree().

<a href = "http://www.umbc.edu/photonics/home/members/jzweck/PhoSSiL/ExampleApps/src/LinearnTyco.cc">LinearnTyco.cc</a>

<a href = "http://www.umbc.edu/photonics/home/members/jzweck/PhoSSiL/ExampleApps/src/StartupLinearn.cc">StartupLinearn.cc</a>

<a href = "http://www.umbc.edu/photonics/home/members/jzweck/PhoSSiL/ExampleApps/include/StartupLinearn.hh">StartupLinearn.hh</a>

<a href = "http://www.umbc.edu/photonics/home/members/jzweck/PhoSSiL/ExampleApps/infiles/TycoLinInfiles.tar.gz">Input parameter files [.tar.gz]</a>


\section SSALG Step Size Selection Algorithms 

 To explain the step-size selection
  algorithms used we first consider the case that there
 is no PMD, ie when OptFiberLocalError::TypeSolver = SCALAR_NLS, or 
 OptFiberLocalError::TypeSolver = VECTOR_MANAKOV_NO_PMD. 
 We use the second-order symmetric
 split-step scheme to solve the PDE, as explained in Agrawal's book and in 
 Oleg Sinkin's JLT paper referenced above. 

 The input parameter OptFiberLocalError::TypeStepSizes specifies 
 which step size selection algorithm to use.  
 
 The WALK_OFF and CONSTANT step size algorithms are explained in 
 OptFiberLocalError::TypeStepSizes.

 We now explain the LOCAL_ERROR_3RD and LOCAL_ERROR_2ND step size 
 selection criteria. We assume the user has read Oleg Sinkin's JLT
 paper on these algorithms referenced above. 

 The most important feature of the algorithm is that for each step 
 along the fiber the step size is automatically chosen to satisfy
 a local error bound. 

 We first take a coarse split-step step of size 
 OptFiberLocalError::DeltaZ, which we compare
 to the result of taking two fine steps each of size DeltaZ/2 starting
 from the same point that we started the coarse step from. 

 We compare the coarse step and fine step solutions and compute the
 local L^2 error of that step which is 

 OptFiberLocalError::RelativeError = sqrt( integral | coarse - fine | ^2 \
                                           integral | fine | ^2 ).

 The integral is done in the frequency domain over the number of modes
 OptSignal::RelevantNumModesOFLE, which is computed in 
 void OptSignal::ComputeRelevantNumModesOFLE(void)

 \warning If you have one channel which is not at or very close to
 the center frequency set 
 OptSignal::ComputeRelevantNumModesOFLEFlag = 0 due to a bug in
 void OptSignal::ComputeRelevantNumModesOFLE(void) in this situation.

 If TypeStepSizes == LOCAL_ERROR_3RD then the solution at the end
 of the step is set to be:

 third-order solution = 4/3 coarse solution - 1/2 fine solution.

 If TypeStepSizes == LOCAL_ERROR_2ND then the solution at the end
 of the step is set to be the fine solution.

 The step sizes are controlled using a single input parameter
 OptFiberLocalError::RelativeErrorGoal.

 The value of RelativeError is used to decide whether to accept the step
 or reject it.

 If RelativeError > 2.0*RelativeErrorGoal 
    -# Reject the step
    -# Halve the step size
    -# Throw away the solution and go back to the beginning of the step

 Else
   -# Accept the step and adjust the step size for the next step.
   -# If RelativeError > RelativeErrorGoal
      -# Reduce DeltaZ by a factor of sqrt(2)
   -# Else 
       -# If RelativeError < 0.5*RelativeErrorGoal
            -# Increase DeltaZ by a factor of sqrt(2)
       -# Else
            -# Keep DeltaZ the same

 Each time we propagate through a particular fiber, we first try to
 take a step of size  OptFiberLocalError::DeltaZInitial. (A step with
 this size may or may not be accepted.)

 We set DeltaZInitial as follows. The very first time through the
 fiber the first step size we try is

 DeltaZInitial = 0.5*OptFiberLocalError::DeltaZMax. (This is very ad-hoc).

 \sa OptFiberLocalError::ComputeDeltaZMax()

 At the end of the first step we actually take in the fiber, 
 DeltaZInitial is set to the coarse step size that we actually took.

 The next time through the fiber we try a coarse step of size DeltaZInitial.

 One final point. At the start of each step we make sure that the
 coarse-step step size DeltaZ does not exceed the input parameter
 OptFiberLocalError::DeltaZMax. We also make sure that the total propagation
 distance does not exceed OptFiberLocalError::LengthFiber.

 Overall, we need 6 FFTs per step: 2 for the coarse step plus 4 for the fine 
 step, plus 1 to update the time domain when the end of the 
 fiber leg is reached.

 \section PMD Modeling PMD

 For simulations with PMD (TypeSolver = VECTOR_MANAKOV_PMD) we use the
 following adaptation by Zweck and Marks of the coarse-step algorithm of 
 Marcuse, Menyuk and Wai (and others). 

 The basic idea is to divide up the fiber into sections. Within each section
 we propagate as in the case of TypeSolver = VECTOR_MANAKOV_NO_PMD as 
 described above, using one of the four options for TypeStepSizes.
 At the start of each section we randomly rotate light on the Poincar'e sphere.

 Also, if TypeSolver = VECTOR_MANAKOV_PMD, we model the birefringence as
 follows. In each step of the split-step algorithm we implement this walk-off
 between orthogonal polarization states in the frequency domain in the
 method OptFiberLocalError::ComputeFreqDomainOperator().

 For a single section without a random rotation the X-polarization state
 is the fast axis and the Y-polarization the slow axis. This means that
 if we launch with Stokes parameters (1,0,0), i.e, X-polarized, then
 the signal is shifted left by PMD/2. Why left? Well we work with retarded
 time, t_ret = t - v_g z. If we launch in the fast axis, then v_g is larger
 and so t_ret of the center of a pulse will be smaller than would be the 
 case without PMD.
 
 We introduce a new parameter OptFiberLocalError::TotalLengthFiber =
 OptFiberLocalError::LengthFiber. The lengths of the sections are
 determined from the input parameter OptFiberLocalError::ScatteringStepSize
 in the method OptFiberLocalError::SetFiberLengthsForPMD().
 This method sets the lengths of the sections using the variables

    -# OptFiberLocalError::NumberScatterings
    -# OptFiberLocalError::LengthFiberExceptLast
    -# OptFiberLocalError::LengthFiberLast

 OptFiberLocalError::SetFiberLengthsForPMD() also allocates memory
 to store the Euler angles parametrizing the random rotation for
 each section, and calls OptFiberLocalError::GetNewFiberRealization()
 to get a new realization of the fiber rotations. We choose the Euler
 angles randomly so that the rotations are uniformily distributed
 on the Poincar'e sphere, as in the method 
 OptFiber::GetFiberSampleCompleteScatt().

 For propagation through each of the NumberScatterings sections we 
 set OptFiberLocalError::LengthFiber using either LengthFiberExceptLast
 or LengthFiberLast. We then call the methods 

    -# OptFiberLocalError::ApplyPMDCompleteScatteringFreqDomainOperator()
    -# OptFiberLocalError::PropagateWithoutPMD()

 For the two LOCAL_ERROR step size selection methods,
 for the first section of fiber we choose DeltaZInitial as in the 
 case of no PMD. For subsequent sections, we choose DeltaZInitial to be
 the size of the last step in the previous section. In this way, if there
 the PMD was negligible we would get the same distribution of step sizes
 as when we would run VECTOR_MANAKOV_NO_PMD.

 In addition to doing propagation the code solves the concatenation equations
 for the output polarization dispersion vector in the method
 OptFiberLocalError::ComputeEulerMatrixAndUpdateOutputPDVector()
 which is called from 
 OptFiberLocalError::ApplyPMDCompleteScatteringFreqDomainOperator().
 
 To compute the OutputPDVector for propagation through an entire system,
 not just through a single fiber, the OutputPDVector is really a variable
 in the OptSignal class, OptSignal::OutputPDVectorOFLE.

 The total DGD up to the present point in the system can be obtained
 by calling OptFiberLocalError::GetDGD(), which just computes the length
 of the OutputPDVector. When you regenerate the optical signal for a 
 new pass through the system in OptSignal::RegenerateSignalString(), the
 OutputPDVector and InputPDVectors are reset by calling 
 OptSignal::ResetPDVectorOFLE().

 \todo Compute InputPDVector, 2nd order PMD, Autocorrelation function
       and test. Add in ability to track oOptSignal->oReduOptSignal 
       for GaussianReceiver class.

*/

class OptFiberLocalError
{

 public:

 //! Constructor

  OptFiberLocalError(string InFileName, OptSignal *oOptSignal, RanNumGen *RNG2,
                    string Job2);

  void SetRamanAmpPointer(RamanAmp * oRaman2);

  //! Destructor

 ~OptFiberLocalError(void);

 //! Propagates OptSignal object through an optical fiber

 void Propagate(double * PropagatedLength,int * ZStepNum);

 //! Propagates OptSignal object through an optical fiber
 //! with Raman amplification

 void PropagateWithRaman(double * PropagatedLength,
			 int * ZStepNum,
			 double  RamanStartingLength2);
 
 string WriteTypeSolver(void);
 string WriteTypeStepSizes(void); 
 string WriteTypeFiberLossGain(void);
 string WriteTypeFiberAmplifierNoise(void);
 string WriteTypeFiberAmplifierGain(void);

 void OpenStepSizesFile(string StepSizesFileName,int Mode);
 void WriteStepSizes(double PropagatedLength);
 void WriteAveragePowerdBm(double PropagatedLength2);

 // Get methods

 double GetFirstOrderDispersion() {return  FstOrDispFiber;};
 double GetSecondOrderDispersion() {return  SndOrDispFiber;};
 double GetFiberLength() {return LengthFiber;};
 double GetNepperAttenuationFiber() {return NepperAttenuationFiber;};
 double GetNonLinIndexFiber() {return NonLinIndexFiber;};
 double GetGamma() {return gamma;};
 double GetDeltaZ() {return DeltaZ;}; 
 int GetNumStepsBetweenScatterings() {return NumStepsBetweenScatterings;};

 void SetNumStepsBetweenScatterings(int value) 
	{NumStepsBetweenScatterings = value;
	 SetFiberLengthsForPMD(); };

  //! Returns value of Dispersion in s/m^2 at Wavelength in m

  double GetDispersion(double Wavelength);

 // Set methods

 void SetTypeStepSizes(int TypeStepSizes2);
 void SetWalkOffParameter(double WalkOffParameter2); 
 void SetConstantDeltaZ(double DeltaZValue);

 // void SetNumStepsBetweenScatterings(int Value);

 void SetOutputStepSizesFlag(int OutputStepSizesFlag2);
 void SetRelativeErrorGoal(double RelativeErrorGoal2);


  /*! Sets First order dispersion, i.e., Beta'', in s^2/m.
      Note that X ps^2/km = X*1e-27 s^2/m
  */

 void SetFstOrDispFiber(double FstOrDispFiber2);

  /*! Sets Second order dispersion, i.e., Beta''', in s^3/m.
      Note that X ps^3/km = X*1e-39 s^2/m
  */

 void SetSndOrDispFiber(double SndOrDispFiber2);
 void RecoverSndOrDispFiber();

 void SetDispersion(double DispersionFiberRefWavelength,
		    double DispSlopeFiberRefWavelength); // s/m^2 and s/m^3

 /*! Sets nonlinear coefficient gamma in m^(-1)W^(-1). 
     X km^(-1)W^(-1) = X*1e-3 m^(-1)W^(-1)
 */

 void SetGammaFiber(double gamma2);
 void RecoverGammaFiber();

 //! Sets optical power attenuation parameter in dB/m. 
 //! \sa AttenuationFiber

 void SetAttenuationdB(double AttenuationFiber2);

 //! Sets linear optical power attenuation parameter in m^(-1).
 //! This is attenuation parameter in NLSE and HME (up to a sign) 
 //! \sa AttenuationFiber \sa NepperAttenuation
 
 void SetNepperAttenuation(double NepperAttenuation2);


 //! Enables user to set unsaturated Nepper gain 

 void SetFiberAmplifierUnsaturatedNepperGain(
           double FiberAmplifierUnsaturatedNepperGain2);

 //! The value of DeltaZmax is determined in this method from
 //! input parameter DeltaZMaxInfile and, if
 //! TypeFiberLossGain == FIBER_AMPLIFIER, from the gain and noise 
 //! parameters of the fiber amplifier. 
 //! \sa enum typeFiberAmplifierGain, enum typeFiberAmplifierNoise

 void ComputeDeltaZMax(void);

 void SetWriteAveragePowerdBmFlag(int Value);

  void SetWritePulseWidthFlag(int Value) {WritePulseWidthFlag = Value;};

  void SetFiberLength(double Value);
  void SetScatteringStepSize(double Value);

  // ####### Methods related to Manakov-PMD equation ######

  void GetNewFiberRealization(void);
  void SetFiberRealization(int SectionNumber, 
			   double Theta2, double Phi2, double Psi2);

  void ApplyPMDCompleteScatteringFreqDomainOperator(
                       int ScatteringSectionIndex);
  
  void ComputeEulerMatrixAndUpdateOutputPDVector(int ScatteringSectionIndex);

  void GetEulerMatrix(double theta,double phi,double psi);

  void SetFiberLengthsForPMD(void);

  void SetMeanDGDPerSqrtLength(double MeanDGDPerSqrtLength2);
  void SetPMD(double PMD2);
  double GetFreqDerivativeOfBirefringence(void);

  double GetDGD(void);  
 
  void PropagateOutputPDVector(double Freq,
                                double theta, double phi, double psi);

  double GetExpectedDGD() 
             {return sqrt(8./(3.*pi)*LengthFiber)*PMD; };
  double GetExpectedSndOrderDGD() 
             {return sq(GetExpectedDGD())/sqrt(3.); };
  double GetMeanDGD_SqrtLength(void) 
             {return sqrt(1e3)/1e-12*MeanDGDPerSqrtLength;};

  // JZ Commented out 
   void UpdateJonesMatrixAfterBiregringence(double StepSizeZ);

  // ########## PUBLIC DATA #####################
  

public:

  RamanAmp * oRaman;

  // ########## PRIVATE METHODS ####################

 private:

  // Private Methods

 void PropagateWithPMD(double * PropagatedLength,
                                           int * ZStepNum);

 void PropagateWithoutPMD(double * PropagatedLength,
                                           int * ZStepNum);

 void PropagateLocalError(double * PropagatedLength, int * ZStepNum);


 void PropagateConstStepSizeOrWalkOff(double * PropagatedLength,
                                        int * ZStepNum);


 void ResetVolatileVariables(void);

 void inline UpdateFreq(void);
 void inline UpdateTime(void);

 void ComputeWalkOffStepSize(void);

 void ComputeDispersionOperator(double StepSizeZ);
 void ApplyDispersionOperator(bool ApplyTwiceFlag);
 void ApplySaturableGainOperator(double StepSizeZ);
 void ApplyFreqDomainOperator(bool ApplyDispersionTwiceFlag,
                              double GainSatStepSizeZ);

 void ApplyRamanOperator(double CurrentZValueWithinFiber2,double ZStep2);
 void ApplyTimeDomainOperator(double StepSizeZ);

 void ApplyKerrNonlinearityOperator(double StepSizeZ);
 void ApplyFastSaturableAbsorptionOperator(double StepSizeZ);
 void ApplyFastSaturableAbsorptionQuadraticAnalytic(double StepSizeZ);
 void ApplyFastSaturableAbsorptionHeunsMethod(double StepSizeZ);
 cplx FastSaturableAbsorptionFunction(cplx u);

 double ComputeMaxPower(void);

 void CopyFftMatricesFreqData(fftMatrices * OrigData,fftMatrices * CopyOfData);

 double ComputeThirdOrderSolutionAndRelativeError(void);
 double ComputeThirdOrderSolutionAndRelativeErrorOLD(void);

  // AccumulatedFirstOrderDispersion is a data member in the 
  // OptSignal class which keeps track of the total dispersion
  // (beta'') in s^2 which the signal has experienced.
  // The UpdateAccumulatedFirstOrderDispersion is called each z-step.

 void UpdateAccumulatedFirstOrderDispersion(double StepSize);
 void UpdateAccumulatedSecondOrderDispersion(double StepSize);

 //! Computes the saturable gain per unit length at start of 
 //! step. Used by FiberAmplifierComputeNoiseAmplitudeFactor and
 //! ComputeFreqDomainOperator. 

 void FiberAmplifierComputeSaturableGain(void);

 //! Computes the saturable gain per unit length in first and second
 //! steps of Heun's method discretization implemented in
 //! ApplySaturableGainOperator. 

 double FiberAmplifierComputeSaturableGainForHeunsMethod(bool FirstStep);

 //! Computes energy in frequency domain, called by 
 //! FiberAmplifierComputeSaturableGainForHeunsMethod

 double FiberAmplifierGetEnergyForHeunsMethod(bool FirstStep);

 //! Computes the noise amplitude factor for SCALAR signals
 //! that are to be propagated through a z-step of size StepSize
 //! in a FIBER_AMPLIFIER. 
 //! Called by OptFiberLocalError::FiberAmplifierAddMonteCarloNoise()

 /*! When we have a saturable amplifier with parabolic gain shape
     we compute the noise amplitude factor at point z along the fiber
     using the gain g(z) at that point. Hence we ignore both the 
     parabolic gain shape (FIX THIS!) and any loss in this computation.

     For a flat gain amplifier we compute the noise amplitude factor
     from the flat gain parameter, which is specified by a negative
     fiber attenuation.

     \sa AttenuationFiber
  */ 

  void FiberAmplifierComputeNoiseAmplitudeFactor(double StepSize);

  //! Adds random noise realization at beginning of current split step
  //! in a FIBER_AMPLIFIER

  void FiberAmplifierAddMonteCarloNoise(double StepSize);

  void AllocateMemoryForEulerAngles(void);

  // Private Raman Methods
  
  void ApplySimpleRamanOperator(double CurrentZValueWithinFiber2,
				double ZStep2);

  void ApplyFullRamanOperator(double CurrentZValueWithinFiber2,double ZStep2);

  // ########## PRIVATE DATA ######################

  // Private Data

 ofstream StepSizesFile;

 int OutputStepSizesFlag;
 int StepSizesFileMode;
 string Job;

 typeSimulation TypeSimulation;

  //! Specifies precisely which PDE is to be solved.
  /* Options are 
     -# SCALAR_NLS = 1: The PDE is the scalar nonlinear Schroedinger equation
        as used in optical communications. This code can be run with or without
        the 8/9-ths factor in the Kerr nonlinearity term. This factor should
        be included when using a scalar PDE to
        model noise-free systems in the limit as the birefringence
        goes to zero, or more precsiely in the limit that the characteristic 
        length scale for PMD is longer than system length (See CRM's paper
        J Eng Math, Eqn 68).
     -# VECTOR_MANAKOV_PMD = 2: The PDE is the vector Manakov equation
        as in CRM's J Eng Math, Eqn (67), in the case that there is no linear
        PMD term and no nonlinear PMD. This equation is appropriate for
        simulations with noise when the characteristic 
        length scale for PMD is longer than system length.
     -# VECTOR_MANAKOV_NO_PMD = 3: The PDE is the vector Manakov equation
        as in CRM's J Eng Math, Eqn (67) with linear PMD (but not nonlinear
        PMD term). Includes the birefringence and random scattering on 
        the Poincare sphere.
  */

  typeSolver TypeSolver;

  typeStepSizes TypeStepSizes; // This variable used to be called
                               // TypeAlgorithm


  typeFiberLossGain TypeFiberLossGain;

  typeFiberAmplifierNoise TypeFiberAmplifierNoise;

  //! The spontaneous emission factor when TypeFiberLossGain = FIBER_AMPLIFIER
  //! and TypeFiberAmplifierNoise != NO_NOISE

  double FiberAmplifierSpontaneousEmissionFactor;

  //! The max step size for a fiber amplifier.
  /*! When TypeFiberLossGain = FIBER_AMPLIFIER
      and TypeFiberAmplifierNoise != FIBER_AMP_NOISE_OFF the max step
      size is given by
   
      DeltaZMax = 
       FiberAmplifierNoiseMaxStepSizeFactor/(abs(NepperAttenuationFiber))
 
      Default value is FiberAmplifierNoiseMaxStepSizeFactor = 0.1

      \sa enum typeFiberAmplifierNoise

      \sa double FiberAmplifierGainSaturationMaxGainPerStep


  */

  double FiberAmplifierNoiseMaxStepSizeFactor;

  //! Used to control maximum step size for saturable gain.
  /*

    Used when TypeFiberLossGain = FIBER_AMPLIFIER and 
      TypeFiberAmplifierGain == FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN

    The step size is chosen to gaurantee that

    Power(z+delta_z)/Power(z) < FiberAmplifierGainSaturationMaxGainPerStep
   
    Note that this gain is specified on a linear scale.

    Default value is 1.01

    This ensures that power is correctly computed in calculation of saturable
    gain.

    \sa enum typeFiberAmplifierGain.

    \sa FiberAmplifierNoiseMaxStepSizeFactor

  */ 

  double FiberAmplifierGainSaturationMaxGainPerStep;

  //! Analogous to OptAmplifier::NoiseAmplitudeFactor

  double FiberAmplifierNoiseAmplitudeFactor;

  //! See enum typeFiberAmplifierGain.

  typeFiberAmplifierGain TypeFiberAmplifierGain;

  //! SuperGaussFiltGainCoefficient for FiberAmplifier.
  //! Determines filter width. 
  //! See enum typeFiberAmplifierGain.

  double FiberAmplifierSuperGaussFiltGainCoefficient;

  //! SuperGaussFiltOrder for  FiberAmplifier. 
  //! Value of 1 gives standard Gaussian filtering.
  //! The higher the value, the more flat-topped the filter.

  int FiberAmplifierSuperGaussFiltOrder;

  //! Unsaturated linear power gain coefficient for saturable gain 
  //! FiberAmplifier \sa enum typeFiberAmplifierGain

  double FiberAmplifierUnsaturatedNepperGain;

  //! Saturation energy in Joules for saturable gain FiberAmplifier
  //! \sa enum typeFiberAmplifierGain

  double FiberAmplifierSaturationEnergy;

  //! Internal parameter for FiberAmplifier with saturable gain

  double FiberAmplifierSaturableGainPerLength;

  typeFastSaturableAbsorption TypeFastSaturableAbsorption;

  //! Coefficient for FastSaturableAbsorption

  /*! In case TypeFastSaturableAbsorption is QUADRATIC or QUADRATIC_ANALYTIC
      we add a term of the form 

      u_z = FastSaturableAbsorptionCoefficient*|u|^2*u

      to the standard NLS. FastSaturableAbsorptionCoefficient should be
      positive and is typically a small compared to the nonlinear Kerr 
      coefficient, gamma.

  */

  double FastSaturableAbsorptionCoefficient;

  //! Internal variable that is set to 1 whenever 
  //! FastSaturableAbsorptionCoefficient is nonzero

  bool DoFastSaturableAbsorptionFlag;

  //! Flag to do time operator exactly by multiplying by exp(arg)
  //! rather than by approximating exp(arg) by a 4th-order Taylor 
  //! polynomial.
  //! Input variable used in method ApplyTimeDomainOperator.

  bool DoTimeOperatorExactlyFlag;

  //! Flag that controls when to recompute DispersionOperator

  /*! We only compute the DispersionOperator at the start of each
       call to the PropagateLocalError() and PropagateWalkOffOrConst()
       methods and if DeltaZ changes. At the start of these methods and 
       whenever DeltaZ changes we set
      DispersionOperatorValid = false.
  */

  bool DispersionOperatorValid;

  /*! Set to 1 if want the scalar NLS solver to have 8/9 factor 
   with nonlinearity, Set to 0 if want factor of 1.
   The factor of 8/9 is appropriate when you are modeling 
   fiber with rapidly and randomly varying birefringence
   but in which the the characteristic length over which 
   PMD becomes important is larger than the system length
   so that one can use a scalar model. The equation with the 8/9
   factor in it is a scalar reduction of the vector Manakov-PMD
   equation. See Menyuk, J. Eng. Math. Vol 36, pp 113-136, 1999, and 
   in particular p 132 of that paper. Also see Wai Menyuk and Chen,
   Optics Letters Vol 16 No 16, 1991 p 1231-1233; Wai and Menyuk, 
   JLT Vol 14 No 2, 1996 pp 148-157, especially pages 154-155; and
   Marcuse Menyuk and Wai, JLT Vol 15 No 9 1997, pp 1735-1746.
   The factor 1 is appropriate when are modeling fiber with
   zero birefringence. 
   */

 int ScalarSolverEightNinthsFlag; 

 OptSignal * oOptSignal;
 RanNumGen * RNG;
 cfftw *fft;
 fftMatrices sfftM;  

 cplx * FreqDomainOperatorX;
 cplx * FreqDomainOperatorY;

 cplx * SqFreqDomainOperatorX;
 cplx * SqFreqDomainOperatorY;

  
 fftMatrices sfftM_Start;
 fftMatrices sfftM_Coarse;
 cplx * LaserWorkX;
 cplx * LaserWorkY;


 double LengthFiber;
 
 double RamanStartingLength;
 double PropagatedLengthStartFiber;
 int NumStepsBetweenScatterings; // For const step size second order algorithm

  double ReferenceFreq;
  double ReferenceWavelength;


  double DispersionFiberCenterWavelength; // s/(m^2)
  double DispSlopeFiberCenterWavelength; // s/(m^3)

 //! If true read in FstOrDispFiber and SndOrDispFiber from infile
 //! rather than reading in Dispersion and DispersionSlope at Ref Freq.

 bool SetFstSndOrDispDirectlyFlag; 

 double FstOrDispFiber; // s^2/m
 double SndOrDispFiber;
 double SndOrDispFiber_Save;
 double NonLinIndexFiber;

 //! Flag that takes value 1 if at least one of
 //! NonLinIndex or FastSaturableAbsorptionCoefficient is nonzero

 int    WithNonLin;      	      

 double EffectAreaFiber;  

 //! The fiber optical power attenuation (or gain) parameter in dB/m

 /*! This parameter plays several differetn rols depending on the context.

	- If TypeFiberLossGain == LOSSY_FIBER 
              this is the fiber attenuation 
	- If TypeFiberLossGain == FIBER_AMPLIFIER and 
             TypeFiberAmplifierGain == FIBER_AMP_UNSATURATED_FLAT_GAIN 
              this is the negative of the fiber gain.
	- If TypeFiberLossGain == FIBER_AMPLIFIER and 
             TypeFiberAmplifierGain == FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN
              this is the attenuation, which we include to implement 
              Haus's Master Equation with distributed gain and loss. 
 */

 double AttenuationFiber;      

 //! The fiber optical power attenuation (or gain) parameter in m^(-1)
 //! The terms of the NLS or Master equation that involve this parameter
 //! are of the form u_z = -0.5*NepperAttenuationFiber*u 

 double NepperAttenuationFiber;

  typeRamanAmplification TypeRamanAmplification;

// Variables for simple raman model

  double RamanOnOffGaindB;
  double NepperRamanGain;
  double NepperRamanPumpLoss;

 double CenterFreq;      
 int    qtPoints;
 double DeltaFreq;     
 double gamma;
 double gamma_Save;

 double CurrentZValueWithinFiber; 
   // Ranges from 0 to LengthFiber during
  // course of propagation within the current fiber

  //! The step size taken in each step
  /*! For the LOCAL_ERROR methods DeltaZ is the size of a coarse step
      The solution is actually the one taken using the fine steps
      of size DeltaZ/2.
  */

 double DeltaZ;
 double DeltaZMaxInfile, DeltaZMax;

  //! Used to set initial step size for the fiber when using Local Error method

 double DeltaZInitial;

 bool InitializedDeltaZInitialFlag;
 
 double WalkOffStepSize;
 double WalkOffLastStepSize;
 int WalkOffNumSteps;
 double WalkOffParameter;
 
 double MaxPower;

 double FWHM_Pulse;

 bool TimeDataValid;
 bool FreqDataValid;

 double RelativeError;
 double RelativeErrorGoal;

 int LocalErrorFreqLimit;

 int WriteAveragePowerdBmFlag;
 int WritePulseWidthFlag;
 
   // ####### Variables related to Manakov-PMD equation ######

  double *OutputPDVector;
  double *InputPDVector;
  
  // Variables related to birefringence strength

 double PMD;
 double MeanDGDPerSqrtLength;

 // Variables related to random rotation of axes of birefringence

 double ScatteringStepSize;
 cplx EulerMatrix[2][2];

  int NumberScatterings;
  double TotalLengthFiber;
  double LengthFiberExceptLast, LengthFiberLast;

  // Arrays for length NumberScatterings for the Euler angles

  double * Theta;
  double * Phi;
  double * Psi;

 double *PDVectorHai;             //added by Hai 8/14/03
 
 
 cplx *JonesMatrix;   //added by Hai 7-1-03
 int NumberOfChannels;  //added by Hai 7-1-03
 double *FreqChannels;   //added by Hai 7-1-03




};

#endif /* _OCS_OPT_FIBER_LOCAL_ERROR_HH_ */





