

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_OPT_SIGNAL_HH_
#define _OCS_OPT_SIGNAL_HH_

#ifndef _OCS_SYSTEM_HDRS_
#define _OCS_SYSTEM_HDRS_

#include <iostream.h>
#include <string.h>
#include <math.h> 
#include <fstream.h>
#include <complex.h>
#include <stdlib.h>
#include <stdio.h>
#include <strstream>

#include <string>
  using namespace std;


#endif /* _OCS_SYSTEM_HDRS_ */

#include "ocsConst.hh"
#include "ocsTools.hh"
#include "ocsReadWriteTools.hh"
#include "ocsRanNumGen.hh"

#include "ModulationFormats.hh"
#include "ocsReduOptSignal.hh"


#ifndef _OCS_TYPE_SOLVER_
#define _OCS_TYPE_SOLVER_

/*!

\brief Specifies the type of signal to be  simulated 

1 = SCALAR_NLS: a scalar optical signal (ie, in a single polarization state)
with fiber propagation done using the scalar nonlinear Schroedinger equation \n
2 = VECTOR_MANAKOV_PMD: a vector signal (ie, with two polarization states)
with fiber propagation done using the Manakov-PMD equation \n
3 = VECTOR_MANAKOV_NO_PMD: a vector signal with no PMD in the fiber 
propagation \n

These propagation options are used in the class OptFiberLocalError.

*/

enum typeSolver { SCALAR_NLS = 1, VECTOR_MANAKOV_PMD = 2, 
                  VECTOR_MANAKOV_NO_PMD = 3 };

#endif /* _OCS_TYPE_SOLVER_ */

#ifndef _OCS_FFT_MATRICES_
#define _OCS_FFT_MATRICES_

//! This public structure consists of 4 cplx arrays which store the
//! time and frequency data in the X and Y polarizations in Jones space

struct fftMatrices 
{
      cplx* tOptSignalX;    
      cplx* fOptSignalX;
      cplx* tOptSignalY;    
      cplx* fOptSignalY;      
};

//! This structure simplifies access to the work space used
//! by the OptFiberLocalError class

//! \sa  OptSignal::UsingOptFiberLocalError 

struct OptFiberLocalErrorWorkSpace
{
 fftMatrices sfftM_Start;
 fftMatrices sfftM_Coarse;
 cplx * LaserWorkX;
 cplx * LaserWorkY;
 cplx * FreqDomainOperatorX;
 cplx * FreqDomainOperatorY;
 cplx * SqFreqDomainOperatorX;
 cplx * SqFreqDomainOperatorY;
 
};


#endif /* _OCS_FFT_MATRICES_ */

#ifndef _PHOSSIL_FFT_POWER_MATRICES_
#define _PHOSSIL_FFT_POWER_MATRICES_

struct fftPowerMatrices
{
      cplx* tPowerSignal;
      cplx* fPowerSignal;
};

#endif /*  _PHOSSIL_FFT_POWER_MATRICES_ */


enum typeEOMElecData {EOM_BINARY_DATA_PERFECT_RECT = 1,
                      EOM_BINARY_DATA_ARCTANH_SMOOTHING = 2,
                      EOM_BINARY_DATA_SUPERGAUSS_SMOOTHING =3,
                      EOM_BINARY_DATA_RCOS=4};

enum typeEOMElecFilter {EOM_NO_ELEC_FILTER = 0,EOM_GAUSSIAN = 1,EOM_BESSEL = 2};


#include "ocsBitString.hh"

// ##### Class OptSignal ############

//! The OptSignal class contains data and methods for the optical signal

/*!

\section OVER Overview

The classes in the library either operate on optical data in the class 
OptSignal or on electrical data in the class Photodetector. As such 
the OptSignal class is the most important in the library.  The class
enables several different modulation formats 
most of which are coded for multiple frequency channels. 
\sa OptSignal::TypePulse

Both SCALAR and VECTOR simulations can performed with the library.
The enumeration type  enum typeSimulation defined in ocsReadWriteTools.hh
specifies the type of simulation. (OptSignal::TypeSimulation).
Half the amount of data is allocated for SCALAR simulations as for VECTOR
ones. Therefore the code runs about twice as fast for SCALAR simulations.  

The main data object in the class is the structure OptSignal::sfftM 
of type fftMatrices which stores the full time and frequency domain
signals. The class contains a pointer to a cfftw object OptSignal::fft.
The cfftw class is an interface to the Fastest Fourier Transform in the
West softpackage fftw which manages and performs 
Fourier transforms in the code. See http:www.fftw.org
 
Data can either be SCALAR or VECTOR.
For VECTOR simulations the initial Stokes parameters can be specified
for each channel. \sa  OptSignal::InitializeInStokes(), 
\sa OptSignal::GetInStokes(), \sa OptSignal::SetInStokes()

\section NOISE Data for Noise Modeling

The OptSignal class also includes data to model additive white Gaussian noise.
The full signal data and the additive white Gaussian noise
can be used by the GaussianReceiver class to compute the Gaussian apprximations
to the electrical pdfs in each 
bit. Accurate electrical pdfs can be computed 
using the ChiSquareReceiver class. 
The ChiSquareReceiver class can be used in conjunction with the NoiseCovariance
class to compute electrical pdfs under the assumption of no noise-noise
interactions during fiber transmission. (Signal-noise interactions in the
fiber are modeled by the NoiseCovariance class.)
The electrical pdfs can also be approximated using basic or advanced
Monte carlo simulations, for which the noise does not need to be 
AWGN.
 
To model unpolarized additive white Gaussian noise 
the noise is accumulated in the double 
OptSignal::AccumulatedTotalNoiseSpectralDensityGaussianStatistics.
With this model the noise is assumed to be identical in each of the
channels. To model partially polarized noise and/or white Gaussian
noise that varies from channel to channel we use a ReduOptSignal object.
The data member OptSignal::oReduOptSignal is a pointer
to a ReduOptSignal object that stores the power and polarization states
of the signal and noise in each channel. The ReduOptSignal is incorporated
into the OptSignal to allow the GaussianReceiver class to receive the
data it needs to model partially polarized additive white Gaussian noise.

\warning Although the ReduOptSignal class was written and tested for
wavelength-division multiplexed (WDM) signals the GaussianReceiver class
has only been carefully tested for single channel signals. 

The OptSignal class also contains data members 
OptSignal::AccumulatedFirstOrderDispersion and
OptSignal::AccumulatedSecondOrderDispersion
that store the first and second order dispersion
accumulated when the signal is propagated through
(possibly several) different OptFiberLocalError 
objects. One application of these methods is to perform
artificial dispersion compensation in the NoiseCovariance class.

\section CLASSES Associated Classes

The OptSignal class includes pointers to a BitString object
OptSignal::oBitString and a RanNumGen object OptSignal::RNG.
The BitString object contains and initializes the data sequence  
in each channel. The RanNumGen object is included to allow
random bit strings and random polarization states in each channel.

\section METHODS Main Methods

The main methods are the constructors, assignment operator,  
and the method OptSignal::RegenerateSignalString() which is used
to regenerate the signal in Monte Carlo simulations. (This method
is also called by the constructors.) 

In addition there are methods to get and set class parameters, 
process accumulated dispersion and noise spectral density,
write various forms of the optical data to output files,
and compute and write to files derived data such as the energy
within a time window and the central time of a pulse specified
within a time window. 

\section INFILE A Sample Input Parameter File

<a href = "http://www.umbc.edu/photonics/home/members/jzweck/PhoSSiL/infiles/Signal.in">Signal.in</a>

\section EXAPP Example Application

\sa OptFiberLocalError (See example application there)

*/

class OptSignal
{

  // ####### Public Methods ########

   public:

      // #### Constructors ####

      // OptSignal();

      // Use these constructors:

      //! The main constructor for the class
      /*! Reads in parameters from the file named InFileName.
          The RanNumGen is used for randomly generated bit sequences
          and for randomly chosen initial polarization states in each channel.
          Initializes all data in the class. 

          The constructor calls OptSignal::ConstructOptSignal() 
          which reads in parameters and does preprocessing. Then it calls
          OptSignal::RegenerateSignalString() which initializes the
          data  OptSignal::fftMatrices used to store the signal in 
          time and frequency.

          If you add another variable/parameter to the class you must
          initialize it in OptSignal::ConstructOptSignal() 
          and also make the appropriate
          addition to the following methods:

          -# OptSignal::OptSignal(OptSignal & OrigOptSignal)
          -# OptSignal::OptSignal & operator=(OptSignal & RHS_Signal)
          -# OptSignal::OptSignal & operator-(OptSignal & Signal2)
          -# OptSignal::OptSignal(OptSignal & Signal1,OptSignal & Signal2,
                                  RanNumGen *RNG3)
          -# friend void AddSignals(OptSignal * Signal1,OptSignal * Signal2,
                            OptSignal * Signal3)
       */ 
 
      OptSignal(string InFileName, RanNumGen *RNG2);
 

      //! A second constructor for the class
 
      /*! Use this constructor if you want to quickly 
          initialize signal in time and
          frequency using data in file named DataFileName
      
          Reads in parameters from the file named InFileName.
          The RanNumGen is used for randomly generated bit sequences
          and for randomly chosen initial polarization states in each channel.
          Initializes all data in the class. 

          The method reads 
          in the signal data arrays from the file with name DataFileName. 

          This constructor calls OptSignal::ConstructOptSignal() 
          which reads in parameters and does preprocessing. Then it calls
          OptSignal::ReadFileComplexFieldsTimeAndFreq() which initializes the
          data  OptSignal::fftMatrices used to store the signal in 
          time and frequency.

          If you add another variable/parameter to the class you must
          initialize it in OptSignal::ConstructOptSignal()
          and also make the appropriate
          addition to the following methods:

          -# OptSignal::OptSignal(OptSignal & OrigOptSignal)
          -# OptSignal::OptSignal & operator=(OptSignal & RHS_Signal)
          -# OptSignal::OptSignal & operator-(OptSignal & Signal2)
          -# OptSignal::OptSignal(OptSignal & Signal1,OptSignal & Signal2,
                                  RanNumGen *RNG3)
          -# friend void AddSignals(OptSignal * Signal1,OptSignal * Signal2,
			    OptSignal * Signal3)

      */

      OptSignal(string InFileName, RanNumGen *RNG2, string DataFileName);


      //! Called by the constructors 
      //! -# OptSignal(string InFileName, RanNumGen *RNG2, string DataFileName)
      //! -# OptSignal(string InFileName, RanNumGen *RNG2)
      
      /*!

          Reads in parameters from the file named InFileName.
          The RanNumGen is used for randomly generated bit sequences
          and for randomly chosen initial polarization states in each channel.
          Initializes all data in the class. 

          If you add another variable/parameter to the class you must
          initialize it in this method 
          and also make the appropriate
          addition to the following methods:

          -# OptSignal::OptSignal(OptSignal & OrigOptSignal)
          -# OptSignal::OptSignal & operator=(OptSignal & RHS_Signal)
          -# OptSignal::OptSignal & operator-(OptSignal & Signal2)
          -# OptSignal::OptSignal(OptSignal & Signal1,OptSignal & Signal2,
                                  RanNumGen *RNG3)
          -# friend void AddSignals(OptSignal * Signal1,OptSignal * Signal2,
                            OptSignal * Signal3)

          \warning To ensure the signal is discretized well the 
          central frequency is each channel is adjusted slightly
          so that it falls on the frequency grid. Actual frequencies used
          are written to the LogFile.

       */

      void ConstructOptSignal(string InFileName, RanNumGen *RNG2);


      //! Copy constructor
 
      OptSignal(OptSignal & OrigOptSignal); 

      //! Assignment operator

      /*! This assignment operator assumes that OptSignal::NumChannels, 
          OptSignal::qtPoints, OptSignal::TimeWindow, OptSignal::StringLength 
          are unchanged so that none
          of the dynamically allocated data arrays/classes changes size.
       */

      OptSignal & operator=(OptSignal & RHS_Signal);
                  

     //! Takes difference of two signals
     /*! First calls copy constructor to initialize parameters for
         returned signal, then initializes data of returned signal
         to be difference of data in this signal and Signal2
     */

     OptSignal & operator-(OptSignal & Signal2); 

     //! Subtracts frequency domain data in oOptSignal2 from this OptSignal
     //! and then calls OptSignal:Freq2Time() 

     void SubtractFreqDomain(OptSignal *oOptSignal2);

     //! Adds frequency domain data in oOptSignal2 to this OptSignal
     //! and then calls OptSignal:Freq2Time() 

     void AddFreqDomain(OptSignal *oOptSignal2);
  
     //! Constructor for sum of signals
     /*! This signal is initialized with parameters of Signal1 and
         data given by sum of data of Signal1 and Signal2.

         \warning Does not work when oReducedOptSignal is important
      */

     OptSignal(OptSignal & Signal1,OptSignal & Signal2,RanNumGen *RNG3);
        

     //! Addition operator: Signal3 = Signal1 + Signal2
     /*! 
         Adds data as well as adding BitStrings using method
         OptSignal::AddBitStrings()

         \warning This method assumes that 
         OptSignal::OptSignal(OptSignal & Signal1,OptSignal & Signal2,
                    RanNumGen *RNG3)  has been previously called

      */

     friend void AddSignals(OptSignal * Signal1,OptSignal * Signal2,
			    OptSignal * Signal3);


  //! Adds a ReduOptSignal containing noise and no signal
  //! to the current OptSignal which is assumed to be noise-free.
  //! By John Zweck, Jan 2006

  void AddReducedNoiseToNoiseFreeSignal(ReduOptSignal *ReduOptSignal1);

      //! Class destructor
      /*! Frees memory allocated in constructor and 
          calls oBitString->~BitString(). 
      */

      ~OptSignal();

      //! Basic transmitter model. Regenerates the signal data.
 
      void RegenerateSignalString(void); 
 
     //! Regenerates the signal using a new randomly chosen bit sequence

      void GetNewRandomSignalString(void);  

      //! Shifts the phase of the signal by the constant amount 
      //! PhaseShiftAmount in degrees

     void ConstantPhaseShift(double PhaseShiftAmount);
   
  //! Used in GaussianReceiver for computing the Q-factor 
  //! Generates an impulse in both polarization states

  void GenerateImpulseSignal(void);  

  //! Used in GaussianReceiver for computing the Q-factor 

  void ChangeAveragePowerChannCurrentSignal(double AveragePowerChann2);

  //! Used in GaussianReceiver for computing the Q-factor 

  void RegenerateSignalWithGivenAveragePowerChann(double AveragePowerChann2);

  //! Written for a particular application. \warning Do not use!
  /*! PhaseShiftAmount is in degrees */

   void GenerateSignalStringGaussianOrSechWithPhaseShift(
                                 double PhaseShiftAmount);

  //! Written for a particular application. \warning Do not use!
  /*! PhaseShiftAmount is in degrees */
 
  void GenerateSignalStringRaisedCosinesWithPhaseShift(
		                 double PhaseShiftAmount);


  //! Reads in time data from file when TypePulse == INPUT_SIGNAL_FROM_FILE

  void InputTimeDataFromFile(void);

  //! Generates a signal using the first order PMD model
  //! for the   Gaussian, Sech and NRZ formats

  /*!

  CAUTION: For NRZ this only works for a single channel signal

  \param DGD: The differential group delay in seconds
  \param PowerSplit: The power splitting parameter in range [0,1].

      The X-polarization component of the signal is shifted right by
      DGD/2 and the Y-polarization component is shifted left by
      DGD/2. The two polarization components are combined so that
      a fraction PowerSplit of the power is in the X-polarization 
      and a fraction 1-PowerSplit of the power is in the Y-polarization.


  */

   void InitializeFirstOrderPMDSignal(double PowerSplit, double DGD);

   void ChangeAbsoluteCenterFreq(double CenterFreq2);

   //! Returns Freqeuncy of chanel with index ChannelArrayIndex
   //! relative to OptSignal::CenterFreq. 

   /*! ChannelArrayIndex is asumed to be in range 0 to OptSignal::NumChanns-1. 
    */

   double GetRelativeCenterFrequency(int ChannelArrayIndex);

  double GetAccumulatedFirstOrderDispersion(void) 
               { return AccumulatedFirstOrderDispersion;};

  double GetAccumulatedSecondOrderDispersion(void) 
               { return AccumulatedSecondOrderDispersion;};

  void ResetAccumulatedFirstOrderDispersionToZero(void) 
               {  AccumulatedFirstOrderDispersion = 0.0;};

  void ResetAccumulatedSecondOrderDispersionToZero(void) 
               {  AccumulatedSecondOrderDispersion = 0.0;};
  
  void SetAccumulatedFirstOrderDispersion(double Value) 
               {  AccumulatedFirstOrderDispersion = Value;};

  void SetAccumulatedSecondOrderDispersion(double Value) 
               {  AccumulatedSecondOrderDispersion = Value;};

  void SetWriteAccumulatedFirstOrderDispersionFlag(int Value)
          { WriteAccumulatedFirstOrderDispersionFlag = Value; }

  //! \sa OptSignal::WriteAccumulatedFirstOrderDispersion()

  void OpenAccumulatedFirstOrderDispersionFile(string FileName);

  //! \sa OptSignal::WriteAccumulatedFirstOrderDispersion()

  void CloseAccumulatedFirstOrderDispersionFile(void);

  //! Writes out accumulated first order dispersion 
  
  /*! Writes out accumulated first order dispersion at distance
      given by PropagatedLength in both ps^2/km
      and ps/(nm-km) to the file opened by 
      OptSignal::OpenAccumulatedFirstOrderDispersionFile()
      \sa OptSignal::OpenAccumulatedFirstOrderDispersionFile()
      \sa OptSignal::CloseAccumulatedFirstOrderDispersionFile()
   */

  void WriteAccumulatedFirstOrderDispersion(double PropagatedLength);


  //! This and related methods are called automatically in the library
  //! and are used for the GaussianReceiver class with unpolarized noise

  double GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics(void)
    {return AccumulatedTotalNoiseSpectralDensityGaussianStatistics;};

  void SetAccumulatedTotalNoiseSpectralDensityGaussianStatistics(double Value)
         {AccumulatedTotalNoiseSpectralDensityGaussianStatistics=Value;};

  void AddToAccumulatedTotalNoiseSpectralDensityGaussianStatistics
         (double Increment)
       {AccumulatedTotalNoiseSpectralDensityGaussianStatistics += Increment;};

  void GainOfAccumulatedTotalNoiseSpectralDensityGaussianStatistics
       (double LinearPowerGainFactor)
            {AccumulatedTotalNoiseSpectralDensityGaussianStatistics 
               *= LinearPowerGainFactor;};

  double GetNoiseBandwidthGaussianStatistics(void)
    {return NoiseBandwidthGaussianStatistics;};

  void SetNoiseBandwidthGaussianStatistics(double Value)
         {NoiseBandwidthGaussianStatistics=Value;};

  // GET METHODS


      typeSimulation GetTypeSimulation(void) { return TypeSimulation;};
      typePulse GetTypePulse(void) {return TypePulse;};

      int GetChannelArrayIndex(int ChannelPhysicalIndex);
      int GetChannelPhysicalIndex(int ChannelArrayIndex);
      double GetFrequency(int ChannelArrayIndex);
      double GetWavelength(int ChannelArrayIndex);

      double GetCenterFreq(void) {return CenterFreq;}; 
      double GetCenterWavelength(void) {return CenterWavelength;};  

      int GetNumChannels(void) {return NumChannels;};
      double GetBitLength(void) {return BitLength;};

      int GetQtPoints(void) {return qtPoints;};
      double GetDeltaTime(void) {return DeltaTime;};
      double GetFWHM_Pulse(void) {return FWHM_Pulse;};
      double GetBitRateChann(void)  {return BitRateChann;};
      int GetStringLength(void) {return oBitString->GetStringLength();}
      double GetChannSpacing(void) {return ChannSpacing;}; 
      double GetOptPowerChann(void) {return OptPowerChann;};
      double GetDeltaFreq(void) {return DeltaFreq;};
      double GetTimeWindow(void) {return TimeWindow;};
      double GetFreqWindow(void) {return FreqWindow;};
      int GetDebugLevel(void) {return DebugLevel;};
      bool GetRandomPolarizationsFlag() { return RandomPolarizationsFlag;};

      bool GetUsingOptFiberLocalError(void) {return UsingOptFiberLocalError;};

      double GetChirpParameter(void) {return ChirpParameter;};
      double GetInStokes(const int ChannelArrayIndex, 
                               const int StokesIndex);
      cplx GetInJones(const int ChannelArrayIndex, int JonesIndex);

      double * GetInStokes(const int ChannelArrayIndex);
      cplx * GetInJones(const int ChannelArrayIndex);

      double GetFreqForPDVectorOFLE(void) {return FreqForPDVectorOFLE;};

  //! Generates a random Jones vector so that the corresponding
  //! Stokes vector is selected from a  uniform distribution
  //!  on the Poincare sphere

   void GetRandomJones(cplx * Jones);
   void TestGetRandomJones(string OutFileName, int NumberPoints);

  //! Used by OptFiberLocalError class 
  //! \warning PMD in OptFiberLocalError class may not be debugged

   void ResetPDVectorOFLE(void);

  //! Used by OptDemuxer class

   void SetNumberChannelsForDemuxer(int NumChannels2); // ## JZ

  //! Used by OptDemuxer class

   void SetNumChannels(const int NumChannels2); //## (IL)

  //! Used by OptDemuxer class

   void ReduceToSingleChannel(int ChannelArrayIndex); //## (IL)

   // SET METHODS

      void SetFWHM_Pulse(double Value)
           { FWHM_Pulse =  Value;};

      void SetRiseTime(double Value)
            { RiseTime = Value;};

      void SetExtinctionRatio(double Value)
            { LinearExtinctionRatio = dB2Linear(Value);};


      void SetOptPowerChannel(const double Power2)
                        { OptPowerChann = Power2; } ; 

      void SetCenterFreq(const double CenterFreq2)
                        { CenterFreq = CenterFreq2; };

      void SetChannSpacing(const double ChannSpacing2)
                        { ChannSpacing = ChannSpacing2; };

      void SetFrequency(int ChannelArrayIndex, double Freq);

      void SetInStokes(const int ChannelArrayIndex, const int StokesIndex,
                       double StokesValue);
      void SetInJones(const int ChannelArrayIndex, const int JonesIndex,
                       cplx JonesValue);

      void SetInStokes(const int ChannelArrayIndex,double * StokesVector);
      void SetInJones(const int ChannelArrayIndex, cplx * JonesVector);

      void SetRandomJonesVectors(void);

      void SetEOM_RiseTime(double Value)
           {EOM_RiseTime = Value;
            EOM_SlopeFactor = (2.0/EOM_RiseTime)*atanh(0.8);};

      void SetEOM_ElecFilter_FreqFWHM(double Value)
           {EOM_ElecFilter_FreqFWHM = Value;};

      //######## Methods to read in InStokes ########
      // Written by John Zweck

      void InitializeInStokes(ifstream * InFile);
      int ReadInStokes(string InStokesName, int ch, ifstream * InFile);

      // #### Methods written by Ivan Lima ####

      //! Sets signal data to zero

      void EliminateSignal(void) {
              EliminateX_Polarization();
	      EliminateY_Polarization();
      };       

      //! Sets signal data to zero

      void ClearSignal(void) {EliminateSignal();};      

      //! Sets signal data in X-polarization to zero

      void EliminateX_Polarization(void);

      //! Sets signal data in Y-polarization to zero

      void EliminateY_Polarization(void);      

      void ScaleSignal(double ScaleFactor);

      void SwitchXandYPolarizations(void);

      //! Returns average power of optical signal computed in time domain
      /*! See separate document on Conventions for the Fourier Transform
          and Nonlinear Schroedinger Equation */

      double GetAveragePower(void);

      //! \sa OptSignal::GetAveragePower() 

      double GetAveragePowerFreqDomain(void);

      //! Returns the total power in frequency interval with center
      //! OptSignal::CenterFreq and width  
      //! OptSignal::NumChannels * OptSignal::ChannelSpacing


      double GetAveragePowerFreqDomainInsideChannels(void);

      //! Returns the total power in frequency interval centered on
      //! channel with index ChannelArrayIndex and full-width   
      //! 2*OptSignal::ChannelSpacing. 

      /*! ChannelArrayIndex ranges from 0 to OptSignalNumChanns - 1 */

      double GetAveragePowerFreqDomainInsideChannel(int ChannelArrayIndex); 

      //! Returns integral of conj(tOptSignalX(t))*tOptSignalX(t+TimeOffset)dt

      cplx GetAutocorrelationFunctionOptSignalX(double TimeOffset);

      double GetDegreeOfPolarization(void);

      //! \note Bandwidth is centered about OptSignal::CenterFreq

      double GetDegreeOfPolarization(double BandWidth);

      double GetDegreeOfPolarization(double RelMinFreq, double RelMaxFreq);

      const char * WriteTypePulse(void);

      //! \sa OptSignal::OpenDegreeOfPolarizationFile()
      //! \sa CloseDegreeOfPolarizationFile()

      void WriteDegreeOfPolarization(double PropLength,
                                     double RelMinFreq, double RelMaxFreq);
 
      void OpenDegreeOfPolarizationFile(string InFileName,bool WriteDOPFlag);
      void CloseDegreeOfPolarizationFile(bool WriteDOPFlag);

      //! Computes the normalized Stokes parameters (4-vector)

      void GetTotalStokesParameters(double *Stokes);

      //! \sa OptSignal::GetTotalStokesParameters

      void GetTotalStokesParametersInBandwidth(double *Stokes,
			    double RelMinFreq, double RelMaxFreq);


      //! \sa OptSignal::GetDegreeOfPolarizationAndTotalStokesFreqDomain()

       void GetTotalStokesParametersInTime(double *Stokes,
					   int RelMinTimeIndex,
					   int RelMaxTimeIndex);

     //! Computes the degree of polarization and total Stokes parameters using 
     //! the frequency domain data only. JZ Sep 7th 2002.

    /*! You can either caluclate the degree of polarization and the
        total Stokes parameters in
        the frequency domain or the time domain. The answers
        are the same. 
     */

      double GetDegreeOfPolarizationAndTotalStokesFreqDomain(double * Stokes);

      //! Computes square of pulse width, using same formula used to
      //! compute variance of a pdf. Assumes signal only has one pulse
      //! \sa OptSignal::ComputeRMSPulseWidth for Simpson's rule version

      double GetSigmaSqOptSignal(void);  

      //! Computes FWHM of a pulse power assuming signal has a single pulse
      //! which is centered roughly in middle of time window. 
      //! Algorithm is very crude.

      double ComputeFWHMPulsePower(void);

      //! Time and frequency data in this signal is obtained 
      //! from oOptSignalOrig

      void ImportOptSignal(OptSignal *oOptSignalOrig);

      //! \warning Do Not Use!

      void DampBounds(double DampTime); 

      //! Does Fourier transform of OptSignal data from Time to Frequency
      void Time2Freq(void);
  
      //! Does Fourier transform of OptSignal data from Time to Frequency
      //! for SCALAR signal

      void Time2FreqScalar(void);

      //! Does Fourier transform of OptSignal data from Frequency to Time 
 
      void Freq2Time(void);  

      //! Does Fourier transform of OptSignal data from Frequency to Time 
      //! for SCALAR signal

      void Freq2TimeScalar(void);
  
      //! Has same effect as function of same name in Matlab 

      void fftshift(void);

      //! Has same effect as function of same name in Matlab 

      void ifftshift(void);
  
      //!   This  method shifts the signal to the right 
      //!   by the amount dTime using the fft

      void TimeShiftSignal(double dTime);

      void FreqShiftSignal(double dFreq);

      //! Writes to a file: Time in column 1 and Optical Power in column 2
      
      void WriteFileTimePower(string outFile);

      //! Writes to a file: Time in column 1 and Power Spectral Density
      //! in column 2

      void WriteFileFreqPower(string outFile);    

      //! Writes frequency data to a file according to the format specified by
      //! TypeOutputFlag. JZAug02

      void WriteFileWavelengthPower(string outFile);    

      //! Writes wavelength data to a file according to the format specified by
      //! TypeOutputFlag. IL_Jun03
  
      void WriteFileFreqPower(string outFile,int TypeOutputFlag);
        
      // ### The following public methods were added by John Zweck #####

     void WriteFileFreqPowerdBm(string outFile);
     
    //! Writes all complex fields to a file for quick restart

    /*! Writes 8 columns of data to file with name OutFileName
        Real and Imaginery parts of X and Y polarizations for
        time and frequency data.
        Use with OptSignal::ReadFileComplexFieldsTimeAndFreq(string InFileName)
     */

     void WriteFileComplexFieldsTimeAndFreq(string OutFileName);

    //! Reads all complex fields to a file for quick restart

   /*! Reads 8 columns of data from file with name InFileName:
       Real and Imaginery parts of X and Y polarizations for
       time and frequency data.
       Use with OptSignal::WriteFileComplexFieldsTimeAndFreq(string InFileName)
   */

     void ReadFileComplexFieldsTimeAndFreq(string InFileName);

     //! Writes to a file: Complex fields real(tX), imag(tX), 
     //! real(tY), imag(tY)

     void WriteComplexFields(string OutFileName, int ForRestartFlag);


 
    //! Needs more debugging!
    void ReadComplexFields(string InFileName);
 
    void ReadComplexFieldsX(string InFileName);

     //! Writes to a file: Complex fields real(tX), imag(tX)

     void WriteComplexFieldsX(string OutFileName, int ForRestartFlag);


     //! Writes to a file: Complex fields real(fX), imag(fX), 
     //! real(fY), imag(fY). Uses OptSignal::fftshift to put DC term in center

     void WriteComplexFieldsFreq(string OutFileName);

     //! Write Phase of X polarization versus Frequency to a file. 
     //! Uses OptSignal::fftshift to put DC term in center

     void WriteFreqPhaseX(string OutFileName);


     //! Writes to file with name outFile the power and phase of the 
     //! optical signal as a function of propagation distance and time
     /*! For VECTOR signals the phase of both X and Y polarizations
         is output. 
 
         File format (column-wise): 
           Distance [m], Time [s], OpticalPower [W], X-Phase (Y-Phase) [rad]
  
         Method can be called at several distances. On first call set
         appendFlag = false, otherwsie set appendFlag = true
      */

      void WriteDistTimePowerPhase(string outFile,  int timeIndexIncrement,
                                   double distance, 
				   int appendFlag);

      //! \sa OptSignal::WriteDistTimePowerPhase()


      void WriteDistTimePower(string outFile,  int timeIndexIncrement,
                                   double distance, 
				   int appendFlag);

      //! Writes out real and imaginary parts of SCALAR signal
      //! as function of distance and time

      /*! \sa OptSignal::WriteDistTimePowerPhase()
          \warning Does not work for VECTOR signals
      */

      void WriteDistTimeCplxWave(string outFile, double distance, 
				 int appendFlag);


     //! \sa OptSignal::GetAveragePower(), OptSignal::WriteDistTimePowerPhase()

     void WriteAveragePowerdBm(string OutFileName,double distance,
			    int appendFlag); 

     //! \sa OptSignal::WriteAveragePowerdBm

     void WritePowerdBmChannel(string OutFileName,double distance,
                                   int appendFlag,int ChannelArrayIndex);
  
     //! \sa OptSignal::WriteAveragePowerdBm

     void WritePowerdBmFreqWindow(string OutFileName,double distance,
                                  int appendFlag,
                                  double LeftFreq,double RightFreq);

      //! Writes amplitude and phase of SCALAR complex wave
      //! in time domain to a file

      void WriteFileTimeAmpPhase(string outFile);

      //! Returns local frequency of SCALAR complex wave at time
      //! given by timeIndex in range 0 to OptSignal::qtPoints 

      /*! Local Frequency of a complex wave E(t) = A(t) exp(i Phi(t)) 
          is defined to be (see for example [Agrawal, Nonlinear Optics,p91])
  
                    LocalFrequency(t) = - dPhi/dt 

          The local frequency at t = timeIndex*DeltaTime is computed 
          from the phase using a central difference to approximate 
          the derivative (for second order accuracy). 

           We wrap the phase difference numerator in doifference quatient)
           to be in the range [-pi,pi)

           Local Frequency is in Hz = Cycles per second

      */

      double GetLocalFrequency(int timeIndex);

      //! Rteurns the chirp of SCALAR complex wave at at time
      //! given by timeIndex in range 0 to OptSignal::qtPoints

      /*! The Chirp of a complex wave E(t) = A(t) exp(i Phi(t)) 
  	  is defined to be 
  
                    Chirp(t) =  -d^2Phi/dt^2 

           Chirp is in Hz^2 whereas LocalFrequency is in Hz.
 
          Usually we are interested in computing the Chirp at time 
          t_0 = central time of a given pulse.
          The negative sign is included so that the sign of the chirp
          agrees with the sign of the chirp parameter C in [NLO,p67]

          The pulse can then be approximated by using a Taylor's 
          series for Phi(t) about t_0

           E(t) ~ A(t) exp( i [Phi(t_0) - 2*pi*LocalFrequency(t_0)*(t-t_0) - 
                          0.5*4*pi^2*Chirp(t_0)*(t-t_0)^2] )  

           The Chirp at t = timeIndex*DeltaTime is computed from the phase
           by  using a central difference to approximate the 2nd derivative
           (for second order accuracy). 
        */

      double GetChirp(int timeIndex);

      //! Computes the average chirp of the marks (1's)
      //! at a given time within the bit slot by making calls to 
      //! OptSignal::GetChirp().

      /*! \warning  Only works for a single channel SCALAR signals.
       */

      double ComputeAverageChirpOfMarks(double TimeWithinBitSlot);

      //! Writes out LocalFrequency and Chirp as function of time
      //! by making calls to OptSignal::GetLocalFrequency() and 
      //! OptSignal::GetChirp()

      void WriteLocalFrequencyAndChirp(string outFile);

      //! \sa OptSignal::WriteLocalFrequencyAndChirp()

      void WriteChirp(string outFile);


     /*! \brief
          Returns the result of wrapping a periodic variable, periodicVariable,
          so that it is in the range [leftEndPoint,rightEndPoint)

          Can use this template for example with T=double by calling as
          
          double WrappedVar = WrapPeriodically<double>(Var,Min,Max);
     */     

     template<class T>  T WrapPeriodically(T periodicVariable, 
				      T leftEndPoint, T rightEndPoint);


   /*! \brief 
       Returns integer index in range [0,qtPoints) corresponding to 
       floor(WrappedTime/DeltaTime), where WrappedTime is 
       in range [0,TimeWindow)

    */

   int GetFloorIndex(double Time);

   /*! \brief 
       Returns wrapped integer index in range [0,qtPoints) corresponding to
       ceil(WrappedTime/DeltaTime), where WrappedTime is 
       in range [0,TimeWindow) 

    */


   int GetCeilIndex(double Time);

   /*!
   \brief
   Returns real part of X-polarization of signal at time Time.
   We use linear interpolation of Time is not on the discrete time grid
   and interpret Time as being periodic in the range [0,TimeWindow)
   */

   double GetRealX(double Time);

   /*! \brief
       Returns imaginary part of X-polarization of signal at time Time. 
       \sa OptSignal::GetRealX()
   */

   double GetImagX(double Time);

   /*! \brief
       Returns real part of Y-polarization of signal at time Time. 
       \sa OptSignal::GetRealX()
   */


   double GetRealY(double Time);

   /*! \brief
       Returns imaginary part of Y-polarization of signal at time Time. 
       \sa OptSignal::GetRealX()
   */


   double GetImagY(double Time);


  //###### Methods to compute Energy, Written by JZ

  //! BitSlotNumber must be in range [0,OptSignal::StringLength) 
  //!    \sa GetEnergyInTimeRange 

  double GetEnergyInBitSlot(int BitSlotNumber);

  //! \sa OptSignal::GetAveragePower()

  double GetTotalEnergy(void);

  //! \sa OptSignal::GetTotalEnergy()

  double GetEnergyInTimeRange(double LeftTime, double RightTime);

 //! Computes mean time pulse width, using same formula used to
 //! compute mean of a pdf. Integration is over specified time window 

  double ComputeMeanTimePulse(void);
 
  double ComputeMeanTimePulse(double StartTime, double EndTime);

  double ComputeMeanFreqSignal(void);

  double ComputeMeanPhasePulse(void);

  double ComputeMeanPhasePulse(double StartTime, double EndTime);

  //! Assumes single pulse
  
  double ComputeRMSPulseWidth(void);

  //! \sa OptSignal::GetSigmaSqOptSignal, OptSignal::ComputeMeanTimePulse()
  //! \sa OptSignal::ComputeRMSPulseWidth

  double ComputeSigmaSqTimeWindow(double StartTime, double EndTime);

  //! Calls OptSignal::ComputeSigmaSqTimeWindow()
  /*! \warning This returns a quantity like a std deviation not a variance
  */

  double ComputeAverageSigmaMarks(void); 

  //! Returns Root Mean Square signal bandwidth in Hz

  double ComputeRMSSpectralWidth(void); 

  double ComputeMaxPower(void);

  double ComputeMaxPower(int & ArgMaxIndex);

  void OpenAveragePulseWidthFile(string FileName,int WriteFlag);
  void CloseAveragePulseWidthFile(int WriteFlag);

  //! Calls OptSignal::GetSigmaSqOptSignal()

  void WriteAveragePulseWidth(double PropagatedLength,int WriteFlag);

  void ResetJonesMatrix(void);


  //! Called by OptSignal::RegenerateSignalString 
  /*! For TypePulse = NRZ, RZ_GAUSS, RZ_SECH, CRZ_GAUSS */

  void GenerateSignalStringGaussianOrSech(void);

  //! Called by OptSignal::RegenerateSignalString
  /*! For TypePulse = RZ_GAUSS_PATTERN */

  void GenerateSignalStringRZGaussianPattern(void);

   //! Out of date. Use GenerateSignalStringRaisedCosinesWithPhaseShift
   
  void GenerateSignalStringRaisedCosines(void);

    //! Called by OptSignal::RegenerateSignalString
  /*! For TypePulse = INPUT_SEQUENCE */

  void GenerateSignalStringUsingInputSequence(void);
 
    //! Called by OptSignal::RegenerateSignalString
  /*! For TypePulse = QPSK */

  void GenerateSignalStringQPSK_Gaussian(void);
  void GenerateSignalStringQPSK_CW(void);

  void GenerateSignalStringNRZ_RCOS_MachZehnderModulator(void);

  void GenerateSignalStringBPSKPhaseModulator(void);
  void GenerateSignalStringBPSKMachZehnderModulator(void);

  void GenerateSignalStringCSRZMachZehnderModulator(void);
  void GenerateSignalStringRZ33MachZehnderModulator(void);
  void GenerateSignalStringRZ50MachZehnderModulator(void);

  void GenerateSignalStringQPSK_SEQ_PM_NRZ(void);
  void GenerateSignalStringQPSK_DPMZ_PM_NRZ(void);

  void GenerateSignalStringQPSK_SEQ_PM_AM_RZ33(void);
  void GenerateSignalStringQPSK_SEQ_PM_AM_RZ50(void);
  void GenerateSignalStringQPSK_DPMZ_PM_AM_RZ33(void);
  void GenerateSignalStringQPSK_DPMZ_PM_AM_RZ50(void);

  void EOM_MachZehnderModulatorOptData(double V_pi1, double V_pi2,
                                double Phi_10, double Phi_20,
				double SinusoidalRate, 
				double SinusoidalThetaShift);

  void EOM_MachZehnderModulatorOptNRZOLD(double V_pi1, double V_pi2,
                                double Phi_10, double Phi_20,
                                double SinusoidalRate,
                                double SinusoidalThetaShift);

  void GenerateSignalStringCSRZMachZehnderModulatorOLD(void);
  void GenerateSignalStringRZ33MachZehnderModulatorOLD(void);
  void GenerateSignalStringRZ50MachZehnderModulatorOLD(void);

  void EOM_MachZehnderModulatorElecNRZ(double V_pi1, double V_pi2,
                                double Phi_10, double Phi_20,
                                double SinusoidalRate,
                                double SinusoidalTimeShift,
                                double SinusoidalVertShift);

  void SetEOM_V_pi_RelError(double EOM_V_pi_RelError2);

  void SetEOM_Phi_10_RelError(double EOM_Phi_10_RelError2);

  void SetEOM_Phi_20_RelError(double EOM_Phi_20_RelError2);

  void SetEOM_SinusoidalVertShift_RelError(
        double EOM_SinusoidalVertShift_RelError2);

  void EOM_WriteFileElecData(string OutFileName);

  void EOM_SinusoidallyModulateElecData(double SinusoidalRate, 
					double SinusoidalTimeShift,
					double SinusoidalVertShift);

  void EOM_PhaseModulator(void);
 
  string EOM_GetTypeElecData(void);
  string EOM_GetTypeElecFilter(void);

  void EOM_GenerateRaisedCosineBinaryElecData(int ChannelArrayIndex,
                                              bool ZeroOneNotMinusPlusFlag);
  double RaisedCosineForNRZ_RCOS_MZ(double Time);

  void EOM_GeneratePerfectRectBinaryElecData(int ChannelArrayIndex);
  void EOM_GenerateSmoothedRectBinaryElecDataArctanh(int ChannelArrayIndex);
  void EOM_GenerateSmoothedRectBinaryElecDataSuperGauss(int ChannelArrayIndex);

  double EOM_SuperGauss(double Time);
  void EOM_FilterElecData(void);

  void EOM_ElectricalFilterBessel4(cplx *my_power_time,double fwhm_norm);
  void EOM_ElectricalFilterGaussian(cplx *my_power_time,
                                             double fwhm_norm);

  cplx EOM_Bessel4(double omega,double FWHM);
  double EOM_Gaussian(double omega,double FWHM);

  //! Computes the RelevantNumModes used in computation of
  //! local error in OptFiberLocalError
 
  void ComputeRelevantNumModesOFLE(void);

  int GetRelevantNumModesOFLE(void) {return RelevantNumModesOFLE;};

  void SetBandwidthForRelevantNumModesOFLE(double BandwidthGHz);

  //! USE WITH EXTREME CAUTION!! (Written for specific application by JZ)

   void SetBitLengthJZ(double Value) {BitLength = Value;}; 
            
  //! USE WITH EXTREME CAUTION!! (Written for specific application by JZ)

   void SetBitRateChannJZ(double Value)  {BitRateChann = Value;};

  
  //! Do not use. Written for application by Brian Marks

  void SetBitRateChannBM(double Value);

  //! Do not use. Written for application by Brian Marks

  void AddSignalBM(OptSignal * oOptSignal2);

  //! Do not use. Written for application by Brian Marks

  void AddCWSignal(double PowerCWChann, double FreqCWChann);
   
  // ####### Public Data ######

  public:

    cfftw *fft;           
    fftMatrices sfftM; 

    //! A BitString object

    BitString *oBitString; 

    RanNumGen * RNG;

  //! A reduced optical signal, to track the Stokes parameters of noise
  //! (and signal) in each channel. 
  //! Within OptSignal class it is set up to work with GaussianReceiver 
  //! Only works for ONE channel for now. 
  //! (See OptSignal::ReduceToSingleChannel())

    ReduOptSignal * oReduOptSignal;

    cplx *JonesMatrixChannel;  
   

    //! Common work space which is used by all OptFiberLocalError objects 
    OptFiberLocalErrorWorkSpace OFLEWork; 

    //! Work space used for calculation of integrals using the 
    //! Quadrature method in ocsTools.hh 

    double * QuadratureWorkSpace;


  //! Output polarization dispersion vector used in the class 
  //! OptFiberLocalError

  /*! For OptFiberLocalError class we keep track of the Polarization dispersion
   vector at the frequency given by FreqForPDVectorOFLE. 
   OutputPDVectorOFLE and InputPDVectorOFLE are initialized to 0 when
   we Regenerate the signal, assuming DGD = 0 then. (JZ, Jan 25th 2002)
  */
       
    double * OutputPDVectorOFLE;

  //! Input polarization dispersion vector used in the class 
  //! OptFiberLocalError

    double * InputPDVectorOFLE;

  //! The transpose of the 3x3 Stokes transfer matrix used in the class 
  //! OptFiberLocalError

  /*!
      Only used if OptSignal::ComputeInputPDVectorFlagOFLE = true

      Used to compute InputPDVectorOFLE from OutputPDVectorOFLE 
      using the equation 

       InputPDVectorOFLE = StokesTransferMatrixTransposeOFLE*OutputPDVectorOFLE

      For fiber with transfer matrix StokesTransferMatrix, the output Stokes
      vector s_out is related to the input Stokes vector s_in by 

      s_out = StokesTransferMatrix s_in.

      With just PMD (no PDL, PDG), the inverse of StokesTransferMatrix is given
      by its transpose. 

      We store the transpose because it enables us to easily rotate its
      column vectors in OptFiberLocalError::PropagateStokesTransferMatrix.

   */

  double ** StokesTransferMatrixTransposeOFLE;

  // ########## Private Methods

  private:

     void AllocateOptFiberLocalErrorWorkSpace(void);
     void DeallocateOptFiberLocalErrorWorkSpace(void);


  // ##### Private Data #######

  private:

      typeSimulation TypeSimulation;

      int NumChannels;
      
      double CenterFreq;
      double CenterWavelength;
      double ChannSpacing; 

      //! Array of length  OptSignal::NumChannels which stores 
      //! frequency of each channel in Hz

      double * Frequency; 

      //! The time length in seconds of a bit

      double BitLength; 
      double BitRateChann;

      //!  \sa ReduOptSignal::typePulse

      typePulse TypePulse; 

      //! Number of bits

      int StringLength;
      double FWHM_Pulse;
      double ChirpParameter;
      double LinearExtinctionRatio;
      double RiseTime;
      double OptPowerChann;

      //! \sa OptSignal::TypePulse (general)

      string InputTimeDataFileName;

      //! \sa OptSignal::TypePulse (idiosyncratic)
    
      string InputSequenceFileName;

  /*!
      Pointer to a 1D array of length  NumChannels*4 
      which stores the Stokes parameters for each 
      channel.  Accessed using the methods
      OptSignal::SetInStokes(), OptSignal::GetInStokes()

  */

      double *InStokes; 

      /*! 
         Pointer to a 1D array of length  NumChannels*2 
         which stores the Jones parameters for each 
         channel.  Accessed using the methods
         OptSignal::SetInJones(), OptSignal::GetInJones()
      */

      cplx *InJones;  

      //! Randomize input polarization states for different channels and
      //! different Monte Carlo runs 

      bool RandomPolarizationsFlag;

      int qtPoints;
      double TimeWindow,DeltaTime;
      double FreqWindow,DeltaFreq;  

      /*!
        AccumulatedFirstOrderDispersion is integral of beta'' in units s^2.
        It is measured at the central frequency.
        It's value is updated by the OptFiberLocalError or OptFiber class 
        after each z-step is taken by the 
        NLS solver. This works out fine as the Signal is passed to each
        optical fiber. There are methods in the Signal class to reset
        to zero and to
        write out the parameter to a file as a function of z.
        CAUTION: At the moment the updates are are being done by the 
        OptFiberLocalError class not the OptFiber class. Feel free to 
        write the code for OptFiber if you like! (JZ, Feb 8th, 2001)

       AccumulatedSecondOrderDispersion is integral of beta''' in units s^3.
      */

      double AccumulatedFirstOrderDispersion;
      double AccumulatedSecondOrderDispersion;

      ofstream AccumulatedFirstOrderDispersionFile;
      int WriteAccumulatedFirstOrderDispersionFlag;

      ofstream AveragePulseWidthFile;

      ofstream DegreeOfPolarizationFile;

  /*! For OptFiberLocalError we keep track of the Polarization dispersion
      vector at the frequency given by FreqForPDVectorOFLE. 
    
     This frequency is an absolute frequency

     FreqForPDVectorOFLE = RelFreqForPDVector + CenterFreq, where

     OptSignal::RelFreqForPDVector is an optional input parameter whose 
     default value is 0.
  */

      double FreqForPDVectorOFLE;     
     
  //! Input parameter that specifies whether or not to compute the
  //! input polarization dispersion vector using the method
  //! OptFiberLocalError::GetInputPDVector. Default value is 0.

  bool ComputeInputPDVectorFlagOFLE;

  //! 

  //! The number of modes used to compute the Local Error in the
  //! OptFiberLocalError class. Default value if qtPoints.

     int RelevantNumModesOFLE;

  //! Flag to compute RelevantNumModesOFLE
  //! \sa int RelevantNumModesOFLE

     bool ComputeRelevantNumModesOFLEFlag;

  //! DebugLevel
  /*! DebugLevel
       -# 0: No output
       -# 1-4: More and more output
  */

      int DebugLevel;
   
      int MemAllocated; 
      bool UsingOptFiberLocalError;

  
  //! Required when using the semi-analytical noise transmission and 
  //! receiver model (JZ, Dec 19, 2002)

  /*! When using the semi-analytical noise transmission and receiver model 
      we need to keep track
      of the TotalNoiseSpectralDensity in W/Hz, which is automatically updated
      by each OptAmplifier, OptFiberLocalError, and LumpedLoss object. 
      (Note that I haven't implemented this for the OptFiber class.)
      For simplicity for now we assume that the noise is unpolarized 
      and so we have not worried about keeping track of the polarization states
      of the noise and how they are affected by polarization sensitive devices
      and fibers.

      Note that 

      TotalNoiseSpectralDensity = 2*NoiseSpectralDensityPerPolarization.

      A major reason for keeping track of TotalNoiseSpectralDensity is
      to ensure that the VECTOR_SATURATED and VECTOR_FIXED_TOTAL_OUTPUT_POWER
      amplifiers work correctly. 
     
      If we want to model partially polarized noise it would be better to
      link the reduced model directly to the full model.
  */

      double AccumulatedTotalNoiseSpectralDensityGaussianStatistics;

  //! The bandwidth of the noise as used in the semi-analytical noise 
  //! transmission and receiver model (in Hz)

  double NoiseBandwidthGaussianStatistics;

  typeEOMElecData TypeEOMElecData;
  typeEOMElecFilter TypeEOMElecFilter;
      
  fftPowerMatrices EOM_sfftPM;
  fftPowerMatrices EOM_sfftPM1;
  fftPowerMatrices EOM_sfftPM2;

  double EOM_RiseTime;
  double EOM_SlopeFactor;
  double EOM_ShiftAmt;
  int EOM_SuperGaussExponent;
  double EOM_ElecFilter_FreqFWHM;

  double EOM_V_pi_RelError;
  double EOM_Phi_10_RelError;
  double EOM_Phi_20_RelError;
  double EOM_SinusoidalVertShift_RelError;

};

#endif /* _OCS_OPT_SIGNAL_HH_ */



