// $Id: ocsRanNumGen.hh,v 1.2 2008/08/08 02:23:11 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################



//## Method for generation of random number for importance 
//#  sampling by Ivan Lima

#ifndef _OCS_RAN_NUM_GEN_HH_
#define  _OCS_RAN_NUM_GEN_HH_

#ifndef _OCS_SYSTEM_HDRS_
#define _OCS_SYSTEM_HDRS_

#include <iostream.h>
#include <string.h>
#include <math.h> 
#include <fstream.h>
#include <complex.h>
#include <stdlib.h>
#include <stdio.h>
#include <strstream>

#include <string>
  using namespace std;


#endif /* _OCS_SYSTEM_HDRS_ */

#include "ocsConst.hh"
#include "ocsReducedModelTools.hh"
#include "ocsReadWriteTools.hh"


//! The Random Number Generator class

/*!

 The RanNumGen class uses the Random Number Generator (RNG)
 ran2 from Numerical Recipes in C. 
 A helpful discussion of this random number generator 
 can be found in the book  Numerical Recipes in C or
 <a href="http://www.nr.com/">Online Numerical Recipes</a>.
 Additional discussion can be found in "Random Number Generation and 
 Monte Carlo Methods" by James E. Gentle (Springer).
 
 The class contains methods that read and write the state
 of the RNG from and to a file. When running multiple jobs
 this feature enables the user to either continue along the 
 random sequence from where you left off at the end of the last
 job or restart the RNG at the point in the sequence where the previous job
 started. The RNG can also be restarted within a job.

 The class also contains methods that are used in importance sampling.
 

*/

class RanNumGen
{

public:

  //! Class constructor
  
  RanNumGen(string InFileName);

  //! Class destructor
  /*! Calls the method   void WriteRanSeqParameters(string OutFileName)
      which writes the final state of the RNG to a file. Hence it should
      be explicitly called in application programs!
      
      \sa void WriteRanSeqParameters(string OutFileName)
  */

  ~RanNumGen();

  //! Saves the current state of the random number generator 
  /*! Should be used in conjunction with RanNumGen::ResetRanNumGen
   */

  void SaveCurrentState(void);

  //! Resets the RNG to the state it was in at the beginning of the job
     //! or that was saved by a call to RanNumGen::SaveCurrentState
  

  void ResetRanNumGen(void);

  /*! \brief Returns the next random number in the sequence. Returns values
      between 0 and 1.  

      This code was taken directly from Numerical Recipes in C, 
      function ran2.\n
      See  http://www.nr.com/ for documention.
  */

  float GetRanNum(void);



  //! This method generates a biased sample for importance sampling.

  float GetRanCosineBiased(double alpha, double beta, 
                           double *LikelihoodRatio);
  float GetLikelihoodRatioForRanCosineBiased(double alpha,
                               double ValFunction);			   
			   
  float GetRanSignBiased(double alpha, 
                         double *LikelihoodRatio);


  /*! The method returns a float which is the value of the biased sample.
     \param alpha: The biasing parameter. \n
     \param LikelihoodRatio: The value of the liklihood ratio
            of the sample that has been generated.
  
      This is the biasing function that we used in the Reduced Model
      Applications.

  */


  float GetRanCosineBiased(double alpha, double *LikelihoodRatio);

  //! Returns a random sample from a the 1-dimensaional Gaussian distribution
  //! with mean 0 and standard deviation 1.
  /*! 
      This code was taken directly from Numerical Recipes in C.
      See  http://www.nr.com/ for documention.
  */

  float GetGaussianDeviate(void);

  //! Initializes the state of the RNG by reading the state parameters
  //! from a file

  void ReadRanSeqParameters(string InFileName);

  //! Writes the current state of the RNG to a file.
  /*! \sa  ~RanNumGen() */

  void WriteRanSeqParameters(string OutFileName);


private:

  static const long IM1 = 2147483563;
  static const long IM2 = 2147483399;
  static const double AM = 1.0/2147483563;
  static const long IMM1 = IM1 - 1;
  static const long IA1 =  40014;
  static const long IA2 = 40692;
  static const long IQ1 = 53668;
  static const long IQ2 = 52774;
  static const long IR1 = 12211;
  static const long IR2 = 3791;
  static const int NTAB = 32;
  static const long NDIV = (1+IMM1/NTAB);
  //static const double EPS = 1.2e-7;
  //static const double RNMX = 1.0-EPS;

  static const double RNMX = 1.0 - 1.2e-7;

  long idum,  idumStart;
  long idum2, idum2Start;
  long iy, iyStart;
  long iv[NTAB], ivStart[NTAB];

  //! The name of the file to which the current state of the RNG is written

  string ContinuationFileName;

  //! The name of the file to which the state of the RNG at the start
  //! of the current job is written.

  string RestartFileName;

  //! DebugLevel

  /*! DebugLevel to LogFile
       -# 0: No output 
       -# 1: Maximum output
  */

  int DebugLevel;

};

#endif  /* _OCS_RAN_NUM_GEN_HH_ */

