// $Id: ocsReceiverSubsystem.hh,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_RECEIVER_SUBSYSTEM_HH_
#define _OCS_RECEIVER_SUBSYSTEM_HH_

#include "ocsOptSignal.hh"

#include "ocsOptFilter.hh"
#include "ocsOptDemuxer.hh"
#include "ocsPhotodetector.hh"
#include "ocsElecFilter.hh"
#include "ocsHistogram2D.hh"
#include "ocsElecSignalStat.hh"
#include "PostDispComp.hh"
#include "ocsOptDataWriter.hh"


class ReceiverSubsystem
{

  // ########### Methods ###################

public:

  // Written by John Zweck, Nov/Dec 2000.

  // This class is designed to take a (WDM) optical signal, and a particular 
  // choice of WDM channel, passes the signal through an optical
  // demuxer which extracts the given channel (using an OptDemuxer
  // object), then through a Photodetector and ElecFilter.
  // Finally signal statistics and eye diagrams are computed.

  // The idea for writing the class was to simplify the way the 
  // existing receiver classes are called in a WDM application.
  // Clearly one can taylor this class for your own application.
  // In this sense it is maybe more of an application class than
  // a library class, but nevertheless should go in the library
  // to help people write their own receiver subsystems.


  // To declare an array of ReceiverSubsystems we need to use a 
  // default constructor which doesn't actually do anything.
  // Once the array has been defined the method ReceiverSubsystemInit
  // is called, which really is the constructor
  // In the event that you want to declare a single ReceiverSubsystem
  // use the constructor (*) below, which simply calls ReceiverSubsystemInit.


//###########################################################
//## In addition to the ReceiverSubsystem class described 
//#  above, the receiver also has a simplified version.
//#  In the simplified version the only task carried out 
//#  is the extraction of the selected channel. Thus, it 
//#  require far less parameters and do not execute other
//#  tasks. There are two modes of operation: one is the 
//#  continuous extraction of the channel that was specified
//#  in the input. If the user want any other channel, he 
//#  only need to send the ChannelArrayIndex as a parameter
//#  to the ReceiveSingleChannel() method.
//#  The receiver subsystem contains all the electrical 
//#  domain modules for the selected channel.
//#  By: Ivan Lima.
//##########################################################


  ReceiverSubsystem();  // default constructor
  
  ReceiverSubsystem(OptSignal * WDMSignal,RanNumGen *RNG2,
                    string InDir, string Job2,
		    int DemuxChannArrayIndex2,OptDataWriter * DataWriter2); 
                               // (*) constructor

  //################################################################
  //## Constructors for the Simplified receiver (Ivan Lima in 7/7/01)
  ReceiverSubsystem(OptSignal *WDMSignal, 
                    int DemuxChannArrayIndex2,
                    string OptFilterFileName,
                    string ElecFilterFileName,
                    string PhotodetectorFileName, 
                    string ElecSignalStatFileName);

  void ReceiverSubsystemInit(OptSignal * WDMSignal,RanNumGen *RNG2,
                    string InDir, string Job2,
		    int DemuxChannArrayIndex2, OptDataWriter * DataWriter2);

  // Use this method if you have many many receivers and want to save
  // memory.

  void ReceiverSubsystemInit(OptSignal * WDMSignal,RanNumGen *RNG2,
                                     string InDir, string Job2, 
                                     int DemuxChannArrayIndex2,
                                     OptDataWriter * DataWriter2,
			             string PostDispCompFileNameSuffix2,
                                     fftPowerMatrices & sfftPM2,
                                    fftPowerMatrices & WorksfftPM2,
			             OptSignal * GlobalSCSignal,
			         OptSpectrumAnalyzer * PowerMeter2,
                                  OptDemuxer *Demuxer2,
				   cplx * MeasuredFilterFreq);

  void SetReceiverMapNumber(int Value) { RecMapNum = Value;};

  //## Initialization methods for the simplified receiver (IL)
  void InitSimplifiedReceiver(OptSignal  *WDMSignal, 
                              int DemuxChannArrayIndex2,
                              string OptFilterFileName,
                              string ElecFilterFileName,
                              string PhotodetectorFileName, 
                              string ElecSignalStatFileName); 

  ~ReceiverSubsystem();

  void ReceiveSignalBackToBack(OptSignal * WDMSignal);

  void ReceiveSignal(OptSignal * WDMSignal, int UpdateIvansEDFlag,
                     int OutputPD_DataFlag,int ExptNum,int MapNum,
                     double PropagatedLength2);

  // The above method is called at the end of each run through the 
  // loop in a Monte Carlo simulation. The flag UpdateIvansEDFlag
  // is used to indicate whether or not to write out data to 
  // the EyeDiagram in the ocsElecSignalStat class written by Ivan Lima.
  // If you do a very large number of runs, the Eye Diagram data file quickly 
  // becomes exceedingly large. 
  // However, the ocsElecSignalStat class still computes Q factors and BERs
  // if the flag set to 0.
  // The OutputPD_DataFlag determines whether or not the electrical signal
  // is output to a data file or not.

   void PostProcessReceiverStats(int ExptNum,int MapNum,int AppendFlag);
   void WriteStats(int ExptNum, int MapNum, int AppendFlag);

  // The above method should be called at the end of the program, after
  // all Monte Carlo simulations are done. 
  // It computes Q factors, BER's and writes Ron's Eye Diagram Histogram
  // data to a file, which can then be further post-processed using 
  // a Matlab program (ask Ron for it!).

   //##############################################################
   //## Single channel extraction for the simplified receiver
   //## The second method allows the user to select any channel (IL)
   void ReceiveSingleChannel(void);
   void ReceiveSingleChannel(int DemuxChannelArrayIndex2);
   void DemultiplexSingleChannnel(int DemuxChannArrayIndex2);
   void ReceivePreviouslyDemultiplexedSingleChannnel(void);

  // ##################################################

void ReceiveSignalBackToBackCenterFreq(
                                                OptSignal * WDMSignal,
                                                double CenterFreqOptFilter,
                                                BitString *SCBitString,
						bool OutputEye,
                                                bool DoDemuxFilter);

void ReceiveSignalCenterFreq(OptSignal * WDMSignal, 
                             double PropagatedLength2,
                             int MapNum,
			       double CenterFreqOptFilter,
			       BitString *SCBitString,
			     bool OutputEye,
			     bool DoDemuxFilter); 


  void WriteStatsCenterFreq(int MapNum,bool AppendFlag);


  // ########## Data ###########
public:

  OptFilter * TunableFilter;
  OptDemuxer *oOptDemuxer;
  OptSignal * SCSignal;
  OptFilter * oOptFilter;
  OptFilter * oOptFilterPD;
  Photodetector * oPhotodetector;
  ElecFilter * oElecFilter;
  ElecSignalStat * oElecSignalStat;
  Histogram2D *RonEyeDiagram;
  PostDispCompSubSystem *oPostDispComp;
  OptDataWriter * DataWriter;

private:
  OptSignal *WDM_OrigSignal;  //## Used by the simplified receiver
  int FlagSimplifiedReceiver;
  int DemuxChannArrayIndex;
  string Job;
  int ObjectInitialized;
  double PropagatedLength;

  int RecMapNum;
  string PostDispCompFileNameSuffix;

};



#endif  




