// $Id: 
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_REDU_OPT_AMPLIFIER_HH_
#define _OCS_REDU_OPT_AMPLIFIER_HH_

#ifndef _OCS_TYPE_AMPLIFIER_
#define _OCS_TYPE_AMPLIFIER_

enum typeAmplifier { NO_AMPLIFICATION = 0,
                     SCALAR_SIMPLE = 1, SCALAR_SATURATED = 2, 
                     VECTOR_SIMPLE = 3, VECTOR_SATURATED = 4, 
                     VECTOR_FIXED_OUTPUT_POWER = 5,
                     VECTOR_FIXED_TOTAL_OUTPUT_POWER = 6 };

#ifndef _TYPE_AMPLIFIER_NOISE_
#define _TYPE_AMPLIFIER_NOISE_

enum typeAmplifierNoise { NOISE_OFF = 0, 
                          NOISE_ON_CONST_POWER_RANDOM_PHASE = 1,
			  NOISE_ON_GAUSSIAN_WHITE = 2, 
                          NOISE_ON_NO_RNG = 3,
			  NOISE_ON_SEMI_ANALYTICAL = 4, 
			  NOISE_ON_BIASED = 5}; 
						  
							  
#endif // _TYPE_AMPLIFIER_NOISE_

#endif /* _OCS_TYPE_AMPLIFIER_ */

#include "ocsReduOptSignal.hh"
#include "ocsReduPolDepAttenuator.hh"

// ##############################################
// #### Classes of objects for the reduced model
// ##############################################

// #### Class ReduOptAmplifier

//! The reduced optical amplifier. \n
//! The model includes a simple model for gain saturated amplifiers
//! as well as polarization dependent gain (PDG).

/*! For a simple application you just need to call the constructor
    ReduOptAmplifier( string InFileName,ReduOptSignal *oReduOptSignalTmp)
    and the method void AmplifyOptSignal(void)\n

*/


class ReduOptAmplifier
{
   public:
  
  //! Constructor
 
     ReduOptAmplifier( string InFileName,
                    ReduOptSignal *oReduOptSignalTmp);

  //! Destructor, calls void ReleaseMemory(void)

      ~ReduOptAmplifier();   

  //! Used to initialize OutputPowerLinear.

  /*! Our simple amplifier gain saturation model keeps the total
      power in the optical signal constant. 
      To determine the constant  output power level per channel the user
      should call this method in an application 
      once just after the (noiseless) signal has been generated.
      It sets OutputPowerLinear to be 
      the average power per channel on a linear scale of the current 
      signal. 

      \sa OutputPowerLinear, \n
          void AmplifyOptSignal(void)

  */

      void SetFixedOutputPowerFromCurrentOptSignal(void);
      void SetFixedTotalOutputPowerFromCurrentOptSignal(void)
           {SetFixedOutputPowerFromCurrentOptSignal();};

  //! Amplifies the reduced optical signal and adds PDG

      void AmplifyOptSignal(void);     

  // ! \sa typeAmplifier TypeAmplifier

      void SetTypeAmplifier(typeAmplifier TypeAmplifier2);

      typeAmplifier GetTypeAmplifier(void)
             {return TypeAmplifier;};

// ! \sa typeAmplifierNoise TypeAmplifierNoise

      void SetTypeAmplifierNoise(typeAmplifierNoise TypeAmplifierNoise2);

      void SetPolDepGain(double PolDepGainOptAmplifier2)
              {PolDepGainOptAmplifier = PolDepGainOptAmplifier2;};

      void SetGain(double dBGain2);

      double SteadyStateSaturatedGain(double InputPower);

      void SetGainCompressionSaturatedAmplifier(
                   double GainTarget_dB, double GainCompression_dB, 
		   double OutputPowerTarget);

  //! Sets NumChannels = oReduOptSignal::NumChannels

      void SyncNumChannels(void);
      
// ! \sa typeAmplifier TypeAmplifier

      string WriteTypeAmplifier(void);

// ! \sa typeAmplifierNoise TypeAmplifierNoise

      string WriteTypeAmplifierNoise(void);

  //! NOT USED by reduced model

      void WriteSteadyStateGainVersusPower(string OutFileName,
                       double InputPowerMaxLinear, int NumPowerSteps);


      void WriteSteadyInputPowerVersusOutputPower(string OutFileName,
                   double InputPowerMaxLinear, int NumPowerSteps);

  //! NOT USED by reduced model

      double GetSaturatingPowerLinear(void){return SaturatingPowerLinear;}; 

      double GetPolDepGain(void) 
                {return PolDepGainOptAmplifier;};

      double GetOutputPowerLinear(void)
             {return OutputPowerLinear;};
	     
      double GetLinearGainOptAmplif(void)
            {return LinearGainOptAmplif;};	     

   private:

  //! Allocates memory for  
  //! double * ReduOptAmplifier::NormalizedNoiseAmplitude and initializes
  //! its values.

      void AllocateNoiseAmplitudeChannels(void);

  //! Releases memory allocated by void AllocateNoiseAmplitudeChannels(void)

      void ReleaseMemory(void);


  //! Adds the PDG. Called by void AmplifyOptSignal(void)

      void AddPolDepGainEffect(void);

  //! This method computes the linear gain of the amplifier.
  /*! The method returns the linear gain required to change the 
      total power from an input value given by the parameter
      InputPower to the required TotalOutputPower = 
      NumChannels*OutputPowerLinear. Note that the gain may actually
      be a loss in some circumstances.

     \sa OutputPowerLinear

  */

      double GetGainFixedOutputPower(double InputPower);

   //// ! Seems not to be used!!
      //double GetGainFixedOutputPowerDiffNoise( double InputPower);

  //! Pointer to the reduced optical signal object that is being amplified

      ReduOptSignal *oReduOptSignal;

  //! A ReduPolDepAttenuator object that is used internally 
  //! to do the PDG

      ReduPolDepAttenuator *oReduPolDepAttenuator;


  //! The bandwidth of the polarized noise in a channel
  //! \sa ReduOptSignal::PolarizNoiseBandwidth

      //double PolarizNoiseBandwidth;

      double StokesSignalTmp[4];
      double StokesNoiseTmp[4];           

      double CenterFreq;
      double ChannSpacing;
      int    NumChannels;

      double dBGain;         


  //! The output power per channel we want in the case of a  noiseless 
  //! amplifier.

  /*! CAUTION: \n
      With ReduOptAmplifier::TypeAmplifier = VECTOR_FIXED_TOTAL_OUTPUT_POWER
      and when noise is added by the amplifier we keep the total power
      TotalPower = NumChannels*OutputPowerLinear constant. 
      \sa void AmplifyOptSignal(void)
  */

      double OutputPowerLinear;

  //! The polarization dependend gain in dB, same as x_PDG in [WM 2001] p 489.

      double PolDepGainOptAmplifier;   

  //! The linear gain of the amplifier
      double LinearGainOptAmplif; 
  
      double NoiseFigOptAmplifier;
      double LinearNoiseFigOptAmplif;
      double SpontaneousEmissionFactor;

      // ## variables for the saturation model 

  //! NOT USED by reduced model
      double AmpLength;  
  //! NOT USED by reduced model
      int NumZSteps;
   //! NOT USED by reduced model
      double SaturatingPower;
  //! NOT USED by reduced model
      double SaturatingPowerLinear;
  //! NOT USED by reduced model
      double SaturatingPowerdBm;
      //! NOT USED by reduced model
      double UnsaturatedGaindB;
  //! NOT USED by reduced model
      double UnsaturatedGainLinear;
  //! NOT USED by reduced model
      double UnsaturatedDifferentialGainLinear;

  //! This double array stores the NormalizedNoiseAmplitude factor
  //!     for each channel.
  /*! This factor tells us how much noise is added
      by the amplifier and depends on the frequency of the channel.
      Note that we are using an amplifier with flat gain.
      The frequency dependence of the amount of noise that is added
      is explained in both of Agrawal's books.
  */

      double *NormalizedNoiseAmplitude;

  //! The type of amplifier. 
  /*! In the reduced model we use 
      TypeAmplifier = 6 = VECTOR_FIXED_TOTAL_OUTPUT_POWER

  */

      typeAmplifier TypeAmplifier;
    
  //! The type of amplifier noise
  /*! In the reduced model we use TypeAmplifierNoise = ON
      which means that we include noise in the system
  */
      typeAmplifierNoise TypeAmplifierNoise;
    
      int DebugLevel;
};

#endif //_OCS_REDU_OPT_AMPLIFIER_HH_
