// $Id:
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_REDU_OPT_FIBER_HH_
#define _OCS_REDU_OPT_FIBER_HH_

#include "ocsReduOptSignal.hh"

// ##### Class ReduOptFiber

//! The reduced optical fiber class.
/*! This class is used to propagate the reduced optical signal
    through an optical fiber with PMD using the coarse step method
    applied to the Stokes parameters of the signal and noise in each
    channel. It also imparts a (lumped) fiber loss to the signal. The
    loss was included to make it simpler to compare the reduced model
    to a full time and frequency domain model. Note that chromatic
    dispersion and thr fiber's Kerr nonlinearity are not included 
    in the reduced optical fiber. 

    Details of the algorithm can be found in Wang and Menyuk, JLT Vol 19,
    Num 4, April 2001 [WM 2001], page 488. The reader unfamiliar with
    the coarse step method is also encouraged to read Ref 8 of [WM 2001]
    and/or Curtis Menyuk's paper in the J. Engineering Math. Vol 36 pp 113-136,
    1999. [JEM99]. The paper "Comparison of Polarization  Mode Dispersion
    Emulators" by Ivan Lima et al. may also be helpful as this code is 
    based on the full model code Ivan used to do the research in that paper.
*/


class ReduOptFiber
{
   public:

  //! Class constructor

  ReduOptFiber(string InFileName,
	       ReduOptSignal *oReduOptSignalTmp,
               RanNumGen *oRanNumGenTmp); 

  //! Copy constructor
  /*! Calls CopyReduOptFiber(const ReduOptFiber &oReduOptFiber) */

  ReduOptFiber(const ReduOptFiber &oReduOptFiber);

  //! Destructor 
  /* Calls \sa ReleaseMemory() */

      ~ReduOptFiber();

  //! Called by the copy constructor
  /*! \sa   ReduOptFiber(const ReduOptFiber &oReduOptFiber) */

      void CopyReduOptFiber(const ReduOptFiber &oReduOptFiber);

  //! Gets a random realization of a fiber with PMD.

      void GetFiberSampleCompleteScatt(void);               

  //! Sets the Euler angles of the fiber section indexed by IndexStep
  /*! \param IndexStep: The index of the fiber section. An int in range
                        0,..,qtSteps-1.\n

      \param Theta: The value of the Euler angle theta 
      \param Phi: The value of the Euler angle phi 
      \param Psi: The value of the Euler angle psi.

      \sa void GetFiberSampleCompleteScatt(void)               
  */


  void SetScatteringMatrix( int IndexStep, double Theta,
                            double Phi, double Psi);


  //! Method enables user to change the length of fiber and number of
  //! sections in the fiber during the course of a simulation.
  /*! Sets 
    -# LengthFiber = Length
    -# qtStepsFiber = numSteps
    -# Calls  ReleaseMemory()
    -# Calls AllocateMemory_SetLengthSteps() 

     \sa ReleaseMemory(), AllocateMemory_SetLengthSteps() 
  */

      void SetLengthFiberAndQtStepsFiber(double Length,int numSteps);

  //! Allows user to set the lengths of the fiber sections.
  /*! \param double *usrStepLength: A user-defined double array of size qtSteps
      that contains the values of the lengths of the fiber 
      sections to be set.\n

      The method also computes the sum of the new section lengths and 
      sets FiberLength to this sum.
  */

      void SetStepLengths(double *usrStepLength); 

      void SetAttenuation(double AttenuationFiber2);
      
      double GetAttenuationFiber() {return AttenuationFiber;};      

      double GetLengthFiber() {return LengthFiber;};

  //! Used to randomly assign the lengths of the fiber sections.
  /*! Currently not used by the reduced model */

      void GetRandomStepLengths(double RelativeRangeLengths); 

  //! Used to set the class paramter NumChannels to be the same
  //! as oReduOptSignal->NumChannels.

      void SyncNumChannels(void);

  //! Propagates the reduced dignal through the fiber

      void PropagateFiberCoupledNLS(double *PropagatedLength);
      void PropagateFiberCoupledNLS(void)
         {double PropagatedLength;
	  PropagateFiberCoupledNLS(&PropagatedLength);}; //## IL (9/6/02)      

  //! Writes the lengths of the fiber sections to a file.
  /*! \param outFile: the name of the output file (char*) \n
   
       The output is given in two columns. Column 1 is the integer
      index of the fiber section, starting from 0. Column 2 is the
      length of the section as stored in StepLength[ii].\n
  */

      void WriteFileSteps(char *outFile); 

  //! Sets the PMD strength as given by the Mean differential group delay
  //!    per square root length. 

  /*! \param MeanDGD_SqrtLength2: The Mean differential group delay
             per square root length in units of ps/sqrt(km)

      Initializes MeanDGD_SqrtLength,PMD_Fiber and dBeta_2dOmega.
      The parameter dBeta_2dOmega is used in 
      void PropagateFiberCoupledNLS(double*)
      
      \sa void SetPMD_Fiber(double PMD_Fiber2)\n
          void PropagateFiberCoupledNLS(double*)\n
          MeanDGD_SqrtLength,PMD_Fiber, dBeta_2dOmega
  */

      void SetMeanDGD_SqrtLength(double MeanDGD_SqrtLength2);

  //! Sets the PMD strength as given by PMD_Fiber

  /*! \param PMD_Fiber2: The PMD of the fiber in ps/sqrt(km)

      Initializes PMD_Fiber, MeanDGD_SqrtLength, and dBeta_2dOmega.
      The parameter dBeta_2dOmega is used in 
      void PropagateFiberCoupledNLS(double*)
      
      \sa void SetPMD_Fiber(double PMD_Fiber2)\n
          void PropagateFiberCoupledNLS(double*)\n
          MeanDGD_SqrtLength,PMD_Fiber, dBeta_2dOmega
  */

      void SetPMD_Fiber(double PMD_Fiber2);


  //! Conversion factor from PMD_Fiber to MeanDGD_SqrtLength
  /*! This conversion factor is different in short length (deterministic)
      regime and in the long length (stochastic) regime. \n

      In the long length regime, we assume DGD is Maxwellian distributed 
      and so the root mean square DGD is related to the mean DGD by 
      a factor of RMS2meanDGD = sqrt(8./(3.*pi)). 
      This conversion factor will be used is qtSteps > 1 \n
      
      Also, if the fiber consists only of one section, qtSteps=1, then
      we are in the short length regime and PMD is deterministic
      not stochastic and so it makes sense to define RMS2meanDGD = 1.\n

     \sa void SetPMD_Fiber(double PMD_Fiber2) \n
         void SetMeanDGD_SqrtLength(double MeanDGD_SqrtLength2)
         OptFiber::rms2meanDGD(void)
  */

    double RMS2meanDGD(void);
  


  //! Returns the differential group delay (DGD) of the 
  //! current random realization of the fiber at the given frequency.
 
  /*! The DGD is given in units of seconds.
      \param double Freq: The frequency at which the DGD is to be computed.
      This frequency must be an absolute frequency, i.e. it is typically 
      on the scale of THz. So for instance the DGD at the central frequency
      of  the simulation window is given by 
      GetDGD(oReduOptSignal->GetCenterFreq()) 

      The method computes the polarization dispersion vector 
      at the given frequency at the output
      of the fiber and returns its length, which is the DGD.

  */

      double GetDGD(double Freq);   

  //! Returns the length of the second order polaroization dispersion vector.
  /*! \sa  double 
           GetSndOrderOutPD_Vector(double Freq, double * SndOrderOutPD_Vector)
  */

      double GetSndOrderDGD(double Freq);

  //! Computes the polarization dispersion vector at the given frequency
  //!  at the fiber input

      void GetInPD_Vector(double Freq,double *InPD_Vector);     

  //! Computes the polarization dispersion vector at the given frequency
  //!  at the fiber output

      void GetOutPD_Vector(double Freq,double *OutPD_Vector);

  //! Computes the second order polarization dispersion vector at the 
  //! given frequency at the fiber output

      void GetSndOrderOutPD_Vector(double Freq,double *SndOrderOutPD_Vector);
    
  //! Returns the expected value of the DGD in seconds for 
  //! a fiber of length LengthFiber and PMD = PMD_Fiber.

      double GetExpectedDGD() 
             {return sqrt(8./(3.*pi)*LengthFiber)*PMD_Fiber; };

  //! Returns the expected value of the second DGD in seconds squared for 
  //! a fiber of length LengthFiber and PMD = PMD_Fiber.
 
      double GetExpectedSndOrderDGD() 
             {return sq(GetExpectedDGD())/sqrt(3.); };

  //! Returns the Mean DGD per square root length in ps/sqrt(km)

      double GetMeanDGD_SqrtLength(void) 
             {return sqrt(1e3)/1e-12*MeanDGD_SqrtLength;};


   private:

  //! Allocates memory to store the  Euler angles and the lengths 
  //! of the fiber sections. 

  /*! Also initializes dBeta_2dOmega.
      \sa  void ReleaseMemory(),\n
           ReduOptFiber(string ,ReduOptSignal*,RanNumGen *)
           
  */

      void AllocateMemory_SetLengthSteps(void);

  //! Releases allocated memory
  /*! Releases memory allocated by void AllocateMemory_SetLengthSteps(void)\n
      Called by the destructor ~ReduOptFiber()\n
  */ 

      void ReleaseMemory();                  

  //! Pointer to the reduced optical signal object which is to be 
  //! propagated through the reduced fiber. 

      ReduOptSignal *oReduOptSignal;

  //! Pointer to a random number generator object which is used 
  //! to obtain the random realization of the fiber in the method
  //! GetFiberSampleCompleteScatt()

      RanNumGen *oRanNumGen;

  //! Number of channels of oReduOptSignal

      int    NumChannels;

  //! Flag that is set to 1 in call to   GetFiberSampleCompleteScatt()

      int    GotFiberSample;

  //! The length of the fiber in meters

      double LengthFiber;

  //! The number of sections in the fiber. 
  /*!
     In the corase step method for simulating propagation in a fiber 
     with PMD we randomly rotate the axes of birefringence qtStepsFiber times.
  */

      int    qtStepsFiber;
  
  //! The PMD of the fiber in s/sqrt(m). 
  /*! Related to the mean DGD per squre root length by 
      PMD_Fiber = sqrt(3.*pi/8.)*MeanDGD_SqrtLength

      \sa MeanDGD_SqrtLength, dBeta_2dOmega
  */
      double PMD_Fiber;

  //! The mean DGD per square root length of the fiber in s/sqrt(m).
  /*! Related to PMD_Fiber by MeanDGD_SqrtLength = sqrt(8./(3.*pi))*PMD_Fiber
      \sa PMD_Fiber, dBeta_2dOmega
  */


      double MeanDGD_SqrtLength;

  //! The fiber loss in dB/meter
  /*! \sa NepperAttenuationFiber */

      double AttenuationFiber;      

  //! The fiber loss expressed on a log scale in base e = 2.718...
  /*!  NepperAttenuationFiber = AttenuationFiber*log(10.0)/10.0; 
       \sa AttenuationFiber
  */

      double NepperAttenuationFiber;

  //! The central frequency of the reduced optical signal

      double CenterFreq;

  //! Arrays for the Euler angles theta,phi,psi for each random rotation
             
      double *psi,*phi,*theta;

  //! Array to store the lengths of the fiber sections.
  /*! In this implementation of the reduced model we require that
      all sections in the same piece of fiber have the same length!!!
  */
      double *StepLength;

      double cvtheta2,cvtheta;      

  //! The length of the fiber section in meters when the fiber sections
  //! all have the same length 
  
      double DeltaZ;         

  //! The birefringence strength 
  /*! Used in PropagateFiberCoupledNLS(double *).\n
      Obtained from PMD_Fiber by 
      dBeta_2dOmega = PMD_Fiber/sqrt(4*DeltaZ)
      where DeltaZ is the length of each the fiber section

      \sa PMD_Fiber,\n
      PropagateFiberCoupledNLS(double *)
  */

      double dBeta_2dOmega;

  //! The type of random rotation. 
  /*! In the reduced model this is set to 1 to indicate that we use
      the Euler angles */

      int TypeScattering;           
};


#endif /* _OCS_REDU_OPT_FIBER_HH_ */
