// $Id: 
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_REDU_POL_DEP_ATTENUATOR_HH_
#define _OCS_REDU_POL_DEP_ATTENUATOR_HH_

#include "ocsReduOptSignal.hh"

// ##############################################
// #### Classes of objects for the reduced model
// ##############################################

// #### Class ReduPolDepAttenuator

//! The reduced polarization dependent attenuator.\n
//! Models a polarization-dependent loss (PDL) element.\n

/*! In the reduced model the action of a PDL element
    on the Stokes parameters of the signal and of the noise
    is given by Equation (6) of 
    Wang and Menyuk, JLT Vol 19, Num4, April 2001 [WM 2001]. \n
    The action of a PDL element on the additional unpolarized noise
    is given by setting (S_1, S_2, S_3) = (0,0,0) in Equation (6) of [WM 2001].

    For a simple application we call the constructor     \n

    ReduPolDepAttenuator(ReduOptSignal *oReduOptSignalTmp
                          ,double PolDepAttenuationXTmp
                          ,double PolDepAttenuationYTmp)\n

    where we usually set PolDepAttenuationX = 0.0 and \n
    PolDepAttenuationY = The amount of PDL in dB.\n
    This is as in [WM 2001, Eq 5].

    Then we call the method \n
    
    void PolDepAttenuate(void) \n
    
    which passes the signal through the PDL element.

*/

class ReduPolDepAttenuator
{
   public:


  //! Default constructor
  /*! Doesn't do anything! 
   */

      ReduPolDepAttenuator() {ObjectInitialized = 0;};

  //! Constructor
  /*! \param oReduOptSignalTmp: An object in the reduced optical signal class
     
      We need to pass a ReduOptSignal because the PDL element acts on the
      Stokes parameters of the signal and noise in each channel and on
      the additional unpolarized noise all of which are data members in 
      ReduOptSignal.

      This constructor calls \n
      InitPolDepAttenuator(ReduOptSignal *oReduOptSignalTmp)

      \sa void InitPolDepAttenuator(ReduOptSignal *oReduOptSignalTmp),\n
          ReduPolDepAttenuator(ReduOptSignal *, double ,double)
                          
                          
  */

      ReduPolDepAttenuator(ReduOptSignal *oReduOptSignalTmp); 

  //! Constructor
  /*! \param oReduOptSignalTmp: An object in the reduced optical signal class
     
      We need to pass a ReduOptSignal because the PDL element acts on the
      Stokes parameters of the signal and noise in each channel and on
      the additional unpolarized noise all of which are data members in 
      ReduOptSignal.

      This constructor calls the two functions:\n
      InitPolDepAttenuator(ReduOptSignal *oReduOptSignalTmp)\n
      SetPolDepAttenuation(double,double)\n
                         

      \sa void InitPolDepAttenuator(ReduOptSignal *oReduOptSignalTmp),\n
          ReduPolDepAttenuator(ReduOptSignal *, double, double)
                         
  */


      ReduPolDepAttenuator(ReduOptSignal *oReduOptSignalTmp
                          ,double PolDepAttenuationXTmp
                          ,double PolDepAttenuationYTmp);

  //! Initializes the parameters. Called by the constructors.
  /*!
      Initializes the class variables oReduOptSignal, NumChannels, 
      and sets ObjectInitialized = 0, since the amount of PDL
      still needs to be set.

      \sa oReduOptSignal, NumChannels, ObjectInitialized \n
          ReduPolDepAttenuator(ReduOptSignal *,double,double)\n
          ReduPolDepAttenuator(ReduOptSignal *oReduOptSignalTmp)\n 


   */
 
      void InitPolDepAttenuator(ReduOptSignal *oReduOptSignalTmp);

  //! Sets the amount of PDL in the X and Y directions in Jones space.
  /*! We usually set 
      \param PolDepAttenuationX = 0.0 dB (no loss in the X direction)
      \param PolDepAttenuationY = Amount of PDL in dB = x_{PDL} in the 
      notation of [Wang and Menyuk, JLT 2001, above Eq (6)].\n

      The method also sets ObjectInitialized = 1.\n
      \sa  ReduPolDepAttenuator(ReduOptSignal *oReduOptSignalTmp
                          ,double PolDepAttenuationXTmp
                          ,double PolDepAttenuationYTmp)\n

          ReduPolDepAttenuator(ReduOptSignal *oReduOptSignalTmp)\n 
          void SetPolDepAttenuationX(double PolDepAttenuationXTmp)\n
          void SetPolDepAttenuationY(double PolDepAttenuationYTmp)\n

  */


      void SetPolDepAttenuation(double PolDepAttenuationXTmp
                              ,double PolDepAttenuationYTmp);

  //! Sets the PDL in the X direction.
  /*! \param PolDepAttenuationXTmp is in dB\n
   
          The factor of 1/2 comes from the fact that 
          the LinearPolDepAttenuationX 
          acts on the field, not on the power.

\sa  void SetPolDepAttenuationY(double PolDepAttenuationYTmp)\n
    void GetTransmMatrixCoefficients(void)  \n 
    void PolDepAttenuate(void)
   
  */

      void SetPolDepAttenuationX(double PolDepAttenuationXTmp);

  //! Sets the PDL in the Y direction.
  /*! \param PolDepAttenuationYTmp is in dB \n
   
          The factor of 1/2 comes from the fact that 
          the LinearPolDepAttenuationY 
          acts on the field, not on the power.

\sa  void SetPolDepAttenuationX(double PolDepAttenuationYTmp)\n
    void GetTransmMatrixCoefficients(void)  \n 
    void PolDepAttenuate(void)
   
  */
      void SetPolDepAttenuationY(double PolDepAttenuationYTmp);

  //! Returns the amount of loss of power due to PDL in the X direction
  //! on  log scale. 

      double GetPolDepAttenuationPowerX(void)
            {return -Linear2dB(sq(LinearPolDepAttenuationX));};

  //! Returns the amount of loss of power due to PDL in the Y direction
  //! on  log scale. 

      double GetPolDepAttenuationPowerY(void)
            {return -Linear2dB(sq(LinearPolDepAttenuationY));};


  //! Returns the amount of loss of power due to PDL in the X direction
  //! on  linear scale. 

      double GetLinearPolDepAttenuationPowerX(void)
            {return sq(LinearPolDepAttenuationX);};

  //! Returns the amount of loss of power due to PDL in the Y direction
  //! on  linear scale. 

      double GetLinearPolDepAttenuationPowerY(void)
            {return sq(LinearPolDepAttenuationY);};

  //! Initializes the non-zero coefficients of the 4 by 4 PDL 
  //! transmission matrix that acts on the 4 Stokes parameters.
  /*! The PDL is initialized in Jones space by  SetPolDepAttenuation
      but it acts by matrix multiplication 
      on the Stokes parameters of the signal and noise.
      This method initializes the non-zero entries of the matrix
      as in [WM 2001] Eq (6). 
      
      The method is called by \n
      void SetPolDepAttenuation(double,double)
                               
 
      \sa Tm00 Tm01 Tm22
  */


      void GetTransmMatrixCoefficients(void);

  //! Synchronizes the class parameter NumChannels with the
  //! corresponding parameter in  oReduOptSignal.
  /*! 
      Used when sweep over the number of channels. 
   */

      void SyncNumChannels(void);

  //! Transmits the Reduced Signal through the PDL element.
  /*!
    The action of a PDL element
    on the Stokes parameters of the signal and of the noise
    is given by Equation (6) of 
    Wang and Menyuk, JLT Vol 19, Num4, April 2001 [WM 2001]. \n
    The action of a PDL element on the additional unpolarized noise
    is given by setting (S_1, S_2, S_3) = (0,0,0) in Equation (6) of [WM 2001].
   */

      void PolDepAttenuate(void);

   private:
      int ObjectInitialized;
  //! A pointer to the ReduOptSignal object to be transmitted through
  //! the PDL element.

      ReduOptSignal *oReduOptSignal;

  //! The number of channels of oReduOptSignal.

      double NumChannels;       

  //! The amount of loss of field-amplitude due to PDL on a linear scale 
  //! in the X dircetion in Jones space
  /*! \sa ReduPolDepAttenuator::SetPolDepAttenuationY(
                              double PolDepAttenuationYTmp)
  */

      double LinearPolDepAttenuationX;
      
   //! The amount of loss of field-amplitude due to PDL on a linear scale 
  //! in the Y dircetion in Jones space
  /*! \sa ReduPolDepAttenuator::SetPolDepAttenuationY(
                              double PolDepAttenuationYTmp)

  */
      double LinearPolDepAttenuationY; 

//! The nonzero entries of the 4 by 4 
//! PDL transmission matrix that acts on the 4 Stokes prameters.
/*!
 \sa      void GetTransmMatrixCoefficients(void)
 */

      double Tm00,Tm01,Tm22;


};

#endif //_OCS_REDU_POL_DEP_ATTENUATOR_HH_










