// $Id:
//###################################################################
//
//       Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_REDU_POLARIZ_TRANSFORMER_HH_
#define _OCS_REDU_POLARIZ_TRANSFORMER_HH_

#include "ocsReduOptSignal.hh"

#ifndef _OCS_TYPE_SCATTERING_
#define _OCS_TYPE_SCATTERING_
enum typeScattering { RX_RY_RX = 1, 
                      RY_RZ_RY = 2, 
                      RX_RZ_RY = 3};
#endif

//## The options RX_RY_RX and RY_RZ_RY can provide a uniform 
//#  rotation. While RX_RZ_RY is adequate for the importance
//#  sampling applied to PDL. (Ivan Lima)

// ##### Class ReduPolarizTransformer

//! Polarization transformer (or controller) that rotates a
//! ReduOptSignal object on the Poincare sphere.
//! Includes methods to use  importance sampling
//! to compute outage probabilities in WDM systems due to the
//! combined effects of PMD, PDL, and PDG.

class ReduPolarizTransformer
{
   public:

  //! The constructor

  /*! 

  The constructor initializes the following:

  \param int TypeScattering2: 
       Initializes ReduPolarizTransformer::TypeScattering

  \param ReduOptSignal *oReduOptSignalTmp:
      Initializes ReduPolarizTransformer::oReduOptSignal

   FlagImportSamplingInitialized = 0, which will be set to 1 when
   user calls:

   ReduPolarizTransformer::InitImportSampPolarizTransformation

  */

      ReduPolarizTransformer( int TypeScattering2, 
                              ReduOptSignal *oReduOptSignalTmp);

  //! Transforms the ReduOptSignal on the sphere

  /*! The particular transformation is specified by 
      ReduPolarizTransformer::TypeScattering and the 
      angles psi,theta,phi that are passed to this method.
  */

      void TransformPolarization( double psi
                                 ,double theta
                                 ,double phi);


  //!  Transforms the ReduOptSignal on the sphere

  /*!

   Calls 

  void ReduPolarizTransformer::TransformPolarization( double psi
                                 ,double theta
                                 ,double phi);
  
  with phi = 0.

  */


      void TransformPolarization( double psi
                                 ,double theta);


 //!  Transforms the ReduOptSignal on the sphere using the angles
  //! used in the previous call to a TransformPolarization method

  /*!

\sa ReduPolarizTransformer::psiLast, ReduPolarizTransformer::thetaLast,
   ReduPolarizTransformer::phiLast

   */

      void TransformPolarization(void);

  //! Undoes the previous transformation

      void UndoLastPolarizTransformation(void);

  //! Returns the type of scattering

      string WriteTypeScattering(void);


  //! Initializes the Polarization Transformer so it can be used
  //! for importance sampling.

  /*!

Sets FlagImportSamplingInitialized = 1 and initializes the following:
  
\sa ReduPolarizTransformer::BiasingParam
    ReduPolarizTransformer::oRanImportSampPolarizTransform


   */
   

      void InitImportSampPolarizTransformation(
                             double BiasingParam2,
                             RanNumGen *oRanImportSampPolarizTransform2); 

      void SetBiasingParam(double BiasingParam2) 
            { BiasingParam = BiasingParam2;};




  //! Performs a biased rotation of the ReduOptSignal object
  //! which should be called just before entering the PDL element.

  /*!

The biasing method is described in Lima OFC 2002, TuI7.

ATTENTION: The likelihood ratio should be initialized 
to 1 each time we start propagating through the system, i.e.
at the start of each Monte Carlo experiment/sample with importance sampling.
The likelihood ratio should be
passed to the Histogram class together with the sample.
If this is not done, the statistical
results will be completelly wrong.


  */

      void GetImportSampPolarizTransformation(
                             int ChannelArrayIndex,
                             double *LikelihoodRatio);

   private:


  //! TypeScattering is an enumeration variable of type typeScattering.

/*! 

The types of scattering are 

- RX_RY_RX = 1: Can be used to generate rotations on the Poincare sphere
              that are selected from a uniform distribution.  
- RY_RZ_RY = 2: Can be used to generate rotations on the Poincare sphere
              that are selected from a uniform distribution
- RX_RZ_RY = 3: Used in importance sampling for PDL

*/

      typeScattering TypeScattering;


  //! The ReduOptSignal object that is to be transformed

      ReduOptSignal *oReduOptSignal;      

  //! The Euler angles for the previous call to the transformation
  //! method for the current object

  double psiLast, thetaLast, phiLast;
  

      //! A Random number generator RanNumGen
  //! object for the  importance sampling
      RanNumGen *oRanImportSampPolarizTransform; 

  //! The biasing parameter for importance sampling.

      double BiasingParam;

      int FlagImportSamplingInitialized;
};

#endif //## _OCS_REDU_POLARIZ_TRANSFORMER_HH_
