// $Id:

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2004):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

// Written by John Zweck, May 2004

#include "ChiSquareReceiver.hh"

extern ofstream *LogFile;

ChiSquareReceiver::ChiSquareReceiver(string InFileName,
				     NoiseCovariance * oNoiseCovariance2)
{

  oNoiseCovariance = oNoiseCovariance2;
  Signal = oNoiseCovariance->Signal;
  Job = oNoiseCovariance->GetJobName();

 // ### Open Input Parameter File #######

  ifstream InFile(InFileName.c_str(),ios::in);    

  if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }

   
  LogFileSeparator();
  *LogFile << InFileName << endl << endl;


   // ####### Read in Parameters ############

   TypeOutput = typeOutput(ReadInt("TypeOutput"," ",1,2,
			  LOWER_AND_UPPER,OPTIONAL_NO_WARNING,1,&InFile));

   DebugLevel = ReadInt("DebugLevel"," ",0,2,
			  LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);


   ReadInSignalAndCovMatrixFlag = bool(ReadInt(
                          "ReadInSignalAndCovMatrixFlag"," ",0,1,
			  LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile));

   if(ReadInSignalAndCovMatrixFlag)
     {

      SignalAndCovMatrixDir = ReadString("SignalAndCovMatrixDir",
						MANDATORY,"",&InFile); 
      SignalAndCovMatrixJob = ReadString("SignalAndCovMatrixJob",
						MANDATORY,"",&InFile); 

      SignalAndCovMatrixJob = SignalAndCovMatrixDir + SignalAndCovMatrixJob;
     }


   SetOSNRdB = ReadDouble("SetOSNRdB"," ",0,-1,
			  LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile);

   if(SetOSNRdB)
     {
       B_OSA = ReadDouble("B_OSA","Hz",0,-1,
			  LOWER_ONLY,MANDATORY,25e9,&InFile);
 
     }

   // ##### Initialize TimeSlices array ####

   NumBitsToPlot = ReadInt("NumBitsToPlot"," ",0,-1,
			   LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile);

   if(NumBitsToPlot == 0)
      NumBitsToPlot = Signal->GetStringLength();


   ReadInClockRecoveryTimeFlag = 
    ReadInt("ReadInClockRecoveryTimeFlag"," ",0,1,LOWER_AND_UPPER,
            OPTIONAL_NO_WARNING,0,&InFile);

   if(ReadInClockRecoveryTimeFlag)
      ClockRecoveryTime = ReadDouble("ClockRecoveryTime","sec",1,-1,
			             NO_BOUNDS,MANDATORY,0,&InFile);

   switch(TypeOutput)
	 {
	  case PDFS:
	    {

          NumBitsBetweenSlices = ReadInt("NumBitsBetweenSlices"," ",1,-1,
			          LOWER_ONLY,OPTIONAL_NO_WARNING,1,&InFile);

          TimeBetweenSlices = NumBitsBetweenSlices*Signal->GetBitLength();

          NumTimeSlices = NumBitsToPlot;

	    }
          break;

         case EYE:
	   {	   
           NumSlicesPerBit = ReadInt("NumSlicesPerBit"," ",1,-1,
			          LOWER_ONLY,OPTIONAL_NO_WARNING,1,&InFile);

	   NumTimeSlices = NumBitsToPlot*NumSlicesPerBit;
           TimeBetweenSlices = Signal->GetBitLength()/NumSlicesPerBit;

 	   }
 
           break;
	 }


  TimeSlices = new double[NumTimeSlices];

   // ## Initialize CurrentVector #####

   MinCurrent = ReadDouble("MinCurrent","V",1,-1,
			   NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);

   MaxCurrent = ReadDouble("MaxCurrent","V",MinCurrent,-1,
			    LOWER_ONLY,MANDATORY,0,&InFile);

   if(MaxCurrent < MinCurrent)
     cerr << "ABORT: ChiSquareReceiver Error: [MinCurrent,MaxCurrent] = [" 
	  << MinCurrent << "," << MaxCurrent << "]" << endl;

   NumPointsCurrent = ReadInt("NumPointsCurrent"," ",0,-1,
			   LOWER_ONLY,MANDATORY,0,&InFile);

   InFile.close();

   DeltaCurrent = (MaxCurrent - MinCurrent)/(NumPointsCurrent-1);

   CurrentVector = new double[NumPointsCurrent];
   
   PDFTimeSlicesStorageVec = new double[NumTimeSlices*NumPointsCurrent];

   PDFTimeSlices = new double*[NumTimeSlices];

   for(int tt=0;tt<NumTimeSlices;tt++)
     PDFTimeSlices[tt] = &(PDFTimeSlicesStorageVec[tt*NumPointsCurrent]);

   if(TypeOutput == PDFS)
     {
      AveragePDFZeros = new double[NumPointsCurrent];
      AveragePDFOnes = new double[NumPointsCurrent];
      PowerInSlice = new double[NumTimeSlices];
      BitString = new int[NumBitsToPlot];
      BER = new double[NumPointsCurrent];
     }

      if(TypeOutput == EYE)
     {
       EyeMatrixStorageVec = new double[NumSlicesPerBit*NumPointsCurrent];

       EyeMatrix = new double*[NumSlicesPerBit];

       for(int tt=0;tt<NumSlicesPerBit;tt++)
         EyeMatrix[tt] = &(EyeMatrixStorageVec[tt*NumPointsCurrent]);

     }

    MomentsIntegrand =  new double[NumPointsCurrent];
    
    MeanCurrent = new double[NumTimeSlices];
    StandardDeviationCurrent = new double[NumTimeSlices];

     // ## Get parameters from NoiseCovariance class ###

     CovMatrixDim = oNoiseCovariance->GetCovMatrixDim();
     CovMatrixFirstIndex = oNoiseCovariance->GetCovMatrixFirstIndex();
     CovMatrixLastIndex = oNoiseCovariance->GetCovMatrixLastIndex();
     
     ReducedNoiseFreeSignalRealVector = new double[2*CovMatrixDim];
     CovarianceMatrix = new double* [2*CovMatrixDim];

     DeltaFreq = Signal->GetDeltaFreq();   // equals ReducedDeltaFreq
     TimeWindow = Signal->GetTimeWindow(); // equals ReducedTimeWindow
     ReducedDeltaTime = TimeWindow/CovMatrixDim;
     ReducedFreqWindow = DeltaFreq*CovMatrixDim;

     PrepareSymmetricEig();

     // ## Allocate memory for matrices of eigenvectors 
     // ## and vectors of eigenvalues
     
     CovarianceEvalues = new double[2*CovMatrixDim];

     CovarianceEvectorsStorageVec = new double[4*CovMatrixDim*CovMatrixDim];

     CovarianceEvectors = new double* [2*CovMatrixDim];

      for(int kk=0;kk< 2*CovMatrixDim;kk++)
	{
          CovarianceEvectors[kk] = 
	    &(CovarianceEvectorsStorageVec[kk*2*CovMatrixDim]);
	} 

     TypeIntegrationMethod = STEEPEST_DESCENT;
     LowerPole = 0;
     UpperPole = 0;

      // Allocate memory to receive noisefree signal. 
      // We need this for debugging, visualization, and clock recovery    

     RNG = Signal->RNG;

     DemuxedNoiseFreeSignal = new OptSignal(*Signal);

     TunableOptFilterForNoiseFreeSignal = 
       new OptFilter("TunableOptFilter.in",DemuxedNoiseFreeSignal);
  
     OptDemuxerForNoiseFreeSignal = 
       new OptDemuxer(Signal,TunableOptFilterForNoiseFreeSignal,
		      DemuxedNoiseFreeSignal);

     PhotodetectorForNoiseFreeSignal = 
     new Photodetector ("Photodetector.in",DemuxedNoiseFreeSignal,RNG);

     ElecFilterForNoiseFreeSignal = 
       new ElecFilter("ElecFilter.in",PhotodetectorForNoiseFreeSignal);

     ElecSignalStatForNoiseFreeSignal = new
      ElecSignalStat(DemuxedNoiseFreeSignal,PhotodetectorForNoiseFreeSignal);
	
      // Allocate memory to store the transfer functions of the
      // optical and electrical filters. This memory is in the
      // form of an optical and electrical signal, which will
      // be acted on by the filter objects.
   
      OptFilterTransferFunction = new OptSignal(*Signal);
      
      TunableOptFilterForTransferFunction =  new OptFilter(
           "TunableOptFilter.in",OptFilterTransferFunction);

      ElecFilterTransferFunction = new Photodetector(
				    "Photodetector.in",
				    OptFilterTransferFunction,RNG);

      ElecFilterTransferFunction->DetectOptSignal();
            
      ElecFilterForTransferFunction = new ElecFilter("ElecFilter.in",
				           ElecFilterTransferFunction);

      FilterMatrixTimeIndependentStorageVec = 
         new cplx[CovMatrixDim*CovMatrixDim];

      FilterMatrixTimeIndependent = new cplx * [CovMatrixDim];

      for(int kk=0;kk<CovMatrixDim;kk++)
	{
          FilterMatrixTimeIndependent[kk] = 
	    &(FilterMatrixTimeIndependentStorageVec[kk*CovMatrixDim]);
	} 


      FilterMatrixStorageVec = 
         new double[2*CovMatrixDim*2*CovMatrixDim];

      FilterMatrix = new double * [2*CovMatrixDim];

      for(int kk=0;kk<2*CovMatrixDim;kk++)
	{
          FilterMatrix[kk] = &(FilterMatrixStorageVec[kk*2*CovMatrixDim]);
	} 

      HMatrixStorageVec = 
         new double[2*CovMatrixDim*2*CovMatrixDim];

      HMatrix = new double * [2*CovMatrixDim];

      for(int kk=0;kk<2*CovMatrixDim;kk++)
	{
          HMatrix[kk] = &(HMatrixStorageVec[kk*2*CovMatrixDim]);
	} 


      VMatrixStorageVec = 
         new double[2*CovMatrixDim*2*CovMatrixDim];

      VMatrix = new double * [2*CovMatrixDim];

      for(int kk=0;kk<2*CovMatrixDim;kk++)
	{
          VMatrix[kk] = &(VMatrixStorageVec[kk*2*CovMatrixDim]);
	} 

      WorkMatrixStorageVec = 
         new double[2*CovMatrixDim*2*CovMatrixDim];

      WorkMatrix = new double * [2*CovMatrixDim];

      for(int kk=0;kk<2*CovMatrixDim;kk++)
	{
          WorkMatrix[kk] = &(WorkMatrixStorageVec[kk*2*CovMatrixDim]);
	} 


      TestEigMatrixStorageVec = 
         new double[2*CovMatrixDim*2*CovMatrixDim];

      TestEigMatrix = new double * [2*CovMatrixDim];

      for(int kk=0;kk<2*CovMatrixDim;kk++)
	{
          TestEigMatrix[kk] = &(TestEigMatrixStorageVec[kk*2*CovMatrixDim]);
	} 

      LambdaVec = new double[2*CovMatrixDim];
      LambdaQSqVec = new double[2*CovMatrixDim];
      WorkVec = new double[2*CovMatrixDim];

      Parameters =  new ChiSquareReceiverParameters;

      Parameters->CovMatrixDim = CovMatrixDim;
      Parameters->LambdaVec =  LambdaVec;
      Parameters->LambdaQSqVec =  LambdaQSqVec;
      Parameters->ActiveCurrentValue = 0;

      InitializeGSL_FunctionToDerivativePsi();

} // ## end Constructor

// ################################################################

ChiSquareReceiver::~ChiSquareReceiver(void)
{

  // Destructor should deallocate memory in opposite order to way
  // it was allocated by constructor

GSL_Function->function = 0;
GSL_Function->params = 0;

GSL_Function->~gsl_function();

Parameters->LambdaVec = 0;
Parameters->LambdaQSqVec = 0;

delete Parameters;

delete [] WorkVec;
delete [] LambdaQSqVec;
delete [] LambdaVec;


for(int kk=0;kk<2*CovMatrixDim;kk++)
  WorkMatrix[kk]= 0;

delete [] WorkMatrix;
delete [] WorkMatrixStorageVec;

for(int kk=0;kk<2*CovMatrixDim;kk++)
  TestEigMatrix[kk]= 0;

delete [] TestEigMatrix;
delete [] TestEigMatrixStorageVec;

for(int kk=0;kk<2*CovMatrixDim;kk++)
  VMatrix[kk]= 0;

delete [] VMatrix;
delete [] VMatrixStorageVec;

for(int kk=0;kk<2*CovMatrixDim;kk++)
  HMatrix[kk]= 0;

delete [] HMatrix;
delete [] HMatrixStorageVec;

for(int kk=0;kk<2*CovMatrixDim;kk++)
  FilterMatrix[kk]= 0;

delete [] FilterMatrix;
delete [] FilterMatrixStorageVec;

for(int kk=0;kk<CovMatrixDim;kk++)
  FilterMatrixTimeIndependent[kk]= 0;

delete [] FilterMatrixTimeIndependent;
delete [] FilterMatrixTimeIndependentStorageVec;

ElecFilterForTransferFunction->~ElecFilter();
ElecFilterTransferFunction->~Photodetector();
TunableOptFilterForTransferFunction->~OptFilter();
OptFilterTransferFunction->~OptSignal();
ElecSignalStatForNoiseFreeSignal->~ElecSignalStat();
ElecFilterForNoiseFreeSignal->~ElecFilter();
PhotodetectorForNoiseFreeSignal->~Photodetector();
OptDemuxerForNoiseFreeSignal->~OptDemuxer();
TunableOptFilterForNoiseFreeSignal->~OptFilter();
DemuxedNoiseFreeSignal->~OptSignal();

RNG->~RanNumGen();

for(int kk=0;kk<2*CovMatrixDim;kk++)
  CovarianceEvectors[kk]= 0;

delete [] CovarianceEvectors;
delete [] CovarianceEvectorsStorageVec;
delete [] CovarianceEvalues;
delete [] SymmetricEigWork;


for(int kk=0;kk<2*CovMatrixDim;kk++)
  CovarianceMatrix[kk]= 0;

delete [] CovarianceMatrix;
delete [] ReducedNoiseFreeSignalRealVector;


for(int tt=0;tt<NumTimeSlices;tt++)
  PDFTimeSlices[tt]= 0;

delete [] PDFTimeSlices;
delete [] PDFTimeSlicesStorageVec;

 delete [] MomentsIntegrand;
 delete [] MeanCurrent;
 delete [] StandardDeviationCurrent;
  
 if(TypeOutput == EYE)
   {

    for(int tt=0;tt<NumSlicesPerBit;tt++)
     EyeMatrix[tt]= 0;

    delete [] EyeMatrix;
    delete [] EyeMatrixStorageVec;
   }

 if(TypeOutput == PDFS)
   {
     delete [] BER;
     delete [] PowerInSlice;
     delete [] BitString;
     delete [] AveragePDFZeros;
     delete [] AveragePDFOnes;
   }

delete [] CurrentVector;
delete [] TimeSlices;

Signal = 0;
oNoiseCovariance = 0;

} // ## end ~ChiSquareReceiver

// ################################################################

void ChiSquareReceiver::ReceiveSignal(string OutFileNamePrefix2)
{

  // For now code only works with channel that is at center freq
  // (in case have evenly spaced freqs)

  DemuxChannArrayIndex = Signal->GetChannelArrayIndex(0);

  OutFileNamePrefix = Job + OutFileNamePrefix2;

  if(abs(Signal->GetCenterFreq() - Signal->GetFrequency(DemuxChannArrayIndex))
     /Signal->GetCenterFreq() > 0.01 )
    {
      cout << "Signal->GetCenterFreq() = " 
	   << Signal->GetCenterFreq() << endl;
      cout << "Signal->GetFrequency(DemuxChannArrayIndex) = "
	   << Signal->GetFrequency(DemuxChannArrayIndex)
	   << endl;

      cerr << "ChiSquareReceiver::ReceiveSignal" << endl
	   << "For now we need center frequency of discrete signal to equal"
	   << endl
	   << "the frequency of the channel being demuxed. ABORTING!"

	   << endl;
      exit(1);
    }

  if(ReadInSignalAndCovMatrixFlag)
    oNoiseCovariance->ReadFileNoiseFreeSignalAndCovarianceMatrix(
					   SignalAndCovMatrixJob);

  ReceiveNoiseFreeSignalAndRecoverClock();
  PrepareToComputeKLModes();

  // With this defn CurrentVector is in normalized units
  // Reason: From equation (2) of JZ notes on ChiSquareReceiver
  // we need s*CurrentValue, s*LambdaVec, and Q to be the same 
  // after normalization as before. 
  // Since s is to be normalized but not Psi itself, I need to
  // normalize derivatives of Psi wrt s. In particular we need to
  // normalize DerivativePsi by NormalizationFactor and
  // SecondDerivativePsi by sq(NormalizationFactor).
  // Before writing to a file must un-normalize all quantities

   for(int vv=0;vv<NumPointsCurrent;vv++)
     CurrentVector[vv] = NormalizationFactor*(MinCurrent+vv*DeltaCurrent);

   // At this point all quantities we need are normalized

   if(DebugLevel ==3)
     {
     cout << "Un-normalized CurrentVector = " << endl;
     for(int vv=0;vv<NumPointsCurrent;vv++)
       cout << CurrentVector[vv]/NormalizationFactor << endl;
     cout << endl;
     }


  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      ComputeKLModes(tt);
      ComputePDFAtTimeValue(tt);
    }
  
  switch(TypeOutput)
    {
     case PDFS:
       WriteFilePDFs();
       break;
     case EYE:
       WriteFileEye();
       break;
    }

}

// ##########################################################

void ChiSquareReceiver::ComputeGaussianPDFsAndBER(void)
{

   double * GaussianPDFTimeSlicesStorageVec;
   double ** GaussianPDFTimeSlices;

   double * AverageGaussianPDFZeros;
   double * AverageGaussianPDFOnes;

   GaussianPDFTimeSlicesStorageVec = new double[NumTimeSlices*NumPointsCurrent];

   GaussianPDFTimeSlices = new double*[NumTimeSlices];

   for(int tt=0;tt<NumTimeSlices;tt++)
     GaussianPDFTimeSlices[tt] = &(GaussianPDFTimeSlicesStorageVec[tt*NumPointsCurrent]);

   AverageGaussianPDFZeros = new double[NumPointsCurrent];
   AverageGaussianPDFOnes = new double[NumPointsCurrent];



   for(int tt=0;tt<NumTimeSlices;tt++)
     {

      for(int vv=0;vv<NumPointsCurrent;vv++)
        GaussianPDFTimeSlices[tt][vv] = exp( -sq(CurrentVector[vv] - MeanCurrent[tt])/
                                     (2.0*sq(StandardDeviationCurrent[tt])))/
                                       (sqrt(2*pi*sq(StandardDeviationCurrent[tt])));

      double IntPDFTimeSlice =
       Quadrature(GaussianPDFTimeSlices[tt],NumPointsCurrent,
                CurrentVector[0],CurrentVector[NumPointsCurrent-1]);

      if(abs(IntPDFTimeSlice-1.0) > 0.05)
        {
          cerr << "ChiSquareReceiver::ComputeGaussuanPDFsAndBER"
               << endl
               << "Error exceeds 1% in integration of PDFTimeSlices." << endl
               << "Change Min or Max Current or increase NumPointsCurrent"
               << endl
               << "ABORTING" << endl;

         cout << "tt for Matlab,IntPDFTimeSlice = "
               << tt+2 << " " << IntPDFTimeSlice << endl;

          exit(1);
        }     

     } // end for tt loop


  NumBitsToPlot = NumTimeSlices;

  int NumZeros = 0;
  int NumOnes = 0;

  for(int vv=0;vv<NumPointsCurrent;vv++)
    AverageGaussianPDFZeros[vv] = 0.0;

  for(int vv=0;vv<NumPointsCurrent;vv++)
    AverageGaussianPDFOnes[vv] = 0.0;

  int NumBitsBetweenSlices = 1;

  for(int bb=0;bb<NumBitsToPlot;bb++)
    {
      int bit = bb*NumBitsBetweenSlices;

      BitString[bb] = DemuxedNoiseFreeSignal->oBitString->GetBit(0,bit);

    }

  double MaxCorrelation = 0.0;
  int OptimalShift = 0;

  for(int ss=0;ss<NumTimeSlices;ss++)
    {
      double Correlation = 0;

      for(int tt=0;tt<NumTimeSlices;tt++)
        {

          int BitIndex = ss+tt;
          while(BitIndex < 0)
            BitIndex += NumBitsToPlot;
          while(BitIndex >= NumBitsToPlot)
            BitIndex -= NumBitsToPlot;

          Correlation += MeanCurrent[tt]*BitString[BitIndex];
        }

      if(Correlation > MaxCorrelation)
        {
         OptimalShift = ss;
         MaxCorrelation = Correlation;
        }

    } // end for ss


  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      int BitIndex = OptimalShift+tt;

      while(BitIndex < 0)
        BitIndex += NumBitsToPlot;
      while(BitIndex >= NumBitsToPlot)
        BitIndex -= NumBitsToPlot;

      if(BitString[BitIndex] == 0)
        {
          NumZeros += 1;

          for(int vv=0;vv<NumPointsCurrent;vv++)
            AverageGaussianPDFZeros[vv] += GaussianPDFTimeSlices[tt][vv];
        }
      else
        {
          NumOnes += 1;

          for(int vv=0;vv<NumPointsCurrent;vv++)
            AverageGaussianPDFOnes[vv] += GaussianPDFTimeSlices[tt][vv];
        }

    } // end for tt


  if(NumZeros != 0)
   for(int vv=0;vv<NumPointsCurrent;vv++)
    AverageGaussianPDFZeros[vv] /= NumZeros;

  if(NumOnes != 0)
   for(int vv=0;vv<NumPointsCurrent;vv++)
    AverageGaussianPDFOnes[vv] /= NumOnes;

  double BitErrorRatio, DecisionThreshold;

  double TotalIntZeros = Quadrature(AverageGaussianPDFZeros,NumPointsCurrent,
                                    CurrentVector[0],
                                    CurrentVector[NumPointsCurrent-1]);

  double TotalIntOnes = Quadrature(AverageGaussianPDFOnes,NumPointsCurrent,
                                   CurrentVector[0],
                                   CurrentVector[NumPointsCurrent-1]);

  double Error = max( abs(TotalIntZeros-1), abs(TotalIntOnes-1) );
 
  if(Error > 0.05)
    {
      cerr << "ChiSquareReceiver::ComputeGaussianPDFsAndBER:"
           <<  endl
           << "Error in integration of AverageGaussianPDFOnes or  AverageGaussianPDFZeros "
           << endl
           << "exceeds 1%. ABORTING" << endl;

      cout << "TotalIntZeros = " << TotalIntZeros << endl;
      cout << "TotalIntOnes = " << TotalIntOnes << endl;

      exit(1);
    }

    // RightEndPt for OnesInt must equal LeftEndPt for Zeros Int

     for(int DecisionIndex=0;DecisionIndex<NumPointsCurrent;DecisionIndex++)
     {
      double ZerosInt = Quadrature(&AverageGaussianPDFZeros[DecisionIndex],
                                   NumPointsCurrent-DecisionIndex,
                                   CurrentVector[DecisionIndex],
                                   CurrentVector[NumPointsCurrent-1])/
                                   TotalIntZeros;

      double OnesInt = Quadrature(AverageGaussianPDFOnes,DecisionIndex +1,
                                  CurrentVector[0],
                                  CurrentVector[DecisionIndex])/
                                  TotalIntOnes;

      double BERTemp = 0.5*(ZerosInt+OnesInt);

      if(DecisionIndex == 0 || BERTemp < BitErrorRatio)
        {
         BitErrorRatio = BERTemp;
         DecisionThreshold = CurrentVector[DecisionIndex];
        }
     } // end for DecisionIndex loop

 string OutFileName = Job  + "BERFromGaussianPDFs.dat";

 ofstream OutFile(OutFileName.c_str(),ios::out);

 if (!OutFile)
   {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  OutFile << "% (1) BERFromGaussianPDFs (2)  DecisionThreshold [mV]" << endl
          <<  BitErrorRatio<< "\t" <<  1e3*DecisionThreshold/NormalizationFactor
          << endl;

  OutFile.close();

} // ## end ChiSquareReceiver::ComputeGaussianPDFsAndBER

// #############################################################

double ChiSquareReceiver::GetOSNR_dB(double B_OSA)
{
   double NoisePower = 
    Signal->GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics()*B_OSA;

   double SignalPower = Signal->GetAveragePower()/Signal->GetNumChannels();

   return Linear2dB(SignalPower/NoisePower);
}

// ###############################################################

void ChiSquareReceiver::SetOSNR(double Value)
{

  if(SetOSNRdB != 0)
    SetOSNRdB = Value;
  else
    {
      cerr << "ChiSquareReceiver::SetOSNR error:" << endl
	   << "Must have SetOSNRdB nonzero in ChiSquareReceiver.in" << endl
	   << "ABORT" 
	   << endl;
      exit(1);

    }

}

// #############################################################

void  ChiSquareReceiver::SetElecFilterBandwidth(double ValueHz)
{

  ElecFilterForTransferFunction->Set_f_3dB_ElecFilter(ValueHz);
  ElecFilterForNoiseFreeSignal->Set_f_3dB_ElecFilter(ValueHz);
}

// ###############################################################

void  ChiSquareReceiver::SetOptFilterBandwidth(double ValueHz)
{

  TunableOptFilterForNoiseFreeSignal->SetFreqFWHM_OptFilter(ValueHz);
  TunableOptFilterForTransferFunction->SetFreqFWHM_OptFilter(ValueHz);

  }
// ###############################################################

void ChiSquareReceiver::ReceiveNoiseFreeSignalAndRecoverClock(void)
{
 
  OptDemuxerForNoiseFreeSignal->DemuxChannel(DemuxChannArrayIndex);

  /***
  
  // added by Walter 08/03/04
  string FileName;

  //write to file demuxed signal
  
  FileName = OutFileNamePrefix + "tNoiseFreeDemuxedSignalComplexFields.dat";
  bool ForRestartFlag=0;
  DemuxedNoiseFreeSignal->WriteComplexFields(FileName,ForRestartFlag);
  
  FileName = OutFileNamePrefix + "fNoiseFreeDemuxedSignalComplexFields.dat";
  DemuxedNoiseFreeSignal->WriteComplexFieldsFreq(FileName);
  //end of added part
  
  ***/

  PhotodetectorForNoiseFreeSignal->DetectOptSignal();
  ElecFilterForNoiseFreeSignal->FilterElecSignal();

  ElecSignalStatForNoiseFreeSignal->ComputePerformanceStatistics();

  if(!ReadInClockRecoveryTimeFlag)
    {
      
      // The ReducedNoiseFreeSignal and the CovarianceMatrix 
      // do not get time shifted. So we need the clock recovery time
      // before the time shift is done in the call
      // ElecSignalStatForNoiseFreeSignal->ComputePerformanceStatistics();

      ClockRecoveryTime = 
       ElecSignalStatForNoiseFreeSignal->GetClockRecoveryTimeBeforeTimeShift();

    }

  // We need this for TypeOutput = PDFS to work out which time slices
  // are zeros and which are ones. Remember that we are always working
  // with the ReducedNoiseFreeSignal which hasn't been time shifted.
  // However, ElecSignalStatForNoiseFreeSignal->ComputePerformanceStatistics()
  // shifts PhotodetectorForNoiseFreeSignal, so we need to undo this shift.

  PhotodetectorForNoiseFreeSignal->TimeShiftSignal(
   ElecSignalStatForNoiseFreeSignal->GetClockRecoveryTimeBeforeTimeShift() );

  ClockRecoveryCurrent = ElecSignalStatForNoiseFreeSignal->GetClockCurrent();

  PhotodetectorForNoiseFreeSignal->WriteFileTimePower(
     OutFileNamePrefix  + "tElecFiltNoiseFreeSignal.dat");
  
  PhotodetectorForNoiseFreeSignal->WriteFileFreqPower(
     OutFileNamePrefix  + "fElecFiltNoiseFreeSignal.dat");

  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      TimeSlices[tt] = ClockRecoveryTime + tt*TimeBetweenSlices;

      if(TypeOutput==EYE)
	TimeSlices[tt] -= Signal->GetBitLength()/2.0;
         	
      // We want TimeSlices[tt] to be in range
      // 0 <= TimeSlices[tt] < TimeWindow

      while(TimeSlices[tt] < 0)
        TimeSlices[tt] += TimeWindow;

      while(TimeSlices[tt] >= TimeWindow)
        TimeSlices[tt] -= TimeWindow;

    } // end for tt

   
  if(DebugLevel >= 1)
    {

     cout << "TimeSlices = " << endl;
      for(int tt=0;tt<NumTimeSlices;tt++)
       cout << TimeSlices[tt] << endl;
     cout << endl;
   
     //cout << "ClockRecoveryTime = " << ClockRecoveryTime << endl;
     cout << "ClockRecoveryCurrent = " << ClockRecoveryCurrent << endl;
    
    }

} //## end ReceiveNoiseFreeSignalAndRecoverClock

// ##############################################################

void ChiSquareReceiver::PrepareToComputeKLModes(void)
{
  GetReducedNoiseFreeSigAndDiagonalizeCovMatrix();
  ComputeTimeIndependentPartOfFilterMatrix();
}

// #############################################################

void ChiSquareReceiver::GetReducedNoiseFreeSigAndDiagonalizeCovMatrix(void)
{

 oNoiseCovariance->GetReducedNoiseFreeSignal(ReceiverReducedNoiseFreeSignal);

 for(int jj=0;jj<CovMatrixDim;jj++)
   ReducedNoiseFreeSignalRealVector[jj] = 
    real(ReceiverReducedNoiseFreeSignal.fOptSignalX[jj]);


 for(int jj=0;jj<CovMatrixDim;jj++)
   ReducedNoiseFreeSignalRealVector[jj+CovMatrixDim] = 
    imag(ReceiverReducedNoiseFreeSignal.fOptSignalX[jj]);


 oNoiseCovariance->GetCovarianceMatrix(CovarianceMatrix,
 				       ReadInSignalAndCovMatrixFlag,
				       SetOSNRdB,B_OSA);

 SymmetrizeMatrix(CovarianceMatrix,2*CovMatrixDim);

 double MeanNoiseVariance = 0.0;

 for(int jj=0;jj<2*CovMatrixDim;jj++)
   MeanNoiseVariance += sq(CovarianceMatrix[jj][jj]/sqrt(2*CovMatrixDim));

 MeanNoiseVariance = sqrt(MeanNoiseVariance); // Units = Watts

 // For AWGN noise NormalizationFactor is independent of CovMatrixDim

 NormalizationFactor = 1.0/MeanNoiseVariance;

 cout << "NormalizationFactor = " << NormalizationFactor <<endl;

 // CovarianceMatrix is in Watts.
 // We normalize and make it unitless:

 DenseMatrixScale(CovarianceMatrix,2*CovMatrixDim,NormalizationFactor);

 // ReceiverReducedNoiseFreeSignal and ReducedNoiseFreeSignalRealVector 
 // are in Watts.
 // We normalize and make them unitless:

 for(int jj=0;jj<CovMatrixDim;jj++)
  ReceiverReducedNoiseFreeSignal.fOptSignalX[jj] *= sqrt(NormalizationFactor);

 for(int jj=0;jj<2*CovMatrixDim;jj++)
  ReducedNoiseFreeSignalRealVector[jj] *= sqrt(NormalizationFactor);

 SymmetricEig(CovarianceMatrix,CovarianceEvalues,CovarianceEvectors,
	      2*CovMatrixDim);

 if(!TestEigDecomposition(CovarianceMatrix,CovarianceEvalues,
                          CovarianceEvectors,2*CovMatrixDim))
   {
     cout << "WARNING re ChiSquareReceiver: " << endl
	  << "Diagonalization of Covariance Matrix not sufficiently accurate"
	  << endl;
   }

 if(DebugLevel >= 1)
 {

 string OutFileName;

 OutFileName = OutFileNamePrefix + "CovarianceEvectors.dat";

 WriteFileSquareMatrix(CovarianceEvectors,2*CovMatrixDim,OutFileName,1);

 OutFileName = OutFileNamePrefix + "CovarianceEvalues.dat";

 WriteFileVector(CovarianceEvalues,2*CovMatrixDim,OutFileName,
		 NormalizationFactor);

 }

 // Form the Matrix CovarianceEvectors*Diag(sqrt(CovarianceEvalues))
 // and assign it to CovarianceEvectors

 for(int jj=0;jj<2*CovMatrixDim;jj++)
   {
     if(CovarianceEvalues[jj] < 0)
       {
         cout << "CovarianceEvalue " << jj << " is negative, value = "
	      << CovarianceEvalues[jj] << endl
	      << "Setting value to 0" << endl;

         CovarianceEvalues[jj] = 0.0;
       }
     }

 for(int ii=0;ii<2*CovMatrixDim;ii++)
  for(int jj=0;jj<2*CovMatrixDim;jj++)
    {
     CovarianceEvectors[ii][jj] = sqrt(CovarianceEvalues[ii])*
                                  CovarianceEvectors[ii][jj];
    }

}

// ############################################################

void ChiSquareReceiver::ComputeTimeIndependentPartOfFilterMatrix(void)
{
   OptFilterTransferFunction->GenerateImpulseSignal();
   ElecFilterTransferFunction->GenerateImpulseCurrent();

   ElecFilterForTransferFunction->FilterElecSignal();
   TunableOptFilterForTransferFunction->FilterOptSignal();

  for(int ii=CovMatrixFirstIndex;ii<=CovMatrixLastIndex;ii++)
   for(int jj=CovMatrixFirstIndex;jj<=CovMatrixLastIndex;jj++)
     {

       /**
        Since N = Signal->GetQtPoints() >= CovMatrixDim
        we have -N/2 <= ii,jj <= N/2 - 1 and so
        -N+1 <= jj-ii <= N-1
        
        Now ii,jj index physical frequencies 
        Omega_ii = DeltaFreq*ii, Omega_jj = DeltaFreq*jj.

        If abs(jj-ii)>= N/2 then abs(Omega_jj-Omega_ii) is large
        and we would want to have the electrical filter zero out
        this frequency entirely. So we don't want to wrap jj-ii 
 
       fft->index : [-N/2,N/2-1] -> [0,N-1]

       **/

       int DiffIndex = jj-ii;
 
       cplx ElecFilterValue;

       if(abs(DiffIndex) >= Signal->GetQtPoints()/2.0)
	 ElecFilterValue = 0.0;
       else
         ElecFilterValue = ElecFilterTransferFunction->sfftPM.fPowerSignal
	   [ElecFilterTransferFunction->fft->index(DiffIndex)];

       FilterMatrixTimeIndependent[ii-CovMatrixFirstIndex]
	                          [jj-CovMatrixFirstIndex] = 
        PhotodetectorForNoiseFreeSignal->GetConvFactOptElecPhotod()*
        conj(OptFilterTransferFunction->sfftM.fOptSignalX[
		OptFilterTransferFunction->fft->index(ii)])*
        OptFilterTransferFunction->sfftM.fOptSignalX[
		OptFilterTransferFunction->fft->index(jj)]*
	ElecFilterValue;


     }

 
  if(DebugLevel >= 2)
   WriteFileComplexMatrix(FilterMatrixTimeIndependent,CovMatrixDim, 
		          Job + "FilterMatrixTimeIndependentReal.dat",
		          Job + "FilterMatrixTimeIndependentImag.dat",1);
 
   string OutFileName;
   
 if(DebugLevel >= 2)
   { 
    OutFileName = Job + "ReducedOptFilterTransferFunction.dat";
    ofstream OutFile;
    OutFile.open(OutFileName.c_str(),ios::out);

   
    for(int ii=CovMatrixFirstIndex;ii<=CovMatrixLastIndex;ii++)
     OutFile << sq(abs(OptFilterTransferFunction->sfftM.fOptSignalX
		       [OptFilterTransferFunction->fft->index(ii)])) << endl;
    
    OutFile.close();

    OutFileName = Job + "ReducedElecFilterTransferFunction.dat";
    OutFile.open(OutFileName.c_str(),ios::out);


    for(int ii=CovMatrixFirstIndex;ii<=CovMatrixLastIndex;ii++)
     OutFile << sq(abs(ElecFilterTransferFunction->sfftPM.fPowerSignal
		       [ElecFilterTransferFunction->fft->index(ii)])) << endl;

    OutFile.close();

    OptFilterTransferFunction->WriteFileFreqPower(
	  Job+"fOptFilterTransferFunction.dat",0);
    OptFilterTransferFunction->WriteFileTimePower(
            Job+"tOptFilterTransferFunction.dat");


   // Note that  Photodetector::WriteFileFreqPower normalizes by DeltaFreq

    ElecFilterTransferFunction->WriteFileFreqPower(
           Job+"fElecFilterTransferFunction.dat",0);
    ElecFilterTransferFunction->WriteFileTimePower(
            Job+"tElecFilterTransferFunction.dat");

   }

}

// ############################################################

void ChiSquareReceiver::ComputeKLModes(int TimeIndex)
{

  double Time = TimeSlices[TimeIndex];

  // Compute FilterMatrix

  for(int kk=0;kk<CovMatrixDim;kk++)
   for(int ll=0;ll<CovMatrixDim;ll++)
     {
       // exp_jc_real(x) = exp(jc*x) is defined in ocsTools.hh
       // We have an additional - sign compared to (4.6.3) of
       // Ron Holzloehner's thesis because OCS uses the Physics
       // convention for ffts and Ron's thesis uses the Math convention

       cplx Value = exp_jc_real(-2*pi*Time*(ll-kk)/TimeWindow)*
	            FilterMatrixTimeIndependent[kk][ll];

       // The next few lines should be sped up using
       // pointers more cleverly!!

       FilterMatrix[kk][ll] = real(Value);
       FilterMatrix[kk][ll+CovMatrixDim] = -imag(Value);

       FilterMatrix[kk+CovMatrixDim][ll] = imag(Value);
       FilterMatrix[kk+CovMatrixDim][ll+CovMatrixDim] = real(Value);

     }


  if(DebugLevel == 2)
   WriteFileSquareMatrix(FilterMatrix,2*CovMatrixDim, 
		   Job + "FilterMatrix.dat",1);

  // Force FilterMatrix to be symmetric!!

  SymmetrizeMatrix(FilterMatrix,2*CovMatrixDim);

  // Compute 
  // HMatrix = CovarianceEvectors^T*FilterMatrix*CovarianceEvectors
  // where now CovarianceEvectors = UD, where U,D are defined similarly
  // to Ron Holzloehner's Appendix A.3

  DenseMatrixSimilarityProduct(CovarianceEvectors,FilterMatrix,HMatrix,
			       WorkMatrix,1,1,2*CovMatrixDim);

  // Diagonalize H = V Diag(Lambda) V^T

 SymmetricEig(HMatrix,LambdaVec,VMatrix,2*CovMatrixDim);

  if(!TestEigDecomposition(HMatrix,LambdaVec,VMatrix,2*CovMatrixDim))
   {
     cout << "WARNING re ChiSquareReceiver: " << endl
	  << "Diagonalization of Matrix H not sufficiently accurate"
	  << endl;
   }


 // Use HMatrix memory to store the matrix E := C^{-1} in Ron's thesis.
 // We have E = U*D*V
 // At this point  CovarianceEvectors stores U*D, so we want
 // HMatrix =  CovarianceEvectors*VMatrix.

 DenseMatrixProduct(CovarianceEvectors, VMatrix, HMatrix,0,0,2*CovMatrixDim);

 // Compute LambdaQ = E^T*W*ReducedNoiseFreeSignalRealVector
 // and store answer in LambdaQSqVec

 DenseMatrixVectorProduct(WorkVec,FilterMatrix,
			  ReducedNoiseFreeSignalRealVector,2*CovMatrixDim);

 TransposeMatrix(HMatrix,2*CovMatrixDim);

 DenseMatrixVectorProduct(LambdaQSqVec,HMatrix,WorkVec,2*CovMatrixDim);

 if(DebugLevel >= 2)
   {
    char  C_String[10];
    sprintf(C_String,"%i.dat",TimeIndex);
    string TimeSuffix = C_String;

    WriteFileVector(LambdaQSqVec,2*CovMatrixDim,Job+"LambdaQVec"+TimeSuffix,1);
   }


 for(int kk=0;kk<2*CovMatrixDim;kk++)
   { 
     if(LambdaVec[kk] == 0)
       LambdaQSqVec[kk] = 0.0;
     else
       LambdaQSqVec[kk] = sq(LambdaQSqVec[kk])/LambdaVec[kk];
   }


 if(DebugLevel >= 1)
   {
      double Tmp = NormalizationFactor;
      NormalizationFactor = 1;

      char  C_String[10];

      sprintf(C_String,"%i.dat",TimeIndex);

      string TimeSuffix = C_String;

     WriteFileVector(LambdaVec,2*CovMatrixDim,
		     Job+"LambdaVec"+TimeSuffix,
	             NormalizationFactor);
     WriteFileVector(LambdaQSqVec,2*CovMatrixDim,
		     Job+"LambdaQSqVec"+TimeSuffix,
	             NormalizationFactor);

     NormalizationFactor = Tmp;
   }

 // Test that sum(LambdaQSqVec) = R^T FilterMatrix R where
 // R = ReducedNoiseFreeSignalRealVector

 double NoiselessCurrent = 0.0;

 for(int kk=0;kk<2*CovMatrixDim;kk++)
   NoiselessCurrent += LambdaQSqVec[kk];

 DenseMatrixVectorProduct(WorkVec,FilterMatrix,
			  ReducedNoiseFreeSignalRealVector,2*CovMatrixDim);

 double QuadForm = InnerProduct(WorkVec,ReducedNoiseFreeSignalRealVector,
			        2*CovMatrixDim);

 if( NoiselessCurrent != 0 )
   {
     double RelError = abs(NoiselessCurrent - QuadForm)/NoiselessCurrent;
     
     if(RelError > 1e-6)
       cout << "WARNING ChiSquareReceiver::ComputeKLModes " << endl
	    << "RelError(sum(LambdaQSqVec)- R^T FilterMatrix R) = " 
	    << RelError << endl;

   }
 else
   {
      if(abs(NoiselessCurrent - QuadForm) > 1e-6)
       cout << "WARNING ChiSquareReceiver::ComputeKLModes " << endl
	    << "Error(sum(LambdaQSqVec)- R^T FilterMatrix R) = " 
	    << abs(NoiselessCurrent - QuadForm) << endl;
   } 


} // ## end ComputeKLModes

// ############################################################

void ChiSquareReceiver::ComputePDFAtTimeValue(int TimeIndex)
{
  /**
  cout << endl
       << "ChiSquareReceiver::ComputePDFAtTimeValue: TimeIndex = "
       << TimeIndex << endl;
  **/

   // At this point all quantities we need are normalized. 

   ComputePDFAtTimeValueFourierTransform(TimeIndex);

   double MaxPosLambda = max( MaxOfArray(LambdaVec,2*CovMatrixDim), 0.0);
   MinNegLambda = min( MinOfArray(LambdaVec,2*CovMatrixDim), 0.0);

   if((MinNegLambda == 0 && CurrentVector[0] < 0.0) || MaxPosLambda == 0)
     {
      cout << "For these receiver filters need current values positive" << endl;
      exit(1);
     }

   if(MinNegLambda == 0)
    {
      LowerPole = 0; // Provisional Value to be changed later
    }
   else
    {
      LowerPole = 0.5/MinNegLambda;   // LowerPole < 0
    }

    UpperPole = 0.5/MaxPosLambda;   // UpperPole > 0


   if(DebugLevel==3)
     cout << "Normalized [MinNegLambda,MaxPosLambda] = [" 
	  << MinNegLambda << "," << MaxPosLambda << "]" << endl;
 
   for(int vv=0;vv<NumPointsCurrent;vv++)
    {  

      TypeIntegrationMethod = STEEPEST_DESCENT; // Use this method if we can

      PDFTimeSlices[TimeIndex][vv] = 
	ComputePDFValueSteepestDescent(TimeIndex,vv);
					 
      if(TypeIntegrationMethod == FOURIER_TRANSFORM)
	PDFTimeSlices[TimeIndex][vv] = 0.0; // FIX!

    }

   // Check pdf integrates to 1.0

   double IntPDFTimeSlice = 
     Quadrature(PDFTimeSlices[TimeIndex],NumPointsCurrent,
		CurrentVector[0],CurrentVector[NumPointsCurrent-1]);

  if(abs(IntPDFTimeSlice-1.0) > 0.05)
    {
      cerr << "ChiSquareReceiver::ComputePDFAtTimeValue" 
	   << endl
	   << "Error exceeds 1% in integration of PDFTimeSlices." << endl
	   << " ABORTING" << endl;

      cout << "tt for Matlab,IntPDFTimeSlice = " 
	   << TimeIndex +2 << " " << IntPDFTimeSlice << endl;
      
      //exit(1);
    }


} // ## end ComputePDFAtTimeValue

// ############################################################

double ChiSquareReceiver::ComputePDFValueSteepestDescent(int TimeIndex, 
	                				 int CurrentIndex)
{

  double PDFValue  = 0.0;

  Parameters->ActiveCurrentValue = CurrentVector[CurrentIndex];

  double LowerBoundValue;
  double UpperBoundValue;
  double LowerBound; 
  double UpperBound; 
  double Epsilon = 1e-3;
  int LowerCount =0;

  if(MinNegLambda == 0)
   {
     LowerPole = -UpperPole;
     double LowerPoleValue = GSL_Function->function(LowerPole,
                                              GSL_Function->params);

     while(LowerCount < 50 && LowerPoleValue > 0)
      {
        LowerCount++;
        LowerPole *= 2.0;
        
        LowerPoleValue = GSL_Function->function(LowerPole,
                                              GSL_Function->params);
       }

     if(DebugLevel >= 2)  
      cout << "LowerPole, LowerPoleValue = " << LowerPole << " " << LowerPoleValue << endl;

    }

  // My theory says that UpperBoundValue > 0 is true.
  // We want LowerBoundValue <= 0 and UpperBoundValue >= 0

  LowerCount = 0;

  do{
     LowerCount++;
     Epsilon *=0.5;
     LowerBound = (1-Epsilon)*LowerPole;  // LowerBound > LowerPole
           
     LowerBoundValue = GSL_Function->function(LowerBound,
					      GSL_Function->params);

     }while(LowerCount < 20 && LowerBoundValue > 0);
     

  Epsilon = 1e-3;
  int UpperCount = 0;

  do{
      UpperCount++;
      Epsilon *=0.5;
      UpperBound = (1-Epsilon)*UpperPole; // UpperBound < UpperPole

      UpperBoundValue = GSL_Function->function(UpperBound,
					       GSL_Function->params);

     }while(UpperCount < 20 && UpperBoundValue < 0);
     
  /**
       cout << "vv,LowerCount,UpperCount,LowerBoundValue,UpperBoundValue = " 
	    << CurrentIndex << " " 
	    << LowerCount << " " << UpperCount << " " 
	    << LowerBoundValue << " " << UpperBoundValue << endl;
  **/

  if( LowerBoundValue > 0 || UpperBoundValue < 0)
   {
     TypeIntegrationMethod = FOURIER_TRANSFORM;

     cout << "ChiSquareReceiver::ComputePDFValueSteepestDescent: " 
          << "NO ZERO CROSSING" << endl;

	 /***
         WriteFileGSLFunction(CurrentVector[CurrentIndex],
			      LowerBound,UpperBound,100,
			      OutFileNamePrefix + "DerivativePsi.dat");
	 ***/
        
   }
 else
   {
     double SaddlePoint = FindRootInterval(LowerBound,UpperBound);

      if(TypeIntegrationMethod == STEEPEST_DESCENT)
         PDFValue = EvaluateGaussianApproximation(SaddlePoint);
   }

  return  PDFValue;

} // ## end ComputePDFValueSteepestDescent

// #################################################################

double ChiSquareReceiver::EvaluateGaussianApproximation(double SaddlePoint)
{

  double PsiSaddlePoint = ChiSquareReceiverPsi(SaddlePoint,(void*) Parameters);

  double SecondDerivativePsiSaddlePoint = 
	  ChiSquareReceiverSecondDerivativePsi(SaddlePoint,(void*) Parameters)/
          sq(NormalizationFactor);

   /***
       cout << "SaddlePoint = " << SaddlePoint 
	    << " PsiSaddlePoint = " << PsiSaddlePoint
	    << " SDPsiSaddlePoint = " << SecondDerivativePsiSaddlePoint  
	    << endl;
   ****/

       // With this normalization integral of PDF with respect to 
       // the Normalized Current in CurrentVector is 1.

       
  return  exp(PsiSaddlePoint)/
	   (NormalizationFactor*sqrt(2*pi*SecondDerivativePsiSaddlePoint));
 
} // ## end EvaluateGaussianApproximation

// ##############################################################
 
//! Not implemented yet as it seems we don't need it

void ChiSquareReceiver::ComputePDFAtTimeValueFourierTransform(int TimeIndex)
{


} // ## end ComputePDFAtTimeValueFourierTransform

// ##########################################################

void ChiSquareReceiver::WriteFilePDFs(void)
{

  WriteFilePDFTimeSlices(OutFileNamePrefix  + "PDFTimeSlices.dat");

  ComputeAveragePDFs();


  // Checked: These two output pdfs integrate to 1, when 
  // the current is measured in mV

  WriteFilePairVectors(CurrentVector,AveragePDFZeros,NumPointsCurrent,
                       OutFileNamePrefix+"AveragePDFZeros.dat", 
		       1e3/NormalizationFactor, 1e3/NormalizationFactor);

  WriteFilePairVectors(CurrentVector,AveragePDFOnes,NumPointsCurrent,
		       OutFileNamePrefix+"AveragePDFOnes.dat", 
		       1e3/NormalizationFactor, 1e3/NormalizationFactor);

  ComputeMinimumBERAndDecisionThreshold();
  ComputeMomentsInEachBit();

} // ## end WriteFilePDFs

// ############################################################

void ChiSquareReceiver::WriteFileEye(void)
{

  WriteFilePDFTimeSlices(OutFileNamePrefix  + "PDFTimeSlices.dat");

  ComputeEye();

  string OutFileName = OutFileNamePrefix + "Eye.dat";

  ofstream OutFile(OutFileName.c_str(),ios::out);

  if (!OutFile) 
    {
      cerr << "Error: " << OutFileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
    }


  //for(int vv=NumPointsCurrent-1;vv>=0;vv--)
   for(int vv=0;vv<NumPointsCurrent;vv++)
    {
      for(int tt=0;tt<NumSlicesPerBit;tt++)
	OutFile << EyeMatrix[tt][vv] << " ";

      OutFile << endl;
    }

  OutFile.close();

} // ## end WriteFileEye

// ##########################################################

void ChiSquareReceiver::ComputeAveragePDFs(void)
{

  // I checked that the correlation is computed correctly
  // by artificially shifting the electrically filtered
  // current as a function of time. 

  if(NumBitsToPlot != NumTimeSlices)
   {
    cerr << "Error in ChiSquareReceiver::ComputeAveragePDFs: "
	 << "NumBitsToPlot != NumTimeSlices. ABORTING" 
	 << endl;
    exit(1);
   }

  int NumZeros = 0;
  int NumOnes = 0;

  for(int vv=0;vv<NumPointsCurrent;vv++)
    AveragePDFZeros[vv] = 0.0;

  for(int vv=0;vv<NumPointsCurrent;vv++)
    AveragePDFOnes[vv] = 0.0;

  for(int bb=0;bb<NumBitsToPlot;bb++)
    {
      int bit = bb*NumBitsBetweenSlices;
     
      BitString[bb] = DemuxedNoiseFreeSignal->oBitString->GetBit(0,bit);
       
    }

  for(int tt=0;tt<NumTimeSlices;tt++)
    {

      int Index = int(TimeSlices[tt]/DemuxedNoiseFreeSignal->GetDeltaTime());

      while(Index < 0)
        Index += PhotodetectorForNoiseFreeSignal->GetQtPoints();
      while(Index >= PhotodetectorForNoiseFreeSignal->GetQtPoints())
        Index -= PhotodetectorForNoiseFreeSignal->GetQtPoints();

      PowerInSlice[tt] = real(PhotodetectorForNoiseFreeSignal->
                         sfftPM.tPowerSignal[Index]);

    }

  double MaxCorrelation = 0.0;
  int OptimalShift = 0;

  for(int ss=0;ss<NumTimeSlices;ss++)
    {
      double Correlation = 0;

      for(int tt=0;tt<NumTimeSlices;tt++)
	{ 
 
          int BitIndex = ss+tt;

          while(BitIndex < 0)
	    BitIndex += NumBitsToPlot;
          while(BitIndex >= NumBitsToPlot)
	    BitIndex -= NumBitsToPlot;

          Correlation += PowerInSlice[tt]*BitString[BitIndex];
	}

      if(Correlation > MaxCorrelation)
	{
         OptimalShift = ss;
         MaxCorrelation = Correlation;
	}

    } // end for ss


  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      int BitIndex = OptimalShift+tt;

      while(BitIndex < 0)
	BitIndex += NumBitsToPlot;
      while(BitIndex >= NumBitsToPlot)
	BitIndex -= NumBitsToPlot;

      if(BitString[BitIndex] == 0)
	{
          NumZeros += 1;

          for(int vv=0;vv<NumPointsCurrent;vv++)
	    AveragePDFZeros[vv] += PDFTimeSlices[tt][vv];
	}
      else
	{
          NumOnes += 1;

          for(int vv=0;vv<NumPointsCurrent;vv++)
	    AveragePDFOnes[vv] += PDFTimeSlices[tt][vv];
	}

    } // end for tt


  if(NumZeros != 0)
   for(int vv=0;vv<NumPointsCurrent;vv++)
    AveragePDFZeros[vv] /= NumZeros;

  if(NumOnes != 0)
   for(int vv=0;vv<NumPointsCurrent;vv++)
    AveragePDFOnes[vv] /= NumOnes;

} // ## end ComputeAveragePDFs

// ############################################################

void ChiSquareReceiver::ComputeMinimumBERAndDecisionThreshold(void)
{

  double TotalIntZeros = Quadrature(AveragePDFZeros,NumPointsCurrent,
				    CurrentVector[0],
				    CurrentVector[NumPointsCurrent-1]);

  double TotalIntOnes = Quadrature(AveragePDFOnes,NumPointsCurrent,
				   CurrentVector[0],
				   CurrentVector[NumPointsCurrent-1]);

  double Error = max( abs(TotalIntZeros-1), abs(TotalIntOnes-1) );
  
  if(Error > 0.05)
    {
      cerr << "ChiSquareReceiver::ComputeMinimumBERAndDecisionThreshold:"
	   <<  endl
	   << "Error in integration of AveragePDFOnes or  AveragePDFZeros "
	   << endl
	   << "exceeds 1%. ABORTING" << endl;

      cout << "TotalIntZeros = " << TotalIntZeros << endl;
      cout << "TotalIntOnes = " << TotalIntOnes << endl;
    
      //exit(1);
    }

    // RightEndPt for OnesInt must equal LeftEndPt for Zeros Int

     for(int DecisionIndex=0;DecisionIndex<NumPointsCurrent;DecisionIndex++)
     {
      double ZerosInt = Quadrature(&AveragePDFZeros[DecisionIndex],
				   NumPointsCurrent-DecisionIndex,
				   CurrentVector[DecisionIndex],
				   CurrentVector[NumPointsCurrent-1])/
	                           TotalIntZeros;

      double OnesInt = Quadrature(AveragePDFOnes,DecisionIndex+1,
				  CurrentVector[0],
				  CurrentVector[DecisionIndex])/
	                          TotalIntOnes;

      BER[DecisionIndex] = 0.5*(ZerosInt+OnesInt);

      
      if(DecisionIndex == 0 || BER[DecisionIndex] < BitErrorRatio)
	{
	 BitErrorRatio = BER[DecisionIndex];
         DecisionThreshold = CurrentVector[DecisionIndex];
	}
     } // end for DecisionIndex loop


     if(DebugLevel >= 1)
       {     
        WriteFilePairVectors(CurrentVector,BER,NumPointsCurrent,
		          OutFileNamePrefix+"BER.dat", 
		          1e3/NormalizationFactor, 1.0);

     
        cout <<"BER = " << BitErrorRatio << endl;
        cout <<"DecisionThreshold [mV] = " 
	     << 1e3*DecisionThreshold/NormalizationFactor 
             << endl;
       }
     

} // ## end ChiSquareReceiver::ComputeMinimumBERAndDecisionThreshold

// ############################################################

void  ChiSquareReceiver::ComputeMomentsInEachBit(void)
{

  for(int tt=0;tt<NumTimeSlices;tt++)
    {

      // JZ: I found that its important for PDF to be normalized to
      // compute mean and standard deviation correctly. Therefore
     // added normalization of MomentsIntegrand by Integral in two
     // places below. Mar 19th 2007

      double Integral = Quadrature(PDFTimeSlices[tt],NumPointsCurrent,
                                   CurrentVector[0],
                                   CurrentVector[NumPointsCurrent-1]);

      for(int vv=0;vv<NumPointsCurrent;vv++)
	{
          MomentsIntegrand[vv] = (CurrentVector[vv]*PDFTimeSlices[tt][vv])/Integral;
	}


       MeanCurrent[tt] = Quadrature(MomentsIntegrand,NumPointsCurrent,
				    CurrentVector[0],
				    CurrentVector[NumPointsCurrent-1]);


      for(int vv=0;vv<NumPointsCurrent;vv++)
	{
          MomentsIntegrand[vv] = (sq(CurrentVector[vv])*PDFTimeSlices[tt][vv])/Integral;
	}

      // Second moment

      StandardDeviationCurrent[tt] = 
                         Quadrature(MomentsIntegrand,NumPointsCurrent,
				    CurrentVector[0],
				    CurrentVector[NumPointsCurrent-1]);

      
      if(sq(MeanCurrent[tt]) > StandardDeviationCurrent[tt])
	StandardDeviationCurrent[tt] = 0.0;
      else
        StandardDeviationCurrent[tt] = sqrt(StandardDeviationCurrent[tt] - 
                                            sq(MeanCurrent[tt]));
    } // end for tt 


  WriteFilePairVectors(TimeSlices,MeanCurrent,NumTimeSlices,
		       OutFileNamePrefix+"ChiSqMeanCurrent.dat", 
		       1e12,NormalizationFactor);

  WriteFilePairVectors(TimeSlices,StandardDeviationCurrent,NumTimeSlices,
		       OutFileNamePrefix+"ChiSqStandardDeviationCurrent.dat", 
		       1e12,NormalizationFactor);

}

// ############################################################

void ChiSquareReceiver::ComputeEye(void)
{

  for(int tt=0;tt<NumSlicesPerBit;tt++)
    for(int vv=0;vv<NumPointsCurrent;vv++)
      EyeMatrix[tt][vv] = 0.0;

  for(int tt=0;tt<NumSlicesPerBit;tt++)
    {
      for(int bb=0;bb<NumBitsToPlot;bb++)
	{
          int ss= tt + bb*NumSlicesPerBit;

          while(ss < 0)
	    ss+=NumTimeSlices;
          while(ss>=NumTimeSlices)
	    ss-=NumTimeSlices;

	  // We use this normalization factor so 
	  // each time slice integrates to 1.0

          for(int vv=0;vv<NumPointsCurrent;vv++)
	    EyeMatrix[tt][vv] += NormalizationFactor*PDFTimeSlices[ss][vv];

	} // end for bb
    }// end for tt

  // Hide info:

  EyeMatrix[0][0] = Signal->GetBitLength();
  EyeMatrix[1][0] = MinCurrent;
  EyeMatrix[2][0] = MaxCurrent;

} // end ComputeEye

// #############################################################

// NOTE: This method really belongs in MatrixTools!!

void ChiSquareReceiver::SymmetricEig(double ** SymmMatrix, double * EigValue,
				     double ** EigVector, int Dim)
{

  // 2D arrays in C++ are in Row Major order in that
  // A[i][j] = Data[j+i*NumCols], i.e. elements on the same
  // row of A are stored contiguously in Data.

  // 2D arrays in Fortran are in Column Major order in that
  // A[i][j] = Data[i+j*NumRows], i.e. elements on the same
  // column of A are stored contiguously in Data.

  // LAPACK is written in Fortran. So when we go from C++ to Fortran
  // we must first take the transpose, unless the matrix is symmetric.
  // Likewise when we go from Fortran to C++ we must take a transpose.
  // after we come back to C++

  // EigVector = SymmMatrix

  CopyMatrix(EigVector,SymmMatrix,Dim);

  char jobz = 'V';
  char uplo = 'U';
  int Info;

  // Call to LAPACK subroutine that computes all eigenvalues and, 
  // optionally, eigenvectors of a real symmetric matrix.

  // Other possibility: dsyevr

  F77NAME(dsyev)(&jobz,&uplo,&Dim,&(EigVector[0][0]),&Dim,EigValue,
  	         SymmetricEigWork,&LengthSymmetricEigWork,&Info);

  TransposeMatrix(EigVector,Dim);
  
 if( Info < 0 )
   {
     cerr << "ChiSquareReceiver::SymmetricEig" << endl
	  << "Error in LAPACK routine dsyev: info = " << Info << endl
           << "(" << Info << "-th argument had an illegal value)" << endl;
      exit(1);
   }
   else if( Info > 0 )
   {
      cerr << "ChiSquareReceiver::SymmetricEig" << endl
	   << "Error in LAPACK routine dsyev: Info = " << Info << endl
           << "(algorithm failed to converge)" << endl;
      exit(1);
   }

 
}  // ## end ChiSquareReceiver::SymmetricEig

// #############################################################

void ChiSquareReceiver::PrepareSymmetricEig(void)
{

     int ispec = 1;
     char * name = "DSYTRD";
     char opts = 'U';
     int n1 = 2*CovMatrixDim;
     int n2=-1;
     int n3=-1;
     int n4=-1;
  
     // The last argument is an int which is the length of the
     // character array name.

     int NB = F77NAME(ilaenv)(&ispec,name,&opts,&n1,&n2,&n3,&n4,6);

     //cout << "NB = " << NB << endl;

     LengthSymmetricEigWork=max(1,(NB+2)*2*CovMatrixDim);

     SymmetricEigWork = new double[LengthSymmetricEigWork];
}

// ##########################################################

bool ChiSquareReceiver::TestEigDecomposition(double ** Matrix, 
					     double * EigValue,
					     double ** EigVector,int Dim)
{

 bool PassesTest = 1;

 DenseMatrixSimilarityProduct(EigVector,Matrix,TestEigMatrix,WorkMatrix,
			      1,0,2*CovMatrixDim);

 AddVectorToPrincipalDiag(TestEigMatrix,EigValue,-1,2*CovMatrixDim);

 double InfNormM = InfNormMatrix(Matrix,2*CovMatrixDim);
 double RelInfNormT = InfNormMatrix(TestEigMatrix,2*CovMatrixDim)/InfNormM;

 if(RelInfNormT > 1e-10)
   PassesTest = 0;

 if(DebugLevel >=2 && !PassesTest)
  cout << "EigDecomp: RelInfNormT = " << RelInfNormT << endl;

 DenseMatrixProduct(EigVector,EigVector,TestEigMatrix,1,0,2*CovMatrixDim);

 AddToPrincipalDiag(TestEigMatrix,2*CovMatrixDim,-1.0);

 RelInfNormT = InfNormMatrix(TestEigMatrix,2*CovMatrixDim)/InfNormM;

 if(DebugLevel >=2 && !PassesTest)
  cout << "OrthoCheck: RelInfNormT = " << RelInfNormT << endl;

  if(RelInfNormT > 1e-10)
   PassesTest = 0;

  return PassesTest;

}

// ######################################################

void ChiSquareReceiver::WriteFileSquareMatrix(double **A, int dim, 
					string OutFileName, 
				        double NormalizationFactor)
{
 
   ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "Error: " << OutFileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   // We use 1.0/NormalizationFactor since we need to un-normalize before
   // we outputb data.

  WriteMatrix(A,dim,&OutFile,1.0/NormalizationFactor);

  OutFile.close();

}

// ###################################################

void ChiSquareReceiver::WriteFileVector(double *V, int dim, 
					string OutFileName, 
	  	                        double NormalizationFactor)
{

   ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "Error: " << OutFileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   // We use 1.0/NormalizationFactor since we need to un-normalize before
   // we output data.

   for(int jj=0;jj<dim;jj++)
     OutFile << V[jj]/NormalizationFactor << endl;

   OutFile.close();


}

//#######################################################

void ChiSquareReceiver::WriteFilePairVectors(double * x, double *y, int dim, 
					string OutFileName, 
	  	                        double xNormalizationFactor,
					double yNormalizationFactor)
{

   ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "Error: " << OutFileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   // We use 1.0/NormalizationFactor since we need to un-normalize before
   // we output data.

   for(int jj=0;jj<dim;jj++)
     OutFile << xNormalizationFactor*x[jj] << "\t" 
	     << y[jj]/yNormalizationFactor << endl;

   OutFile.close();


}

//#######################################################

void ChiSquareReceiver::WriteFileComplexMatrix(cplx **A, int dim, 
					string RealOutFileName, 
					string ImagOutFileName, 
				        double NormalizationFactor)
{
 
   ofstream OutFile;

   OutFile.open(RealOutFileName.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "Error: " << RealOutFileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   // We use 1.0/NormalizationFactor since we need to un-normalize before
   // we output data.

  for(int ii=0;ii<dim;ii++)
    {
     for(int jj=0;jj<dim;jj++)
       OutFile << real(A[ii][jj]) << "\t";

     OutFile << endl;
    }

  OutFile.close();

  OutFile.open(ImagOutFileName.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "Error: " << ImagOutFileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   // We use 1.0/NormalizationFactor since we need to un-normalize before
   // we output data.

  for(int ii=0;ii<dim;ii++)
    {
     for(int jj=0;jj<dim;jj++)
       OutFile << imag(A[ii][jj]) << "\t";

     OutFile << endl;
    }

  OutFile.close();


}

// ############################################################

void ChiSquareReceiver::WriteFilePDFTimeSlices(string OutFileName)
{

   ofstream OutFile;

   OutFile.open(OutFileName.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "Error: " << OutFileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   
  OutFile << "% (1) Current [mV] (t+1) PDF of current in t-th TimeSlice" 
	  << endl;

  // The choices of NoramlizationFactor ensure that pdf integrates to 1.0

  for(int vv=0;vv<NumPointsCurrent;vv++)
    {

      OutFile << 1e3*CurrentVector[vv]/NormalizationFactor << "\t";

      for(int tt=0;tt<NumTimeSlices;tt++)
       OutFile << NormalizationFactor*PDFTimeSlices[tt][vv] << "\t";

      OutFile << endl;
    }

  OutFile.close();

}

// ############################################################

void ChiSquareReceiver::WriteFileGSLFunction(double CurrentValue, 
					     double s_min,double s_max,
					     int NumPts,string OutFileName)
{

   ofstream OutFile;

   OutFile.open(OutFileName.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "Error: " << OutFileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   if(NumPts==0)
     {
       cerr << "Error in ChiSquareReceiver::WriteFileGSLFunction" << endl;
       exit(1);
     }


   Parameters->ActiveCurrentValue = CurrentValue;
   double Delta_s = (s_max-s_min)/(NumPts-1);

   OutFile.precision(16);

   for(int ss=0;ss<NumPts;ss++)
     {   
       double s_value = s_min + ss*Delta_s;
       OutFile << s_value << "\t"
	       << GSL_Function->function(s_value,GSL_Function->params)
	       << endl;
     }

   OutFile.close();

}

// ############################################################

void ChiSquareReceiver::InitializeGSL_FunctionToDerivativePsi(void)
{
  GSL_Function =  new gsl_function;

  GSL_Function->function = &ChiSquareReceiverDerivativePsi;
  GSL_Function->params = Parameters;

  //double Value = GSL_Function->function(2,GSL_Function->params);
  //cout << "GSL = " << Value << endl;

}

// ############################################################

double ChiSquareReceiver::FindRootInterval(double &LowerBound,
					   double &UpperBound)
{

  // We impose very stringent tests for root since we need the root
  // to be very accurate to calculate the pdf accurately.

  RootFinderMaxIterations = 200;
  double ResidualEpsilon = 1e-8;
  double IntervalAbsoluteEpsilon =  1e-8;
  double IntervalRelativeEpsilon =  1e-8;
  double GaussianApproximationEpsilon = 1e-8;

  int status;
  int status1,status2,status3;

  int iter = 0;

  const gsl_root_fsolver_type *T;

  gsl_root_fsolver *s;

  double r;
  
  T = gsl_root_fsolver_brent;
  s = gsl_root_fsolver_alloc (T);

  gsl_root_fsolver_set(s, GSL_Function,LowerBound,UpperBound );

  double NewLowerBound;
  double NewUpperBound;

  double ResidualFunctionValue;
  double RelGaussianError;

  do
    {
      iter++;

      status = gsl_root_fsolver_iterate (s);
      r = gsl_root_fsolver_root (s);

      NewLowerBound = gsl_root_fsolver_x_lower (s);
      NewUpperBound = gsl_root_fsolver_x_upper (s);
    
      ResidualFunctionValue = GSL_Function->function(r,GSL_Function->params);

      // Test that |ResidualFunctionValue| < ResidualEpsilon

      status1 = gsl_root_test_residual(ResidualFunctionValue,ResidualEpsilon);
   
      // Tests that 
      // | NewUpperBound - NewLowerBound | < 
      // IntervalAbsoluteEpsilon + 
      // IntervalRelativeEpsilon min(| NewUpperBound | , | NewLowerBound | )
      
      status2 = gsl_root_test_interval(NewLowerBound,NewUpperBound,
				       IntervalAbsoluteEpsilon,
				       IntervalRelativeEpsilon);


      status3 = GSL_CONTINUE;

      double LowerBoundPDFValue = EvaluateGaussianApproximation(NewLowerBound);
      double UpperBoundPDFValue = EvaluateGaussianApproximation(NewUpperBound);

      RelGaussianError = 2.0*abs(LowerBoundPDFValue - UpperBoundPDFValue)/
	                 (abs(LowerBoundPDFValue) + abs(UpperBoundPDFValue));

      if(RelGaussianError < GaussianApproximationEpsilon)
         status3 = GSL_SUCCESS;

    }
  while ( (status1 == GSL_CONTINUE || status2 == GSL_CONTINUE ||
	   status3 == GSL_CONTINUE )
	 && iter < RootFinderMaxIterations);

if (status1 != GSL_SUCCESS || status2 != GSL_SUCCESS || status3 != GSL_SUCCESS)
  {

    if(status1 != GSL_SUCCESS)
       cout << "ChiSquareReceiver::FindRootInterval FAILED: " 
	    << "ResidualFunctionValue = " << ResidualFunctionValue 
	    << endl;

    if(status2 != GSL_SUCCESS)
       cout << "ChiSquareReceiver::FindRootInterval FAILED: "
	    << "Interval Width = "
	    << abs(NewUpperBound - NewLowerBound) << endl;

    if(status3 != GSL_SUCCESS)
       cout << "ChiSquareReceiver::FindRootInterval FAILED: "
	    << "RelGaussianError = " << RelGaussianError
	    << endl;    

   /***
   WriteFileGSLFunction(Parameters->ActiveCurrentValue,
			LowerBound,UpperBound,100,
			OutFileNamePrefix + "DerivativePsi.dat");

   WriteFileGSLFunction(Parameters->ActiveCurrentValue,
			NewLowerBound,NewUpperBound,100,
			OutFileNamePrefix + "DerivativePsiZoom.dat");
   ***/

   TypeIntegrationMethod = FOURIER_TRANSFORM;
   
  }
    
  return r;

} // ## end FindRootInterval

// #############################################################

cplx ChiSquareReceiverPsiCplx(cplx s, void * InParameters)
{

  struct ChiSquareReceiverParameters *Parameters
    = (struct ChiSquareReceiverParameters *) InParameters;

  cplx Value = -s*Parameters->ActiveCurrentValue;

  for(int kk=0;kk<2*Parameters->CovMatrixDim;kk++)
    {

      cplx z = 1.0-2.0*s*Parameters->LambdaVec[kk];
      double theta = atan2(imag(z),real(z));
      
      Value += -0.5*(log(abs(z)) + jc*theta) 
	       + s*Parameters->LambdaQSqVec[kk]/z;
    }

  return Value; 

}

// ###########################################################

double ChiSquareReceiverPsi(double s, void * InParameters)
{

  struct ChiSquareReceiverParameters *Parameters
    = (struct ChiSquareReceiverParameters *) InParameters;

  double Value = -s*Parameters->ActiveCurrentValue;

  for(int kk=0;kk<2*Parameters->CovMatrixDim;kk++)
    {

      double z = 1-2*s*Parameters->LambdaVec[kk];
      // theta = 0 as s and hence z are real
      
      Value += -0.5*log(abs(z))
	       + s*Parameters->LambdaQSqVec[kk]/z;
    }

  return Value; 

}

// ############################################################

double ChiSquareReceiverDerivativePsi(double s, void * InParameters)
{

  struct ChiSquareReceiverParameters *Parameters
    = (struct ChiSquareReceiverParameters *) InParameters;

  double Value = -Parameters->ActiveCurrentValue;

  //cout << "Term0 = " << Value << endl;

  for(int kk=0;kk<2*Parameters->CovMatrixDim;kk++)
    {

      double z = 1-2*s*Parameters->LambdaVec[kk];
      Value += (Parameters->LambdaVec[kk]*z + 
		Parameters->LambdaQSqVec[kk])/(sq(z));

      /**
      cout << "Term = " << (Parameters->LambdaVec[kk]*z + 
			    Parameters->LambdaQSqVec[kk])/(sq(z)) 
	   << endl;
      **/

    }

  return Value; 

}

// ############################################################

double ChiSquareReceiverSecondDerivativePsi(double s, void * InParameters)
{

  struct ChiSquareReceiverParameters *Parameters
    = (struct ChiSquareReceiverParameters *) InParameters;

  double Value=0;

  for(int kk=0;kk<2*Parameters->CovMatrixDim;kk++)
    {

      double z = 1-2*s*Parameters->LambdaVec[kk];
      Value += (2*sq(Parameters->LambdaVec[kk])*z 
                + 4*Parameters->LambdaVec[kk]*
		    Parameters->LambdaQSqVec[kk])/(cube(z));
    }

  return Value; 

}

// ############################################################

void ChiSquareReceiver::ReceiveSignalCWNoLossNoDisp(string OutFileNamePrefix2,
                        double P_0, double z, double gamma, double N_ASE,double NF)
{

DemuxChannArrayIndex = Signal->GetChannelArrayIndex(0);

OutFileNamePrefix = Job + OutFileNamePrefix2;

ReceiveNoiseFreeSignalAndRecoverClock();
PrepareToComputeKLModes();

for(int vv=0;vv<NumPointsCurrent;vv++)
 CurrentVector[vv] = NormalizationFactor*(MinCurrent+vv*DeltaCurrent);

double Phi = gamma*P_0*z;

cout << "Phi = " << Phi << endl;

NormalizationFactor = NF;

cout << "ReceiveSignalCWNoLossNoDisp, NormalizationFactor = " << NormalizationFactor << endl;

double LambdaPlus =  N_ASE*(0.5 + Phi*(Phi+sqrt(1+Phi*Phi)))*(4*Signal->GetDeltaFreq());
double LambdaMinus = N_ASE*(0.5 + Phi*(Phi-sqrt(1+Phi*Phi)))*(4*Signal->GetDeltaFreq());

cout << "LambdaPlus, LambdaMinus = " << NF*LambdaPlus << " " << NF*LambdaMinus << endl;

for(int kk=0;kk<CovMatrixDim;kk++)
 LambdaVec[2*kk] = NormalizationFactor*LambdaPlus;

for(int kk=0;kk<CovMatrixDim;kk++)
 LambdaVec[2*kk+1] = NormalizationFactor*LambdaMinus;

for(int kk=0;kk<2*CovMatrixDim;kk++)
  LambdaQSqVec[kk] = 0.0;

LambdaQSqVec[0] = NormalizationFactor*P_0*sq(cos(Phi));
LambdaQSqVec[1] = NormalizationFactor*P_0*sq(sin(Phi));

cout << "LambdaQSq0, LambdaQSq1 = " << LambdaQSqVec[0] << " " << LambdaQSqVec[1] << endl;

 if(DebugLevel >= 1)
   {
      double Tmp = NormalizationFactor;
      NormalizationFactor = 1;

      char  C_String[10];

      sprintf(C_String,"%i.dat",0);

      string TimeSuffix = C_String;

      WriteFileVector(LambdaVec,2*CovMatrixDim,
                     Job+"LambdaVecAnal"+TimeSuffix,
                     NormalizationFactor);
     WriteFileVector(LambdaQSqVec,2*CovMatrixDim,
                     Job+"LambdaQSqVecAnal"+TimeSuffix,
                     NormalizationFactor);

     NormalizationFactor = Tmp;
   }



  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      ComputePDFAtTimeValue(tt);
    }

  switch(TypeOutput)
    {
     case PDFS:
       WriteFilePDFs();
       break;
     case EYE:
       WriteFileEye();
       break;
    }

} // ## end ReceiveSignalCWNoLossNoDisp


