// $Id:

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2006):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

// Written by John Zweck, Jan 2006

#include "GaussianReceiver.hh"

extern ofstream *LogFile;

// ############################################################

GaussianReceiver::GaussianReceiver(OptSignal * WDMSignal2,
		   RanNumGen *RNG2,
		   int DemuxChannArrayIndex2,
		   string InDir, 
		   string Job2)
{

LogFileSeparator();

*LogFile << "GaussianReceiver Constructor" << endl;
 
WDMSignal = WDMSignal2;
DemuxChannArrayIndex = DemuxChannArrayIndex2;
Job = Job2;

 // ### Open Input Parameter File #######

 string InFileName = "GaussianReceiver.in";

  ifstream InFile(InFileName.c_str(),ios::in);    

  if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }

   
  LogFileSeparator();
  *LogFile << InFileName << endl << endl;


 // ####### Read in Parameters ############

 DebugLevel = ReadInt("DebugLevel"," ",0,2,
			  LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);

 TypeOutput = typeOutput(ReadInt("TypeOutput"," ",1,2,
			 LOWER_AND_UPPER,OPTIONAL_NO_WARNING,1,&InFile));

 TypeGaussianNoise = typeGaussianNoise(ReadInt("TypeGaussianNoise"," ",1,3,
			 LOWER_AND_UPPER,OPTIONAL_NO_WARNING,1,&InFile));

 TypePMDonSignal = typePMDonSignal(ReadInt("TypePMDonSignal"," ",1,2,
			 LOWER_AND_UPPER,OPTIONAL_NO_WARNING,1,&InFile));

 TypeClockRecovery = typeClockRecoveryGaussianReceiver
                         (ReadInt("TypeClockRecovery"," ",1,2,
			 LOWER_AND_UPPER,OPTIONAL_NO_WARNING,1,&InFile));


 if(TypeClockRecovery==MANUALLY_SET)
    ClockRecoveryTime = ReadDouble("ClockRecoveryTime","sec",1,-1,
			            NO_BOUNDS,MANDATORY,0,&InFile);

   // ## Initialize CurrentVector #####

   MinCurrent = ReadDouble("MinCurrent","V",1,-1,
			   NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);

   MaxCurrent = ReadDouble("MaxCurrent","V",MinCurrent,-1,
			    LOWER_ONLY,MANDATORY,0,&InFile);

   if(MaxCurrent < MinCurrent)
     cerr << "ABORT: ChiSquareReceiver Error: [MinCurrent,MaxCurrent] = [" 
	  << MinCurrent << "," << MaxCurrent << "]" << endl;

   NumPointsCurrent = ReadInt("NumPointsCurrent"," ",0,-1,
			   LOWER_ONLY,MANDATORY,0,&InFile);

  InFile.close();

*LogFile << "TypeOutput = " << WriteTypeOutput() << endl;
*LogFile << "TypeGaussianNoise = " << WriteTypeGaussianNoise() << endl;
*LogFile << "TypePMDonSignal = " << WriteTypePMDonSignal() << endl;
*LogFile << "TypeClockRecovery = " << WriteTypeClockRecovery() << endl;

 DeltaCurrent = (MaxCurrent - MinCurrent)/(NumPointsCurrent-1);

 CurrentVector = new double[NumPointsCurrent];

 for(int vv=0;vv<NumPointsCurrent;vv++)
   CurrentVector[vv] = MinCurrent+vv*DeltaCurrent;


  // ######## Initialize Memory #####################

SingleChannelSignal = new OptSignal(*WDMSignal);

TunableOptFilter = new OptFilter(InDir + "TunableOptFilter.in",
			         SingleChannelSignal);
  
oOptDemuxer = new OptDemuxer(WDMSignal,TunableOptFilter,SingleChannelSignal);

// Make following call so that in constructor the call to 
// oElecSignalStat->SetOptElecFiltersForGaussianStatistics(
//		                   TunableOptFilter,oElecFilter);
// is correct.

 oOptDemuxer->PrepareToDemuxChannel(DemuxChannArrayIndex);

oPhotodetector = new Photodetector (InDir + "Photodetector.in",
				    SingleChannelSignal,RNG2);

oElecFilter = new ElecFilter(InDir + "ElecFilter.in",oPhotodetector);

oElecSignalStat = new  ElecSignalStat(SingleChannelSignal,oPhotodetector);
	
oElecSignalStat->SetOptElecFiltersForGaussianStatistics(
   	                   TunableOptFilter,oElecFilter);


 switch(TypeOutput)
   {

    case PDFS:
      NumTimeSlices = SingleChannelSignal->GetStringLength();
      TimeBetweenSlices = SingleChannelSignal->GetTimeWindow()/NumTimeSlices;
      break;
  
    case EYE:
      {	   
        int NumPtsPerBit = SingleChannelSignal->GetQtPoints()/
	                   SingleChannelSignal->GetStringLength();

        NumSlicesPerBit = ReadInt("NumSlicesPerBit"," ",1,-1,
			          LOWER_ONLY,OPTIONAL_NO_WARNING,NumPtsPerBit,
				  &InFile);

	NumTimeSlices = SingleChannelSignal->GetStringLength()*NumSlicesPerBit;
        TimeBetweenSlices =SingleChannelSignal->GetBitLength()/NumSlicesPerBit;

       }
 
      break;
   } // end switch


TimeSlices = NewDouble(NumTimeSlices);
TimeSliceIndices = NewInt(NumTimeSlices);
VarianceCurrent = NewDouble(NumTimeSlices);
MeanCurrent = new double[NumTimeSlices];

NumBitsToPlot = NumTimeSlices;

if(TypeOutput == PDFS)
     {
      AveragePDFZeros = new double[NumPointsCurrent];
      AveragePDFOnes = new double[NumPointsCurrent];
      BitString = new int[NumBitsToPlot];
      BER = new double[NumPointsCurrent];

      PDFTimeSlicesStorageVec = new double[NumTimeSlices*NumPointsCurrent];

      PDFTimeSlices = new double*[NumTimeSlices];

      for(int tt=0;tt<NumTimeSlices;tt++)
       PDFTimeSlices[tt] = &(PDFTimeSlicesStorageVec[tt*NumPointsCurrent]);

     }


ObjectInitialized = 1;
FirstOrderPMDParametersSet = 0;

 if(TypePMDonSignal == FIRST_ORDER_PMD && 
    TypeGaussianNoise == PARTIALLY_POLARIZED)
   {  
     StokesPlusPlus = new double[3];
     StokesPlusMinus = new cplx[3];
     WDMSignalTmp = new OptSignal(*WDMSignal);
   }

} // ## end GaussianReceiver

// #######################################################

GaussianReceiver::~GaussianReceiver()
{
 if(ObjectInitialized == 1)
    {
     oOptDemuxer->~OptDemuxer();
     TunableOptFilter->~OptFilter();    
     oElecFilter->~ElecFilter();
     oElecSignalStat->~ElecSignalStat();
     oPhotodetector->~Photodetector();
     SingleChannelSignal->~OptSignal();
     delete [] VarianceCurrent;
     delete [] MeanCurrent;

  if(TypeOutput == PDFS)
   {
     delete [] BER;
     delete [] BitString;
     delete [] AveragePDFZeros;
     delete [] AveragePDFOnes;

     for(int tt=0;tt<NumTimeSlices;tt++)
       PDFTimeSlices[tt]= 0;

     delete [] PDFTimeSlices;
     delete [] PDFTimeSlicesStorageVec;
   }

  delete [] CurrentVector;
  delete [] TimeSlices;

  if(TypePMDonSignal == FIRST_ORDER_PMD && 
     TypeGaussianNoise == PARTIALLY_POLARIZED)
   {
     delete [] StokesPlusPlus;
     delete [] StokesPlusMinus;
     WDMSignalTmp->~OptSignal();
   }

    }
}

// #######################################################

string GaussianReceiver::WriteTypeOutput(void)
{

  string ReturnString;

  switch(TypeOutput)
    {
     case PDFS: 
      ReturnString = "PDFS";
      break;

     case EYE: 
      ReturnString = "EYE";
      break;

     default:
       cerr << "Switch error in GaussianReceiver::WriteTypeOutput"
	    << endl
	    << "ABORTING NOW!!" << endl;
       exit(1);

    }

  return ReturnString;

} // ## end WriteTypeOutput

// ############################################################

string GaussianReceiver::WriteTypeGaussianNoise(void)
{

  string ReturnString;

  switch(TypeGaussianNoise)
    {
     case UNPOLARIZED: 
      ReturnString = "UNPOLARIZED";
      break;

     case PARTIALLY_POLARIZED: 
      ReturnString = "PARTIALLY_POLARIZED";
      break;

     case PARTIALLY_POLARIZED_WITH_DGD: 
      ReturnString = "PARTIALLY_POLARIZED_WITH_DGD";
      break;

     default:
       cerr << "Switch error in GaussianReceiver::WriteTypeGaussianNoise"
	    << endl
	    << "ABORTING NOW!!" << endl;
       exit(1);

    }

  return ReturnString;

} // ## end WriteTypeGaussianNoise

// ############################################################

string GaussianReceiver::WriteTypePMDonSignal(void)
{

  string ReturnString;

  switch(TypePMDonSignal)
    {
     case NO_PMD: 
      ReturnString = "NO_PMD";
      break;

     case FIRST_ORDER_PMD: 
      ReturnString = "FIRST_ORDER_PMD";
      break;

     default:
       cerr << "Switch error in GaussianReceiver::WriteTypePMDonSignal"
	    << endl
	    << "ABORTING NOW!!" << endl;
       exit(1);

    }

  return ReturnString;

} // ## end WriteTypePMDonSignal

// ############################################################

string GaussianReceiver::WriteTypeClockRecovery(void)
{

  string ReturnString;

  switch(TypeClockRecovery)
    {
    case  AUTOMATIC:
      ReturnString = "AUTOMATIC";
      break;

    case MANUALLY_SET:
      ReturnString = "MANUALLY_SET";
      break;

     default:
       cerr << "Switch error in GaussianReceiver::WriteTypeClockRecovery"
	    << endl
	    << "ABORTING NOW!!" << endl;
       exit(1);

    }

  return ReturnString;

} // ## end WriteTypeClockRecovery

// ############################################################

void GaussianReceiver::ReceiveSignal(string FileID)
{

  
  if(TypeGaussianNoise == UNPOLARIZED)
    ReceiveSignalUnpolarizedNoise(FileID);
  else
   if(TypePMDonSignal == NO_PMD && TypeGaussianNoise == PARTIALLY_POLARIZED)
     // Signal must be X-polarized!
    ReceiveSignalNoPMD_NoisePartiallyPolarized(FileID);
   else
    if(TypePMDonSignal == FIRST_ORDER_PMD && 
       TypeGaussianNoise == PARTIALLY_POLARIZED)
         ReceiveSignalDGD_NoisePartiallyPolarized(FileID);
     else
      {
       cout << "GaussianReceiver::ReceiveSignal: "
	    <<" This option not yet implemented:" << endl
	    << "TypePMDonSignal = " << WriteTypePMDonSignal() 
	    << " and TypeGaussianNoise = " << WriteTypeGaussianNoise() << endl
	    << "ABORTING NOW!" << endl;
       exit(1);
      }
  

  if(TypeOutput == PDFS)
    {
     ComputeAveragePDFs(FileID);
     ComputeMinimumBERAndDecisionThreshold(FileID);
    }

}

// ############################################################

void GaussianReceiver::RecoverClockAndSetTimeSlices(void)
{

 if(TypeClockRecovery == AUTOMATIC)
   {
    oOptDemuxer->DemuxChannel(DemuxChannArrayIndex);
    oPhotodetector->DetectOptSignal();
    oElecFilter->FilterElecSignal();
    oElecSignalStat->ComputePerformanceStatistics();
    ClockRecoveryTime = oElecSignalStat->GetClockRecoveryTimeBeforeTimeShift();
    cout << "ClockRecoveryTime [ps] = " << ClockRecoveryTime*1e12 << endl;
    
   }


 SetTimeSlices();
}

// ####################################################

void GaussianReceiver::SetTimeSlices(void)
{

 for(int tt=0;tt<NumTimeSlices;tt++)
    {
      TimeSlices[tt] = ClockRecoveryTime + tt*TimeBetweenSlices;

      if(TypeOutput==EYE)
	TimeSlices[tt] -= SingleChannelSignal->GetBitLength()/2.0;
         	
      // We want TimeSlices[tt] to be in range
      // 0 <= TimeSlices[tt] < TimeWindow

      while(TimeSlices[tt] < 0)
        TimeSlices[tt] += SingleChannelSignal->GetTimeWindow();

      while(TimeSlices[tt] >= SingleChannelSignal->GetTimeWindow())
        TimeSlices[tt] -= SingleChannelSignal->GetTimeWindow();

      TimeSliceIndices[tt] =  int(floor(TimeSlices[tt]/
	                      SingleChannelSignal->GetDeltaTime()));

    } // end for tt

   
  if(DebugLevel >= 2)
    {

     cout << "TimeSlices = " << endl;
      for(int tt=0;tt<NumTimeSlices;tt++)
       cout << TimeSlices[tt] << endl;
     cout << endl;
   
    }

} // ## end GaussianReceiver::SetTimeSlices(void)

// ###########################################################

void GaussianReceiver::ReceiveSignalUnpolarizedNoise(string FileID)
{

  RecoverClockAndSetTimeSlices();

  /* The following call to PrepareToDemuxChannel
     sets the number of channels of the SingleChannelSignal
     to 1, reduces the WDM BitString to a SingleChannel BitString,
     sets the carrier frequency of the only channel in the SingleChannel 
     signal and sets the center frequency of the optical filter. It does
     not do the actual filtering since we need to do that in the 
     semianalytical receiver. It also resets 
     *SingleChannelSignal = *WDM_Signal
  */

  oOptDemuxer->PrepareToDemuxChannel(DemuxChannArrayIndex);

  // The following call does not allocate any new memory
  
  oElecSignalStat->ResetOptElecFiltersForGaussianStatistics();

  double TotalNoiseSpectralDensity = 
  SingleChannelSignal->GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics();

  cout << "TotalNoiseSpectralDensity [W/Hz] = " 
       << TotalNoiseSpectralDensity << endl;

 string OutFileName = Job + FileID + "TotalNoiseSpectralDensity.dat";

 ofstream OutFileN_ASE(OutFileName.c_str(),ios::out);

  if (!OutFileN_ASE)
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

 OutFileN_ASE << "%TotalNoiseSpectralDensity [W/Hz] " << endl
              << TotalNoiseSpectralDensity << endl;

 OutFileN_ASE.close();

  if(SingleChannelSignal->GetTypeSimulation()==SCALAR) 
    {
      cerr << "GaussianReceiver::ReceiveSignalUnpolarizedNoise: " 
	   << endl
	   << "SCALAR Signal not implemented (yet)"
	   << endl
	   << "ABORTING" << endl;
      exit(1);
    }
 
 
 MeanNoise = oElecSignalStat->GetMeanNoiseGaussianStatistics(
				    TotalNoiseSpectralDensity);

 double DOP_Noise = 0.0;

 double NoiseNoiseVariance = 
   oElecSignalStat->GetNoiseNoiseVarianceGaussianStatistics(
			TotalNoiseSpectralDensity,DOP_Noise);
 
 for(int tt=0;tt < NumTimeSlices; tt++)
   {
    VarianceCurrent[tt] = NoiseNoiseVariance + oElecSignalStat->
      GetSignalNoiseVarianceForUnpolarizedNoiseGaussianStatistics(
	 TimeSlices[tt], TotalNoiseSpectralDensity);
   }
  
  OutFileName = Job + FileID + "StdDevCurrent.dat";

  ofstream OutFile(OutFileName.c_str(),ios::out);
  
  if (!OutFile) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }


  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      OutFile << TimeSlices[tt] << "\t" 
	      << sqrt(VarianceCurrent[tt]) << endl;
    }

  OutFile.close();

  // In case we have changed the data in SingleChannelSignal in the 
  //   previous function calls we re-initialize the SingleChannelSignal
  //   from the WDM signal, and we then apply the demuxing filter to it.
  

  oOptDemuxer->DemuxChannel(DemuxChannArrayIndex);

  oPhotodetector->DetectOptSignal();
  oElecFilter->FilterElecSignal();

  oPhotodetector->WriteFileTimePower(Job+"tOutElec.dat");

  OutFileName = Job + FileID + "MeanCurrent.dat";

  ofstream OutFile2(OutFileName.c_str(),ios::out);

  if (!OutFile2) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  for(int tt=0;tt<NumTimeSlices;tt++)
    {

      MeanCurrent[tt] = MeanNoise + 
	                real(oPhotodetector->sfftPM.tPowerSignal[
					    TimeSliceIndices[tt]]);

      OutFile2 << TimeSlices[tt] << "\t" <<  MeanCurrent[tt] <<  endl;
    }

 
  OutFile2.close();

  OutFileName = Job + FileID + "Clock.dat";

  ofstream OutFile3(OutFileName.c_str(),ios::out);

  if (!OutFile3) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  OutFile3 << "% (1) ClockTime [ps]" << endl
	   <<  1e12*ClockRecoveryTime << endl;
 
  OutFile3.close();


} // ## end ReceiveSignalUnpolarizedNoise

// #########################################################

void GaussianReceiver::ReceiveSignalNoPMD_NoisePartiallyPolarized
                        (string FileID)
{

  if(SingleChannelSignal->GetTypeSimulation()==SCALAR) 
    {
      cerr << "GaussianReceiver::ReceiveSignalNoPMD_NoisePartiallyPolarized: " 
	   << endl
	   << "Signal must be VECTOR."
	   << endl
	   << "ABORTING." << endl;
      exit(1);
    }

  RecoverClockAndSetTimeSlices();

  /* The following call to PrepareToDemuxChannel
     sets the number of channels of the SingleChannelSignal
     to 1, reduces the WDM BitString to a SingleChannel BitString,
     sets the carrier frequency of the only channel in the SingleChannel 
     signal and sets the center frequency of the optical filter. It does
     not do the actual filtering since we need to do that in the 
     semianalytical receiver. It also resets 
     *SingleChannelSignal = *WDM_Signal
  */

 oOptDemuxer->PrepareToDemuxChannel(DemuxChannArrayIndex);

  // The following call does not allocate any new memory
  
  oElecSignalStat->ResetOptElecFiltersForGaussianStatistics();
 
double epsilon = 1e-3; // a number small compared to 1.0
  
double UnitStokesVectorPolarizedPartNoise[3];

for(int ii=0;ii<3;ii++)
  UnitStokesVectorPolarizedPartNoise[ii] = 
    SingleChannelSignal->oReduOptSignal->GetStokesNoise(ii+1,0);

// Now we have a 3x1 Stokes vector S_n normalized by StokesNoise[0]
// So its length (regardless of how the 4x1 vector StokesNoise is normalized)
// is DOP_Noise

 double DOP_Noise = 0.0;

for(int ii=0;ii<3;ii++)
  DOP_Noise += sq(UnitStokesVectorPolarizedPartNoise[ii]);

 DOP_Noise = sqrt(DOP_Noise);

 if(DOP_Noise < epsilon)
   {
    for(int ii=0;ii<3;ii++)
     UnitStokesVectorPolarizedPartNoise[ii] = 0.0;
   }
 else
   {
    for(int ii=0;ii<3;ii++)
      UnitStokesVectorPolarizedPartNoise[ii] /=  DOP_Noise;
   }

double UnitStokesVectorSignal[4];

double DOP_Signal =  SingleChannelSignal->
  GetDegreeOfPolarizationAndTotalStokesFreqDomain(UnitStokesVectorSignal);

 if(DOP_Signal == 0)
   {
   cerr << "GaussianReceiver::ReceiveSignalNoPMD_NoisePartiallyPolarized"
	<< endl
	<< "DOP_signal = 0. ABORT.";
     exit(1);
   }

 // Normalize unit Stokes vector on Poincare sphere

 for(int ii=1;ii<4;ii++)
   UnitStokesVectorSignal[ii] /= DOP_Signal;


 if(DOP_Signal < 1-epsilon)
   {
   cerr << "GaussianReceiver::ReceiveSignalNoPMD_NoisePartiallyPolarized"
	<< endl
	<< "DOP_signal != 1. ABORT.";
     exit(1);
   }

 // Get N_ASE

 double TotalNoiseSpectralDensity = 
 SingleChannelSignal->oReduOptSignal->GetTotalNoiseSpectralDensity();

 cout << "TotalNoiseSpectralDensity (Reduced Model) [W/Hz] = " 
       << TotalNoiseSpectralDensity << endl;
 
 cout << "TotalNoiseSpectralDensity (Full Model) [W/Hz] = "
      << SingleChannelSignal->
         GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics()
      << endl;

 string OutFileName = Job + FileID + "TotalNoiseSpectralDensity.dat";

 ofstream OutFileN_ASE(OutFileName.c_str(),ios::out);

  if (!OutFileN_ASE)
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

 OutFileN_ASE << "%TotalNoiseSpectralDensity [W/Hz] " << endl
              << TotalNoiseSpectralDensity << endl;

 OutFileN_ASE.close();
 
 MeanNoise = oElecSignalStat->GetMeanNoiseGaussianStatistics(
				    TotalNoiseSpectralDensity);

 double NoiseNoiseVariance = 
   oElecSignalStat->GetNoiseNoiseVarianceGaussianStatistics(
			TotalNoiseSpectralDensity,DOP_Noise);
 
 for(int tt=0;tt < NumTimeSlices; tt++)
   {
    VarianceCurrent[tt] = NoiseNoiseVariance + oElecSignalStat->
       GetSignalNoiseVarianceForPolarizedSignalGaussianStatistics(
          TimeSlices[tt],
          &UnitStokesVectorSignal[1],
          UnitStokesVectorPolarizedPartNoise,
          TotalNoiseSpectralDensity,
          DOP_Noise);
   }
  
  OutFileName = Job + FileID + "StdDevCurrent.dat";

  ofstream OutFile(OutFileName.c_str(),ios::out);
  
  if (!OutFile) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }


  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      OutFile << TimeSlices[tt] << "\t" 
	      << sqrt(VarianceCurrent[tt]) << endl;
    }

  OutFile.close();

  // In case we have changed the data in SingleChannelSignal in the 
  //   previous function calls we re-initialize the SingleChannelSignal
  //   from the WDM signal, and we then apply the demuxing filter to it.
  

  oOptDemuxer->DemuxChannel(DemuxChannArrayIndex);

  oPhotodetector->DetectOptSignal();
  oElecFilter->FilterElecSignal();

  oPhotodetector->WriteFileTimePower(Job+"tOutElec.dat");

  OutFileName = Job + FileID + "MeanCurrent.dat";

  ofstream OutFile2(OutFileName.c_str(),ios::out);

  if (!OutFile2) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  for(int tt=0;tt<NumTimeSlices;tt++)
    {

      MeanCurrent[tt] = MeanNoise + 
	                real(oPhotodetector->sfftPM.tPowerSignal[
					    TimeSliceIndices[tt]]);

      OutFile2 << TimeSlices[tt] << "\t" <<  MeanCurrent[tt] <<  endl;
    }

 
  OutFile2.close();

  OutFileName = Job + FileID + "Clock.dat";

  ofstream OutFile3(OutFileName.c_str(),ios::out);

  if (!OutFile3) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  OutFile3 << "% (1) ClockTime [ps]" << endl
	   <<  1e12*ClockRecoveryTime << endl;
 
  OutFile3.close();

}

// ##############################################################

//! The OptFiber PMF must be associated to the OptSignal WDMSignal.
//! You must have called PMF->GetFiberSampleCompleteScatt() before
//! calling this method. The implementation of the theory given in this method
//! assumes that WDMSignal has only been generated using a signal constructor
//! or OptSignal::RegenerateSignalString() and is to be propagated through
//! the PMF. In particular, the signal has not been sent through a 
//! PolarizTransformer. Instead, you can use OptSignal::SetInJones() and
//! OptSignal::RegenerateSignalString()

void GaussianReceiver::ComputeFirstOrderPMDParameters(OptFiber * PMF)
{

 DGD = PMF->GetDGD(WDMSignal->GetCenterFreq());

 cout << "DGD = " << DGD << endl;

 double OutPDVectorPlus[3],OutPDVectorMinus[3];
 double StokesParametersPlusPlus[4];
 double StokesParametersMinusMinus[4];
	  
 for(int kk = 0; kk < 3; kk++)
  OutPDVectorPlus[kk] = 0;
	
 PMF->GetOutPD_Vector(WDMSignal->GetCenterFreq(),OutPDVectorPlus); 

 cout << "OutPD_Vector = ";
  
 for(int kk = 0; kk < 3; kk++)
   cout << OutPDVectorPlus[kk] << " ";

 cout << endl;

 for(int kk = 0; kk < 3; kk++)
  {
    OutPDVectorMinus[kk] = -OutPDVectorPlus[kk];
    StokesParametersPlusPlus[kk+1] = OutPDVectorPlus[kk]/DGD;
    StokesParametersMinusMinus[kk+1] = -StokesParametersPlusPlus[kk+1];
  }
	  
 StokesParametersPlusPlus[0] = 1.0;
 StokesParametersMinusMinus[0] = 1.0;

 cplx JonesVectorPlus[2], JonesVectorMinus[2];
 cplx StokesParametersPlusMinus[4];
	  
 Stokes2Jones(StokesParametersPlusPlus,JonesVectorPlus);
 Stokes2Jones(StokesParametersMinusMinus,JonesVectorMinus);

  JonesPair2MixedStokes(JonesVectorPlus,JonesVectorMinus,
			StokesParametersPlusMinus);

  // Calculate Power Splitting factors c_plus and c_minus using the
  // formula c_plus = 
  // conjugatetranspose(JonesVectorPlus)(Rotation thru PMF of Signal->InJones)

  cplx OutJones[2]; // Will store Signal->InJones rotated through the PMF.

  OutJones[0] = WDMSignal->GetInJones(0,0);
  OutJones[1] = WDMSignal->GetInJones(0,1);

  PMF->TransformJonesVectorThroughFiber(0.0,OutJones);

 cplx c_plus =
   conj(JonesVectorPlus[0])*OutJones[0] + conj(JonesVectorPlus[1])*OutJones[1];

 cplx c_minus =
   conj(JonesVectorMinus[0])*OutJones[0] + 
   conj(JonesVectorMinus[1])*OutJones[1];

 SetFirstOrderPMDParameters(StokesParametersPlusPlus,StokesParametersPlusMinus,
			    c_plus,c_minus);

}

// ##############################################################

//! StokesParametersPlusPlus and StokesParametersPlusMinus are 4x1 vectors
//! for the first-order PMD fiber. 
//! c_plus2 and c_minus2 are the cplx splitting factors.

void GaussianReceiver::SetFirstOrderPMDParameters(
		double * StokesParametersPlusPlus,
		cplx * StokesParametersPlusMinus, cplx c_plus2, cplx c_minus2)
{

  FirstOrderPMDParametersSet = 1;
  
  // DGD = StokesParametersPlusPlus[0]; //This is wrong. Debugged by Hua.

  for(int ii=0;ii<3;ii++)
    StokesPlusPlus[ii] = StokesParametersPlusPlus[ii+1];

  for(int ii=0;ii<3;ii++)
    StokesPlusMinus[ii] = StokesParametersPlusMinus[ii+1];

  c_plus = c_plus2;
  c_minus = c_minus2;

}

// #############################################################

//! Receives a signal that is the sum of (1) a noise-free signal that has
//! gone through first-order PMD fiber (ie PMF) and (2) Partially polarized
//! noise.

/*! For each different fiber realization and input polarization state, you
    must call GaussianReceiver::ComputeFirstOrderPMDParameters. You must also
    propagate the signal through an OptFiber object
    before calling this method. 
    The clock will be recovered based on the signal after the PMF unless you
    use TypeClockRecovery = MANUALLY_SET.
*/

void GaussianReceiver::ReceiveSignalDGD_NoisePartiallyPolarized(string FileID)
{

  if(SingleChannelSignal->GetTypeSimulation()==SCALAR) 
    {
      cerr << "GaussianReceiver::ReceiveSignalNoPMD_NoisePartiallyPolarized: " 
	   << endl
	   << "Signal must be VECTOR."
	   << endl
	   << "ABORTING." << endl;
      exit(1);
    }

  // We recover the clock using the noise-free signal after the PMF.

  RecoverClockAndSetTimeSlices();

  /* The following call to PrepareToDemuxChannel
     sets the number of channels of the SingleChannelSignal
     to 1, reduces the WDM BitString to a SingleChannel BitString,
     sets the carrier frequency of the only channel in the SingleChannel 
     signal and sets the center frequency of the optical filter. It does
     not do the actual filtering since we need to do that in the 
     semianalytical receiver. It also resets 
     *SingleChannelSignal = *WDM_Signal
  */

  oOptDemuxer->PrepareToDemuxChannel(DemuxChannArrayIndex);  
  oElecSignalStat->ResetOptElecFiltersForGaussianStatistics();
 
  double epsilon = 1e-3; // a number small compared to 1.0
  
  double UnitStokesVectorPolarizedPartNoise[3];

  for(int ii=0;ii<3;ii++)
   UnitStokesVectorPolarizedPartNoise[ii] = 
    SingleChannelSignal->oReduOptSignal->GetStokesNoise(ii+1,0);

// Now we have a 3x1 Stokes vector S_n normalized by StokesNoise[0]
// So its length (regardless of how the 4x1 vector StokesNoise is normalized)
// is DOP_Noise

 double DOP_Noise = 0.0;

for(int ii=0;ii<3;ii++)
  DOP_Noise += sq(UnitStokesVectorPolarizedPartNoise[ii]);

 DOP_Noise = sqrt(DOP_Noise);

 if(DOP_Noise < epsilon)
   {
    for(int ii=0;ii<3;ii++)
     UnitStokesVectorPolarizedPartNoise[ii] = 0.0;
   }
 else
   {
    for(int ii=0;ii<3;ii++)
      UnitStokesVectorPolarizedPartNoise[ii] /=  DOP_Noise;
   }

 // Get N_ASE

 double TotalNoiseSpectralDensity = 
 SingleChannelSignal->oReduOptSignal->GetTotalNoiseSpectralDensity();

 cout << "TotalNoiseSpectralDensity (reduced Model) [W/Hz] = " 
       << TotalNoiseSpectralDensity << endl;
 
 cout << "TotalNoiseSpectralDensity (Full Model) [W/Hz] = "
      << SingleChannelSignal->
         GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics()
      << endl;

 string OutFileName = Job + FileID + "TotalNoiseSpectralDensity.dat";

 ofstream OutFileN_ASE(OutFileName.c_str(),ios::out);

  if (!OutFileN_ASE)
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

 OutFileN_ASE << "%TotalNoiseSpectralDensity [W/Hz] " << endl
              << TotalNoiseSpectralDensity << endl;

 OutFileN_ASE.close();
 
 MeanNoise = oElecSignalStat->GetMeanNoiseGaussianStatistics(
				    TotalNoiseSpectralDensity);

 double NoiseNoiseVariance = 
   oElecSignalStat->GetNoiseNoiseVarianceGaussianStatistics(
			TotalNoiseSpectralDensity,DOP_Noise);

 // In case the preceeding calls have changed it, 
 // we re-initialize the SingleChannelSignal
 // from the WDM signal, and apply the demuxing filter to it.
 // Since WDMSignal has passed through the PMF the mean current
 // is calculated from the signal aftre the PMD as it needs to be!

  oOptDemuxer->DemuxChannel(DemuxChannArrayIndex);

  oPhotodetector->DetectOptSignal();
  oElecFilter->FilterElecSignal();

  oPhotodetector->WriteFileTimePower(Job+FileID+"tOutElec.dat");

  OutFileName = Job + FileID + "MeanCurrent.dat";

  ofstream OutFile2(OutFileName.c_str(),ios::out);

  if (!OutFile2) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  for(int tt=0;tt<NumTimeSlices;tt++)
    {

      MeanCurrent[tt] = MeanNoise + 
	                real(oPhotodetector->sfftPM.tPowerSignal[
					    TimeSliceIndices[tt]]);

      OutFile2 << TimeSlices[tt] << "\t" <<  MeanCurrent[tt] <<  endl;
    }

 
  OutFile2.close();

  // For the calculation of the scalar integrals in the receiver formula,
  // we need the signal to be the one
  // at the transmitter, and moreover, we need this signal to be
  // X-polarized. Since this may not be the case we proceed as follows:

  // In case we need WDMSignal later on in the application code we 
  // first save a temporary copy of it:

  *WDMSignalTmp = *WDMSignal;

  for(int ch=0;ch<WDMSignal->GetNumChannels();ch++)
    {
     WDMSignal->SetInJones(ch,0,1);
     WDMSignal->SetInJones(ch,1,0);
    }

  WDMSignal->RegenerateSignalString();

  // WDMSignal is now X-polarized and as at the transmitter

  oOptDemuxer->PrepareToDemuxChannel(DemuxChannArrayIndex);
  oElecSignalStat->ResetOptElecFiltersForGaussianStatistics(); 

for(int tt=0;tt < NumTimeSlices; tt++)
 {
 
  double I_Plus = oElecSignalStat->GetIntegralForSignalNoiseVariance(
		   SingleChannelSignal,SingleChannelSignal,0.0,0.0,
		   TimeSlices[tt]-DGD/2.0);

  double I_Minus = oElecSignalStat->GetIntegralForSignalNoiseVariance(
		   SingleChannelSignal,SingleChannelSignal,0.0,0.0,
		   TimeSlices[tt]+DGD/2.0);
  
  cplx I_PlusMinus = oElecSignalStat->GetIntegralForSignalNoiseVariance(
		   SingleChannelSignal,SingleChannelSignal,DGD/2.0,-DGD/2.0,
		   TimeSlices[tt]);

  double InnerProduct = 0.0;

  // In formula for InnerProduct below last term has - sign. Debugged by Hua.

  for(int ii=0;ii<3;ii++)
   {
    InnerProduct += UnitStokesVectorPolarizedPartNoise[ii]*
     ( (sq(abs(c_plus))*I_Plus - sq(abs(c_minus))*I_Minus)*StokesPlusPlus[ii]
       + 2*real(I_PlusMinus*conj(c_plus)*c_minus)*real(StokesPlusMinus[ii])
       - 2*imag(I_PlusMinus*conj(c_plus)*c_minus)*imag(StokesPlusMinus[ii]));
   }

  VarianceCurrent[tt] = NoiseNoiseVariance + 
    0.5*sq(oPhotodetector->GetConvFactOptElecPhotod())
      *TotalNoiseSpectralDensity
	*(sq(abs(c_plus))*I_Plus + sq(abs(c_minus))*I_Minus 
	  + DOP_Noise*InnerProduct);

 } // end for int tt loop
  
  OutFileName = Job + FileID + "StdDevCurrent.dat";

  ofstream OutFile(OutFileName.c_str(),ios::out);
  
  if (!OutFile) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }


  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      OutFile << TimeSlices[tt] << "\t" 
	      << sqrt(VarianceCurrent[tt]) << endl;
    }

  OutFile.close();

  *WDMSignal = *WDMSignalTmp;


  OutFileName = Job + FileID + "Clock.dat";

  ofstream OutFile3(OutFileName.c_str(),ios::out);

  if (!OutFile3) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  OutFile3 << "% (1) ClockTime [ps]" << endl
	   <<  1e12*ClockRecoveryTime << endl;
 
  OutFile3.close();


}

//#######################################################

void GaussianReceiver::ComputeAveragePDFs(string FileID)
{

  // Adapted from ChiSquareReceiver::ComputeAveragePDFs()

  // I checked that the correlation is computed correctly
  // by artificially shifting the electrically filtered
  // current as a function of time. 

   for(int tt=0;tt<NumTimeSlices;tt++)
     {

      for(int vv=0;vv<NumPointsCurrent;vv++)
	PDFTimeSlices[tt][vv] = exp( -sq(CurrentVector[vv] - MeanCurrent[tt])/
				     (2.0*VarianceCurrent[tt]))/
	                               (sqrt(2*pi*VarianceCurrent[tt]));
 
      double IntPDFTimeSlice = 
       Quadrature(PDFTimeSlices[tt],NumPointsCurrent,
		CurrentVector[0],CurrentVector[NumPointsCurrent-1]);

      if(abs(IntPDFTimeSlice-1.0) > 0.05)
	{
	  cerr << "GaussianReceiver::ComputeAveragePDFs" 
	       << endl
	       << "Error exceeds 1% in integration of PDFTimeSlices." << endl
	       << "Change Min or Max Current or increase NumPointsCurrent"
	       << endl
	       << "ABORTING" << endl;

	  cout << "tt for Matlab,IntPDFTimeSlice = " 
	       << tt+2 << " " << IntPDFTimeSlice << endl;

	  //exit(1);
	}      

     } // end for tt loop

  if(DebugLevel >= 1)
   WriteFilePDFTimeSlices(Job+FileID+"PDFTimeSlices.dat");

  NumBitsToPlot = NumTimeSlices;

  int NumZeros = 0;
  int NumOnes = 0;

  for(int vv=0;vv<NumPointsCurrent;vv++)
    AveragePDFZeros[vv] = 0.0;

  for(int vv=0;vv<NumPointsCurrent;vv++)
    AveragePDFOnes[vv] = 0.0;

  int NumBitsBetweenSlices = 1;

  for(int bb=0;bb<NumBitsToPlot;bb++)
    {
      int bit = bb*NumBitsBetweenSlices;
     
      BitString[bb] = SingleChannelSignal->oBitString->GetBit(0,bit);
       
    }

  double MaxCorrelation = 0.0;
  int OptimalShift = 0;

  for(int ss=0;ss<NumTimeSlices;ss++)
    {
      double Correlation = 0;

      for(int tt=0;tt<NumTimeSlices;tt++)
	{ 
 
          int BitIndex = ss+tt;

          while(BitIndex < 0)
	    BitIndex += NumBitsToPlot;
          while(BitIndex >= NumBitsToPlot)
	    BitIndex -= NumBitsToPlot;

          Correlation += (MeanCurrent[tt]-MeanNoise)*BitString[BitIndex];
	}

      if(Correlation > MaxCorrelation)
	{
         OptimalShift = ss;
         MaxCorrelation = Correlation;
	}

    } // end for ss


  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      int BitIndex = OptimalShift+tt;

      while(BitIndex < 0)
	BitIndex += NumBitsToPlot;
      while(BitIndex >= NumBitsToPlot)
	BitIndex -= NumBitsToPlot;

      if(BitString[BitIndex] == 0)
	{
          NumZeros += 1;

          for(int vv=0;vv<NumPointsCurrent;vv++)
	    AveragePDFZeros[vv] += PDFTimeSlices[tt][vv];
	}
      else
	{
          NumOnes += 1;

          for(int vv=0;vv<NumPointsCurrent;vv++)
	    AveragePDFOnes[vv] += PDFTimeSlices[tt][vv];
	}

    } // end for tt


  if(NumZeros != 0)
   for(int vv=0;vv<NumPointsCurrent;vv++)
    AveragePDFZeros[vv] /= NumZeros;

  if(NumOnes != 0)
   for(int vv=0;vv<NumPointsCurrent;vv++)
    AveragePDFOnes[vv] /= NumOnes;


  if(DebugLevel >= 1)
   {
    WriteFilePairVectors(CurrentVector,AveragePDFOnes,NumPointsCurrent,
                         Job+FileID+"AveragePDFOnes.dat");

    WriteFilePairVectors(CurrentVector,AveragePDFZeros,NumPointsCurrent,
                         Job+FileID+"AveragePDFZeros.dat");

    }


} // ## end ComputeAveragePDFs

// ###########################################################

void GaussianReceiver::ComputeMinimumBERAndDecisionThreshold(string FileID)
{

  // Adapted from ChiSquareReceiver::ComputeMinimumBERAndDecisionThreshold

  double BitErrorRatio, DecisionThreshold;

  double TotalIntZeros = Quadrature(AveragePDFZeros,NumPointsCurrent,
				    CurrentVector[0],
				    CurrentVector[NumPointsCurrent-1]);

  double TotalIntOnes = Quadrature(AveragePDFOnes,NumPointsCurrent,
				   CurrentVector[0],
				   CurrentVector[NumPointsCurrent-1]);

  double Error = max( abs(TotalIntZeros-1), abs(TotalIntOnes-1) );
  
  if(Error > 0.05)
    {
      cerr << "GaussianReceiver::ComputeMinimumBERAndDecisionThreshold:"
	   <<  endl
	   << "Error in integration of AveragePDFOnes or  AveragePDFZeros "
	   << endl
	   << "exceeds 1%. ABORTING" << endl;

      cout << "TotalIntZeros = " << TotalIntZeros << endl;
      cout << "TotalIntOnes = " << TotalIntOnes << endl;
    
      exit(1);
    }

    // RightEndPt for OnesInt must equal LeftEndPt for Zeros Int

     for(int DecisionIndex=0;DecisionIndex<NumPointsCurrent;DecisionIndex++)
     {
      double ZerosInt = Quadrature(&AveragePDFZeros[DecisionIndex],
				   NumPointsCurrent-DecisionIndex,
				   CurrentVector[DecisionIndex],
				   CurrentVector[NumPointsCurrent-1])/
	                           TotalIntZeros;

      double OnesInt = Quadrature(AveragePDFOnes,DecisionIndex,
				  CurrentVector[0],
				  CurrentVector[DecisionIndex])/
	                          TotalIntOnes;

      BER[DecisionIndex] = 0.5*(ZerosInt+OnesInt);
      
      if(DecisionIndex == 0 || BER[DecisionIndex] < BitErrorRatio)
	{
	 BitErrorRatio = BER[DecisionIndex];
         DecisionThreshold = CurrentVector[DecisionIndex];
	}
     } // end for DecisionIndex loop


     if(DebugLevel >= 1)
       {     
        WriteFilePairVectors(CurrentVector,BER,NumPointsCurrent,
		             Job+FileID+"BERVector.dat");
       }
     

 string OutFileName = Job + FileID + "BER.dat";

 ofstream OutFile(OutFileName.c_str(),ios::out);

 if (!OutFile) 
   {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  OutFile << "% (1) BER (2)  DecisionThreshold [mV]" << endl
	  <<  BitErrorRatio<< "\t" <<  1e3*DecisionThreshold 
	  << endl;
 
  OutFile.close();

} // ## end GaussianReceiver::ComputeMinimumBERAndDecisionThreshold

// ###############################################################

void GaussianReceiver::WriteFilePairVectors(double * x, double *y, int dim, 
					    string OutFileName)
{

  // Adapted from ChiSquareReceiver::WriteFilePairVectors

   ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "Error: " << OutFileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   for(int jj=0;jj<dim;jj++)
     OutFile << x[jj] << "\t" << y[jj] << endl;

   OutFile.close();


}

// ############################################################

void GaussianReceiver::WriteFilePDFTimeSlices(string OutFileName)
{

   ofstream OutFile;

   OutFile.open(OutFileName.c_str(),ios::out);

   if (!OutFile)
    {
      cerr << "Error: " << OutFileName
           << "could not be created. ABORTING"<<endl;
      exit(1);
   }


  OutFile << "% (1) Current [V] (t+1) PDF of current in t-th TimeSlice"
          << endl;

  for(int vv=0;vv<NumPointsCurrent;vv++)
    {

      OutFile << CurrentVector[vv] << "\t";

      for(int tt=0;tt<NumTimeSlices;tt++)
       OutFile << PDFTimeSlices[tt][vv] << "\t";

      OutFile << endl;
    }

  OutFile.close();

}

// ############################################################

//! Receives a signal that is the sum of (1) a noise-free signal that has
//! gone through first-order PMD fiber (ie PMF) and (2) Partially polarized
//! noise. To be used when know input parameters, say from an experiment.
//! See Jan 5th 2006 "How to validate receiver formula" by J. Zweck

/*! UnitStokesPlusPlus2, UnitAverageStokesSignal 
    and UnitStokesVectorPolarizedPartNoise
    are 3x1 and have length one.
*/

void GaussianReceiver::ReceiveSignalDGD_NoisePartiallyPolarized(
     double DGD2, 
     double ClockRecoveryTime2, 
     double *UnitStokesPlusPlus2, 
     double *UnitAverageStokesSignal,
     double DOP_Signal,
     double TotalNoiseSpectralDensity,
     double DOP_Noise, 
     double * UnitStokesVectorPolarizedPartNoise,
     string OutFileID)
{

  DGD = DGD2;
  ClockRecoveryTime = ClockRecoveryTime2;
  SetTimeSlices();

  for(int ii=0;ii<3;ii++)
   StokesPlusPlus[ii] = UnitStokesPlusPlus2[ii];

  cplx InJones[2];

  InJones[0]=1.0;
  InJones[1]=0.0;

  *WDMSignalTmp = *WDMSignal;

  WDMSignal->SetInJones(0,InJones);
  WDMSignal->RegenerateSignalString();

  double P_0 = real(WDMSignal->GetAutocorrelationFunctionOptSignalX(0));
  cplx P_DGD = WDMSignal->GetAutocorrelationFunctionOptSignalX(DGD);

  cout << "P_0, P_DGD = " << P_0 << " " << P_DGD << endl;

  double AverageStokesSignal[3];

  for(int ii=0;ii<3;ii++)
    AverageStokesSignal[ii] = DOP_Signal*P_0*UnitAverageStokesSignal[ii];

  double SqAbsCplus = 0.0;

  for(int ii=0;ii<3;ii++)
    {
      SqAbsCplus += AverageStokesSignal[ii]*StokesPlusPlus[ii];
    }

 // SqAbsCplus = 0.5*(1+SqAbsCplus/P_0);
    SqAbsCplus = 0.5*(1+SqAbsCplus/P_0/DOP_Signal); // Hua bebugged this.

  if(SqAbsCplus > 1 || SqAbsCplus < 0)
    {
      cerr << "Out of bounds: SqAbsCplus = " << SqAbsCplus << endl;
      double epsilon = 1e-3;
          if (SqAbsCplus > 1+epsilon)
            {
        cerr << "ABORT";
        exit(1);
           }
          else
                {
             SqAbsCplus = 1.0;
                 cout << " Set SqAbsCplus to 1.0." << endl;
                }
   } 

  double SqAbsCminus = 1 - SqAbsCplus;

  cout << "SqAbsCplus SqAbsCminus = " 
       << SqAbsCplus << " " 
       << SqAbsCminus << endl;

  double StokesParametersPlusPlus[4];
  double StokesParametersMinusMinus[4];

  StokesParametersPlusPlus[0] = 1.0;
  StokesParametersMinusMinus[0] = 1.0;

  for(int kk = 0; kk < 3; kk++)
  {
    StokesParametersPlusPlus[kk+1] = StokesPlusPlus[kk];
    StokesParametersMinusMinus[kk+1] = -StokesParametersPlusPlus[kk+1];
  }
	  
 cplx JonesVectorPlus[2], JonesVectorMinus[2];
 cplx StokesParametersPlusMinus[4];
	  
 Stokes2Jones(StokesParametersPlusPlus,JonesVectorPlus);
 Stokes2Jones(StokesParametersMinusMinus,JonesVectorMinus);

 cout << "The following lengths should be 1.0" << endl;
 cout << "Length JonesVectorPlus = " 
      << sqrt(sq(abs(JonesVectorPlus[0])) + sq(abs(JonesVectorPlus[1])))
      << endl;
 cout << "Length JonesVectorMinus = " 
      << sqrt(sq(abs(JonesVectorMinus[0])) + sq(abs(JonesVectorMinus[1])))
      << endl;


  JonesPair2MixedStokes(JonesVectorPlus,JonesVectorMinus,
			StokesParametersPlusMinus);

  cout << "StokesParametersPlusMinus = " 
       << StokesParametersPlusMinus[0] << " "
       << StokesParametersPlusMinus[1] << " "
       << StokesParametersPlusMinus[2] << " "
       << StokesParametersPlusMinus[3] << " "
       << endl;

  cout << "Length StokesPlusMinus = "
       << sqrt(sq(abs(StokesParametersPlusMinus[1])) + 
	       sq(abs(StokesParametersPlusMinus[2])) +
	       sq(abs(StokesParametersPlusMinus[3])))
       << endl;

  cout << "Inner product real and imag parts StokesPlusMinus should be zero = "
       << real(StokesParametersPlusMinus[1])*imag(StokesParametersPlusMinus[1])
         +real(StokesParametersPlusMinus[2])*imag(StokesParametersPlusMinus[2])
         +real(StokesParametersPlusMinus[3])*imag(StokesParametersPlusMinus[3])
       << endl;

  double SqLengthRealPlusMinus = 0.0;

  //Next few lines were debugged by Hua

  for(int kk = 0; kk < 3; kk++)
    SqLengthRealPlusMinus += sq(real(StokesParametersPlusMinus[kk+1]));



  double RealTerm = 0.0;

  for(int kk = 0; kk < 3; kk++)
   RealTerm += AverageStokesSignal[kk]*real(StokesParametersPlusMinus[kk+1]);

  RealTerm = RealTerm/(2.0*SqLengthRealPlusMinus);

  double SqLengthImagPlusMinus = 0.0;

  for(int kk = 0; kk < 3; kk++)
    SqLengthImagPlusMinus += sq(imag(StokesParametersPlusMinus[kk+1]));

  double ImagTerm = 0.0;

  for(int kk = 0; kk < 3; kk++)
   ImagTerm += AverageStokesSignal[kk]*imag(StokesParametersPlusMinus[kk+1]);

  ImagTerm = ImagTerm/(2.0*SqLengthImagPlusMinus);

  cout << "RealTerm ImagTerm = " << RealTerm << " " << ImagTerm << endl;

  double Real_cPlusMinus = (real(P_DGD)*RealTerm + imag(P_DGD)*ImagTerm)/
                            sq(abs(P_DGD));
  double Imag_cPlusMinus = (-imag(P_DGD)*RealTerm + real(P_DGD)*ImagTerm)/
                            sq(abs(P_DGD));

  cplx cPlusMinus = Real_cPlusMinus + jc*Imag_cPlusMinus;

  cout << "cPlusMinus = " << cPlusMinus << endl;

  cout << "abs(cPlusMinus)/sqrt(SqAbsCplus*SqAbsCminus) should be 1.0 = "
       <<  abs(cPlusMinus)/sqrt(SqAbsCplus*SqAbsCminus) << endl;
//  cout << "However that will only be true if parameters come from a realistic"
//       << "signal ???" << endl;

  /* The following call to PrepareToDemuxChannel
     sets the number of channels of the SingleChannelSignal
     to 1, reduces the WDM BitString to a SingleChannel BitString,
     sets the carrier frequency of the only channel in the SingleChannel 
     signal and sets the center frequency of the optical filter. It does
     not do the actual filtering since we need to do that in the 
     semianalytical receiver. It also resets 
     *SingleChannelSignal = *WDM_Signal
  */

  oOptDemuxer->PrepareToDemuxChannel(DemuxChannArrayIndex);  
  oElecSignalStat->ResetOptElecFiltersForGaussianStatistics();
    
  MeanNoise = oElecSignalStat->GetMeanNoiseGaussianStatistics(
				    TotalNoiseSpectralDensity);

 double NoiseNoiseVariance = 
   oElecSignalStat->GetNoiseNoiseVarianceGaussianStatistics(
			TotalNoiseSpectralDensity,DOP_Noise);


 cout << "MeanNoise, NoiseNoiseVariance = " << MeanNoise << " " 
      << NoiseNoiseVariance << endl;

 WDMSignal->SetInJones(0,InJones);
 WDMSignal->RegenerateSignalString();

 oOptDemuxer->DemuxChannel(DemuxChannArrayIndex);

 oPhotodetector->DetectOptSignal();
 oElecFilter->FilterElecSignal();

 // Sign in next line changed by Hua

 oPhotodetector->TimeShiftSignal(-DGD/2.0);

 
  for(int tt=0;tt<NumTimeSlices;tt++)
      MeanCurrent[tt] = MeanNoise + 
	                SqAbsCplus*real(oPhotodetector->sfftPM.tPowerSignal[
						       TimeSliceIndices[tt]]);
 
 // Sign in next line changed by Hua
   
 oPhotodetector->TimeShiftSignal(DGD);

  for(int tt=0;tt<NumTimeSlices;tt++)
   MeanCurrent[tt] += SqAbsCminus*real(oPhotodetector->sfftPM.tPowerSignal[
         				               TimeSliceIndices[tt]]);

 string  OutFileName = Job + OutFileID + "MeanCurrent.dat";

  ofstream OutFile2(OutFileName.c_str(),ios::out);

  if (!OutFile2) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      OutFile2 << TimeSlices[tt] << "\t" <<  MeanCurrent[tt] <<  endl;
    }

 
  OutFile2.close();

  // For the calculation of the scalar integrals in the receiver formula,
  // we need the signal to be the one
  // at the transmitter, and moreover, we need this signal to be
  // X-polarized. Since this may not be the case we proceed as follows:

  // In case we need WDMSignal later on in the application code we 
  // first save a temporary copy of it:

  for(int ch=0;ch<WDMSignal->GetNumChannels();ch++)
    {
     WDMSignal->SetInJones(ch,0,1);
     WDMSignal->SetInJones(ch,1,0);
    }

  WDMSignal->RegenerateSignalString();

  // WDMSignal is now X-polarized and as at the transmitter

  oOptDemuxer->PrepareToDemuxChannel(DemuxChannArrayIndex);
  oElecSignalStat->ResetOptElecFiltersForGaussianStatistics(); 

for(int tt=0;tt < NumTimeSlices; tt++)
 {
 
  double I_Plus = oElecSignalStat->GetIntegralForSignalNoiseVariance(
		   SingleChannelSignal,SingleChannelSignal,0.0,0.0,
		   TimeSlices[tt]-DGD/2.0);

  double I_Minus = oElecSignalStat->GetIntegralForSignalNoiseVariance(
		   SingleChannelSignal,SingleChannelSignal,0.0,0.0,
		   TimeSlices[tt]+DGD/2.0);
  
  cplx I_PlusMinus = oElecSignalStat->GetIntegralForSignalNoiseVariance(
		   SingleChannelSignal,SingleChannelSignal,DGD/2.0,-DGD/2.0,
		   TimeSlices[tt]);

  double InnerProduct = 0.0;

  for(int ii=0;ii<3;ii++)
   {
     InnerProduct += UnitStokesVectorPolarizedPartNoise[ii]*(
     (SqAbsCplus*I_Plus - SqAbsCminus*I_Minus)*StokesPlusPlus[ii]
       + 2*real(I_PlusMinus*cPlusMinus)*real(StokesParametersPlusMinus[ii+1])
       - 2*imag(I_PlusMinus*cPlusMinus)*imag(StokesParametersPlusMinus[ii+1]));
   } // The two ii+1's used to ii's. Debugged by Hua. JZ: Should we have 
     // - sign on last term as above?? I think so.

  VarianceCurrent[tt] = NoiseNoiseVariance + 
    0.5*sq(oPhotodetector->GetConvFactOptElecPhotod())
      *TotalNoiseSpectralDensity
	*(SqAbsCplus*I_Plus + SqAbsCminus*I_Minus 
	  + DOP_Noise*InnerProduct);

 } // end for int tt loop
  
  OutFileName = Job + OutFileID + "StdDevCurrent.dat";

  ofstream OutFile(OutFileName.c_str(),ios::out);
  
  if (!OutFile) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }


  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      OutFile << TimeSlices[tt] << "\t" 
	      << sqrt(VarianceCurrent[tt]) << endl;
    }

  OutFile.close();

  *WDMSignal = *WDMSignalTmp;


  OutFileName = Job + OutFileID + "Clock.dat";

  ofstream OutFile3(OutFileName.c_str(),ios::out);

  if (!OutFile3) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  OutFile3 << "% (1) ClockTime [ps]" << endl
	   <<  1e12*ClockRecoveryTime << endl;
 
  OutFile3.close();


  if(TypeOutput == PDFS)
    {
     ComputeAveragePDFs(OutFileID);
     ComputeMinimumBERAndDecisionThreshold(OutFileID);
    }


}

