// $Id:

#include "GeneralRealMatrix.hh"

extern ofstream *LogFile;

// ############################################################

GeneralRealMatrix::GeneralRealMatrix(int NumRows2, int NumCols2)
{
 
  if(NumRows2 <= 0 || NumCols2 <= 0)
    {  
     cerr << "GeneralRealMatrix::GeneralRealMatrix: "
	  << "Matrix dimensions not positive. ABORTING." << endl;
     exit(1);
    }

  NumRows = NumRows2;
  NumCols = NumCols2;  
  Data = new double[ NumRows * NumCols ];


}

// ############################################################

GeneralRealMatrix::GeneralRealMatrix(const GeneralRealMatrix & Copy)
{
  NumRows = Copy.NumRows;
  NumCols = Copy.NumCols;

  Data = new double[ NumRows*NumCols ];

  memcpy(Data, Copy.Data, NumRows*NumCols*sizeof(double));
  
}

// ##############################################################

GeneralRealMatrix::~GeneralRealMatrix(void)
{

  delete [] Data;

}

// ##############################################################

void GeneralRealMatrix::WriteFileMatrix(string OutFileName)
{

  ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

 for(int i=1; i <= NumRows; i++)
  {
    for(int j=1; j <= NumCols; j++)
      OutFile << Get(i,j) << "\t";
    OutFile << endl;
  }

 OutFile.close();

}

// ##############################################################

void GeneralRealMatrix::SetAll(double Val)
{
    
  for(int ii=0; ii<NumRows*NumCols; ii++)
    Data[ii] = Val;
}

// #############################################################

void GeneralRealMatrix::Zero(void)
{
  SetAll(0.0);
}

// ##############################################################

void GeneralRealMatrix::Multiply(const GeneralRealMatrix & A, 
				 const GeneralRealMatrix & B)
{

  if( A.NumRows != NumRows)
    {
      cerr << "GeneralRealMatrix::Multiply" << endl
	   << "Range of A and this matrix do not agree. ABORT"
	   << endl;
      exit(1);
    }

  if( B.NumCols != NumCols)
    {
      cerr << "GeneralRealMatrix::Multiply" << endl
	   << "Domain of B and this matrix do not agree. ABORT"
	   << endl;
      exit(1);
    }

  if( A.NumCols != B.NumRows)
    {
      cerr << "GeneralRealMatrix::Multiply" << endl
	   << "Domain of A and range of B do not agree. ABORT"
	   << endl;
      exit(1);
    }

  Zero();

  for( int Row = 1; Row <= NumRows; Row++ )
    {
      for( int Col = 1; Col <= NumCols; Col++ )
	{
	  double Val = 0;
	  for( int i = 1; i <= A.NumCols;i++ )
	    {
	      Val += A.Get( Row, i ) * B.Get( i, Col );
	    }
	  Element(Row,Col) += Val;
	}
    }
  
}

//##########################################################

void GeneralRealMatrix::MultiplyTrans(const GeneralRealMatrix & A,
			              const GeneralRealMatrix & B)
{
  
  if( A.NumCols != NumRows)
    {
      cerr << "GeneralRealMatrix::MultiplyTrans" << endl
	   << "A.NumCols does not equal NumRows. ABORT"
	   << endl;
      exit(1);
    }

  if( B.NumCols != NumCols)
    {
      cerr << "GeneralRealMatrix::MultiplyTrans" << endl
	   << "Domain of B and this matrix do not agree. ABORT"
	   << endl;
      exit(1);
    }

  if( A.NumRows != B.NumRows)
    {
      cerr << "GeneralRealMatrix::MultiplyTrans" << endl
	   << "Domains of A and B do not agree. ABORT"
	   << endl;
      exit(1);
    }

  Zero();

  for( int j = 1; j <= NumCols; j++ )
    {
      for( int i = 1; i <= NumRows; i++ )
	{
	  double Val = 0.0;
	  for( int k = 1; k <= B.NumRows; k++ )
	    {
	      Val += A.Get( k, i ) * B.Get( k, j );
	    }
	  Element( i, j ) += Val;
	}
    }

}

// #################################################################


//! Produce image, Y = A * X, where A is this.

void GeneralRealMatrix::Image( const GeneralRealMatrix & X,
			       GeneralRealMatrix & Y) const
{

  if( NumCols != X.NumRows)
    {
      cerr << "GeneralRealMatrix::Image" << endl
	   << "NumCols does not equal X.NumRows. ABORT"
	   << endl;
      exit(1);
    }

  if( NumRows != Y.NumRows)
    {
      cerr << "GeneralRealMatrix::MultiplyTrans" << endl
	   << "NumRows does not equal Y.NumRows. ABORT"
	   << endl;
      exit(1);
    }

  if( X.NumCols != Y.NumCols)
    {
      cerr << "GeneralRealMatrix::MultiplyTrans" << endl
	   << "NumCols of X and Y are not equal. ABORT"
	   << endl;
      exit(1);
    }


  Y.Zero();     
  
  // Iterate over each column of X,Y
  for(int k=1; k <= X.NumCols; k++)
    {
      // Iterate over columns of the coeff. matrix (this)
      for(int j=1; j <= NumCols; j++)
	{
	  // Iterate over rows of coeff. matrix
	  for(int i=1; i <= NumRows; i++)
	    {
	      Y.Element( i, k ) += Get( i, j ) * X.Get( j, k );
	    }
	}// for j
    }// for k
  
}

// #######################################################################

void GeneralRealMatrix::Symmetrize(void)
{

  if(NumRows != NumCols)
    {
      cerr << "GeneralRealMatrix::Symmetrize" << endl
	   << "NumRows != NumCols. ABORT" << endl;
      exit(1);
    }

  for(int i=1;i<=NumRows;i++)
    for(int j=i+1;j<=NumCols;j++)
      {
	Element(i,j) = 0.5*(Get(i,j) + Get(j,i));
        Element(j,i) = Element(i,j);
      }
}

// ######################################################################

void GeneralRealMatrix::Transpose(GeneralRealMatrix & Atrans) const
{
  if(this == &Atrans)
    {
      cerr << "GeneralRealMatrix::Transpose" << endl
	   << "Atrans may not be this matrix. ABORT" << endl;
      exit(1);
    }

  if(NumRows != Atrans.NumCols || NumCols != Atrans.NumRows)
    {
      cerr << "GeneralRealMatrix::Transpose" << endl
	   << "Dimensions do not agree. ABORT" << endl;
      exit(1);
    }

  int i, j;
  
  for(j=1; j <= NumCols; j++)
    {
      for(i=1; i <= NumRows; i++)
	{
	  Atrans.Element(j,i) = Get(i,j);
	}
    }
  
}

// ################################################################

int GeneralRealMatrix::Solve(GeneralRealMatrix & X,
			     const GeneralRealMatrix & B) const
{
 
  if( NumRows != NumCols)
    {
      cerr << "GeneralRealMatrix::Solve" << endl
	   << "This matrix is not square. ABORT"
	   << endl;
      exit(1);
    }

  if( X.NumRows != NumCols)
    {
      cerr << "GeneralRealMatrix::Solve" << endl
	   << "X.NumRows != NumRows. ABORT"
	   << endl;
      exit(1);
    }

  if( B.NumRows != NumRows)
    {
      cerr << "GeneralRealMatrix::Solve" << endl
	   << "B.NumRows != NumRows. ABORT"
	   << endl;
      exit(1);
    }


  if( X.NumCols != B.NumCols)
    {
      cerr << "GeneralRealMatrix::Solve" << endl
	   << "X.NumCols != B.NumCols. ABORT"
	   << endl;
      exit(1);
    }

  int i;
  
  int info;
  int n = NumRows;
  
  double* Factors = new double[ NumRows*NumCols ];
  int*    Pivots  = new int[ NumRows ];

  for(i=0; i < NumRows*NumCols; i++)
    Factors[i] = Data[i];
  
  // Factor the coefficient matrix
  
  F77NAME(dgetrf)( &n, &n, Factors, &n, Pivots, &info );

  if(info != 0)
    {
     cerr << "GeneralRealMatrix::Solve" << endl
	  << "Factorization of system failed. ABORT." << endl;
     exit(1);
    }

  // Copy B to X 
  if( &X != &B )
    {
      for(i=0; i < B.NumRows*B.NumCols; i++)
	X.Data[i] = B.Data[i];
    }
  
  
  F77NAME(dgetrs)("N", &n, (int*)&X.NumCols, Factors,
		  &n, Pivots, X.Data, &n, &info);
  
  if(info != 0)
    {
     cerr << "GeneralRealMatrix::Solve" << endl
	  << "Solution of system failed. ABORT." << endl;
     exit(1);
    }
  
  delete[] Factors;
  delete[] Pivots;

  return info;
}

// ##################################################################

void GeneralRealMatrix::SymmetricEig(GeneralRealMatrix & EigValue,
				     GeneralRealMatrix & EigVector) const
{

  if(NumRows != NumCols)
    {
      cerr << "GeneralRealMatrix::SymmetricEig" << endl
	   << "NumRows != NumCols. ABORT" << endl;
      exit(1);

    }

  if(EigVector.NumRows != NumRows || EigVector.NumCols != NumCols )
    {
      cerr << "GeneralRealMatrix::SymmetricEig" << endl
	   << "EigVector dimensions inccorect. ABORT" << endl;
      exit(1);

    }

  if(EigValue.NumRows != NumRows || EigValue.NumCols != 1 )
    {
      cerr << "GeneralRealMatrix::SymmetricEig" << endl
	   << "EigValue dimensions inccorect. ABORT" << endl;
      exit(1);

    }

  int ispec = 1;
  char * name = "DSYTRD";
  char opts = 'U';
  int n1 = NumRows;
  int n2=-1;
  int n3=-1;
  int n4=-1;
  
  // The last argument is an int which is the length of the
  // character array name.

  int NB = F77NAME(ilaenv)(&ispec,name,&opts,&n1,&n2,&n3,&n4,6);

  //cout << "NB = " << NB << endl;

  int LengthSymmetricEigWork=max(1,(NB+2)*NumRows);

  double * SymmetricEigWork;

  SymmetricEigWork = new double[LengthSymmetricEigWork];

  // EigVector = SymmMatrix
  
  memcpy(EigVector.Data, Data, NumRows*NumCols*sizeof(double));

  char jobz = 'V';
  char uplo = 'U';
  int Info;
  int Dim = NumRows;

  // Call to LAPACK subroutine that computes all eigenvalues and, 
  // optionally, eigenvectors of a real symmetric matrix.

  // Other possibility: dsyevr

  F77NAME(dsyev)(&jobz,&uplo,&Dim,EigVector.Data,&Dim,EigValue.Data,
  	         SymmetricEigWork,&LengthSymmetricEigWork,&Info);
  
 if( Info < 0 )
   {
     cerr << "GeneralRealMatrix::SymmetricEig" << endl
	  << "Error in LAPACK routine dsyev: info = " << Info << endl
           << "(" << Info << "-th argument had an illegal value)" << endl;
      exit(1);
   }
   else if( Info > 0 )
   {
      cerr << "GeneralRealMatrix::SymmetricEig" << endl
	   << "Error in LAPACK routine dsyev: Info = " << Info << endl
           << "(algorithm failed to converge)" << endl;
      exit(1);
   }

 
 delete [] SymmetricEigWork;

}  // ## end GeneralRealMatrix::SymmetricEig

// ###################################################################

