// $Id: MachZehnder.cc,v 1.2 2008/08/08 02:25:23 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "MachZehnder.hh"

extern ofstream *LogFile;

// ############################################################

MachZehnder::MachZehnder(string InFileName, OptSignal * oOptSignal2)
{
  // This class was originally created by John Zweck.  
  // Starting 9 Sept 2003, Brian Marks has started to add a lot
  // of the functionality in order to apply Mach-Zehnder modulators
  // to analog optical fiber systems.  BM is attempting to maintain
  // backward compatibility.

ifstream InFile(InFileName.c_str(),ios::in);    
   
if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

LogFileSeparator();
*LogFile << InFileName << endl << endl;

// Ultimately, the output field one gets from the MZ, as it is set up
// now (10 Sep 2003), is
//
//   E_{out}(t) = E_0 \exp(i\omega t) *
//                cos[ V_\pi P_s(t) + (1/2)(\phi_{10} + \phi_{20}) ] *
//                \exp[ (i/2)(\phi_{10} - \phi_{20}) ]
//
// where E_0, \omega = input field amplitude and frequency
//             V_\pi = angle swing of the RF modulator (in degrees)
//         \phi_{k0} = angle corresponding to bias voltages of the LiNbO3's
//            P_s(t) = RF driving signal (0 <= P_s(t) <= 1)
//
// Note that this field is chirp-free!  
// If \phi_{10} = \phi_{20}, the final exponential is 1.

 TypeMachZehnder = typeMachZehnder(ReadInt("TypeMachZehnder"," ",1,4,
					   LOWER_AND_UPPER,MANDATORY,
					   &InFile));

Phi_10 = (pi/180.0)*ReadDouble("Phi_10"," ",0,-1,NO_BOUNDS,
			       OPTIONAL_NO_WARNING,-1000*180/pi,&InFile);
Phi_20 = (pi/180.0)*ReadDouble("Phi_20"," ",0,-1,NO_BOUNDS,
			       OPTIONAL_NO_WARNING,-1000*180/pi,&InFile);
V_pi = (pi/180.0)*ReadDouble("V_pi"," ",0,-1,NO_BOUNDS,
			     OPTIONAL_NO_WARNING,-1000*180/pi,&InFile);
V_pi1 = (pi/180.0)*ReadDouble("V_pi1"," ",0,-1,NO_BOUNDS,
			     OPTIONAL_NO_WARNING,-1000*180/pi,&InFile);
V_pi2 = (pi/180.0)*ReadDouble("V_pi2"," ",0,-1,NO_BOUNDS,
			     OPTIONAL_NO_WARNING,-1000*180/pi,&InFile);
DeltaPhaseModulator = (pi/180.0)*ReadDouble("DeltaPhaseModulator"," ",
					    0,-1,NO_BOUNDS,
					    OPTIONAL_NO_WARNING,0,&InFile);
OptModulationDepth = ReadDouble("OptModulationDepth"," ",0,1,LOWER_AND_UPPER,
				OPTIONAL_NO_WARNING,0.1,&InFile);


 if ( Phi_10 < -999 && Phi_20 < -999 ) 
   {
     if ( TypeMachZehnder == SINUSOIDAL_SIGNAL )
       {
	 //  Impose quadrature bias for chirp-free Mach-Zehnder
	 Phi_10 = 0.25*pi;
	 Phi_20 = 0.25*pi;
       }
     else
       {
	 cerr << "Error:  MachZehnder " << endl
	      << "One of Phi_10 or Phi_20 must be specified!" << endl
	      << "Aborting now!!" << endl;
	 exit(1);
       }
   }
 else if ( Phi_10 < -999 ) 
   {
     Phi_10 = Phi_20;
   }
 else if ( Phi_20 < -999 )
   {
     Phi_20 = Phi_10;
   }

 if ( V_pi < -999 && V_pi1 < -999 && V_pi2 < -999 ) 
   {
     if ( TypeMachZehnder == SINUSOIDAL_SIGNAL )
       {
	 V_pi = 0.5*asin(OptModulationDepth);
	 V_pi1 = V_pi2 = V_pi;
       }
     else
       {
	 cerr << "Error:  MachZehnder " << endl
	      << "At least one of V_pi, V_pi1, or V_pi2 must be specified!" 
	      << endl << "Aborting now!!" << endl;
	 exit(1);
       }
   }
 else if ( V_pi1 < -999 && V_pi2 < -999 )
   {
     V_pi1 = V_pi;
     V_pi2 = V_pi;
   }
 else if ( V_pi < -999 && V_pi1 < -999 )
   { 
     V_pi1 = V_pi2;
   }
 else if ( V_pi < -999 && V_pi2 < -999 )
   {
     V_pi2 = V_pi1;
   }
 else if ( V_pi1 < -999 )
   {
     V_pi1 = V_pi;
   }
 else if ( V_pi2 < -999 )
   {
     V_pi2 = V_pi;
   }



SlopeFactor = ReadDouble("SlopeFactor"," ",0,-1,NO_BOUNDS,&InFile);
ShiftAmt = ReadDouble("ShiftAmt"," ",0,-1,NO_BOUNDS,&InFile);
SuperGaussExponent = ReadInt("SuperGaussExponent"," ",0,-1,LOWER_ONLY,&InFile);
 
DoEOMFlag = ReadInt("DoEOMFlag"," ",0,1,LOWER_AND_UPPER,OPTIONAL_NO_WARNING,
                     1,&InFile);
EAMWithHump = ReadInt("EAMWithHump"," ",0,1,LOWER_AND_UPPER,
                       OPTIONAL_NO_WARNING,
                       1,&InFile);


 if(SuperGaussExponent%2) // SuperGaussExponent is odd
   {
     cerr << "MachZehnder::MachZehnder: SuperGaussExponent must be even"
	  << endl
	  << "ABORTING NOW!!"
	  << endl;

     exit(1); 

   }

InFile.close();

oOptSignal = oOptSignal2;
oBitString = oOptSignal->oBitString;

fft =  oOptSignal->fft;
sfftM = oOptSignal->sfftM;
TypeSimulation = oOptSignal->GetTypeSimulation();

qtPoints = oOptSignal->GetQtPoints();
StringLength = oOptSignal->GetStringLength();
TimeWindow = oOptSignal->GetTimeWindow();
FreqWindow = oOptSignal->GetFreqWindow();
DeltaTime = oOptSignal->GetDeltaTime();
DeltaFreq = oOptSignal->GetDeltaFreq();
NumChannels = oOptSignal->GetNumChannels();
CenterFreq = oOptSignal->GetCenterFreq();
OptPowerChann = oOptSignal->GetOptPowerChann();
BitRateChann =  oOptSignal->GetBitRateChann();
BitLength =  oOptSignal->GetBitLength();
RandomPolarizationFlag = oOptSignal->GetRandomPolarizationsFlag();

 SlopeFactor *=  BitRateChann; // Normalization is appropriate


sfftPM.tPowerSignal = NewCplx(qtPoints);
sfftPM.fPowerSignal = NewCplx(qtPoints);

if(!sfftPM.tPowerSignal || !sfftPM.fPowerSignal)
  {
    *LogFile << "Malloc error in MachZehnder" << endl
	    << "Aborting Now" << endl << flush;

    exit(1);

  }

oElecFilter = new ElecFilter(InFileName,fft,sfftPM);

}

// ##############################################################

void MachZehnder::GeneratePerfectRectBinaryElecData(int ChannelArrayIndex)
{

  // FIX: Make this more like the new RegenerateSignlaString method!!!

   int ii,jj,kk; 
   double Value;
   
    for (ii = 0; ii < qtPoints; ii++) 
      {
        sfftPM.tPowerSignal[ii] = 0.0;
        sfftPM.fPowerSignal[ii] = 0.0;

      } // ## end for ii-loop


   for (ii = 0; ii < qtPoints; ii++) 
   {

     // Comments by JZ (I think this is correct):
     // (1) Given an index ii we want to know which 
     // bit slot kk the index ii is in, and also
     // what the index jj of the center of this kk-th bit slot is.
     // The reason jj is the index of the center of the bit slot
     // is that 
     // jj = (int) (-1 + (qtPoints/StringLength)*(SomeInteger + 0.5) 
     //
     // and qtPoints/StringLength = Number of points in a bit slot
     // (I'm assuming this ratio is an integer for simplicity).


     // (2) The kk-th bit slot goes from 
     //  ii = kk*M to kk*M + M-1
     // where M =  qtPoints/StringLength is as above.


      jj = (int)  (-1 + qtPoints/StringLength 
                 *(floor( (1.*ii)/(qtPoints/StringLength)) + 0.5) );
      kk = (int) (jj/(qtPoints/StringLength)); 

      // ## It can happen that kk == StringLength in which case 
      // ## it should be set to kk = 0 to be in range for the
      // ## if (oBitString.GetBit(ch,kk)) test (JZ)

      if(kk >= StringLength)
	kk -= StringLength;
     
   

      if(oBitString->GetBit(ChannelArrayIndex,kk))
	 Value = 1.0;
      else
         Value = 0.0;

     sfftPM.tPowerSignal[ii] = Value;
                                         
   } // ## end for ii-loop
   
fft->FFT(sfftPM.tPowerSignal,sfftPM.fPowerSignal);  

} // ## end GeneratePerfectRectBinaryElecData

// ############################################################

void MachZehnder::GenerateSmoothedRectBinaryElecDataArctanh
                  (int ChannelArrayIndex)
{
   int ii,jj,kk; 
   int kkm1,kkp1;
   double Value;
   
    for (ii = 0; ii < qtPoints; ii++) 
      {
        sfftPM.tPowerSignal[ii] = 0.0;
        sfftPM.fPowerSignal[ii] = 0.0;

      } // ## end for ii-loop


   for (ii = 0; ii < qtPoints; ii++) 
   {

     // Comments by JZ (I think this is correct):
     // (1) Given an index ii we want to know which 
     // bit slot kk the index ii is in, and also
     // what the index jj of the center of this kk-th bit slot is.
     // The reason jj is the index of the center of the bit slot
     // is that 
     // jj = (int) (-1 + (qtPoints/StringLength)*(SomeInteger + 0.5) 
     //
     // and qtPoints/StringLength = Number of points in a bit slot
     // (I'm assuming this ratio is an integer for simplicity).


     // (2) The kk-th bit slot goes from 
     //  ii = kk*M to kk*M + M-1
     // where M =  qtPoints/StringLength is as above.


      jj = (int)  (-1 + qtPoints/StringLength 
                 *(floor( (1.*ii)/(qtPoints/StringLength)) + 0.5) );
      kk = (int) (jj/(qtPoints/StringLength)); 

      // ## It can happen that kk == StringLength in which case 
      // ## it should be set to kk = 0 to be in range for the
      // ## if (oBitString.GetBit(ch,kk)) test (JZ)

      if(kk >= StringLength)
	kk -= StringLength;
     
      if(ii <= jj)
	{
          if(kk==0)
	    kkm1 = StringLength-1;
          else
            kkm1 = kk-1;

          if(oBitString->GetBit(ChannelArrayIndex,kk) &&
	     oBitString->GetBit(ChannelArrayIndex,kkm1)) // 1,1
	       Value = 1.0;
          else
	    {
             if(!oBitString->GetBit(ChannelArrayIndex,kk)&&
                !oBitString->GetBit(ChannelArrayIndex,kkm1)) // 0,0
                 Value = 0.0;
             else
	     {
               if(oBitString->GetBit(ChannelArrayIndex,kk)&&
                  !oBitString->GetBit(ChannelArrayIndex,kkm1)) // 0,1
	        {
                  Value = 
		       0.5*
		(1+tanh(SlopeFactor*(ii*DeltaTime - (kk+ShiftAmt)*BitLength)));
                  
	        } // # end 1,0 if 
	       else // 1,0
	        {
		  Value = 1 -
	 0.5*(1+ tanh(SlopeFactor*(ii*DeltaTime - (kk-ShiftAmt)*BitLength)));

		
	        }
	     } // end else
	    }//end else

	}// end if ii<=jj


        if(ii > jj)
	 {
          if(kk==StringLength-1)
	    kkp1 = 0;
          else
            kkp1 = kk+1;

          if(oBitString->GetBit(ChannelArrayIndex,kk) &&
	     oBitString->GetBit(ChannelArrayIndex,kkp1)) // 1,1
	       Value = 1.0;
          else
	    {
             if(!oBitString->GetBit(ChannelArrayIndex,kk)&&
                !oBitString->GetBit(ChannelArrayIndex,kkp1)) // 0,0
                 Value = 0.0;
             else
	     {
               if(oBitString->GetBit(ChannelArrayIndex,kk)&&
                  !oBitString->GetBit(ChannelArrayIndex,kkp1)) // 1,0
	        {

                 Value = 
		          1-0.5*(1+
	       tanh(SlopeFactor*(ii*DeltaTime - (kk+1-ShiftAmt)*BitLength)));

	        } // # end 1,0 if 
	       else // 0,1
	        {
                 Value = 
		        0.5*(1+
		 tanh(SlopeFactor*(ii*DeltaTime - (kk+1+ShiftAmt)*BitLength)));
	        }
	     } // end else
	    }//end else

	}// end if ii>jj   


      sfftPM.tPowerSignal[ii] = Value;
                                         
   } // ## end for ii-loop
   
fft->FFT(sfftPM.tPowerSignal,sfftPM.fPowerSignal);    

} // ## end GenerateSmoothedRectBinaryElecData

// #########################################################

void MachZehnder::GenerateSmoothedRectBinaryElecDataSuperGauss
                  (int ChannelArrayIndex)
{
 for(int ii = 0; ii < qtPoints; ii++) 
      {
        sfftPM.tPowerSignal[ii] = 0.0;
        sfftPM.fPowerSignal[ii] = 0.0;

      } // ## end for ii-loop

for(int ii = 0; ii < qtPoints; ii++)
  {

    double Value;

    double Time = ii*DeltaTime;
    int BitSlot = int(floor(Time/BitLength));

    // Claim: 0 <= BitSlot < StringLength

    int LeftBitSlot = BitSlot - 1;
    int RightBitSlot = BitSlot + 1;

    if(LeftBitSlot < 0)
      LeftBitSlot += StringLength;

    if(RightBitSlot >= StringLength)
      RightBitSlot -= StringLength;

    // Work out whether Time is in left or right half of the bit slot.

    if(Time - BitSlot*BitLength < 0.5*BitLength) // ii is in left half
      {
        int BitValue = oBitString->GetBit(ChannelArrayIndex,BitSlot);
        int LeftBitValue = oBitString->GetBit(ChannelArrayIndex,LeftBitSlot);

        if(LeftBitValue == 1 && BitValue == 1)
	  Value = 1;
        if(LeftBitValue == 0 && BitValue == 0)
	  Value = 0;
        if(LeftBitValue == 1 && BitValue == 0)
          Value = SuperGauss(Time - (BitSlot-0.5)*BitLength);
        if(LeftBitValue == 0 && BitValue == 1)
          Value = SuperGauss(Time - (BitSlot+0.5)*BitLength);
        
      } // ## end ii is in left half
    else // ii is in right half
      {
        int BitValue = oBitString->GetBit(ChannelArrayIndex,BitSlot);
        int RightBitValue = oBitString->GetBit(ChannelArrayIndex,RightBitSlot);

        if(BitValue == 1 && RightBitValue == 1)
	  Value = 1;
        if(BitValue == 0 && RightBitValue == 0)
	  Value = 0;
        if(BitValue == 1 && RightBitValue == 0)
          Value = SuperGauss(Time - (BitSlot+0.5)*BitLength);
        if(BitValue == 0 && RightBitValue == 1)
          Value = SuperGauss(Time - (BitSlot+1.5)*BitLength);

      }// ## end ii is in right half

    
     sfftPM.tPowerSignal[ii] = Value;
  
  } // ## end for ii-loop
   
fft->FFT(sfftPM.tPowerSignal,sfftPM.fPowerSignal);  

} // ## end GenerateSmoothedRectBinaryElecDataSuperGauss

// ######################################################

double MachZehnder::SuperGauss(double Time)
{

  while(Time < -0.5*TimeWindow)
    Time += TimeWindow;
  while(Time >= 0.5*TimeWindow)
    Time -= TimeWindow;

  // -0.5*TimeWindow <= Time < 0.5*TimeWindow

  return pow(2,-pow(Time/(0.5*BitLength), SuperGaussExponent));


} // ## end SuperGauss

// #########################################################

void MachZehnder::GenerateSinusoidalElecSignal(int ChannelArrayIndex)
{

  // GenerateSinusoidalElecSignal by Brian Marks
  // This function generates a sinusoidal electrical signal 
  // that can be used to drive the Mach-Zehnder interferometer.
  // The period of the sinusoidal drive is that of BitLength, and 
  // 0 <= P_s(t) <= 1.

  for(int ii = 0; ii < qtPoints; ii++) 
      {
        sfftPM.tPowerSignal[ii] = 0.0;
        sfftPM.fPowerSignal[ii] = 0.0;

      } // ## end for ii-loop

  for(int ii = 0; ii < qtPoints; ii++)
  {

    double Value;

    double Time = ii*DeltaTime;
    //    int BitSlot = floor(Time/BitLength);

    //    Value = 0.5*(1.0 + cos(2.0*pi*Time/BitLength));
    Value = sin(2.0*pi*Time/BitLength);
    
    sfftPM.tPowerSignal[ii] = Value;

  } // ## end for ii-loop
   
  fft->FFT(sfftPM.tPowerSignal,sfftPM.fPowerSignal);  

} // ## end GenerateSinusoidalElecSignal

// #########################################################

void MachZehnder::FilterBinaryElecData(void)
{

   oElecFilter->FilterElecSignal();

} // ## end FilterBinaryData

// #########################################################

void MachZehnder::MZModulator(void)
{

  // Written by John Zweck, Dec 2000 - Jan 2001

  // This method is used to construct the "Ciena" NRZ source.
  // It seems OK to use GeneratePerfectRectangularWave
  // to generate a perfect rectangular wave which is then 
  // filtered by a fourth order Bessel filter. The 
  // electrical current so obtained is then used to drive a 
  // Mach-Zehnder Modulator which takes incoming CW light
  // and modulates it using this current.

  // For perfect extinction ratio one should use 
  // V_pi = -90 degrees and Phi_10 = Phi_20 = 90 degrees

  // For an extinction ratio of -20 dB use 
  // V_pi = -84.262 degrees and Phi_10 = Phi_20 = 84.262 degrees
  


int ii;

double ChannelPhaseOffset;
double Phi_1,Phi_2;
cplx Envelope;

for (ii = 0; ii < qtPoints; ii++) 
 {
    sfftM.tOptSignalX[ii] = 0.0;
    sfftM.tOptSignalY[ii] = 0.0;
    sfftM.fOptSignalX[ii] = 0.0;
    sfftM.fOptSignalY[ii] = 0.0;

 } // ## end for ii-loop


for(int ch=0;ch<NumChannels;ch++)
 {

   // First set the Jones vector for the channel.

   cplx Jones[2];
   

   if(RandomPolarizationFlag)
     {
      oOptSignal->GetRandomJones(Jones);
     }
   else
     {
       Jones[0] = oOptSignal->GetInJones(ch,0);
       Jones[1] = oOptSignal->GetInJones(ch,1);
     }

   switch ( TypeMachZehnder ) 
     {
     case BINARY_DATA_PERFECT_RECT:
       GeneratePerfectRectBinaryElecData(ch);
       break;

     case BINARY_DATA_ARCTANH_SMOOTHING:
       GenerateSmoothedRectBinaryElecDataArctanh(ch);
       break;

     case BINARY_DATA_SUPERGAUSS_SMOOTHING:
       GenerateSmoothedRectBinaryElecDataSuperGauss(ch);  
       break;

     case SINUSOIDAL_SIGNAL:
       GenerateSinusoidalElecSignal(ch);
       break;
     } // end switch

   FilterBinaryElecData();

  for (ii = 0; ii < qtPoints; ii++) 
   {

  // ## Compute the Phase Offset for current channel

      // ## We need the 2*pi factor to convert from Hz = cycles per second to
      // ## Radians per second

  ChannelPhaseOffset = 2*pi*(oOptSignal->GetFrequency(ch) - CenterFreq)*
                       (ii)*DeltaTime;

  Phi_1 = V_pi1*real(sfftPM.tPowerSignal[ii]) + Phi_10;

  Phi_2 = - V_pi2*real(sfftPM.tPowerSignal[ii]) - Phi_20;
  

  Envelope = 0.5*(exp(jc*Phi_1) + exp(jc*Phi_2));


  sfftM.tOptSignalX[ii] += Jones[0]*sqrt(OptPowerChann)*
                           Envelope*exp(-jc*ChannelPhaseOffset); 

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] += Jones[1]*sqrt(OptPowerChann)*
                              Envelope*exp(-jc*ChannelPhaseOffset); 

   } // ## end for ii-loop
   } // ## end for ch-loop


oOptSignal->Time2Freq();

} // ## end MZModulator

// ########################################################

void MachZehnder::WriteFileTimePower(string outFile)
{

   ofstream OutFile(outFile.c_str(),ios::out);
   
   if (!OutFile) 
    {
      *LogFile << "Error in MachZehnder: " << outFile 
              << "could not be created."<<endl << flush;
      exit(1);
    }

   for (int ii = 0; ii < qtPoints; ii++) 
      OutFile <<  ii*DeltaTime  << "   " 
              << abs(sfftPM.tPowerSignal[ii]) << endl;     

  OutFile.close();
}

// ###############################################################

void MachZehnder::WriteFileFreqPower(string outFile)
{

   ofstream OutFile(outFile.c_str(),ios::out);

   if (!OutFile) 
    {
      *LogFile << "Error in MachZehnder: " << outFile 
              << "could not be created."<<endl << flush;
      exit(1);
   }


  fft->fftshift(sfftPM.fPowerSignal);

 for (int ii = 0; ii < qtPoints; ii++) 
   {
      // The DC frequency is dropped out
     //if (ii != qtPoints/2) 
       {  
         OutFile <<  ii*DeltaFreq - FreqWindow/2  << "   " 
                 << abs(sfftPM.fPowerSignal[ii])/sq(DeltaFreq)
                 << endl;   
      }           
   }   

   fft->ifftshift(sfftPM.fPowerSignal);                 
 
  OutFile.close();
}

// ##############################################################

double MachZehnder::EAM34PulseShape(double TimeRelPulseCenter)
{

double PulseShape;

double time = 1e12*TimeRelPulseCenter+50; // Want 0 < time < 100 in ps
 
 
   if(EAMWithHump && (time > 65 || time < 35))
   {
     PulseShape = EAMPoly(time);
   }
  else
    {
       PulseShape = exp(-sq( TimeRelPulseCenter 
                               *2*sqrt(log(2.))/oOptSignal->GetFWHM_Pulse() ));
    }
   

return sqrt(PulseShape);


} // ## end MachZehnder::EAM34PulseShape

// ######################################################

double MachZehnder::EAMPoly(double time)
{

  // time in ps in range 0 < time < 100.
  
  // For poly we need it in range 50 < time < 150

  

  if(time < 50)
    time += 100;

  double mean = 99.26190476;
  double sigma = 22.40626734;

  double x = (time - mean)/sigma;  
  
 double p[] =
{-27.20596079 , 
 4.79367642 , 
 192.7913544 , 
 -10.95093374 , 
 -419.2086537 , 
 -65.38333651 , 
 391.526997 , 
 214.9023855 , 
 -163.081588 , 
 -212.9002355 , 
   123.2396385 };
 
 double y;
  
 y = (((((((((p[0]*x+p[1])*x + p[2])*x + p[3])*x + p[4])*x + p[5])*x + p[6])*x
      + p[7])*x + p[8])*x + p[9])*x + p[10];


 y = y/1800.0;

 return y;


} // ## end MachZehnder::EAMPoly(double time)

// #################################################

void MachZehnder::PhaseModulator(void)
{

int ii;

double ChannelPhaseOffset;
double Phi;
cplx Envelope;

for (ii = 0; ii < qtPoints; ii++)
 {
    sfftM.tOptSignalX[ii] = 0.0;
    sfftM.tOptSignalY[ii] = 0.0;
    sfftM.fOptSignalX[ii] = 0.0;
    sfftM.fOptSignalY[ii] = 0.0;

 } // ## end for ii-loop

for(int ch=0;ch<NumChannels;ch++)
 {

   // First set the Jones vector for the channel.

   cplx Jones[2];


   if(RandomPolarizationFlag)
     {
      oOptSignal->GetRandomJones(Jones);
     }
   else
     {
       Jones[0] = oOptSignal->GetInJones(ch,0);
       Jones[1] = oOptSignal->GetInJones(ch,1);
     }

   switch ( TypeMachZehnder ) 
     {
     case BINARY_DATA_PERFECT_RECT:
       GeneratePerfectRectBinaryElecData(ch);
       break;
  
     case BINARY_DATA_ARCTANH_SMOOTHING:
       GenerateSmoothedRectBinaryElecDataArctanh(ch);
       break;
  
     case BINARY_DATA_SUPERGAUSS_SMOOTHING:
       GenerateSmoothedRectBinaryElecDataSuperGauss(ch);
       break;

     case SINUSOIDAL_SIGNAL:
       GenerateSinusoidalElecSignal(ch);
       break;
     } // end switch

   FilterBinaryElecData();

  for (ii = 0; ii < qtPoints; ii++)
   {
 
  // ## Compute the Phase Offset for current channel
    
      // ## We need the 2*pi factor to convert from Hz = cycles per second to
      // ## Radians per second

  ChannelPhaseOffset = 2*pi*(oOptSignal->GetFrequency(ch) - CenterFreq)*
                       (ii)*DeltaTime;


  Phi = V_pi*real(sfftPM.tPowerSignal[ii]) + Phi_10;

  Envelope = exp(jc*Phi);

  sfftM.tOptSignalX[ii] += Jones[0]*sqrt(OptPowerChann)*
                           Envelope*exp(-jc*ChannelPhaseOffset);

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] += Jones[1]*sqrt(OptPowerChann)*
                              Envelope*exp(-jc*ChannelPhaseOffset);

   } // ## end for ii-loop
   } // ## end for ch-loop


oOptSignal->Time2Freq();

} // ## end PhaseModulator

