#include "MatrixTools.hh"
#include <iostream.h>

/*

  // Testing sequence: Successful, 31 Jan 01

  double AA[16] = 
  {1.0, 2.0, 3.0, 4.0,
  -5.0, 6.0, 7.0, 8.0, 
  -2.0, -3.0, 4.0, -1.0,
  4.0, 6.0, -2.0, 3.0 };
  double *A2[4] = {AA, AA+4, AA+8, AA+12};

  double BB[16] = 
  {6.0, 2.0, 9.0, 1.0,
  4.0, -2.0, 3.0, 1.0, 
  7.0, 4.0, 2.0, -6.0,
  3.0, 1.0, -5.0, 7.0 };
  double *B2[4] = {BB, BB+4, BB+8, BB+12};

  double RR[16] = 
  {6.0, 2.0, 9.0, 1.0,
  4.0, -2.0, 3.0, 1.0, 
  7.0, 4.0, 2.0, -6.0,
  3.0, 1.0, -5.0, 7.0 };
  double *R2[4] = {RR, RR+4, RR+8, RR+12};

  dense_matrix_product(A2, B2, R2, true, 4);
  dense_matrix_scale(R2, 4, -3.0);
  add_to_principal_diag(R2, 4, -17.0);

  for(int a1=0; a1<4; a1++)
  {
  for(int b1=0; b1<4; b1++)
  cout << (R2[a1][b1]) << "\t";
  cout << endl;
  } 
  exit(0);

*/

// ################################################################

void CopyMatrix(double ** LHS_Matrix, double ** RHS_Matrix, int dim)
{
  double *p_RHS = &(RHS_Matrix[0][0]);
  const double *const p_RHS_lim = &(RHS_Matrix[dim-1][dim-1]);

  double *p_LHS = &(LHS_Matrix[0][0]);
  const double *const p_LHS_lim = &(LHS_Matrix[dim-1][dim-1]);


  for(; p_LHS <= p_LHS_lim; p_LHS++, p_RHS++)
    *p_LHS = *p_RHS; 
}

// ###########################################################

//! Computes the matrix product R = A * B or R = A * B^T depending
//! on whether the flag Transpose_B is 0 or 1.

/*! A,B, and R are all real matrices of dimension dim x dim
    where indices run from 0 to dim-1. 
    It is assumed that all matrices are double 2D arrays with 
    M[i][j] being the (row,col) = (i,j) of the Matrix M. 
*/

void DenseMatrixProduct(double **A, double **B, double **R,
                        bool Transpose_B, bool R_is_symmetric, int dim)

{
  for(int i=0; i < dim; i++) // row by row
  {

    // if we already know that R[][] will be symmetric,
    // only compute lower half (plus principal diagonal)

    const int j_lim = (R_is_symmetric)? i:dim-1;

    double *p_R = &(R[i][0]);  // R[row][column]
    double *const p_A_start = &(A[i][0]);
    double *const p_A_lim   = &(A[i][dim-1]);
    double *p_B_start = &(B[0][0]);

    if(!Transpose_B)
    {
      for(int j=0; j <= j_lim; j++, p_R++, p_B_start++)
      {
	register double *p_A = p_A_start;
	register double *p_B = p_B_start;
	register double sum_l = 0.0;
	for( ;p_A <= p_A_lim; p_A++, p_B+=dim)
	  sum_l += (*p_A) * (*p_B); // scalar product
	*p_R = sum_l;
      }
    }
    else // Transpose_B
    {
      for(int j=0; j <= j_lim; j++, p_R++, p_B_start+=dim) // row by row in B
      {
	register double *p_A = p_A_start;
	register double *p_B = p_B_start;
	register double sum_l = 0.0;
	for( ;p_A <= p_A_lim; p_A++, p_B++)  
	  sum_l += (*p_A) * (*p_B); // scalar product
	*p_R = sum_l;
      }
    }
  } // for(i)

  if(R_is_symmetric)
    CopyLowerToUpperHalf(R,dim);
}

// ##############################################################

//! Computes the matrix product R = A*B*A^T or R = A^T*B*A depending
//! on whether the flag Transpose_A is 0 or 1, respectively

/*! A,B,R, and Work are all real matrices of dimension dim x dim
    where indices run from 0 to dim-1. 
    It is assumed that all matrices are double 2D arrays with 
    M[i][j] being the (row,col) = (i,j) of the Matrix M. 
*/

void DenseMatrixSimilarityProduct(double **A, double **B, double **R,
				  double ** Work, bool Transpose_A, 
				  bool R_is_symmetric,int dim)
{

  // Work = B*A when Transpose_A = 1
  // Work = B*A^T when Transpose_A = 0

 DenseMatrixProduct(B,A,Work,!Transpose_A,0,dim);

 // R = A^T*Work when Transpose_A = 1
 // R = A*Work when Transpose_A = 0

  if(Transpose_A)
   TransposeMatrix(A,dim);

  DenseMatrixProduct(A,Work,R,0,R_is_symmetric,dim);

  if(Transpose_A)
   TransposeMatrix(A,dim);

}

// ###############################################################

//! Multiply all elements of the dim x dim double matrix A by the double scale.

/*! A is a  real matrix of dimension dim x dim where
    indices run from 0 to dim-1. 
    A is a double 2D array with 
    A[i][j] being the (row,col) = (i,j) of the Matrix A. 
*/


void DenseMatrixScale(double **A, int dim, double scale)
{
  double *p_A = &(A[0][0]);
  const double *const p_A_lim = &(A[dim-1][dim-1]);

  for(; p_A <= p_A_lim; p_A++)
    *p_A *= scale;
}

// #####################################################


//!  Copies the lower half of the matrix A to the upper half, so that the
//!  resulting matrix is symmetric.
 
/*!  A: real matrix of dimension dim x dim
     indices run from 0 to dim-1
     A[row][column]
*/



void CopyLowerToUpperHalf(double **A, int dim)
{
  for(int i=1; i< dim; i++)
  {
    double *p_A1 = &(A[i][0]); // go along row
    const double *const p_A1_end = &(A[i][i]);
    double *p_A2 = &(A[0][i]); // go along column
    for(; p_A1 < p_A1_end; p_A1++, p_A2 +=dim)
      *p_A2 = *p_A1;
  }
}

// #################################################

void SymmetrizeMatrix(double **A, int dim)
{

  for(int i=0;i<dim;i++)
    for(int j=i+1;j<dim;j++)
      {
	A[i][j] = 0.5*(A[i][j]+	A[j][i]);
        A[j][i] = A[i][j];
      }

}

// ###############################################

void TransposeMatrix(double **A, int dim)
{

  for(int i=0;i<dim;i++)
    for(int j=i+1;j<dim;j++)
      {
        double temp = A[i][j];
	A[i][j] = A[j][i];
	A[j][i] = temp;
      }
}

// ##################################################

//! Sets all entries of Matrix A to 0.

/*! A: real matrix of dimension dim x dim
    indices run from 0 to dim-1
*/

void ZeroMatrix(double **A, int dim)
{
  double *p_A = &(A[0][0]);
  const double *const p_A_lim = &(A[dim-1][dim-1]);
  for(; p_A <= p_A_lim; p_A++)
    *p_A = 0.0;
}

//###################################################

// A: real matrix of dimension dim x dim
// indices run from 0 to dim-1

//! Filters the matrix A with filter.

/*!

   If A[i][j] is the (i,j) entry of A, and filter[i] is the i-th
   element of filter, then the filtering of A is defined to be

   A_new[i][j] = A_old[i][j]*filter[i]*filter[j]

   where * is regular multiplication of numbers.

 */

void DenseMatrixFilter(double **A, int dim, double *filter)
{
  const double *filter_i = &(filter[0]);

  double *p_A = &(A[0][0]);

  for(int i=0; i < dim; i++, filter_i++)
  {
    const double *const p_A_lim = &(A[i][dim-1]); // end of this row
    const double *filter_j = &(filter[0]);

    for(; p_A <= p_A_lim; p_A++, filter_j++)
      *p_A *= (*filter_i) * (*filter_j);

  }
}

// ########################################################

//! Writes the Matrix scaling*A of size dim x dim to the file F  

/*! The i-th row of the file contains the i-th row of A */

void WriteMatrix(double **A, int dim, ofstream *F, double scaling)
{
  F->precision(4);
  double *p_A = &(A[0][0]);

  for(int i=0; i < dim; i++)
  {
    for(int j=0; j < dim; j++, p_A++)
      *F << scaling*(*p_A) << "\t";
    *F << endl;
  }
}

// ########################################################

//! Reads in the Matrix scaling*A of size dim x dim from the file F  
//! that was previously written using WriteMatrix

/*! The i-th row of the file contains the i-th row of A */

void ReadMatrix(double **A, int dim, ifstream *F, double scaling)
{
  F->precision(4);
  double *p_A = &(A[0][0]);

  for(int i=0; i < dim; i++)
  {
    for(int j=0; j < dim; j++, p_A++)
      {
        double Temp;

	*F >> Temp;
	*p_A = Temp/scaling;
      }      
    }
}

// ######################################################

//! Computes the matrix-vector product Out = A*In where In and Out are vectors
//! and A is square of size dim x dim

void DenseMatrixVectorProduct(double * Out, double **A, double * In, int dim)
{
  for(int jj=0; jj < dim; jj++)
  {
    double *p_A = &(A[jj][0]); // beginning of row jj
    const double *const p_A_lim = &(A[jj][dim-1]); // end of row jj
    const double *in_ptr = &(In[0]);
    register double accum = 0.0;
    for(; p_A <= p_A_lim; p_A++, in_ptr++)
      accum += (*p_A)*(*in_ptr);

    Out[jj] = accum;
  }
}

// #######################################################

//! Computes the matrix-vector product Out = In*A where In and Out are vectors
//! and A is square of size dim x dim

void DenseVectorMatrixProduct(double *Out, double *In, double **A, int dim)
{
  for(int j=0; j < dim; j++)
  {
    double *p_A = &(A[0][j]); // beginning of this column
    const double *const p_A_lim = &(A[dim-1][j]); // end of this column
    const double *in_i = &(In[0]);
    register double accum = 0.0;
    for(; p_A <= p_A_lim; p_A+=dim, in_i++)
      accum += (*in_i) * (*p_A);

    Out[j] = accum;
  }
}

// #####################################################

//! Adds the double add to the principal diagonal of Matrix A.

void AddToPrincipalDiag(double **A, int dim, double add)
{
  double *p_A = &(A[0][0]);
  const double *const p_A_lim = &(A[dim-1][dim-1]); // end of this row

  for(; p_A <= p_A_lim; p_A += dim+1)
    *p_A += add;

}

// #####################################################

//! Adds the array ScalarMultiple*V, where V is a vector 
//! to the principal diagonal of Matrix A.

void AddVectorToPrincipalDiag(double **A, double *V, 
			      double ScalarMultiple,int dim)
{
  double *p_A = &(A[0][0]);
  const double *const p_A_lim = &(A[dim-1][dim-1]); // end of this row

  for(int kk=0; p_A <= p_A_lim; kk++, p_A += dim+1)
    *p_A += ScalarMultiple*V[kk];

}

// ################################################################

//! Returns the double
//! (aa,bb)/(bb,bb) for aa, bb complex vectors of length len
//! where (aa,bb) := Re[conj(aa) * bb] = Re(aa)*Re(bb) + Im(aa)*Im(bb)

double ScalarProductNormalized(cplx *aa, cplx *bb, int len)
{
  double bb_norm = 0.0;
  double ab = 0.0;
  const cplx *const aa_end = aa + len;

  cplx *aa_p = aa;
  cplx *bb_p = bb;

  for(;aa_p < aa_end; aa_p++, bb_p++)
  {
    bb_norm += norm(*bb_p);

    // Re[conj(field)*noise] = Re[field*conj(noise)] =
    //  1/2 * (conj(field)*noise + field*conj(noise)) = Re Re + Im Im

    ab += real(*aa_p) * real(*bb_p) + imag(*aa_p) * imag(*bb_p);
  }
  return ab / bb_norm;
}

// #####################################################################

//! Returns the angle between complex vectors aa and bb of length len
//! using the inner product 
//! (aa,bb) := Re[conj(aa) * bb] = Re(aa)*Re(bb) + Im(aa)*Im(bb)

/*! Computes acos((aa,bb)/(|bb|*|aa|)) for aa, bb complex vectors of length len
    where (aa,bb) := Re[conj(aa) * bb] = Re(aa)*Re(bb) + Im(aa)*Im(bb)
*/

double VectorAngle(cplx *aa, cplx *bb, int len)
{
  double aa_norm = 0.0, bb_norm = 0.0;
  double ab = 0.0;
  const cplx *const aa_end = aa + len;

  cplx *aa_p = aa;
  cplx *bb_p = bb;
  for(;aa_p < aa_end; aa_p++, bb_p++)
  {
    aa_norm += norm(*aa_p);
    bb_norm += norm(*bb_p);
    ab += real(*aa_p) * real(*bb_p) + imag(*aa_p) * imag(*bb_p);
  }
  return acos(ab / sqrt(aa_norm*bb_norm));
}

// ####################################################################

//! Performs Gram-Schmidt orthogonalization of num vectors of length len
//! where vec[i][] is the i-th vector 
//! i.e., vec is a cplx array of dim num x len
//! alpha[] must be a double array of length (num^2-num)/2
//! Can start orthogonalization proceedure with the k1-th vector
//! instead of the 1-st vector if you like. (k1 is an optional argument).
//! Uses the inner product 
//! (aa,bb) := Re[conj(aa) * bb] = Re(aa)*Re(bb) + Im(aa)*Im(bb)

/*! 
  Input: Set of num vectors of length len that span a linear space

  Output: Orthogonalized set of num vectors in vec. vec[0][]
   is untouched, the component of vec[1][] that is parallel
   to vec[0][] is removed from vec[1][], then the component of
   vec[2][] that is parallel to vec[0][] is removed from vec[2][], then
   the component of vec[2][] that is parallel to vec[1][] is removed etc...
 
  Method: Schmidt's orthogonalization procedure, see
  "Taschenbuch der Mathematik", p. 147
  alpha[] must be a double array of length (num^2-num)/2
  that will contain the normalized scalar products
    (a_2,b_1)/(b_1,b_1), (a_3,b_1)/(b_1,b_1), (a_3,b_2)/(b_2,b_2),
    (a_4,b_1)/(b_1,b_1), (a_4,b_2)/(b_2,b_2), (a_4,b_3)/(b_3,b_3),...
 
  vec[][]: cplx array of dim num x len
  k1: optional argument, start index (index of b_i to start with)
 
  Successfully tested 25 Feb 2001 R.H. for vectors of length 3:
  vec[0][0] = 2.0; vec[0][1] = 3.0; vec[0][2] = -1.0;
  vec[1][0] = -5.0; vec[1][1] = 7.0; vec[1][2] = -3.0;
  vec[2][0] = 1.0; vec[2][1] = 2.0; vec[2][2] = 9.0;
  schmidt_orthogonalizer(vec, 3, 3, alpha);

*/


void SchmidtOrthogonalizer(cplx **vec, int num, int len,
                           double *alpha)
{
SchmidtOrthogonalizer(vec,num,len,alpha,1);
}

// ##############################################################

void SchmidtOrthogonalizer(cplx **vec, int num, int len, 
                           double *alpha, int k1=1)
{
  int j=0; // counts scalar products

  for(int k=k1; k < num; k++) // loop over out[] vectors
  {
    int i;
    for(i=0; i < k; i++)
      alpha[j++] = ScalarProductNormalized(vec[k], vec[i], len);

    for(i=0; i < k; i++)
    {
      const double help = alpha[j-k+i];
      // this happens if a vector is zero. Just skip in this case.
      if(isnan(help)) continue;
      cplx *vec_k = &(vec[k][0]);
      cplx *vec_i = &(vec[i][0]);
      const cplx *const vec_k_end = &(vec[k][len]);
      // b_k = a_k - \sum_{i=1}^{k-1} (a_k,b_i)/(b_i,b_i) b_i
      for(;vec_k < vec_k_end; vec_k++, vec_i++)
	*vec_k -= help * (*vec_i); 
           // note that help is defined with minus sign
    } // for(i)
  }// for(k)
}

// ##########################################################

//! The infinity norm of a square matrix A is the largest row sum
//! of the matrix whose entries are the absolute values of A 


double InfNormMatrix(double **A,int dim)
 {
    double MaxRowSum = 0.0;

    for(int ii=0;ii<dim;ii++)
     {
       double RowSum = 0.0;

       for(int jj=0;jj<dim;jj++)
         RowSum += abs(A[ii][jj]);

       if(RowSum > MaxRowSum)
	 MaxRowSum = RowSum;
     }

    return MaxRowSum;
}

// ###############################################

//! The one norm of a square matrix A is the largest column sum
//! of the matrix whose entries are the absolute values of A 

double OneNormMatrix(double **A,int dim)
 {
   TransposeMatrix(A,dim);
   double OneNorm = InfNormMatrix(A,dim);
   TransposeMatrix(A,dim);
   return OneNorm;
}

// #################################################

double InnerProduct(double * V, double *W, int Length)
{
  double InnerProd = 0.0;

  const double *const V_end = V + Length;

  double *V_p = V;
  double *W_p = W;

  for(;V_p < V_end; V_p++, W_p++)
    InnerProd += (*V_p)*(*W_p);

  return InnerProd;

}








