// $Id: NoiseCovariance.cc,v 1.5 2003/11/05 23:13:34 zweck Exp 

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2002):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "NoiseCovariance.hh"

extern ofstream *LogFile;

NoiseCovariance::NoiseCovariance(string InFileName, string Job2,
				 OptSignal * Signal2)
{

  Job = Job2;
  Signal = Signal2;

  fft = Signal->fft;
  DataToPropagate = Signal->sfftM;

  SignalVecDim = Signal->GetQtPoints();
  TypeSimulation = Signal->GetTypeSimulation();

  NumberOfChannels=Signal->GetNumChannels();

  if (!(NumberOfChannels%2))
     {
     cerr << "NoiseCovariance::NoiseCovariance \n" 
     << "Error:\n" 
     << "Covariance Matrix Method only works with odd number of channels!\n"
     << " ABORTING NOW !!\n";
     //exit(1);
     }
  
     
  // Here I choose the central channel becuase I want the center block of an
  // ideal complete big Covariance Matrix for WDM systems
  // I can choose another channel. This means that I can compute a block matrix
  // on the diagonal of the big ideal complete Covariance Matrix for 
  // WDM systems.
  // ATTENTION. This case has still to be tested!
  // ATTENTION2 (NumberOfChannel-1)/2 is the central channel in case of 
  // Evenly Spaced Channels. In the other two types of Channel allocation
  // it may be but it may be not
  // BOTTOM LINE: the program has been tested only with evenly spaced channels!
  
  /*
  ChannelOfCovarianceMatrix=(NumberOfChannels-1)/2;
   
  // ChanelOfCovarianceMatrix has to be a number between 
  // 0 to NumberOfChannels -1
  
  if ((ChannelOfCovarianceMatrix<0) ||
      (ChannelOfCovarianceMatrix>=NumberOfChannels))
     {
     cerr << "NoiseCovariance::NoiseCovariance(...). Error!\n" 
     << "Wrong ChannelOfCovarianceMatric value.\n"
     << "ChannelOfCovarianceMatrix has to be in range "
     << "[0," << NumberOfChannels-1 
     <<"]!\n" << "ABORTING NOW!!!\n";
     exit(1);
     }
   */ 
    
   // ### Open Input Parameter File #######

  ifstream InFile(InFileName.c_str(),ios::in);    

  if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }

   
  LogFileSeparator();
  *LogFile << InFileName << endl << endl;

  // ####### Read in Parameters ############

  TypeNoiseMethod = typeNoiseMethod(ReadInt("TypeNoiseMethod"," ",0,3,
                                    LOWER_AND_UPPER,MANDATORY,0,&InFile));


  CovMatrixDim = ReadInt("CovMatrixDim"," ",0,-1,LOWER_ONLY,MANDATORY,1,
			 &InFile);

  DeltaFrac = ReadDouble("DeltaFrac"," ",0,1,LOWER_AND_UPPER,
			 OPTIONAL_NO_WARNING,0.01,
			 &InFile);

  TypeJitterSeparation = typeJitterSeparation(
                         ReadInt("TypeJitterSeparation"," ",0,2,
                                  LOWER_AND_UPPER,OPTIONAL_WARNING,1,&InFile));

  TypeJitterSeparationInput = TypeJitterSeparation;

  AlsoRemoveZerosJitter = bool(ReadInt("AlsoRemoveZerosJitter"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_WARNING,1,&InFile));

  InFile.close();

  *LogFile << "TypeJitterSeparation = " << WriteTypeJitterSeparation() << endl;

  // ##### Compute derived parameters ##########

  if (NumberOfChannels>1)
     {
     // WDM case
     // the number of frequencies taken into account into the Covariance
     // Matrix Method has to be equal or less than the number of frequencies
     // contained in the channel bandwidth
     // It could be less because in the case of very spaced channels, it makes
     // no sense use a Covariance Matrix as big as the channel spacing where
     // most of the frequency components of the signal are equal to zero 
     // Note: all the WDM part of the Covariance Matrix has been developed
     // only for evenly spaced channels! 
     int NumFreqInChanSpac= (int) floor(Signal->GetChannSpacing()/
                                        Signal->GetDeltaFreq());
     if (NumFreqInChanSpac%2)
        NumFreqInChanSpac-=1;
						
     if (CovMatrixDim>NumFreqInChanSpac && TypeNoiseMethod==ODE)
        {
        cerr << "NoiseCovariance::NoiseCovariance: CovMatrixDim is bigger"
	     << " than the number of frequencies in the channel spacing" <<endl
	     << "It should be less or equal to " << NumFreqInChanSpac << endl
	     << "ABORTING" << endl;
		  
    *LogFile << "NoiseCovariance::NoiseCovariance: CovMatrixDim is bigger"    
	     << " than the number of frequencies in the channel spacing" <<endl
	     << "It should be less or equal to " << NumFreqInChanSpac
	     << "ABORTING" << endl;
        exit(1);
	} 					 
     }

  if(CovMatrixDim%2)
    {
      cerr << "NoiseCovariance::NoiseCovariance: CovMatrixDim mus be even"
	   << endl
	   << "ABORTING" << endl;
      exit(1);
    }

  CovMatrixFirstIndex = -CovMatrixDim/2;
  CovMatrixLastIndex =  CovMatrixDim/2 - 1;

  AllocateMemory();

  ResetVolatileParameters();


} // end NoiseCovariance::NoiseCovariance

// ##############################################################

/*!

Notes:

-# The OptSignal objects of LHS and RHS NoiseCovariance objects
   need not be the same, though it must be possible to make the call 
   *LHS_Signal = *RHS_Signal, which is done in NoiseCovariance::operator=

-# We copy the CovarianceMatrix and PropagatorMatrix. 
-# A careful inspection of code shows no other data arrays need be
   copied, as they are temporary arrays used in Propagate()
   that are initialized from
   the Signal and CovarianceMatrix before they are used.

-# The fiber and amplfier pointers are not assigned since they
   are set each time Propagate or Amplify is called

 */

NoiseCovariance & NoiseCovariance::operator=(NoiseCovariance & 
					     RHS_NoiseCovariance)
{

  if(TypeSimulation != RHS_NoiseCovariance.TypeSimulation || 
     TypeNoiseMethod != RHS_NoiseCovariance.TypeNoiseMethod ||
     TypeJitterSeparationInput != RHS_NoiseCovariance.TypeJitterSeparationInput ||
     SignalVecDim != RHS_NoiseCovariance.SignalVecDim ||
     CovMatrixDim != RHS_NoiseCovariance.CovMatrixDim ||
     AlsoRemoveZerosJitter != RHS_NoiseCovariance.AlsoRemoveZerosJitter)
    {
      cerr << "NoiseCovariance::operator= : Incompatible Objects"
	   << "ABORTING NOW!!" << endl;
      exit(1);
    }

  Job = RHS_NoiseCovariance.Job;

  TypeJitterSeparation = RHS_NoiseCovariance.TypeJitterSeparation;

  PropagatedLength = RHS_NoiseCovariance.PropagatedLength;
  PropagatedLengthSaved = RHS_NoiseCovariance.PropagatedLengthSaved;
  ZStepNum = RHS_NoiseCovariance.ZStepNum;
  ZStepNumSaved = RHS_NoiseCovariance.ZStepNumSaved;

  AccumulatedFirstOrderDispersion = 
   RHS_NoiseCovariance.AccumulatedFirstOrderDispersion;

  AccumulatedSndOrderDispersion = 
   RHS_NoiseCovariance.AccumulatedSndOrderDispersion;

  DeltaFrac = RHS_NoiseCovariance.DeltaFrac;


  *Signal = *(RHS_NoiseCovariance.Signal);

  CovMatrixFirstIndex = RHS_NoiseCovariance.CovMatrixFirstIndex;
  CovMatrixLastIndex = RHS_NoiseCovariance.CovMatrixLastIndex;


  if(TypeNoiseMethod == MC_WITH_K || TypeNoiseMethod == ODE)
  {
   CopyMatrix(CovarianceMatrix,RHS_NoiseCovariance.CovarianceMatrix,
	     2*CovMatrixDim);

   CopyMatrix(PropagatorMatrix,RHS_NoiseCovariance.PropagatorMatrix,
	     2*CovMatrixDim);
  }

  return *this;

} // ## end operator=

// ############################################################

void NoiseCovariance::AllocateMemory(void)
{

  Reduced_fft = new cfftw(CovMatrixDim, Signal->GetTimeWindow());

  ReducedNoiseFreeSignalStorage.fOptSignalX = NewCplx(CovMatrixDim);
  ReducedNoiseFreeSignalStorage.tOptSignalX = NewCplx(CovMatrixDim);

  ReducedNoiseFreeSignal.fOptSignalX = 
    ReducedNoiseFreeSignalStorage.fOptSignalX - CovMatrixFirstIndex;

  ReducedNoiseFreeSignal.tOptSignalX = 
    ReducedNoiseFreeSignalStorage.tOptSignalX - CovMatrixFirstIndex;

  if(TypeNoiseMethod == MC_WITH_K || TypeNoiseMethod == ODE ||
     TypeNoiseMethod == AWGN)
    {
      CovarianceStorageVec = new double[4*CovMatrixDim*CovMatrixDim];

      if(!CovarianceStorageVec)
    {
      cerr << "NoiseCovariance::AllocateMemory: "  
	   << "Malloc error for CovarianceStorageVec" 
	   << "Aborting Now!"
           << endl;
       
      exit(1); 
    }

      CovarianceMatrix = new double* [2*CovMatrixDim];

      for(int kk=0;kk< 2*CovMatrixDim;kk++)
	{
          CovarianceMatrix[kk] = &(CovarianceStorageVec[kk*2*CovMatrixDim]);
	} 

      ZeroMatrix(CovarianceMatrix,2*CovMatrixDim);

    } // ## end if TypeNoiseMethod
  else
    {
      CovarianceStorageVec = 0;
      CovarianceMatrix = 0;
    }

 if(TypeNoiseMethod == ODE)
  {    

    // Initialize Matrices 

    PropagatorStorageVec = new double[4*CovMatrixDim*CovMatrixDim];
    TempProductStorageVec = new double[4*CovMatrixDim*CovMatrixDim];

    if(!PropagatorStorageVec || !TempProductStorageVec)
    {
      cerr << "NoiseCovariance::AllocateMemory: " 
	   << "Malloc error for PropagatorStorageVec or TempProductStorageVec" 
	   << "Aborting Now!"
           << endl;
       
      exit(1); 
    }

    PropagatorMatrix = new double*[2*CovMatrixDim];
    TempProductMatrix = new double*[2*CovMatrixDim];

    for(int kk=0;kk<2*CovMatrixDim;kk++)
    {
      PropagatorMatrix[kk] = &(PropagatorStorageVec[kk*2*CovMatrixDim]);
      TempProductMatrix[kk] = &(TempProductStorageVec[kk*2*CovMatrixDim]);
    }

    ZeroMatrix(PropagatorMatrix,2*CovMatrixDim);
    ZeroMatrix(TempProductMatrix,2*CovMatrixDim);
    
    
    // Initialize Signal data used in Propagator method

    if(Signal->GetTypeSimulation() == SCALAR)
      {

       NoiseFreeSignalStart.fOptSignalX = NewCplx(SignalVecDim);
       
       NoiseFreeSignalEnd.tOptSignalX = NewCplx(SignalVecDim);
       NoiseFreeSignalEnd.fOptSignalX = NewCplx(SignalVecDim);
 
       jcNoiseFreeSignalEnd.tOptSignalX = NewCplx(SignalVecDim);
       jcNoiseFreeSignalEnd.fOptSignalX = NewCplx(SignalVecDim);

       jcOmegaNoiseFreeSignalEnd.tOptSignalX = NewCplx(SignalVecDim);
       jcOmegaNoiseFreeSignalEnd.fOptSignalX = NewCplx(SignalVecDim);

       PerturbationEnd.tOptSignalX = NewCplx(SignalVecDim);
       PerturbationEnd.fOptSignalX = NewCplx(SignalVecDim);
       
       if (NumberOfChannels>1)
          {
	  PerturbationEnd_saved.tOptSignalX = NewCplx(SignalVecDim);
          PerturbationEnd_saved.fOptSignalX = NewCplx(SignalVecDim);
          }

       // Reduced signals have size CovMatrixDim. The reduced signal 
       // in frequency is obtained
       // from the full length signal by extracting the central
       // CovMatrixDim frequencies. 

       ReducedPerturbationEnd.fOptSignalX = NewCplx(CovMatrixDim);

       ArtificialDispersionCompensationOperator = NewCplx(SignalVecDim);
      }
    else
      {
        cerr << "NoiseCovariance doesn't work for TypeSimulation == VECTOR" 
	       << endl
	       << "ABORTING" << endl;

        exit(1);
      }


  }
  else
  {
    PropagatorStorageVec = TempProductStorageVec = 0;
    PropagatorMatrix = TempProductMatrix = 0;    

  }


} // ## end NoiseCovariance:AllocateMemory()

// #############################################################

//! Called at start of propagation through a system

void NoiseCovariance::ResetVolatileParameters(void)
{

  PropagatedLength = 0;
  ZStepNum = 0;
  PropagatedLengthSaved = 0;
  ZStepNumSaved = 0;
  AccumulatedFirstOrderDispersion = 0.0;
  AccumulatedSndOrderDispersion = 0.0;

  ZeroMatrix(CovarianceMatrix,2*CovMatrixDim);


} // ## end NoiseCovariance::ResetVolatileParameters

// ############################################################

NoiseCovariance::~NoiseCovariance()
{

 ReducedNoiseFreeSignal.fOptSignalX = 0;
 ReducedNoiseFreeSignal.tOptSignalX = 0;
 FreeCplx(ReducedNoiseFreeSignalStorage.fOptSignalX);
 FreeCplx(ReducedNoiseFreeSignalStorage.tOptSignalX);

 Reduced_fft->~cfftw();

 if(TypeNoiseMethod == MC_WITH_K || TypeNoiseMethod == ODE ||
    TypeNoiseMethod == AWGN)
  {

    for(int kk=0;kk< 2*CovMatrixDim;kk++)
      CovarianceMatrix[kk]=0;

    delete [] CovarianceMatrix;

    delete [] CovarianceStorageVec;
  }

 if(TypeNoiseMethod == ODE)
  {

    for(int kk=0;kk< 2*CovMatrixDim;kk++)
      PropagatorMatrix[kk]=0;

    delete [] PropagatorMatrix;

    for(int kk=0;kk< 2*CovMatrixDim;kk++)
      TempProductMatrix[kk]=0;
    
    delete [] TempProductMatrix;
    delete [] PropagatorStorageVec;
    delete [] TempProductStorageVec;

  }
 // fixed bug adding "&& TypeNoiseMethod == ODE" to the following if statement
 // this because the following vectors are used only in ODE case
 // Walter 08/02/04
 if(Signal->GetTypeSimulation() == SCALAR && TypeNoiseMethod == ODE)
  {
    FreeCplx(NoiseFreeSignalStart.fOptSignalX);
    FreeCplx(NoiseFreeSignalEnd.tOptSignalX);
    FreeCplx(NoiseFreeSignalEnd.fOptSignalX);
    FreeCplx(jcNoiseFreeSignalEnd.tOptSignalX);
    FreeCplx(jcNoiseFreeSignalEnd.fOptSignalX);
    FreeCplx(jcOmegaNoiseFreeSignalEnd.tOptSignalX);
    FreeCplx(jcOmegaNoiseFreeSignalEnd.fOptSignalX);
    FreeCplx(PerturbationEnd.tOptSignalX);
    FreeCplx(PerturbationEnd.fOptSignalX);
    if (NumberOfChannels>1)
          {
	  FreeCplx(PerturbationEnd_saved.tOptSignalX);
          FreeCplx(PerturbationEnd_saved.fOptSignalX);
          }
    FreeCplx(ReducedPerturbationEnd.fOptSignalX);
    FreeCplx(ArtificialDispersionCompensationOperator);
  }
  
  

DataToPropagate.tOptSignalX = 0;
DataToPropagate.fOptSignalX = 0;

fft = 0;
Signal = 0;

} // ## end NoiseCovariance::~NoiseCovariance()

// ############################################################

void NoiseCovariance::Propagate(OptFiberLocalError * Fiber2)
{

 Fiber = Fiber2;

 switch(TypeNoiseMethod)
   {

    case AWGN: case MC_WITHOUT_K: case MC_WITH_K:

     Fiber->Propagate(&PropagatedLength,&ZStepNum);
     break;

    case ODE:
     Propagator();
     UpdateCovMatrixODE();   
     break;

    default:
      cerr << "Switch error in NoiseCovariance::Propagate" << endl;
      cerr << "ABORTING!!" << endl;
      exit(1);

   } // end switch


} // ## end NoiseCovariance::Propagate

// ########################################################

//! Computes covariance matrix in case of CW signal for an amplifier 
//! followed by a  single fiber with no loss and no dispersion.
//! The power of the CW signal is specified by Signal->GetOptPowerChann()
//! Based on a special case of a formula of Carena et al. 
//! in PTL Vol 9 Num 4 April 1997, p535.

void NoiseCovariance::PropagateAnalyticCWSignalNoLossNoDispersion(
			       OptFiberLocalError * Fiber2, 
			       OptAmplifier * Amplifier2)
{

 Fiber = Fiber2;
 Amplifier = Amplifier2;

 double Phi = Fiber->GetGamma()*Signal->GetOptPowerChann()*
              Fiber->GetFiberLength();

 cout << "Gamma = " << Fiber->GetGamma() 
      << "Peak Power = " << Signal->GetOptPowerChann()
      << "FiberLength = " << Fiber->GetFiberLength() << endl;

 cout << "Phi = " << Phi << endl;

 // Should have units of W = W/Hz * Hz :
 
 double N_ASE = 
    Amplifier->GetSpectralNoisePowerDensityAddedPerPolarization()*
    Signal->GetDeltaFreq();

  cout << "NoisePerFreq = " << N_ASE << endl;


  double DiagTerm = 0.5*N_ASE*(1 + 2*Phi*Phi);
  double AntiDiagTerm = N_ASE*Phi*(sin(2*Phi) + Phi*cos(2*Phi));
  double RIBlockTerm = N_ASE*Phi*(cos(2*Phi) - Phi*sin(2*Phi));

  ZeroMatrix(CovarianceMatrix,2*CovMatrixDim);
  AddToPrincipalDiag(CovarianceMatrix,2*CovMatrixDim,DiagTerm);

  CovarianceMatrix[0][0] -= AntiDiagTerm;
  CovarianceMatrix[0][CovMatrixDim] += RIBlockTerm;
  CovarianceMatrix[CovMatrixDim][0] += RIBlockTerm;
  CovarianceMatrix[CovMatrixDim][CovMatrixDim] += AntiDiagTerm;

  for(int kk=1;kk<CovMatrixDim;kk++)
    {
      int ii,jj;
 
      // (R,R)-block

      ii=kk;
      jj=CovMatrixDim-kk;
      CovarianceMatrix[ii][jj] -= AntiDiagTerm;

      // (R,I)-block

      ii=kk;
      jj=2*CovMatrixDim-kk;
      CovarianceMatrix[ii][jj] +=  RIBlockTerm;

      // (I,R)-block

      ii=kk+CovMatrixDim;
      jj=CovMatrixDim-kk;
      CovarianceMatrix[ii][jj] +=  RIBlockTerm;


      // (I,I)-block

      ii=kk+CovMatrixDim;
      jj=2*CovMatrixDim-kk;
      CovarianceMatrix[ii][jj] += AntiDiagTerm;

    } // end for kk loop

} // ## end PropagateAnalyticNoLossNoDispersion

// #########################################################

void NoiseCovariance::Amplify(OptAmplifier * Amplifier2)
{

 Amplifier = Amplifier2;

 switch(TypeNoiseMethod)
   {

    case AWGN:

     Amplifier->SetTypeAmplifierNoise(NOISE_ON_SEMI_ANALYTICAL);
     Amplifier->AmplifyOptSignal();

     break;

    case MC_WITHOUT_K: case MC_WITH_K:

     Amplifier->AmplifyOptSignal();

     break;

    case ODE:

     // Amplify noise-free signal

     /* We use NOISE_ON_SEMI_ANALYTICAL so that saturated and 
        fixed output power amplifiers work correctly compared to
        what you'd have with Monte Carlo, at least for the case
        of no nonlinearity. For now that will have to do. Later
        I should fix this by using Diag(CovarianceMatrix) data
        instead of AccumulatedTotalNoiseSpectralDensity. But this
        will require more careful work. JZ 19/11/2003
     */


     //Amplifier->SetTypeAmplifierNoise(NOISE_OFF);
     Amplifier->SetTypeAmplifierNoise(NOISE_ON_SEMI_ANALYTICAL);
     Amplifier->AmplifyOptSignal();
 
     Amplifier->SetTypeAmplifierNoise(NOISE_ON_GAUSSIAN_WHITE);
     AmplifyCovarianceMatrix(Amplifier->GetLinearGainOptAmplif());

     break;

    default:
      cerr << "Switch error in NoiseCovariance::Amplify" << endl;
      cerr << "ABORTING!!" << endl;
      exit(1);

   } // end switch
  
} // ## end NoiseCovariance::Amplify

// ##########################################################

/*!

This LumpedLoss object must act on the same signal
that the NoiseCovariance object is acting on.
I am not explicitly enforcing this at the moment.

*/

void NoiseCovariance::Attenuate(LumpedLoss * oLumpedLoss)
{

 switch(TypeNoiseMethod)
   {

    case AWGN: case MC_WITHOUT_K: case MC_WITH_K:

     oLumpedLoss->ApplyLumpedLoss();

     break;

    case ODE:

     oLumpedLoss->ApplyLumpedLoss();

     LumpedLossCovarianceMatrix(oLumpedLoss->GetLossLinear());

     break;

    default:
      cerr << "Switch error in NoiseCovariance::Amplify" << endl;
      cerr << "ABORTING!!" << endl;
      exit(1);

   } // end switch

} // ## end NoiseCovariance::Attenuate

// ###########################################################

void NoiseCovariance::TurnOffJitterSeparation(void)
{
 TypeJitterSeparation = NONE;
}

// #######################################################

void NoiseCovariance::TurnOnJitterSeparation(void)
{
 TypeJitterSeparation = TypeJitterSeparationInput;
}



// ###########################################################

string NoiseCovariance::WriteTypeNoiseMethod(void)
{

string name;

switch(TypeNoiseMethod)
  {
  case AWGN: name = "AWGN"; break;
  case MC_WITHOUT_K: name = "MC_WITHOUT_K"; break;
  case MC_WITH_K: name = "MC_WITH_K"; break;
  case ODE: name = "ODE"; break;
  default: name = "Invalid TypeNoiseMethod";
  }

return name;

}

// ############################################################

string NoiseCovariance::WriteTypeJitterSeparation(void)
{

string name;

switch(TypeJitterSeparation)
  {
  case NONE: name = "NONE"; break;
  case PHASE_ONLY: name = "PHASE_ONLY"; break;
  case PHASE_AND_TIMING: name = "PHASE_AND_TIMING"; break;
  default: name = "Invalid TypeJitterSeparation";
  }

return name;

}

// ###########################################################

void NoiseCovariance::ComputeReducedSignal(fftMatrices Signal2,
					   fftMatrices ReducedSignal2)
{

  // Reduced Signals must be defined as is fftMatrices ReducedNoiseFreeSignal

  fft->fftshift(Signal2.fOptSignalX);

  for(int jj=CovMatrixFirstIndex;jj<=CovMatrixLastIndex;jj++)
    ReducedSignal2.fOptSignalX[jj] = Signal2.fOptSignalX[SignalVecDim/2 + jj];

  Reduced_fft->ifftshift(&(ReducedSignal2.fOptSignalX[CovMatrixFirstIndex]));

  Reduced_fft->IFFT(&(ReducedSignal2.fOptSignalX[CovMatrixFirstIndex]),
		    &(ReducedSignal2.tOptSignalX[CovMatrixFirstIndex]));

  Reduced_fft->fftshift(&(ReducedSignal2.fOptSignalX[CovMatrixFirstIndex]));

  fft->ifftshift(Signal2.fOptSignalX);

}

// #########################################################

void NoiseCovariance::WriteFileNoiseFreeSignalAndCovarianceMatrix(
			                 string OutFileNamesSuffix)
{

ComputeCovarianceMatrixAWGN();

if(TypeNoiseMethod == ODE || TypeNoiseMethod == AWGN) 
 {

   string FileName;

   /**************

   FileName = Job + OutFileNamesSuffix + "tNoiseFreeSignal" + ".dat";

   Signal->WriteFileTimePower(FileName);

   FileName = Job + OutFileNamesSuffix + "fNoiseFreeSignal" + ".dat";

   Signal->WriteFileFreqPower(FileName);

   ******************/


  //write to file complex signal
 

  FileName = Job + OutFileNamesSuffix + "tNoiseFreeSignalComplexFields.dat";
  bool ForRestartFlag=0;
  Signal->WriteComplexFields(FileName,ForRestartFlag);
  
  FileName = Job + OutFileNamesSuffix + "fNoiseFreeSignalComplexFields.dat";
  Signal->WriteComplexFieldsFreq(FileName);
  

   // Write out NoiseFreeSignal to a file so that we can potentially
   // read it back in again in ChiSquareReceiver


  FileName = Job + OutFileNamesSuffix + "tCplxNoiseFreeSignal.dat";
  ForRestartFlag = 1;

  Signal->WriteComplexFields(FileName,ForRestartFlag);
  
  
  ComputeReducedSignal(DataToPropagate,ReducedNoiseFreeSignal);

  // First: Output the reduced noisefree signal in time

  FileName = Job + OutFileNamesSuffix + "tNoiseFreeReducedSignal" + ".dat";

   ofstream File(FileName.c_str(),ios::out);

   if (!File) 
    {
      cerr << "Error: " << FileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   File 
     << "% (1) Time [ps] (2) Power [W] (3) Real Part (4) Imag Part [W^0.5]"
     << endl;


   for(int jj=CovMatrixFirstIndex;jj<=CovMatrixLastIndex;jj++)  
     {
       
       File 
	 << 1e12*double(jj-CovMatrixFirstIndex)*Reduced_fft->delta_t << "\t"
	 << sq(abs(ReducedNoiseFreeSignal.tOptSignalX[jj])) << "\t"
	 << real(ReducedNoiseFreeSignal.tOptSignalX[jj]) << "\t"
	 << imag(ReducedNoiseFreeSignal.tOptSignalX[jj]) << endl;

     }

   File.close();

  // Second: Output the noisefree signal in frequency

   FileName = Job + OutFileNamesSuffix + "fNoiseFreeReducedSignal" + ".dat";

   File.open(FileName.c_str(),ios::out);

   if (!File) 
    {
      cerr << "Error: " << FileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   File 
     << "% (1) Freq [GHz] (2) Power [W] (3) Real Part (4) Imag Part"
     << endl;

   //  If you use fftshift here then DC term is at first index of output
   //  rather than being in the center.

   //  Reduced_fft->fftshift(
   //  &(ReducedNoiseFreeSignal.fOptSignalX[CovMatrixFirstIndex]));


   for(int jj=CovMatrixFirstIndex;jj<=CovMatrixLastIndex;jj++)  
     {
       
       File 
	 << 1e-9*jj*Reduced_fft->delta_f << "\t"
	 << sq(abs(ReducedNoiseFreeSignal.fOptSignalX[jj])) << "\t"
	 << real(ReducedNoiseFreeSignal.fOptSignalX[jj]) << "\t"
	 << imag(ReducedNoiseFreeSignal.fOptSignalX[jj]) << endl;
     }

   //Reduced_fft->ifftshift(
   //  &(ReducedNoiseFreeSignal.fOptSignalX[CovMatrixFirstIndex]));


   File.close();

 // Third: Output the covariance matrix

   FileName = Job + OutFileNamesSuffix + "CovarianceMatrix" + ".dat";

   File.open(FileName.c_str(),ios::out);

   if (!File) 
    {
      cerr << "Error: " << FileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   File.precision(12);

   WriteMatrix(CovarianceMatrix,2*CovMatrixDim,&File,1.0);

   File.close();

   /**********

 // Fourth: Output the propagator matrix

   FileName = Job + OutFileNamesSuffix + "PropagatorMatrix" + ".dat";

   File.open(FileName.c_str(),ios::out);

   if (!File) 
    {
      cerr << "Error: " << FileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   WriteMatrix(PropagatorMatrix,2*CovMatrixDim,&File,1.0);

   File.close();

   ************/

 } // end if TypeNoiseMethod==ODE

}

// ##############################################################

void NoiseCovariance::ReadFileNoiseFreeSignalAndCovarianceMatrix(
                                        string InFileNamesPrefix)
{

   // First: Read in NoiseFreeSignal
   
   Signal->ReadComplexFields(InFileNamesPrefix+"_tCplxNoiseFreeSignal.dat");
 
   string FileName;
   ifstream File;

  // Second: Read in the covariance matrix
   
   FileName = InFileNamesPrefix + "_CovarianceMatrix.dat";

   File.open(FileName.c_str(),ios::in);

   if (!File) 
    {
      cerr << "Error: " << FileName 
	   << " could not be opened. ABORTING"<<endl;
      exit(1);
   }

   ReadMatrix(CovarianceMatrix,2*CovMatrixDim,&File,1.0);

   File.close();
 
   /***************

   // First read in ReducedNoiseFreeSignal in frequency

   FileName = InFileNamesPrefix + "fNoiseFreeReducedSignal.dat";

   if (!File) 
    {
      cerr << "Error: " << FileName 
	   << "could not be opened. ABORTING"<<endl;
      exit(1);
   }

   char TempChar[1];

   File.getline(TempChar,1);
     
   for(int jj=CovMatrixFirstIndex;jj<=CovMatrixLastIndex;jj++)  
     {
       double Temp, RealValue, ImagValue;       

       File >> Temp >> Temp >> RealValue >> ImagValue;
       ReducedNoiseFreeSignal.fOptSignalX[jj] = RealValue + jc*ImagValue;
     }

  File.close();
   
  Reduced_fft->ifftshift(&(ReducedNoiseFreeSignal.fOptSignalX[
                                          CovMatrixFirstIndex]));

  Reduced_fft->IFFT(&(ReducedNoiseFreeSignal.fOptSignalX[CovMatrixFirstIndex]),
		   &(ReducedNoiseFreeSignal.tOptSignalX[CovMatrixFirstIndex]));

  Reduced_fft->fftshift(&(ReducedNoiseFreeSignal.fOptSignalX[
					 CovMatrixFirstIndex]));

   ***********/

} // ## end ReadFileReducedNoiseFreeSignalAndCovarianceMatrix


// ############################################################

  //! No real need to use this method anymore.

void NoiseCovariance::WriteFileAWGNNoiseFreeSignalAndNoiseSpectralDensity(
						 string OutFileNameSuffix)
{

  string FileName1 = Job + OutFileNameSuffix + "TotalNoiseSpectralDensity.dat";

   ofstream OutFile(FileName1.c_str(),ios::out);

   if (!OutFile)
    {
       cerr << "Error: " << FileName1 << "could not be created."<<endl;
       exit(1);
    }

   OutFile << "% NoiseSpectralDensity (SCALAR) [W/Hz] " << endl
        << Signal->GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics();

   OutFile.close();


 if(TypeNoiseMethod==AWGN)
  {


   string FileName = Job + OutFileNameSuffix + "tNoiseFreeSignal" + ".dat";

   Signal->WriteFileTimePower(FileName);

   FileName = Job + OutFileNameSuffix + "fNoiseFreeSignal" + ".dat";

   Signal->WriteFileFreqPower(FileName);
  
   /**
   FileName = Job + OutFileNameSuffix + "TotalNoiseSpectralDensity.dat";

   ofstream OutFile(FileName.c_str(),ios::out);

   if (!OutFile)
    {
       cerr << "Error: " << FileName << "could not be created."<<endl;
       exit(1);
    }

   OutFile << "% NoiseSpectralDensity (SCALAR) [W/Hz] " << endl
         <<Signal->GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics();

   OutFile.close();
   **/

   FileName = Job + OutFileNameSuffix + "TotalNoiseSpectralDensity.dat";

   // ## Next output reduced noise free signal 
   // ## so can use ChiSquare receiver model    

   ComputeReducedSignal(DataToPropagate,ReducedNoiseFreeSignal);

   FileName = Job + OutFileNameSuffix + "fNoiseFreeReducedSignal" + ".dat";

   OutFile.open(FileName.c_str(),ios::out);

   if (!OutFile)
    {
      cerr << "Error: " << FileName
           << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   OutFile
     << "% (1) Freq [GHz] (2) Power [W] (3) Real Part (4) Imag Part"
     << endl;

   //  If you use fftshift here then DC term is at first index of output
   //  rather than being in the center.
 
   //  Reduced_fft->fftshift(
   //  &(ReducedNoiseFreeSignal.fOptSignalX[CovMatrixFirstIndex]));
 
 
   for(int jj=CovMatrixFirstIndex;jj<=CovMatrixLastIndex;jj++)
     {
 
       OutFile
         << 1e-9*jj*Reduced_fft->delta_f << "\t"
         << sq(abs(ReducedNoiseFreeSignal.fOptSignalX[jj])) << "\t"
         << real(ReducedNoiseFreeSignal.fOptSignalX[jj]) << "\t"
         << imag(ReducedNoiseFreeSignal.fOptSignalX[jj]) << endl;
     }

   //Reduced_fft->ifftshift(
   //  &(ReducedNoiseFreeSignal.fOptSignalX[CovMatrixFirstIndex]));

  } // ## end if TypeNoiseMethod==AWGN

} // ## end WriteFileAWGNNoiseSpectralDensity

// #############################################################

void NoiseCovariance::WriteFileTimeField(string OutFileName,fftMatrices Field)
{

   ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

   if(TypeSimulation == VECTOR)
     {
      for (int ii = 0; ii < SignalVecDim; ii++) 
          OutFile << ii*Signal->GetDeltaTime() << "   " 
                  <<  sq(abs(Field.tOptSignalX[ii]))
	             +sq(abs(Field.tOptSignalY[ii]))  <<endl;   
     }
    else
     {
      for (int ii = 0; ii < SignalVecDim; ii++) 
          OutFile << ii*Signal->GetDeltaTime() << "   " 
                  <<  sq(abs(Field.tOptSignalX[ii])) <<endl;   
     }

   OutFile.close();


} // ## end WriteFileTimeField

// ########################################################

void NoiseCovariance::WriteFileFreqField(string OutFileName, fftMatrices Field)
{

  
   ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
   {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

 if(TypeSimulation == VECTOR)
  {

   fft->fftshift(Field.fOptSignalX);
   fft->fftshift(Field.fOptSignalY);
      
   for (int ii = 0; ii < SignalVecDim; ii++)  
    {    
      OutFile <<  ii*Signal->GetDeltaFreq() - Signal->GetFreqWindow()/2.0 
              << "   " 
              <<  sq(abs(Field.fOptSignalX[ii]))
	         +sq(abs(Field.fOptSignalY[ii])) 
              << endl;   
    }

   fft->ifftshift(Field.fOptSignalX);
   fft->ifftshift(Field.fOptSignalY);    
  }
 else
   {
    fft->fftshift(Field.fOptSignalX);
       
    for (int ii = 0; ii < SignalVecDim; ii++)  
     {    
      OutFile <<  ii*Signal->GetDeltaFreq() - Signal->GetFreqWindow()/2.0 
              << "   " 
              << sq(abs(Field.fOptSignalX[ii]))
              << endl;   
    }

   fft->ifftshift(Field.fOptSignalX);
   
   }

   OutFile.close();   

} // ## end WriteFileFreqField

// ############################################################

void NoiseCovariance::GetReducedNoiseFreeSignal(
                       fftMatrices & ReceiverReducedNoiseFreeSignal)
{

  ComputeReducedSignal(DataToPropagate,ReducedNoiseFreeSignal);

   ReceiverReducedNoiseFreeSignal.tOptSignalX = 
   ReducedNoiseFreeSignalStorage.tOptSignalX;

  ReceiverReducedNoiseFreeSignal.fOptSignalX = 
   ReducedNoiseFreeSignalStorage.fOptSignalX;

}

// ###########################################################

void NoiseCovariance::ComputeCovarianceMatrixAWGN(void)
{

  if(TypeNoiseMethod==AWGN)
    {

     ZeroMatrix(CovarianceMatrix,2*CovMatrixDim);

     double NoiseVarianceRealOrImag;

     if(Signal->GetTypeSimulation()==SCALAR)
       {
         NoiseVarianceRealOrImag = 
          Signal->GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics()*
          Signal->GetDeltaFreq()/2.0;
       }
     else
       {
        NoiseVarianceRealOrImag = 
          Signal->GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics()*
          Signal->GetDeltaFreq()/4.0;
       }


     AddToPrincipalDiag(CovarianceMatrix,2*CovMatrixDim,
			NoiseVarianceRealOrImag);

    }

}

// ########################################################

void NoiseCovariance::SetCovarianceMatrixAWGN(double OSNRdB,double B_OSA)
{

 double NoisePower = Signal->GetAveragePower()/
                    (Signal->GetNumChannels()*dB2Linear(OSNRdB));

 Signal->SetAccumulatedTotalNoiseSpectralDensityGaussianStatistics(
					      NoisePower/B_OSA);

 ComputeCovarianceMatrixAWGN();

}

// #######################################################

void NoiseCovariance::InitializeCovarianceMatrix(
                double TotalNoiseSpectralDensity)
{

     Signal->SetAccumulatedTotalNoiseSpectralDensityGaussianStatistics(
                                              TotalNoiseSpectralDensity);

     ZeroMatrix(CovarianceMatrix,2*CovMatrixDim);

     AddToPrincipalDiag(CovarianceMatrix,2*CovMatrixDim,
                        0.5*TotalNoiseSpectralDensity*Signal->GetDeltaFreq());

}


// #########################################################

void NoiseCovariance::GetCovarianceMatrix(double ** ReceiverCovarianceMatrix,
					  bool ReadInSignalAndCovMatrixFlag,
                                          double SetOSNRdB, double B_OSA)
{

  if(!ReadInSignalAndCovMatrixFlag)
     ComputeCovarianceMatrixAWGN();

  if(SetOSNRdB)
    {
      if(TypeNoiseMethod != AWGN)
	{
	 cerr << "SetOSNRdB != 0 requires TypeNoiseMethod == AWGN"
	     << "ABORTING"  << endl;
         exit(1);
	}

     SetCovarianceMatrixAWGN(SetOSNRdB,B_OSA);

    }

  for(int kk=0;kk< 2*CovMatrixDim;kk++)
        ReceiverCovarianceMatrix[kk] = CovarianceMatrix[kk];
    
}

// ############################################################

void NoiseCovariance::Propagator(void)
{

  double AccumulatedTotalNoiseSpectralDensity = 
   Signal->GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics();

  // Compute accumulated dispersion at end of this fiber

  //cout << AccumulatedFirstOrderDispersion << endl;

  AccumulatedFirstOrderDispersion += 
    Fiber->GetFiberLength()*Fiber->GetFirstOrderDispersion();

  AccumulatedSndOrderDispersion += 
    Fiber->GetFiberLength()*Fiber->GetSecondOrderDispersion();

  // Initialize NoiseFreeSignalStart. This contains all channels
  // All fields at End only contain a single channel.

  memcpy(NoiseFreeSignalStart.fOptSignalX,DataToPropagate.fOptSignalX,
	 SignalVecDim*sizeof(cplx));

  
  // Allow step sizes to stabilize by propagating three times.  
  // Moreover the following lines create the NoiseFreeSignalEnd required to
  // calculate the perturbation du_m(t,L) (Ron's thesis pag. 51)
  // Note: Perturb propagates the signal from NoiseFreeSignalStart to
  // NoiseFreeSignalEnd. These three lines do not propagate 3 times farther.
  
  
  Perturb(0,0.0,1);
  Perturb(0,0.0,1);
  Perturb(0,0.0,1);
  
  /*
  // these lines should be used with this version of Perturb
  Perturb(0,0.0,1,ChannelOfCovarianceMatrix);
  Perturb(0,0.0,1,ChannelOfCovarianceMatrix);
  Perturb(0,0.0,1,ChannelOfCovarianceMatrix);
  */

  // Initialize NoiseFreeSignalEnd 

  memcpy(NoiseFreeSignalEnd.tOptSignalX,DataToPropagate.tOptSignalX,
	 SignalVecDim*sizeof(cplx));

  memcpy(NoiseFreeSignalEnd.fOptSignalX,DataToPropagate.fOptSignalX,
	 SignalVecDim*sizeof(cplx));

    
  // Initialize jcNoiseFreeSignalEnd and jcOmegaNoiseFreeSignalEnd

  for(int ii=0;ii<SignalVecDim;ii++)
    {
      jcNoiseFreeSignalEnd.fOptSignalX[ii] = 
	jc*NoiseFreeSignalEnd.fOptSignalX[ii];

      jcOmegaNoiseFreeSignalEnd.fOptSignalX[ii] =
	jcNoiseFreeSignalEnd.fOptSignalX[ii]*2.0*pi*fft->freq(ii);

    }

  fft->IFFT(jcNoiseFreeSignalEnd.fOptSignalX,jcNoiseFreeSignalEnd.tOptSignalX);

  fft->IFFT(jcOmegaNoiseFreeSignalEnd.fOptSignalX,
	    jcOmegaNoiseFreeSignalEnd.tOptSignalX);


  ComputeArtificialDispersionCompensationOperator();
  ApplyArtificialDispersionCompensation(jcNoiseFreeSignalEnd,true);
  ApplyArtificialDispersionCompensation(jcOmegaNoiseFreeSignalEnd,true);

  
  // Compute the pertrubation amount Delta 

  double Delta = sqrt(Signal->GetAveragePower())*DeltaFrac;

  //cout << "Delta = " << Delta << endl;

  // Compute total energy for use in ComputeColumnPropagator calls below

  double TotalEnergy = 0.0;

  for(int jj = 0; jj < SignalVecDim; jj++)
    TotalEnergy += sq(abs(jcNoiseFreeSignalEnd.tOptSignalX[jj]));

  // Compute PropagatorMatrix
  // This for loop does one column at a time.

 for(int kk=CovMatrixFirstIndex; kk <= CovMatrixLastIndex; kk++)
 {
   // We compute column kk of PropagatorMatrix, i.e.,
   // \Psi_{j,k} for 1 <= j <= 2*CovMatrixDim

   bool RealNotImag = 1;
   Perturb(kk,Delta,RealNotImag);
   //Perturb(kk,Delta,RealNotImag,ChannelOfCovarianceMatrix);
   ComputeColumnPropagator(kk,Delta,RealNotImag,TotalEnergy);

   RealNotImag = 0;
   Perturb(kk,Delta,RealNotImag);
   //Perturb(kk,Delta,RealNotImag,ChannelOfCovarianceMatrix);
   ComputeColumnPropagator(kk,Delta,RealNotImag,TotalEnergy);

 } // end for kk

 // Finally compute noise-free solution with all channels

 memcpy(DataToPropagate.fOptSignalX,NoiseFreeSignalStart.fOptSignalX,
	 SignalVecDim*sizeof(cplx));

 fft->IFFT(DataToPropagate.fOptSignalX,DataToPropagate.tOptSignalX);

 PropagatedLength = PropagatedLengthSaved;
 ZStepNum = ZStepNumSaved;

 Signal->SetAccumulatedTotalNoiseSpectralDensityGaussianStatistics(
                           AccumulatedTotalNoiseSpectralDensity);

 Fiber->Propagate(&PropagatedLength,&ZStepNum);

 } // ## end NoiseCovariance::Propagator

// #############################################################

void NoiseCovariance::Perturb(int PerturbedFreqIndex, 
			      double Delta, bool RealNotImag)
{

  // This will need to be modified for VECTOR.
  
  memcpy(DataToPropagate.fOptSignalX,NoiseFreeSignalStart.fOptSignalX,
	 SignalVecDim*sizeof(cplx));

  int Index_k = fft->index(PerturbedFreqIndex); 

  DataToPropagate.fOptSignalX[Index_k] += (RealNotImag) ? 
                                cplx(Delta,0.0) : cplx(0.0,Delta);

  fft->IFFT(DataToPropagate.fOptSignalX,DataToPropagate.tOptSignalX);

  // In a given fiber the first time we call Perturb we have Delta == 0
  // and we propagate the noise-free signal. Subsequent calls traverse
  // the same fiber.

 // JZ Nov 2003: THIS NEXT if/else IS BUGGY as cll Pertub three times at start.

  if(Delta == 0)
    {
     PropagatedLengthSaved = PropagatedLength;
     ZStepNumSaved = ZStepNum;
     //cout << "NoiseCovariance::Perturb: Delta == 0" << endl;
    }
  else
    {
      PropagatedLength = PropagatedLengthSaved;
      ZStepNum = ZStepNumSaved;
      //cout << "NoiseCovariance::Perturb: Delta != 0" << endl;

    }

  Fiber->Propagate(&PropagatedLength,&ZStepNum);

} // ## end NoiseCovariance::Perturb

// #############################################################

void NoiseCovariance::ComputeColumnPropagator(int PerturbedFreqIndex,
					      double Delta, 
					      bool RealNotImag,
					      double TotalEnergy)
{
  for(int jj=0; jj < SignalVecDim; jj++)
    PerturbationEnd.tOptSignalX[jj] = 
      (DataToPropagate.tOptSignalX[jj] - NoiseFreeSignalEnd.tOptSignalX[jj])/
       Delta;

  fft->FFT(PerturbationEnd.tOptSignalX,PerturbationEnd.fOptSignalX);
  
  ApplyArtificialDispersionCompensation(PerturbationEnd,true);
  
  if (NumberOfChannels>1)
      ApplyArtificialCentralChannelDemux(PerturbationEnd);
  
  // Jitter Separation

  int StringLength = Signal->GetStringLength();

  double alpha[(3*3-3)/2]; // See SchmidtOrthogonalizer in MatrixTools

  if(TypeJitterSeparation != NONE && StringLength > 1)
    {
      if(SignalVecDim%StringLength)
	{
          cerr << "NoiseCovariance::ComputeColumnPropagator:" << endl
	       << "qtPoints must be a multiple of StringLength" << endl
	       << "ABORTING" << endl;
          exit(1);
	}

      const int PointsPerBit = SignalVecDim/StringLength;
      const double AveragePulseEnergy = 2.0*TotalEnergy/StringLength;

      for(int bit=0; bit < StringLength; bit++)
	{
          int StartBitIndex = bit*PointsPerBit; // JZ:Correct

          double BitEnergy = 0.0;
 
	  // We use jcNoiseFreeSignalEnd since it has been disentangled
	  // in Propagator method. JZ:Correct

          for(int kk=StartBitIndex;kk<StartBitIndex+PointsPerBit;kk++)
	    BitEnergy += sq(abs(jcNoiseFreeSignalEnd.tOptSignalX[kk]));
     
	  if(!AlsoRemoveZerosJitter && BitEnergy < 0.05*AveragePulseEnergy)
	    {

	      /****
              cout << "BitEnergy in bit #" << bit << " too small, skipping"
		   << endl
		   << "This bit is a " << Signal->oBitString->GetBit(0,bit)
		   << endl;
	      ****/

              continue; // i.e. skip rest of for loop for this bit
            }

          if(TypeJitterSeparation == PHASE_ONLY)
	    {

	      // For each bit we make the vector that is the 
	      // restriction of the perturbation
	      // to that bit be orthogonal to the vector that is the 
	      // restriction of the phase jitter vector  to that bit
	      
              cplx *vec[] = { &jcNoiseFreeSignalEnd.tOptSignalX[StartBitIndex],
                              &PerturbationEnd.tOptSignalX[StartBitIndex] };

	      SchmidtOrthogonalizer(vec,2,PointsPerBit,alpha);
	      
              for(int g=0;g < (2*2-2)/2;g++) //See SchmidtOrthogonalizer 
		{
		  // Alpha is nan iff one of vectors you are orthogonalizing
		  // wrt is zero. See SchmidtOrthogonalizer 

		  if(isnan(alpha[g]))
		    {
                      cerr << "NoiseCovariance::ComputeColumnPropagator:" 
			   << endl
			   << "alpha element is nan, so a vector was zero"
			   << endl;

                   exit(1);
		    }

		} // end for g

	    } // end PHASE_ONLY
          else if( TypeJitterSeparation == PHASE_AND_TIMING)
	    {
	      // For each bit we make the vector that is the 
	      // restriction of the perturbation
	      // to that bit be orthogonal to span of the vectors that are the 
	      // restrictions of the phase and timing jitter vectors  
	      // to that bit.
	      // We put the timing jitter vector first so that it is not
	      // modified by Gramm Schmidt, since we may want it later on (??)

              cplx *vec[] = 
	               { &jcOmegaNoiseFreeSignalEnd.tOptSignalX[StartBitIndex],
			 &jcNoiseFreeSignalEnd.tOptSignalX[StartBitIndex],
                         &PerturbationEnd.tOptSignalX[StartBitIndex] };

	      SchmidtOrthogonalizer(vec,3,PointsPerBit,alpha);

              for(int g=0;g < (3*3-3)/3;g++) //See SchmidtOrthogonalizer 
		{
		  // Alpha is nan iff one of vectors you are orthogonalizing
		  // wrt is zero. See SchmidtOrthogonalizer 

		  if(isnan(alpha[g]))
		    {
                      cerr << "NoiseCovariance::ComputeColumnPropagator:" 
			   << endl
			   << "alpha element is nan, so a vector was zero"
			   << endl;

                   exit(1);
		    }

		} // end for g


	    } // ## end  PHASE_AND_TIMING

	} // ## end for bit loop

    } // ## end if(TypeJitterSeparation != NONE && StringLength > 1)

      // Finally we also orthogonalize the entire perturbation vector wrt
      // the phase and/or timing jitter vectors. This ensures that in addition
      // to removing the jitters from each of the marks we also remove the
      // total jitters in the zeros.
     
      if(TypeJitterSeparation == PHASE_ONLY)
	    {

              cplx *vecB[] = { jcNoiseFreeSignalEnd.tOptSignalX,
                              PerturbationEnd.tOptSignalX };

	      SchmidtOrthogonalizer(vecB,2,SignalVecDim,alpha);

              for(int g=0;g < (2*2-2)/2;g++) //See SchmidtOrthogonalizer 
		{
		  // Alpha is nan iff one of vectors you are orthogonalizing
		  // wrt is zero. See SchmidtOrthogonalizer 

		  if(isnan(alpha[g]))
		    {
                      cerr << "NoiseCovariance::ComputeColumnPropagator:" 
			   << endl
			   << "alpha element is nan, so a vector was zero"
			   << endl;

                   exit(1);
		    }

		} // end for g

	    } // end PHASE_ONLY
          else if( TypeJitterSeparation == PHASE_AND_TIMING)
	    {
	      
              cplx *vecB[] = 
	               { jcOmegaNoiseFreeSignalEnd.tOptSignalX,
			 jcNoiseFreeSignalEnd.tOptSignalX,
                         PerturbationEnd.tOptSignalX };

	      SchmidtOrthogonalizer(vecB,3,SignalVecDim,alpha);

              for(int g=0;g < (3*3-3)/3;g++) //See SchmidtOrthogonalizer 
		{
		  // Alpha is nan iff one of vectors you are orthogonalizing
		  // wrt is zero. See SchmidtOrthogonalizer 

		  if(isnan(alpha[g]))
		    {
                      cerr << "NoiseCovariance::ComputeColumnPropagator:" 
			   << endl
			   << "alpha element is nan, so a vector was zero"
			   << endl;

                   exit(1);
		    }

		} // end for g


	    } // ## end  PHASE_AND_TIMING

      fft->FFT(PerturbationEnd.tOptSignalX,PerturbationEnd.fOptSignalX);
      
      if (NumberOfChannels>1)
         RestoreChannels(PerturbationEnd);
      
      // Re-entangle pulses of perturbation

      ApplyArtificialDispersionCompensation(PerturbationEnd,false);
       
      memcpy(ReducedPerturbationEnd.fOptSignalX,
	      &PerturbationEnd.fOptSignalX[SignalVecDim - CovMatrixDim/2],
	      sizeof(cplx)*CovMatrixDim/2);

      memcpy(&ReducedPerturbationEnd.fOptSignalX[CovMatrixDim/2],
	      PerturbationEnd.fOptSignalX,
	      sizeof(cplx)*CovMatrixDim/2);


      // Fill in column k+offset in PropagatorMatrix[][]

      const int offset = (RealNotImag) ? 0:CovMatrixDim;

      double *Propagator_j_re_k =
	&(PropagatorMatrix[0][PerturbedFreqIndex+offset-CovMatrixFirstIndex]);

      double *Propagator_j_re_N_k =
        &(PropagatorMatrix[CovMatrixDim]
	                  [PerturbedFreqIndex+offset-CovMatrixFirstIndex]);

	
	  // The PropagatorMatrix is really stored as a vector
	  // The offset between an element and the one in the 
	  // same column and next row down is 2*CovMatrixDim
 
      for(int j_re = CovMatrixFirstIndex; j_re <= CovMatrixLastIndex; 
	  j_re++, Propagator_j_re_k += 2*CovMatrixDim, 
	          Propagator_j_re_N_k += 2*CovMatrixDim)
        {
          cplx noise_j_re = 
	    ReducedPerturbationEnd.fOptSignalX[j_re - CovMatrixFirstIndex];

          *Propagator_j_re_k   = real(noise_j_re); // Psi[j_re][k+offset]
          *Propagator_j_re_N_k = imag(noise_j_re); 
	                              // Psi[j_re+CovMatrixDim][k+offset]
    
        } //## for j_re
    
} // ## end ComputeColumnPropagator

// ###########################################################

void NoiseCovariance::ComputeArtificialDispersionCompensationOperator(void)
{

  // Based on OptFiberLocalError::ComputeFreqDomainOperator

  register const double FstOrDispDivTwo = -AccumulatedFirstOrderDispersion/2.0;
  register const double SndOrDispDivSix = -AccumulatedSndOrderDispersion/6.0;

  for(int jj=0;jj<SignalVecDim;jj++)
    {
      double omega = 2.0*pi*fft->freq(jj); // omega is freq in radians
	
      ArtificialDispersionCompensationOperator[jj] = 
	     exp_jc_real(sq(omega)*(FstOrDispDivTwo + omega*SndOrDispDivSix));
      
          } // end for-jj loop

} // ## end ComputeArtificialDispersionCompensationOperator

// ###########################################################

void NoiseCovariance::ApplyArtificialDispersionCompensation(
		      fftMatrices SignalData, bool DisentangleFlag)
{

  //static int CallNum = 1;

  if(DisentangleFlag)
   for(int jj=0;jj<SignalVecDim;jj++)
    SignalData.fOptSignalX[jj] *= ArtificialDispersionCompensationOperator[jj];
  else
   for(int jj=0;jj<SignalVecDim;jj++)
    SignalData.fOptSignalX[jj] /= ArtificialDispersionCompensationOperator[jj];

  fft->IFFT(SignalData.fOptSignalX,SignalData.tOptSignalX);


/**
  if(DisentangleFlag && CallNum%100 == 0)
  {

    char C_string[80];
    sprintf(C_string,"ArtDisp%i.dat",CallNum);
    string FileName = Job + C_string;

    WriteFileTimeField(FileName,SignalData);  

    
  }

    CallNum++;

    if(CallNum == 10)
        exit(1);

**/

} // ## end ApplyArtificialDispersionCompensation

// #########################################################

void NoiseCovariance::AmplifyCovarianceMatrix(double LinearGain)
{

  // Amplify covariance matrix

  DenseMatrixScale(CovarianceMatrix,2*CovMatrixDim,LinearGain);

  // Add lumped noise after amplification

  double NoiseVarianceRealOrImag = 
    Amplifier->GetSpectralNoisePowerDensityAddedPerPolarization()*
    Signal->GetDeltaFreq()/2.0;

   //cout << "LinearGain = " << LinearGain << endl;
   //cout << "NoiseVarianceRealOrImag = " <<  NoiseVarianceRealOrImag << endl;

                                 
  AddToPrincipalDiag(CovarianceMatrix,2*CovMatrixDim,NoiseVarianceRealOrImag);

} // ## end AmplifyCovarianceMatrix

// #######################################################

void NoiseCovariance::LumpedLossCovarianceMatrix(double LossLinear)
{

  // Attenuate covariance matrix. In LumpledLoss class when attenuate
  // LossLinear > 1, so we must scale by 1.0/LossLinear

  DenseMatrixScale(CovarianceMatrix,2*CovMatrixDim,1.0/LossLinear);

} // ## end LumpedLossCovarianceMatrix

// ############################################################

void NoiseCovariance::UpdateCovMatrixODE(void)
{

  // Compute CovarianceMatrix_new = PropagatorMatrix*CovarianceMatrix_old*
  //                               (PropagatorMatrix^T)

  DenseMatrixProduct(PropagatorMatrix,CovarianceMatrix,TempProductMatrix,
		     false,false,2*CovMatrixDim);

  DenseMatrixProduct(TempProductMatrix,PropagatorMatrix,CovarianceMatrix,
		     true,true,2*CovMatrixDim);


} // ## end NoiseCovariance::UpdateCovMatrixODE

// #############################################################

// added by Walter to extend CovarianceMethod to WDM

// #############################################################

void NoiseCovariance::Perturb(int PerturbedFreqIndex, 
			      double Delta, bool RealNotImag, int ChannelIndex)
{

  // This will need to be modified for VECTOR.
  
 double DisplacementCenterFrequency = Signal->GetRelativeCenterFrequency(ChannelIndex);
 int index_displacement =(int)
 floor(DisplacementCenterFrequency/Signal->GetDeltaFreq());
     
 PerturbedFreqIndex+=index_displacement;

  memcpy(DataToPropagate.fOptSignalX,NoiseFreeSignalStart.fOptSignalX,
	 SignalVecDim*sizeof(cplx));

  int Index_k = fft->index(PerturbedFreqIndex); 

  DataToPropagate.fOptSignalX[Index_k] += (RealNotImag) ? 
                                cplx(Delta,0.0) : cplx(0.0,Delta);

  fft->IFFT(DataToPropagate.fOptSignalX,DataToPropagate.tOptSignalX);

  // In a given fiber the first time we call Perturb we have Delta == 0
  // and we propagate the noise-free signal. Subsequent calls traverse
  // the same fiber.

 // JZ Nov 2003: THIS NEXT if/else IS BUGGY as cll Pertub three times at start.

  if(Delta == 0)
    {
     PropagatedLengthSaved = PropagatedLength;
     ZStepNumSaved = ZStepNum;
     //cout << "NoiseCovariance::Perturb: Delta == 0" << endl;
    }
  else
    {
      PropagatedLength = PropagatedLengthSaved;
      ZStepNum = ZStepNumSaved;
      //cout << "NoiseCovariance::Perturb: Delta != 0" << endl;

    }

  Fiber->Propagate(&PropagatedLength,&ZStepNum);

} // ## end NoiseCovariance::Perturb


// #############################################################

void NoiseCovariance::ApplyArtificialCentralChannelDemux(fftMatrices& PerturbationEnd)
   {
   // demux out all the channels of the WDM signal but the central channel
   
   // we have at least 3 channels because I have imposed odd number of channels
   // and becuase this method is called only if NumberOfChannels is >1
   
   // in the FFT-unshifted PerturbationEnd the central channel positive
   // frequencies are stored in the first NumPositiveFreq positions of the
   // PerturbationEnd array whilethe negative frequencies are stored in the
   // last NumNegativeFreq positions of PerturbationEnd array
   
    
   int NumPositiveFreq = 0;
   int NumNegativeFreq = 0;
   
    // ############## WARNING #######################
    // the following line would be the correct one but since I couldn't insert
    // it in the ocsOptSignal class I suppose that the user uses
    // EvenlySpacedChannels 
  
    //if (Signal->GetChannelAllocationMethod()==1)
    if (1)
       {
       // Evenly Spaced Channel Allocation
       double ChannelSpacing=Signal->GetChannSpacing();
       double delta_freq= Signal->GetDeltaFreq();
       NumPositiveFreq= (int) floor((ChannelSpacing/2)/delta_freq);  
       NumNegativeFreq= NumPositiveFreq;
       }
   else
       {
       //Either allocation by frequencies or wavelengths
       cerr << "NoiseCovariance::ApplyArtificialCentralChannelDemux() \n"
       << "So far only the Evenly Spaced Channel Allocation Method is implemented \n"
       << "ABORTING NOW !! \n";
       *LogFile << "NoiseCovariance::ApplyArtificialCentralChannelDemux() \n"
       << "So far only the Evenly Spaced Channel Allocation Method is implemented \n"
       << "ABORTING NOW !! \n";
       exit(1);
       
       // the following works well just if the channel are set up in the
       // following way:
       // 1)the channel with the lowest frequency is named Chann1 in Signal.in 
       // 2)the channel named Chann2 has the central freq. greater than Chann1
       // frequency
       // 3) so on 
       // In fact in this case the central channel of a 5 channel wdm signal
       // has index 2
       
       
       // Doing this we assume that the neihboring channels 
       //have the same bandwidth as the central one
 
       double CenterFreqRightChannel=
	 Signal->GetRelativeCenterFrequency((NumberOfChannels-1)/2+1); 
  
       double CenterFreqLeftChannel=
	 Signal->GetRelativeCenterFrequency((NumberOfChannels-1)/2-1);  
 
       double CenterFrequency=Signal->GetCenterFreq();
     
       double PositiveFreqBWCentralChannel=
	 (CenterFreqRightChannel-CenterFrequency)/2;

       double NegativeFreqBWCentralChannel=
	 (CenterFreqLeftChannel-CenterFrequency)/2;
   
       double delta_freq= Signal->GetDeltaFreq();
   
       NumPositiveFreq = (int) floor(PositiveFreqBWCentralChannel/delta_freq);
       NumNegativeFreq = (int) floor(NegativeFreqBWCentralChannel/delta_freq);
       }
    
   // filtering out all the channels except the central one and saving them
   // into fftMatrices PerturbationEnd_saved    
   
   for(int jj=0; jj < SignalVecDim; jj++)
      {
      if ((jj>NumPositiveFreq) && jj<(SignalVecDim-NumNegativeFreq))
         {
         PerturbationEnd_saved.fOptSignalX[jj] = 
	   PerturbationEnd.fOptSignalX[jj];

         PerturbationEnd.fOptSignalX[jj] = cplx(0.0,0.0);
	 }
      else
         {
	 // just to be sure (in case it haven't been set to zero before)
	 PerturbationEnd_saved.fOptSignalX[jj] = cplx(0.0,0.0);
         }
      }	  
   
   fft->IFFT(PerturbationEnd.fOptSignalX,PerturbationEnd.tOptSignalX);  
   
   // at this point in PerturbationEnd there is only the central channel
   // and outside its bandwidth there are just zero values
   
   // in  PerturbationEnd_saved.fOptSignalX[jj] I have saved all the
   // others channel but not the central one where there are zeros
            

   }  //end ApplyArtificialCentralChannelDemux(fftMatrices& PerturbationEnd)

// #############################################################

void NoiseCovariance::RestoreChannels(fftMatrices& PerturbationEnd)
   {
   // insert in PerturbationEnd signal, which has the jitter 
   // free central channel
   // the others channels which have still the jitter
   // it's the counter part of ApplyArtificialCentralChannelDemux(.)
   
   for(int jj=0; jj < SignalVecDim; jj++)
      {
      PerturbationEnd.fOptSignalX[jj] += PerturbationEnd_saved.fOptSignalX[jj];
      }	  	  
   fft->IFFT(PerturbationEnd.fOptSignalX,PerturbationEnd.tOptSignalX);

   
   }  //end RestoreChannels(fftMatrices& PerturbationEnd)     



