// $Id: PulseEvolutionAnalyzer.cc,v 1.2 2008/08/08 02:28:50 zweck Exp $


//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2001):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "PulseEvolutionAnalyzer.hh"

extern ofstream *LogFile;

PulseEvolutionAnalyzer::PulseEvolutionAnalyzer(OptSignal * PropagatingSignal2, 
                                               string InDir, string Job2,
                                               bool ComputeTimeFreqShifts2,
                                               bool ComputeTimingJitter2,
                                             bool ComputeMeanZerosStdDevOnes2,
					       int ComputeEnergyInBitSlot2)
{

  // We assume PropagatingSignal has one channel and that 
  // the channel frequency of the signal equals the center frequency
  // of the discretization of the signal (ie equals the carrier freq).

  ComputeTimeFreqShifts = ComputeTimeFreqShifts2;
  ComputeTimingJitter = ComputeTimingJitter2;
  ComputeMeanZerosStdDevOnes = ComputeMeanZerosStdDevOnes2;
  ComputeEnergyInBitSlot = ComputeEnergyInBitSlot2;

  PropagatingSignal = PropagatingSignal2;
  Job = Job2;


  if(ComputeTimeFreqShifts||ComputeTimingJitter||
     ComputeMeanZerosStdDevOnes || ComputeEnergyInBitSlot>=0)
      CompressedSignal = new OptSignal(*PropagatingSignal);


if(ComputeTimeFreqShifts)
 {
  TrailingPulse = new OptSignal(*PropagatingSignal);
       // So that memory allocation works

 
  // Get bit slots for leading and trailing pulses

  LeadingPulseBitSlot = 0;

  int kk = 0;

  while(!PropagatingSignal->oBitString->GetBit(0,kk))
    {
      kk++;
    }

  LeadingPulseBitSlot = kk;
  kk = kk+1;

  while(!PropagatingSignal->oBitString->GetBit(0,kk))
    {
      kk++;
    }

  TrailingPulseBitSlot = kk;
  
  if(LeadingPulseBitSlot >= PropagatingSignal->GetStringLength() ||
     TrailingPulseBitSlot >= PropagatingSignal->GetStringLength())
    {
      cerr << "Error in PulseEvolutionAnalyzer::PulseEvolutionAnalyzer"
	   << endl
	   << "ABORTING NOW!!" << endl;
      exit(1);
    }

  
  /*
  cout << "LeadingPulseBitSlot = " <<  LeadingPulseBitSlot << endl;
  cout << "TrailingPulseBitSlot = " << TrailingPulseBitSlot << endl;
  */
 }

  RanNumGen RNG = *(CompressedSignal->RNG);

  LinearFiber = new OptFiberLocalError(InDir + "LinearFiber.in",
                                       CompressedSignal,
                                       &RNG,Job);

  LinearFiber->SetFiberLength(1);


if(ComputeMeanZerosStdDevOnes || ComputeEnergyInBitSlot>=0)
  {

    
       HistogramZeros = new Histogram(1,0,1);
       HistogramOnes = new Histogram(1,0,1);
    

    string InFileName = InDir + "Startup.in";

    ifstream InFile(InFileName.c_str(),ios::in);   

    if (!InFile) 
     {
       cerr << "Error: Input file "<< InFileName 
            << " could not be opened."<< endl;
       exit(1);
     }

    LeftTimeZero = ReadDouble("LeftTimeZero"," ",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile);

    RightTimeZero = ReadDouble("RightTimeZero"," ",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile);

    LeftTimeOne = ReadDouble("LeftTimeOne"," ",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile);

    RightTimeOne = ReadDouble("RightTimeOne"," ",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile);

  } // ## end if ComputeMean...

  
} // ## end PulseEvolutionAnalyzer::PulseEvolutionAnalyzer

// ################################################

void PulseEvolutionAnalyzer::SetBitString(BitString * oBitString2)
{
  oBitString = new BitString(*oBitString2);
}



// ##############################################

void PulseEvolutionAnalyzer::ApplyPulseAnalyzer(double PropagatedLength2)
{
  PropagatedLength = PropagatedLength2;

  if(ComputeTimeFreqShifts||ComputeTimingJitter||
     ComputeMeanZerosStdDevOnes||ComputeEnergyInBitSlot>=0)
     LinearlyCompressPulses();

  if(ComputeTimeFreqShifts)
    {
     SeparatePulses();
     ComputeTimeShift();
     ComputeFreqShift();
     ComputeMaxPower();
     WriteTimeFreqShifts(Job+"TimeFreqShifts.dat");
    }

  if(ComputeTimingJitter)
    {
     
    }

  if(ComputeMeanZerosStdDevOnes)
    {
      ComputePulseEnergyStats();
      WriteMeanZerosStdDevOnes(Job + "ZerosOnesEvolution.dat");
    }

  if(ComputeEnergyInBitSlot>=0)
     WriteEnergyInBitSlot(Job + "EnergyBitSlotEvolution.dat");


} // end PulseEvolutionAnalyzer::ApplyPulseAnalyzer

// ##################################################

void PulseEvolutionAnalyzer::LinearlyCompressPulses(void)
{

   *CompressedSignal = *PropagatingSignal;

   LinearFiber->SetFstOrDispFiber(
               -CompressedSignal->GetAccumulatedFirstOrderDispersion());
  
   LinearFiber->SetSndOrDispFiber(
               -CompressedSignal->GetAccumulatedSecondOrderDispersion());
  
   double PropagatedLength3 = 0;
   int StepNum = 0;

   LinearFiber->Propagate(&PropagatedLength3,&StepNum);


   static int AppendFlag = 0;

   // CompressedSignal->WriteDistTimePowerPhase(Job + "CompressedDTP.dat", 
   //                                      1,PropagatedLength,AppendFlag);

   AppendFlag = 1;

} // ## end LinearlyCompressPulses

// #######################################

void PulseEvolutionAnalyzer::SeparatePulses(void)
{

  // We assume that we have two pulses which are in the Leading
  // and Trailing bit slots.

  IsolatePulse(CompressedSignal,TrailingPulse,TrailingPulseBitSlot);
  IsolatePulse(CompressedSignal,CompressedSignal,LeadingPulseBitSlot);

  /*
  CompressedSignal->WriteFileTimePower(Job+"tLead.dat");
  CompressedSignal->WriteFileFreqPower(Job+"fLead.dat",0);
  TrailingPulse->WriteFileTimePower(Job+"tTrail.dat");
  TrailingPulse->WriteFileFreqPower(Job+"fTrail.dat",0);
  */

} // end SeparatePulses

// #####################################


void PulseEvolutionAnalyzer::IsolatePulse(OptSignal * FullSignal,
                                          OptSignal *SinglePulse, 
                                          int BitSlotSinglePulse)
{

  // We use a mask to isolate the pulse

  *SinglePulse = *FullSignal;

  int PointsPerBit = int(floor(SinglePulse->GetQtPoints()/
                               float(SinglePulse->GetStringLength())));


  int Left_ii = BitSlotSinglePulse*PointsPerBit;
  int Right_ii = Left_ii + PointsPerBit-1;

  /*
  cout << "PointsPerBit = " << PointsPerBit << endl;
  cout << "BitSlotSinglePulse = " << BitSlotSinglePulse << endl;
  cout << "Left_ii,Right_ii = " << Left_ii << " " << Right_ii << endl;
  */

  for(int ii=0;ii<SinglePulse->GetQtPoints();ii++)
    {
      if(ii<=Left_ii || ii > Right_ii)
	{
          SinglePulse->sfftM.tOptSignalX[ii] = 0.0;
          SinglePulse->sfftM.tOptSignalY[ii] = 0.0;
	}
    }// end for


  SinglePulse->Time2Freq();

}// ## end IsolatePulse

// ###########################################

void PulseEvolutionAnalyzer::ComputeTimeShift(void)
{

  double TimeWindow = CompressedSignal->GetTimeWindow();

  LeadPulseTimeShift = CompressedSignal->ComputeMeanTimePulse(0,TimeWindow);
  TrailPulseTimeShift = TrailingPulse->ComputeMeanTimePulse(0,TimeWindow);

  LeadPulseTimeShift -= (LeadingPulseBitSlot+0.5)*
                         CompressedSignal->GetBitLength();

  TrailPulseTimeShift -= (TrailingPulseBitSlot+0.5)*
                          CompressedSignal->GetBitLength();

} // ## end ComputeTimeShift

// ################################################

void PulseEvolutionAnalyzer::ComputeFreqShift(void)
{

  LeadPulseFreqShift = CompressedSignal->ComputeMeanFreqSignal();
  TrailPulseFreqShift = TrailingPulse->ComputeMeanFreqSignal();

}

// ##########################################

void PulseEvolutionAnalyzer::ComputeMaxPower(void)
{
  LeadPulseMaxPower = CompressedSignal->ComputeMaxPower();
  TrailPulseMaxPower = TrailingPulse->ComputeMaxPower();

  double PulseEnergy = CompressedSignal->GetTotalEnergy();

  LeadPulseMaxPower = LeadPulseMaxPower/PulseEnergy;
  TrailPulseMaxPower = TrailPulseMaxPower/PulseEnergy;

}

// ########################################

void  PulseEvolutionAnalyzer::WriteTimeFreqShifts(string OutFileName)
{

  static bool FirstTimeCalled = 1;

  ofstream OutFile;

  if(FirstTimeCalled)
    {
      OutFile.open(OutFileName.c_str(),ios::out);
      OutFile << "% PropagatedLength [km], LeadPulseTimeShift [ps], "
	      << "TrailPulseTimeShift [ps], "
	      << "LeadPulseFreqShift [GHz],TrailPulseFreqShift [GHz]" 
	      << " LeadPulseMaxPower [mW], TrailPulseMaxPower [mW]"
              << endl;

      FirstTimeCalled = 0;
    }
  else
    OutFile.open(OutFileName.c_str(),ios::app);

  OutFile << 1e-3*PropagatedLength << "\t" << 1e12*LeadPulseTimeShift << "\t" 
	  << 1e12*TrailPulseTimeShift << "\t" 
          << 1e-9*LeadPulseFreqShift << "\t"
	  << 1e-9*TrailPulseFreqShift << "\t"
          << 1e3*LeadPulseMaxPower << "\t"
          << 1e3*TrailPulseMaxPower << endl;

  OutFile.close();

}

// ## end WriteTimeFreqShifts

// ##############################################

void  PulseEvolutionAnalyzer::WriteMeanZerosStdDevOnes(string OutFileName)
{

  static bool FirstTimeCalled = 1;

  ofstream OutFile;

  if(FirstTimeCalled)
    {
      OutFile.open(OutFileName.c_str(),ios::out);
      OutFile << "% PropagatedLength [km], Rel MeanZeros, Rel StdDevOnes"
              << endl;

      FirstTimeCalled = 0;
    }
  else
    OutFile.open(OutFileName.c_str(),ios::app);

  OutFile << 1e-3*PropagatedLength << "\t" 
          << HistogramZeros->GetMean()/HistogramOnes->GetMean() << "\t" 
          << HistogramOnes->GetStdDev()/HistogramOnes->GetMean() << endl;

  OutFile.close();

}

// ## end WriteMeanZerosStdDevOnes

// ##############################################

void PulseEvolutionAnalyzer::ComputePulseEnergyStats(void)
{

// This method computes the mean and std_dev of the energy in the 0's and 1's
// in a signal. It was designed to be used for my FWM research.

// Because of the Wiggling we need to pass the Bitstring separately
// from the Signal. 

// We assume that there is only 1 channel in the BitString

// We assume that  Startup.in contains variables:
// LeftTimeZero, RightTimeZero in range [0,BitLength]
// LeftTimeOne, RightTimeOne in range [0,BitLength]

int StringLength = CompressedSignal->GetStringLength();
double TimeWindow =  CompressedSignal->GetTimeWindow();

HistogramOnes->ClearHistogram();
HistogramZeros->ClearHistogram();

// Using RegenerateSignalString() the pulses are centered on 
// (kk+0.5)*BitLength


for(int kk=0;kk < StringLength;kk++)
 {

   double PulseStartTime = kk*CompressedSignal->GetBitLength();

  if(oBitString->GetBit(0,kk)) // Bit is a 1
   {
    
     double LeftTime = PulseStartTime + LeftTimeOne;
     double RightTime = PulseStartTime + RightTimeOne;
     
     HistogramOnes->AddSample(CompressedSignal->GetEnergyInTimeRange(
                                         LeftTime,RightTime));
      
   } // end Bit is a 1

  else // Bit is a 0
    {

     double LeftTime = PulseStartTime + LeftTimeZero;
     double RightTime = PulseStartTime + RightTimeZero;
     
     HistogramZeros->AddSample( CompressedSignal->GetEnergyInTimeRange(
                                         LeftTime,RightTime));

    } // end Bit is  0


 } // end kk loop


} // ## end UpdatePulseEnergyStats

// #############################################################

void  PulseEvolutionAnalyzer::WriteEnergyInBitSlot(string OutFileName)
{

  static bool FirstTimeCalled = 1;

  ofstream OutFile;

  if(FirstTimeCalled)
    {
      OutFile.open(OutFileName.c_str(),ios::out);
      OutFile << "% PropagatedLength [km], EnergyInBitSlot"
              << endl;



      FirstTimeCalled = 0;
    }
  else
    OutFile.open(OutFileName.c_str(),ios::app);


   double ZeroPulseStartTime = ComputeEnergyInBitSlot*
                           CompressedSignal->GetBitLength();

   double ZeroLeftTime = ZeroPulseStartTime + LeftTimeZero;
   double ZeroRightTime = ZeroPulseStartTime + RightTimeZero;


   // cout << "Left, Right = " << LeftTime << " " << RightTime << endl;

  OutFile << 1e-3*PropagatedLength << "\t" 
          
          << CompressedSignal->GetEnergyInTimeRange(
                                         ZeroLeftTime,ZeroRightTime)/
             CompressedSignal->GetTotalEnergy()  << "\t"
	  << CompressedSignal->GetEnergyInTimeRange(
                               ZeroLeftTime,ZeroRightTime) << "\t"
	  << CompressedSignal->GetTotalEnergy() 
<< endl;

  OutFile.close();

}

// ############################################################








