// #####################################################################
//
//    cfftw3.cc
//
//    Methods module for the class of objects used as
//    an interface to the FFTW toolkit.
//
//    Copyright:
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
// #####################################################################
//    Author: Ronald Holzloehner  <holzloehner@umbc.edu>
//
//    Adapted by John Zweck, Feb 12, 2003 to make code 
//    conform to the Physics convention
//
//    Adapted August 2008 to be compatible with fftw3
//
// #####################################################################

#include "cfftw.hh"


#ifdef _USING_FFTW3_


#include <string>
   using namespace std;   

// ################### Methods for the class cfftw
//  Implements interface to the FFTW toolkit
//  (Fast Fourier Transforms)
//  Ronald Holzloehner, 11 Nov 99


/*! Constructor.
 
    \param N_in integer size of discrete Fourier transform
    \param T_in Total time window in seconds

 */

cfftw::cfftw(int N_in, double T_in)
{

  FILE *input_file;

  N = N_in;  
  N_half = N/2;  
  T = T_in; // T is total time window

  delta_t = T / double(N);  // time step
  delta_f = 1.0 / T;        // frequency step

  F = N * delta_f;  // F is total frequency range

  fftw_in = (fftw_complex*) NewCplx(N);
  fftw_out = (fftw_complex*) NewCplx(N);

  work = (cplx*) fftw_in;

  string FileNameFFTWwisdom = "FFTWwisdom";

  string Suffix;

  if(!getenv("HOSTNAME"))
  {
   cerr << "WARNING:" 
       	<< "cfftw::cfftw: Can't read environment variable HOSTNAME"
        << endl;
  }
  else
  {   
   Suffix = getenv("HOSTNAME");
   FileNameFFTWwisdom += "." + Suffix;
   }

  input_file = NULL; 
  input_file=fopen(FileNameFFTWwisdom.c_str(),"r");	

  if(input_file==NULL)
   {
    cout<<"The file " <<FileNameFFTWwisdom
	<<" is missing or not valid wisdom file."<<
	"\nFFTW_MEASURE procedure will be performed and a new"
	<< " file will be generated with this name."<<endl;

   }	
  else
   {
    fftw_import_wisdom_from_file(input_file);
    fclose(input_file);
   }

	
  forw  = fftw_plan_dft_1d(N,fftw_in,fftw_out,FFTW_FORWARD,
                             FFTW_MEASURE);

  backw  = fftw_plan_dft_1d(N,fftw_in,fftw_out,FFTW_BACKWARD,
                              FFTW_MEASURE);

	
  input_file=fopen(FileNameFFTWwisdom.c_str(),"w");
  fftw_export_wisdom_to_file(input_file);
  fclose(input_file);


  if(!forw || !backw)
  {
    cout << "\n cfftw::cfftw(): FFTW allocation failed!" << endl;
    exit(-1);
  }

  shift_len2 = N/2;  // if datlen_l is even, len1==len2,
  shift_len1 = N - shift_len2;  // otherwise  len1 == len2 + 1

  shift_work = NewCplx(shift_len1+1);

}


cfftw::~cfftw()
{
  fftw_destroy_plan(forw);
  fftw_destroy_plan(backw);

  work = NULL;
  FreeCplx(shift_work);
  FreeCplx((cplx*)fftw_in);
  FreeCplx((cplx*)fftw_out);

}

// ##### FFT forward transform
// CAUTION: No fftshift() is done!!
// FFT(IFFT()) and vice versa are the identity
// frequencies go like
// (i < N/2)? (i):(-N+i)

/*!

   Transforms from time to frequency using the Physics convention

   FreqVector = FFT(TimeVector) := 1/N IFFT_fftw (TimeVector)

   where IFFT_fftw is the inverse Fourier transform in FFTW.

   Note that cfftw::fftshift() is not called by cfftw::FFT()

   -# FFT(IFFT(FreqVector)) = FreqVector
   -# IFFT(FFT(TimeVector)) = TimeVector

   \param in TimeVector
   \param out FreqVector

 */

void cfftw::FFT(cplx *in, cplx *out)
{
  in = ((in)?in:(cplx*)fftw_in);
  out = ((out)?out:(cplx*)fftw_out);
  fftw_execute_dft(backw, (fftw_complex*)in, (fftw_complex*)out);  

  cplx *p = out;
  cplx *const end_p = &(out[N]);

  for(; p < end_p; p++)  // normalize
    *p *= 1.0/double(N);

  // NO FFTSHIFT() HERE !!

}

// ##### FFT backward transform
// CAUTION: No fftshift() is done!!
// IFFT(FFT()) and vice versa is the identity

/*!

   Transforms from frequency to time using the Physics convention

   TimeVector = IFFT(FreqVector) := FFT_fftw (FreqVector)

   where FFT_fftw is the Fourier transform in FFTW.

   Note that cfftw::fftshift() is not called by cfftw::IFFT()

   -# FFT(IFFT(FreqVector)) = FreqVector
   -# IFFT(FFT(TimeVector)) = TimeVector

   \param in FreqVector
   \param out TimeVector


 */

void cfftw::IFFT(cplx *in, cplx *out)
{
  in = ((in)?in:(cplx*)fftw_in);
  out = ((out)?out:(cplx*)fftw_out);  

  fftw_execute_dft(forw, (fftw_complex*)in, (fftw_complex*)out);

  // NO FFTSHIFT() HERE !!
}

// ##### cfftw::fftshift() works like its namesake in MATLAB
// and swaps the upper and lower halves of a vector
// If vector length is odd, first "half" is LARGER than second
// CAUTION: behavior is different from sort()!!

/*!
cfftw::fftshift() works like its namesake in MATLAB.

It swaps the upper and lower halves of cplx vector called in.
If vector length is odd, first "half" is LARGER than second.
 */

void cfftw::fftshift(cplx *in)
{
  in = ((in)?in:(cplx*)fftw_in);

  // save first half (length len1)

  memcpy(shift_work, in, sizeof(cplx)*shift_len1);    

 // shift second half up

  memcpy(in, in+shift_len1,sizeof(cplx)*shift_len2); 

  // copy first half back

  memcpy(in+shift_len2, shift_work ,sizeof(cplx)*shift_len1); 
}

// ##### Ron_fftw::ifftshift() works like its namesake in MATLAB
// and swaps the upper and lower halves of a vector
// Undoes effect of fftshift()
// If vector length is odd, first "half" is SHORTER than second
// CAUTION: behavior is different from resort()!!

/*!

cfftw::ifftshift() works like its namesake in MATLAB.

It swaps the upper and lower halves of cplx vector called in
Undoes effect of cfftw::fftshift()
If vector length is odd, first "half" SHORTER than second.
 */


void cfftw::ifftshift(cplx *in)
{
  in = ((in)?in:(cplx*)fftw_in);

  // save second half (length len1)

  memcpy(shift_work, in+shift_len2, sizeof(cplx)*shift_len1);    

  // shift first half down

  memcpy(in+shift_len1, in, sizeof(cplx)*shift_len2); 

  // copy second half back

  memcpy(in,shift_work ,sizeof(cplx)*shift_len1); 

}

#endif /* _USING_FFTW3_ */
// #####################################################################
//
//    cfftw.cc
//
//    Methods module for the class of objects used as
//    an interface to the FFTW toolkit.
//
//    Copyright:
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
// #####################################################################
//    Author: Ronald Holzloehner  <holzloehner@umbc.edu>
//
//    Adapted by John Zweck, Feb 12, 2003 to make code 
//    conform to the Physics convention
//
// #####################################################################


#ifndef _USING_FFTW3_

#include <string>
   using namespace std;  // Required for Sue's code to compile 

// ################### Methods for the class cfftw
//  Implements interface to the FFTW toolkit
//  (Fast Fourier Transforms)
//  Ronald Holzloehner, 11 Nov 99


/*! Constructor.
 
    \param N_in integer size of discrete Fourier transform
    \param T_in Total time window in seconds

 */

cfftw::cfftw(int N_in, double T_in)
{

 int method = FFTW_USE_WISDOM;

	FILE *input_file;
	bool WIS_EXIST;

  N = N_in;  
  N_half = N/2;  
  T = T_in; // T is total time window

  delta_t = T / double(N);  // time step
  delta_f = 1.0 / T;        // frequency step

  F = N * delta_f;  // F is total frequency range

        //## Change by Ivan Lima: now the FFTWwisdom is loaded from
        //#  a separate file for each computer in the network 
 
        string FileNameFFTWwisdom = "FFTWwisdom";

        string Suffix;

        if(!getenv("HOSTNAME"))
	  {
	   cerr << "WARNING:" 
		<< "cfftw::cfftw: Can't read environment variable HOSTNAME"
	        << endl;
	  }
        else
	  {   
           Suffix = getenv("HOSTNAME");
           FileNameFFTWwisdom += "." + Suffix;
	  }

	if(method==FFTW_USE_WISDOM)
	{

		input_file=fopen(FileNameFFTWwisdom.c_str(),"r");	
		
		if(FFTW_FAILURE==fftw_import_wisdom_from_file(input_file))
		{
			cout<<"The file " <<FileNameFFTWwisdom
			<<" is missing or not valid wisdom file."<<
			"\nFFTW_MEASURE procedure will be performed and a new"
			<< " file will be generated with this name."<<endl;
		
			WIS_EXIST=false;
		}	
		else
		{
			WIS_EXIST=true;
			if(input_file!=NULL)
			fclose(input_file);
		}
	
		forw  = fftw_create_plan(N, FFTW_FORWARD, 
					 FFTW_MEASURE |FFTW_USE_WISDOM );
		backw = fftw_create_plan(N, FFTW_BACKWARD, 
					 FFTW_MEASURE |FFTW_USE_WISDOM);
	
	if(WIS_EXIST==false)
	{
	         input_file=fopen(FileNameFFTWwisdom.c_str(),"w");

		fftw_export_wisdom_to_file(input_file);
		fclose(input_file);
	}	

	}
	else
	{
		forw  = fftw_create_plan(N, FFTW_FORWARD,  method);
		backw = fftw_create_plan(N, FFTW_BACKWARD, method);
	}

  if(!forw || !backw)
  {
    cout << "\n cfftw::cfftw(): FFTW allocation failed!" << endl;
    exit(-1);
  }

  //work = new cplx[N];

  work = NewCplx(N);

  shift_len2 = N/2;  // if datlen_l is even, len1==len2,
  shift_len1 = N - shift_len2;  // otherwise  len1 == len2 + 1

  //shift_work = new cplx[shift_len1+1];

  shift_work = NewCplx(shift_len1+1);

}


cfftw::~cfftw()
{
  fftw_destroy_plan(forw);
  fftw_destroy_plan(backw);
  // delete [] work;
  // delete [] shift_work;

  FreeCplx(work);
  FreeCplx(shift_work);

}

// ##### FFT forward transform
// CAUTION: No fftshift() is done!!
// FFT(IFFT()) and vice versa are the identity
// frequencies go like
// (i < N/2)? (i):(-N+i)

/*!

   Transforms from time to frequency using the Physics convention

   FreqVector = FFT(TimeVector) := 1/N IFFT_fftw (TimeVector)

   where IFFT_fftw is the inverse Fourier transform in FFTW.

   Note that cfftw::fftshift() is not called by cfftw::FFT()

   -# FFT(IFFT(FreqVector)) = FreqVector
   -# IFFT(FFT(TimeVector)) = TimeVector

   \param in TimeVector
   \param out FreqVector

 */

void cfftw::FFT(cplx *in, cplx *out)
{
  in = ((in)?in:work);
  out = ((out)?out:work);
  fftw_one(backw, (fftw_complex*)in, (fftw_complex*)out);  

  cplx *p = out;
  cplx *const end_p = &(out[N]);

  for(; p < end_p; p++)  // normalize
    *p *= 1.0/double(N);

  // NO FFTSHIFT() HERE !!

}

// ##### FFT backward transform
// CAUTION: No fftshift() is done!!
// IFFT(FFT()) and vice versa is the identity

/*!

   Transforms from frequency to time using the Physics convention

   TimeVector = IFFT(FreqVector) := FFT_fftw (FreqVector)

   where FFT_fftw is the Fourier transform in FFTW.

   Note that cfftw::fftshift() is not called by cfftw::IFFT()

   -# FFT(IFFT(FreqVector)) = FreqVector
   -# IFFT(FFT(TimeVector)) = TimeVector

   \param in FreqVector
   \param out TimeVector


 */

void cfftw::IFFT(cplx *in, cplx *out)
{
  in = ((in)?in:work);
  out = ((out)?out:work);  

  fftw_one(forw, (fftw_complex*)in, (fftw_complex*)out);

  // NO FFTSHIFT() HERE !!
}

// ##### cfftw::fftshift() works like its namesake in MATLAB
// and swaps the upper and lower halves of a vector
// If vector length is odd, first "half" is LARGER than second
// CAUTION: behavior is different from sort()!!

/*!
cfftw::fftshift() works like its namesake in MATLAB.

It swaps the upper and lower halves of cplx vector called in.
If vector length is odd, first "half" is LARGER than second.
 */

void cfftw::fftshift(cplx *in)
{
  in = ((in)?in:work);

  // save first half (length len1)

  memcpy(shift_work, in, sizeof(cplx)*shift_len1);    

 // shift second half up

  memcpy(in, in+shift_len1,sizeof(cplx)*shift_len2); 

  // copy first half back

  memcpy(in+shift_len2, shift_work ,sizeof(cplx)*shift_len1); 
}

// ##### Ron_fftw::ifftshift() works like its namesake in MATLAB
// and swaps the upper and lower halves of a vector
// Undoes effect of fftshift()
// If vector length is odd, first "half" is SHORTER than second
// CAUTION: behavior is different from resort()!!

/*!

cfftw::ifftshift() works like its namesake in MATLAB.

It swaps the upper and lower halves of cplx vector called in
Undoes effect of cfftw::fftshift()
If vector length is odd, first "half" SHORTER than second.
 */


void cfftw::ifftshift(cplx *in)
{
  in = ((in)?in:work);

  // save second half (length len1)

  memcpy(shift_work, in+shift_len2, sizeof(cplx)*shift_len1);    

  // shift first half down

  memcpy(in+shift_len1, in, sizeof(cplx)*shift_len2); 

  // copy second half back

  memcpy(in,shift_work ,sizeof(cplx)*shift_len1); 

}

// in:	pointer to field
// T_half:  50% - point of window
// T_001:   1%  - point of window, measured from T_half

/*!

Not used by PhoSSiL

 */


void cfftw::timewindow(cplx *in, double T_half, double T_001)
{ // 3.0 is a heuristic limit
  int num = int(rint(3.0*max(T_half, T_001)/T*double(N)));
  if(num > N/2)
    num = N/2;
  // CAUTION! Remember that t=+-T/2 lies in the middle of the data vector
  cplx *field_p1 = in + N/2;
  cplx *field_p2 = field_p1-1;
  cplx *end_p = field_p1 + num;
  //  Remark: 0.5*(tanh(T_001/T_slope)+1.0) == 0.01
// ********* #warning T_0018 instead of T_001
//  double T_slope = fabs(T_001) / 2.2975599; // 2.2975599 == | atanh(2*0.01-1) |
  double T_slope = fabs(T_001) / 1.9996; // 1.9996 == | atanh(2*0.018-1) |
  double t_l = -T_half;
  for(; field_p1 < end_p; field_p1++, field_p2--, t_l += delta_t)
  {
    double val = (tanh(t_l/T_slope)+1.0)*0.5;
    *field_p1 *= val;
    *field_p2 *= val;
  }
}

#endif /* ifndef _USING_FFTW3_ */


