// $Id:
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsBeamSplitter.hh"

// ################### Methods for the class BeamSplitter      
// #### Initializes the object BeamSplitter
// ## By: Ivan Lima
BeamSplitter::BeamSplitter( OptSignal *oOptSignalATmp
                           ,OptSignal *oOptSignalBTmp)
// ## It is assumed that the original signal is in oOptSignalATmp. 
// #  It is important that oOptSignalBTmp have the same # of points
// #  and the same # of bits of oOptSignalATmp.
{
   oOptSignalA = oOptSignalATmp;
   oOptSignalB = oOptSignalBTmp; 
   if (oOptSignalA->GetQtPoints() 
       != oOptSignalB->GetQtPoints() ) {
      cerr << "Error: The two OptSignal objects" 
           << "should have the same size."
           << endl;
      exit(1);
   }         
   qtPoints = oOptSignalA->GetQtPoints();        
   
   // ## Sets the default power splitting factor as 0.5,
   // #  that means equal parts.
   PowerSplittingFactor = 0.5;
}

// #### Defines the new splitting polarization of the polarizer
// ## By: Ivan Lima
void BeamSplitter::SetPowerSplittingFactor(double PowerSplittingFactorTmp)
{
   PowerSplittingFactor = PowerSplittingFactorTmp;
   if (PowerSplittingFactor < 0 || PowerSplittingFactor > 1) {
      cerr << "Error: The parameter PowerSplittingFactor can only "<<endl 
           << "be in the range [0,1]. The value selected was: "
           << PowerSplittingFactor
           << endl;
      exit(1);
   }                        
}                       


// #### Split the optical signal
// ## By: Ivan Lima
void BeamSplitter::SplitOptSignal()
// ## The object oOptSignalA contains the original signal.
// ## oOptSignalA will end up with the component of 
// #  the signal with the power proportional to the power 
// #  splitting factor.
// #  object oOptSignalB will end up with the complement of
// #  the power originally contained in oOptSignalA. 
{   
   cout << "PowerSplittingFactor = " << PowerSplittingFactor << endl; 
   cplx Jones[2];
   for (int ii = 0; ii < qtPoints; ii++) {
      Jones[0] = oOptSignalA->sfftM.tOptSignalX[ii];
      Jones[1] = oOptSignalA->sfftM.tOptSignalY[ii];      
      oOptSignalA->sfftM.tOptSignalX[ii] = 
                    sqrt(PowerSplittingFactor)*Jones[0];   
      oOptSignalA->sfftM.tOptSignalY[ii] = 
                    sqrt(PowerSplittingFactor)*Jones[1];
      oOptSignalB->sfftM.tOptSignalX[ii] = 
                  sqrt(1. - PowerSplittingFactor)*Jones[0];   
      oOptSignalB->sfftM.tOptSignalY[ii] = 
                  sqrt(1. - PowerSplittingFactor)*Jones[1];
   }   
   oOptSignalA->Time2Freq();
   oOptSignalB->Time2Freq();
}



// #### Split the optical signal in the scalar model
// ## By: Ivan Lima
void BeamSplitter::SplitOptSignalScalar()
// ## The object oOptSignalA contains the original signal.
// ## oOptSignalA will end up with the component of 
// #  the signal with the power proportional to the power 
// #  splitting factor.
// #  object oOptSignalB will end up with the complement of
// #  the power originally contained in oOptSignalA. 
{   
   cplx Jones[2];
   for (int ii = 0; ii < qtPoints; ii++) {
      Jones[0] = oOptSignalA->sfftM.tOptSignalX[ii];
      Jones[1] = oOptSignalA->sfftM.tOptSignalY[ii];      
      oOptSignalA->sfftM.tOptSignalX[ii] = 
                        PowerSplittingFactor*Jones[0];   
      oOptSignalB->sfftM.tOptSignalX[ii] = 
                        (1. - PowerSplittingFactor)*Jones[0];   
   }   
   oOptSignalA->Time2FreqScalar();
   oOptSignalB->Time2FreqScalar();
   int StringLength = oOptSignalA->GetStringLength();
   for (int ii = 0; ii < StringLength; ii++) {
      oOptSignalB->oBitString->SetBit(0,ii,
         oOptSignalA->oBitString->GetBit(0,ii) );
   }      
}

               
