// $Id: ocsElecEqualizer.cc,v 1.2 2006/05/11 22:29:20 zweck Exp $
//###################################################################
//
//       Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################
#include "ocsElecEqualizer.hh"
extern ofstream *LogFile;

// ################### Methods for the class Polarizer                    

// ##########################################################
// #### Constructor of ElecEqualizer with initialization
// #    from a file
// ##   By: Aurenice Lima
ElecEqualizer::ElecEqualizer(string InFileName, 
                             ElecSignalStat *oElecSignalStat2)
{
   IsComplexEqualization = 0;
   InitElecEqualizer(InFileName, oElecSignalStat2);
}


// ##########################################################
// #### Constructor of Complex ElecEqualizer with initialization
// #    from a file
// ##   By: Aurenice Lima
ElecEqualizer::ElecEqualizer(string InFileName, 
                             ElecSignalStat *oElecSignalStat2,
                             ElecSignalStat *oElecSignalStatIm2)
{
   IsComplexEqualization = 1;
   InitElecEqualizer(InFileName, oElecSignalStat2);
   oElecSignalStatIm = oElecSignalStatIm2;
}


// ##########################################################
// #    Actual initialization of the object
// ##   By: Aurenice Lima
void ElecEqualizer::InitElecEqualizer(string InFileName, 
                             ElecSignalStat *oElecSignalStat2)
// ## EqualizationTap must be between 1 and qtInputTaps
// ## EqualizationTap == qtInputTaps corresponds to the 
// ## Multipath case                             
{
   ifstream InFile(InFileName.c_str(),ios::in);       
   if (!InFile) 
   {
      cout << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
   }

   oElecSignalStat = oElecSignalStat2;
   
   StringLength = oElecSignalStat->GetStringLength();

   LogFileSeparator();
   *LogFile << InFileName << endl;

   LogFileSeparator();   
   *LogFile <<"Initializing object of class ElecEqualizer" 
           << endl << endl;

   qtInputTaps  = ReadInt("qtInputTaps"," ",1,-1,LOWER_ONLY,
                          MANDATORY,&InFile);                          
//   *LogFile << "qtInputTaps = " << qtInputTaps << endl;                          
   EqualizationTap = ReadInt("EqualizationTap"," ",1,qtInputTaps,LOWER_AND_UPPER,
                             MANDATORY,&InFile);
//   *LogFile << "EqualizationTap = " << EqualizationTap << endl;
   FeedBackCorrectedBits = ReadInt("FeedBackCorrectedBits"," ",0,1,LOWER_AND_UPPER,
                             MANDATORY,&InFile);
//   *LogFile << "FeedBackCorrectedBits = " << FeedBackCorrectedBits << endl;                             
   UseFinalWeights = ReadInt("UseFinalWeights"," ",0,1,LOWER_AND_UPPER,
                          MANDATORY,&InFile);
//   *LogFile << "UseFinalWeights = " << UseFinalWeights << endl;
   NormalizeWeights = ReadInt("NormalizeWeights"," ",0,1,LOWER_AND_UPPER,
                          MANDATORY,&InFile);
//   *LogFile << "NormalizeWeights = " << NormalizeWeights << endl;                          
   SymmetricEqualization = ReadInt("SymmetricEqualization"," ",0,1,LOWER_AND_UPPER,
                          MANDATORY,&InFile);
//   *LogFile << "SymmetricEqualization = " << SymmetricEqualization << endl;  
   UseExactDesired = ReadInt("UseExactDesired"," ",0,1,LOWER_AND_UPPER,
                          OPTIONAL_WARNING,&InFile);                                                    
//   *LogFile << "UseExactDesired = " << UseExactDesired << endl;                          
   
   if (FeedBackCorrectedBits == 1) {
      *LogFile << "From ElecEqualizer: Error "
              << "Option FeedBackCorrectedBits "
              << "is not available at this time." <<endl;
      cerr << "Error! Look at the log file (*.log)." << endl;        
      exit(1);           
   }   

   if (NormalizeWeights == 1 && UseFinalWeights != 1) {
      *LogFile << "From ElecEqualizer: Error "
              << "If NormalizeWeights == 1 => UseFinalWeights "
              << "should also be equal to 1" <<endl;
      cerr << "Error! Look at the log file (*.log)." << endl;        
      exit(1);      
   }
   
   if (StringLength < qtInputTaps) {
      *LogFile << "From ElecEqualizer."  << endl
              << "Error: StringLength < qtInputTaps " << endl
              << "   StringLength " << StringLength << endl
              << "   qtInputTaps     " << qtInputTaps << endl;
      exit(1);
   }

   InFile.close();

   AllocateMemory();
   //FixBack2Back = 1; //## default
   
}


// ##########################################################
// #### Memory allocation for the ElecEqualizer
// ##   By: Aurenice Lima
void ElecEqualizer::AllocateMemory(void)
{   
    if (EqualizationTap < 1 || EqualizationTap > qtInputTaps)
    {
       *LogFile << "ERROR: EqualizationTap = " 
               << EqualizationTap << endl  
               << "It  must be  between 1 and qtInputTaps."
               << endl << "ABORTING NOW"
               << endl; 
       exit(1);   
    }      
   // #### For the Wiener solution
   CorrelationVectorQtAverage = new int[qtInputTaps];    
   if (! IsComplexEqualization) {
      CrossCorrelationVector = new double[qtInputTaps];  //double[sq(qtInputTaps)]; 
      CorrelationVector = new double[qtInputTaps];  
      TapWeight      = new double[qtInputTaps];          
      TapWeightTmp = dmatrix(1, qtInputTaps, 1, 1);
      CorrelationMatrix = dmatrix(1, qtInputTaps, 1, qtInputTaps);      
   }
   else {
      CrossCorrelationVectorCplx = new cplx[qtInputTaps];   //cplx[sq(qtInputTaps)]; 
      CorrelationVectorCplx = new cplx[qtInputTaps];   
      TapWeightCplx      = new cplx[qtInputTaps];          
      TapWeightTmpCplx = cmatrix(1, qtInputTaps, 1, 1);
      CorrelationMatrixCplx = cmatrix(1, qtInputTaps, 1, qtInputTaps);     
   }      
   ResetEqualizer();  
}


// ##########################################################
// #### Destructor of ElecEqualizer
// ##   By: Aurenice Lima
ElecEqualizer::~ElecEqualizer()
{
   delete [] CorrelationVectorQtAverage; 
   if (! IsComplexEqualization) {
      delete [] TapWeight;
      delete [] CrossCorrelationVector;   
      delete [] CorrelationVector;
      free_dmatrix(CorrelationMatrix, 1, qtInputTaps, 1, qtInputTaps);   
      free_dmatrix(TapWeightTmp, 1, qtInputTaps, 1, 1);    
   }
   else {
      delete [] TapWeightCplx;
      delete [] CrossCorrelationVectorCplx;   
      delete [] CorrelationVectorCplx;
      free_cmatrix(CorrelationMatrixCplx, 1, qtInputTaps, 1, qtInputTaps);   
      free_cmatrix(TapWeightTmpCplx, 1, qtInputTaps, 1, 1);       
   } 
}


// ### Reset the equalizer
// ##  By: Aurenice Lima
void ElecEqualizer::ResetEqualizer(void)
{
   for (int ii = 0; ii < qtInputTaps; ii++) {
      CorrelationVectorQtAverage[ii] = 0;       
      if (! IsComplexEqualization) {
         TapWeight[ii] = 0;
         CrossCorrelationVector[ii] = 0;
         CorrelationVector[ii] = 0;

      }
      else {
         TapWeightCplx[ii] = 0;
         CrossCorrelationVectorCplx[ii] = 0;
         CorrelationVectorCplx[ii] = 0;    
      }   
   }   
}


// ##########################################################
// #### Equalization selection using the Wiener solution
// #    The options are real or complex equalization
// ## By: Aurenice Lima
void ElecEqualizer::EqualizeBitIntensityWiener(void)
{
   if (! IsComplexEqualization)
      EqualizeBitIntensityWienerReal();
   else
      EqualizeBitIntensityWienerCplx();   
}


// ##########################################################
// #### Equalization using the Wiener algorithm for a real quantity
// ## By: Aurenice Lima
void ElecEqualizer::EqualizeBitIntensityWienerReal(void)
{
   // ## The back to back is used here just to be adjusted 
   // #  in order to make possible to look at the 
   // #  equalized error

   if (StringLength < qtInputTaps) {
      *LogFile << "From ElecEqualizer."  << endl
              << "Error: StringLength < qtInputTaps " << endl
              << "   StringLength " << StringLength << endl
              << "   qtInputTaps     " << qtInputTaps << endl;
      exit(1);
   }

   double *DesiredBitIntensity = new double[StringLength];
   double *DistortedBitIntensity; 
   double *EqualizedBitIntensity = oElecSignalStat->BitIntensity;

   if (FeedBackCorrectedBits == 1) {
      //## The corrected values will be used for the equalization
      DistortedBitIntensity = oElecSignalStat->BitIntensity;
   }
   else {
      DistortedBitIntensity = new double[StringLength];
      for (int ii = 0; ii < StringLength; ii++) {     
         DistortedBitIntensity[ii] = oElecSignalStat->BitIntensity[ii];
      }
   }

   MeanBitIntensity = 0.;
   for (int ii = 0; ii < StringLength; ii++) {  
      MeanBitIntensity += DistortedBitIntensity[ii];
   }   
   MeanBitIntensity /= StringLength;


   //## This is the time to grab the actual desired bit intensity,
   //#  in case we want to do it. It can be done using the 
   //#  autocorrelation function of the bit sequence that is 
   //#  in the ElecSignalStat class.
   //#  I can get that information from the class ElecSignalStat
   //#  class that is used in the initialization of the 
   //#  ElecEqualizer class.
   //## ATTENTION: You must have extracted the signal by
   //#  convoluting. The effect of the convolution is 
   //#  to match the BitIntensity vector with the 
   //#  original BitString vector.
      
   
   if (SymmetricEqualization == 1) {
      // ## Deducting the mean value (symetrical)
      //cout << "Doing symmetrical equalization" << endl;      
      for (int ii = 0; ii < StringLength; ii++) {
         DistortedBitIntensity[ii] -= MeanBitIntensity;
         //Back2BackBitIntensity[ii] -= MeanBitIntensity;
         if ( ! UseExactDesired && DistortedBitIntensity[ii] < 0
            ||  UseExactDesired 
               && oElecSignalStat->GetBitFromString(ii)==0) 
            DesiredBitIntensity[ii] = -MeanBitIntensity; //MeanBit0;   
         else
            DesiredBitIntensity[ii] =  MeanBitIntensity;  //MeanBit1;         
         //if (Back2BackBitIntensity[ii] < 0) 
         //   Back2BackBitIntensity[ii] = -MeanBitIntensity; //MeanBit0;   
         //else
         //   Back2BackBitIntensity[ii] =  MeanBitIntensity;  //MeanBit1;            
      }
   } 
   else {
      // ## No deduction of the mean value (non-symetric)
      for (int ii = 0; ii < StringLength; ii++) {
         if ( ! UseExactDesired 
               && DistortedBitIntensity[ii] < MeanBitIntensity
             || UseExactDesired 
               &&  oElecSignalStat->GetBitFromString(ii)==0) 
              
            DesiredBitIntensity[ii] = 0; //MeanBit0;   
         else
            DesiredBitIntensity[ii] = 2*MeanBitIntensity;  //MeanBit1; 
         // ## Adjust the back to back values
         //if (Back2BackBitIntensity[ii] < MeanBitIntensity) 
         //   Back2BackBitIntensity[ii] = 0; //MeanBit0;   
         //else
         //   Back2BackBitIntensity[ii] = 2*MeanBitIntensity;  //MeanBit1;            
      }    
   }
   

       
   int AveragingPeriod = 10000;   
   int SweepTime;
   int IndexTmp;
   for (int ii = 0; ii < StringLength; ii++) {      
      EqualizedBitIntensity[ii] = 0;   
      for (int jj = 0; jj < qtInputTaps; jj++) {
         SweepTime = ii - (EqualizationTap-1 - jj);
         if (CorrelationVectorQtAverage[jj] >= AveragingPeriod)
            CorrelationVectorQtAverage[jj] = AveragingPeriod - 1;
            
         CorrelationVector[jj] = ( CorrelationVector[jj]
               * CorrelationVectorQtAverage[jj]
               + DistortedBitIntensity[ii]
               * DistortedBitIntensity
                    [Periodicity(ii-jj,StringLength)] ) // SweepTime
               / (1. + CorrelationVectorQtAverage[jj]);
         CrossCorrelationVector[jj] = (CrossCorrelationVector[jj]
               * CorrelationVectorQtAverage[jj]
               //+ Back2BackBitIntensity[ii]
               + DesiredBitIntensity[ii]
               * DistortedBitIntensity
                    [Periodicity(SweepTime,StringLength)] )
               / (1. + CorrelationVectorQtAverage[jj]);                        
         CorrelationVectorQtAverage[jj]++;      
               
         //            cout << "ii = " << ii << "   SweepTime =  " << SweepTime << endl;
         EqualizedBitIntensity[ii] += TapWeight[jj]
                        *DistortedBitIntensity
                           [Periodicity(SweepTime,StringLength)];                      
      }

      //cout << endl;
      for (int kk = 0; kk < qtInputTaps; kk++) {      
         //cout << "CrossCorrelationVector["<< kk <<"] = " 
         //     << CrossCorrelationVector[kk] << endl;      
         //cout << "CorrelationVector["<< kk <<"] = " 
         //     << CorrelationVector[kk] << endl;
              
         for (int ll = 0; ll < qtInputTaps; ll++) {                           
            if (ll - kk >= 0)
               IndexTmp = ll - kk;
            else
               IndexTmp = kk - ll; //ll -kk + qtInputTaps;                  
            CorrelationMatrix[kk+1][ll+1] = CorrelationVector[IndexTmp];            
            //cout << "CorrelationMatrix["<< kk+1 <<"]["<<ll+1<<"] = "
            //     << CorrelationMatrix[kk+1][ll+1] << endl;
 
         }   
         TapWeightTmp[kk+1][1] = CrossCorrelationVector[kk];      
      }
      
      
      if (ii > StringLength/4) {  //## Wait for a minimum number of 
                                  //##samples to compute the matrices and vectors      
         gaussj(CorrelationMatrix, qtInputTaps, TapWeightTmp, 1);       
   
         for (int kk = 0; kk < qtInputTaps; kk++) {
            TapWeight[kk] = TapWeightTmp[kk+1][1];
         }     
      }	 
   } 

   if (NormalizeWeights == 1) {
      double NormalizationFactor = TapWeight[EqualizationTap - 1];
      //for (int jj = 0; jj < qtInputTaps; jj++) {       
      //   NormalizationFactor += abs(TapWeight[jj]);   
      //}         
      //## Normalize the weights in order to make a fair 
      //#  equalization
      for (int jj = 0; jj < qtInputTaps; jj++) {       
         TapWeight[jj] /= NormalizationFactor;   
      }              
   }
   
   // ## It is incompatible with FeedBackCorrectedBits
   if (UseFinalWeights == 1) {
      //## Equalize the whole thing, again, using the final weights
      for (int ii = 0; ii < StringLength; ii++) {      
         EqualizedBitIntensity[ii] = 0;   
         for (int jj = 0; jj < qtInputTaps; jj++) {
            SweepTime = ii - (EqualizationTap-1 - jj);
            EqualizedBitIntensity[ii] += TapWeight[jj]
                              *DistortedBitIntensity
                                 [Periodicity(SweepTime,StringLength)];                      
         }
      } 

   }   

   // ## It is incompatible with FeedBackCorrectedBits
   if (SymmetricEqualization == 1) {
       // ## Add the mean values back to the signal
      // #  to remove the symetry
      for (int ii = 0; ii < StringLength; ii++) {
         DistortedBitIntensity[ii] += MeanBitIntensity;
         //Back2BackBitIntensity[ii] += MeanBitIntensity;
         EqualizedBitIntensity[ii] += MeanBitIntensity;
      }   
   }

   //for (int ii = 0; ii < StringLength; ii++) {
   //   cout << oElecSignalStat->BitIntensity[ii] << endl;  
   //}
   //cout <<"MeanBitIntensity = " << MeanBitIntensity << endl;   
  
   oElecSignalStat->ComputeEqualizedPerformanceStatistics();
 
   //oElecSignalStat->GetSingleStringElecSignalStatEqualizedConvoluting();   
   delete [] DesiredBitIntensity;       
} //## End of EqualizeBitIntensityWienerReal


// #######################################################
// #### Equalization using the complex Wiener algorithm
// ## By: Aurenice Lima
void ElecEqualizer::EqualizeBitIntensityWienerCplx(void)
{
   // ## The back to back is used here just to be adjusted 
   // #  in order to make possible to look at the 
   // #  equalized error

   //cout << " I am trying to equalize " << endl;

   if (StringLength < qtInputTaps) {
      *LogFile << "From ElecEqualizer."  << endl
              << "Error: StringLength < qtInputTaps " << endl
              << "   StringLength " << StringLength << endl
              << "   qtInputTaps     " << qtInputTaps << endl;
      exit(1);
   }

   cplx *DesiredBitIntensity   = new cplx[StringLength];
   cplx *DistortedBitIntensity = new cplx[StringLength]; 
   cplx *EqualizedBitIntensity = new cplx[StringLength]; //oElecSignalStat->BitIntensity;

   //if (FeedBackCorrectedBits == 1) {
   //   //## The corrected values will be used for the equalization
   //   DistortedBitIntensity = oElecSignalStat->BitIntensity;
   //}
   //else {
   //   DistortedBitIntensity = new double[StringLength];
   //   for (int ii = 0; ii < StringLength; ii++) {     
   //      DistortedBitIntensity[ii] = oElecSignalStat->BitIntensity[ii];
   //   }
   //}
   
   for (int ii = 0; ii < StringLength; ii++) {  
      DistortedBitIntensity[ii] = oElecSignalStat->BitIntensity[ii]
                              +jc*oElecSignalStatIm->BitIntensity[ii];
   }   

   MeanBitIntensity = 0.;
   MeanBitIntensityIm = 0.;
   for (int ii = 0; ii < StringLength; ii++) {  
      MeanBitIntensity   += real(DistortedBitIntensity[ii]);
      MeanBitIntensityIm += real(-jc*DistortedBitIntensity[ii]);      
   }   
   MeanBitIntensity   /= StringLength;
   MeanBitIntensityIm /= StringLength;   

   if (SymmetricEqualization == 1) {
      // ## Deducting the mean value (symetrical)
      //cout << "Doing symmetrical equalization" << endl;      
      for (int ii = 0; ii < StringLength; ii++) {
         DistortedBitIntensity[ii] -= MeanBitIntensity
                                     +jc*MeanBitIntensityIm;         
         //Back2BackBitIntensity[ii] -= MeanBitIntensity;
         //if ( ( real(DistortedBitIntensity[ii])/MeanBitIntensity 
         //      +imag(DistortedBitIntensity[ii])/MeanBitIntensityIm)/2. < 0) 
         if ( ( real(DistortedBitIntensity[ii])+imag(DistortedBitIntensity[ii]) )
               /(MeanBitIntensity+MeanBitIntensityIm) < 0)                
            DesiredBitIntensity[ii] = -( MeanBitIntensity //MeanBit0;   
                                        +jc*MeanBitIntensityIm); //MeanBit0;               
         else
            DesiredBitIntensity[ii] = +( MeanBitIntensity //MeanBit0;   
                                        +jc*MeanBitIntensityIm); //MeanBit0;                        
      }
   } 
   else {
      // ## No deduction of the mean value (non-symetric)
      for (int ii = 0; ii < StringLength; ii++) {
         //if ( ( real(DistortedBitIntensity[ii])/MeanBitIntensity 
         //      +imag(DistortedBitIntensity[ii])/MeanBitIntensityIm)/2. < 1.)                
         if ( ( real(DistortedBitIntensity[ii])+imag(DistortedBitIntensity[ii]) )
               /(MeanBitIntensity + MeanBitIntensityIm) < 1.)          
            DesiredBitIntensity[ii] = 0; //MeanBit0;   
         else
            DesiredBitIntensity[ii] = 2.0*( MeanBitIntensity
                                         +jc*MeanBitIntensityIm);  //MeanBit1; 
      }    
   }

/******
   if (SymmetricEqualization == 1) {
      // ## Deducting the mean value (symetrical)
      //cout << "Doing symmetrical equalization" << endl;      
      for (int ii = 0; ii < StringLength; ii++) {
         DistortedBitIntensity[ii] -= MeanBitIntensity
                                     +jc*MeanBitIntensityIm;         
         //Back2BackBitIntensity[ii] -= MeanBitIntensity;
         if (real(DistortedBitIntensity[ii]) < 0) 
            DesiredBitIntensity[ii] = -MeanBitIntensity; //MeanBit0;   
         else
            DesiredBitIntensity[ii] =  MeanBitIntensity;  //MeanBit1;    
         if (real(-jc*DistortedBitIntensity[ii]) < 0) 
            DesiredBitIntensity[ii] += -jc*MeanBitIntensityIm; //MeanBit0;   
         else
            DesiredBitIntensity[ii] +=  jc*MeanBitIntensityIm;  //MeanBit1;                
                 
         //if (Back2BackBitIntensity[ii] < 0) 
         //   Back2BackBitIntensity[ii] = -MeanBitIntensity; //MeanBit0;   
         //else
         //   Back2BackBitIntensity[ii] =  MeanBitIntensity;  //MeanBit1;            
      }
   } 
   else {
      // ## No deduction of the mean value (non-symetric)
      for (int ii = 0; ii < StringLength; ii++) {
         if (real(DistortedBitIntensity[ii]) < MeanBitIntensity) 
            DesiredBitIntensity[ii] = 0; //MeanBit0;   
         else
            DesiredBitIntensity[ii] = 2*MeanBitIntensity;  //MeanBit1; 
         if (real(-jc*DistortedBitIntensity[ii]) < MeanBitIntensityIm) 
            DesiredBitIntensity[ii] += 0*jc; //MeanBit0;   
         else
            DesiredBitIntensity[ii] += 2*jc*MeanBitIntensityIm;  //MeanBit1; 
         // ## Adjust the back to back values
         //if (Back2BackBitIntensity[ii] < MeanBitIntensity) 
         //   Back2BackBitIntensity[ii] = 0; //MeanBit0;   
         //else
         //   Back2BackBitIntensity[ii] = 2*MeanBitIntensity;  //MeanBit1;            
      }    
   }   
*********/   

       
   int AveragingPeriod = 10000;   
   int SweepTime;
   int IndexTmp;
   for (int ii = 0; ii < StringLength; ii++) {      
      EqualizedBitIntensity[ii] = 0;   
      for (int jj = 0; jj < qtInputTaps; jj++) {
         SweepTime = ii - (EqualizationTap-1 - jj);
         if (CorrelationVectorQtAverage[jj] >= AveragingPeriod)
            CorrelationVectorQtAverage[jj] = AveragingPeriod - 1;
            
         CorrelationVectorCplx[jj] = ( CorrelationVectorCplx[jj]
               * CorrelationVectorQtAverage[jj]
               + DistortedBitIntensity[ii]
               * conj(DistortedBitIntensity
                    [Periodicity(ii-jj,StringLength)]) ) // SweepTime
               / (1. + CorrelationVectorQtAverage[jj]);
         //cout << jj << "SweepTime = " << SweepTime << endl;      
         CrossCorrelationVectorCplx[jj] = (CrossCorrelationVectorCplx[jj]
               * CorrelationVectorQtAverage[jj]
               //+ Back2BackBitIntensity[ii]
               + DesiredBitIntensity[ii]  // conj (does not work with conj)
               * conj(DistortedBitIntensity
                    [Periodicity(SweepTime,StringLength)] )  )
               / (1. + CorrelationVectorQtAverage[jj]);                  
         CorrelationVectorQtAverage[jj]++;      
               
         //            cout << "ii = " << ii << "   SweepTime =  " << SweepTime << endl;
         EqualizedBitIntensity[ii] += TapWeightCplx[jj]
                        *DistortedBitIntensity
                           [Periodicity(SweepTime,StringLength)];                      
      }

      //cout << endl;
      for (int kk = 0; kk < qtInputTaps; kk++) {      
         //cout << "CrossCorrelationVector["<< kk <<"] = " 
         //     << CrossCorrelationVector[kk] << endl;      
         //cout << "CorrelationVector["<< kk <<"] = " 
         //     << CorrelationVector[kk] << endl;
              
         for (int ll = 0; ll < qtInputTaps; ll++) {                           
            if (ll - kk >= 0)
               IndexTmp = ll - kk;
            else
               IndexTmp = kk - ll; //ll -kk + qtInputTaps;                  
               
            if ( kk <= ll )  
               CorrelationMatrixCplx[kk+1][ll+1] = 
                                (CorrelationVectorCplx[IndexTmp]);            
            else 
               CorrelationMatrixCplx[kk+1][ll+1] = 
                               conj(CorrelationVectorCplx[IndexTmp]); // conj                          
            //cout << "CorrelationMatrix["<< kk+1 <<"]["<<ll+1<<"] = "
            //     << CorrelationMatrix[kk+1][ll+1] << endl;
 
         }   
         TapWeightTmpCplx[kk+1][1] = CrossCorrelationVectorCplx[kk];      
      }
      
      gaussj_complex(CorrelationMatrixCplx, qtInputTaps, TapWeightTmpCplx, 1);       

      for (int kk = 0; kk < qtInputTaps; kk++) {
         TapWeightCplx[kk] = TapWeightTmpCplx[kk+1][1];
      }     
   } 
   
   if (NormalizeWeights == 1) {
      double NormalizationFactor = abs(TapWeightCplx[EqualizationTap - 1]);
      //for (int jj = 0; jj < qtInputTaps; jj++) {       
      //   NormalizationFactor += abs(TapWeight[jj]);   
      //}         
      //## Normalize the weights in order to make a fair 
      //#  equalization
      for (int jj = 0; jj < qtInputTaps; jj++) {       
         TapWeightCplx[jj] /= NormalizationFactor;   
         //cout << "TapWeightCplx [" << jj<<"] = " << TapWeightCplx[jj] << endl;
      }              
   }
   
   
   // ## It is incompatible with FeedBackCorrectedBits
   if (UseFinalWeights == 1) {
      //## Equalize the whole thing, again, using the final weights
      for (int ii = 0; ii < StringLength; ii++) {      
         EqualizedBitIntensity[ii] = 0;   
         for (int jj = 0; jj < qtInputTaps; jj++) {
            SweepTime = ii - (EqualizationTap-1 - jj);
            EqualizedBitIntensity[ii] += TapWeightCplx[jj]
                              *DistortedBitIntensity
                                 [Periodicity(SweepTime,StringLength)];                      
         }
      } 

   }   

   // ## It is incompatible with FeedBackCorrectedBits
   if (SymmetricEqualization == 1) {
       // ## Add the mean values back to the signal
      // #  to remove the symetry
      for (int ii = 0; ii < StringLength; ii++) {
         DistortedBitIntensity[ii] += MeanBitIntensity
                                 + jc*MeanBitIntensityIm;         
         EqualizedBitIntensity[ii] += MeanBitIntensity
                                 + jc*MeanBitIntensityIm;                                
         //Back2BackBitIntensity[ii] += MeanBitIntensity;         
      }         
   }
   

   
   for (int ii = 0; ii < StringLength; ii++) {  
     //cout << real(EqualizedBitIntensity[ii]) << endl;
     oElecSignalStat->BitIntensity[ii] =    real(    EqualizedBitIntensity[ii]);
     oElecSignalStatIm->BitIntensity[ii] =  real(-jc*EqualizedBitIntensity[ii]);    
   }    
   
     oElecSignalStat->ComputeEqualizedPerformanceStatistics();
     oElecSignalStat->ComputeEqualizedPerformanceStatistics();

   //oElecSignalStat->GetSingleStringElecSignalStatEqualizedConvoluting();  
   //oElecSignalStatIm->GetSingleStringElecSignalStatEqualizedConvoluting();      
   delete [] DistortedBitIntensity;      
   delete [] DesiredBitIntensity;       
   delete [] EqualizedBitIntensity;          
} //## End of EqualizeBitIntensityWiener


// ##########################################################
// #### Equalization using the Wiener algorithm
// #    Out of date. It is not trully
// #    integrated to the OCS library.
// #    It can be said as for general purposes
// ##   The back to back is not adjusted
// ## By: Aurenice Lima
void ElecEqualizer::EqualizeBitIntensityWiener( int StringLength
                                            ,double *DistortedBitIntensity
                                            ,double *EqualizedBitIntensity)
{
   double *Back2BackBitIntensity = new double[StringLength];
   EqualizeBitIntensityWiener(StringLength, 
                              DistortedBitIntensity,
                              EqualizedBitIntensity,
                              Back2BackBitIntensity);
   delete [] Back2BackBitIntensity;                              
}


// ##########################################################
// #### Equalization using the Wiener algorithm
// ##   Out of date.
// ##   It automatically adjusts the back to back data
// ## By: Aurenice Lima
void ElecEqualizer::EqualizeBitIntensityWiener( int StringLength
                                            ,double *DistortedBitIntensity
                                            ,double *EqualizedBitIntensity
                                            ,double *Back2BackBitIntensity) 
{
   // ## The back to back is used here just to be adjusted 
   // #  in order to make possible to look at the 
   // #  equalized error

   if (StringLength < qtInputTaps) {
      *LogFile << "From ElecEqualizer."  << endl
              << "Error: StringLength < qtInputTaps " << endl
              << "   StringLength " << StringLength << endl
              << "   qtInputTaps     " << qtInputTaps << endl;
      exit(1);
   }

   double *DesiredBitIntensity = new double[StringLength];

   MeanBitIntensity=0.;

   for (int ii = 0; ii < StringLength; ii++) {  
      MeanBitIntensity += DistortedBitIntensity[ii];
   }   
  MeanBitIntensity /= StringLength;


   // // ## Deducting the mean value (symetrical)
   //for (int ii = 0; ii < StringLength; ii++) {
   //   DistortedBitIntensity[ii] -= MeanBitIntensity;
   //   Back2BackBitIntensity[ii] -= MeanBitIntensity;
   //   if (DistortedBitIntensity[ii] < 0) 
   //      DesiredBitIntensity[ii] = -MeanBitIntensity; //MeanBit0;   
   //   else
   //      DesiredBitIntensity[ii] =  MeanBitIntensity;  //MeanBit1;         
   //   if (Back2BackBitIntensity[ii] < 0) 
   //      Back2BackBitIntensity[ii] = -MeanBitIntensity; //MeanBit0;   
   //   else
   //      Back2BackBitIntensity[ii] =  MeanBitIntensity;  //MeanBit1;            
   //} 

   // ## No deduction of the mean value (non-symetric)
   for (int ii = 0; ii < StringLength; ii++) {
      if (DistortedBitIntensity[ii] < MeanBitIntensity) 
         DesiredBitIntensity[ii] = 0; //MeanBit0;   
      else
         DesiredBitIntensity[ii] = 2*MeanBitIntensity;  //MeanBit1; 
      // ## Adjust the back to back values
      if (Back2BackBitIntensity[ii] < MeanBitIntensity) 
         Back2BackBitIntensity[ii] = 0; //MeanBit0;   
      else
         Back2BackBitIntensity[ii] = 2*MeanBitIntensity;  //MeanBit1;            
   }    
       
   int AveragingPeriod = 10000;   
   int SweepTime;
   int IndexTmp;
   for (int ii = 0; ii < StringLength; ii++) {      
      EqualizedBitIntensity[ii] = 0;   
      for (int jj = 0; jj < qtInputTaps; jj++) {
         SweepTime = ii - (EqualizationTap-1 - jj);
         if (CorrelationVectorQtAverage[jj] >= AveragingPeriod)
            CorrelationVectorQtAverage[jj] = AveragingPeriod - 1;
            
         CorrelationVector[jj] = ( CorrelationVector[jj]
               * CorrelationVectorQtAverage[jj]
               + DistortedBitIntensity[ii]
               * DistortedBitIntensity
                    [Periodicity(ii-jj,StringLength)] ) // SweepTime
               / (1. + CorrelationVectorQtAverage[jj]);
         CrossCorrelationVector[jj] = (CrossCorrelationVector[jj]
               * CorrelationVectorQtAverage[jj]
               //+ Back2BackBitIntensity[ii]
               + DesiredBitIntensity[ii]
               * DistortedBitIntensity
                    [Periodicity(SweepTime,StringLength)] )
               / (1. + CorrelationVectorQtAverage[jj]);                        
         CorrelationVectorQtAverage[jj]++;      
               
         //            cout << "ii = " << ii << "   SweepTime =  " << SweepTime << endl;
         EqualizedBitIntensity[ii] += TapWeight[jj]
                        *DistortedBitIntensity
                           [Periodicity(SweepTime,StringLength)];                      
      }

      //cout << endl;
      for (int kk = 0; kk < qtInputTaps; kk++) {      
         //cout << "CrossCorrelationVector["<< kk <<"] = " 
         //     << CrossCorrelationVector[kk] << endl;      
         //cout << "CorrelationVector["<< kk <<"] = " 
         //     << CorrelationVector[kk] << endl;
              
         for (int ll = 0; ll < qtInputTaps; ll++) {                           
            if (ll - kk >= 0)
               IndexTmp = ll - kk;
            else
               IndexTmp = kk - ll; //ll -kk + qtInputTaps;                  
            CorrelationMatrix[kk+1][ll+1] = CorrelationVector[IndexTmp];            
            //cout << "CorrelationMatrix["<< kk+1 <<"]["<<ll+1<<"] = "
            //     << CorrelationMatrix[kk+1][ll+1] << endl;
 
         }   
         TapWeightTmp[kk+1][1] = CrossCorrelationVector[kk];      
      }
      
      gaussj(CorrelationMatrix, qtInputTaps, TapWeightTmp, 1);       

      for (int kk = 0; kk < qtInputTaps; kk++) {
         TapWeight[kk] = TapWeightTmp[kk+1][1];
      }     
   } 

   // // ## Add the mean value back to the signal
   // // #  to remove the symetry
   //for (int ii = 0; ii < StringLength; ii++) {
   //   DistortedBitIntensity[ii] += MeanBitIntensity;
   //   Back2BackBitIntensity[ii] += MeanBitIntensity;
   //   EqualizedBitIntensity[ii] += MeanBitIntensity;
   //}   
   
   delete [] DesiredBitIntensity;       
}


// ##########################################################
// #### Equalization using the LMS algorithm
// ## This might be out of date, since it hasn't 
// ## been used
// ## By: Aurenice Lima
void ElecEqualizer::EqualizeBitIntensityLMS( int StringLength
                                            ,double *DistortedBitIntensity
                                            ,double *EqualizedBitIntensity
                                            ,double *Back2BackBitIntensity
                                            ,double StepSize)
{
   if (StringLength < qtInputTaps) {
      *LogFile << "From ElecEqualizer."  << endl
              << "Error: StringLength < qtInputTaps " << endl
              << "   StringLength " << StringLength << endl
              << "   qtInputTaps     " << qtInputTaps << endl;
      exit(1);
   }

   //ResetEqualizer();
   
   int SweepTime;
   double EstimatedError;
   for (int ii = 0; ii < StringLength; ii++) {      
      EqualizedBitIntensity[ii] = 0;   
      for (int jj = 0; jj < qtInputTaps; jj++) {
         //SweepTime = ii - (qtInputTaps - jj - 1);
         SweepTime = ii - (EqualizationTap - jj - 1);
         if ( SweepTime  >= 0 && SweepTime < StringLength) {
            //            cout << "ii = " << ii << "   SweepTime =  " << SweepTime << endl;
            EqualizedBitIntensity[ii] += TapWeight[jj]
                              *DistortedBitIntensity[SweepTime];                      
         }
      }
      
      EstimatedError = Back2BackBitIntensity[ii] 
                      -EqualizedBitIntensity[ii]; 
      
      for (int jj = 0; jj < qtInputTaps; jj++) {
         //SweepTime = ii - (qtInputTaps - jj - 1);  
         SweepTime = ii - (EqualizationTap - jj - 1);    
         if (SweepTime >= 0) {                                
            TapWeight[jj] += StepSize * DistortedBitIntensity[SweepTime]
                                      * EstimatedError;
         }         
      }      
   }        
}


// ##### Get the weight for user purpose
// ## By: Aurenice Lima
double ElecEqualizer::GetTapWeight(int IndexTap)
{
   if (IndexTap < 1 || IndexTap > qtInputTaps) {
      cout << "From ElecEqualizer::GetTapWeight(.)" << endl
           << "Error: The tap index must be "
           << "Between 1 and " << qtInputTaps << endl
           << "You selected "<< IndexTap << endl
           << "Aborting now" << endl;
      exit(1);        
   }
   // ## Note that the return will range from 1 to qtInputTaps
   return TapWeight[IndexTap+1];
   
}


// ##########################################################
// ##### Writes the eye diagram to a file using the real part
// #     of the current. It is adequate for the signal 
// #     combination to see the result of the electrical 
// #     equalization
// ## By: Aurenice and Ivan Lima
void ElecEqualizer::WriteFileEyeDiagram(string outFile) 
//                        ElecSignalStat *oElecSignalStat)
//                     int StringLength,
//                     int qtPoints,
//                     cplx *SignalTimeDomainOrig,
//                     double DeltaTime)
{
   //int StringLength = oElecSignalStat->GetStringLength();
   int qtPoints = oElecSignalStat->GetQtPoints();
   cplx *SignalTimeDomainOrig = oElecSignalStat->GetSignalTimeDomain(); //sfftPM.tPowerSignal;
   double DeltaTime = oElecSignalStat->GetDeltaTime();

   double *SignalTimeDomain = new double[qtPoints];

   // These lines are not necessary
   //if (SymmetricalEqualization) {
   //   for (int ii = 0; ii < qtPoints; ii++)
   //      SignalTimeDomain[ii] = SignalTimeDomainOrig[ii];
   //}   

   //TapWeight[0] = 0;
   //TapWeight[1] = 0;
   //TapWeight[2] = 1;
   //TapWeight[3] = 0;
   //TapWeight[4] = 0;

   for ( int ii = 0; ii < qtPoints; ii++) {
      SignalTimeDomain[ii] = 0;
      for (int jj = 0; jj < qtInputTaps; jj++) {
         if (ii == 1000)
            cout << jj << "  " << TapWeight[jj] << endl;
                // << Periodicity(ii + (int) ((EqualizationTap-1 - jj)
                //    *qtPoints/StringLength)  ,qtPoints) << endl;

         if (SymmetricEqualization == 1) {
            // ## Symetrical  
            SignalTimeDomain[ii] += TapWeight[jj]
                 *(abs(SignalTimeDomainOrig[
                  Periodicity(ii - (int) ((EqualizationTap-1 - jj)
                  *qtPoints/StringLength)  ,qtPoints)])
                  - MeanBitIntensity);
         }
         else {
            // ##Non-symetrical
            SignalTimeDomain[ii] += TapWeight[jj]
                 *(abs(SignalTimeDomainOrig[
                 Periodicity(ii - (int) ((EqualizationTap-1 - jj)
                 *qtPoints/StringLength)  ,qtPoints)]) ); 
         }                            
      }   
   }   
   
   if (SymmetricEqualization) {   
      // ## Symetrical: Restores the mean value
      for ( int ii = 0; ii < qtPoints; ii++) {
         SignalTimeDomain[ii] += MeanBitIntensity;         
      }   
   }   


   
   ofstream OutFile;   
//   if (FirstRunEyeDiagram) {    
//      OutFile.open(outFile.c_str(),ios::out);
//      FirstRunEyeDiagram = 0;
//   }   
//   else    
   OutFile.open(outFile.c_str(),ios::out);   
   if (!OutFile) {
      cerr << "Error: " << outFile << " could not be created."<< endl;
      exit(1);
   }
      
   int ii,jj,iiMin,iiMax,iiScroll;
   for (jj = 0; jj < StringLength; jj++) {
      iiMin =  jj*qtPoints/StringLength;
      iiMax = (jj+2)*qtPoints/StringLength;
      for (ii = iiMin; ii < iiMax; ii++) {
         if (ii < qtPoints)
            iiScroll = ii;
         else
            iiScroll = ii - qtPoints;   
         OutFile << (ii-iiMin)*DeltaTime << "   " 
                 << SignalTimeDomain[iiScroll] << endl;   
      }
      OutFile << endl; // ## End of a datablock           
   }             
   OutFile.close();    
} // end WriteEyeDiagramFile
              
