

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsElecSignalStat.hh"
#include <math.h>
#include "nrutil.h"

extern ofstream *LogFile;

//######################################################

//! Constructor without input file

ElecSignalStat::ElecSignalStat(OptSignal *oOptSignal2,
                               Photodetector *oPhotodetector2)
{
      InitElecSignalStat(oOptSignal2,oPhotodetector2);
}			       

//######################################################

 //!Constructor that reads parameters from an input file

ElecSignalStat::ElecSignalStat(string InFileName,OptSignal *oOptSignal2,
                               Photodetector *oPhotodetector2)
{   
   InitElecSignalStat(oOptSignal2,oPhotodetector2);

   ifstream InFile(InFileName.c_str(),ios::in);    

   if (!InFile)  {
       *LogFile << "WARNING: Input file "<< InFileName 
               << " could not be opened."<< endl;
        cout << "Error: See *.log for more information." << endl;
        exit(1);     
   }


     LogFileSeparator();
     *LogFile << InFileName << endl << endl;

   int SetTypeClockRecoveryFlag = ReadInt("SetTypeClockRecoveryFlag",
                                            " ",0,1,LOWER_AND_UPPER,
					  OPTIONAL_NO_WARNING,0,&InFile);


     if(SetTypeClockRecoveryFlag)
       {
         TypeClockRecovery = typeClockRecovery(ReadInt("TypeClockRecovery",
                                            " ",0,8,LOWER_AND_UPPER,
                                            MANDATORY,0,&InFile) );   

         *LogFile << "TypeClockRecovery = " 
		  << WriteTypeClockRecovery() << endl;

         ComputeTargetFreqIndex();

         UseSmoothingToRecoverClockFlag = 
                                    ReadInt("UseSmoothingToRecoverClockFlag",
                                            " ",0,1,LOWER_AND_UPPER,
                                            OPTIONAL_NO_WARNING,0,&InFile);

         GaussianSmoothingFWHM = ReadDouble("GaussianSmoothingFWHM",
                                            " ",0,-1,LOWER_ONLY,
                                            OPTIONAL_NO_WARNING,0.0,&InFile);

         if(TypeClockRecovery == USER_SET)
            InputClockRecoveryTime = ReadDouble("ClockRecoveryTime","sec ",0,
		    oOptSignal->GetBitLength(),LOWER_AND_UPPER,
                    MANDATORY,0,&InFile);  


       } // ## end  if(SetTypeClockRecoveryFlag)


      ComputePatternStatsFlag = bool(ReadInt("ComputePatternStatsFlag",
                                         " ",0,1,LOWER_AND_UPPER,
					     OPTIONAL_NO_WARNING,0,&InFile));

     if(ComputePatternStatsFlag)
       {

         PatternLength = ReadInt("PatternLength",
                                 " ",0,StringLength,LOWER_AND_UPPER,
				 MANDATORY,1,&InFile); 

         if(PatternLength%2==0)
           {
             cerr << "ElecSignalStat constructor: PatternLength must be odd."
		  << endl 
		  << "ABORTING NOW!!" << endl;

             exit(1);
	   }

         InitPatternStats();

       } // ## end if(ComputePatternStatsFlag)


DoGaussianPDFsMonteCarloFlag = bool(ReadInt("DoGaussianPDFsMonteCarloFlag",
                                         " ",0,1,LOWER_AND_UPPER,
					     OPTIONAL_NO_WARNING,0,&InFile));


 if(DoGaussianPDFsMonteCarloFlag)
   {

   MinCurrent = ReadDouble("MinCurrent","V",1,-1,
			   NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);

   MaxCurrent = ReadDouble("MaxCurrent","V",MinCurrent,-1,
			    LOWER_ONLY,MANDATORY,0,&InFile);

   if(MaxCurrent < MinCurrent)
     cerr << "ABORT: ChiSquareReceiver Error: [MinCurrent,MaxCurrent] = [" 
	  << MinCurrent << "," << MaxCurrent << "]" << endl;

   NumPointsCurrent = ReadInt("NumPointsCurrent"," ",0,-1,
			   LOWER_ONLY,MANDATORY,0,&InFile);
    
       NumTimeSlices = StringLength;
       DeltaCurrent = (MaxCurrent - MinCurrent)/(NumPointsCurrent-1);

       CurrentVector = new double[NumPointsCurrent];

       for(int vv=0;vv<NumPointsCurrent;vv++)
	 CurrentVector[vv] = MinCurrent+vv*DeltaCurrent;

       BitString = NewInt(StringLength);
       TimeSlices = NewDouble(StringLength);
       TimeSliceIndices = NewInt(StringLength);
       MeanCurrentMonteCarlo = new double[StringLength];
       StdDevCurrentMonteCarlo = new double[StringLength];
       PowerSumBit = new double[StringLength];
       PowerSquareSumBit = new double[StringLength];
       qtBit = new int[StringLength];
       AverageGaussianPDFZeros = new double[NumPointsCurrent];
       AverageGaussianPDFOnes = new double[NumPointsCurrent];
       BERGaussianPDFs = new double[NumPointsCurrent];

       GaussianPDFTimeSlicesStorageVec = 
	 new double[NumTimeSlices*NumPointsCurrent];

       GaussianPDFTimeSlices = new double*[NumTimeSlices];

       for(int tt=0;tt<NumTimeSlices;tt++)
        GaussianPDFTimeSlices[tt] = 
	  &(GaussianPDFTimeSlicesStorageVec[tt*NumPointsCurrent]);

  }

  InFile.close();

  ClearElecSignalStat();

} // ## end ElecSignalStat::ElecSignalStat

// ###################################################################

//! This function is called by both constructors

void ElecSignalStat::InitElecSignalStat(OptSignal *oOptSignal2,
					Photodetector *oPhotodetector2)
{

  // Written by John Zweck, Nov 8 2000
  // Update (CLK recovery) by Ivan Lima, 7/6/01

   oOptSignal = oOptSignal2; 
   oPhotodetector = oPhotodetector2; 

   UseAbsCurrentFlag = 0; // # Added by JZ, See comments at start of file.
   *LogFile << "UseAbsCurrentFlag = " << UseAbsCurrentFlag << endl;

  if (oOptSignal->GetTypePulse() == NRZ)
     TypeClockRecovery = FREQ_DOUBLING;
  else
     TypeClockRecovery = RZ_SIGNAL_FREQ;
        
  *LogFile << "TypeClockRecovery (default for this format) = " 
          << WriteTypeClockRecovery() << endl;  
  UseSmoothingToRecoverClockFlag = 0;
  GaussianSmoothingFWHM  = 0.0;

   fft = oOptSignal->fft;
   sfftPM = oPhotodetector->sfftPM;
   TypePulse = oOptSignal->GetTypePulse();
   BitLength = oOptSignal->GetBitLength();

   qtPoints = fft->give_N();
   TimeWindow = 1/fft->delta_f;
   DeltaTime = TimeWindow/qtPoints;  
   StringLength = oOptSignal->oBitString->GetStringLength();
   BitRateChann = StringLength/TimeWindow;
   DeltaFreq = oOptSignal->GetDeltaFreq();
   
   ComputeTargetFreqIndex();

   BitIntensity = new double[StringLength];      
 
   ComputePatternStatsFlag = 0;
   DoGaussianPDFsMonteCarloFlag = 0;   

   ClearElecSignalStat();
   ObjectInitialized = 1;
   InitializedGaussianStatisticsFlag = 0; 

   

   // add by Jonathan for debug
    Q_FactorPre = 2;

} // end ElecSignalStat

//######################################################

ElecSignalStat::~ElecSignalStat(void)
{
   if (ObjectInitialized)

  {
    delete [] BitIntensity;

    if(DoGaussianPDFsMonteCarloFlag)
    {
      delete [] MeanCurrentMonteCarlo;
      delete [] StdDevCurrentMonteCarlo;

      delete [] PowerSumBit;
      delete [] PowerSquareSumBit;
      delete qtBit;

      delete [] TimeSlices;
      delete [] TimeSliceIndices;
      delete [] BitString;
      delete [] BERGaussianPDFs;
      delete [] AverageGaussianPDFZeros;
      delete [] AverageGaussianPDFOnes;

      for(int tt=0;tt<NumTimeSlices;tt++)
       GaussianPDFTimeSlices[tt]= 0;

      delete [] GaussianPDFTimeSlices;
      delete [] GaussianPDFTimeSlicesStorageVec;


      delete [] CurrentVector;
  
    }


   if(InitializedGaussianStatisticsFlag) {
      delete oOptSignal1; 
      delete oOptSignal2; 
      delete oOptSignal3;
      delete oOptSignal4;
      delete oOptSignalX;

      delete oOptFilter2;
      delete oPhotodetector2;
      delete oElecFilter2;
      delete oElecSignalStat2;    
   }      

   if(ComputePatternStatsFlag)
     ReleaseMemoryForPatternStats();

  }
}

// ####################################################


 /*!
 Instead of taking abs(Current) to get stats we take real(Current).
 This allows for negative currents, which could arise because
 of response function of electric filter or because of shot noise
 in photodetector. 

 To allow for backward compatibility, there is an optional
 flag called UseAbsCurrentFlag. 
 If this flag is 1, then the code takes abs(Current).
 If the flag is 0 the code takes real(Current). 

 The constructor sets UseAbsCurrentFlag = 0 since this is
 the physically correct thing to do.
 If you want UseAbsCurrentFlag = 1
 SetUseAbsCurrentFlag(1) straight after the constructor is called.

  */

 void ElecSignalStat::SetUseAbsCurrentFlag(bool Value)
{
  UseAbsCurrentFlag =  Value;
  *LogFile << "RESET: UseAbsCurrentFlag = " << UseAbsCurrentFlag << endl;

}

// #######################################################

//! Returns the type of the clock recovery method

string ElecSignalStat::WriteTypeClockRecovery(void)
{
   string strTypeClockRecovery;
   switch (TypeClockRecovery) {
      case USER_SET:       
         strTypeClockRecovery = "USER_SET"; break;     
      case FREQ_DOUBLING:       
         strTypeClockRecovery = "FREQ_DOUBLING"; break;         
      case RZ_SIGNAL_FREQ:     
         strTypeClockRecovery = "RZ_SIGNAL_FREQ"; break;     
      case NRZ_HALF_SIGNAL_FREQ:
         strTypeClockRecovery = "NRZ_HALF_SIGNAL_FREQ"; break;
      case BEST_Q_FACTOR:   
         strTypeClockRecovery = "BEST_Q_FACTOR"; break;
      case DIFFERENTIATE_SIGNAL: 
         strTypeClockRecovery = "DIFFERENTIATE_SIGNAL"; break;
      case CONTINUOUS_EXOR:    
         strTypeClockRecovery = "CONTINUOUS_EXOR"; break;
      case BINARY_EXOR:       
         strTypeClockRecovery = "BINARY_EXOR"; break;
      case BEST_MIN_EYE_OPENING:   
         strTypeClockRecovery = "BEST_MIN_EYE_OPENING"; break;         
      default:
         strTypeClockRecovery  = "ERROR";  
         cerr << "ElecSignalStat::WriteTypeClockRecovery Error. ABORTING!"
	      << endl;
         exit(1);  
   }
   return strTypeClockRecovery;
}

// ######################################################

  //! Resets variables to correctly calculate statistics
  //! Use when doing parameter studies before change parameters

void ElecSignalStat::ClearElecSignalStat()
{
   qtRunsEyeDiagram = 0;
   qtRunsEyeDiagramFixedCLK = 0;   
  
   qtBit0 = 0;
   qtBit1 = 0;   
   Bit0_PowerSum = 0;
   Bit1_PowerSum = 0;
   Bit0_PowerSquareSum  = 0;   
   Bit1_PowerSquareSum  = 0;     
   Bit0_PowerMax = -1e100; //## Startup Value
   Bit1_PowerMin =  1e100; //## Startup Value
   
   qtIsolatedBit0 = 0;
   qtIsolatedBit1 = 0;   
   IsolatedBit0_PowerSum = 0;
   IsolatedBit1_PowerSum = 0;
   IsolatedBit0_PowerSquareSum  = 0;   
   IsolatedBit1_PowerSquareSum  = 0;     
   IsolatedBit0_PowerMax = -1e100; //## Startup Value
   IsolatedBit1_PowerMin =  1e100; //## Startup Value

   NumCallsToUpdateClock = 0;         
   ClockRecoveryTime = 0.0; 

   if(ComputePatternStatsFlag)
     ClearPatternStats();

   if(DoGaussianPDFsMonteCarloFlag)
     ClearGaussianPDFsMonteCarlo();

}

// ##########################################################

  //! Computes the intensity of received current in each bit
  //! and the mean and standard deviation of the received current
  //! in each bit at clock recovery time. Used in either noise-free
  //! simulations or with noise in Monte Carlo simulations.

void ElecSignalStat::ComputePerformanceStatistics(void)
{
   ClearElecSignalStat();
   ClockRecoveryTime = GetTargetTime();
   ClockRecoveryTimeBeforeTimeShift = ClockRecoveryTime;
   oPhotodetector->TimeShiftSignal( -ClockRecoveryTime );

   UpdatePerformanceStatistics();

}

// ##########################################################

  //! A version of ElecSignalStat::ComputePerformanceStatistics
  //! that does not involve time-shifting the signal. 
  //! A private function that should ONLY be used by semianalytical methods

void ElecSignalStat::ComputePerformanceStatisticsNoTimeShift(void)
{
   ClearElecSignalStat();
   UpdatePerformanceStatistics();
}

// ########################################################

//! Calculates intensity of received current in each bit
//! at clock recovery time and updates variables that
//! store mean and standard deviation of current in each bit.
//! Used in Monte Carlo simulations.

/*! Method first finds optimal correlatation between the bit string
    and the bit intensity array which gives the 
    power of the elecectrically filtered signal at the recovery time.

     The idea is that we want to find the shift in the bit sequence
     which maximizes the correlation between these two vectors.

     If the match is not optimal then
     (1) Mean Ones decreases from optimal value
         as some physical ones are not lined up with logical ones.
     (2) Mean Zeros increases. 
     (3) So Mean Ones - Means Zeros decreases
     (4) StdDev ones and zeros will both increase
     (5) So Q = Ratio (3)/(4) will be forced to decrease
  
     So our algorithm is to walk through the shifts of the bit string
     computing Q values and choose the maximal Q value.

     One reason you need to do this correlation is that different
     channels travel at different speeds relative to the central
     frequency. One way to test this method is to see that the 
     computation of Q for a given physical channel is invariant 
     to shifts in the central frequency of the slowly-varying envelope
     approximation. To get good agreement though you need DeltaTime
     to be small enough. 
  
     Another reason why one needs to do this correlation 
     is because the PMD can shift the bits beyound one bit period. 
     Hence, one could find a Q factor much smaller than it really is.

*/

void ElecSignalStat::UpdatePerformanceStatistics(void)
{
 
 
   if (! ObjectInitialized) {
      cerr << "Error: Object of class ElecSignalStat not initialized"
           << endl;
      exit(1);
   }
   
   int TargetIndexTime = GetTargetIndexTime();
          
   double PowerElecSignal;  

   double Q_FactorTmp;
   double qtBit0_Tmp, qtBit1_Tmp;
   double Bit0_PowerSumTmp, Bit1_PowerSumTmp;
   double Bit0_PowerSquareSumTmp, Bit1_PowerSquareSumTmp;
   int OptimumBitMatch;
   double MaxQ_Factor;

   for (int kk = 0; kk < StringLength; kk++) 
    {
      qtBit0_Tmp = qtBit1_Tmp = 0;
      Bit0_PowerSumTmp = Bit1_PowerSumTmp = 0;
      Bit0_PowerSquareSumTmp = Bit1_PowerSquareSumTmp = 0;
    
      for (int jj = 0; jj < StringLength; jj++)  
       {
         if(UseAbsCurrentFlag)
	   {
             PowerElecSignal = 
               abs(sfftPM.tPowerSignal[Periodicity(TargetIndexTime
               +qtPoints/StringLength*
                Periodicity(jj+kk,StringLength),qtPoints)]);
	   }
          else
	   {
            PowerElecSignal = 
               real(sfftPM.tPowerSignal[Periodicity(TargetIndexTime
               +qtPoints/StringLength*
                Periodicity(jj+kk,StringLength),qtPoints)]);
	   }
 
         BitIntensity[jj] = PowerElecSignal;  
                
         if (oOptSignal->oBitString->GetBit(0,jj) == 0) 
           {           
            Bit0_PowerSumTmp   += PowerElecSignal;
            Bit0_PowerSquareSumTmp += sq(PowerElecSignal);  
            qtBit0_Tmp += 1;
           }   
         else 
	   {       
            Bit1_PowerSumTmp   += PowerElecSignal;
            Bit1_PowerSquareSumTmp += sq(PowerElecSignal);  
            qtBit1_Tmp += 1;
	   } 
                 
      } //# End for jj

      Q_FactorTmp = ( Bit1_PowerSumTmp/qtBit1_Tmp
                     -Bit0_PowerSumTmp/qtBit0_Tmp)
                   /( sqrt( (Bit1_PowerSquareSumTmp
                     -qtBit1_Tmp*sq(Bit1_PowerSumTmp/qtBit1_Tmp))
                     /(qtBit1_Tmp - 1) ) 
                     +sqrt( (Bit0_PowerSquareSumTmp
                     -qtBit0_Tmp*sq(Bit0_PowerSumTmp/qtBit0_Tmp))
                     /(qtBit0_Tmp - 1) ) );
      
      if (Q_FactorTmp > MaxQ_Factor || kk == 0) 
	{
         MaxQ_Factor = Q_FactorTmp;
         OptimumBitMatch = kk;
		 
	}  
 
   } //# End for kk  


        //## The bit string shift is stored in a variable of the class

   BitStringShift = OptimumBitMatch;    

   //## It now gets the contribution of the optimum match

   for (int jj = 0; jj < StringLength; jj++) 
     {

       if(UseAbsCurrentFlag)
	 {
          PowerElecSignal = 
            abs(sfftPM.tPowerSignal[Periodicity(TargetIndexTime
            +qtPoints/StringLength*
            Periodicity(jj+OptimumBitMatch,StringLength),qtPoints)]);
	 }
        else
	 {
          PowerElecSignal = 
            real(sfftPM.tPowerSignal[Periodicity(TargetIndexTime
            +qtPoints/StringLength*
            Periodicity(jj+OptimumBitMatch,StringLength),qtPoints)]);
	 }


      BitIntensity[jj] = PowerElecSignal;  

      UpdateMoments(jj,PowerElecSignal);

   } // end for jj

   UpdateClock();

} //## End  UpdateComputePerformanceStatistics

// ############################################################

//! Internal function called by UpdatePerformanceStatistics

void ElecSignalStat::UpdateMoments(int BitIndex,double PowerElecSignal)
{
   if (oOptSignal->oBitString->GetBit(0,BitIndex) == 0) 
     {           
      Bit0_PowerSum   += PowerElecSignal;
      Bit0_PowerSquareSum += sq(PowerElecSignal);  
      qtBit0 += 1;
      if (PowerElecSignal > Bit0_PowerMax) 
         Bit0_PowerMax = PowerElecSignal;
     
     }   
   else 
     {       
      Bit1_PowerSum   += PowerElecSignal;
      Bit1_PowerSquareSum += sq(PowerElecSignal);  
      qtBit1 += 1;
      if (PowerElecSignal < Bit1_PowerMin) 
         Bit1_PowerMin = PowerElecSignal;
     }        
   
   //## Do the statistics for the isolated bits, which should produce
   //#  the worst distribution at the receiver
 
   int BitIndexPrevious  = Periodicity(BitIndex-1,StringLength);
   int BitIndexPosterior = Periodicity(BitIndex+1,StringLength);      

   if (oOptSignal->oBitString->GetBit(0,BitIndex) == 0
          && oOptSignal->oBitString->GetBit(0,BitIndexPrevious) == 1
          && oOptSignal->oBitString->GetBit(0,BitIndexPosterior) == 1 ) 
     {           
      IsolatedBit0_PowerSum   += PowerElecSignal;
      IsolatedBit0_PowerSquareSum += sq(PowerElecSignal);  
      qtIsolatedBit0 += 1;
      if (PowerElecSignal > IsolatedBit0_PowerMax) 
	IsolatedBit0_PowerMax = PowerElecSignal;
     }   
   else if (oOptSignal->oBitString->GetBit(0,BitIndex) == 1
          && oOptSignal->oBitString->GetBit(0,BitIndexPrevious) == 0
          && oOptSignal->oBitString->GetBit(0,BitIndexPosterior) == 0 ) 
     {  
      IsolatedBit1_PowerSum   += PowerElecSignal;
      IsolatedBit1_PowerSquareSum += sq(PowerElecSignal);  
      qtIsolatedBit1 += 1;
      if (PowerElecSignal < Bit1_PowerMin) 
         IsolatedBit1_PowerMin = PowerElecSignal;
     
     }     
        
   if(ComputePatternStatsFlag)
     UpdatePatternStats(BitIndex,PowerElecSignal);  

   if(DoGaussianPDFsMonteCarloFlag)
     {
      PowerSumBit[BitIndex]   += PowerElecSignal;
      PowerSquareSumBit[BitIndex] += sq(PowerElecSignal);  
      qtBit[BitIndex] += 1;
     }
    
} // ## end UpdateMoments

// ################################################################

//! A version of UpdatePerformanceStatistics() used in conjunction with
//! an ElecEqualizer object. The BitIntensity array must have been
//! previously set by the  ElecEqualizer object.

void ElecSignalStat::UpdateEqualizedPerformanceStatistics(void)
{
 
   if (! ObjectInitialized) 
     {
      cerr << "Error: Object of class ElecSignalStat not initialized"
           << endl;
      exit(1);
     }
            
   double PowerElecSignal;  

   double Q_FactorTmp;
   double qtBit0_Tmp, qtBit1_Tmp;
   double Bit0_PowerSumTmp, Bit1_PowerSumTmp;
   double Bit0_PowerSquareSumTmp, Bit1_PowerSquareSumTmp;
   int OptimumBitMatch;
   double MaxQ_Factor = 0;
   for (int kk = 0; kk < StringLength; kk++) {
      qtBit0_Tmp = qtBit1_Tmp = 0;
      Bit0_PowerSumTmp = Bit1_PowerSumTmp = 0;
      Bit0_PowerSquareSumTmp = Bit1_PowerSquareSumTmp = 0;
      for (int jj = 0; jj < StringLength; jj++)  
        {
         PowerElecSignal = BitIntensity[Periodicity(jj+kk,StringLength)];
            
         if (oOptSignal->oBitString->GetBit(0,jj) == 0) 
	   {           
            Bit0_PowerSumTmp   += PowerElecSignal;
            Bit0_PowerSquareSumTmp += sq(PowerElecSignal);  
            qtBit0_Tmp += 1;
	   }   
         else 
	   {       
            Bit1_PowerSumTmp   += PowerElecSignal;
            Bit1_PowerSquareSumTmp += sq(PowerElecSignal);  
            qtBit1_Tmp += 1;
	   } 
                 
	} //# End for jj

      Q_FactorTmp = ( Bit1_PowerSumTmp/qtBit1_Tmp
                     -Bit0_PowerSumTmp/qtBit0_Tmp)
                   /( sqrt( (Bit1_PowerSquareSumTmp
                     -qtBit1_Tmp*sq(Bit1_PowerSumTmp/qtBit1_Tmp))
                     /(qtBit1_Tmp - 1) ) 
                     +sqrt( (Bit0_PowerSquareSumTmp
                     -qtBit0_Tmp*sq(Bit0_PowerSumTmp/qtBit0_Tmp))
                     /(qtBit0_Tmp - 1) ) );

      if (Q_FactorTmp > MaxQ_Factor || kk == 0) 
	{
         MaxQ_Factor = Q_FactorTmp;
         OptimumBitMatch = kk;
	}                    
   } //# End for kk  
   
   //## It now gets the contribution of the optimum match

   for (int jj = 0; jj < StringLength; jj++) 
     {
      PowerElecSignal = BitIntensity[jj];
      
      UpdateMoments(jj,PowerElecSignal);      
      
     } // end for jj
   

   UpdateClock();

} //# End of UpdateEqualizedPerformanceStatistics

// ##########################################################

//! A version of ComputePerformanceStatistics() used in conjunction with
//! an ElecEqualizer object. The BitIntensity array must have been
//! previously set by the  ElecEqualizer object.

void ElecSignalStat::ComputeEqualizedPerformanceStatistics(void)
{
   ClearElecSignalStat();  
   UpdateEqualizedPerformanceStatistics();
   
}

// ###############################################################

//! Returns power difference between smallest one and largest zero.
//! Useful when studying PMD.
//! Only to be used in noise-free simulations! 

double ElecSignalStat::GetMinAmplitudeMargin()
{
   if (Bit1_PowerMin - Bit0_PowerMax > 0) {
      return Bit1_PowerMin - Bit0_PowerMax;
   }   
   else {
      return 0;
   }
}

// ###########################################################

  //! For use in conjunction with Monte Carlo simulations with noise.
  //! Must set DoGaussianPDFsMonteCarloFlag = 1 and other required parameters
  //! in ElecSignalStat infile and call constructor:
  //! ElecSignalStat::ElecSignalStat(string InFileName,OptSignal *oOptSignal2,
  //!                             Photodetector *oPhotodetector2)
  //! 
  //! First call UpdatePerformanceStatistics() many times within
  //! a loop.
  //! Outputs mean and stddev of current in each bit, computes
  //! Gaussian approximation to pdf in each bit, computes
  //! average pdfs of zeros and ones, based on the individual Gaussian
  //! approximations, and finally computes BER from these average pdfs.
  //! Outputs all the above statistics to  files. 

void ElecSignalStat::OutputGaussianStatisticsFromMonteCarlo(string FileID)
{

  if(!DoGaussianPDFsMonteCarloFlag)
    {
      cerr << "ElecSignalStat::OutputGaussianStatisticsFromMonteCarlo"
	   << endl
	   << "Must set DoGaussianPDFsMonteCarloFlag = 1."
	   << endl
	   << "See documentation for method. ABORTING!" << endl;
      exit(1);
    }

  SetTimeSlices();
  ComputeAndOutputMomentsFromMonteCarlo(FileID);
  ComputeAveragePDFs();
  WriteFilePDFTimeSlicesAndAveragePDFs(FileID);
  ComputeMinimumBERAndDecisionThreshold(FileID);
}

// ##########################################################

void ElecSignalStat::SetTimeSlices(void)
{

  double TimeBetweenSlices = 
    oOptSignal->GetTimeWindow()/oOptSignal->GetStringLength();

 for(int tt=0;tt<NumTimeSlices;tt++)
    {
      TimeSlices[tt] = GetClockRecoveryTime() + tt*TimeBetweenSlices;
         	
      // We want TimeSlices[tt] to be in range
      // 0 <= TimeSlices[tt] < TimeWindow

      while(TimeSlices[tt] < 0)
        TimeSlices[tt] += oOptSignal->GetTimeWindow();

      while(TimeSlices[tt] >= oOptSignal->GetTimeWindow())
        TimeSlices[tt] -= oOptSignal->GetTimeWindow();

      TimeSliceIndices[tt] =  int(floor(TimeSlices[tt]/
	                      oOptSignal->GetDeltaTime()));

    } // end for tt

} // ## end ElecSignalStat::SetTimeSlices(void)

// ##################################################

  //! Mean of current in bit with index BitIndex at clock recovery time.
  //! Must first call UpdatePerformanceStatistics(). For Monte Carlo.

  double ElecSignalStat::GetPowerMeanBit(int BitIndex) 
    {return PowerSumBit[BitIndex]/qtBit[BitIndex];};

// #######################################################

  //! Standard deviation of current in bit with index BitIndex 
  //! at clock recovery time.
  //! Must first call 
  //! UpdatePerformanceStatistics(). For Monte Carlo.

double ElecSignalStat::GetPowerStdDevBit(int BitIndex)
{
  double SqNum = PowerSquareSumBit[BitIndex] - 
                 qtBit[BitIndex]*sq(GetPowerMeanBit(BitIndex));
 
  if(SqNum < 0)
    { 
     cout << "SqNum = " << SqNum << endl;
     SqNum = 0;
     cout << "ElecSignalStat::GetPowerStdDevBit: SqNum < 0" <<endl;
    }
     
  int SqDen = qtBit[BitIndex] - 1;

  if(SqDen <= 0)
    {
     cout << "ElecSignalStat::GetPowerStdDevBit: SqDen <= 0" <<endl;
    }

  return sqrt(SqNum/double(SqDen));
  
              /* return sqrt( (PowerSquareSumBit[BitIndex]
                      -qtBit[BitIndex]*sq(GetPowerMeanBit(BitIndex)))
                      /(qtBit[BitIndex] - 1.) )
               */

};

// ####################################################

void ElecSignalStat::ComputeAndOutputMomentsFromMonteCarlo(string FileID)
{

  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      MeanCurrentMonteCarlo[tt] = GetPowerMeanBit(tt);
      StdDevCurrentMonteCarlo[tt] = GetPowerStdDevBit(tt);
    }


  string  OutFileName = FileID + "StdDevCurrent.dat";

  ofstream OutFile(OutFileName.c_str(),ios::out);
  
  if (!OutFile) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }


  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      OutFile << TimeSlices[tt] << "\t" 
	      << StdDevCurrentMonteCarlo[tt] << endl;
    }

  OutFile.close();

  OutFileName = FileID + "MeanCurrent.dat";

  ofstream OutFile2(OutFileName.c_str(),ios::out);

  if (!OutFile2) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      OutFile2 << TimeSlices[tt] << "\t" <<  MeanCurrentMonteCarlo[tt] 
	       <<  endl;
    }

 
  OutFile2.close();

  OutFileName = FileID + "Clock.dat";

  ofstream OutFile3(OutFileName.c_str(),ios::out);

  if (!OutFile3) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  OutFile3 << "% (1) ClockTime [ps]" << endl
	   <<  1e12*GetClockRecoveryTime() << endl;
 
  OutFile3.close();

}

// ##################################################

void ElecSignalStat::ComputeAveragePDFs(void)
{

  // Adapted from ChiSquareReceiver::ComputeAveragePDFs()

  // I checked that the correlation is computed correctly
  // by artificially shifting the electrically filtered
  // current as a function of time. 

   for(int tt=0;tt<NumTimeSlices;tt++)
     {

      for(int vv=0;vv<NumPointsCurrent;vv++)
	GaussianPDFTimeSlices[tt][vv] = 
	  exp( -sq(CurrentVector[vv] - MeanCurrentMonteCarlo[tt])/
		     (2.0*sq(StdDevCurrentMonteCarlo[tt])))/
	                (sqrt(2*pi)*StdDevCurrentMonteCarlo[tt]);
 
      double IntPDFTimeSlice = 
       Quadrature(GaussianPDFTimeSlices[tt],NumPointsCurrent,
		CurrentVector[0],CurrentVector[NumPointsCurrent-1]);

      if(abs(IntPDFTimeSlice-1.0) > 0.05)
	{
	  cerr << "ElecSignalStat::ComputeAveragePDFs" 
	       << endl
	       << "Error exceeds 1% in integration of PDFTimeSlices." << endl
	       << "Change Min or Max Current or increase NumPointsCurrent"
	       << endl
	       << "ABORTING" << endl;

	  cout << "tt for Matlab,IntPDFTimeSlice = " 
	       << tt+2 << " " << IntPDFTimeSlice << endl;

	  exit(1);
	}      

     } // end for tt loop


  int NumBitsToPlot = NumTimeSlices;

  int NumZeros = 0;
  int NumOnes = 0;

  for(int vv=0;vv<NumPointsCurrent;vv++)
    AverageGaussianPDFZeros[vv] = 0.0;

  for(int vv=0;vv<NumPointsCurrent;vv++)
    AverageGaussianPDFOnes[vv] = 0.0;

  int NumBitsBetweenSlices = 1;

  for(int bb=0;bb<NumBitsToPlot;bb++)
    {
      int bit = bb*NumBitsBetweenSlices;
     
      BitString[bb] = oOptSignal->oBitString->GetBit(0,bit);
       
    }

  double MaxCorrelation = 0.0;
  int OptimalShift = 0;

  for(int ss=0;ss<NumTimeSlices;ss++)
    {
      double Correlation = 0;

      for(int tt=0;tt<NumTimeSlices;tt++)
	{ 
 
          int BitIndex = ss+tt;

          while(BitIndex < 0)
	    BitIndex += NumBitsToPlot;
          while(BitIndex >= NumBitsToPlot)
	    BitIndex -= NumBitsToPlot;

          Correlation += (MeanCurrentMonteCarlo[tt])*BitString[BitIndex];
	}

      if(Correlation > MaxCorrelation)
	{
         OptimalShift = ss;
         MaxCorrelation = Correlation;
	}

    } // end for ss


  for(int tt=0;tt<NumTimeSlices;tt++)
    {
      int BitIndex = OptimalShift+tt;

      while(BitIndex < 0)
	BitIndex += NumBitsToPlot;
      while(BitIndex >= NumBitsToPlot)
	BitIndex -= NumBitsToPlot;

      if(BitString[BitIndex] == 0)
	{
          NumZeros += 1;

          for(int vv=0;vv<NumPointsCurrent;vv++)
	    AverageGaussianPDFZeros[vv] += GaussianPDFTimeSlices[tt][vv];
	}
      else
	{
          NumOnes += 1;

          for(int vv=0;vv<NumPointsCurrent;vv++)
	    AverageGaussianPDFOnes[vv] += GaussianPDFTimeSlices[tt][vv];
	}

    } // end for tt


  if(NumZeros != 0)
   for(int vv=0;vv<NumPointsCurrent;vv++)
    AverageGaussianPDFZeros[vv] /= NumZeros;

  if(NumOnes != 0)
   for(int vv=0;vv<NumPointsCurrent;vv++)
    AverageGaussianPDFOnes[vv] /= NumOnes;


} // ## end ComputeAveragePDFs

// #################################################################

void ElecSignalStat::WriteFilePDFTimeSlicesAndAveragePDFs(string FileID)
{

  WriteFilePairVectors(CurrentVector,AverageGaussianPDFZeros,NumPointsCurrent,
                       FileID+"AveragePDFZeros.dat");

  WriteFilePairVectors(CurrentVector,AverageGaussianPDFOnes,NumPointsCurrent,
                       FileID+"AveragePDFOnes.dat");

   ofstream OutFile;

   string OutFileName = FileID+"PDFTimeSlices.dat";

   OutFile.open(OutFileName.c_str(),ios::out);

   if (!OutFile)
    {
      cerr << "Error: " << OutFileName
           << "could not be created. ABORTING"<<endl;
      exit(1);
   }


  OutFile << "% (1) Current [mV] (t+1) PDF of current in t-th TimeSlice"
          << endl;


  for(int vv=0;vv<NumPointsCurrent;vv++)
    {

      OutFile << 1e3*CurrentVector[vv] << "\t";

     for(int tt=0;tt<NumTimeSlices;tt++)
       OutFile << GaussianPDFTimeSlices[tt][vv] << "\t";

      OutFile << endl;
    }

  OutFile.close();

} // ## end WriteFilePDFTimeSlicesAndAveragePDFs

// ############################################################

void ElecSignalStat::ComputeMinimumBERAndDecisionThreshold(string FileID)
{

  // Adapted from ChiSquareReceiver::ComputeMinimumBERAndDecisionThreshold

  double BitErrorRatioGaussianPDFs, DecisionThreshold;

  double TotalIntZeros = Quadrature(AverageGaussianPDFZeros,NumPointsCurrent,
				    CurrentVector[0],
				    CurrentVector[NumPointsCurrent-1]);

  double TotalIntOnes = Quadrature(AverageGaussianPDFOnes,NumPointsCurrent,
				   CurrentVector[0],
				   CurrentVector[NumPointsCurrent-1]);

  double Error = max( abs(TotalIntZeros-1), abs(TotalIntOnes-1) );
  
  if(Error > 0.05)
    {
      cerr << "ElecSignalStat::ComputeMinimumBERAndDecisionThreshold:"
	   <<  endl
	   << "Error in integration of AveragePDFOnes or  AveragePDFZeros "
	   << endl
	   << "exceeds 1%. ABORTING" << endl;

      cout << "TotalIntZeros = " << TotalIntZeros << endl;
      cout << "TotalIntOnes = " << TotalIntOnes << endl;
    
      exit(1);
    }

    // RightEndPt for OnesInt must equal LeftEndPt for Zeros Int

     for(int DecisionIndex=0;DecisionIndex<NumPointsCurrent;DecisionIndex++)
     {
      double ZerosInt = Quadrature(&AverageGaussianPDFZeros[DecisionIndex],
				   NumPointsCurrent-DecisionIndex,
				   CurrentVector[DecisionIndex],
				   CurrentVector[NumPointsCurrent-1])/
	                           TotalIntZeros;

      double OnesInt = Quadrature(AverageGaussianPDFOnes,DecisionIndex,
				  CurrentVector[0],
				  CurrentVector[DecisionIndex])/
	                          TotalIntOnes;

      BERGaussianPDFs[DecisionIndex] = 0.5*(ZerosInt+OnesInt);
      
      if(DecisionIndex == 0 || 
	 BERGaussianPDFs[DecisionIndex] < BitErrorRatioGaussianPDFs)
	{
	 BitErrorRatioGaussianPDFs = BERGaussianPDFs[DecisionIndex];
         DecisionThreshold = CurrentVector[DecisionIndex];
	}
     } // end for DecisionIndex loop


  
        WriteFilePairVectors(CurrentVector,BERGaussianPDFs,NumPointsCurrent,
		             FileID+"BERVector.dat");

    
 string OutFileName = FileID + "BERFromGaussianPDFs.dat";

 ofstream OutFile(OutFileName.c_str(),ios::out);

 if (!OutFile) 
   {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  OutFile << "% (1) BER (2)  DecisionThreshold [mV]" << endl
	  <<  BitErrorRatioGaussianPDFs<< "\t" <<  1e3*DecisionThreshold 
	  << endl;
 
  OutFile.close();

} // ## end GaussianReceiver::ComputeMinimumBERAndDecisionThreshold

// ###############################################################

void ElecSignalStat::WriteFilePairVectors(double * x, double *y, int dim, 
					    string OutFileName)
{

  // Adapted from ChiSquareReceiver::WriteFilePairVectors

   ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "Error: " << OutFileName 
	   << "could not be created. ABORTING"<<endl;
      exit(1);
   }

   for(int jj=0;jj<dim;jj++)
     OutFile << x[jj] << "\t" << y[jj] << endl;

   OutFile.close();


}

// ###########################################################

//! Sets the target frequency for the clock recovery 

void ElecSignalStat::ComputeTargetFreqIndex(void)
{
   if (TypeClockRecovery == NRZ_HALF_SIGNAL_FREQ ||
       (TypeClockRecovery == USER_SET && TypePulse == NRZ)) 
     {
      TargetFreqIndex = (int) ( (BitRateChann/2.)*TimeWindow + 0.1);   
   }
   else {
      TargetFreqIndex = (int) ( (BitRateChann)*TimeWindow + 0.1);      
   }

   if (   ! (TypePulse == NRZ)
      && TypeClockRecovery == NRZ_HALF_SIGNAL_FREQ) {
      string ErrorMsg = "From: ElecSignalStat::GetTargetFreqIndex()";
      ErrorMsg       += " The TypeClockRecovery selected is";
      ErrorMsg       += " incompatible with the RZ (or CRZ) format.";
      ErrorMsg       += "\nPlease, fix it in the input files."; 
      ErrorMsg       += "\nIt can also be that the source code of this";
      ErrorMsg       += " function is out of date";
      LogFileAbortWithErrorMsg(ErrorMsg);      
   }
   
   if (  ! (   TypePulse == RZ_GAUSS || TypePulse ==  RZ_SECH 
         || TypePulse == CRZ_GAUSS || TypePulse == CRZ_RCOS
         || TypePulse == CRZ_SIN_MZ || TypePulse == UMBC_RZ
	 || TypePulse == CW || TypePulse == RZ_GAUSS 
	 || TypePulse == RZ_GAUSS_PATTERN 
	 || TypePulse == INPUT_SEQUENCE )  
        && TypeClockRecovery == RZ_SIGNAL_FREQ) {
      string ErrorMsg = "From: GetTargetFreqIndex()";        
      ErrorMsg       += " The TypeClockRecovery selected is";
      ErrorMsg       += " incompatible with the NRZ format.";
      ErrorMsg       += "\nPlease, fix it in the input files."; 
      ErrorMsg       += "\nIt can also be that the source code of this";
      ErrorMsg       += " function is out of date";      
      LogFileAbortWithErrorMsg(ErrorMsg);      
   }
}

// ########################################################

//! Internal function that returns integer index of the clock recovery time

int ElecSignalStat::GetTargetIndexTime(void)
{
   int TargetIndexTime;

   double PhaseTargetFreq = ComputePhaseTargetFreq();

   switch (TypeClockRecovery) 
     {
      case USER_SET:  
	TargetIndexTime = (int)(InputClockRecoveryTime/DeltaTime);
        break;

      case FREQ_DOUBLING: 
         if ( PhaseTargetFreq <= 0)     
            TargetIndexTime = (int) ( qtPoints / (double) StringLength
                              *( -0.25    - PhaseTargetFreq/(2*pi) ) + 0.5 );
         else   
            TargetIndexTime = (int) ( qtPoints / (double) StringLength
                          *( -0.25 + 1. - PhaseTargetFreq/(2*pi) ) + 0.5 );  
       
         break;          
      case NRZ_HALF_SIGNAL_FREQ:
      case CONTINUOUS_EXOR:
      case BINARY_EXOR:
      
         if(PhaseTargetFreq <= 0)
            TargetIndexTime = (int) ( qtPoints / (double) StringLength
                                   *(    - PhaseTargetFreq/pi ) + 0.5);  
         else
            TargetIndexTime = (int) ( qtPoints / (double) StringLength
                                   *( 1.0 - PhaseTargetFreq/pi ) + 0.5); 
         
         break;
      case DIFFERENTIATE_SIGNAL:
      
         // We shift the target time by a half a bit period 
	 // In the time domain the 
	 // differentiated signal has peaks at the transitions
	 // from 0's to 1's which occurs half way between two bits
	 // rather than in the middle of each bit.
         if(PhaseTargetFreq <= 0)
            TargetIndexTime = (int) ( qtPoints / (double) StringLength
                                   *(0.5 - PhaseTargetFreq/pi ) + 0.5);      
                     
         else
            TargetIndexTime = (int) ( qtPoints / (double) StringLength
                                   *(0.5 + 1.0 - PhaseTargetFreq/pi ) + 0.5);
         break;                          
      case RZ_SIGNAL_FREQ:  
         if ( PhaseTargetFreq <= 0)     
            TargetIndexTime = (int) ( qtPoints / (double) StringLength
                                   *(    - PhaseTargetFreq/(2*pi) ) +0.5);
         else   
            TargetIndexTime = (int) ( qtPoints / (double) StringLength
                                   *( 1. - PhaseTargetFreq/(2*pi) ) +0.5);  
       
         break;
      case BEST_Q_FACTOR:
         TargetIndexTime = GetTargetIndexTimeBestQ_Factor();
         break;
      case BEST_MIN_EYE_OPENING:
         TargetIndexTime = GetTargetIndexTimeBestMinEyeOpening();         
         break;             
      default:
         string strErrorMsg = "From: ElecSignalStat::GetTargetIndexTime()";
         strErrorMsg       += "\nUnkwown clock recovery time."; 
         LogFileAbortWithErrorMsg(strErrorMsg);
     }
   
   return TargetIndexTime;
} // ## end GetTargetIndexTime

// ###########################################################

//! Gets the double-precision target time associated with the clock recovery.  

double ElecSignalStat::GetTargetTime(void)
{
   //## This function returns the index of the clock recovery time 

   double TargetTime;

   double PhaseTargetFreq = ComputePhaseTargetFreq();;

   switch (TypeClockRecovery) {
      case USER_SET: 
	TargetTime = InputClockRecoveryTime;
        break;
      case FREQ_DOUBLING: 
         if ( PhaseTargetFreq <= 0)     
            TargetTime = ( 1.0 / BitRateChann ) 
                              *( -0.25    - PhaseTargetFreq/(2*pi) );
         else   
            TargetTime = ( 1.0 / BitRateChann )
                          *( -0.25 + 1. - PhaseTargetFreq/(2*pi) );  
       
         break;          
      case NRZ_HALF_SIGNAL_FREQ:
      case CONTINUOUS_EXOR:
      case BINARY_EXOR:
      
         if(PhaseTargetFreq <= 0)
            TargetTime = ( 1.0 / BitRateChann )
                                   *(    - PhaseTargetFreq/pi );  
         else
            TargetTime = ( 1.0 / BitRateChann )
                                   *( 1.0 - PhaseTargetFreq/pi ); 
         
         break;
      case DIFFERENTIATE_SIGNAL:
      
         // We shift the target time by a half a bit period 
	 // In the time domain the 
	 // differentiated signal has peaks at the transitions
	 // from 0's to 1's which occurs half way between two bits
	 // rather than in the middle of each bit.
         if(PhaseTargetFreq <= 0)
            TargetTime = ( 1.0 / BitRateChann )
                                   *(0.5 - PhaseTargetFreq/pi );      
                     
         else
            TargetTime = ( 1.0 / BitRateChann )
                                   *(0.5 + 1.0 - PhaseTargetFreq/pi );
         break;                          
      case RZ_SIGNAL_FREQ:  
      // THE FOLLOWING IS THE ONLY BIT OF THIS CODE THAT'S RIGHT!!!
         if ( PhaseTargetFreq <= 0)     
            TargetTime = ( 1. / BitRateChann )
                                   *(    - PhaseTargetFreq/(2*pi) );
         else   
            TargetTime = ( 1. / BitRateChann )
                                   *( 1. - PhaseTargetFreq/(2*pi) );         
         break;
      case BEST_Q_FACTOR:
         TargetTime = DeltaTime*GetTargetIndexTimeBestQ_Factor();
         break;
      case BEST_MIN_EYE_OPENING:
         TargetTime = DeltaTime*GetTargetIndexTimeBestMinEyeOpening();         
         break;             
      default:
         string strErrorMsg = "From: ElecSignalStat::GetTargetIndexTime()";
         strErrorMsg       += "\nUnkwown clock recovery time."; 
         LogFileAbortWithErrorMsg(strErrorMsg);
   }
   
   return TargetTime;

} // ## end GetTargetTime

//#############################################################


//! Returns the phase that will be used to compute the clock recovery time

double ElecSignalStat::ComputePhaseTargetFreq(void)
{
 
  // Written on April 5th 2001 by John Zweck

  // Memory leak fixed by JZ and Oleg Sinkin, Feb 21, 2002

  // JZFeb03: For Physics convention we need to change the sign of the
  // phase of the target frequency. We do this in the return statement. 

  // The purpose of this method is to compute the phase of the target
  // frequency more reliably.

  // In the case of NRZ signals, Ivan suggests that because isolated
  // ones may not occur very frequently that the frequency
  // component corresponding to the clock recovery may very small.
  // To increase it somewhat we can differentiate and square 
  // the signal in time.
  // Alternatively you can use a binary EXOR or continuous EXOR method
  // which is closer in spirit to what real NRZ clock revovery systems
  // use.   

  // THE BINARY EXOR METHOD SEEMS TO WORK BEST (JZ Apr 10 2001)

  // If the differentiated signal is used I think (and have coded up)
  // you need to shift the TargetTime by half a bit period.


  // For filtered noisy signals just recovering the phase
  // at the target frequency may not yield the clock recovery
  // time accurately.  To solve this problem Ron Holzloehner suggested that we 
  // compute the phase of a weighted average of frequencies near the 
  // target frequency. The weighting is done using a Gaussian.
  // The FWHM of the Gaussian is set in the input file and should be 
  // chosen to be less than the Target Frequency. 

  // NOTE: Subsequently I have found that for now its not necessary
  // to smooth in the simulations I am working with. So I won't
  // implemnent this just now. But the commented out code is there
  // if we want to fix it up later.

  // Note that the user can choose whether or not to use this smoothing
  // method by setting the flag UseSmoothingToRecoverClockFlag in the
  // input parameter file for the class.

   double PhaseTargetFreq;

   if (TypeClockRecovery == NRZ_HALF_SIGNAL_FREQ ||
       TypeClockRecovery == RZ_SIGNAL_FREQ ||
       TypeClockRecovery == USER_SET) 
      {
         PhaseTargetFreq = GetPhase(sfftPM.fPowerSignal[TargetFreqIndex]);
      }      
   else if(TypeClockRecovery == CONTINUOUS_EXOR) 
          {
          
             oPhotodetector->ComputeWorkContEXORElecSignal();
         
             PhaseTargetFreq = GetPhase(
                 oPhotodetector->WorksfftPM.fPowerSignal[TargetFreqIndex]);

   }
   else if(TypeClockRecovery == BINARY_EXOR) 
        {
            oPhotodetector->ComputeWorkBinaryEXORElecSignal();
         
             PhaseTargetFreq = GetPhase(
                 oPhotodetector->WorksfftPM.fPowerSignal[TargetFreqIndex]);

         
        }
   else if(TypeClockRecovery == DIFFERENTIATE_SIGNAL) 
        {
         
         oPhotodetector->ComputeWorkDifferentiatedThenSquaredElecSignal();
         
         PhaseTargetFreq = GetPhase(
                 oPhotodetector->WorksfftPM.fPowerSignal[TargetFreqIndex]);
	
        }                                               
   else if (TypeClockRecovery == FREQ_DOUBLING) 
        {
         
         oPhotodetector->ComputeWorkSumDelayedHalfBitPeriodThenSquare();      
         PhaseTargetFreq = GetPhase(
               oPhotodetector->WorksfftPM.fPowerSignal[TargetFreqIndex]);   
                 
       }
   else if (   TypeClockRecovery == BEST_Q_FACTOR
            || TypeClockRecovery ==  BEST_MIN_EYE_OPENING ) 
      {
      //## Do nothing. The phase is not an issue for this option
      PhaseTargetFreq = 0;
   }  
   else {
         string strErrorMsg = "From: ElecSignalStat::ComputePhaseTargetFreq()";
         strErrorMsg       += "\nUnkwown clock recovery time."; 
         LogFileAbortWithErrorMsg(strErrorMsg);   
                                        
        }   
      
   // JZFeb03: Physics convention requires negative sign.

   return -PhaseTargetFreq;

} // ## end ComputePhaseTargetFreq

// ####################################################

int ElecSignalStat::GetTargetIndexTimeBestQ_Factor(void)
{
   int NumPointsTimeDiscretization = 200;
   if (qtPoints/StringLength < NumPointsTimeDiscretization) {
      NumPointsTimeDiscretization = qtPoints/StringLength;
   }
   double PowerElecSignal;

   double Q_FactorTmp;
   double qtBit0_Tmp, qtBit1_Tmp;
   double Bit0_PowerSumTmp, Bit1_PowerSumTmp;
   double Bit0_PowerSquareSumTmp, Bit1_PowerSquareSumTmp;
   int OptimumBitMatch;
   int OptimumIndexTime;
   double MaxQ_Factor;
   int SweepIndexTime;

   for (int ii = 0; ii < NumPointsTimeDiscretization; ii++) 
     {
      SweepIndexTime = ii*(qtPoints/StringLength)
                         /NumPointsTimeDiscretization;
      for (int kk = 0; kk < StringLength; kk++) 
	{
         qtBit0_Tmp = qtBit1_Tmp = 0;
         Bit0_PowerSumTmp = Bit1_PowerSumTmp = 0;
         Bit0_PowerSquareSumTmp = Bit1_PowerSquareSumTmp = 0;
        
	 for (int jj = 0; jj < StringLength; jj++)
           {

             if(UseAbsCurrentFlag)
               {

                  PowerElecSignal =
                     abs(sfftPM.tPowerSignal[Periodicity(SweepIndexTime
                         +qtPoints/StringLength*
                          Periodicity(jj+kk,StringLength),qtPoints)]);
               }
              else
                {
                  PowerElecSignal =
                     real(sfftPM.tPowerSignal[Periodicity(SweepIndexTime
                         +qtPoints/StringLength*
                          Periodicity(jj+kk,StringLength),qtPoints)]);
                }

            if (oOptSignal->oBitString->GetBit(0,jj) == 0) 
	      {
               Bit0_PowerSumTmp   += PowerElecSignal;
               Bit0_PowerSquareSumTmp += sq(PowerElecSignal);
               qtBit0_Tmp += 1;
	      }
            else 
	      {
               Bit1_PowerSumTmp   += PowerElecSignal;
               Bit1_PowerSquareSumTmp += sq(PowerElecSignal);
               qtBit1_Tmp += 1;
	      }

	   } //# End for jj
       
	 Q_FactorTmp = ( Bit1_PowerSumTmp/qtBit1_Tmp
                        -Bit0_PowerSumTmp/qtBit0_Tmp)
                      /( sqrt( (Bit1_PowerSquareSumTmp
                        -qtBit1_Tmp*sq(Bit1_PowerSumTmp/qtBit1_Tmp))
                        /(qtBit1_Tmp - 1) )
                        +sqrt( (Bit0_PowerSquareSumTmp
                     -qtBit0_Tmp*sq(Bit0_PowerSumTmp/qtBit0_Tmp))
                        /(qtBit0_Tmp - 1) ) );
         
         if (   Q_FactorTmp > MaxQ_Factor || (ii == 0 && kk == 0) ) 
	   {
            MaxQ_Factor = Q_FactorTmp;
            OptimumBitMatch = kk;
            OptimumIndexTime = SweepIndexTime;
           }
      } //# End for kk  
   } //# End of ii
   return OptimumIndexTime;
} //# End of GetTargetIndexTimeBestQ_Factor()

// ###################################################

int ElecSignalStat::GetTargetIndexTimeBestMinEyeOpening(void)
{
   int NumPointsTimeDiscretization = 200;
   if (qtPoints/StringLength < NumPointsTimeDiscretization) 
     {
      NumPointsTimeDiscretization = qtPoints/StringLength;
     }
   
   double PowerElecSignal;

   double Q_FactorTmp;
   double MinAmplitudeMarginTmp;
   double qtBit0_Tmp, qtBit1_Tmp;
   double Bit0_PowerSumTmp, Bit1_PowerSumTmp;
   double Bit0_PowerSquareSumTmp, Bit1_PowerSquareSumTmp;
   double Bit0_PowerMaxTmp, Bit1_PowerMinTmp;
   int OptimumBitMatch;
   int OptimumIndexTime;
   double Maximum_MinAmplitudeMarginTmp;
   double MaxQ_Factor;
   int SweepIndexTime;

   for (int ii = 0; ii < NumPointsTimeDiscretization; ii++) 
     {
      SweepIndexTime = ii*(qtPoints/StringLength)
                         /NumPointsTimeDiscretization;
     
      for (int kk = 0; kk < StringLength; kk++) 
	{
         qtBit0_Tmp = qtBit1_Tmp = 0;
         Bit0_PowerSumTmp = Bit1_PowerSumTmp = 0;
         Bit0_PowerSquareSumTmp = Bit1_PowerSquareSumTmp = 0;
        
	 for (int jj = 0; jj < StringLength; jj++)
            {

             if(UseAbsCurrentFlag)
               {
                 PowerElecSignal =
                   abs(sfftPM.tPowerSignal[Periodicity(SweepIndexTime
                   +qtPoints/StringLength*
                   Periodicity(jj+kk,StringLength),qtPoints)]);
               }
              else
               {
                  PowerElecSignal =
                   real(sfftPM.tPowerSignal[Periodicity(SweepIndexTime
                   +qtPoints/StringLength*
                   Periodicity(jj+kk,StringLength),qtPoints)]);
               }

            if (oOptSignal->oBitString->GetBit(0,jj) == 0) 
	      {
               Bit0_PowerSumTmp   += PowerElecSignal;
               Bit0_PowerSquareSumTmp += sq(PowerElecSignal);
               qtBit0_Tmp += 1;
	      }
            else 
	      {
               Bit1_PowerSumTmp   += PowerElecSignal;
               Bit1_PowerSquareSumTmp += sq(PowerElecSignal);
               qtBit1_Tmp += 1;
	      }

         } //# End for jj
        
	 Q_FactorTmp = ( Bit1_PowerSumTmp/qtBit1_Tmp
                        -Bit0_PowerSumTmp/qtBit0_Tmp)
                      /( sqrt( (Bit1_PowerSquareSumTmp
                        -qtBit1_Tmp*sq(Bit1_PowerSumTmp/qtBit1_Tmp))
                        /(qtBit1_Tmp - 1) )
                        +sqrt( (Bit0_PowerSquareSumTmp
                        -qtBit0_Tmp*sq(Bit0_PowerSumTmp/qtBit0_Tmp))
                        /(qtBit0_Tmp - 1) ) );
         
         if (Q_FactorTmp > MaxQ_Factor ||  kk == 0 ) 
	   {
            MaxQ_Factor = Q_FactorTmp;
            OptimumBitMatch = kk;
            }         
      } //# End for kk        

      //## Now that what is 0 and what is 1 is known
      //#  for this particular sweep, we can get 
      //#  MinAmpMargin.   
      Bit0_PowerMaxTmp = -1e99;
      Bit1_PowerMinTmp =  1e99;
     //## It now gets the contribution of the optimum match
      for (int jj = 0; jj < StringLength; jj++)
        {

          if(UseAbsCurrentFlag)
            {
             PowerElecSignal =
                 abs(sfftPM.tPowerSignal[Periodicity(SweepIndexTime
                 +qtPoints/StringLength*
                  Periodicity(jj+OptimumBitMatch,StringLength),qtPoints)]);
            }
          else
            {
             PowerElecSignal =
                 real(sfftPM.tPowerSignal[Periodicity(SweepIndexTime
                 +qtPoints/StringLength*
                  Periodicity(jj+OptimumBitMatch,StringLength),qtPoints)]);
            }

         if (oOptSignal->oBitString->GetBit(0,jj) == 0  ) 
	   {
            if (   Bit0_PowerMaxTmp < sq(PowerElecSignal) ) 
	      {
               Bit0_PowerMaxTmp = sq(PowerElecSignal);
              }
	   }
         else 
	   {
            if (   Bit1_PowerMinTmp > sq(PowerElecSignal) ) 
	      {
               Bit1_PowerMinTmp = sq(PowerElecSignal);
              }
	   }

      } // end for jj

       
      MinAmplitudeMarginTmp = Bit1_PowerMinTmp - Bit0_PowerMaxTmp;

      if (MinAmplitudeMarginTmp > Maximum_MinAmplitudeMarginTmp || ii == 0 )
        {
         Maximum_MinAmplitudeMarginTmp = MinAmplitudeMarginTmp;
         
         OptimumIndexTime = SweepIndexTime;
         
	}
   } //# End of ii
   return OptimumIndexTime;
} //# End of GetTargetIndexTimeBestMinEyeOpening()

// ######################################################

// ##### Writes the eye diagram to a file
// ## The first eye will be centralized, and the subsequent 
// #  ones will suffer the same relative shift
// ## By: Ivan Lima
void ElecSignalStat::WriteFileEyeDiagram(string outFile)
{
   if (! ObjectInitialized) {
      cerr << "Error: Object of class ElecSignalStat not initialized" 
           << endl;
      exit(1);
   }
   ofstream OutFile;   
   

   

   if (qtRunsEyeDiagram == 0) {
      OutFile.open(outFile.c_str(),ios::out);
      if (!OutFile) {
         cerr << "Error: " << outFile << " could not be open."<< endl;
         exit(1);
      }      
      FirstTargetIndexTime = GetTargetIndexTime();      
      SumRelTargetIndexTime = 0;
   }
   else   {    
      OutFile.open(outFile.c_str(),ios::app);   
      if (!OutFile) {
         cerr << "Error: " << outFile << " could not be created."<< endl;
         exit(1);
      }
   }   
         
   qtRunsEyeDiagram++;         

   SumRelTargetIndexTime += GetTargetIndexTime() - FirstTargetIndexTime;      

      
   int ii,jj,iiMin,iiMax,iiScroll;
   for (jj = 0; jj < StringLength; jj++) {
      iiMin =  jj*qtPoints/StringLength;
      iiMax = (jj+2)*qtPoints/StringLength;
      for (ii = iiMin; ii < iiMax; ii++) {
         iiScroll = Periodicity(ii + FirstTargetIndexTime 
                                - qtPoints/StringLength/2,qtPoints);
         OutFile << (ii-iiMin)*DeltaTime << "   " 
                 << real(sfftPM.tPowerSignal[iiScroll]) << endl;   
      }
      OutFile << endl; // ## End of a datablock           
   }             
   OutFile.close();    
   
   string outFileCR = outFile + "_CR";

   ofstream OutFileCR(outFileCR.c_str(),ios::out);

   if (!OutFileCR) {
      cerr << "Error: " << outFileCR << "could not be created."<< endl;
      exit(1);
   }      
     
   
   OutFileCR <<  DeltaTime*(qtPoints/StringLength/2 
                          + SumRelTargetIndexTime/ (double) qtRunsEyeDiagram ) 
             << "   " 
             << (GetBit1_PowerMean() -GetBit0_PowerMean() )/2. 
                +GetBit0_PowerMean()    
             << endl;      

   OutFileCR.close();    

} // end WriteFileEyeDiagram



// ##########################################################

// ##### Writes the eye diagram to a file
// ## By: Ivan Lima

void ElecSignalStat::WriteFileEyeDiagramFixedCLK(string outFile)
{
   if (! ObjectInitialized) {
      cerr << "Error: Object of class ElecSignalStat not initialized" 
           << endl;
      exit(1);
   }
   ofstream OutFile;   

  if (qtRunsEyeDiagramFixedCLK == 0) {   
      OutFile.open(outFile.c_str(),ios::out);
      if (!OutFile) {
         cerr << "Error: " << outFile << " could not be open."<< endl;
         exit(1);
      }      
   }   
   else {   
      OutFile.open(outFile.c_str(),ios::app);   
      if (!OutFile) {
         cerr << "Error: " << outFile << " could not be created."<< endl;
         exit(1);
      }
   }   
   
   qtRunsEyeDiagramFixedCLK++;
   int TargetIndexTime = GetTargetIndexTime();
      
   int ii,jj,iiMin,iiMax,iiScroll;
   for (jj = 0; jj < StringLength; jj++) {
      iiMin =  jj*qtPoints/StringLength;
      iiMax = (jj+2)*qtPoints/StringLength;
      for (ii = iiMin; ii < iiMax; ii++) {
         iiScroll = Periodicity(ii + TargetIndexTime 
                                - qtPoints/StringLength/2,qtPoints);
         OutFile << (ii-iiMin)*DeltaTime << "   " 
                 << real(sfftPM.tPowerSignal[iiScroll]) << endl;   
      }
      OutFile << endl; // ## End of a datablock           
   }             
   OutFile.close();    
   
   string outFileCR = outFile + "_CR";

   ofstream OutFileCR(outFileCR.c_str(),ios::out);

   if (!OutFileCR) {
      cerr << "Error: " << outFileCR << "could not be created."<< endl;
      exit(1);
   }      
   
   
   OutFileCR <<  DeltaTime*qtPoints/StringLength/2 << "   " 
             << (GetBit1_PowerMean() -GetBit0_PowerMean() )/2. 
                +GetBit0_PowerMean()    
             << endl;      

   OutFileCR.close();    

} // end WriteFileEyeDiagramFixedCLK

// ##########################################################


void ElecSignalStat::UpdateClock(void)
{
  // By JZ, adapted from Code of Ivan Lima
  // Returns the values of ClockTime [ps] and ClockCurrent [mV]
  // Useful in conjunction with contour eyes computed using Histogram2D
  // Called from  and UpdatePerformanceStatistics
  // UpdateEqualizedPerformanceStatistics
  
   if(NumCallsToUpdateClock == 0) 
    {  
      SumTargetIndexTimesForUpdateClock = 0;
    }
   
   NumCallsToUpdateClock++;
   
   SumTargetIndexTimesForUpdateClock += GetTargetIndexTime();
                    
} // end UpdateClock

// ##########################################################

  //! Use in conjunction with WriteFileEyeDiagram() to generate
  //! eye diagram that can plot in Matlab  

double ElecSignalStat::GetClockTimeForMonteCarloEye(void)
{
  // Clock Recovery time in ps between [0,BitWindow]
  // Use in conjunction with Histogram2D 

 return    1e12*DeltaTime*(qtPoints/double(StringLength) - Periodicity(
        int(SumTargetIndexTimesForUpdateClock/double(NumCallsToUpdateClock)),
        int(qtPoints/double(StringLength))));
}

// ##############################################################

double ElecSignalStat::GetClockCurrent(void)
{
  // Clock Recovery current in mV

   return    1e3*(GetBit0_PowerMean() + 
             (GetBit1_PowerMean() - GetBit0_PowerMean())/2.0);
 
}

//#############################################################

//! Returns the Q-factor computed from the mean and the standard
//! deviation of the marks and spaces
//! By Ivan Lima

double ElecSignalStat::GetQ_Factor(void)
{
   double Q_FactorStd = ( GetBit1_PowerMean()
                         -GetBit0_PowerMean() )
                       /( GetBit1_PowerStdDev()
                          +GetBit0_PowerStdDev() );
   if (Q_FactorStd < 0) {
      Q_FactorStd = 0;
   }
   return Q_FactorStd;
}




//########################################################
//##### Get the Q-factor based on the mean and the standard
//##    deviation of the isolated marks and spaces
//## By: Ivan Lima
double ElecSignalStat::GetQ_FactorFromIsolatedBits(void) 
{ 
   double Q_FactorStdIsolatedBits = ( GetIsolatedBit1_PowerMean() 
                         -GetIsolatedBit0_PowerMean() )
                       /( GetIsolatedBit1_PowerStdDev() 
     	                  +GetIsolatedBit0_PowerStdDev() );
   if (Q_FactorStdIsolatedBits < 0) {
      Q_FactorStdIsolatedBits = 0;   
   }   
   return Q_FactorStdIsolatedBits;        		       		 
}


//###############################################################
//###############################################################
//###############################################################
//###############################################################
//## Codes to obtain the semi-analytical Q-factor
//## These codes were written based on the equations derived
//## by Aurenice Lima and Ivan Lima. These equations were published
//## only in an internal report as of 7/16/02, but will be soon
//## be available in the technical literature.
//## The true variance of the signal and of the noise were given
//## by P. Winzer, et al., JLT 2001, though we found slightly 
//## different results in our derivation.
//##    YOU MUST TRANSFORM THE POLARIZED PART OF THE NOISE TO THE
//##    X POLARIZATION. CAREFUL TO APPLY THE SAME TRANSFORMATION
//##    TO THE SIGNAL TO GET THE CORRECT SIGNAL TO NOISE BEATING.
//##    You should not worry about it if the noise is unpolarized.
//## Codified in C++ by Ivan T. Lima Jr. (5/16/02)
//###############################################################

//! Constructs auxilliary optical and the electrical filters that will
//! be used to compute the semi-analytical statistics. 

/*! The method initializes auxiliary objects that are used 
 throughout the computation. The method should be called
 prior to the first time in the application code that
 the  semi-analytical statistics are calculated .
 See also ElecSignalStat::ResetOptElecFiltersForGaussianStatistics().

 By: Ivan Lima  (5/16/02)
*/

void ElecSignalStat::SetOptElecFiltersForGaussianStatistics(
                           OptFilter *oOptFilter, 
                           ElecFilter *oElecFilter)
{
   if (! InitializedGaussianStatisticsFlag) 
     {
   
      LogFileSeparator();
      *LogFile << "ElecSignalStat::"
	       << "SetOptElecFiltersForGaussianStatistics"
	       << endl << endl;
      *LogFile << "The semi-analytical statistics in ElecSignalStat class"
	       << endl << "has been initialized."<< endl
	       << "To avoid incorrect results, make sure that the OptSignal"
	       << endl << "object that you used to initialize this "
	       << "ElecSignalStat object" << endl
	       << "be absolutelly noiseless." << endl
	       << "The OptSignal object might contain the signal right"
	       << endl << "before the optical filter." << endl << endl;
      *LogFile << "I recommend you to rotate the signal to the X polarization,"
              << " (1,0,0),"<< endl 
	      << "if you are studying polarization effects."<< endl
	      << "In that case the X component of the noise is the copolarized"
	      << " noise,"<< endl << "and the Y component is the orthogonal "
	      << "noise." << endl << endl;
      
      oOptSignal1 = new OptSignal(*oOptSignal);

      oOptSignal2 = new OptSignal(*oOptSignal);

      oOptSignal3 = new OptSignal(*oOptSignal);

      oOptSignal4 = new OptSignal(*oOptSignal); 
      oOptSignalX = new OptSignal(*oOptSignal); 

      if (oOptSignal->GetNumChannels() != 1) 
	{   
         *LogFile << "The OptSignal object has " 
		  << oOptSignal->GetNumChannels()
		  << " channels." << endl
		  << "The temporary OptSignalMethods will be set to single"
		  << " channel." << endl;   
       	      		 
         oOptSignal1->ReduceToSingleChannel(0);      
         oOptSignal2->ReduceToSingleChannel(0);   	 
         oOptSignal3->ReduceToSingleChannel(0);   	 
         oOptSignal4->ReduceToSingleChannel(0); 
         oOptSignalX->ReduceToSingleChannel(0); 
      }
              
      oPhotodetector2 = new Photodetector(*oPhotodetector);
      oPhotodetector2->SetOptSignal(oOptSignal2);
      oElecSignalStat2 = new ElecSignalStat(oOptSignal2,oPhotodetector2);   

      oPhotodetector4 = new Photodetector(*oPhotodetector);
      oPhotodetector4->SetOptSignal(oOptSignal4);
      oElecSignalStat4 = new ElecSignalStat(oOptSignal4,oPhotodetector4);   
              
     } 
   else 
     {
      delete oOptFilter2;
      delete oElecFilter2;
      delete [] VariancesCurrentSignalNoiseBeating;

      delete oOptFilter4;
      delete oElecFilter4;
     }
   InitializedGaussianStatisticsFlag = 1;   
   

   //## Attribution methods should be created to OptFilter and 
   //#  ElecFilter in case massive memory allocations start
   //#  occuring in these objects. This is not the case at this point.

   oOrigOptFilter = oOptFilter; // Added by JZ April 2003
   
   oOptFilter2 = new OptFilter(oOptFilter,oOptSignal2);
   oElecFilter2 = new ElecFilter(oElecFilter,oPhotodetector2);     
   
   oOptFilter4 = new OptFilter(oOptFilter,oOptSignal4);
   oElecFilter4 = new ElecFilter(oElecFilter,oPhotodetector4);  
      
   VariancesCurrentSignalNoiseBeating = new double[StringLength];
   for (int ii = 0; ii < StringLength; ii++) 
      VariancesCurrentSignalNoiseBeating[ii] = 0;
   
   
}			   

// #########################################################

//! Use this method if you want to use the same ElecSignalStat object
//! to compute the GaussianStatistics for oOptSignal
//! at different times during your simulation (JZ, April 2003)
//!
//! Call each time system parameters change.

void ElecSignalStat::ResetOptElecFiltersForGaussianStatistics(void)
{

*oOptSignal1 = *oOptSignal;
*oOptSignal2 = *oOptSignal;
*oOptSignal3 = *oOptSignal;
*oOptSignal4 = *oOptSignal;
*oOptSignalX = *oOptSignal;

oPhotodetector2->SetOptSignal(oOptSignal2);
oOptFilter2->SetRelativeCenterFreq(
  oOrigOptFilter->GetRelativeCenterFreqOptFilter());

}

// #########################################################

void ElecSignalStat::ResetOptElecFilterBandwidthsForGaussianStatistics(
                      double FreqFWHM_OptFilter2, 
                      double f3dB_ElecFilter2)
{
    CheckGaussianStatisticsInitialization(
         "ResetOptElecFilterBandwidthsForGaussianStatistics");

    oOptFilter2->SetFreqFWHM_OptFilter(FreqFWHM_OptFilter2);
    oElecFilter2->SetFreqFWHM_ElecFilter(2.0*f3dB_ElecFilter2);    
}

//##########################################################
//##### Check the initialization of the semi-analytical
//##    objects that are required for the statistics
//## By: Ivan Lima  (5/16/02)
void ElecSignalStat::CheckGaussianStatisticsInitialization(
                           string CallingMethod)
{
   if (! InitializedGaussianStatisticsFlag) {
      string OutputBuffer = "\n************************************\n";
      OutputBuffer += "Error From ElecSignalStat::";
      OutputBuffer += CallingMethod;
      OutputBuffer +="(.):\n You should execute the method ";
      OutputBuffer += "SetOptElecFiltersForGaussianStatistics.\n";
      //OutputBuffer += "SetFiltersForGaussianStatistics before.\n";
      OutputBuffer += " Program aborting now.\n\n";
      LogFileSeparator();
      *LogFile << OutputBuffer << endl;           
      cout << OutputBuffer << endl;
      //## Abort the program
      exit(1);
   }
}		   			   

// ###########################################################

//! Returns mean of noise <i_n> given by JLT05_22
/*!

The only difference is that we also multiply by real(H_e(0)) which
was assumed to be 1 in the paper.

*/

double ElecSignalStat::GetMeanNoiseGaussianStatistics(
		     double TotalNoiseSpectralDensity)
{

  // Adapted by J. Zweck from Ivan Lima's method GetMeanASE_ASE

 CheckGaussianStatisticsInitialization("GetMeanASE_ASE");
	                
 oOptSignal2->GenerateImpulseSignal();
 oOptFilter2->FilterOptSignal();
	   
 int qtPoints = oOptSignal2->GetQtPoints();
 double DeltaFreq = oOptSignal2->GetDeltaFreq();

 *oOptSignal3 = *oOptSignal2; 

 //oOptSignal3 now contains transfer function of optical filter

   oOptSignal2->GenerateImpulseSignal();

   for (int ii = 0; ii < qtPoints; ii++) 
    oPhotodetector2->sfftPM.tPowerSignal[ii] = 
      oOptSignal2->sfftM.tOptSignalX[ii];
   
   oPhotodetector2->Time2Freq();
   
   oElecFilter2->FilterElecSignal();
   
   cplx ElecFilterDC_Component = oPhotodetector2->sfftPM.fPowerSignal[0];
   
  
  // The transfer function of the 
  // optical filter is in the X polarization state of oOptSignal3

  double NormMeanNoise = 0;

  for (int ii = 0; ii < qtPoints; ii++) 
    NormMeanNoise += sq(abs(oOptSignal3->sfftM.tOptSignalX[ii]));
                       
   NormMeanNoise *= TotalNoiseSpectralDensity*real(ElecFilterDC_Component);
   
   NormMeanNoise *= oPhotodetector2->GetDeltaFreq()
                       *oPhotodetector2->GetConvFactOptElecPhotod()
                       /qtPoints;				
				  
   return NormMeanNoise;
         
} // ## end GetMeanNoiseGaussianStatistics

// ########################################################

//! Returns noise-noise beating variance given by JLT05_24

double ElecSignalStat::GetNoiseNoiseVarianceGaussianStatistics(
			       double TotalNoiseSpectralDensity,
			       double DOP_Noise)
{

   if(oOptSignal->GetTypeSimulation()==SCALAR)
    {
      cerr << "ElecSignalStat::"
	   << "GetNoiseNoiseVarianceGaussianStatistics"
	   << endl
	   << "is not implemented for SCALAR signals (yet)."
	   << endl
	   << "ABORTING!" << endl;
      exit(1);
    }


  double Variance = 0.5*sq(oPhotodetector2->GetConvFactOptElecPhotod()*
			   TotalNoiseSpectralDensity)*
                        (1+sq(DOP_Noise))*
                        GetIntegralForNoiseNoiseVarianceGaussianStatistics();

  return Variance;
                          
}

// ########################################################

 //! Returns I_{ASE-ASE} given by JLT05_26 using formula in JLT05_53
 //! This integral represents the factor in the noise-noise beating
 //! variance that is independent of the noise statistics, and only
 //! depends on the optical and electrical filters.

double ElecSignalStat::GetIntegralForNoiseNoiseVarianceGaussianStatistics(void)
 {
   CheckGaussianStatisticsInitialization("GetIntegralForNoiseNoiseVariance");
	                  
   oOptSignal2->GenerateImpulseSignal();
   oOptFilter2->FilterOptSignal();
	   
   // Calculate |\tilde{H_0}(-\omega)|^2 and store in X-polarization
   // state of oOptSignal3

   oOptSignal3->sfftM.fOptSignalX[0] = 
	    sq(abs(oOptSignal2->sfftM.fOptSignalX[0]));

   for(int ii = 1; ii < qtPoints; ii++) 
     oOptSignal3->sfftM.fOptSignalX[ii] = 
      sq(abs(oOptSignal2->sfftM.fOptSignalX[qtPoints - ii]));  

   // Calculate InverseFourierTransform of |\tilde{H_0}(-\omega)|^2 to
   // get a function of time
      
   oOptSignal3->Freq2Time();

   oOptSignal2->GenerateImpulseSignal();

   for(int ii = 0; ii < qtPoints; ii++)
    oPhotodetector2->sfftPM.tPowerSignal[ii] = 
      oOptSignal2->sfftM.tOptSignalX[ii];
   
   oPhotodetector2->Time2Freq();
   oElecFilter2->FilterElecSignal();
   
   // Calculate |H_e|^2 
 
   for(int ii = 0; ii < qtPoints; ii++) 
    oPhotodetector2->sfftPM.fPowerSignal[ii] =    
     sq(abs(oPhotodetector2->sfftPM.fPowerSignal[ii]));   

   oPhotodetector2->Freq2Time();      
  
   double Integral = 0;
  
   for(int ii = 0; ii < qtPoints; ii++)
      Integral += sq(abs(oOptSignal3->sfftM.tOptSignalX[ii]))
	          *real(oPhotodetector2->sfftPM.tPowerSignal[ii]);

   // JZ Check is it really a sq in the following line? (I guess comes from
   // fact have two FFTs in this integral.)
                              
   Integral *= sq(oPhotodetector2->GetDeltaFreq())/qtPoints;

   return Integral;  
                                      
}

// #######################################################


//! Returns the signal-noise beating variance at time Time 
//! for an arbitrary signal in case the noise is  unpolarized 
//! Gaussian white noise, using (a variant of) JLT05_32.

double 
ElecSignalStat::GetSignalNoiseVarianceForUnpolarizedNoiseGaussianStatistics
                       (double Time,
                        double TotalNoiseSpectralDensity)
{

  double Variance;

  if(oOptSignal->GetTypeSimulation()==SCALAR)
    {
      cerr << "ElecSignalStat::"
	   << "GetSignalNoiseVarianceUnpolarizedNoiseGaussianStatistics"
	   << endl
	   << "is not implemented for SCALAR signals (yet)."
	   << endl
	   << "ABORTING!" << endl;
      exit(1);
    }
  else
    {
     double epsilon = 5e-4;
     double PowerX = 0.0;
     double PowerY = 0.0;

     for(int ii=0;ii<qtPoints;ii++)
      PowerX += sq(abs(oOptSignal->sfftM.tOptSignalX[ii]));
     for(int ii=0;ii<qtPoints;ii++)
      PowerY += sq(abs(oOptSignal->sfftM.tOptSignalY[ii]));
   
     if( (PowerX + PowerY) < ((1+epsilon)*PowerX)) // All power in X-pol
      {       
        Variance = GetIntegralForSignalNoiseVariance(oOptSignal,
	  					     oOptSignal,0.0,0.0,Time);
      }
     else
       {
         *oOptSignalX = *oOptSignal;
	 oOptSignalX->EliminateY_Polarization();

         Variance = GetIntegralForSignalNoiseVariance(oOptSignalX,
						      oOptSignalX,0.0,0.0,Time);

         *oOptSignalX = *oOptSignal;
	 oOptSignalX->SwitchXandYPolarizations();
	 oOptSignalX->EliminateY_Polarization();

         Variance += GetIntegralForSignalNoiseVariance(oOptSignalX,
	  				               oOptSignalX,0.0,0.0,Time);
       }

     double GammaS_ASE = 0.5;
   
     Variance *= GammaS_ASE*sq(oPhotodetector2->GetConvFactOptElecPhotod())
                 *TotalNoiseSpectralDensity;
    }

 return Variance;

} // ## end GetSignalNoiseVarianceForUnpolarizedNoiseGaussianStatistics

// ###############################################################

//! Returns the signal-noise beating variance at time Time 
//! for a polarized VECTOR signal in case of (possibly) partially 
//! polarized Gaussian white noise, using JLT05_32.
//! oOptSignal is assumed to be X-polarized!

double 
ElecSignalStat::GetSignalNoiseVarianceForPolarizedSignalGaussianStatistics
                       (double Time,
			double * UnitStokesSignal,
			double * UnitStokesNoise,
                        double TotalNoiseSpectralDensity,
			double DOP_Noise)
{

 double Variance = GetIntegralForSignalNoiseVariance(oOptSignal,
						     oOptSignal,0.0,0.0,Time);

 double GammaS_ASE = 0.0;

 for(int ii=0;ii<3;ii++)
   GammaS_ASE += UnitStokesSignal[ii]*UnitStokesNoise[ii];

 GammaS_ASE = 0.5*(1+DOP_Noise*GammaS_ASE);

 Variance *= GammaS_ASE*sq(oPhotodetector2->GetConvFactOptElecPhotod())
             *TotalNoiseSpectralDensity;

 return Variance;

}

// ########################################################

 //! Returns I_{S-ASE}(t) given by JLT05_29 using formula in JLT05_54
 //! where e_(\tau) is replaced by OptSignalA and e_(\tau') by OptSignalB
 //! and Time is the time t. 

 /*! Integrals of this sort represent terms in the formula for the
     signal-noise beating variance that are independent of the 
     noise statistics, and only depend on the optical and electrical 
     filters, and the noise-free signal. This function is called by
     the various versions of GetSignalNoiseVariance with appropriate
     choices of OptSignalA, OptSignalB, and Time.

     It is assumed that OptSignalA and OptSignalB are X-polarized!

     The integral is 

     I(t) = 2\int\int conj(SigA(w))SigB(z)r_0(z-w)h_e(t-w)h_e(t-z) dw dz

     which we compute using the formula

    I(t) = 2 invFT[ H_e FT[SigB invFT( FT(G_s) |\tilde H_0|^2) ]](t) 
   
   
    evaluated at s=t, where G_s(w) = conj(SigA(w))h_e(s-w). 

 */

double ElecSignalStat::GetIntegralForSignalNoiseVariance(OptSignal *OptSignalA,
							 OptSignal *OptSignalB,
							 double TimeShiftA, 
							 double TimeShiftB,
	                                                 double Time)
{
   CheckGaussianStatisticsInitialization("GetIntegralForSignalNoiseVariance");

   double epsilon = 1e-3;

   double PowerX = 0.0;
   double PowerY = 0.0;

   for(int ii=0;ii<qtPoints;ii++)
     PowerX += sq(abs(OptSignalA->sfftM.tOptSignalX[ii]));
   for(int ii=0;ii<qtPoints;ii++)
     PowerY += sq(abs(OptSignalA->sfftM.tOptSignalY[ii]));
   
   if( (PowerX + PowerY) > ((1+epsilon)*PowerX))
     {
       cerr << "ElecSignalStat::GetIntegralForSignalNoiseVariance " << endl
	    << "OptSignalA must be X-polarized!!" << endl
	    << "ABORTING!" << endl;

       cerr << "PowerX, PowerY = " << PowerX << " " <<  PowerY << endl;

       exit(1);
     }

   PowerX=0.0;
   PowerY=0.0;

   for(int ii=0;ii<qtPoints;ii++)
     PowerX += sq(abs(OptSignalB->sfftM.tOptSignalX[ii]));
   for(int ii=0;ii<qtPoints;ii++)
     PowerY += sq(abs(OptSignalB->sfftM.tOptSignalY[ii]));
   
   if( (PowerX + PowerY) > ((1+epsilon)*PowerX))
     {
       cerr << "ElecSignalStat::GetIntegralForSignalNoiseVariance " << endl
	    << "OptSignalB must be X-polarized!!" << endl
	    << "ABORTING!" << endl;
       cerr << "PowerX, PowerY = " << PowerX << " " <<  PowerY << endl;
       exit(1);
     }

   while(Time < 0)
    Time += oOptSignal->GetTimeWindow();
   while(Time >= oOptSignal->GetTimeWindow())
    Time -= oOptSignal->GetTimeWindow();

   int TimeIndex = int(floor(Time/oOptSignal->GetDeltaTime()));

   // Set oOptSignal1 to be optically filtered OptSignalA

   *oOptSignal2 = *OptSignalA;
   oOptSignal2->TimeShiftSignal(TimeShiftA); // U_s(t) = U(t-s), shift to right
   oOptFilter2->FilterOptSignal();
  
   *oOptSignal1 = *oOptSignal2; 

   // Set oOptSignal4 to be optically filtered OptSignalB

   *oOptSignal2 = *OptSignalB;
   oOptSignal2->TimeShiftSignal(TimeShiftB);
   oOptFilter2->FilterOptSignal();
  
   *oOptSignal4 = *oOptSignal2; 

   // Set oOptSignal3 to be transfer function of optical filter

   oOptSignal2->GenerateImpulseSignal();
   oOptFilter2->FilterOptSignal();

   *oOptSignal3 = *oOptSignal2;
	   
   // Set oPhotodetector2 to be transfer function of electrical filter
               
   oOptSignal2->GenerateImpulseSignal();

   for(int ii = 0; ii < qtPoints; ii++) 
    oPhotodetector2->sfftPM.tPowerSignal[ii] = 
          oOptSignal2->sfftM.tOptSignalX[ii];
   
   oPhotodetector2->Time2Freq();
   oElecFilter2->FilterElecSignal();
   
   // Compute FT(G_s(w))

   for (int ii = 0; ii < qtPoints; ii++) 
    {
      oOptSignal2->sfftM.tOptSignalX[ii] = 
          conj(oOptSignal1->sfftM.tOptSignalX[ii])
         *oPhotodetector2->sfftPM.tPowerSignal[
	  Periodicity(TimeIndex - ii,qtPoints)];
    }

   for (int ii = 0; ii < qtPoints; ii++)
      oOptSignal2->sfftM.tOptSignalY[ii] = 0.0;

   oOptSignal2->Time2Freq();

   // Compute invFT( FT(G_s(w)) |\tilde H_0|^2 )

 oOptSignal2->sfftM.fOptSignalX[0] *= 
   sq(abs(oOptSignal3->sfftM.fOptSignalX[0]));

 for(int ii = 1; ii < qtPoints; ii++) 
   oOptSignal2->sfftM.fOptSignalX[ii] *= 
    sq(abs(oOptSignal3->sfftM.fOptSignalX[qtPoints - ii]));  
            
 oOptSignal2->Freq2Time();

 // Compute FT[SigB invFT( FT(G_s) |\tilde H_0|^2) ]

   for (int ii = 0; ii < qtPoints; ii++) 
      oOptSignal2->sfftM.tOptSignalX[ii] *= oOptSignal4->sfftM.tOptSignalX[ii];
    	
   oOptSignal2->Time2Freq();   


   // Compute invFT[ H_e FT[SigB invFT( FT(G_s) |\tilde H_0|^2) ]]

for(int ii = 0; ii < qtPoints; ii++) 
oOptSignal2->sfftM.fOptSignalX[ii] *= oPhotodetector2->sfftPM.fPowerSignal[ii];
   
oOptSignal2->Freq2Time();      
 
// This result is only good at TimeIndex
    
 double Integral = 2.0*real(oOptSignal2->sfftM.tOptSignalX[TimeIndex])
	              *oPhotodetector2->GetDeltaFreq();
 if(Integral < 0)
    Integral = 0.0;

   return Integral;

} // ## end GetIntegralForSignalNoiseVariance

// #####################################################



//##########################################################
//##### Get the enhancement factor for a given format
//##    This factor is used to convert OSNR into ESNR
//#     in order to compute the Q-factor.
//#     The optical and the electrical filters are the
//#     ones that comprize the receiver, which is modeled
//#     as a square law detector.
//##    Warning: it is very important that oOptSignal
//#     be absolutely noiseless at this point. It should be
//#     as if it were in front of the receiver.
//## By: Ivan Lima  (5/3/02)
double ElecSignalStat::GetEnhancementFactor(double OSA_NoiseBandwidth)
{

   CheckGaussianStatisticsInitialization("GetEnhancementFactor");

   //#######
   //## Get the ratio between the normalized MaxCurrent 
   //#  and the MeanOpticalPower
   /////oOptSignal2->RegenerateSignalString();   
   *oOptSignal2 = *oOptSignal;
   //## It is very important to have equal marks and spaces in
   //#  the bit string.
   double MeanOpticalPower = oOptSignal2->GetAveragePower();
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatistics();   
       
   double MinCurrentBit1 =  oElecSignalStat2->GetMinCurrentBit1(); 
   //double MaxCurrentBit0 =  oElecSignalStat2->GetMaxCurrentBit0();

   //cout << "$#$# MinCurrentBit1 = " << MinCurrentBit1 << endl;
   //cout << "#$#$ MaxCurrentBit0 = " << MaxCurrentBit0 << endl;
   //cout << "#$#$ AverageSignalPower = " << oOptSignal2->GetAveragePower() << endl;
   //oOptSignal2->WriteFileTimePower("tOutOptSignal.dat");   
   //oOptSignal2->WriteFileFreqPower("fOutOptSignal.dat");

   double PeakCurrent_AverageOpticalPower = (  MinCurrentBit1
                               /oPhotodetector2->GetConvFactOptElecPhotod() )
			      /MeanOpticalPower; 

   //cout << "PeakCurrent_AverageOpticalPower = " 
   //     << PeakCurrent_AverageOpticalPower  << endl;
   
   //double PeakCurrent_AverageOpticalPower = ( ( MinCurrentBit1
   //                              -MaxCurrentBit0 )
   //                            /oPhotodetector2->GetConvFactOptElecPhotod() )
   //			      /MeanOpticalPower; 
  
   double EffectiveOpticalFilterBandwidth_OSA_NoiseBandwidth =
               GetMeanASE_ASE(1.)
	       /(oPhotodetector2->GetConvFactOptElecPhotod() 
	         *OSA_NoiseBandwidth);

   //## These lines are useful for debugging
   //cout << "   EffectiveNoiseBandwidth (GHz) = " 
   //     << GetMeanASE_ASE(1.)
   //	   /oPhotodetector2->GetConvFactOptElecPhotod()*1e-9 
   //	<< endl;
   //cout << "   OSA_NoiseBandwidth (GHz) = " << OSA_NoiseBandwidth*1e-9 << endl;
   //cout << "   EffectiveOpticalFilterBandwidth_OSA_NoiseBandwidth = "
   //     <<EffectiveOpticalFilterBandwidth_OSA_NoiseBandwidth << endl;


   double EnhancementFactor = PeakCurrent_AverageOpticalPower
                       /EffectiveOpticalFilterBandwidth_OSA_NoiseBandwidth;  
  
   if (EnhancementFactor < 0. ) {
      cout << "Warning from ElecSignalStat::GetEnhancementFactor(.)"<<endl
           << "Enhancement factor was found to be negative."
           << "The filters are probably too narrow" << endl;
      cout << "EnhancementFactor will artificially be set to zero."<< endl;	   
      EnhancementFactor = 0.;
   }   
   
   /**
   //$$$$$$$$ This is a test where the effective noise bandwidth is 
   //$$       considered. I should comment it afterwards.
   //$$       But do not delete these comments, please.
   //$$  Definitively, this should not be here (IL 5/14/02)
   double RatioEffectiveNoiseBandwidth_BandwidthOptFilter 
          = GetMeanASE_ASE(1.,oOptFilter,oElecFilter)
	   /oOptFilter->GetFreqFWHM_OptFilter();
   cout << "RatioEffectiveNoiseBandwidth_BandwidthOptFilter = " 
        << RatioEffectiveNoiseBandwidth_BandwidthOptFilter << endl;	   
   EnhancementFactor /= RatioEffectiveNoiseBandwidth_BandwidthOptFilter;
   //$$ End of test
   **/
  
   return EnhancementFactor;
}


// #################################################################

double ElecSignalStat::GetNormalizedEnhancementFactor(void)
{

  // By JZ adapted from code by IL, April 2004

 //## It is very important to have an equal number marks and spaces in
 //#  the bit string for the mean optical power

   CheckGaussianStatisticsInitialization(
    "GetNormalizedEnhancementFactor");

   *oOptSignal2 = *oOptSignal;
   
   double MeanOpticalPower = oOptSignal2->GetAveragePower();
   
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatistics();   
       
   double MinCurrentBit1 =  oElecSignalStat2->GetMinCurrentBit1(); 
  

   double PeakCurrent_AverageOpticalPower = (  MinCurrentBit1
                               /oPhotodetector2->GetConvFactOptElecPhotod() )
			      /MeanOpticalPower; 

   return PeakCurrent_AverageOpticalPower;

} // ## end GetNormalizedEnhancementFactor


//########################################################
//##### Get the normalized Kcoeff that is used by the semi-analytical
//##    Q-factor based on Marcuse's formula.
//##    In order to get the actual Kcoeff you need to multiply it
//##    by
//##    NoiseSpectralDensityParalel
//##   *(NoiseSpectralDensityParalel + NoiseSpectralDensityOrthog)/sqrt(2.)
//##   /( sq(NoiseSpectralDensityX) + sq(NoiseSpectralDensityY) );
//##    Note that in Marcuse's formula this number is equal to 2.
//## By: Ivan Lima
double ElecSignalStat::GetNormKcoeff1_GaussianStatistics(void)
{
   CheckGaussianStatisticsInitialization(
         "GetNormKcoeff1_GaussianStatistics");

   //## Get MinCurrentBit1 and its the BitSlot
   *oOptSignal2 = *oOptSignal;
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#     
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatistics();
   int WorstBitSlot1;
   double MinCurrentBit1 = oElecSignalStat2->GetMinCurrentBit1(&WorstBitSlot1);

   double NumberNoiseModes = sq( GetMeanASE_ASE(0.5,0.5)
                                /GetStdDevASE_ASE(0.5,0.5) ); 
 
   double NormKcoeff1 = sqrt(NumberNoiseModes)
                  *sq(GetStdDevSignal_ASE_BitSlot(WorstBitSlot1,0.5,0.5) )
		  /( MinCurrentBit1*GetStdDevASE_ASE(0.5,0.5) );   
   //cout << "NumberNoiseModes = " << NumberNoiseModes << endl;		  
   //cout << "Kcoeff = " << Kcoeff << endl;  
//   cout << "WorstBitSlot1 = " << WorstBitSlot1 << endl;
//   cout << "Sig_ASE / ASE_ASE = " 
//        << GetStdDevSignal_ASE(WorstBitSlot1,0.5,0.5) 
//	  /GetStdDevASE_ASE(1.) << endl;
   		  
   return NormKcoeff1;
}



//########################################################
//##### Get the Kcoeff that is used by the semi-analytical
//##    Q-factor that has finite extinctio ratio.
//##    In order to get the actual Kcoeff you need to multiply it
//##    by
//##    NoiseSpectralDensityParalel
//##   *(NoiseSpectralDensityParalel + NoiseSpectralDensityOrthog)/sqrt(2.)
//##   /( sq(NoiseSpectralDensityX) + sq(NoiseSpectralDensityY) );
//## By: Ivan Lima
double ElecSignalStat::GetNormKcoeff0_GaussianStatistics(void)
{
   CheckGaussianStatisticsInitialization(
         "GetNormKcoeff0_GaussianStatistics");

   //## Get MinCurrentBit1 and its the BitSlot
   *oOptSignal2 = *oOptSignal;
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#     
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatistics();
   int WorstBitSlot0;
   double MaxCurrentBit0 = oElecSignalStat2->GetMaxCurrentBit0(&WorstBitSlot0);

   double NumberNoiseModes = sq( GetMeanASE_ASE(0.5,0.5)
                                /GetStdDevASE_ASE(0.5,0.5) ); 
 
   double NormKcoeff0 = sqrt(NumberNoiseModes)
                  *sq(GetStdDevSignal_ASE_BitSlot(WorstBitSlot0,0.5,0.5) )
		  /( MaxCurrentBit0*GetStdDevASE_ASE(0.5,0.5) );   
   //cout << "NumberNoiseModes = " << NumberNoiseModes << endl;		  
   //cout << "Kcoeff = " << Kcoeff << endl;  
//   cout << "WorstBitSlot1 = " << WorstBitSlot1 << endl;
//   cout << "Sig_ASE / ASE_ASE = " 
//        << GetStdDevSignal_ASE(WorstBitSlot1,0.5,0.5) 
//	  /GetStdDevASE_ASE(1.) << endl;
   		  
   return NormKcoeff0;
}



//########################################################
//##### Get the normalized number of noise modes that is 
//##    used by the semi-analytical
//##    Q-factor based on Marcuse's formula.
//##    To get the actual number of noise modes you need to 
//##    multiply this normalized number to
//##    sq(NoiseSpectralDensityX + NoiseSpectralDensityY)/2
//##   /( sq(NoiseSpectralDensityX) + sq(NoiseSpectralDensityY) )
//##    Note that in Marcuse's formula this number is equal to 2T*Bo
//## By: Ivan Lima
double ElecSignalStat::GetNormNumberNoiseModes_GaussianStatistics(void)
{
   CheckGaussianStatisticsInitialization(
         "GetNormNumberNoiseModes_GaussianStatistics");

   //## Get MinCurrentBit1 and its the BitSlot
   *oOptSignal2 = *oOptSignal;
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$# 
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatistics();
   //int WorstBitSlot1;
   //double MinCurrentBit1 = oElecSignalStat2->GetMinCurrentBit1(&WorstBitSlot1);

   double NormNumberNoiseModes = sq( GetMeanASE_ASE(0.5,0.5)
                                    /GetStdDevASE_ASE(0.5,0.5) );    		  
   return NormNumberNoiseModes;
}



//########################################################
//##### Get the extinctio ratio after the optical and 
//##    electrical filters for a noiseless signal.
//## By: Ivan Lima
double ElecSignalStat::GetCurrentExtinctionRatio_GaussianStatistics(void)
{
   CheckGaussianStatisticsInitialization(
         "GetCurrentExtinctionRatio_GaussianStatistics");

   //## Get MinCurrentBit1 and its the BitSlot
   *oOptSignal2 = *oOptSignal;
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#    
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatistics();
   double MinCurrentBit1 = oElecSignalStat2->GetMinCurrentBit1();
   double MaxCurrentBit0 = oElecSignalStat2->GetMaxCurrentBit0();
 
   double ExtinctionRatio = MaxCurrentBit0/MinCurrentBit1;
   		  
   return ExtinctionRatio;
}








//##########################################################
//##### This is just an interface for the case where the 
//#     noise is unpolarized. This corresponds to the case
//#     where there is no PDL and no PDG in the system.
//## By: Ivan Lima  (5/25/02)
double ElecSignalStat::GetMeanASE_ASE(double NoiseSpectralDensity)
{
   return GetMeanASE_ASE(NoiseSpectralDensity/2.,
                         NoiseSpectralDensity/2.);
}



//##########################################################
//##### Get the mean of the ASE noise-noise 
//##    beating that can be used to compute the Q_factor
//#     semi-analitically from the noise spectral density.
//#     This is useful for the case of infinite extintio ratio.
//###   This is the term that goes as noise in the SNR 
//#     expression in Marcuse's JLT'90 paper.
//#     The optical and the electrical filters are the
//#     ones that comprize the receiver, which is modeled
//#     as a square law detector.
//## By: Ivan Lima  (5/17/02)
double ElecSignalStat::GetMeanASE_ASE(double NoiseSpectralDensityX,
                                      double NoiseSpectralDensityY)
{
   CheckGaussianStatisticsInitialization("GetMeanASE_ASE");
	      
   //#######
   //## First, I should obtain the effective noise bandwidth
   //## Get the impulse response of the optical and electrical filters
   //## It is required to obtain the effective noise bandwidth
            
   oOptSignal2->GenerateImpulseSignal();
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#     
   oOptFilter2->FilterOptSignal();
	   
   int qtPoints = oOptSignal2->GetQtPoints();
   double DeltaFreq = oOptSignal2->GetDeltaFreq();


   /*
   //##  These lines are good for debugging      
   //#### Get the effective optical bandwidth
   double EffectiveOptFilterBandwidth = 0;
   for (int ii = 0; ii < qtPoints; ii++) {
      EffectiveOptFilterBandwidth +=       
          sq(abs(oOptSignal2->sfftM.fOptSignalX[ii]));  
   }
   EffectiveOptFilterBandwidth *= oOptSignal2->GetDeltaFreq();            
   cout << "Effective OptFilter bandwidth (GHz) = " 
        << EffectiveOptFilterBandwidth*1e-9 << endl; 	   
   oOptSignal2->WriteFileTimePower("tOutOptSignal2.dat");
   oOptSignal2->WriteFileFreqPower("fOutOptSignal2.dat");  
   */

   //## Transfer the inpulse response of the optical filter to oOptSignal3
   *oOptSignal3 = *oOptSignal2;

/*
   //## Get the Fourier transform of the square of the abs of the 
   //#  impulse response of the optical filter
   for (int ii = 0; ii < qtPoints; ii++) {
      oOptSignal3->sfftM.tOptSignalX[ii] = 
           sq(abs(oOptSignal2->sfftM.tOptSignalX[ii]));
      oOptSignal3->sfftM.tOptSignalY[ii] = 
            sq(abs(oOptSignal2->sfftM.tOptSignalY[ii]));	    
   }     
   oOptSignal3->Time2Freq(); 
*/
     

   oOptSignal2->GenerateImpulseSignal();

   for (int ii = 0; ii < qtPoints; ii++) {
      oPhotodetector2->sfftPM.tPowerSignal[ii] = 
            oOptSignal2->sfftM.tOptSignalX[ii];
   }   
   oPhotodetector2->Time2Freq();
   
   oElecFilter2->FilterElecSignal();
   
   cplx ElecFilterDC_Component = oPhotodetector2->sfftPM.fPowerSignal[0];
   
   //cout << "oPhotodetector2->sfftPM.fPowerSignal[0] = "
   //     << oPhotodetector2->sfftPM.fPowerSignal[0] << endl;

   /***
   //## Get the integral of the inpulse response of the electrical filter
   //## For pure low pass electrical filters this step is unecessary,
   //## But I am leaving it here for the sake of completeness. 
   cplx IntElecFilter = 0;
   for (int ii = 0; ii < qtPoints; ii++) {
      IntElecFilter += 
         oPhotodetector2->sfftPM.tPowerSignal[ii];
   }
   IntElecFilter /= qtPoints;   
   //cout << "IntElecFilter = " 
   //     << IntElecFilter << endl;	
   ***/
   
   /***  
   //##  These lines are good for debugging
   //##  Get the effective electrical filter bandwidth
   double EffectiveElecFilterBandwidth = 0;
   for (int ii = 0; ii < qtPoints; ii++) {
      EffectiveElecFilterBandwidth += 
         sq(abs(oPhotodetector2->sfftPM.fPowerSignal[ii]));
   }
   EffectiveElecFilterBandwidth *= oPhotodetector2->GetDeltaFreq();

   oPhotodetector2->WriteFileTimePower("tOutSignal2.dat");
   oPhotodetector2->WriteFileFreqPower("fOutSignal2.dat");  
   cout << "Effective ElecFilter bandwidth (GHz) = " 
        << EffectiveElecFilterBandwidth*1e-9 << endl;
   ***/
   
   /*
   for (int ii = 0; ii < qtPoints; ii++) {
      oPhotodetector2->sfftPM.fPowerSignal[ii] = 
            NoiseSpectralDensityX*  
            oOptSignal3->sfftM.fOptSignalX[ii]
            *conj(oPhotodetector2->sfftPM.fPowerSignal[ii])
           +NoiseSpectralDensityY*  
            oOptSignal3->sfftM.fOptSignalY[ii]
            *conj(oPhotodetector2->sfftPM.fPowerSignal[ii]);
   }   
   oPhotodetector2->Freq2Time();
   */

   /*
   for (int ii = 0; ii < qtPoints; ii++) {
      oPhotodetector2->sfftPM.fPowerSignal[ii] = 
            NoiseSpectralDensityX*  
            oOptSignal3->sfftM.fOptSignalX[ii]
           +NoiseSpectralDensityY*  
            oOptSignal3->sfftM.fOptSignalY[ii]
            *conj(oPhotodetector2->sfftPM.fPowerSignal[ii]);
   }   
   oPhotodetector2->Freq2Time();
   */

   /*
  cplx cNormMeanASE_ASE = 0;
  for (int ii = 0; ii < qtPoints; ii++) {
     cNormMeanASE_ASE += oPhotodetector2->sfftPM.tPowerSignal[ii];    
   } 
   */   

  //## Note: The optical sygnalX has to be in both X and Y
  //## contributions because the impulse response of the
  //## optical filter was generated only in the X polarization
  cplx cNormMeanASE_ASE = 0;
  for (int ii = 0; ii < qtPoints; ii++) {
     cNormMeanASE_ASE += 
            NoiseSpectralDensityX*  
            sq(abs(oOptSignal3->sfftM.tOptSignalX[ii]))
           +NoiseSpectralDensityY*  
            sq(abs(oOptSignal3->sfftM.tOptSignalX[ii]));  
   } 
   //cNormMeanASE_ASE *= IntElecFilter;

   cNormMeanASE_ASE *= ElecFilterDC_Component;
   
   cNormMeanASE_ASE *= oPhotodetector2->GetDeltaFreq()
                                /qtPoints;				
				   
   double NormMeanASE_ASE = real(cNormMeanASE_ASE);

   return NormMeanASE_ASE
         *oPhotodetector2->GetConvFactOptElecPhotod();                                        
//   return NoiseSpectralDensity*EffectiveNoiseBandwidth
//         *oPhotodetector2->GetConvFactOptElecPhotod();
}

//##########################################################
//##### This is just an interface for the case where the noise
//##    is totaly depolarized, as in the absence of PDL and PDG.
//## By: Ivan Lima  (5/11/02)
double ElecSignalStat::GetStdDevASE_ASE(double NoiseSpectralDensity)
{
      return GetStdDevASE_ASE(NoiseSpectralDensity/2.,
                              NoiseSpectralDensity/2.);
}



//##########################################################
//##### Get the standard deviation of the ASE noise-noise 
//##    beating that can be used to compute the Q_factor
//#     semi-analitically from the noise spectral density
//#     The optical and the electrical filters are the
//#     ones that comprize the receiver, which is modeled
//#     as a square law detector.
//## By: Ivan Lima  (5/11/02)
double ElecSignalStat::GetStdDevASE_ASE(double NoiseSpectralDensityX,
                                        double NoiseSpectralDensityY)
{
   CheckGaussianStatisticsInitialization("GetStdDevASE_ASE");
	      
   //#######
   //## First, I should obtain the effective noise bandwidth
   //## Get the impulse response of the optical and electrical filters
   //## It is required to obtain the effective noise bandwidth
            
   oOptSignal2->GenerateImpulseSignal();
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#     
   oOptFilter2->FilterOptSignal();
	   
   int qtPoints = oOptSignal2->GetQtPoints();
   double DeltaFreq = oOptSignal2->GetDeltaFreq();

   /***
   //##  These lines are good for debugging      
   //#### Get the effective optical bandwidth
   double EffectiveOptFilterBandwidth = 0;
   for (int ii = 0; ii < qtPoints; ii++) {
      EffectiveOptFilterBandwidth +=       
          sq(abs(oOptSignal2->sfftM.fOptSignalX[ii]));
   }
   EffectiveOptFilterBandwidth *= oOptSignal2->GetDeltaFreq();            
   cout << "Effective OptFilter bandwidth (GHz) = " 
        << EffectiveOptFilterBandwidth*1e-9 << endl; 	   
   oOptSignal2->WriteFileTimePower("tOutOptSignal2.dat");
   oOptSignal2->WriteFileFreqPower("fOutOptSignal2.dat");  
   ***/


   //## Get the autocorrelation of the optical filter: abs(B(-\omega))^2
   for (int ii = 0; ii < qtPoints; ii++) {
      if (ii == 0 ) {   
         oOptSignal3->sfftM.fOptSignalX[ii] = 
	    sq(abs(oOptSignal2->sfftM.fOptSignalX[ii]));      
      } else {
         oOptSignal3->sfftM.fOptSignalX[ii] = 
	    sq(abs(oOptSignal2->sfftM.fOptSignalX[qtPoints - ii]));  
      }       
   }      
   oOptSignal3->Freq2Time();



/*
   //## Get the autocorrelation of the optical filter: B(-\omega)B*(\omega)
   for (int ii = 0; ii < qtPoints; ii++) {
      //## If ii = 0 || ii = qtPoints/2 it is just this
      cplx cOptSignal2_X_MinusOmega = oOptSignal2->sfftM.fOptSignalX[ii];
      cplx cOptSignal2_Y_MinusOmega = oOptSignal2->sfftM.fOptSignalY[ii];      
      if (ii != 0 && ii != qtPoints/2) {
         cOptSignal2_X_MinusOmega = 
	       oOptSignal2->sfftM.fOptSignalX[qtPoints - ii];
         cOptSignal2_Y_MinusOmega = 
	       oOptSignal2->sfftM.fOptSignalY[qtPoints - ii];      
      }      
      oOptSignal3->sfftM.fOptSignalX[ii] = 
            cOptSignal2_X_MinusOmega 
           *conj(oOptSignal2->sfftM.fOptSignalX[ii]);
      oOptSignal3->sfftM.fOptSignalY[ii] = 
            cOptSignal2_Y_MinusOmega
            *conj(oOptSignal2->sfftM.fOptSignalY[ii]);	    
   }     
   oOptSignal3->Freq2Time();
*/



   /*
   //## Get the autocorrelation of the optical filter: |B(\omega)|^2
   for (int ii = 0; ii < qtPoints; ii++) {
      oOptSignal3->sfftM.fOptSignalX[ii] = 
           sq(abs(oOptSignal2->sfftM.fOptSignalX[ii]));
      oOptSignal3->sfftM.fOptSignalY[ii] = 
            sq(abs(oOptSignal2->sfftM.fOptSignalY[ii]));	    
   }     
   oOptSignal3->Freq2Time();
   */
      
   //####
   //## At this point the time domain vector of oOptSignal3 contains
   //#  the autocorrelation function of the impulse response of 
   //#  the optical filter.

   oOptSignal2->GenerateImpulseSignal();

   for (int ii = 0; ii < qtPoints; ii++) {
      oPhotodetector2->sfftPM.tPowerSignal[ii] = 
            oOptSignal2->sfftM.tOptSignalX[ii];
   }   
   oPhotodetector2->Time2Freq();
   
   oElecFilter2->FilterElecSignal();
   
   //## Plot the impulse response of the electrical filter, if necessary
   //oPhotodetector2->WriteFileTimePower("tOutSignal.dat");
   //exit(1);
   
   /***  
   //##  These lines are good for debugging
   //##  Get the effective electrical filter bandwidth
   double EffectiveElecFilterBandwidth = 0;
   for (int ii = 0; ii < qtPoints; ii++) {
      EffectiveElecFilterBandwidth += 
         sq(abs(oPhotodetector2->sfftPM.fPowerSignal[ii]));
   }
   EffectiveElecFilterBandwidth *= oPhotodetector2->GetDeltaFreq();

   oPhotodetector2->WriteFileTimePower("tOutSignal2.dat");
   oPhotodetector2->WriteFileFreqPower("fOutSignal2.dat");  
   cout << "Effective ElecFilter bandwidth (GHz) = " 
        << EffectiveElecFilterBandwidth*1e-9 << endl;
   ***/

    
   //## Get the autocorrelation of the electrical filter: H(-\omega)H(\omega)
   //## Because the electrical signal is real -> H(-\omega) = conj(H(\omega))
   //## Thus, I will get simply H(\omega)conj(H(\omega)) = sq(abs(H(\omega)))
   for (int ii = 0; ii < qtPoints; ii++) {
      oPhotodetector2->sfftPM.fPowerSignal[ii]      
            = sq(abs(oPhotodetector2->sfftPM.fPowerSignal[ii]));   
//      oPhotodetector2->sfftPM.fPowerSignal[ii]      
//            *= conj(oPhotodetector2->sfftPM.fPowerSignal[ii]);	    
   }     
   oPhotodetector2->Freq2Time();      


   /***   
   for (int ii = 0; ii < qtPoints; ii++) {
      cplx cPhotodetector2 = oPhotodetector2->sfftPM.fPowerSignal[ii];      
      if (ii != 0 && ii != qtPoints/2) {
         cPhotodetector2 = oPhotodetector2->sfftPM.fPowerSignal[qtPoints - ii];
      }	 
      oPhotodetector2->sfftPM.fPowerSignal[ii]      
            = cPhotodetector2 
	     *oPhotodetector2->sfftPM.fPowerSignal[ii];	    
   }     
   oPhotodetector2->Freq2Time();      
   ***/   
   
   /*
   //## Get the autocorrelation of the electrical filter: |H(\omega)|^2
   for (int ii = 0; ii < qtPoints; ii++) {
      oPhotodetector2->sfftPM.fPowerSignal[ii]
            = sq(abs(oPhotodetector2->sfftPM.fPowerSignal[ii]));	    
   }     
   oPhotodetector2->Freq2Time();      
   */

   cplx sqNormASE_ASE_Beating = 0;
   for (int ii = 0; ii < qtPoints; ii++) {
      //## Note that the impulse response is only in the X polarization
      sqNormASE_ASE_Beating += sq(NoiseSpectralDensityX)
           *sq(abs(oOptSignal3->sfftM.tOptSignalX[ii]))
            *oPhotodetector2->sfftPM.tPowerSignal[ii]
                              +sq(NoiseSpectralDensityY)
           *sq(abs(oOptSignal3->sfftM.tOptSignalX[ii]))
            *oPhotodetector2->sfftPM.tPowerSignal[ii];
   } 
   
   sqNormASE_ASE_Beating *= sq(oPhotodetector2->GetDeltaFreq())
                                /qtPoints;

   //cout << "sqNormASE_ASE_Beating = " << sqNormASE_ASE_Beating << endl;
   
   double NormASE_ASE_Beating = sqrt( real(sqNormASE_ASE_Beating));

//   cout <<endl<< "sqrt(sqNormASE_ASE_Beating) (StdDev) = "
//        <<  sqrt(sqNormASE_ASE_Beating)<<endl;

   return NormASE_ASE_Beating*oPhotodetector2->GetConvFactOptElecPhotod();                                        
//   return NoiseSpectralDensity*NormASE_ASE_Beating
//          *oPhotodetector2->GetConvFactOptElecPhotod();
}



//##########################################################
//##### Get the standard deviation of the signal-ASE 
//##    beating for a given bit (which can be a logical ZERO)
//##    Warning: It is very important that signal be
//#     noiseless and that it is not optically filtered.
//##    This method is convenient for unpolarized noise.
//## By: Ivan Lima  (5/11/02)
double ElecSignalStat::GetStdDevSignal_ASE_BitSlot( int BitSlot,
                                             double NoiseSpectralDensity)
{
   CheckGaussianStatisticsInitialization("GetStdDevSignal_ASE_BitSlot");
   return GetStdDevSignal_ASE_BitSlot(BitSlot, NoiseSpectralDensity/2.,
				           NoiseSpectralDensity/2.);
}					



//##########################################################
//##### Get the standard deviation of the signal-ASE 
//##    beating for a given bit (which can be a logical ZERO)
//##    Warning: It is very important that signal be
//#     noiseless and that it is not optically filtered.
//## By: Ivan Lima  (5/11/02)
double ElecSignalStat::GetStdDevSignal_ASE_BitSlot( int BitSlot,
                                        double NoiseSpectralDensityX,
					double NoiseSpectralDensityY)
{ 
   CheckGaussianStatisticsInitialization("GetStdDevSignal_ASE_BitSlot");

   ///////oOptSignal2->RegenerateSignalString();   
//   change by Jonathan Hu
// 	oOptSignal2 = oOptSignal;  
   *oOptSignal2 = *oOptSignal;   
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#    
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatisticsNoTimeShift(); 
   //### I can't use this method here, because it would cause a shift in time      
   ///oElecSignalStat2->ComputePerformanceStatistics();      

   //cout <<"** TargetIndexTime = "  <<oElecSignalStat2->GetTargetIndexTime()<<endl;
   //cout <<"** GetBitStringShift = "<<oElecSignalStat2->GetBitStringShift()<<endl; 
   
//   //## We now have the coordinates of the bits   
//   int TargetIndexTimeBitSlot = Periodicity(
//                      oElecSignalStat2->GetTargetIndexTime()
//	           +( Periodicity(BitSlot,StringLength) 
//		     +oElecSignalStat2->GetBitStringShift() )
//                   *qtPoints/StringLength,
//                    qtPoints);

    int TargetIndexTimeBitSlot = Periodicity(oElecSignalStat2->GetTargetIndexTime()
            +qtPoints/StringLength*
            Periodicity(BitSlot+oElecSignalStat2->GetBitStringShift()
	       ,StringLength),qtPoints);		    
   
   /*
   //$$$ Temporary lines   
   //cout << " $#$# Bit intensity at " << BitSlot <<" = " 
   //     << real(oElecSignalStat2->sfftPM.tPowerSignal[
   //	      TargetIndexTimeBitSlot])
   //     << "  " << TargetIndexTimeBitSlot << endl;	
   for (int ii = 0; ii < qtPoints; ii++) {  
      cout << ii << "  StdDevSignal_ASE = " 
           << GetStdDevSignal_ASE(ii,
                       NoiseSpectralDensity,
                       oOptFilter, 
                       oElecFilter) << endl;
   }
   */
  
   return GetStdDevSignal_ASE_TimeIndex(TargetIndexTimeBitSlot,
                       NoiseSpectralDensityX,NoiseSpectralDensityY); 
}


//##########################################################
//##### Get the standard deviation of the signal-ASE 
//##    beating that can be used to compute the Q_factor
//#     semi-analitically from the noise spectral density
//#     The optical and the electrical filters are the
//#     ones that comprize the receiver, which is modeled
//#     as a square law detector.
//#     Note that this function does not uses the Photodetector
//#     object of the ElecSignalStat, but only the noiseless 
//#     OpticalSignal object to make the computation.
//## By: Ivan Lima  (5/11/02)
double ElecSignalStat::GetStdDevSignal_ASE_TimeIndex(int TargetIndexTime2,
                                        double NoiseSpectralDensityX,
					double NoiseSpectralDensityY)
{
   CheckGaussianStatisticsInitialization("GetStdDevSignal_ASE_TimeIndex");
	      
   //#######
   //## First, I should obtain the spectrum of the noiseless signal.
   //## Second, the impulse response of the optical and electrical filters.
   //## I should use both to integrate over the spectrum domain to obtain
   //## StdDevSignal_ASE
   
   //## Spectrum of the noiseless signal optically filtered in oOptSignal1
   //## Note that it will filter it optically here. Thus, it cannot be
   //#  filtered before.
   // change by Jonathan
//	oOptSignal2 = oOptSignal;

   *oOptSignal2 = *oOptSignal;
   //////oOptSignal2->RegenerateSignalString();   
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#    
   oOptFilter2->FilterOptSignal();
   
   // change by Jonathan
//	oOptSignal1 = oOptSignal2;   
   *oOptSignal1 = *oOptSignal2;


  //## Frequency response of the optical filter: oOptSignal3           
   oOptSignal2->GenerateImpulseSignal();
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#    
   oOptFilter2->FilterOptSignal();

   // change by Jonathan
//	oOptSignal3 = oOptSignal2;   
   *oOptSignal3 = *oOptSignal2;
	   
   int qtPoints = oOptSignal2->GetQtPoints();
   double DeltaFreq = oOptSignal2->GetDeltaFreq();

   /***
   //##  These lines are good for debugging      
   //#### Get the effective optical bandwidth
   double EffectiveOptFilterBandwidth = 0;
   for (int ii = 0; ii < qtPoints; ii++) {
      EffectiveOptFilterBandwidth +=       
          sq(abs(oOptSignal3->sfftM.fOptSignalX[ii]));
   }
   EffectiveOptFilterBandwidth *= oOptSignal3->GetDeltaFreq();            
   cout << "Effective OptFilter bandwidth (GHz) = " 
        << EffectiveOptFilterBandwidth*1e-9 << endl; 	   
   oOptSignal3->WriteFileTimePower("tOutOptSignal2.dat");
   oOptSignal3->WriteFileFreqPower("fOutOptSignal2.dat");  
   ***/
               
   //## Now get the electrical filter response: oPhotodetector2;
   oOptSignal2->GenerateImpulseSignal();

   for (int ii = 0; ii < qtPoints; ii++) {
      oPhotodetector2->sfftPM.tPowerSignal[ii] = 
            oOptSignal2->sfftM.tOptSignalX[ii];
   }   
   oPhotodetector2->Time2Freq();
   
   oElecFilter2->FilterElecSignal();
  
   /***  
   //##  These lines are good for debugging
   //##  Get the effective electrical filter bandwidth
   double EffectiveElecFilterBandwidth = 0;
   for (int ii = 0; ii < qtPoints; ii++) {
      EffectiveElecFilterBandwidth += 
         sq(abs(oPhotodetector2->sfftPM.fPowerSignal[ii]));
   }
   EffectiveElecFilterBandwidth *= oPhotodetector2->GetDeltaFreq();

   oPhotodetector2->WriteFileTimePower("tOutSignal2.dat");
   oPhotodetector2->WriteFileFreqPower("fOutSignal2.dat");  
   cout << "Effective ElecFilter bandwidth (GHz) = " 
        << EffectiveElecFilterBandwidth*1e-9 << endl;
   ***/

   //## The multiplication between eout*(t) and h(t'-\tau) to get g1,G1
   for (int ii = 0; ii < qtPoints; ii++) {
      oOptSignal2->sfftM.tOptSignalX[ii] = 
          conj(oOptSignal1->sfftM.tOptSignalX[ii])
         *oPhotodetector2->sfftPM.tPowerSignal[
	        Periodicity(TargetIndexTime2 - ii,qtPoints)];
      oOptSignal2->sfftM.tOptSignalY[ii] = 
          conj(oOptSignal1->sfftM.tOptSignalY[ii])
         *oPhotodetector2->sfftPM.tPowerSignal[
	        Periodicity(TargetIndexTime2 - ii,qtPoints)];		
   }
   oOptSignal2->Time2Freq();



   //## Multiply G1(\omega) and sq(abs(B(-\omega))) to get i2, I2
   for (int ii = 0; ii < qtPoints; ii++) {
      if (ii == 0 ) {   
         oOptSignal2->sfftM.fOptSignalX[ii] *= 
	    sq(abs(oOptSignal3->sfftM.fOptSignalX[ii]));      
	 //## Note that the impulse response is only in the X polarization
         oOptSignal2->sfftM.fOptSignalY[ii] *= 
	    sq(abs(oOptSignal3->sfftM.fOptSignalX[ii]));
      } else {
         oOptSignal2->sfftM.fOptSignalX[ii] *= 
	    sq(abs(oOptSignal3->sfftM.fOptSignalX[qtPoints - ii]));  
	 //## Note that the impulse response is only in the X polarization
         oOptSignal2->sfftM.fOptSignalY[ii] *= 
	    sq(abs(oOptSignal3->sfftM.fOptSignalX[qtPoints - ii])); 
      }             
   }      
   oOptSignal2->Freq2Time();


   /*** COMMENTED OUT BY JZ on Jan 13th 2006

   This block of code is identical to the one above it.
   An examination of the formula shows that the code should
   only appear once not twice. Looking through the CVS files
   it appears that this bug has been in the code since 
   it was originally added to the CVS archive.

   The effect of the bug was to multiply by |\tilde H_0|^4 instead
   of |\tilde H_0|^2. For an RZ signal with a 60GHz Gaussian optical
   filter and 8GHz electrical Bessel filter removing the bug 
   increases the variance of the current and decreases Q from 7.33
   to 7.16 for the OSNR I tried. Not a big deal. This would make the
   solid Q versus OSNR curve in Fig 1 of Ivan Lima's PTL on the receiver
   lower slightly, but not enough to substantially change the results.

   //## Multiply G1(\omega) and sq(abs(B(-\omega))) to get i2, I2
   for (int ii = 0; ii < qtPoints; ii++) {
      if (ii == 0 ) {   
         oOptSignal2->sfftM.fOptSignalX[ii] *= 
	    sq(abs(oOptSignal3->sfftM.fOptSignalX[ii]));      
	 //## Note that the impulse response is only in the X polarization
         oOptSignal2->sfftM.fOptSignalY[ii] *= 
	    sq(abs(oOptSignal3->sfftM.fOptSignalX[ii]));
      } else {
         oOptSignal2->sfftM.fOptSignalX[ii] *= 
	    sq(abs(oOptSignal3->sfftM.fOptSignalX[qtPoints - ii]));  
	 //## Note that the impulse response is only in the X polarization
         oOptSignal2->sfftM.fOptSignalY[ii] *= 
	    sq(abs(oOptSignal3->sfftM.fOptSignalX[qtPoints - ii])); 
      }             
   }      
   oOptSignal2->Freq2Time();

   ****/

/*
   //## Multiply G1(\omega) and B(-\omega)B*(\omega) to get i2, I2
   for (int ii = 0; ii < qtPoints; ii++) {
      //## It is just this if ii = 0 || ii = qtPoints/2
      cplx cOptSignal3_X_MinusOmega = oOptSignal3->sfftM.fOptSignalX[ii];
      cplx cOptSignal3_Y_MinusOmega = oOptSignal3->sfftM.fOptSignalY[ii];      
      if (ii != 0 && ii != qtPoints/2) {
         cOptSignal3_X_MinusOmega = 
	       oOptSignal3->sfftM.fOptSignalX[qtPoints - ii];
         cOptSignal3_Y_MinusOmega = 
	       oOptSignal3->sfftM.fOptSignalY[qtPoints - ii];      
      }
      oOptSignal2->sfftM.fOptSignalX[ii] = 
            oOptSignal2->sfftM.fOptSignalX[ii]
           *cOptSignal3_X_MinusOmega	    
	   *conj(oOptSignal3->sfftM.fOptSignalX[ii]); 
      oOptSignal2->sfftM.fOptSignalY[ii] = 
            oOptSignal2->sfftM.fOptSignalY[ii]
           *cOptSignal3_Y_MinusOmega
	   *conj(oOptSignal3->sfftM.fOptSignalY[ii]);
   }
   oOptSignal2->Freq2Time();   
*/   
/*   
   //## Multiply G1(\omega) and |B(\omega)|^2 to get i2, I2
   for (int ii = 0; ii < qtPoints; ii++) {
      oOptSignal2->sfftM.fOptSignalX[ii] = 
            oOptSignal2->sfftM.fOptSignalX[ii]
	   *sq(abs(oOptSignal3->sfftM.fOptSignalX[ii])); 
      oOptSignal2->sfftM.fOptSignalY[ii] = 
            oOptSignal2->sfftM.fOptSignalY[ii]
	   *sq(abs(oOptSignal3->sfftM.fOptSignalY[ii]));
   }
   oOptSignal2->Freq2Time();   
*/
   
   //## The multiplication between eout(t) and i2(t) to get g2,G2
   for (int ii = 0; ii < qtPoints; ii++) {
      oOptSignal2->sfftM.tOptSignalX[ii] = 
             oOptSignal2->sfftM.tOptSignalX[ii] 
            *oOptSignal1->sfftM.tOptSignalX[ii];
      oOptSignal2->sfftM.tOptSignalY[ii] = 
             oOptSignal2->sfftM.tOptSignalY[ii] 
            *oOptSignal1->sfftM.tOptSignalY[ii];	
   }
   oOptSignal2->Time2Freq();   

   //## Multiply G2(\omega) and H(\omega) to get i1, I1
   for (int ii = 0; ii < qtPoints; ii++) {
      oOptSignal2->sfftM.fOptSignalX[ii] = 
            oOptSignal2->sfftM.fOptSignalX[ii]
	   *oPhotodetector2->sfftPM.fPowerSignal[ii]; 
      oOptSignal2->sfftM.fOptSignalY[ii] = 
            oOptSignal2->sfftM.fOptSignalY[ii]
	   *oPhotodetector2->sfftPM.fPowerSignal[ii]; 	
   }
   oOptSignal2->Freq2Time();      
   //## This result is only good at TargetIndexTime2, only.

   //## As expected, the Sigma_S-ASE has a negligible imaginary part
   //## Thus, the real part is the Sigma_S-ASE
   //cout << "oOptSignal2->sfftM.tOptSignalX[TargetIndexTime2] = "
   //     << oOptSignal2->sfftM.tOptSignalX[TargetIndexTime2] << endl;           
   
   double normVarianceSignal_ASE = 
        (  NoiseSpectralDensityX
	  *real(oOptSignal2->sfftM.tOptSignalX[TargetIndexTime2])
	 + NoiseSpectralDensityY
	  *real(oOptSignal2->sfftM.tOptSignalY[TargetIndexTime2]) )
	*oPhotodetector2->GetDeltaFreq();

   if ( normVarianceSignal_ASE < 0) {
      //cout << "VarianceSignal_ASE < 0; = " 
      //     << VarianceSignal_ASE << endl;
      normVarianceSignal_ASE = 0;
   }

   double StdDevSignal_ASE = sqrt(2*normVarianceSignal_ASE )
                            *oPhotodetector2->GetConvFactOptElecPhotod();
   
//   double StdDevSignal_ASE = sqrt(2*NoiseSpectralDensity/2.
//                                  *normVarianceSignal_ASE );

   return StdDevSignal_ASE;
}



//###########################################################
//##  Get the Q-factor in a semi analytical form as outlined
//#   by P. Winzer. This module is returning the worse Q-factor
//#   in the pattern. That comes from the lowest noiseless bit 
//#   one and the highest noiseless bit zero. Thus, only two
//#   bits are used in this computation.
//#   WARNING: oOptSignal has to contain a noiseless signal 
//#   prior to the optical filter. It is very important.
//##  This is just an interface to the actual method for the
//#   particular case in which the noise is unpolarized.
//## By: Ivan Lima (2/5/14)
double ElecSignalStat::GetQ_FactorGaussianStatisticsWorstPattern(
                             double NoiseSpectralDensity)
{
   return GetQ_FactorGaussianStatisticsWorstPattern(NoiseSpectralDensity/2.,
                                                NoiseSpectralDensity/2.);
}						



//###########################################################
//##  Get the Q-factor in a semi analytical form as outlined
//#   by P. Winzer. This module is returning the worse Q-factor
//#   in the pattern. That comes from the lowest noiseless bit 
//#   one and the highest noiseless bit zero. Thus, only two
//#   bits are used in this computation.
//#   WARNING: oOptSignal has to contain a noiseless signal 
//#   prior to the optical filter. It is very important.
//## By: Ivan Lima (5/14/02)
double ElecSignalStat::GetQ_FactorGaussianStatisticsWorstPattern(
                             double NoiseSpectralDensityX,
			     double NoiseSpectralDensityY)
{  CheckGaussianStatisticsInitialization(
         "GetQ_FactorGaussianStatisticsWorstPattern");
		         
   //## Start the actual computation
   ///////oOptSignal2->RegenerateSignalString();   
   
   //cout << "Copying oOptSignal "<< endl;
// Changed by Jonathan Hu this has problem
//      oOptSignal2 = oOptSignal;
   *oOptSignal2 = *oOptSignal;

/*   
   //cout << "$#$#$# GetFrequency(0) = " << oOptSignal2->GetFrequency(0) << endl;
   cout << "$#$#$ GetBit (Ch0): 0, 1, 2, 3 -> " 
        << oOptSignal->oBitString->GetBit(0,0) << "  "
        << oOptSignal->oBitString->GetBit(0,1) << "  " 
	<< oOptSignal->oBitString->GetBit(0,2) << endl;
   cout << "$#$#$ GetBit2 (Ch0): 0, 1, 2, 3 -> " 
        << oOptSignal2->oBitString->GetBit(0,0) << "  "
        << oOptSignal2->oBitString->GetBit(0,1) << "  " 
	<< oOptSignal2->oBitString->GetBit(0,2) << endl;
*/
   //cout << "oOptSignal copied. Now start filtering "<< endl;   

/*   //oPhotodetector2->WriteFileTimePower("tOutSignal.dat"); 
   oOptSignal2->WriteFileTimePower("tOutOptSignal2b.dat");       
   oOptSignal2->WriteFileFreqPower("fOutOptSignal2b.dat"); 
   oOptSignal->WriteFileTimePower("tOutOptSignal.dat");       
   oOptSignal->WriteFileFreqPower("fOutOptSignal.dat");    
   //exit(1); //#$#$#$#
*/

   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#  
   oOptFilter2->FilterOptSignal();
/*      for( int ii = 0; ii< qtPoints;ii++)
   		*LogFile << ii <<  "  " <<   abs(fft.tPowerSignal[ii]) << endl;
	exit(1);
*/  
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatistics();


/*   for (int jj = 0; jj < StringLength; jj++) 
     {
	cout << "GetBit" << jj << " = " << GetBit(jj) << " bitIntefgsfdgsnsity " << jj << " = " << BitIntensity[jj] << endl;
	}
	
	exit(1);
*/
/*
   //oPhotodetector2->WriteFileTimePower("tOutSignal.dat"); 
   oOptSignal2->WriteFileTimePower("tOutOptSignal2a.dat");       
   //oOptSignal->WriteFileTimePower("tOutOptSignal.dat");       
   exit(1); //#$#$#$#
*/
   //cout << "The eye opening will be obtained now " << endl;
   
   int WorstBitSlot1;
   double MinCurrentBit1 = oElecSignalStat2->GetMinCurrentBit1(&WorstBitSlot1);
   int WorstBitSlot0; 
   double MaxCurrentBit0 = oElecSignalStat2->GetMaxCurrentBit0(&WorstBitSlot0);

/*
   for (int ii = 0; ii < StringLength; ii++) {        
      cout << "$# " << ii << " LogicalBit = " << GetBit(ii) 
           << " -> " << BitIntensity[ii] << "  ";
      //cout << "Sig_ASE = " <<  GetStdDevSignal_ASE_Bit(ii,0.5,0.5)  
      //     << "  ASE_ASE = " << GetStdDevASE_ASE(1) << endl;
      
      cout << "$#$# Sig_ASE / ASE_ASE = " 
           << GetStdDevSignal_ASE_Bit(ii,0.5,0.5) 
           /GetStdDevASE_ASE(1) 
           << endl;	                
   }	   
*/    

   double AmplitudeMargin = MinCurrentBit1 - MaxCurrentBit0;

/*
   cout << "$#$#$ GetBit_OptSignal2 (Ch0): 0, 1, 2, 3 -> " 
        << oOptSignal2->oBitString->GetBit(0,0) << "  "
        << oOptSignal2->oBitString->GetBit(0,1) << "  " 
	<< oOptSignal2->oBitString->GetBit(0,2) << endl;
*/
//   oOptSignal2->WriteFileTimePower("tOutOptSignal.dat"); 
//   exit(1);

   //cout << "$#$#$ AmplitudeMargin = " << AmplitudeMargin << endl;

   //cout <<"Getting SumStdDevs..." << endl;
   
  double sqStdDevASE_ASE_Tmp = sq(GetStdDevASE_ASE(
				          NoiseSpectralDensityX,
					  NoiseSpectralDensityY ) );

   double SumStdDevs = sqrt( sq(GetStdDevSignal_ASE_BitSlot(WorstBitSlot1,
			                 NoiseSpectralDensityX,
                                         NoiseSpectralDensityY))
   			     +sqStdDevASE_ASE_Tmp )
    		       +sqrt( sq(GetStdDevSignal_ASE_BitSlot(WorstBitSlot0,
			                 NoiseSpectralDensityX,
					 NoiseSpectralDensityY))
 			     +sqStdDevASE_ASE_Tmp );
//	SumStdDevs += 0.0000206033;
/*
   double SumStdDevs = sqrt( sq(GetStdDevSignal_ASE_BitSlot(WorstBitSlot1,
			                 NoiseSpectralDensityX,
                                         NoiseSpectralDensityY))
   				     +sq(GetStdDevASE_ASE(
				          NoiseSpectralDensityX,
				          NoiseSpectralDensityY) ) )
     		        +sqrt( sq(GetStdDevSignal_ASE_BitSlot(WorstBitSlot0,
			                 NoiseSpectralDensityX,
					 NoiseSpectralDensityY))
   				     +sq(GetStdDevASE_ASE(
				          NoiseSpectralDensityX,
					  NoiseSpectralDensityY ) ) );
*/

   //cout <<"I got SumStdDevs..." << endl;
				       
   //## Now get the Q-factor from WorstBit0 and WorstBit1;
   double Q_FactorGaussianStatistics = AmplitudeMargin
	                          /SumStdDevs;
	cout << "( " << MinCurrentBit1 << " - " << MaxCurrentBit0  << ") / " << SumStdDevs << " = " << Q_FactorGaussianStatistics  << endl;
   if (Q_FactorGaussianStatistics < 0 ) {
      Q_FactorGaussianStatistics = 0;
   }

   return Q_FactorGaussianStatistics;
}				 






//###########################################################
//##  Get the Q-factor in a semi analytical form as outlined
//#   by P. Winzer. This module is returning the  Q-factor
//#   using the first mark and the first space that appear
//#   in the pattern.  Thus, only two
//#   bits are used in this computation.
//#   WARNING: oOptSignal has to contain a noiseless signal 
//#   prior to the optical filter. It is very important.
//## By: Ivan Lima (2/4/03)
double ElecSignalStat::GetQ_FactorGaussianStatisticsFirstPattern(
                             double NoiseSpectralDensityX,
			     double NoiseSpectralDensityY)
{
   CheckGaussianStatisticsInitialization(
         "GetQ_FactorGaussianStatisticsFirstPattern");
		         
   //## Start the actual computation
   ///////oOptSignal2->RegenerateSignalString();     
      
   *oOptSignal2 = *oOptSignal;

   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#  
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatistics();
   
   int FirstBitSlot1;
   double CurrentFirstBit1 = oElecSignalStat2->GetCurrentFirstBit1(&FirstBitSlot1);
   int FirstBitSlot0; 
   double CurrentFirstBit0 = oElecSignalStat2->GetCurrentFirstBit0(&FirstBitSlot0);

   double AmplitudeMargin = CurrentFirstBit1 - CurrentFirstBit0;
   
   double sqStdDevASE_ASE_Tmp = sq(GetStdDevASE_ASE(
				          NoiseSpectralDensityX,
					  NoiseSpectralDensityY ) );

   double SumStdDevs = sqrt( sq(GetStdDevSignal_ASE_BitSlot(FirstBitSlot1,
			                 NoiseSpectralDensityX,
                                         NoiseSpectralDensityY))
   			     +sqStdDevASE_ASE_Tmp )
    		       +sqrt( sq(GetStdDevSignal_ASE_BitSlot(FirstBitSlot0,
			                 NoiseSpectralDensityX,
					 NoiseSpectralDensityY))
 			     +sqStdDevASE_ASE_Tmp );
				       
   //## Now get the Q-factor from FirstBit0 and FirstBit1;
   double Q_FactorGaussianStatistics = AmplitudeMargin
	                          /SumStdDevs;
	cout << "( " << CurrentFirstBit1 << " - " <<  CurrentFirstBit0   << ") / " << SumStdDevs << " = " << Q_FactorGaussianStatistics  << endl;

   if (Q_FactorGaussianStatistics < 0 ) {
      Q_FactorGaussianStatistics = 0;
   }

   return Q_FactorGaussianStatistics;
}	

//###########################################################
//##  This is just an interface for the case where the noise
//##  is completely unpolarized.
//## By: Ivan Lima (5/25/02)
double ElecSignalStat::GetStdDevWorstBit1_GaussianStatistics(
                             double NoiseSpectralDensity)
{
   return GetStdDevWorstBit1_GaussianStatistics(
                             NoiseSpectralDensity/2.,
			     NoiseSpectralDensity/2.);
}



//###########################################################
//##  Get the standard deviation of the worse bit one
//##  using the semi-analytical method.
//## By: Ivan Lima (5/25/02)
double ElecSignalStat::GetStdDevWorstBit1_GaussianStatistics(
                             double NoiseSpectralDensityX,
			     double NoiseSpectralDensityY)
{
   CheckGaussianStatisticsInitialization(
         "GetStdDevWorstBit0_GaussianStatistics");
		         
   //## Start the actual computation
   ///////oOptSignal2->RegenerateSignalString();   
   *oOptSignal2 = *oOptSignal;
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#     
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatistics();

   int WorstBitSlot1;
   double MinCurrentBit1 = oElecSignalStat2->GetMinCurrentBit1(&WorstBitSlot1);
   //int WorstBitSlot0; 
   //double MaxCurrentBit0 = oElecSignalStat2->GetMaxCurrentBit0(&WorstBitSlot0);
   
   //## Now get StdDev of the WorstBit1;
   double StdDevWorstBit1 = sqrt( sq(GetStdDevSignal_ASE_BitSlot(WorstBitSlot1,
			                 NoiseSpectralDensityX,
                                         NoiseSpectralDensityY))
   				     +sq(GetStdDevASE_ASE(
				          NoiseSpectralDensityX,
				          NoiseSpectralDensityY) ) );
   
   
   return StdDevWorstBit1;
}	



//###########################################################
//##  This is just an interface for the case where the noise
//##  is completely unpolarized.
//## By: Ivan Lima (5/25/02)
double ElecSignalStat::GetStdDevWorstBit0_GaussianStatistics(
                             double NoiseSpectralDensity)
{
   return GetStdDevWorstBit0_GaussianStatistics(
                             NoiseSpectralDensity/2.,
			     NoiseSpectralDensity/2.);
}



//###########################################################
//##  Get the standard deviation of the worse bit one
//##  using the semi-analytical method.
//## By: Ivan Lima (5/25/02)
double ElecSignalStat::GetStdDevWorstBit0_GaussianStatistics(
                             double NoiseSpectralDensityX,
			     double NoiseSpectralDensityY)
{
   CheckGaussianStatisticsInitialization(
         "GetStdDevWorstBit0_GaussianStatistics");
		         
   //## Start the actual computation
   ///////oOptSignal2->RegenerateSignalString();   
   *oOptSignal2 = *oOptSignal;
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#     
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatistics();

   //int WorstBitSlot1;
   //double MinCurrentBit1 = oElecSignalStat2->GetMinCurrentBit1(&WorstBitSlot1);
   int WorstBitSlot0; 
   double MaxCurrentBit0 = oElecSignalStat2->GetMaxCurrentBit0(&WorstBitSlot0);
   
   //## Now get StdDev of the WorstBit1;
   double StdDevWorstBit0 = sqrt( sq(GetStdDevSignal_ASE_BitSlot(WorstBitSlot0,
			                 NoiseSpectralDensityX,
                                         NoiseSpectralDensityY))
   				     +sq(GetStdDevASE_ASE(
				          NoiseSpectralDensityX,
				          NoiseSpectralDensityY) ) );
   
   
   return StdDevWorstBit0;
}	



//########################################################
//##### Get the ESNR of the worse mark using the semi-analytical approach
//##    You should use this one if you want to get the Q-factor
//##    from the ESNR.
//## By: Ivan Lima
double ElecSignalStat::GetESNR_Mark_GaussianStatistics(double NoiseSpectralDensity)
{
   CheckGaussianStatisticsInitialization(
         "GetESNR_GaussianStatistics");

   *oOptSignal2 = *oOptSignal;
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#     
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatistics();

   double MinCurrentBit1 = oElecSignalStat2->GetMinCurrentBit1();
   //double MaxCurrentBit0 = oElecSignalStat2->GetMaxCurrentBit0();
/*
   int WorstBitSlot1;
   double MinCurrentBit1 = oElecSignalStat2->GetMinCurrentBit1(&WorstBitSlot1);
   int WorstBitSlot0; 
   double MaxCurrentBit0 = oElecSignalStat2->GetMaxCurrentBit0(&WorstBitSlot0);
*/

   double ESNR_Mark = MinCurrentBit1 
                /GetMeanASE_ASE(NoiseSpectralDensity);

   //double ESNR = (MinCurrentBit1 - MaxCurrentBit0)
   //             /GetMeanASE_ASE(NoiseSpectralDensity);

   if (ESNR_Mark < 0 ) {
      //## This can only happend if either the filters are too narow or
      //#  if there is something that's really wrong
      ESNR_Mark = 0;
   }   
   return ESNR_Mark;
}



//########################################################
//##### Get the ESNR using the semi-analytical approach
//##    You should NOT use this one if you want to get the Q-factor
//##    from the ESNR.
//## By: Ivan Lima
double ElecSignalStat::GetESNR_GaussianStatistics(double NoiseSpectralDensity)
{
   CheckGaussianStatisticsInitialization(
         "GetESNR_GaussianStatistics");

   *oOptSignal2 = *oOptSignal;
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#     
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatistics();

   double MinCurrentBit1 = oElecSignalStat2->GetMinCurrentBit1();
   double MaxCurrentBit0 = oElecSignalStat2->GetMaxCurrentBit0();
/*
   int WorstBitSlot1;
   double MinCurrentBit1 = oElecSignalStat2->GetMinCurrentBit1(&WorstBitSlot1);
   int WorstBitSlot0; 
   double MaxCurrentBit0 = oElecSignalStat2->GetMaxCurrentBit0(&WorstBitSlot0);
*/

   double ESNR = (MinCurrentBit1 - MaxCurrentBit0)
                /GetMeanASE_ASE(NoiseSpectralDensity);

   if (ESNR < 0 ) {
      //## This can only happend if either the filters are too narow or
      //#  if there is something that's really wrong
      ESNR = 0;
   }   
   return ESNR;
}



/*
//########################################################
//##### Get the Q-factor based on Marcuse's 
//##    Using the semi-analytical approach
//##    It assumes that the noise is unpolarized.
//##    You can use the enhancement factor to get the Q-factor
//##    as a function of the OSNR.
//## By: Ivan Lima
double ElecSignalStat::GetQ_FactorFromESNR_GaussianStatistics( double ESNR)
{
   CheckGaussianStatisticsInitialization(
         "GetQ_FactorFromESNR_GaussianStatistics");

   //## Get MinCurrentBit1 and its the BitSlot
   *oOptSignal2 = *oOptSignal;
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#     
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatisticsNoTimeShift();

   int WorstBitSlot1;
   double MinCurrentBit1 = oElecSignalStat2->GetMinCurrentBit1(&WorstBitSlot1);
   int WorstBitSlot0;
   double MaxCurrentBit0 = oElecSignalStat2->GetMaxCurrentBit0(&WorstBitSlot0);

   double ExtinctionRatio = MaxCurrentBit0/MinCurrentBit1;
   double NormNumberNoiseModes = sq( GetMeanASE_ASE(0.5,0.5)
                                /GetStdDevASE_ASE(0.5,0.5) ); 
 
   double NormKcoeff1 = sqrt(NormNumberNoiseModes)
                  *sq(GetStdDevSignal_ASE_BitSlot(WorstBitSlot1,0.5,0.5) )
		  /( MinCurrentBit1*GetStdDevASE_ASE(0.5,0.5) );   
 
   double NormKcoeff0 = sqrt(NormNumberNoiseModes)
                  *sq(GetStdDevSignal_ASE_BitSlot(WorstBitSlot0,0.5,0.5) )
		  /( MaxCurrentBit0*GetStdDevASE_ASE(0.5,0.5) );   
      
	                   
      double Q_factorFromESNR =  (1. - ExtinctionRatio)*ESNR
                                *sqrt(NormNumberNoiseModes)
                             /( sqrt(1. + NormKcoeff1*ESNR) 
                               +sqrt(1. + NormKcoeff0*ExtinctionRatio*ESNR)); 
		                	
      return Q_factorFromESNR;	   
}
*/



//########################################################
//##### Get the Q-factor based on the generalization of Marcuse's 
//##    approach by Aurenice Lima and Ivan Lima using the 
//##    semi-analytical approach.
//##    It assumes that the noise is unpolarized.
//##    You can use the enhancement factor to get the Q-factor
//##    as a function of the OSNR, or directly from GetESNR_GaussianStatistics().
//##    This is just an interface.
//## By: Ivan Lima
double ElecSignalStat::GetQ_FactorFromESNR_GaussianStatistics( double ESNR)
{
   return GetQ_FactorFromESNR_GaussianStatistics( ESNR, 1., 1.); 
}



//########################################################
//##### Get the Q-factor based on the generalization of Marcuse's 
//##    approach by Aurenice Lima and Ivan Lima using the 
//##    semi-analytical approach.
//##    You can use the enhancement factor to get the Q-factor
//##    as a function of the OSNR, or directly from GetESNR_GaussianStatistics().
//##    This function takes into account polarized noise, but
//##    YOU MUST TRANSFORM THE POLARIZED PART OF THE NOISE TO THE
//##    X POLARIZATION. BE CAREFUL TO APPLY THE SAME TRANSFORMATION
//##    TO THE SIGNAL TO GET THE CORRECT SIGNAL TO NOISE BEATING.
//##    The values of the noise spectral density are only used to get
//##    the ratio between the X and the Y component. The absolute value
//##    does not matter, since it will be extracted from the ESNR.
//## By: Ivan Lima
double ElecSignalStat::GetQ_FactorFromESNR_GaussianStatistics( double ESNR_Mark,
                       double NoiseSpectralDensityX,
	   	       double NoiseSpectralDensityY)
{
   CheckGaussianStatisticsInitialization(
         "GetQ_FactorFromESNR_GaussianStatistics");

   //## Get MinCurrentBit1 and its the BitSlot
   *oOptSignal2 = *oOptSignal;
   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#     
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatistics();

   int WorstBitSlot1;
   double MinCurrentBit1 = oElecSignalStat2->GetMinCurrentBit1(&WorstBitSlot1);
   int WorstBitSlot0;
   double MaxCurrentBit0 = oElecSignalStat2->GetMaxCurrentBit0(&WorstBitSlot0);

   double ExtinctionRatio = MaxCurrentBit0/MinCurrentBit1;
   double NormNumberNoiseModes = sq( GetMeanASE_ASE(0.5,0.5)
                                    /GetStdDevASE_ASE(0.5,0.5) ); 
 
   double NormKcoeff1 = sqrt(NormNumberNoiseModes)
                  *sq(GetStdDevSignal_ASE_BitSlot(WorstBitSlot1,0.5,0.5) )
		  /( MinCurrentBit1*GetStdDevASE_ASE(0.5,0.5) );   
 
   double NormKcoeff0 = sqrt(NormNumberNoiseModes)
                  *sq(GetStdDevSignal_ASE_BitSlot(WorstBitSlot0,0.5,0.5) )
		  /( MaxCurrentBit0*GetStdDevASE_ASE(0.5,0.5) );   
   /******* ## For debugging
         cout << "#$#$ Begin" << endl;
         //cout << "EnhancementFactor = " << EnhancementFactor << endl;
	 //cout << "NormalizedEnhancementFactor = " 
	 //     << NormalizedEnhancementFactor << endl;
         //cout << "PowerEquivalentSpectralWidthOptFilter (GHz)= "
	 //     << PowerEquivalentSpectralWidthOptFilter*1e-9 << endl;	      
         cout << "ExtinctionRatio = " << ExtinctionRatio << endl;
         cout << "NormNumberNoiseModes = " << NormNumberNoiseModes << endl;
         cout << "NormKcoeff1 = " << NormKcoeff1 << endl;
         cout << "NormKcoeff0 = " << NormKcoeff0 << endl;     
         cout << "#$#$ End" << endl;
   *******/

   //## oOptSignal2 was used as temporary space for the internal
   //## computation
   
   double StokesSignal[4];   
   oOptSignal->GetTotalStokesParameters(StokesSignal);

   double TransPlus  = (1 + StokesSignal[1])/2.;
   double TransMinus = (1 - StokesSignal[1])/2.;   

   double gammaPlus = NoiseSpectralDensityX/sqrt(2.)
                     /sqrt( sq(NoiseSpectralDensityX)
                     +sq(NoiseSpectralDensityY) );
   double gammaMinus = NoiseSpectralDensityY/sqrt(2.)
                      /sqrt( sq(NoiseSpectralDensityX)
	              +sq(NoiseSpectralDensityY) );
   //cout <<"StokesSignal[1] = " << StokesSignal[1] << endl;
   //cout <<"StokesSignal[2] = " << StokesSignal[2] << endl;   
   //cout <<"StokesSignal[3] = " << StokesSignal[3] << endl;   
   //cout <<"TransPlus = "<<TransPlus<<"   TransMinus = "<<TransMinus<<endl;
   //cout <<"gammaPlus = "<<gammaPlus<<"   gammaMinus = "<<gammaMinus<<endl;

   double Delta = 2*(TransPlus*gammaPlus + TransMinus*gammaMinus);
		      
				            
   double Q_FactorFromESNR =  (1. - ExtinctionRatio)*ESNR_Mark
                        *sqrt(NormNumberNoiseModes)
                        *(gammaPlus + gammaMinus)
            /( sqrt( 1. + NormKcoeff1
	            *Delta
		    *(gammaPlus + gammaMinus) * ESNR_Mark ) 
              +sqrt( 1. + NormKcoeff0
     	            *Delta
		    *(gammaPlus + gammaMinus)* ExtinctionRatio * ESNR_Mark ) );
		     
      /* 
      ..## The old thing    	                   
      double Q_FactorFromESNR =  (1. - ExtinctionRatio)*ESNR
                                *sqrt(NormNumberNoiseModes)
                             /( sqrt(1. + NormKcoeff1*ESNR) 
                               +sqrt(1. + NormKcoeff0*ExtinctionRatio*ESNR)); 
       */			       
		                	

      return Q_FactorFromESNR;	   

}

//###################################################################
//#### Get the minimum current of a noiseless mark, which MUST be 
//#    obtained from a noiseless simulation. This is only an interface
//## By: Ivan Lima  (5/17/02)
double ElecSignalStat::GetMinCurrentBit1(void)
{
   int BitSlotTmp; 
   return GetMinCurrentBit1(&BitSlotTmp);
}


					
//###################################################################
//#### Get the minimum current of a noiseless mark, which MUST be 
//#    obtained from a noiseless simulation.
//## By: Ivan Lima  (5/17/02)
double ElecSignalStat::GetMinCurrentBit1(int *BitSlotMinCurrentBit1)
{
   //#  I need to recover the clock for that
   ComputePerformanceStatistics();  
   double MinCurrentBit1 = 0; 
   bool GotFirstBit1_Flag = 0;
   for (int ii = 0; ii < StringLength; ii++) {
      if (   GetBit(ii) == 1) {
         if (   BitIntensity[ii] < MinCurrentBit1
	     || ! GotFirstBit1_Flag) {
	  GotFirstBit1_Flag = 1;   
	  MinCurrentBit1 = BitIntensity[ii];
	  *BitSlotMinCurrentBit1 = ii;
	 } 
     }    

   //   cout << "$# " << ii << " LogicalBit = " << GetBit(ii) 
  //         << " -> " << BitIntensity[ii] << "  ";
      /*	   
      cout << "$#$# sq(Sig_ASE) / ASE_ASE = " 
           << sq(GetStdDevSignal_ASE(ii,0.5,0.5) ) 
   	     /GetStdDevASE_ASE(1) 
           << endl;	                
      */	   
   } 
 //  cout << endl << "$# " << *BitSlotMinCurrentBit1 << " -> " << MinCurrentBit1 << endl;
     /***	  	  
      //$$$$$$ These lines should be commented
      //$$ Awesome! It is now syncronized
      int TargetIndexTime = GetTargetIndexTime();
      cout << ii<< " BitIntensity = " << BitIntensity[ii] 
           << "  LogicalBit = " << oOptSignal->oBitString->GetBit(0,ii) << endl;
      cout << "      " 
           << real(sfftPM.tPowerSignal[Periodicity(TargetIndexTime
	           +(Periodicity(ii,StringLength) + BitStringShift)
                   *qtPoints/StringLength,
                 qtPoints)]) << endl; 	   
      cout << "   BitStringShift  = " << BitStringShift
           << "  TargetIndexTime = " << TargetIndexTime << endl;
      ***/
   
   return(MinCurrentBit1);
}  

					
//###################################################################
//#### Get the current of the first noiseless mark, which MUST be 
//#    obtained from a noiseless simulation.
//## By: Ivan Lima  (5/17/02)
double ElecSignalStat::GetCurrentFirstBit1(int *BitSlotCurrentFirstBit1)
{
   //#  I need to recover the clock for that
   ComputePerformanceStatistics();  
   double CurrentFirstBit1 = 0; 
   for (int ii = 0; ii < StringLength; ii++) {       
      if (   GetBit(ii) == 1) {
	  CurrentFirstBit1 = BitIntensity[ii];
	  *BitSlotCurrentFirstBit1 = ii;
      }    
   }    
   return(CurrentFirstBit1);
}  



//###################################################################
//#### Get the current of the selected noiseless mark
//#    obtained from a noiseless simulation.
//## The clock must have been previously recovered
//## By: Ivan Lima  (4/13/03)
double ElecSignalStat::GetCurrentBitFromRecoveredSignal(int BitSlotCurrentBit)
{
   ////////ComputePerformanceStatistics();  
   double CurrentBit = BitIntensity[BitSlotCurrentBit];
   return(CurrentBit);
}  



//###################################################################
//#### Get the current of the selected noiseless mark
//#    obtained from a noiseless simulation.
//## By: Ivan Lima  (2/27/03)
double ElecSignalStat::GetCurrentBit(int BitSlotCurrentBit)
{
   //#  I need to recover the clock for that
   ComputePerformanceStatistics();  
   double CurrentBit = BitIntensity[BitSlotCurrentBit];
   return(CurrentBit);
}  


//###################################################################
//#### Get the maximum current of a noiseless space, which MUST be 
//#    obtained from a noiseless simulation. This is just an interface.
//## By: Ivan Lima (5/17/02)
double ElecSignalStat::GetMaxCurrentBit0(void)
{
   int BitSlotTmp; 
   return GetMaxCurrentBit0(&BitSlotTmp);
}



//###################################################################
//#### Get the maximum current of a noiseless space, which MUST be 
//#    obtained from a noiseless simulation.
//## By: Ivan Lima (5/17/02)
double ElecSignalStat::GetMaxCurrentBit0(int *BitSlotMaxCurrentBit0)
{
   //#  I need to recover the clock for that
   ComputePerformanceStatistics();  
   double MaxCurrentBit0 = 0; 
   bool GotFirstBit0_Flag = 0;
   for (int ii = 0; ii < StringLength; ii++) {       
      if (   GetBit(ii) == 0) {
         if (   BitIntensity[ii] > MaxCurrentBit0
	     || ! GotFirstBit0_Flag) {
	  GotFirstBit0_Flag = 1;   
	  MaxCurrentBit0 = BitIntensity[ii];
	  *BitSlotMaxCurrentBit0 = ii;
	 } 	
      }       
   } 
   

      /***	  	  
      //$$$$$$ These lines should be commented
      //$$ Awesome! It is now syncronized
      int TargetIndexTime = GetTargetIndexTime();
      cout << ii<< " BitIntensity = " << BitIntensity[ii] 
           << "  LogicalBit = " << oOptSignal->oBitString->GetBit(0,ii) << endl;
      cout << "      " 
           << real(sfftPM.tPowerSignal[Periodicity(TargetIndexTime
	           +(Periodicity(ii,StringLength) + BitStringShift)
                   *qtPoints/StringLength,
                 qtPoints)]) << endl; 	   
      cout << "   BitStringShift  = " << BitStringShift
           << "  TargetIndexTime = " << TargetIndexTime << endl;
      ***/
   
   return(MaxCurrentBit0);
}  




//###################################################################
//#### Get the current of the first noiseless mark, which MUST be 
//#    obtained from a noiseless simulation.
//## By: Ivan Lima  (5/17/02)
double ElecSignalStat::GetCurrentFirstBit0(int *BitSlotCurrentFirstBit0)
{
   //#  I need to recover the clock for that
   ComputePerformanceStatistics();  
   double CurrentFirstBit0 = 0; 
   for (int ii = 0; ii < StringLength; ii++) {       
      if (   GetBit(ii) == 0) {
	  CurrentFirstBit0 = BitIntensity[ii];
	  *BitSlotCurrentFirstBit0 = ii;
      }    
   }    
   return(CurrentFirstBit0);
}  				
					
					
//###################################################################
//#### Get the Max current of the signal, which comes from 
//##   the isolated marks in the bit string. It gets the average
//##   value of the isolated marks, which should be obtained from
//#    a noiseless simulation.
//## By: Ivan Lima
double ElecSignalStat::GetMeanCurrentIsolatedBit1(void)
{
   
   
   //## I need get the average value of the current of
   //#  the isolated ones at the clock recovery time
   //#  I need to recover the clock for that
   ComputePerformanceStatistics();  
   double MeanCurrent = 0; 
   int qtIsolatedBit1 = 0;
   int iiPeriodicPrevious;
   int iiPeriodicPosterior;
   for (int ii = 0; ii < StringLength; ii++) {   
      iiPeriodicPrevious = Periodicity(ii-1,StringLength);
      iiPeriodicPosterior = Periodicity(ii+1,StringLength);      
      if (   oOptSignal->oBitString->GetBit(0,ii) == 1
          && oOptSignal->oBitString->GetBit(0,iiPeriodicPrevious) == 0
          && oOptSignal->oBitString->GetBit(0,iiPeriodicPosterior) == 0 ) {
	  qtIsolatedBit1++;
	  MeanCurrent += BitIntensity[ii];
	  //cout << "* ";
      } 
      
      /***	  	  
      //$$$$$$ These lines should be commented
      //$$ Awesome! It is now syncronized
      int TargetIndexTime = GetTargetIndexTime();
      cout << ii<< " BitIntensity = " << BitIntensity[ii] 
           << "  LogicalBit = " << oOptSignal->oBitString->GetBit(0,ii) << endl;
      cout << "      " 
           << real(sfftPM.tPowerSignal[Periodicity(TargetIndexTime
	           +(Periodicity(ii,StringLength) + BitStringShift)
                   *qtPoints/StringLength,
                 qtPoints)]) << endl; 	   
      cout << "   BitStringShift  = " << BitStringShift
           << "  TargetIndexTime = " << TargetIndexTime << endl;
      ***/
   } 
   
   if (qtIsolatedBit1 == 0) {
      cout << "Error from ElecSignalStat::GetMeanCurrentIsolatedBit1()"<<endl
           << "There must be at least one isolated bit 1 in the bit string."
	   << endl << "Please, fix it in your OptSignal.in." << endl;
      exit(1);
   }
   
   MeanCurrent /= qtIsolatedBit1;
   
//   for (int ii = 0; ii < qtPoints; ii++) { 
//      if ( MeanCurrent < real(oPhotodetector->sfftPM.tPowerSignal[ii]) )  {
//         MeanCurrent =  real(oPhotodetector->sfftPM.tPowerSignal[ii]);            
//       }	
//   }
   return(MeanCurrent);
}  



//###################################################################
//#### Get the min current of the signal, which comes from 
//##   the isolated spaces in the bit string. It gets the average
//##   value of the isolated spaces, which should be obtained from
//#    a noiseless simulation.
//## By: Ivan Lima
double ElecSignalStat::GetMeanCurrentIsolatedBit0(void)
{
   //## I need get the average value of the current of
   //#  the isolated ones at the clock recovery time
   //#  I need to recover the clock for that
   ComputePerformanceStatistics();  
   double MeanCurrent = 0; 
   int qtIsolatedBit0 = 0;
   int iiPeriodicPrevious;
   int iiPeriodicPosterior;
   for (int ii = 0; ii < StringLength; ii++) {   
      iiPeriodicPrevious = Periodicity(ii-1,StringLength);
      iiPeriodicPosterior = Periodicity(ii+1,StringLength);      
      if (   oOptSignal->oBitString->GetBit(0,ii) == 0
          && oOptSignal->oBitString->GetBit(0,iiPeriodicPrevious) == 1
          && oOptSignal->oBitString->GetBit(0,iiPeriodicPosterior) == 1 ) {
	  qtIsolatedBit0++;
	  MeanCurrent += BitIntensity[ii];
	  //cout << "* ";
      }	  	  
      //cout << ii<< " BitIntensity = " << BitIntensity[ii] 
      //     << "  LogicalBit = " << oOptSignal->oBitString->GetBit(0,ii) << endl;
   } 
   
   if (qtIsolatedBit0 == 0) {
      cout << "Error from ElecSignalStat::GetMeanCurrentIsolatedBit1()"<<endl
           << "There must be at least one isolated bit 0 in the bit string."
	   << endl << "Please, fix it in your OptSignal.in." << endl;
      exit(1);
   }
   
   MeanCurrent /= qtIsolatedBit0;
   
//   for (int ii = 0; ii < qtPoints; ii++) { 
//      if ( MeanCurrent < real(oPhotodetector->sfftPM.tPowerSignal[ii]) )  {
//         MeanCurrent =  real(oPhotodetector->sfftPM.tPowerSignal[ii]);            
//       }	
//   }
   return(MeanCurrent);
}  

// ###### METHODS FOR COMPUTING PATTERN STATS #####################

void ElecSignalStat::InitPatternStats(void)
{

  NumPatterns = int(pow(2.,PatternLength));

  LogFileSeparator();
  *LogFile << "ElecSignalStat::InitPatternStats" << endl << endl;
  *LogFile << "PatternLength = " << PatternLength << endl;
  *LogFile << "NumPatterns = " << NumPatterns << endl;

  AllocateMemoryForPatternStats();
  ClearPatternStats();
  EnumeratePatterns(LogFile); 

} // ## end InitPatternStats

// #############################################################

void ElecSignalStat::AllocateMemoryForPatternStats(void)
{

PatternArray = new bool[PatternLength];
NumOccurencesPattern = new int[NumPatterns];

MeanPattern = new double[NumPatterns];
StdDevPattern = new double[NumPatterns];
VoltageSumPattern = new double[NumPatterns];
VoltageSumSquaresPattern = new double[NumPatterns];
MaximumVoltagePattern = new double[NumPatterns];
MinimumVoltagePattern = new double[NumPatterns];

 if(!PatternArray || !NumOccurencesPattern || !MeanPattern ||
    !StdDevPattern || !VoltageSumPattern || !VoltageSumSquaresPattern ||
    !MaximumVoltagePattern || !MinimumVoltagePattern)
   {
     cerr << "ElecSignalStat::AllocateMemoryForPatternStats, Malloc error"
	  << endl << "ABORTING NOW!" << endl;
    
     exit(1);
   }

}

// ####################################################

void ElecSignalStat::ReleaseMemoryForPatternStats(void)
{

  delete [] PatternArray;
  delete [] NumOccurencesPattern;
  delete [] MeanPattern;
  delete [] StdDevPattern;
  delete [] VoltageSumPattern;
  delete [] VoltageSumSquaresPattern;
  delete [] MaximumVoltagePattern;
  delete [] MinimumVoltagePattern;

}

// ####################################################

void ElecSignalStat::ClearPatternStats(void)
{

  for(int ii=0;ii<NumPatterns;ii++)
      PatternArray[ii] = 0;

  for(int ii=0;ii<NumPatterns;ii++)
    NumOccurencesPattern[ii] = 0;

  for(int ii=0;ii<NumPatterns;ii++)
    MeanPattern[ii] = 0;

  for(int ii=0;ii<NumPatterns;ii++)
    StdDevPattern[ii] = 0;

  for(int ii=0;ii<NumPatterns;ii++)
    VoltageSumPattern[ii] = 0;

  for(int ii=0;ii<NumPatterns;ii++)
    VoltageSumSquaresPattern[ii] = 0;

  for(int ii=0;ii<NumPatterns;ii++)
    MaximumVoltagePattern[ii] = -1e100;

  for(int ii=0;ii<NumPatterns;ii++)
    MinimumVoltagePattern[ii] = 1e100;

} // ## end ClearPatternStats

// #####################################################


// Added by Hua on Jan. 24, 2006
void ElecSignalStat::ClearGaussianPDFsMonteCarlo(void)
{

  for(int ii=0;ii<StringLength;ii++)
    { 
          qtBit[ii] = 0;
	  PowerSumBit[ii] = 0;
	  PowerSquareSumBit[ii] = 0;
	 } 

} // ## end ClearGaussianPDFsMonteCarlo

// #######################################################

//! Writes all bit patterns of length PatternLength to a file
//! together with their PatternIndicies which are the decimal
//! integers corresponding to each of the bit patterns


void ElecSignalStat::EnumeratePatterns(ofstream * OutFile)
{

  *OutFile << "%% (0) PatternIndex (1-PatternLength) The Pattern " 
	   << "(Last) Check of PatternIndex" 
	  << endl << endl;

  for(int PatternIndex = 0; PatternIndex < NumPatterns; PatternIndex++)
    {
      *OutFile << PatternIndex << "\t"; 

      for(int BitIndex = PatternLength-1; BitIndex >= 0; BitIndex--)
	{
          PatternArray[BitIndex] = GetPattern(PatternIndex,BitIndex);
          *OutFile << PatternArray[BitIndex] << "\t";
	}

      *OutFile << GetPatternIndex(PatternArray,PatternLength) << endl;   
      
    } // ## end for loop

} // ## end EnumeratePatterns

// ################################################

//! Converts the integer PatternIndex into binary format
//! returning coefficient of pow(2,BitIndex) in this binary expansion

bool ElecSignalStat::GetPattern(int PatternIndex,int BitIndex)
{

  int N = PatternIndex;

  for(int ii= PatternLength-1; ii > BitIndex;ii--)
    N = N - GetPattern(PatternIndex,ii)*int(pow(2.,ii));

  if(N >= pow(2.,BitIndex))
     return 1;
  else
     return 0;

} // ## end GetPattern

// #######################################################
  
bool ElecSignalStat::GetCenterBitPattern(int PatternIndex)
{

  return GetPattern(PatternIndex,(PatternLength-1)/2);

}

// #######################################################

int ElecSignalStat::GetPatternIndex(bool * LocalPatternArray, 
                                    int LengthOfPattern) 
{

  int PatternIndex = 0;

  for(int ii=0;ii<LengthOfPattern;ii++)
    {
      PatternIndex += LocalPatternArray[ii]*int(pow(2.,ii));
    }

  return PatternIndex;

} // ## end GetPatternIndex

// ####################################################

void ElecSignalStat::UpdatePatternStats(int BitIndex,double PowerElecSignal)
{

  // First we find the PatternIndex of the 
  // sub-bitstring whose length is PatternLength and whose center bit 
  // has index BitIndex

  int MinBitIndex = BitIndex - (PatternLength - 1)/2;
  int MaxBitIndex = BitIndex + (PatternLength - 1)/2;

  // Remember for the elec signal stats we only have one channel

  for(int ii=0;ii<PatternLength;ii++)
    {
      PatternArray[ii] = oOptSignal->oBitString->GetBit(0,
			 oOptSignal->WrapPeriodically(MinBitIndex+ii,
                                                      0,StringLength));
    }

  int PatternIndex = GetPatternIndex(PatternArray,PatternLength);

  
  NumOccurencesPattern[PatternIndex] += 1;

  VoltageSumPattern[PatternIndex] += PowerElecSignal;
  VoltageSumSquaresPattern[PatternIndex] += sq(PowerElecSignal);

  if(PowerElecSignal > MaximumVoltagePattern[PatternIndex])
    MaximumVoltagePattern[PatternIndex] = PowerElecSignal;

  if(PowerElecSignal < MinimumVoltagePattern[PatternIndex])
    MinimumVoltagePattern[PatternIndex] = PowerElecSignal;
  
  /***********  
  cout << PatternIndex << " " 
       << PatternArray[0] << " " 
       << PatternArray[1] << " "
       << PatternArray[2] << " "
       << NumOccurencesPattern[PatternIndex] << " "
       << VoltageSumPattern[PatternIndex] << " "
       << VoltageSumSquaresPattern[PatternIndex] << " "
       << MaximumVoltagePattern[PatternIndex] << " "
       << MinimumVoltagePattern[PatternIndex] << endl;
  **************/

} // ## end UpdatePatternStats

// ####################################################

bool ElecSignalStat::PatternEquality(bool * PatternArray1, 
                                     bool * PatternArray2, 
				     int LengthOfPattern)
{

  for(int ii=0;ii<LengthOfPattern;ii++)
    if(PatternArray1[ii] != PatternArray2[ii])
      return 0;

  return 1;

} // ## end PatternEquality

// #####################################################

void ElecSignalStat::PreparePatternStatsForOutput(void)
{

  for(int ii=0;ii<NumPatterns;ii++)
    {

      if(NumOccurencesPattern[ii]==0)
	{
          cerr << "ElecSignalStat::PreparePatternStatsForOutput" << endl
	       << "The pattern with index " << ii
	       << " has not occurred:" << endl 
	       << "ABORTING!" << endl;

          exit(1);

	}

      MeanPattern[ii] = VoltageSumPattern[ii]/double(NumOccurencesPattern[ii]);
      
      if(NumOccurencesPattern[ii] > 1)
       StdDevPattern[ii] = sqrt( (VoltageSumSquaresPattern[ii] - 
			         NumOccurencesPattern[ii]*sq(MeanPattern[ii]))
				/(NumOccurencesPattern[ii]-1.0));

      else
      	StdDevPattern[ii] = 0;

    } // ## end for

} // ## end PreparePatternStatsForOutput

// ###############################################################

void ElecSignalStat::WriteFilePatternStats(string OutFileName)
{

  PreparePatternStatsForOutput();

  ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "ElecSignalStat::WriteFilePatternStats" 
	   << endl
	   << "Error: " << OutFileName << "could not be created." << endl;
      exit(1);
	}      


 OutFile << "%% (1) PatternIndex, (2) Mean (3) StdDev (4) Max (5) Min " << endl
         << "%% (6) Number of occurences (7-End) The Pattern" << endl;

 for(int PatternIndex = 0; PatternIndex < NumPatterns; PatternIndex++)
    {
      OutFile << PatternIndex << "\t"; 
      OutFile << MeanPattern[PatternIndex] << "\t"; 
      OutFile << StdDevPattern[PatternIndex] << "\t"; 
      OutFile << MaximumVoltagePattern[PatternIndex] << "\t"; 
      OutFile << MinimumVoltagePattern[PatternIndex] << "\t"; 
      OutFile << NumOccurencesPattern[PatternIndex] << "\t"; 

      for(int BitIndex = PatternLength-1; BitIndex >= 0; BitIndex--)
	{
          PatternArray[BitIndex] = GetPattern(PatternIndex,BitIndex);
          OutFile << PatternArray[BitIndex] << "\t";
	}

          OutFile << endl;
  

    } // ## end for loop

} // ## end WriteFilePatternStats

// ###############################################################

// add by Jonathan Hu
// adopt the code from Ivan Lima

//######################################################################
//## Get the BER for the full model for the specified decision current
//## using the semi-analytical algorithm
//## This method was described by Winzer et al.
//## This method is based on ::GetQ_FactorGaussianStatistics()
//##  ATTENTION: the noiseless signal has to be already in
//##             the object oReceiverSubsystem->SCSignal
//##             and its clock has to be previously recovered
//## The parameter PublicPowerPenaltyFactor must be equal to one
//## if you are not computing the power penalty.
//## From CompensationPMD
//## By: Ivan Lima (2/27/03) 
double ElecSignalStat::GetBER_GaussianStatisticsFromDecisionThreshold(double DecisionCurrent)
{
   int StringLength = oOptSignal->GetStringLength();

   //cout << "DecisionCurrent = " << DecisionCurrent << endl;


   double const PublicPowerPenaltyFactor = 1; //## Not used here
   int qtBit1 = 0;
   int qtBit0 = 0;
   double Sum_erfc1 = 0;
   double Sum_erfc0 = 0;
   for (int ii = 0; ii < StringLength; ii++ ) {
/*      double StdDevCurrentBit = sqrt( VarSignal_NoiseBeating[ii]
                                      /PublicPowerPenaltyFactor
                                     +VarNoise_NoiseBeating);
									 
									 
      //double CurrentBit =
      //      oReceiverSubsystem->oElecSignalStat->GetCurrentBit(ii)
      //                   /PublicPowerPenaltyFactor;
      double CurrentBit =

oReceiverSubsystem->oElecSignalStat->GetCurrentBitFromRecoveredSignal(ii)
                         /PublicPowerPenaltyFactor;
	 
*/						 
	double sqStdDevASE_ASE_Tmp = sq(GetStdDevASE_ASE(
				          NoiseSpectralDensityX_PD,
					  NoiseSpectralDensityY_PD ) );
//	sqStdDevASE_ASE_Tmp += 0.00005 ; //946225077/8; 0.00004-10
		// 0.00034 is the best for 70 ps raise time if you are using SHF
		// 0.00004 is the best for 70 ps raise time if you are using anritsu

    double StdDevCurrentBit = sqrt( sq(GetStdDevSignal_ASE_BitSlot(ii,
			                 NoiseSpectralDensityX_PD,
                                         NoiseSpectralDensityY_PD))
   			     +sqStdDevASE_ASE_Tmp );		
//	StdDevCurrentBit +=  0.0000084;
	double CurrentBit  = oElecSignalStat4->GetCurrentBit(ii);


      if ( GetBit(ii) == 1)
		{
         qtBit1++;
         double Arg_erfc = (CurrentBit - DecisionCurrent )
	       /(sqrt(2.0)*StdDevCurrentBit );
         Sum_erfc1 += 0.5*erfccExtendedRange( (float) Arg_erfc );
//	  	cout << " Arg_erfc=" << Arg_erfc << endl;
	  } else { //## Space (Bit "0")

         qtBit0++;
         double Arg_erfc = (DecisionCurrent - CurrentBit )
	       /(sqrt(2.0)*StdDevCurrentBit );
         Sum_erfc0 += 0.5*erfccExtendedRange( (float) Arg_erfc );
//	  cout << " Arg_erfc=" << Arg_erfc << endl;		 
      }

   }

   return (Sum_erfc1/qtBit1 + Sum_erfc0/qtBit0)/2.;
}

// ###############################################################

// add by Jonathan Hu
// adopt the code from Ivan Lima
//######################################################################
//## Get the BER for the full model for the optimized decision current
//## using the semi-analytical algorithm
//## The parameter PublicPowerPenaltyFactor must be equal to one
//## if you are not computing the power penalty.
//## Imported from CompensationPMD
//## By: Ivan Lima (2/27/02)

#define ITMAX 1000
#define CGOLD 0.3819660
#define ZEPS 1.0e-60
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);

double ElecSignalStat::GetBER_GaussianStatistics(void)
{
   double DecisionCurrent, Dec;
   double const PublicPowerPenaltyFactor = 1; //## Not used
/*   double MaxCurrentBit0 =
oReceiverSubsystem->oElecSignalStat->GetMaxCurrentBit0()
                          /PublicPowerPenaltyFactor;
   double MinCurrentBit1 =
oReceiverSubsystem->oElecSignalStat->GetMinCurrentBit1()
                          /PublicPowerPenaltyFactor;   
*/
   int WorstBitSlot1;
   double MinCurrentBit1 = oElecSignalStat4->GetMinCurrentBit1(&WorstBitSlot1);
   int WorstBitSlot0; 
   double MaxCurrentBit0 = oElecSignalStat4->GetMaxCurrentBit0(&WorstBitSlot0);
   double BER_dB = 0;
   
// Add by Jonathan Hu for debug
/*   for (int jj = 0; jj < StringLength; jj++) 
     {
	cout << "GetBit" << jj << " = " << GetBit(jj) << " bitIntensity " << jj << " = " << oElecSignalStat2->GetCurrentBit(jj) << endl;
	}
   cout <<  "MinCurrentBit1"  << "  " << MinCurrentBit1;
   cout <<  "   MaxCurrentBit0"  << "  " << MaxCurrentBit0  << endl;;

   double Dec;
   for (int jj = 0; jj < 30; jj++) 
     {
 		Dec = MaxCurrentBit0 + jj/ 30.0 * (MinCurrentBit1-MaxCurrentBit0) ;
		cout << Dec << "->" <<  GetBER_GaussianStatisticsFromDecisionThreshold(Dec) << "         " ;
	}   
*/   
   if (MaxCurrentBit0 < MinCurrentBit1) {
      double Tolerance = 1e-3;
	  
	  
/*     BER_dB = brentd(MaxCurrentBit0,
	  				(MaxCurrentBit0+MinCurrentBit1)/2.,
                    MinCurrentBit1,
					&GetBER_GaussianStatisticsFromDecisionThreshold,
					Tolerance,
		            &Dec); 
*/
//   DecisionCurrent = Dec;
   	double ax = MaxCurrentBit0;
	double bx = (MaxCurrentBit0+MinCurrentBit1)/2.;
	double cx = MinCurrentBit1;
	double tol = 1e-3;
	double *xmin;
	xmin = &DecisionCurrent;
	double returnfx;
	 

	int iter;
	double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
	double e=0.0;

	a=(ax < cx ? ax : cx);
	b=(ax > cx ? ax : cx);
	x=w=v=bx;
	fw=fv=fx=GetBER_GaussianStatisticsFromDecisionThreshold(x);
		for (iter=1;iter<=ITMAX;iter++) {
		xm=0.5*(a+b);
		tol2=2.0*(tol1=tol*fabs(x)+ZEPS);
		if (fabs(x-xm) <= (tol2-0.5*(b-a))) {
			*xmin=x;
			returnfx = fx;
			iter = ITMAX+10;
		}
		if (fabs(e) > tol1) {
			r=(x-w)*(fx-fv);
			q=(x-v)*(fx-fw);
			p=(x-v)*q-(x-w)*r;
			q=2.0*(q-r);
			if (q > 0.0) p = -p;
			q=fabs(q);
			etemp=e;
			e=d;
			if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
				d=CGOLD*(e=(x >= xm ? a-x : b-x));
			else {
				d=p/q;
				u=x+d;
				if (u-a < tol2 || b-u < tol2)
					d=SIGN(tol1,xm-x);
			}
		} else {
			d=CGOLD*(e=(x >= xm ? a-x : b-x));
		}
		u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
		fu=GetBER_GaussianStatisticsFromDecisionThreshold(u);
		if (fu <= fx) {
			if (u >= x) a=x; else b=x;
			SHFT(v,w,x,u)
			SHFT(fv,fw,fx,fu)
		} else {
			if (u < x) a=u; else b=u;
			if (fu <= fw || w == x) {
				v=w;
				w=u;
				fv=fw;
				fw=fu;
			} else if (fu <= fv || v == x || v == w) {
				v=u;
				fv=fu;
			}
		}
	} 
 	if ( iter < ITMAX+5){
		nrerror("Too many iterations in brent");
		*xmin=x;
		returnfx = fx;
	}

//	cout << " run after brentd" << endl;
//  cout << *xmin << "  " << returnfx << "  " << Q_Factor << endl;
  BER_dB = returnfx;
  
   }
//cout << " BER_dB = " << BER_dB << endl << "dB2Linear(BER_dB)= " << dB2Linear(BER_dB);
//exit(1);

   return BER_dB;

   // return dB2Linear(BER_dB);
}

/*
//################################################################
//## The function used by the 1-D minimization algorithm
//## By: Ivan Lima (4/8/03)
float GetBER_dB_FromDecisionThresholdPolariz(float DecisionCurrent)
{
   double BER_dB =    Linear2dB( GetBER_GaussianStatistics(
     	                 DecisionCurrent) );
   //cout << " ** DecisionCurrent = " << DecisionCurrent;
   //cout << "   BER = " << dB2Linear(BER_dB) << endl;
   return BER_dB;
}
*/



//###########
//You should include this line in your program that gets the Q-factor
//that corresponds to a given BER:

//## That's it
// ###############################################################

// add by Jonathan Hu
// adopt the code from Ivan Lima

//###########################################################
//##  Get the Q-factor in a semi analytical form as outlined
//#   by P. Winzer. This module is returning the worse Q-factor
//#   in the pattern. That comes from the lowest noiseless bit 
//#   one and the highest noiseless bit zero. Thus, only two
//#   bits are used in this computation.
//#   WARNING: oOptSignal has to contain a noiseless signal 
//#   prior to the optical filter. It is very important.
//##  This is just an interface to the actual method for the
//#   particular case in which the noise is unpolarized.
//## By: Ivan Lima (2/5/14)
double ElecSignalStat::GetQ_FactorGaussianStatisticsFromBER(
                             double NoiseSpectralDensity)
{
   return GetQ_FactorGaussianStatisticsFromBER(NoiseSpectralDensity/2.,
                                                NoiseSpectralDensity/2.);
}						

// ###############################################################

// add by Jonathan Hu
// adopt the code from Ivan Lima

//###########################################################
//##  Get the Q-factor in a semi analytical form as outlined
//#   by P. Winzer. This module is returning the worse Q-factor
//#   in the pattern. That comes from the lowest noiseless bit 
//#   one and the highest noiseless bit zero. Thus, only two
//#   bits are used in this computation.
//#   WARNING: oOptSignal has to contain a noiseless signal 
//#   prior to the optical filter. It is very important.
//## By: Ivan Lima (5/14/02)
double ElecSignalStat::GetQ_FactorGaussianStatisticsFromBER(
                             double NoiseSpectralDensityX,
			     double NoiseSpectralDensityY)
{ // CheckGaussianStatisticsInitialization(
   //      "GetQ_FactorGaussianStatisticsWorstPattern");
		         
   //## Start the actual computation
   ///////oOptSignal2->RegenerateSignalString();   
   
   //cout << "Copying oOptSignal "<< endl;
// Changed by Jonathan Hu this has problem
//      oOptSignal2 = oOptSignal;
   *oOptSignal2 = *oOptSignal;
/*   
   //cout << "$#$#$# GetFrequency(0) = " << oOptSignal2->GetFrequency(0) << endl;
   cout << "$#$#$ GetBit (Ch0): 0, 1, 2, 3 -> " 
        << oOptSignal->oBitString->GetBit(0,0) << "  "
        << oOptSignal->oBitString->GetBit(0,1) << "  " 
	<< oOptSignal->oBitString->GetBit(0,2) << endl;
   cout << "$#$#$ GetBit2 (Ch0): 0, 1, 2, 3 -> " 
        << oOptSignal2->oBitString->GetBit(0,0) << "  "
        << oOptSignal2->oBitString->GetBit(0,1) << "  " 
	<< oOptSignal2->oBitString->GetBit(0,2) << endl;
*/
   //cout << "oOptSignal copied. Now start filtering "<< endl;   
/*
   //oPhotodetector2->WriteFileTimePower("tOutSignal.dat"); 
   oOptSignal2->WriteFileTimePower("tOutOptSignal2b.dat");       
   oOptSignal2->WriteFileFreqPower("fOutOptSignal2b.dat"); 
   oOptSignal->WriteFileTimePower("tOutOptSignal.dat");       
   oOptSignal->WriteFileFreqPower("fOutOptSignal.dat");    
   //exit(1); //#$#$#$#
*/

   //oOptFilter2->SetRelativeCenterFreq(oOptSignal2->GetRelativeCenterFrequency(0)); //#$$#  
   oOptFilter2->FilterOptSignal();
   oPhotodetector2->DetectOptSignal();
   oElecFilter2->FilterElecSignal();
   oElecSignalStat2->ComputePerformanceStatistics();



// add following for PD
   *oOptSignal4 = *oOptSignal;
   NoiseSpectralDensityX_PD = NoiseSpectralDensityX;
   NoiseSpectralDensityY_PD = NoiseSpectralDensityY;

   oOptFilter4->FilterOptSignal();
   oPhotodetector4->DetectOptSignal();
   oElecFilter4->FilterElecSignal();
   oElecSignalStat4->ComputePerformanceStatistics();
      
/*	for (double jj = 100; jj >0.01; jj/=sqrt(sqrt(10))) 
     {
 		cout << jj << "->" <<  erfccExtendedRange(jj) << endl ;
	} 
   exit(1);
*/ //	return GetBER_GaussianStatistics();
/*
   for (int jj = 0; jj < StringLength; jj++) 
     {
	cout << "GetBit" << jj << " = " << GetBit(jj) << " bitIntensity " << jj << " = " << oElecSignalStat2->GetCurrentBit(jj) << endl;
	}
*/	
//	exit(1);
	
/*
   //oPhotodetector2->WriteFileTimePower("tOutSignal.dat"); 
   oOptSignal2->WriteFileTimePower("tOutOptSignal2a.dat");       
   //oOptSignal->WriteFileTimePower("tOutOptSignal.dat");       
   exit(1); //#$#$#$#
*/
   //cout << "The eye opening will be obtained now " << endl;
    double Q_Factor;
    double Tolerance = 1e-3;
// double MinimizeSqQ_FactorVersusBER(double Q_FactorTmp);

//      brentd(0.5,2+Q_FactorPre,40., &MinimizeSqQ_FactorVersusBER,Tolerance,
// 		         &Q_Factor);
//     cout.precision(10);
//       brentHu(-7,6.,9., &testsq,Tolerance,&xmin); 
//	cout << " input Q_FactorPre ?=" ;
//	cin >> Q_FactorPre ;
	double ax = 0.5;
	double bx = 2+Q_FactorPre;
	double cx = 40;
	double tol = 1e-3;
	double *xmin;
	xmin = &Q_Factor;
	double returnfx;
	
cout.precision(10);
	int iter;
	double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
	double e=0.0;

	a=(ax < cx ? ax : cx);
	b=(ax > cx ? ax : cx);
	x=w=v=bx;
	BER_GaussianStatistics = GetBER_GaussianStatistics();
	fw=fv=fx=MinimizeSqQ_FactorVersusBER(x);
		for (iter=1;iter<=ITMAX;iter++) {
		xm=0.5*(a+b);
		tol2=2.0*(tol1=tol*abs(x)+ZEPS);
//		cout << " x = " <<  x  << "(" << erfccExtendedRange(x/sqrt(2.))/2. << "," << MinimizeSqQ_FactorVersusBER(x) << ")"  << " xm = " << xm 
//				<< "(" << erfccExtendedRange(xm/sqrt(2.))/2. << "," << MinimizeSqQ_FactorVersusBER(xm) <<  ") tol2= " << tol2 <<  " -0.5(b-a)= " << -0.5*(b-a) << "  "  << abs(x-xm) 
//				<< "   " << (tol2-0.5*(b-a)) << endl;
		if (abs(x-xm) <= (tol2-0.5*(b-a))) {
			*xmin=x;
			returnfx = fx;
			iter = ITMAX+10;
		}
		if (abs(e) > tol1) {
			r=(x-w)*(fx-fv);
			q=(x-v)*(fx-fw);
			p=(x-v)*q-(x-w)*r;
			q=2.0*(q-r);
			if (q > 0.0) p = -p;
			q=abs(q);
			etemp=e;
			e=d;
			if (abs(p) >= abs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
				d=CGOLD*(e=(x >= xm ? a-x : b-x));
			else {
				d=p/q;
				u=x+d;
				if (u-a < tol2 || b-u < tol2)
					d=SIGN(tol1,xm-x);
			}
		} else {
			d=CGOLD*(e=(x >= xm ? a-x : b-x));
		}
		u=(abs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
//		BER_GaussianStatistics = GetBER_GaussianStatistics();
		fu=MinimizeSqQ_FactorVersusBER(u);
		if (fu <= fx) {
			if (u >= x) a=x; else b=x;
			SHFT(v,w,x,u)
			SHFT(fv,fw,fx,fu)
		} else {
			if (u < x) a=u; else b=u;
			if (fu <= fw || w == x) {
				v=w;
				w=u;
				fv=fw;
				fw=fu;
			} else if (fu <= fv || v == x || v == w) {
				v=u;
				fv=fu;
			}
		}
	} 
 	if ( iter < ITMAX+5){
		nrerror("Too many iterations in brent");
		*xmin=x;
		returnfx = fx;
	}

//	cout << " run after brentd" << endl;
//  cout << *xmin << "  " << returnfx << "  " << Q_Factor << endl;
	
cout << "Q=" << Q_Factor << "  erfc=" << erfccExtendedRange(Q_Factor/sqrt(2.))/2. 
		<< " GetBer=" << BER_GaussianStatistics
//		<< " I# = " << iter  // << endl;
//exit(1);

	<< " Min Fun = " << MinimizeSqQ_FactorVersusBER(Q_Factor) << endl << endl;
//   cout << Q_Factor << endl;
	Q_FactorPre = Q_Factor;
   return Q_Factor;
}				 


double ElecSignalStat::testsq(double ax)
{
	return sq(ax)*sq(ax) -ax*50;
// 	return sq(ax);

}


double ElecSignalStat::brentHu(double ax, double bx, double cx, double (*f)(double), double tol,
	double *xmin)
{

	int iter;
	float a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
	float e=0.0;

	a=(ax < cx ? ax : cx);
	b=(ax > cx ? ax : cx);
	x=w=v=bx;
	fw=fv=fx=(*f)(x);
		for (iter=1;iter<=ITMAX;iter++) {
		xm=0.5*(a+b);
		tol2=2.0*(tol1=tol*fabs(x)+ZEPS);
		if (fabs(x-xm) <= (tol2-0.5*(b-a))) {
			*xmin=x;
			return fx;
		}
		if (fabs(e) > tol1) {
			r=(x-w)*(fx-fv);
			q=(x-v)*(fx-fw);
			p=(x-v)*q-(x-w)*r;
			q=2.0*(q-r);
			if (q > 0.0) p = -p;
			q=fabs(q);
			etemp=e;
			e=d;
			if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
				d=CGOLD*(e=(x >= xm ? a-x : b-x));
			else {
				d=p/q;
				u=x+d;
				if (u-a < tol2 || b-u < tol2)
					d=SIGN(tol1,xm-x);
			}
		} else {
			d=CGOLD*(e=(x >= xm ? a-x : b-x));
		}
		u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
		fu=(*f)(u);
		if (fu <= fx) {
			if (u >= x) a=x; else b=x;
			SHFT(v,w,x,u)
			SHFT(fv,fw,fx,fu)
		} else {
			if (u < x) a=u; else b=u;
			if (fu <= fw || w == x) {
				v=w;
				w=u;
				fv=fw;
				fw=fu;
			} else if (fu <= fv || v == x || v == w) {
				v=u;
				fv=fu;
			}
		}
	}
	nrerror("Too many iterations in brent");
	*xmin=x;
	return fx;
}

/* (C) Copr. 1986-92 Numerical Recipes Software *1%:?!$:V3+). */


// ###############################################################

// add by Jonathan Hu
// adopt the code from Ivan Lima
//################################################################
//## The function used by the root search algorithm
//## By: Ivan Lima (4/8/03)
//double ElecSignalStat::MinimizeSqQ_FactorVersusBER(double Q_FactorTmp)
double ElecSignalStat::MinimizeSqQ_FactorVersusBER(double Q_FactorTmp)
{   
   //cout << "Q_FactorTmp = " << Q_FactorTmp << endl;
   double Root =  abs( BER_GaussianStatistics
-erfccExtendedRange(Q_FactorTmp/sqrt(2.))/2.);
   //cout << " BER_dB = " << Linear2dB(oPolarizEffects->PublicBER_Tmp);
   //cout << " BER_dB_FromQ = " << Linear2dB(erfccExtendedRange(Q_FactorTmp/sqrt(2.))/2.);
   //cout << " Root = " << Root;
   //cout << " ** Q_FactorTmp = " << Q_FactorTmp << endl;
    return Root;
}

