// $Id: ocsHistogram.cc,v 1.2 2005/07/06 16:29:59 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsHistogram.hh"

extern ofstream *LogFile;

// ################### Methods for the class Histogram


/*!    
       This constructor calls the private method InitHistogram 

       \param qtBins2 The number of bins in the histogram
       \param MinRange2 Minimum value of the domain of the Histogram  
       \param MaxRange2 Maximum value of the domain of the Histogram 
       \sa Histogram(string InFileName) ~Histogram() 
           InitHistogram(int qtBins2,double MinRange2, double MaxRange2)
  */


// #### Constructor of the class Histogram with initialization
// ## By: Ivan Lima

Histogram::Histogram(int qtBins2 ,double MinRange2,double MaxRange2)
{
   ObjectInitialized = 0;
   InitHistogram(qtBins2,MinRange2,MaxRange2);   
   MakeFileWithSamples = 0;
}

// #############################################

/*!    
       This constructor reads parameters from the file 

       \param InFileName: Name of parameter input file (a C++ string) 
     
       This constructor just calls InitHistogram(string InFileName). 

       \sa InitHistogram(string InFileName)
           Histogram(int qtBins2,double MinRange2, double MaxRange2)
           ~Histogram() 
           InitHistogram
           AddImportantSample
           MakeFileWithSamples
           FileNameWithSamples
  */

Histogram::Histogram(string InFileName)
{
  InitHistogram(InFileName);
}

/*!    
       This method reads parameters from the file 

       \param InFileName: Name of parameter input file (a C++ string) 
     
       This constructor calls 
         InitHistogram(int qtBins2,double MinRange2, double MaxRange2). 
       It also initializes the flag MakeFileWithSamples 
       and the output file name FileNameWithSamples

       \sa InitHistogram(string InFileName)
           Histogram(int qtBins2,double MinRange2, double MaxRange2)
           ~Histogram() 
           InitHistogram
           AddImportantSample
           MakeFileWithSamples
           FileNameWithSamples
  */


// #################################################

void Histogram::InitHistogram(string InFileName)
{

   // Written by John Zweck Oct 25th 2000
   // ## Modified by Ivan Lima in 3/4/01
   
   
   HistInFileName = InFileName; //## This is used by other methods
   
   ifstream InFile(InFileName.c_str(),ios::in);  

   if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }

   DebugLevel =    ReadInt("DebugLevel"," ",0,1,
			    LOWER_AND_UPPER,OPTIONAL_WARNING,1,&InFile,0);

   if(DebugLevel)
     {
      LogFileSeparator();
      *LogFile << InFileName << endl << endl;
     }


   int qtBins2 = ReadInt("qtBins"," ",0,-1,LOWER_ONLY,&InFile,DebugLevel);
   double MinRange2 = ReadDouble("MinRange"," ",0,0,NO_BOUNDS,&InFile,
				 DebugLevel);
   double MaxRange2 = ReadDouble("MaxRange"," ",0,0,NO_BOUNDS,&InFile,
				 DebugLevel);

   
   MakeFileWithSamples =    ReadInt("MakeFileWithSamples"," ",0,1,
				    LOWER_AND_UPPER,OPTIONAL_WARNING,&InFile,
				    DebugLevel);

   
   FileNameWithSamples = ReadString("FileNameWithSamples",
                                    OPTIONAL_WARNING, &InFile,DebugLevel);
   
   if (MakeFileWithSamples && FileNameWithSamples == "") {
      cout << endl;
      cout  << "Error: you should include the field" 
           << " FileNameWithSamples if you are setting"<< endl
           << "MakeFileWithSamples == "<< MakeFileWithSamples
           << endl; 
      exit(1);      
   }
   
   if (MakeFileWithSamples) {
      OutFileWithSamples = new ofstream(FileNameWithSamples.c_str(),ios::out);
   }
      
   InFile.close();

   ObjectInitialized = 0;
   InitHistogram(qtBins2,MinRange2,MaxRange2);  
}

// ############################################

// #### Destructor of the class Histogram 
// ## By: Ivan Lima
Histogram::~Histogram()
{
//   if (ObjectInitialized)
   ReleaseMemory();   
}


// ############################################
//## Allow change in the file that is generated with the samples
//## By: Ivan Lima (4/9/03)
void Histogram::SetFileNameWithSamples(string FileNameWithSamples2)
{
   FileNameWithSamples = FileNameWithSamples2;
   if (! MakeFileWithSamples) {
      OutFileWithSamples = new ofstream(FileNameWithSamples.c_str(),ios::out);
      MakeFileWithSamples = 1;
   } else {
      OutFileWithSamples->close();
      OutFileWithSamples->open(FileNameWithSamples.c_str());   
   }
   *LogFile << "From Histogram: FileNameWithSamples changed to: " 
           << FileNameWithSamples << endl;
}





// ##############################################

// #### Releases the memory (internal)
// ## By: Ivan Lima
void Histogram::ReleaseMemory()
{
   delete [] pdfX;  
   delete [] pdfY;
   delete [] pdfConfidence;
   delete [] Bin;
   delete [] BinSquare;

   delete [] cdfConfidence;
   delete [] cdfBin;
   delete [] cdfBinSquare;   

   delete [] cdfcConfidence;
   delete [] cdfcBin;
   delete [] cdfcBinSquare; 
   
   delete [] qtSamplesBin;
   delete [] ExpPolynomialCoef;  
   if (MakeFileWithSamples) {
      delete OutFileWithSamples;
   }   
}
   
// #############################################

// #### Initializes the object Histogram
// ## By: Ivan Lima   
void Histogram::InitHistogram(int qtBins2 ,double MinRange2,double MaxRange2)

                       
{
   qtBins    = qtBins2;   
   MinRange      = MinRange2;
   MaxRange      = MaxRange2;
   if (MaxRange <= MinRange) {
      cout << endl;
      cout << "Error from class Histogram: MinRange cannot be > MaxRange."
           << endl << "MinRange = " << MinRange 
           << "     MaxRange = " << MaxRange 
           << endl;
      exit(1);   
   
   }   
   
   if (ObjectInitialized && qtBins != qtBinsLast) {
      ReleaseMemory();
      ObjectInitialized = 0;
   }   
   if (! ObjectInitialized ) {
      ObjectInitialized = 1;        
      pdfX = new double[qtBins];
      pdfY = new double[qtBins];
      pdfConfidence = new double[qtBins];      
      Bin  = new double[qtBins];
      BinSquare = new double[qtBins];      

      cdfConfidence = new double[qtBins];      
      cdfBin  = new double[qtBins];
      cdfBinSquare = new double[qtBins];    

      cdfcConfidence = new double[qtBins];      
      cdfcBin  = new double[qtBins];
      cdfcBinSquare = new double[qtBins];  
      
      qtSamplesBin = new long[qtBins];
      qtBinsLast = qtBins;
      ExpPolynomialCoef = new double[qtBins];
   }
   PreviouslyComputed_pdfConfidenceFlag = 0; //## Default value   
   //BalanceHeuristicInitialized = 0;  
   ClearHistogram();
}

// ########################################################

// #### Clear the data to produce a histogram with the initial parameters
// ## By: Ivan Lima
void Histogram::ClearHistogram()
{
   // ## Initialize the variables

   qtSamples = 0;
   qtSamplesInRange = 0;
   SumLikelihoodRatio = 0;
   SumUnitaryLikelihoodRatio = 0;
   MeanTmp =  MeanSquareTmp =0; 
   for (int ii=0; ii < qtBins; ii++) {
      qtSamplesBin[ii] = 0;

      Bin[ii] = 0;    
      BinSquare[ii] = 0;       
      pdfConfidence[ii] = 0;

      cdfBin[ii] = 0;    
      cdfBinSquare[ii] = 0; 
      cdfConfidence[ii] = 0;   

      cdfcBin[ii] = 0;    
      cdfcBinSquare[ii] = 0;
      cdfcConfidence[ii] = 0;               
   }   
}   

// #######################################################

// #### Adds a sample to the statistics
// ## By: Ivan Lima
void Histogram::AddSample(double vSample)
{
   //## The regular histogram has unitary 
   //#  likelihood ratio.
   AddImportantSample(vSample,1.);

}   

// #######################################################

/*!
      Algorithm:
      -# If the flag MakeFileWithSamples == 1, then writes sample
      and likelihood ratio to the file whose name is given by the string
      FileNameWithSamples 
      -# If the sample is in the domain of the histogram, the 
      variables used internally to compute the mean and standard deviation
      of the histogram  are updated 
      -# If the sample is in the domain of the histogram,
       the histogram itself is updated by placing
       the sample in the correct bin a weight that is given by the 
       LikelihoodRatio
      -# The histogram of the squares of the samples is also updated.
      -# If the sample is out of range the MinSample and MaxSample
        values are updated.

    \param  vSample: Value of sample to be added (double)
    \param   LikelihoodRatio: Likelihood this sample would occur
             in a standard Monte Carlo simulation

    \sa MakeFileWithSamples Bin

 */


//###########################################################
// #### Adds an important sample to the statistics
// ## By: Ivan Lima
void Histogram::AddImportantSample(double vSample, double LikelihoodRatio)
{
   if (MakeFileWithSamples) 
    {
      *OutFileWithSamples << vSample << "  " << LikelihoodRatio << endl;
    }
      
   qtSamples++;
   SumLikelihoodRatio += LikelihoodRatio;   

   if(LikelihoodRatio == 1.) 
    {
      SumUnitaryLikelihoodRatio++;
    }

   MeanTmp += vSample*LikelihoodRatio;
   MeanSquareTmp += sq(vSample)*LikelihoodRatio;       


   Index_pdf = GetIndex_pdf(vSample);
   //Index_pdf = (int) ( qtBins*(vSample - MinRange)
   //                   /( MaxRange - MinRange ) );
   if( MinRange < vSample && vSample < MaxRange) 
    {  
      qtSamplesInRange++;
      Bin[Index_pdf] += LikelihoodRatio;
      BinSquare[Index_pdf] += sq(LikelihoodRatio); 
     
      //## Since the random variable at stake here is 
      //#  the indicator function of a bin, it is the 
      //#  thing that is squared (=1), which should 
      //#  be weighted by the likelihood ratio.
      
      //cout << BinSquare[Index_pdf]/qtSamples << " " 
      //<<sq(Bin[Index_pdf]/qtSamples) << endl;
 
      qtSamplesBin[Index_pdf] += 1;
   }  
   //## Now compute the bins of the cdf and the cdfc
   for (int ii = 0; ii < qtBins; ii++) {
      if (Index_pdf >= ii) {
         cdfcBin[ii] += LikelihoodRatio;
	 cdfcBinSquare[ii] += sq(LikelihoodRatio);
      } else { //## Index_pdf < ii
         cdfBin[ii] += LikelihoodRatio;
	 cdfBinSquare[ii] += sq(LikelihoodRatio);      
      }
   }

   if (qtSamples == 1)
      MinSample = MaxSample = vSample;
               
   if (vSample < MinSample)
      MinSample = vSample;
   if (vSample > MaxSample)
      MaxSample = vSample;    
    
}   

//###########################################################
//#### Get the index of the pdf
//## By: Ivan Lima
int Histogram::GetIndex_pdf(double vSample)
{
   int Index_pdfTmp = (int) ( qtBins*(vSample - MinRange)
                      /( MaxRange - MinRange ) );
   return Index_pdfTmp;		      
}
// ############################################################

/*! 

   This method converts the discrete histogram data into a pdf.\n
   It sets the values of three arrays: 
   -# pdfX: Stores the discrete values of the independent variable of the pdf. 
      These values are computed using the centers of the bins.\n
   -# pdfY: Stores the dependent values of the pdf. The pdf is normalized
      to have integral 1.
   -# pdfConfidence: ASK IVAN

 */


// #### Obtain the pdf of the distribution
// #    It puts the data in the vectors such that 
// #    they can be used by the other modules.
// ## By: Ivan Lima

void Histogram::Obtain_pdf()
{
   if (qtSamplesInRange == 0) {
      cout << "From Histogram::Obtain_pdf:" << endl
           << "Warning: Object of class Histogram without"
           << "any sample inside the range." << endl
           << "The final range was: " << endl
           << "     Min = " << GetMinSample() << endl
           << "     Max = " << GetMaxSample() << endl
           << endl;
      //exit(1);      
   }
   
   int IsNormalizedSumLikelihoodRatio = 1;   
   
   /***
   if (SumUnitaryLikelihoodRatio < 0.001*qtSamples) { //## Less than 0.1%
      //## Only purely biased distributions can satisfy this condition
      //#  It sets the flag to provide a better approximation for the pdf
      //#  But the user should combine biased and unbiased to 
      //#  resolve the pdf adequatly
      IsNormalizedSumLikelihoodRatio = 0;
      LogFileSeparator();      
      *LogFile << "Histogram: SumUnitaryLikelihoodRatio = "
              << SumUnitaryLikelihoodRatio/ (double) qtSamples
              << " * qtSamples ( <0.1% )." << endl;
      *LogFile << "   You should combine biased and unbiased"
              << " pdfs using MultipleSelectiveImport." 
              << endl << endl;        
      cout << " (From Histogram: See *.log) " << endl;        
   }
   *****/
   
   for (int ii = 0; ii < qtBins; ii++)  {
     pdfX[ii] = (ii + 0.5)*(MaxRange - MinRange)/qtBins + MinRange;
     
     //## The first case will be removed soon. I have to fix the
     //#  methods that merge multiple biased distributions.
     ////if (IsNormalizedSumLikelihoodRatio) { 
     ////   pdfY[ii] = Bin[ii]/SumLikelihoodRatio*qtBins/(MaxRange - MinRange);
     ////}
     ////else {
 
     pdfY[ii] = Bin[ii]/qtSamples*qtBins/(MaxRange - MinRange);

     ////}
     //cout << ii << " "<< BinSquare[ii]/qtSamples<< " "
     //<< sq(Bin[ii]/qtSamples)<< endl;

     if (! PreviouslyComputed_pdfConfidenceFlag ) {
        //## The confidence is the StdDev of pdfY divided by qtSamples.
	//## The actual confidence inteval is the mean plus and 
	//## minus the Confidence obtained here.
	pdfConfidence[ii] = 0;
        if (BinSquare[ii] > sq(Bin[ii])/qtSamples )  {
           pdfConfidence[ii] = sqrt( ( BinSquare[ii] 
                              -qtSamples*sq(Bin[ii]/qtSamples) )
                               /( qtSamples-1) 
                               /qtSamples );   
        } 
        //## Normalization of the confidence interval for file output
        pdfConfidence[ii] *= qtBins/(MaxRange - MinRange);	
	
	//## Computation of the confidence interval of the cdf
	cdfConfidence[ii] = 0;
        if (cdfBinSquare[ii] > sq(cdfBin[ii])/qtSamples )  {
           cdfConfidence[ii] = sqrt( ( cdfBinSquare[ii] 
                              -qtSamples*sq(cdfBin[ii]/qtSamples) )
                               /( qtSamples-1) 
                               /qtSamples );   
        } 
	//## Computation of the confidence interval of the cdfc
	cdfcConfidence[ii] = 0;
        if (cdfcBinSquare[ii] > sq(cdfcBin[ii])/qtSamples )  {
           cdfcConfidence[ii] = sqrt( ( cdfcBinSquare[ii] 
                              -qtSamples*sq(cdfcBin[ii]/qtSamples) )
                               /( qtSamples-1) 
                               /qtSamples );   
        } 
     }

     //cout << ii << "  " << pdfY[ii] << "  "
     //<< Bin[ii]/qtSamples << "  " << pdfConfidence[ii] << endl;    
     //cout << pdfY[ii] <<" "<< Bin[ii] <<"  "<< SumLikelihoodRatio << endl;

   } // ## end for ii loop 
             
  //cout << "qtSamples " << qtSamples << endl;
  //cout << "SumLikelihoodRatio " << SumLikelihoodRatio << endl;     


} // ## end Obtain_pdf


//#################################################################
//##### Imports the data from files created by the 
//##    histogram class applied to PMD.
//##    The heuristic balance weights need to be previously generated 
//##    A description of this method can be found in 
//##    Dr. Eric Veach's Ph.D. thesis
//##    This is, absolutely, the right way to merge 
//##    samples obtained from multiple bias.
//##    The confidence interval is corect (1/17/02).
//##    Convert2dB_Scale must be -1 for -10 log10(x) or -2 for -20 log10(x)
//## By: Ivan Lima   07/27/02
void Histogram::MultipleHeuristicBalanceImport(string InFileName, 
                                               int Convert2dB_Scale)
{
   //## Structure of InFileName
   //##  SampDataFileName

   //## Clear the histogram, just in case...
   ClearHistogram();
   
   ifstream * InFile;

   InFile = new ifstream;

   InFile->open(InFileName.c_str(),ios::in);

   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }  

   double Sample, LikelihoodRatio;
   //double MinRangeTmp, MaxRangeTmp;
   string SampFileName;
   string SampHeuristicBalanceWeightFileName;
   
   
   //## First, count the number of samples per ensemble
   ifstream * InFileWithImpSampWeights;
   InFileWithImpSampWeights = new ifstream;
    
   #define MAX_FILES_HEURISTIC_BALANCE 100
   double qtSamplesFile[MAX_FILES_HEURISTIC_BALANCE];
   Histogram *oHistogramBias[MAX_FILES_HEURISTIC_BALANCE];
   double qtSamplesTotal = 0;
   int FileCounter = 0;

   while ( *InFile >> SampFileName) {
      *InFile >> SampHeuristicBalanceWeightFileName;              

      InFileWithImpSampWeights->open(SampFileName.c_str(),ios::in);
   				     
      if (! InFileWithImpSampWeights ) {
         continue;
      }	      

      InFileWithImpSampWeights->clear();
      InFileWithImpSampWeights->seekg(0);

      double UselessValue, LikelihoodRatioFile;
      qtSamplesFile[FileCounter] = 0;    
     
      while (*InFileWithImpSampWeights >> UselessValue ) {         
         *InFileWithImpSampWeights >> LikelihoodRatioFile; 
         qtSamplesFile[FileCounter]++;   
      }	 
      qtSamplesTotal += qtSamplesFile[FileCounter];
      cout << SampFileName 	 
	   << "\t"<< qtSamplesFile[FileCounter]
           << endl;  

      InFileWithImpSampWeights->close();      

      oHistogramBias[FileCounter] = new Histogram(qtBins, MinRange, MaxRange);         
      FileCounter++;     
   } // end while   

   //## Do the actual processing
   //InFile.seekg(0); //## Resets the file pointer to the start of the file.
   InFile->close();
   InFile->open(InFileName.c_str(),ios::in);
   
   InFile->clear();
   InFile->seekg(0);
   
   FileCounter = 0;

   ifstream * SampFile;
   SampFile = new ifstream;

   ifstream * SampHeuristicBalanceFile;
   SampHeuristicBalanceFile = new ifstream;

   while ( *InFile >> SampFileName) {
      *InFile >> SampHeuristicBalanceWeightFileName;
      //cout << SampFileName << endl;
      //InFile >> MinRangeTmp;
      //InFile >> MaxRangeTmp;

      SampFile->open(SampFileName.c_str(),ios::in);

      if ( ! SampFile) {
         cout << endl;
         cout << "Warning: " << SampFileName << " could not be open."
              << endl;
         continue;	      
         //exit(1);      
      }       

     SampFile->clear();
     SampFile->seekg(0);

     SampHeuristicBalanceFile->open(SampHeuristicBalanceWeightFileName.c_str(),ios::in);

      if ( ! SampHeuristicBalanceFile) {
         cout << endl;
         cout << "Error: " << SampHeuristicBalanceWeightFileName 
	      << " could not be open."
              << endl;	      
         exit(1);      
      }   

     SampHeuristicBalanceFile->clear();
     SampHeuristicBalanceFile->seekg(0);

      cout << SampFileName << "   " 
           << SampHeuristicBalanceWeightFileName << endl;
      double LikelihoodRatioHeuristicBalance;
      double HeuristicBalanceWeight;
      double TrueCombinedWeight;
      while ( *SampFile >> Sample) {
         *SampFile >> LikelihoodRatio;
	 *SampHeuristicBalanceFile >> LikelihoodRatioHeuristicBalance;
	 *SampHeuristicBalanceFile >> HeuristicBalanceWeight;	 
	 *SampHeuristicBalanceFile >> TrueCombinedWeight;
       
         if ( abs( (LikelihoodRatio - LikelihoodRatioHeuristicBalance)
	      /LikelihoodRatio) > 1e-4 ) {   	 
	 //if (LikelihoodRatio != LikelihoodRatioHeuristicBalance ) {
	    cout << "Error: the LikelihoodRatio of the sample and " << endl
	         << "the one reported in the heuristic balance do not match."
		 << endl << "Check " << InFileName << endl
		 << "Aborting now ..." << endl;
            exit(1);		 	   
	 }
         //if (Sample >= MinRangeTmp && Sample < MaxRangeTmp) {
            if (LikelihoodRatio > 0) {
               //## Samples can be filtered by the user 
               //#  doing the LikelihoodRatio < 0
               //AddImportantSample(Sample,LikelihoodRatio);          
               /*
               cout << "Actual weight = " << LikelihoodRatio
		                     *HeuristicBalanceWeight 
					*qtSamplesTotal
					/qtSamplesFile[FileCounter] << endl;	       					
               cout << "**" << FileCounter 
	            << "  " << qtSamplesFile[FileCounter]					
		    << "  " << qtSamplesTotal
		    << "  " << HeuristicBalanceWeight
		    << "  " << LikelihoodRatio << endl;
		 */   
		/* 
               if (Convert2dB_Scale) {
                  AddImportantSample( (Convert2dB_Scale)
                                     *Linear2dB(Sample),LikelihoodRatio
			                *HeuristicBalanceWeight
					*qtSamplesTotal
					/qtSamplesFile[FileCounter]);
               }
               else {
                  AddImportantSample(Sample,LikelihoodRatio
		                     *HeuristicBalanceWeight 
					*qtSamplesTotal
					/qtSamplesFile[FileCounter] );
               }  
	       */   
               if (Convert2dB_Scale) {
                  AddImportantSample( (Convert2dB_Scale)
                                     *Linear2dB(Sample),
		                     TrueCombinedWeight);
                  oHistogramBias[FileCounter]->AddImportantSample( 
		                     (Convert2dB_Scale)
                                     *Linear2dB(Sample),
	   		HeuristicBalanceWeight * LikelihoodRatio);				     
				     //TrueCombinedWeight);				     				     
               } else {
                  AddImportantSample(Sample,TrueCombinedWeight );
                  oHistogramBias[FileCounter]->AddImportantSample(
		        Sample, HeuristicBalanceWeight * LikelihoodRatio );		  
		        //Sample, TrueCombinedWeight );		  
               }  
            }
         //}   
      }   
      SampFile->close();  
      SampHeuristicBalanceFile->close();       
      oHistogramBias[FileCounter]->Obtain_pdf();      
      FileCounter++;
   }   
   InFile->close();
   Obtain_pdf();
   
   //## It is time to get the variance of the estimator in order to 
   //## be able to create a file containing the confidence interval

   for (int ii = 0; ii < qtBins; ii++) {
      double VarianceBin = 0;
      double Variance_cdfBin = 0;
      double Variance_cdfcBin = 0;      
      for (int jj = 0; jj < FileCounter; jj++) {
         VarianceBin += sq(oHistogramBias[jj]->Get_pdfConfidence(ii));
         Variance_cdfBin += sq(oHistogramBias[jj]->Get_cdfConfidence(ii));	 
         Variance_cdfcBin += sq(oHistogramBias[jj]->Get_cdfcConfidence(ii));	 
      } 
      //## The standard deviation (confidence) is already normalized      
      pdfConfidence[ii] = sqrt(VarianceBin);
      cdfConfidence[ii] = sqrt(Variance_cdfBin);      
      cdfcConfidence[ii] = sqrt(Variance_cdfcBin);      
   }
   //## Prevent an incorrect computation of the variance in Obtain_pdf()
   PreviouslyComputed_pdfConfidenceFlag = 1;

   for (int ii = 0; ii < FileCounter; ii++) {
      delete oHistogramBias[ii];
   }   
   //WriteFileGaussian_pdf("Gaussian_pdf.dat",1000);   
} //## End MultipleHeuristicBalanceImport(.)

// #################################################################

/*! 
     Computes the probability density function 
     f(X=x| z_min < Z < z_max), 
     Samples of the random variables X,Z have been obtained from an 
     importance sampling simulation, and that the names of the files containing
     the samples are stored in the file InFileName. 
   */

void Histogram::MergeMultipleConditionalHeuristicBalance(string InFileName, 
					  double z_min, double z_max,
                                          int Convert2dB_Scale)
{
   //## Structure of InFileName
   //##  SampDataFileName

   //## Clear the histogram, just in case...
   ClearHistogram();
   
   ifstream * InFile;

   InFile = new ifstream;

   InFile->open(InFileName.c_str(),ios::in);

   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }  

   double Sample, LikelihoodRatio;
   double SampleZ, LikelihoodRatioZ;

   string SampFileName;
   string SampZFileName; // For Constraint pdf
   string SampHeuristicBalanceWeightFileName;
   
   //## First, count the number of samples per ensemble
   ifstream * InFileWithImpSampWeights;
   InFileWithImpSampWeights = new ifstream;
    
   #define MAX_FILES_HEURISTIC_BALANCE 100
   double qtSamplesFile[MAX_FILES_HEURISTIC_BALANCE];
   Histogram *oHistogramBias[MAX_FILES_HEURISTIC_BALANCE];
   double qtSamplesTotal = 0;
   int FileCounter = 0;

   while ( *InFile >> SampFileName) {
      *InFile >> SampZFileName;
      *InFile >> SampHeuristicBalanceWeightFileName;              
       
      InFileWithImpSampWeights->open(SampFileName.c_str(),ios::in);
   				     
      if (! InFileWithImpSampWeights ) {
         continue;
      }	      

      InFileWithImpSampWeights->clear();
      InFileWithImpSampWeights->seekg(0);

      double UselessValue, LikelihoodRatioFile;
      qtSamplesFile[FileCounter] = 0;    
     
      while (*InFileWithImpSampWeights >> UselessValue ) {         
         *InFileWithImpSampWeights >> LikelihoodRatioFile; 
         qtSamplesFile[FileCounter]++;   
      }	 
      qtSamplesTotal += qtSamplesFile[FileCounter];
      cout << SampFileName 	 
	   << "\t"<< qtSamplesFile[FileCounter]
           << endl;  

      InFileWithImpSampWeights->close();      

      oHistogramBias[FileCounter] = new Histogram(qtBins, MinRange, MaxRange);         
      FileCounter++;     
   } // end while   

   //## Do the actual processing
   //InFile.seekg(0); //## Resets the file pointer to the start of the file.
   InFile->close();
   InFile->open(InFileName.c_str(),ios::in);
   
   InFile->clear();
   InFile->seekg(0);
   
   FileCounter = 0;

   ifstream * SampFile;
   SampFile = new ifstream;

   ifstream * SampZFile;
   SampZFile = new ifstream;

   ifstream * SampHeuristicBalanceFile;
   SampHeuristicBalanceFile = new ifstream;

   while ( *InFile >> SampFileName) {
      *InFile >> SampZFileName;
      *InFile >> SampHeuristicBalanceWeightFileName;

      SampFile->open(SampFileName.c_str(),ios::in);

      if ( ! SampFile) {
         cout << endl;
         cout << "Warning: " << SampFileName << " could not be open."
              << endl;
         continue;	      
         exit(1);      
      }       

     SampFile->clear();
     SampFile->seekg(0);

     SampZFile->open(SampZFileName.c_str(),ios::in);

      if ( ! SampZFile) {
         cout << endl;
         cout << "Warning: " << SampZFileName << " could not be open."
              << endl;
         continue;
         exit(1);      
      }

     SampZFile->clear();
     SampZFile->seekg(0);

     SampHeuristicBalanceFile->open(SampHeuristicBalanceWeightFileName.c_str(),
				    ios::in);

      if ( ! SampHeuristicBalanceFile) {
         cout << endl;
         cout << "Error: " << SampHeuristicBalanceWeightFileName 
	      << " could not be open."
              << endl;	      
         exit(1);      
      }   

     SampHeuristicBalanceFile->clear();
     SampHeuristicBalanceFile->seekg(0);

      cout << SampFileName << "   " 
	   << SampZFileName << "   "
           << SampHeuristicBalanceWeightFileName << endl;
      double LikelihoodRatioHeuristicBalance;
      double HeuristicBalanceWeight;
      double TrueCombinedWeight;
      while ( *SampFile >> Sample) {
         *SampFile >> LikelihoodRatio;
         *SampZFile >> SampleZ;
         *SampZFile >> LikelihoodRatioZ;
	 *SampHeuristicBalanceFile >> LikelihoodRatioHeuristicBalance;
	 *SampHeuristicBalanceFile >> HeuristicBalanceWeight;	 
	 *SampHeuristicBalanceFile >> TrueCombinedWeight;
       
         if (  LikelihoodRatio !=  LikelihoodRatioZ || 
		abs( (LikelihoodRatio - LikelihoodRatioHeuristicBalance)
	      /LikelihoodRatio) > 1e-4 ) 
           {   	 
	    cout << "Error: the LikelihoodRatio of the sample and " << endl
	         << "the one reported in the heuristic balance do not match."
		 << endl 
		 << "OR LikelihoodRatio != LikelihoodRatioZ " << endl
                 << "Check " << InFileName << endl
	         << "Aborting now ..." << endl;
             
            exit(1);		 	   
	   }
            
         if (LikelihoodRatio > 0 && SampleZ > z_min && SampleZ < z_max) 
             {
               if (Convert2dB_Scale) {
                  AddImportantSample( (Convert2dB_Scale)
                                     *Linear2dB(Sample),
		                     TrueCombinedWeight);
                  oHistogramBias[FileCounter]->AddImportantSample( 
		                     (Convert2dB_Scale)
                                     *Linear2dB(Sample),
	   		HeuristicBalanceWeight * LikelihoodRatio);				     
               } else {
                  AddImportantSample(Sample,TrueCombinedWeight );
                  oHistogramBias[FileCounter]->AddImportantSample(
		        Sample, HeuristicBalanceWeight * LikelihoodRatio );		  
               }  
            }
      }
   
      SampFile->close();  
      SampZFile->close();
      SampHeuristicBalanceFile->close();       
      oHistogramBias[FileCounter]->Obtain_pdf();      
      FileCounter++;
   }   
   InFile->close();
   Obtain_pdf();
   
   //## It is time to get the variance of the estimator in order to 
   //## be able to create a file containing the confidence interval

   for (int ii = 0; ii < qtBins; ii++) {
      double VarianceBin = 0;
      double Variance_cdfBin = 0;
      double Variance_cdfcBin = 0;      
      for (int jj = 0; jj < FileCounter; jj++) {
         VarianceBin += sq(oHistogramBias[jj]->Get_pdfConfidence(ii));
         Variance_cdfBin += sq(oHistogramBias[jj]->Get_cdfConfidence(ii));	 
         Variance_cdfcBin += sq(oHistogramBias[jj]->Get_cdfcConfidence(ii));	 
      } 
      //## The standard deviation (confidence) is already normalized      
      pdfConfidence[ii] = sqrt(VarianceBin);
      cdfConfidence[ii] = sqrt(Variance_cdfBin);      
      cdfcConfidence[ii] = sqrt(Variance_cdfcBin);      
   }
   //## Prevent an incorrect computation of the variance in Obtain_pdf()
   PreviouslyComputed_pdfConfidenceFlag = 1;

   for (int ii = 0; ii < FileCounter; ii++) {
      delete oHistogramBias[ii];
   }   

} //## End MultipleHeuristicBalanceImport(.)

// #################################################################


/*! 
     Computes the probability density function 
     f(X=x| z_min < Z < z_max, w_min < W < w_max), 
     Samples of the random variables X,Z,W have been obtained from an 
     importance sampling simulation, and that the names of the files containing
     the samples are stored in the file InFileName. 
   */

void Histogram::MergeMultipleConditionalHeuristicBalance(string InFileName,
                                          double z_min, double z_max,
                                          double w_min, double w_max,
                                          int Convert2dB_Scale)
{
   //## Structure of InFileName
   //##  SampDataFileName

   //## Clear the histogram, just in case...
   ClearHistogram();
  
   ifstream * InFile;

   InFile = new ifstream;

   InFile->open(InFileName.c_str(),ios::in);

   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   } 

   double Sample, LikelihoodRatio;
   double SampleZ, LikelihoodRatioZ;
   double SampleW, LikelihoodRatioW;

   string SampFileName;
   string SampZFileName; // For 1st Constraint pdf
   string SampWFileName; // For 2nd Constraint pdf
   string SampHeuristicBalanceWeightFileName;

   //## First, count the number of samples per ensemble
   ifstream * InFileWithImpSampWeights;
   InFileWithImpSampWeights = new ifstream;

   #define MAX_FILES_HEURISTIC_BALANCE 100
   double qtSamplesFile[MAX_FILES_HEURISTIC_BALANCE];
   Histogram *oHistogramBias[MAX_FILES_HEURISTIC_BALANCE];
   double qtSamplesTotal = 0;
   int FileCounter = 0;


   while ( *InFile >> SampFileName) {
      *InFile >> SampZFileName;
      *InFile >> SampWFileName;
      *InFile >> SampHeuristicBalanceWeightFileName;

      InFileWithImpSampWeights->open(SampFileName.c_str(),ios::in);

      if (! InFileWithImpSampWeights ) {
         continue;
      }     

      InFileWithImpSampWeights->clear();
      InFileWithImpSampWeights->seekg(0);

      double UselessValue, LikelihoodRatioFile;
      qtSamplesFile[FileCounter] = 0;

      while (*InFileWithImpSampWeights >> UselessValue ) {
         *InFileWithImpSampWeights >> LikelihoodRatioFile;
         qtSamplesFile[FileCounter]++;
      } 
      qtSamplesTotal += qtSamplesFile[FileCounter];
      cout << SampFileName
           << "\t"<< qtSamplesFile[FileCounter]
           << endl;

      InFileWithImpSampWeights->close();

      oHistogramBias[FileCounter] = new Histogram(qtBins, MinRange, MaxRange);    
      FileCounter++;
   } // end while

   //## Do the actual processing
   //InFile.seekg(0); //## Resets the file pointer to the start of the file.
   InFile->close();
   InFile->open(InFileName.c_str(),ios::in);
   InFile->clear();
   InFile->seekg(0);

   FileCounter = 0;

   ifstream * SampFile;
   SampFile = new ifstream;

   ifstream * SampZFile;
   SampZFile = new ifstream;

   ifstream * SampWFile;
   SampWFile = new ifstream;

   ifstream * SampHeuristicBalanceFile;
   SampHeuristicBalanceFile = new ifstream;

   while ( *InFile >> SampFileName) {
      *InFile >> SampZFileName;
      *InFile >> SampWFileName;
      *InFile >> SampHeuristicBalanceWeightFileName;

      SampFile->open(SampFileName.c_str(),ios::in);

      if ( ! SampFile) {
         cout << endl;
         cout << "Warning: " << SampFileName << " could not be open."
              << endl;
         continue;
         exit(1);
      }     

     SampFile->clear();
     SampFile->seekg(0);

     SampZFile->open(SampZFileName.c_str(),ios::in);

      if ( ! SampZFile) {
         cout << endl;
         cout << "Warning: " << SampZFileName << " could not be open."
              << endl;
         continue;
         exit(1);
      }

     SampZFile->clear();
     SampZFile->seekg(0);

     SampWFile->open(SampWFileName.c_str(),ios::in);

      if ( ! SampWFile) {
         cout << endl;
         cout << "Warning: " << SampWFileName << " could not be open."
              << endl;
         continue;
         exit(1);
      }

     SampWFile->clear();
     SampWFile->seekg(0);

     SampHeuristicBalanceFile->open(SampHeuristicBalanceWeightFileName.c_str(),
                                    ios::in);

      if ( ! SampHeuristicBalanceFile) {
         cout << endl;
         cout << "Error: " << SampHeuristicBalanceWeightFileName
              << " could not be open."
              << endl;
         exit(1);
      }  

     SampHeuristicBalanceFile->clear();
     SampHeuristicBalanceFile->seekg(0);

      cout << SampFileName << "   "
           << SampZFileName << "   "
           << SampWFileName << "   "
           << SampHeuristicBalanceWeightFileName << endl;
      double LikelihoodRatioHeuristicBalance;
      double HeuristicBalanceWeight;
      double TrueCombinedWeight;
      while ( *SampFile >> Sample) {
         *SampFile >> LikelihoodRatio;
         *SampZFile >> SampleZ;
         *SampZFile >> LikelihoodRatioZ;
         *SampWFile >> SampleW;
         *SampWFile >> LikelihoodRatioW;
         *SampHeuristicBalanceFile >> LikelihoodRatioHeuristicBalance;
         *SampHeuristicBalanceFile >> HeuristicBalanceWeight;
         *SampHeuristicBalanceFile >> TrueCombinedWeight;

         if (  LikelihoodRatio !=  LikelihoodRatioZ ||
	       LikelihoodRatio !=  LikelihoodRatioW ||
                abs( (LikelihoodRatio - LikelihoodRatioHeuristicBalance)
              /LikelihoodRatio) > 1e-4 )
           {
            cout << "Error: the LikelihoodRatio of the sample and " << endl
                 << "the one reported in the heuristic balance do not match."
                 << endl
                 << "OR LikelihoodRatio != LikelihoodRatioZ " << endl
                 << "Check " << InFileName << endl
                 << "Aborting now ..." << endl;

            cout << "LikelihoodRatio = " << LikelihoodRatio << endl
		 << "LikelihoodRatioZ = " << LikelihoodRatioZ << endl
                 << "LikelihoodRatioW = " << LikelihoodRatioW << endl
                 << "LikelihoodRatioHeuristicBalance = " 
	         << LikelihoodRatioHeuristicBalance << endl;


            exit(1);
           }

         if (LikelihoodRatio > 0 && SampleZ > z_min && SampleZ < z_max 
				 && SampleW > w_min && SampleW < w_max )
             {
               if (Convert2dB_Scale) {
                  AddImportantSample( (Convert2dB_Scale)
                                     *Linear2dB(Sample),
                                     TrueCombinedWeight);
                  oHistogramBias[FileCounter]->AddImportantSample(
                                     (Convert2dB_Scale)
                                     *Linear2dB(Sample),
                        HeuristicBalanceWeight * LikelihoodRatio);                
               } else {
                  AddImportantSample(Sample,TrueCombinedWeight );
                  oHistogramBias[FileCounter]->AddImportantSample(
                        Sample, HeuristicBalanceWeight * LikelihoodRatio );       
               }
            }
      }

      SampFile->close();
      SampZFile->close();
      SampWFile->close();
      SampHeuristicBalanceFile->close();
      oHistogramBias[FileCounter]->Obtain_pdf();
      FileCounter++;
   }  
   InFile->close();
   Obtain_pdf();

   //## It is time to get the variance of the estimator in order to
   //## be able to create a file containing the confidence interval

   for (int ii = 0; ii < qtBins; ii++) {
      double VarianceBin = 0;
      double Variance_cdfBin = 0;
      double Variance_cdfcBin = 0;
      for (int jj = 0; jj < FileCounter; jj++) {
         VarianceBin += sq(oHistogramBias[jj]->Get_pdfConfidence(ii));
         Variance_cdfBin += sq(oHistogramBias[jj]->Get_cdfConfidence(ii));       
         Variance_cdfcBin += sq(oHistogramBias[jj]->Get_cdfcConfidence(ii));     
      }
      //## The standard deviation (confidence) is already normalized
      pdfConfidence[ii] = sqrt(VarianceBin);
      cdfConfidence[ii] = sqrt(Variance_cdfBin);
      cdfcConfidence[ii] = sqrt(Variance_cdfcBin);
   }
   //## Prevent an incorrect computation of the variance in Obtain_pdf()
   PreviouslyComputed_pdfConfidenceFlag = 1;

   for (int ii = 0; ii < FileCounter; ii++) {
      delete oHistogramBias[ii];
   }
} //## End MultipleHeuristicBalanceImport(.)


// ###################################################################

/**************************
// This was the code used until 1/15/03
//#################################################################
//##### Imports the data from files created by the 
//##    histogram class applied to PMD.
//##    The heuristic balance weights need to be previously generated 
//##    A description of this method can be found in 
//##    Dr. Eric Veach's Ph.D. thesis
//##    This is, absolutely, the right way to merge 
//##    samples obtained from multiple bias.
//## By: Ivan Lima   07/27/02
void Histogram::MultipleHeuristicBalanceImport(string InFileName, 
                                               int Convert2dB_Scale)
{
   //## Structure of InFileName
   //##  SampDataFileName

   //## Clear the histogram, just in case...
   ClearHistogram();
   
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }  
   double Sample, LikelihoodRatio;
   //double MinRangeTmp, MaxRangeTmp;
   string SampFileName;
   string SampHeuristicBalanceWeightFileName;
   
   
   //## First, count the number of samples per ensemble
   ifstream InFileWithImpSampWeights("",ios::in);    
   #define MAX_FILES_HEURISTIC_BALANCE 100
   double qtSamples[MAX_FILES_HEURISTIC_BALANCE];
   Histogram *oHistogramBias[MAX_FILES_HEURISTIC_BALANCE];
   double qtSamplesTotal = 0;
   int FileCounter = 0;
   while ( InFile >> SampFileName) {
      InFile >> SampHeuristicBalanceWeightFileName;              

      InFileWithImpSampWeights.open(SampFileName.c_str(),
                                     ios::in);   				     
      if (! InFileWithImpSampWeights ) {
         continue;
      }	      

      double UselessValue, LikelihoodRatioFile;
      qtSamples[FileCounter] = 0;         
      while (InFileWithImpSampWeights >> UselessValue ) {         
         InFileWithImpSampWeights >> LikelihoodRatioFile; 
         qtSamples[FileCounter]++;   
      }	 
      qtSamplesTotal += qtSamples[FileCounter];
      cout << SampFileName 	 
	   << "\t"<< qtSamples[FileCounter]
           << endl;  
      InFileWithImpSampWeights.close();      
      oHistogramBias[FileCounter] = new Histogram(qtBins, MinRange, MaxRange);         
      FileCounter++;     
   }   

   //## Do the actual processing
   //InFile.seekg(0); //## Resets the file pointer to the start of the file.
   InFile.close();
   InFile.open(InFileName.c_str(),ios::in);
   
   FileCounter = 0;
   while ( InFile >> SampFileName) {
      InFile >> SampHeuristicBalanceWeightFileName;
      //cout << SampFileName << endl;
      //InFile >> MinRangeTmp;
      //InFile >> MaxRangeTmp;
      ifstream SampFile(SampFileName.c_str(),ios::in);
      if ( ! SampFile) {
         cout << endl;
         cout << "Warning: " << SampFileName << " could not be open."
              << endl;
         continue;	      
         //exit(1);      
      }       
       
      ifstream SampHeuristicBalanceFile(
            SampHeuristicBalanceWeightFileName.c_str(),ios::in);
      if ( ! SampHeuristicBalanceFile) {
         cout << endl;
         cout << "Error: " << SampHeuristicBalanceWeightFileName 
	      << " could not be open."
              << endl;	      
         exit(1);      
      }   
      cout << SampFileName << "   " 
           << SampHeuristicBalanceWeightFileName << endl;
      double LikelihoodRatioHeuristicBalance;
      double HeuristicBalanceWeight;
      double TrueCombinedWeight;
      while ( SampFile >> Sample) {
         SampFile >> LikelihoodRatio;
	 SampHeuristicBalanceFile >> LikelihoodRatioHeuristicBalance;
	 SampHeuristicBalanceFile >> HeuristicBalanceWeight;	 
	 SampHeuristicBalanceFile >> TrueCombinedWeight;
         if ( abs( (LikelihoodRatio - LikelihoodRatioHeuristicBalance)
	      /LikelihoodRatio) > 1e-4 ) {   	 
	 //if (LikelihoodRatio != LikelihoodRatioHeuristicBalance ) {
	    cout << "Error: the LikelihoodRatio of the sample and " << endl
	         << "the one reported in the heuristic balance do not match."
		 << endl << "Check " << InFileName << endl
		 << "Aborting now ..." << endl;
            exit(1);		 	   
	 }
         //if (Sample >= MinRangeTmp && Sample < MaxRangeTmp) {
            if (LikelihoodRatio > 0) {
               //## Samples can be filtered by the user 
               //#  doing the LikelihoodRatio < 0
               //AddImportantSample(Sample,LikelihoodRatio);          

  
               if (Convert2dB_Scale) {
                  AddImportantSample( sgn(Convert2dB_Scale)
                                     *Linear2dB(Sample),
				     TrueCombinedWeight);
                  oHistogramBias[FileCounter]->AddImportantSample( 
		                     sgn(Convert2dB_Scale)
                                     *Linear2dB(Sample),
				     TrueCombinedWeight);				     
		   //		     LikelihoodRatio);				     
               }
               else {
                  AddImportantSample(Sample,TrueCombinedWeight );
                  oHistogramBias[FileCounter]->AddImportantSample(
		        Sample, TrueCombinedWeight );		  
		  //      Sample, LikelihoodRatio );
               }  
            }
         //}   
      }   
      SampFile.close();                        
      SampHeuristicBalanceFile.close();       
      oHistogramBias[FileCounter]->Obtain_pdf();      
      FileCounter++;
   }   
   InFile.close();
   
   //## It is time to get the variance of the estimator in order to 
   //## be able to create a file containing the confidence interval

   for (int ii = 0; ii < qtBins; ii++) {
      double VarianceBin = 0;
      for (int jj = 0; jj < FileCounter; jj++) {
         VarianceBin += sq(oHistogramBias[jj]->Get_pdfConfidence(ii));
      } 
      pdfConfidence[ii] = sqrt(VarianceBin);
   }
   //## Prevent an incorrect computation of the variance in Obtain_pdf()
   PreviouslyComputed_pdfConfidenceFlag = 1;

   for (int ii = 0; ii < FileCounter; ii++) {
      delete oHistogramBias[ii];
   }      
} //## End MultipleHeuristicBalanceImport(.)

***************/





/*
//#############################################################
//##  This function should be run after the initialization
//#   to initialize the database that is required to apply
//#   the balance heuristic. 
//## By: Ivan Lima (01/04/02)
void Histogram::InitBalanceHeuristicDataStorage(double BiasingParam2,
      int qtRamdomSourcesPerSample2) 
{
   BiasingParam = BiasingParam2;
   qtRandomSourcesPerSample = qtRamdomSourcesPerSample2;
   BalanceHeuristicInitialized = 1;   
   
   string FileNameWithBalanceHeuristicData = FileNameWithSamples 
         + "BalanceHeuristic.dat";

}
*/


/*
//##### Imports the data from files created by the 
//##    histogram class applied to PMD. 
//##    The bias is described in thh OptFiber class.
//##    This method merges multiple sets of files weighting each 
//##    sample according to the balance-heuristic method.
//##    A description of this method can be found in 
//##    Dr. Eric Veach's Ph.D. thesis
//##    This is, absolutely, the right way to merge 
//##    samples obtained from multiple bias.
//##    One big problem is that we have to record the 
//##    biased parameters to be able to compute the 
//##    likelihood ratio for all the biased distributions
//##    for all the samples. That complicate things a lot.
//## By: Ivan Lima   08/21/01
void Histogram::MultipleBalanceHeuristicImport(string InFileName)
{
   //## Structure of InFileName
   //##  SampDataFileName    

   cout << "MultipleBalanceHeuristicImport(string InFileName)" 
        << endl << "Not available."<< endl;
   exit(1);     
         
   //## Clear the histogram, just in case...
   ClearHistogram();
   
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }     
   
   //## It can merge up to 100 biased files. You are welcome
   //#  to increase this number, if you feel you need to.
   string SampFileName[100];
   int NumSamplesFile[100];
   
   string SampFileNameBiasParam;
   //Histogram* oHist_pdf[100];  

   //## The first thing to do is to get the file names
   //#  and to count the total number of samples      
   //#  The files with the biased parameters must 
   //#  be available in the same directory with 
   //#  extension ".BiasParam" beyond the name of 
   //#  the file with the samples.
      
   *LogFile << "List of file names to be merged:" << endl;
   int FileCounter = 0;
   int TotNumSamples = 0;
   double Sample, LikelihoodRatio;      
   while (InFile >> SampFileName[FileCounter]) {
      *LogFile << "   "<< SampFileName[FileCounter] << endl;       
    
      ifstream SampFile(SampFileName[FileCounter].c_str(),ios::in);  
      if(SampFile) {          
         FileCounter++;    
         NumSamplesFile[FileCounter] = 0;
         while ( SampFile >> Sample) {
            SampFile >> LikelihoodRatio;
            if (LikelihoodRatio > 0) {
               //## Samples can be filtered by the user 
               //#  doing the LikelihoodRatio < 0
               NumSamplesFile[FileCounter]++;          
            }
         }                
         TotNumSamples += NumSamplesFile[FileCounter];
         SampFile.close();         
      }
      else if ( ! SampFile) {
         cout << endl;
         cout << "Warning: " << SampFileName 
              << " could not be open." << endl;
         //exit(1);      
      }       

   }
   InFile.close();


   //## We already have the number of samples per 
   //#  file and, consequently, the total number of samples.
   //#  This module is unfinished!!! I should continue from here.

   
   for (int ii = 0; ii < FileCounter; ii++) {
      //oHist_pdf[ii] = new Histogram(HistInFileName.c_str());
      //cout << "File Name: "<< ii <<" = " <<SampFileName[ii] << endl;
      ifstream SampFile(SampFileName[ii].c_str(),ios::in);      
      if ( ! SampFile) {
         cout << endl;
         cout << "Warning: " << SampFileName << " could not be open."
              << endl;
         //exit(1);      
      }             
      while ( SampFile >> Sample) {
         SampFile >> LikelihoodRatio;
         if (LikelihoodRatio > 0) {
            //## Samples can be filtered by the user 
            //#  doing the LikelihoodRatio < 0
            //oHist_pdf[ii]->AddImportantSample(Sample,LikelihoodRatio);          
         }
      }          
      //oHist_pdf[ii]->Obtain_pdf();       
      SampFile.close();
   } 
   
   //## Second, get the total number of samples per bin
   long *qtSamplesBinTmp = new long[qtBins];
   for (int jj = 0; jj < qtBins; jj++) {   
      qtSamplesBinTmp[jj] = 0;
   }      
   for (int ii = 0; ii < FileCounter; ii++) {
      for (int jj = 0; jj < qtBins; jj++) {   
         //qtSamplesBinTmp[jj] += oHist_pdf[ii]->GetQtSamplesBin(jj);
      }
   }   
   
   //## Third, merge the pdfs weighting
   //#  each bin by the respective number of samples 
   //#  in each biased distribution.
   //#  I have to do it in such a way that Obtain_pdf
   //#  won't mess things up.
   for (int ii = 0; ii < FileCounter; ii++) {
      for (int jj = 0; jj < qtBins; jj++) {
         if (qtSamplesBinTmp[jj] != 0) {
            //Bin[jj] += oHist_pdf[ii]->Get_pdfY(jj)
            //   *oHist_pdf[ii]->GetQtSamplesBin(jj)
            //   /qtSamplesBinTmp[jj];
         }      
      }   
   }   

   //## Debug lines
   //for (int jj = 0; jj < qtBins; jj++) {
   //   cout << jj<< " " << Bin[jj] << endl;
   //}      

   //## Fourth, and finally, force the number of samples to be 
   //#  unitary to get the right statistics, since all the 
   //#  pdfs used were already normalized.
   //#  Obviously, I need to denormilize the Bins, since the
   //#  Obtain_pdf method will do it
   SumUnitaryLikelihoodRatio = SumLikelihoodRatio = qtSamples = 1;
   qtSamplesInRange = 1;    
   for (int jj = 0; jj < qtBins; jj++) {
      Bin[jj] *= (double) qtSamples/qtBins*(MaxRange - MinRange);  
      //cout <<  (double) qtSamples/qtBins*(MaxRange - MinRange) << endl; 
      //cout << qtBins << "  " << qtSamples << " " << MaxRange - MinRange <<endl;      
      
   }   
 
   //cout << "Now the final pdf will be obtained :" << endl;     
   //cout << qtBins << "  " << qtSamples << endl;
   
   Obtain_pdf();

   *LogFile << "The pdf from "<< InFileName.c_str()<<" was obtained "
           << "from a merging process." << endl;
   *LogFile << "Thus, don't expect fully functionality from the methods: "
           << "GetMean(), GetMax(), GetQtSamples(), etc." << endl;
   
} //## End MultipleBalanceHeuristicImport(.)
*/


// #########################################################

//##### Imports the data from files created by the histogram class
//##    and merge multiple sets of files weighting each bin
//##    from each of the distributions
//##    by the number of samples in it. 
//##    This method was suggested by Biondinin and Kath 
//##    from Northwestern University.
//##    This is not the optimum. The optimum is the 
//##    Balance-Heuristic. But the latter requires 
//##    the set of parameters to calculate the likelihood 
//##    ratio for all the biased distributions for each 
//##    of the samples. 
//## By: Ivan Lima   06/27/01

void Histogram::MultipleWeightedImport(string InFileName)
{
   //## Structure of InFileName
   //##  SampDataFileName    
         
   //## Clear the histogram, just in case...
   ClearHistogram();
   
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }     
   
   //## It can merge up to 100 biased files
   string SampFileName[100];
   Histogram* oHist_pdf[100];  
      
   *LogFile << "List of file names to be merged:" << endl;
   int FileCounter = 0;
   while (InFile >> SampFileName[FileCounter]) {
      *LogFile << "   "<< SampFileName[FileCounter] << endl;       

      FileCounter++;
   }
   InFile.close();

   //## First, get the pdf for each case
   double Sample, LikelihoodRatio;      
   for (int ii = 0; ii < FileCounter; ii++) {
      oHist_pdf[ii] = new Histogram(HistInFileName.c_str());
      cout << "File Name: "<< ii <<" = " <<SampFileName[ii] << endl;
      ifstream SampFile(SampFileName[ii].c_str(),ios::in);      
      if ( ! SampFile) {
         cout << endl;
         cout << "Warning: " << SampFileName << " could not be open."
              << endl;
         //exit(1);      
      }             
      while ( SampFile >> Sample) {
         SampFile >> LikelihoodRatio;
         if (LikelihoodRatio > 0) {
            //## Samples can be filtered by the user 
            //#  doing the LikelihoodRatio < 0
            oHist_pdf[ii]->AddImportantSample(Sample,LikelihoodRatio);          
         }
      }          
      oHist_pdf[ii]->Obtain_pdf();       
      SampFile.close();
   } 
   
   //## Second, get the total number of samples per bin
   long *qtSamplesBinTmp = new long[qtBins];
   for (int jj = 0; jj < qtBins; jj++) {   
      qtSamplesBinTmp[jj] = 0;
   }      
   for (int ii = 0; ii < FileCounter; ii++) {
      for (int jj = 0; jj < qtBins; jj++) {   
         qtSamplesBinTmp[jj] += oHist_pdf[ii]->GetQtSamplesBin(jj);
      }
   }   
   
   //## Third, merge the pdfs weighting
   //#  each bin by the respective number of samples 
   //#  in each biased distribution.
   //#  I have to do it in such a way that Obtain_pdf
   //#  won't mess things up.
   for (int ii = 0; ii < FileCounter; ii++) {
      for (int jj = 0; jj < qtBins; jj++) {
         if (qtSamplesBinTmp[jj] != 0) {
            Bin[jj] += oHist_pdf[ii]->Get_pdfY(jj)
               *oHist_pdf[ii]->GetQtSamplesBin(jj)
               /qtSamplesBinTmp[jj];
         }      
      }   
   }   

   //## Debug lines
   //for (int jj = 0; jj < qtBins; jj++) {
   //   cout << jj<< " " << Bin[jj] << endl;
   //}      

   //## Fourth, and finally, force the number of samples to be 
   //#  unitary to get the right statistics, since all the 
   //#  pdfs used were already normalized.
   //#  Obviously, I need to denormilize the Bins, since the
   //#  Obtain_pdf method will do it
   SumUnitaryLikelihoodRatio = SumLikelihoodRatio = qtSamples = 1;
   qtSamplesInRange = 1;    
   for (int jj = 0; jj < qtBins; jj++) {
      Bin[jj] *= (double) qtSamples/qtBins*(MaxRange - MinRange);  
      //cout <<  (double) qtSamples/qtBins*(MaxRange - MinRange) << endl; 
      //cout << qtBins << "  " << qtSamples << " " << MaxRange - MinRange <<endl;      
      
   }   
 
   //cout << "Now the final pdf will be obtained :" << endl;     
   //cout << qtBins << "  " << qtSamples << endl;
   
   Obtain_pdf();

   *LogFile << "The pdf from "<< InFileName.c_str()<<" was obtained "
           << "from a merging process." << endl;
   *LogFile << "Thus, don't expect fully functionality from the methods: "
           << "GetMean(), GetMax(), GetQtSamples(), etc." << endl;
   
} //## End MultipleWeightedImport(.)




//########################################################
//##### Imports the data from files created by the histogram class
//##    Convert2dB_Scale also determines the sign of the conversion to dB
//## By: Ivan Lima   09/22/01

void Histogram::MultipleSelectMaxRelNumHitsImport(string InFileName,
                                                  int Convert2dB_Scale)
{
   //## Structure of InFileName
   //##  SampDataFileName    
         
   //## Clear the histogram, just in case...

   ClearHistogram();
   
   ifstream InFile(InFileName.c_str(),ios::in);

   if ( ! InFile) 
    {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }     
   
   //## It can merge up to 100 biased files

   string SampFileName[100];
   Histogram* oHist_pdf[100];  
      
   *LogFile << "List of file names to be merged:" << endl;

   int FileCounter = 0;

   while (InFile >> SampFileName[FileCounter]) 
    {
      *LogFile << "   "<< SampFileName[FileCounter] << endl;       

      FileCounter++;
   }

   InFile.close();

   //## First, get the pdf for each case

   double Sample, LikelihoodRatio;      

   for (int ii = 0; ii < FileCounter; ii++) 
    {
      oHist_pdf[ii] = new Histogram(HistInFileName.c_str());
    
      cout << "File Name: "<< ii <<" = " <<SampFileName[ii] << endl;

      ifstream SampFile(SampFileName[ii].c_str(),ios::in);      

      if ( ! SampFile) 
       {
         cout << endl;
         cout << "Warning: " << SampFileName << " could not be open."
              << endl;
         //exit(1);      
      }
             
      while ( SampFile >> Sample) 
       {
         SampFile >> LikelihoodRatio;
         if (LikelihoodRatio > 0) 
           {
            //## Samples can be filtered by the user 
            //#  doing the LikelihoodRatio < 0
           
             if (Convert2dB_Scale) 
               {
               oHist_pdf[ii]->AddImportantSample( (Convert2dB_Scale)
                                     *Linear2dB(Sample),LikelihoodRatio);
            }
            else 
             {
              oHist_pdf[ii]->AddImportantSample(Sample,LikelihoodRatio);
            }   
         }
      }      
    
      oHist_pdf[ii]->Obtain_pdf();       
      
      cout << "     qtSamples = " << oHist_pdf[ii]->GetQtSamples() << endl;
      
      SampFile.close();
   } 
   
//   //## Second, get the total number of samples per bin
//   long *qtSamplesBinTmp = new long[qtBins];
//   for (int jj = 0; jj < qtBins; jj++) {   
//      qtSamplesBinTmp[jj] = 0;
//   }      
//   for (int ii = 0; ii < FileCounter; ii++) {
//      for (int jj = 0; jj < qtBins; jj++) {   
//         qtSamplesBinTmp[jj] += oHist_pdf[ii]->GetQtSamplesBin(jj);
//      }
//   }   
   
   //## Third, select the bin that had the maximum relative
   //#  number of hits from all the distributions
   //#  I have to do it in such a way that Obtain_pdf
   //#  won't mess things up.
   //double pdfConfidenceTmp[qtBins];

   double MaxRelNumHits;

   for (int jj = 0; jj < qtBins; jj++) 
    {   
      MaxRelNumHits = 0;

      for (int ii = 0; ii < FileCounter; ii++) {
         if ( (double) oHist_pdf[ii]->GetQtSamplesBin(jj)
                      /oHist_pdf[ii]->GetQtSamples() 
             > MaxRelNumHits )  {
            MaxRelNumHits =  (double) 
                  oHist_pdf[ii]->GetQtSamplesBin(jj)
                  /oHist_pdf[ii]->GetQtSamples();              
		  
            Bin[jj] = oHist_pdf[ii]->GetBin(jj)/oHist_pdf[ii]->GetQtSamples();
            pdfConfidence[ii] = oHist_pdf[ii]->Get_pdfConfidence(ii);
            //cdfBin[jj] = oHist_pdf[ii]->Get_cdfBin(jj);
            //cdfBin[jj] = oHist_pdf[ii]->Get_cdfcBin(jj);	    
         }

         //if (oHist_pdf[ii]->Get_pdfY(jj) > Bin[jj]) {
         //   Bin[jj] = oHist_pdf[ii]->Get_pdfY(jj);
         //}         
//         if (qtSamplesBinTmp[jj] != 0) {
//            Bin[jj] += oHist_pdf[ii]->Get_pdfY(jj)
//               *oHist_pdf[ii]->GetQtSamplesBin(jj)
//               /qtSamplesBinTmp[jj];
//         }      

      }   
      
    } // ## end for jj loop over bins  
    //## Prevent from overriding the pdfConfidence given here
    PreviouslyComputed_pdfConfidenceFlag = 1;

   //## Debug lines
   //for (int jj = 0; jj < qtBins; jj++) {
   //   cout << jj<< " " << Bin[jj] << endl;
   //}      

   //## Fourth, and finally, force the number of samples to be 
   //#  unitary to get the right statistics, since all the 
   //#  pdfs used were already normalized.
   //#  Obviously, I need to denormilize the Bins, since the
   //#  Obtain_pdf method will do it

   SumUnitaryLikelihoodRatio = SumLikelihoodRatio = qtSamples = 1;

   qtSamplesInRange = 1;    

   /****
   for (int jj = 0; jj < qtBins; jj++) {
      Bin[jj] *= (double) qtSamples/qtBins*(MaxRange - MinRange);  
      //cout <<  (double) qtSamples/qtBins*(MaxRange - MinRange) << endl; 
      //cout << qtBins << "  " << qtSamples << " " 
      //<< MaxRange - MinRange <<endl;      
      
   }   
   ****/

   //## Now, compute the cdf and the cdfc
   cdfBin[0] = Bin[0];   
   for (int ii = 1; ii < qtBins; ii++) {
      cdfBin[ii] = cdfBin[ii - 1] + Bin[ii];
   }             
   cdfcBin[qtBins - 1] = Bin[qtBins - 1];   
   for (int ii = qtBins - 2; ii >= 0; ii--) {
      cdfcBin[ii] = cdfcBin[ii + 1] + Bin[ii];
   }        
 
   //cout << "Now the final pdf will be obtained :" << endl;     
   //cout << qtBins << "  " << qtSamples << endl;
   
   Obtain_pdf();

   *LogFile << "The pdf from "<< InFileName.c_str()<<" was obtained "
           << "from a merging process." << endl;
   *LogFile << "Thus, don't expect fully functionality from the methods: "
           << "GetMean(), GetMax(), GetQtSamples(), etc." << endl;
   
} //## End MultipleSelectMaxRelNumHitsImport(.)


// ###############################################################





//##### Imports the data from files created by the histogram class
//##    and merge multiple sets of files within the specified 
//##    range in the InFileName. The number of samples
//##   MUST be equal in all the files.
//##    Do not use this function!!! Unless the recomended 
//##    method does not work. This is not the optimum method.
//## By: Ivan Lima
void Histogram::MultipleSelectiveImport(string InFileName)
{
   //## Structure of InFileName
   //##  SampDataFileName    MinRangeToExtract  MaxRangeToExtract

   //## Clear the histogram, just in case...
   ClearHistogram();
   
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }  
   #define MAX_FILES_HEURISTIC_BALANCE 100
   double qtSamplesFile[MAX_FILES_HEURISTIC_BALANCE];
   Histogram *oHistogramBias[MAX_FILES_HEURISTIC_BALANCE];   
   double Sample, LikelihoodRatio;
   double MinRangeTmp, MaxRangeTmp;
   string SampFileName;
   int FileCounter = 0;
   while ( InFile >> SampFileName) {
      InFile >> MinRangeTmp;
      InFile >> MaxRangeTmp;
      ifstream SampFile(SampFileName.c_str(),ios::in);
      if ( ! SampFile) {
         cout << endl;
         cout << "Warning: " << SampFileName << " could not be open."
              << endl;
         //exit(1);      
	 continue;
      }       
      oHistogramBias[FileCounter] = new Histogram(qtBins, MinRange, MaxRange);                 
      while ( SampFile >> Sample) {
         SampFile >> LikelihoodRatio;
         if (Sample >= MinRangeTmp && Sample < MaxRangeTmp) {
            if (LikelihoodRatio > 0) {
               //## Samples can be filtered by the user 
               //#  doing the LikelihoodRatio < 0
               AddImportantSample(Sample,LikelihoodRatio);          
	       oHistogramBias[FileCounter]->AddImportantSample(Sample,
	             LikelihoodRatio);
	       
            }
         }   
      }   
      oHistogramBias[FileCounter]->Obtain_pdf(); 
      FileCounter++;
      SampFile.close();                        
   }   
   InFile.close();
   Obtain_pdf();
   
   //## It is time to get the variance of the estimator in order to 
   //## be able to create a file containing the confidence interval

   //## Note that the actual weights are "FileCounter" times larger 
   //## that they should be.   

   for (int ii = 0; ii < qtBins; ii++) {
      double VarianceBin = 0;
      for (int jj = 0; jj < FileCounter; jj++) {
         VarianceBin += sq(oHistogramBias[jj]->Get_pdfConfidence(ii));
      } 
      //## The standard deviation (confidence) is already normalized      
      //## The division by "FileCounter" is because of the lack
      //## of the normalization of the weights
      pdfConfidence[ii] = sqrt(VarianceBin)/FileCounter;
   }
   //## Prevent an incorrect computation of the variance in Obtain_pdf()
   PreviouslyComputed_pdfConfidenceFlag = 1;

   for (int ii = 0; ii < FileCounter; ii++) {
      delete oHistogramBias[ii];
   }
         
} //## End MultipleSelectiveImport(.)





/*************** Commented in 1/17/03
//##### Imports the data from files created by the histogram class
//##    and merge multiple sets of files within the specified 
//##    range in the InFileName. The number of samples
//##   MUST be equal in all the files.
//##    Do not use this function!!! Unless the recomended 
//##    method does not work. This is not the optimum method.
//## By: Ivan Lima
void Histogram::MultipleSelectiveImport(string InFileName)
{
   //## Structure of InFileName
   //##  SampDataFileName    MinRangeToExtract  MaxRangeToExtract

   //## Clear the histogram, just in case...
   ClearHistogram();
   
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }  
   double Sample, LikelihoodRatio;
   double MinRangeTmp, MaxRangeTmp;
   string SampFileName;
   while ( InFile >> SampFileName) {
      InFile >> MinRangeTmp;
      InFile >> MaxRangeTmp;
      ifstream SampFile(SampFileName.c_str(),ios::in);
      if ( ! SampFile) {
         cout << endl;
         cout << "Warning: " << SampFileName << " could not be open."
              << endl;
         //exit(1);      
       }       
      while ( SampFile >> Sample) {
         SampFile >> LikelihoodRatio;
         if (Sample >= MinRangeTmp && Sample < MaxRangeTmp) {
            if (LikelihoodRatio > 0) {
               //## Samples can be filtered by the user 
               //#  doing the LikelihoodRatio < 0
               AddImportantSample(Sample,LikelihoodRatio);          
            }
         }   
      }   
      SampFile.close();                        
   }   
   InFile.close();
} //## End MultipleSelectiveImport(.)
****************************/




//##### Imports the data from files created by the histogram class
//#     and merge multiple sets of files.
//#     This method is useful to merge all the samples of independent 
//#     unbiased simulations.
//## By: Ivan Lima (12/5/01)
void Histogram::MultipleImport(string InFileName, int Convert2dB_Scale)
{
   //## Structure of InFileName
   //##  SampDataFileName

   //## Clear the histogram, just in case...
   ClearHistogram();
   
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }  
   #define MAX_FILES_HEURISTIC_BALANCE 100
   double qtSamplesFile[MAX_FILES_HEURISTIC_BALANCE];
   Histogram *oHistogramBias[MAX_FILES_HEURISTIC_BALANCE]; 
   int FileCounter = 0;
   double Sample, LikelihoodRatio;
   //double MinRangeTmp, MaxRangeTmp;
   string SampFileName;
   while ( InFile >> SampFileName) {
      //InFile >> MinRangeTmp;
      //InFile >> MaxRangeTmp;
      ifstream SampFile(SampFileName.c_str(),ios::in);
      if ( ! SampFile) {
         cout << endl;
         cout << "Warning: " << SampFileName << " could not be open."
              << endl;
	 continue;     
         //exit(1);      
      }       
      oHistogramBias[FileCounter] = new Histogram(qtBins, MinRange, MaxRange);        
      while ( SampFile >> Sample) {
         SampFile >> LikelihoodRatio;
         //if (Sample >= MinRangeTmp && Sample < MaxRangeTmp) {
            if (LikelihoodRatio > 0) {
               //## Samples can be filtered by the user 
               //#  doing the LikelihoodRatio < 0
               //AddImportantSample(Sample,LikelihoodRatio);          
               if (Convert2dB_Scale) {
                  AddImportantSample( (Convert2dB_Scale)
                                   *Linear2dB(Sample),LikelihoodRatio);
   	          oHistogramBias[FileCounter]->AddImportantSample(
		        (Convert2dB_Scale)*Linear2dB(Sample),
	                LikelihoodRatio);				   				   
               } else {
                  AddImportantSample(Sample,LikelihoodRatio);		  
	          oHistogramBias[FileCounter]->AddImportantSample(Sample,
	                LikelihoodRatio);		  
               }     
            }
         //}   
      }   
      oHistogramBias[FileCounter]->Obtain_pdf(); 
      FileCounter++;
      SampFile.close();                        
   }   
   InFile.close();
   Obtain_pdf();

   
   //## It is time to get the variance of the estimator in order to 
   //## be able to create a file containing the confidence interval

   //## Note that the actual weights are "FileCounter" times larger 
   //## that they should be.   

   for (int ii = 0; ii < qtBins; ii++) {
      double VarianceBin = 0;
      for (int jj = 0; jj < FileCounter; jj++) {
         VarianceBin += sq(oHistogramBias[jj]->Get_pdfConfidence(ii));
      } 
      //## The standard deviation (confidence) is already normalized      
      //## The division by "FileCounter" is because of the lack
      //## of the normalization of the weights
      pdfConfidence[ii] = sqrt(VarianceBin)/FileCounter;
   }
   //## Prevent an incorrect computation of the variance in Obtain_pdf()
   PreviouslyComputed_pdfConfidenceFlag = 1;

   for (int ii = 0; ii < FileCounter; ii++) {
      delete oHistogramBias[ii];
   }
         
   
} //## End MultipleSelectiveImport(.)


// ####################################################################

/*! Imports the data from files created by the histogram class
    and merge multiple sets of files.
    This method is useful to merge all the samples of independent 
    unbiased simulations, i.e., we assume LikelihoodRatio=1.0

    By: Ivan Lima (12/5/01). Adapted by JZ Aug 2004
*/

void Histogram::MultipleUnbiasedImport(string InFileName)
{
   //## Structure of InFileName
   //##  SampDataFileName

   //## Clear the histogram, just in case...

   ClearHistogram();
   
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }  
   
   #define MAX_FILES 100

   double qtSamplesFile[MAX_FILES];

   Histogram *oHistogramUnbiased[MAX_FILES]; 

   int FileCounter = 0;

   double Sample, LikelihoodRatio;
  
   string SampFileName;
 
  while ( InFile >> SampFileName) 
    {
      ifstream SampFile(SampFileName.c_str(),ios::in);
      if ( ! SampFile) 
        {
         cout << endl;
         cout << "Warning: " << SampFileName << " could not be open."
              << endl;
	 //continue;     
         exit(1);      
      }       

      oHistogramUnbiased[FileCounter] = new Histogram(qtBins, 
						      MinRange, MaxRange); 
       
      while (SampFile >> Sample) 
       {
         SampFile >> LikelihoodRatio;
        
         AddImportantSample(Sample,1.0);		  
	 oHistogramUnbiased[FileCounter]->AddImportantSample(Sample,1.0);
		  
       }     
               
      oHistogramUnbiased[FileCounter]->Obtain_pdf(); 
      FileCounter++;
      SampFile.close();                        
    }
  
   InFile.close();
   Obtain_pdf();

   for (int ii = 0; ii < FileCounter; ii++) 
      delete oHistogramUnbiased[ii];
     
} //## End MultipleUnbiasedImport


// ####################################################################


/********* This module was used until 1/15/03
//##### Imports the data from files created by the histogram class
//#     and merge multiple sets of files.
//#     This method is useful to merge all the samples of independent 
//#     unbiased simulations.
//## By: Ivan Lima (12/5/01)
void Histogram::MultipleImport(string InFileName, int Convert2dB_Scale)
{
   //## Structure of InFileName
   //##  SampDataFileName

   //## Clear the histogram, just in case...
   ClearHistogram();
   
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }  
   double Sample, LikelihoodRatio;
   //double MinRangeTmp, MaxRangeTmp;
   string SampFileName;
   while ( InFile >> SampFileName) {
      //InFile >> MinRangeTmp;
      //InFile >> MaxRangeTmp;
      ifstream SampFile(SampFileName.c_str(),ios::in);
      if ( ! SampFile) {
         cout << endl;
         cout << "Warning: " << SampFileName << " could not be open."
              << endl;
         //exit(1);      
       }       
      while ( SampFile >> Sample) {
         SampFile >> LikelihoodRatio;
         //if (Sample >= MinRangeTmp && Sample < MaxRangeTmp) {
            if (LikelihoodRatio > 0) {
               //## Samples can be filtered by the user 
               //#  doing the LikelihoodRatio < 0
               //AddImportantSample(Sample,LikelihoodRatio);          
               if (Convert2dB_Scale) {
                  AddImportantSample( sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample),LikelihoodRatio);
               }
               else {
                  AddImportantSample(Sample,LikelihoodRatio);
               }     
            }
         //}   
      }   
      SampFile.close();                        
   }   
   InFile.close();
} //## End MultipleSelectiveImport(.)
******************/


//##### Imports the data from files created by the histogram class
//##   and merge multiple sets of files within the specified 
//#    range in the InFileName
//##  This function builds a pdf of a random parameter 
//#   selectivelly based on another parameter, which 
//#   was the one biased. It is assumed that there is 
//#   a substantial correlation between the two  
//#   quantities, but there are other parameters 
//#   that determine the dependencie.
//#   Attention: this module requires that all the 
//#   different bias have the same number of samples
//#   before the condition is applied.
//##  This function is obsolete, but is necessary
//#   while I do not put together the balance heuristic.
// ## By: Ivan Lima
void Histogram::ConditionalMultipleSelectiveImport(string InFileName,
                                                  int Convert2dB_Scale)
{
   //## Structure of InFileName
   //##  SampDataFileName    MinRangeToExtract  MaxRangeToExtract

   //## Clear the histogram, just in case...
   ClearHistogram();
   
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }  
   
   int qtSampleFiles = 0;
   int MaxQtSampleFiles = 100;
   double StatQtSamplesSumWeights[MaxQtSampleFiles][3];
   for (int ii = 0; ii< MaxQtSampleFiles; ii ++) {
      StatQtSamplesSumWeights[ii][0] = 0;
      StatQtSamplesSumWeights[ii][1] = 0;
      StatQtSamplesSumWeights[ii][2] = 0;
   }
   
   double Sample, SampleRange;
   double LikelihoodRatio, LikelihoodRatioRange;
   double MinRangeTmp, MaxRangeTmp;
   string SampFileName, RangeFileName;
   while ( InFile >> SampFileName) {

      InFile >> RangeFileName;
      InFile >> MinRangeTmp;
      InFile >> MaxRangeTmp;
      ifstream SampFile(SampFileName.c_str(),ios::in);
      if ( ! SampFile) {
         cout << endl;
         cout << "Warning: " << SampFileName << " could not be open."
              << endl;
         //exit(1);      
       }       
      ifstream RangeFile(RangeFileName.c_str(),ios::in);
      if ( ! RangeFile) {
         cout << endl;
         cout << "Warning: " << RangeFileName << " could not be open."
              << endl;
         //exit(1);      
       }              
       if (SampFile && RangeFile) {
          qtSampleFiles++;
       }   

      while ( SampFile >> Sample) {
         SampFile >> LikelihoodRatio;
         if (LikelihoodRatio > 0) {
           StatQtSamplesSumWeights[qtSampleFiles - 1][0] += 1;
           StatQtSamplesSumWeights[qtSampleFiles - 1][1] += LikelihoodRatio;
         }
         RangeFile >> SampleRange;
         RangeFile >> LikelihoodRatioRange;
         if (abs(LikelihoodRatio) != abs(LikelihoodRatioRange) ) {
            cout << endl;
            cout << "Error: " << SampFileName <<" and " 
                 << RangeFileName << " are not syncronized."
                 << endl << "Their likelihoodration do not match."
                 << endl;
            exit(1);             
         }                  
         if (SampleRange >= MinRangeTmp && SampleRange < MaxRangeTmp) {
            if (LikelihoodRatio > 0) {
               StatQtSamplesSumWeights[qtSampleFiles - 1][2] += 1;
               //## Samples can be filtered by the user 
               //#  doing the LikelihoodRatio < 0         
               if (Convert2dB_Scale) {
                  AddImportantSample( sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample),LikelihoodRatio);
               }
               else {
                  AddImportantSample(Sample,LikelihoodRatio);
               }                  
               //   AddImportantSample(Sample,LikelihoodRatio);          
            }   
         }   
      }   
      SampFile.close();                        
      RangeFile.close();
   }   
   InFile.close();
   
   //## Now check if everything is OK with respect
   //#  to StatQtSamplesSumWeights;
   
   int FoundUnbiasedSet = 0;
   int IndexUnbiasedSet;
   //## First, search for the unbiased case
   cout << endl << "File#   qtSamples SumWeights  qtUsedSamples" 
        << endl;
   for (int ii = 0; ii < qtSampleFiles; ii++) {
      if ( StatQtSamplesSumWeights[ii][0] > 0
          && (  StatQtSamplesSumWeights[ii][0]
              == StatQtSamplesSumWeights[ii][1]) ) {
         FoundUnbiasedSet++;
         IndexUnbiasedSet = ii;

      }        
      cout << ii+1 << "  " << StatQtSamplesSumWeights[ii][0]
           << "  " << StatQtSamplesSumWeights[ii][1] 
           << "  " << StatQtSamplesSumWeights[ii][2]
           << endl;
      if (    StatQtSamplesSumWeights[ii][0] 
           != StatQtSamplesSumWeights[ 0][0] ) {
         cout << "Error	from: Histogram::ConditionalMultipleSelectiveImport()"
              << endl << "The number of samples in each bias must be equal"
	      << endl;
	 //     << endl << "Aborting now!!!" << endl;
         //exit(1);	       	       
      }	   	   
	   
   }
   cout << endl;

   /**********   
   if ( ! FoundUnbiasedSet ) {
      cout << endl;
      cout << "Error: There should at least one Unbiased file to merge."
           << endl;
      exit(1);
   }  
   else if ( FoundUnbiasedSet > 1 ) {
      cout << endl;
      cout << "Warning: There more than one Unbiased file to merge."<< endl 
           << "         The total number is:" << FoundUnbiasedSet << endl;
   }
   
   cout << "#$#$#$ Now, check ..." << endl;
   //## Now check if all the sets of samples have the same number of samples
   double RatioQtSamples;
   for (int ii = 0; ii < qtSampleFiles; ii++) {
      if ( StatQtSamplesSumWeights[ii][0] > 0) {
         RatioQtSamples = StatQtSamplesSumWeights[ii][0]
                          /StatQtSamplesSumWeights[IndexUnbiasedSet][0];
         if  (RatioQtSamples > 1.+1e-3 || RatioQtSamples < 1.-1e-3 ) {
            cout << endl;
	    cout << "ERROR !!!!!!" << endl;
            cout << "Error: All the sets of samples must have the same number of samples."
                 << endl;
            //exit(1);         
         }
      }        
   } 

   cout << "#$#$#$ Now, check 2222..." << endl;   
   //## Now check if the sum of the weights is close to 
   //#  the number of samples per bias (all of whom must be equal)
   
   RatioQtSamples = GetSumLikelihoodRatio()
                   /StatQtSamplesSumWeights[IndexUnbiasedSet][0];
   cout << "Ratio (Sum Weigths)/(samples per bias) = "
        << RatioQtSamples << endl;
   if  (RatioQtSamples > 1.+1e-2 || RatioQtSamples < 1.-1e-2 ) {
      cout << endl;
      cout << "################################################" << endl;
      cout << "Warning: There is a substantial deviation between" << endl
           << " the expected and the actual sum of the weights." << endl
           << " This might affect the reliability of the results."
           << endl;
   } 
   
   //## Forcing the number of samples to be equal to the 
   //#  number of samples in the unbiased simulation
   //#  This is a very delicade issue.
   qtSamples = (long) StatQtSamplesSumWeights[IndexUnbiasedSet][0];
   
   cout << "Normalizing factor qtSamples set to "<<qtSamples<<endl<<endl;
   ***************/    

   qtSamples = (long) StatQtSamplesSumWeights[0][0];
   cout << "Normalizing factor qtSamples set to "<<qtSamples<<endl<<endl;
   
} //## End ConditionalMultipleSelectiveImport(.)


// ######################################################

/*!

   Imports (Sample,LikelihoodRatio)-pairs from the file.
   \param InFileName: The input file name (string)
   \sa  ImportData(string InFileName, int Column,int qtColumns)
        AddImportantSample
 */

// ##### Imports the data from a file created by the histogram class
// ## By: Ivan Lima
void Histogram::ImportData(string InFileName)
{
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }  
   double Sample, LikelihoodRatio;
   while ( InFile >> Sample) {
      InFile >> LikelihoodRatio;
      if (LikelihoodRatio >= 0) {
         AddImportantSample(Sample,LikelihoodRatio);          
      }   
   }   
   InFile.close();
}


// #########################################################

/*!

   Imports a column of sample data from the file and
   adds the samples to the Histogram.
   \param InFileName: Name of input file (string)
   \param Column: The index of the column of data to be input
   \param qtColumns: The number of columns in the file
   \sa  ImportData(string InFileName, int Column,int qtColumns)
        AddSample
 */

// ##### Imports the data from an arbitrary file
// ## By: Ivan Lima
void Histogram::ImportData(string InFileName, int Column,int qtColumns)
{

   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) 
    {
      cout << endl;
      cout << "Error: " << InFileName << " could not be created."
           << endl;
      exit(1);
    }

   //string strBuffer;

   double Sample = 0;
   int LocalColumn = 0;

   // ## I should investigate a way to know when 
   // #  is the line beginning to get rid of qtColumns.   

   while ( InFile >> Sample) 
    {
      LocalColumn++;
      if (LocalColumn > qtColumns)
         LocalColumn = 1;
      if (LocalColumn == Column)   
         AddSample(Sample);          
   }
   
   InFile.close();

}

// #################################################

/*! 

The method first calls Obtain_pdf to compute the pdf

The output file is a matrix of size qtBins by 3:

-# Column 1 is the vector pdfX 
-# Column 2 is the vector pdfY
-# Column 3 is the vector pdfConfidence (relative to the pdf)

\sa Obtain_pdf

 */

// ##### Write pdf from the histogram to a file
//#  The confidence is of the pdf
// ## By: Ivan Lima
void Histogram::WriteFile_pdf(string OutFileName)
{
   Obtain_pdf();
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << OutFileName << "could not be created."
           << endl;
      exit(1);
   }
   for (int ii = 0; ii < qtBins; ii++) {
      double pdfRelativeStdDevTmp = 99;
      if (pdfY[ii] != 0) {
         pdfRelativeStdDevTmp = pdfConfidence[ii] / pdfY[ii];
      }   
      OutFile << pdfX[ii] << "   " 
              << pdfY[ii] << "   "  
              << pdfConfidence[ii] << "   "
	      << pdfRelativeStdDevTmp << endl;   
   }           
   OutFile.close();
}

// ###########################################################

/*! 

The method first calls Obtain_pdf to compute the pdf

The output file is a matrix of size qtBins by 3:

-# Column 1 is the vector pdfX 
-# Column 2 is the cdf vector
-# Column 3 is the vector pdfConfidence, currently set to zero.

\sa Obtain_pdf

 */


//##### Write cdf from the histogram to a file
//## By: Ivan Lima (01/11/11)
void Histogram::WriteFile_cdf(string OutFileName)
{
   Obtain_pdf();
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << OutFileName << "could not be created."
           << endl;
      exit(1);
   }
   double deltaX = (MaxRange - MinRange)/qtBins;
   for (int ii = 0; ii < qtBins; ii++) {	      
      //## Note (IL): the shift by +deltaX/2 is due to the fact that 
      //#  the rectangular integration is correct at the left edge
      //#  of the rectangle.
      double cdfRelativeStdDevTmp = 99;
      if (cdfBin[ii]/qtSamples != 0) {
         cdfRelativeStdDevTmp = cdfConfidence[ii] 
	                       /(cdfBin[ii]/qtSamples);
      }        
      OutFile << pdfX[ii] + deltaX/2. << "   " 
              << cdfBin[ii]/qtSamples << "   "  
              << cdfConfidence[ii]    << "   "
	      << cdfRelativeStdDevTmp << endl;	      
   }             
   OutFile.close();
}



/***************** This method was used until 01/18/03
//##### Write cdf from the histogram to a file
//#  The confidence is absolute, not relative
//## By: Ivan Lima (01/11/11)
void Histogram::WriteFile_cdf(string OutFileName)
{
   Obtain_pdf();
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << OutFileName << "could not be created."
           << endl;
      exit(1);
   }
   double cdf = 0;
   double cdfConfidence = 0;   
   //double RenormCoeff;
   double deltaX = (MaxRange - MinRange)/qtBins;
   for (int ii = 0; ii < qtBins; ii++) {
      cdf += pdfY[ii]*deltaX;
      //if (ii > 1  && pdfY[ii]!= 0) {
      //   RenormCoeff = pdfY[ii-1]/pdfY[ii];      
      //}                
      //else {
      //   RenormCoeff = 1;
      //}      
      //cdfConfidence = cdfConfidence*RenormCoeff 
      //               +pdfConfidence[ii];
      //cdfConfidence += pdfConfidence[ii]*pdfY[ii]*deltaX;
      cdfConfidence += pdfConfidence[ii]*deltaX;
		      
      //cdfConfidence = 0; //## The confidence is not being generated		      
      //## Note (IL): the shift by +deltaX/2 is due to the fact that 
      //#  the rectangular integration is correct at the left edje
      //#  of the rectangle.
      double cdfRelativeStdDevTmp = 99;
      if (cdf != 0) {
         cdfRelativeStdDevTmp = cdfConfidence / cdf;
      }        
      OutFile << pdfX[ii] + deltaX/2. << "   " << cdf << "   "  
              << cdfConfidence << "   "
	      << cdfRelativeStdDevTmp << endl;	      
   }             
   OutFile.close();
}
**************/


// #############################################

/*! 

This method is used in applications to compute outage probabilities!

The method first calls Obtain_pdf to compute the pdf

The output file is a matrix of size qtBins by 3:

-# Column 1 is the vector pdfX 
-# Column 2 is the cdfc vector
-# Column 3 is the vector pdfConfidence, currently set to zero.

\sa Obtain_pdf

*/



//##### Write cdfc from the histogram to a file
//##  This method the indicator functions of the cdfc
//## By: Ivan Lima (01/11/11)
void Histogram::WriteFile_cdfc(string OutFileName)
{
   Obtain_pdf();
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << OutFileName << "could not be created."
           << endl;
      exit(1);
   }           
   double deltaX = (MaxRange - MinRange)/qtBins;
   for (int ii = 0; ii < qtBins; ii++) {
      //## Note (IL): the shift by -deltaX/2 is due to the fact that 
      //#  the rectangular integration is correct at the left edje
      //#  of the rectangle.
      double cdfcRelativeStdDevTmp = 99;
      if (cdfcBin[ii]/qtSamples != 0) {
         cdfcRelativeStdDevTmp = cdfcConfidence[ii] 
	                        /(cdfcBin[ii]/qtSamples);
      }        
      OutFile << pdfX[ii] - deltaX/2. << "   " 
              << cdfcBin[ii]/qtSamples << "   "  
              << cdfcConfidence[ii] << "   "
	      << cdfcRelativeStdDevTmp << endl; 
   }             
   OutFile.close();
}


/******************* This method was unsed until 01/18/03
//##### Write cdfc from the histogram to a file
//##  This method uses the rectangular integration
//## By: Ivan Lima (01/11/11)
void Histogram::WriteFile_cdfc(string OutFileName)
{
   Obtain_pdf();
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << OutFileName << "could not be created."
           << endl;
      exit(1);
   }
   double *cdfc = new double[qtBins];
   double *cdfcConfidence = new double[qtBins];
   //double cdfc[qtBins];
   //double cdfcConfidence[qtBins];
   double deltaX =  (MaxRange - MinRange)/qtBins;

   cdfc[qtBins - 1] = pdfY[qtBins - 1] * deltaX; //## IL 2/20/02
   cdfcConfidence[qtBins - 1] = pdfConfidence[qtBins - 1]*deltaX;
   //cdfcConfidence[qtBins - 1] = pdfConfidence[qtBins - 1]
   //                            *pdfY[qtBins - 1]*deltaX;
   //double RenormCoeff;                               
   for (int ii = qtBins - 2; ii >=  0; ii--) {
      cdfc[ii] = cdfc[ii+1] + pdfY[ii] * deltaX;
      //if (pdfY[ii]!= 0) {
      //   RenormCoeff = pdfY[ii+1]/pdfY[ii];      
      //}                
      //else {
      //   RenormCoeff = 1;
      //}
      //cdfcConfidence[ii] = cdfcConfidence[ii + 1]
      //                    *RenormCoeff
      //                    +pdfConfidence[ii];      
      //cdfcConfidence[ii] = cdfcConfidence[ii + 1]
      //                    +pdfConfidence[ii]*pdfY[ii]*deltaX;
      cdfcConfidence[ii] = cdfcConfidence[ii + 1]
                          +pdfConfidence[ii]*deltaX;      
   }

   for (int ii = 0; ii < qtBins; ii++) {
      //cdfcConfidence[ii] = 0; //## The confidence is not being generated
      //## Note (IL): the shift by -deltaX/2 is due to the fact that 
      //#  the rectangular integration is correct at the left edje
      //#  of the rectangle.
      double cdfcRelativeStdDevTmp = 99;
      if (cdfc[ii] != 0) {
         cdfcRelativeStdDevTmp = cdfcConfidence[ii] / cdfc[ii];
      }        
      OutFile << pdfX[ii] - deltaX/2. << "   " << cdfc[ii] << "   "  
              << cdfcConfidence[ii] << "   "
	      << cdfcRelativeStdDevTmp << endl; 
   }             
   OutFile.close();
   delete [] cdfc;
   delete [] cdfcConfidence;
}
***********/


// ####################################################
/*! 

The method first calls Obtain_pdf to compute the pdf

The output file is a matrix of size qtBins by 3:

-# Column 1 is the vector pdfX 
-# Column 2 is the number of hits per bin
-# Column 3 is the vector pdfConfidence (not necessarily debugged)

\sa Obtain_pdf

 */

// ##### Write the number of hits per bin to a file
// ## By: Ivan Lima (01/11/29)
void Histogram::WriteFileQtSamplesBin(string OutFileName)
{
   Obtain_pdf();
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << OutFileName << "could not be created."
           << endl;
      exit(1);
   }
   for (int ii = 0; ii < qtBins; ii++) {
      OutFile << pdfX[ii] << "   " << qtSamplesBin[ii] << "   "  
              << pdfConfidence[ii] << endl;   
   }           
   OutFile.close();
}

// #####################################################

void Histogram::WriteFileBinValues(string OutFileName)
{
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << OutFileName << "could not be created."
           << endl;
      exit(1);
   }
   for (int ii = 0; ii < qtBins; ii++) {
      OutFile << pdfX[ii] << "   " << Bin[ii] << endl;
   }
   OutFile.close();
}



//##############################################
//## Fail safe StdDev
//## By: Ivan Lima

double Histogram::GetStdDev(void) 
{ 
   double StdDev;
   if (  GetMeanSquare() > sq(GetMean()) 
      && qtSamples > 1) {
      StdDev = sqrt( qtSamples*(GetMeanSquare() 
                                - sq(GetMean()) )
                  /(qtSamples - 1.) );
   }
   else {
      StdDev = 0;
   }
   return StdDev;
}   


//##############################################
//## Get the sum of all the probabilities
//#  This method is just a debugging method
//## By: Ivan Lima

double Histogram::GetSumProbabilities(void) 
{ 
   Obtain_pdf();
   double SumProbabilities = 0;
   for (int ii = 0; ii < qtBins; ii++) {
      SumProbabilities += pdfY[ii]/qtBins*(MaxRange - MinRange);
   }
   return SumProbabilities;
}   

// #############################################

void Histogram::WriteFilePDF(string OutFileName)
 {
   // OLD CODE. Not maintained anymore.

   // This method duplicates WriteFile_pdf except that the file name
   // is passed as a string not char * (JZ)

   cout << "Number Samples = " << qtSamples << endl;
   cout << "Number Samples Out of Range = " 
        << qtSamples - qtSamplesInRange << endl;
// ## It is no longer required to Obtain_pdf()
// #  It will do automatically (Ivan Lima, 2/22/01) 
//   if (! pdfObtained) 
//     {
//      cerr << "Error: Execute first Obtain_pdf()" 
//           << " before WriteFilePDF(...)."
//           << endl;
//      exit(1);
//   }      

   Obtain_pdf();
   ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "Error: " << OutFileName << "could not be created."
           << endl;
      exit(1);
    }

   for (int ii = 0; ii < qtBins; ii++) 
      OutFile << pdfX[ii] << "   " << pdfY[ii]   << endl;   

   OutFile.close(); 
 }


/*!

The domain of the Maxwellian pdf is the same as the domain of the
histogram

\param  OutFileName: Name of the output file for the Maxwellian pdf
\param  qtPoints: Number of points in the discretization of the Maxwellian
                  pdf
 */

//##############################################
//##### Write Maxwellian pdf, with the same mean, to a file
//## By: Ivan Lima
void Histogram::WriteFileMaxwell_pdf(string OutFileName, int qtPoints)
{
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << OutFileName << "could not be created."
           << endl;
      exit(1);
   }
   double vSample;  
   double Alpha = GetMean()/(2*sqrt(2/pi));
   for (int ii = 0; ii < qtPoints; ii++)  {
      vSample = ii/ (double) qtPoints*(MaxRange - MinRange) + MinRange;
      OutFile << vSample << "   "
              << sqrt(2/pi)*sq(vSample)/cube(Alpha)
                *exp(-sq(vSample)/(2*sq(Alpha))) << endl;
   }            
   OutFile.close();
}


//##############################################

/*!

 Writes to a file the theoretical pdf of the length of the
 second order component of the polarization 
 dispersion vector, with the same mean as that of the Histogram

*/

//##### Write theoretical pdf of the length of the
//##    second order component of the polarization 
//##    dispersion vector, with the same mean, to a file.
//## By: Ivan Lima (11/27/01)

void Histogram::WriteFileSndOrderPDV_pdf(string OutFileName, int qtPoints)
{
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << OutFileName << "could not be created."
           << endl;
      exit(1);
   }
   double vSample;  
   double gamma=2.332487232/GetMean();     //## from Maple
   for (int ii = 0; ii < qtPoints; ii++)  {
      vSample = ii/ (double) qtPoints*(MaxRange - MinRange) + MinRange;
      OutFile << vSample << "   "
              <<(2./pi)*sq(gamma)*vSample
                *sinh(gamma*vSample)/sq(cosh(gamma*vSample))
              << endl;  
   }       
   OutFile.close();
}
 
 
//################################################

/*!

The domain of the Gaussian pdf is the same as the domain of the
histogram

\param  OutFileName: Name of the output file for the Gaussian pdf
\param  qtPoints: Number of points in the discretization of the Gaussian
                  pdf

The first column of the output is the discretized domain of the histogram.
The second column are the values of the pdf.

*/

 
//################################################
//##### Write Gaussian pdf, with the same mean and deviation, to a file
//## By: Ivan Lima
void Histogram::WriteFileGaussian_pdf(string OutFileName, int qtPoints)
{
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << OutFileName << "could not be created."
           << endl;
      exit(1);
   }
   double vSample;
   
   //cout << "GetMean() = " << GetMean() << endl;
   //cout << "GetStdDev() = " << GetStdDev() << endl;
     
   for (int ii = 0; ii < qtPoints; ii++)  {
      vSample = ii/ (double) qtPoints*(MaxRange - MinRange) + MinRange;
      OutFile << vSample << "   "
              << 1./(sqrt(2*pi)*GetStdDev())
                *exp(-sq((vSample-GetMean())
                         /GetStdDev())/2.) << endl;
   }               
   OutFile.close();
}



//################################################
//##### Write Gaussian cdf, with the same mean and deviation, to a file
//## By: Ivan Lima
void Histogram::WriteFileGaussian_cdf(string OutFileName, int qtPoints)
{
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << OutFileName << "could not be created."
           << endl;
      exit(1);
   }
   double vSample, vGaussian;  
   double cdfcGaussian = 0.;
   for (int ii = 0; ii < qtPoints; ii++)  {
      vSample = ii/ (double) qtPoints*(MaxRange - MinRange) + MinRange;
      vGaussian = 1./(sqrt(2*pi)*GetStdDev())
                *exp(-sq((vSample-GetMean())
                         /GetStdDev())/2.);
      cdfcGaussian += vGaussian*(MaxRange - MinRange)/qtPoints;  	 		 
      OutFile << vSample << "   " << cdfcGaussian << endl;
   }               
   OutFile.close();
}


//################################################
//##### Write Gaussian cdfc, with the same mean and deviation, to a file
//## By: Ivan Lima
void Histogram::WriteFileGaussian_cdfc(string OutFileName, int qtPoints)
{
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << OutFileName << "could not be created."
           << endl;
      exit(1);
   }
   double vSample, vGaussian;  
   double cdfcGaussian = 1.;
   for (int ii = 0; ii < qtPoints; ii++)  {
      vSample = ii/ (double) qtPoints*(MaxRange - MinRange) + MinRange;
      vGaussian = 1./(sqrt(2*pi)*GetStdDev())
                *exp(-sq((vSample-GetMean())
                         /GetStdDev())/2.);
      cdfcGaussian -= vGaussian*(MaxRange - MinRange)/qtPoints;  	 		 
      OutFile << vSample << "   " << cdfcGaussian << endl;
   }               
   OutFile.close();
}


