//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsHistogram2D.hh"

extern ofstream *LogFile;

// ###########################################

Histogram2D::Histogram2D(string InFileName,float x_min2,float x_max2,
                         float y_min2, float y_max2)
{

 HistInFileName = InFileName;

 ifstream InFile(InFileName.c_str(),ios::in);  

 if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName
          << " could not be opened."<< endl;
      exit(1);
    }

   
   LogFileSeparator();
   *LogFile << InFileName << endl << endl;

  NumBinsX =  ReadInt("NumBinsX"," ",0,-1,LOWER_ONLY,&InFile);
  NumBinsY =  ReadInt("NumBinsY"," ",0,-1,LOWER_ONLY,&InFile);

  InFile.close();

  x_min = x_min2;
  x_max = x_max2;
  y_min = y_min2;
  y_max = y_max2;

  InitializeParameters();
  
}

// ###################################################################

void Histogram2D::InitializeParameters(void)
{

  x_window_size = x_max - x_min;
  y_window_size = y_max - y_min;

  delta_x = x_window_size/float(NumBinsX);
  delta_y = y_window_size/float(NumBinsY);

  Bin = new double[NumBinsX*NumBinsY]; 
  BinSquare = new double[NumBinsX*NumBinsY]; 
  pdfConfidence = new double[NumBinsX*NumBinsY];   

  PDF = new double[NumBinsX*NumBinsY];

  XData = new double[NumBinsX];
  YData = new double[NumBinsY];
  XMarginal = new double[NumBinsX];
  YMarginal = new double[NumBinsY];

  PreviouslyComputed_pdfConfidenceFlag = 0;  

  if(!Bin || !BinSquare || !pdfConfidence || !PDF || 
     !XData || !YData || !XMarginal || !YMarginal)
    {
      *LogFile << "Memory Allocation error in Histogram2D constructor" << endl;
      exit(1);
    }

  //## For use in importance sampling studies

  qtSamplesBin = 0;

  qtSamplesBin = new int[NumBinsX*NumBinsY]; 

  if(!qtSamplesBin)
    {
      *LogFile << "Memory Allocation error in Histogram2D constructor" << endl;
      exit(1);
    }   
   
  Clear();

} // ## end InitializeParameters()

//#########################################

Histogram2D::Histogram2D(string InFileName)
{

 HistInFileName = InFileName;

 ifstream InFile(InFileName.c_str(),ios::in);  

 if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName
          << " could not be opened."<< endl;
      exit(1);
    }

   
   LogFileSeparator();
   *LogFile << InFileName << endl << endl;


  NumBinsX =  ReadInt("NumBinsX"," ",0,-1,LOWER_ONLY,&InFile);
  NumBinsY =  ReadInt("NumBinsY"," ",0,-1,LOWER_ONLY,&InFile);

  x_min = (float) ReadDouble("x_min"," ",-1,-1,NO_BOUNDS,&InFile);
  x_max = (float) ReadDouble("x_max"," ",-1,-1,NO_BOUNDS,&InFile);  
  y_min = (float) ReadDouble("y_min"," ",-1,-1,NO_BOUNDS,&InFile);
  y_max = (float) ReadDouble("y_max"," ",-1,-1,NO_BOUNDS,&InFile);  

  InFile.close();

  InitializeParameters();

} // ## end Histogram2D constructor


// #########################################

Histogram2D::Histogram2D(string InFileName,OptSignal *oOptSignal)
{

 HistInFileName = InFileName;

 ifstream InFile(InFileName.c_str(),ios::in);  

 if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName
          << " could not be opened."<< endl;
      exit(1);
    }

   
   LogFileSeparator();
   *LogFile << InFileName << endl << endl;


  NumBinsX =  ReadInt("NumBinsX"," ",0,-1,LOWER_ONLY,&InFile);
  NumBinsY =  ReadInt("NumBinsY"," ",0,-1,LOWER_ONLY,&InFile);
  double MaxPDPower = ReadDouble("MaxPDPower"," ",0,-1,LOWER_ONLY,&InFile);
  double MinPDPower = ReadDouble("MinPDPower"," ",0,-1,LOWER_ONLY,&InFile);

  x_min = 0.0;
  x_max = oOptSignal->GetBitLength();
  y_min = MinPDPower;
  y_max = MaxPDPower;

  InFile.close();

  InitializeParameters();

} // ## end constructor

// #########################################

//! Copy constructor

Histogram2D::Histogram2D(Histogram2D & OrigHist)
{

x_min = OrigHist.x_min;
x_max = OrigHist.x_max;
y_min = OrigHist.y_min;
y_max = OrigHist.y_max;

NumBinsX = OrigHist.NumBinsX;
NumBinsY = OrigHist.NumBinsY;

InitializeParameters();

for(int x_index=0;x_index<NumBinsX;x_index++)
for(int y_index=0;y_index<NumBinsY;y_index++)
 SetPDF(x_index,y_index,OrigHist.GetPDF(x_index,y_index));

for(int x_index=0;x_index<NumBinsX;x_index++)
for(int y_index=0;y_index<NumBinsY;y_index++)
 SetBin(x_index,y_index,OrigHist.GetBin(x_index,y_index));

for(int ii=0;ii<NumBinsX*NumBinsY;ii++)
 BinSquare[ii]=OrigHist.BinSquare[ii];

for(int ii=0;ii<NumBinsX*NumBinsY;ii++)
 qtSamplesBin[ii]=OrigHist.qtSamplesBin[ii];

for(int ii=0;ii<NumBinsX*NumBinsY;ii++)
 pdfConfidence[ii]=OrigHist.pdfConfidence[ii];

PreviouslyComputed_pdfConfidenceFlag = 
 OrigHist.PreviouslyComputed_pdfConfidenceFlag;

NumSamplesTotal = OrigHist.NumSamplesTotal;
NumSamplesInRange = OrigHist.NumSamplesInRange;
NumSamplesWithOutOfRangeX = OrigHist.NumSamplesWithOutOfRangeX;
NumSamplesWithOutOfRangeY = OrigHist.NumSamplesWithOutOfRangeY;
SamplesMinX =  OrigHist.SamplesMinX;
SamplesMaxX =  OrigHist.SamplesMaxX;
SamplesMinY =  OrigHist.SamplesMinY;
SamplesMaxY =  OrigHist.SamplesMaxY;
HistInFileName = OrigHist.HistInFileName;
MeanX=OrigHist.MeanX;
MeanY=OrigHist.MeanY;


}

// ##############################################

Histogram2D & Histogram2D::operator=(Histogram2D & RHSHist)
{ 
  
x_min = RHSHist.x_min;
x_max = RHSHist.x_max;
y_min = RHSHist.y_min;
y_max = RHSHist.y_max;

NumBinsX = RHSHist.NumBinsX;
NumBinsY = RHSHist.NumBinsY;

x_window_size = RHSHist.x_window_size; 
y_window_size = RHSHist.y_window_size;

delta_x = RHSHist.delta_x;
delta_y = RHSHist.delta_y;


for(int x_index=0;x_index<NumBinsX;x_index++)
for(int y_index=0;y_index<NumBinsY;y_index++)
 SetPDF(x_index,y_index,RHSHist.GetPDF(x_index,y_index));

for(int x_index=0;x_index<NumBinsX;x_index++)
for(int y_index=0;y_index<NumBinsY;y_index++)
 SetBin(x_index,y_index,RHSHist.GetBin(x_index,y_index));
 
for(int ii=0;ii<NumBinsX*NumBinsY;ii++)
 BinSquare[ii]=RHSHist.BinSquare[ii];
    
for(int ii=0;ii<NumBinsX*NumBinsY;ii++)
 qtSamplesBin[ii]=RHSHist.qtSamplesBin[ii];

for(int ii=0;ii<NumBinsX*NumBinsY;ii++)
 pdfConfidence[ii]=RHSHist.pdfConfidence[ii];

PreviouslyComputed_pdfConfidenceFlag =
 RHSHist.PreviouslyComputed_pdfConfidenceFlag;

NumSamplesTotal = RHSHist.NumSamplesTotal;
NumSamplesInRange = RHSHist.NumSamplesInRange;
NumSamplesWithOutOfRangeX = RHSHist.NumSamplesWithOutOfRangeX;
NumSamplesWithOutOfRangeY = RHSHist.NumSamplesWithOutOfRangeY;
SamplesMinX =  RHSHist.SamplesMinX;
SamplesMaxX =  RHSHist.SamplesMaxX;
SamplesMinY =  RHSHist.SamplesMinY;
SamplesMaxY =  RHSHist.SamplesMaxY;
HistInFileName = RHSHist.HistInFileName;
MeanX=RHSHist.MeanX;
MeanY=RHSHist.MeanY;
PreviouslyComputed_pdfConfidenceFlag = 
 RHSHist.PreviouslyComputed_pdfConfidenceFlag;
 
return *this;

} 



// ########################################

void Histogram2D::SetMaxY(double y_max2)
{

  y_max = y_max2;
  y_window_size = y_max - y_min;

  LogFileSeparator();
  *LogFile << "Resetting maximum y-value in 2D Histogram and clearing histogram"
	  << endl;

  *LogFile << "[x_min,x_max] = [" << x_min << " , " << x_max << "]" << endl;
  *LogFile << "[y_min,y_max] = [" << y_min << " , " << y_max << "]" << endl;


  delta_x = x_window_size/float(NumBinsX);
  delta_y = y_window_size/float(NumBinsY);

  Clear();

} // ## end SetMaxY

// #############################################################

void Histogram2D::SetMinMaxXY(double x_min2, double x_max2,
			      double y_min2, double y_max2)
{
  x_max = x_max2;
  x_min = x_min2;
  y_max = y_max2;
  y_min = y_min2;


  x_window_size = x_max - x_min;
  y_window_size = y_max - y_min;

/*
  LogFileSeparator();
  *LogFile << "Resetting maximum y-value in 2D Histogram and clearing histogram"
          << endl;

  *LogFile << "[x_min,x_max] = [" << x_min << " , " << x_max << "]" << endl;
  *LogFile << "[y_min,y_max] = [" << y_min << " , " << y_max << "]" << endl;
*/

  delta_x = x_window_size/float(NumBinsX);
  delta_y = y_window_size/float(NumBinsY);

  Clear();

} // ## end SetMinMaxXY

// ###########################################

Histogram2D::~Histogram2D()
{
  delete [] Bin;
  delete [] BinSquare;  
  delete [] pdfConfidence;
  delete [] qtSamplesBin;
}

// #########################################

void Histogram2D::SetBin(int x_index, int y_index, double value)
{

  // We use this means of indexing as typically want to access
  // Bins in a double loop like
  // for x
  // for y
  // Bin(x,y) = value

  Bin[y_index + NumBinsY*x_index] = value;
}

// #######################################

double Histogram2D::GetBin(int x_index, int y_index)
{
  return  Bin[y_index + NumBinsY*x_index];
}

//##################################################
//## By: Ivan Lima (10/07/02)
double Histogram2D::GetBinSquare(int x_index, int y_index)
{
  return  BinSquare[y_index + NumBinsY*x_index];
}

//##################################################
//## By: Ivan Lima (10/06/02)
double Histogram2D::Get_pdfConfidence(int x_index, int y_index)
{
   double pdfConfidenceTmp = 0;
   if ( GetBinSquare(x_index, y_index) 
	     > sq(GetBin(x_index, y_index) )/NumSamplesTotal )  
      {
      pdfConfidenceTmp = sqrt( ( GetBinSquare(x_index, y_index)  
                     -NumSamplesTotal*sq(GetBin(x_index, y_index)
		     /NumSamplesTotal) )
                        /( NumSamplesTotal-1) /NumSamplesTotal );
      }			   
   /////## Normalization of the confidence interval
   /////pdfConfidenceTmp *= NumBinsX/(x_max - x_min)*NumBinsY/(y_max - y_min);   
   return pdfConfidenceTmp;
}			       

//##################################################
//## Compute the confidence for all the bins in the pdf
//## By: Ivan Lima (10/06/02)
void Histogram2D::ObtainConfidence(void)
{
   if ( ! PreviouslyComputed_pdfConfidenceFlag) {
      for (int x_index = 0; x_index < NumBinsX; x_index++) {
         for (int y_index = 0; y_index < NumBinsY; y_index++) {
            pdfConfidence[GetIndex(x_index, y_index)] =	    
   	          Get_pdfConfidence(x_index, y_index);
         }
      }    
   }
}


//##################################

int Histogram2D::GetQtSamplesBin(int x_index, int y_index)
{
  return  qtSamplesBin[y_index + NumBinsY*x_index];
}


// ######################################


void Histogram2D::IncrementBin(int x_index, int y_index, double LikelihoodRatio)
{
  Bin[y_index + NumBinsY*x_index] += LikelihoodRatio; 
  BinSquare[y_index + NumBinsY*x_index] += sq(LikelihoodRatio);   
  qtSamplesBin[y_index + NumBinsY*x_index]++;  
}


//void Histogram2D::IncrementBin(int x_index, int y_index)
//{
//  Bin[y_index + NumBinsY*x_index] += 1;
//}

// #######################################

void Histogram2D::Clear()
{

  LogFileSeparator();
  *LogFile << "Clearing Histogram2D" << endl;

  for(int x_index = 0; x_index < NumBinsX; x_index ++)
     for(int y_index = 0; y_index < NumBinsY; y_index ++) {
        SetBin(x_index,y_index,0);
	qtSamplesBin[y_index + NumBinsY*x_index] = 0;
        BinSquare[y_index + NumBinsY*x_index] = 0;
        pdfConfidence[y_index + NumBinsY*x_index] = 0;

     }  

  NumSamplesTotal = 0;
  NumSamplesInRange = 0;
  NumSamplesWithOutOfRangeX = 0;
  NumSamplesWithOutOfRangeY = 0;

  MeanX = 0;
  MeanY = 0;

  for(int i=0;i<NumBinsX;i++)
    XData[i]=0.0;

  for(int i=0;i<NumBinsY;i++)
    YData[i]=0.0;

  for(int i=0;i<NumBinsX;i++)
    XMarginal[i]=0.0;

  for(int i=0;i<NumBinsY;i++)
    YMarginal[i]=0.0;

  for(int i=0;i<NumBinsX;i++)
   for(int j=0;j<NumBinsY;j++)
    SetPDF(i,j,0.0);

}

// ##########################################

void Histogram2D::AddSample(double x, double y)
{
   AddSample(x,y,1.); //## For unbiased simulations
}

// ###########################################

void Histogram2D::AddSample(double x, double y, double LikelihoodRatio)
{

  // The bottom left corner (x_min,y_min) of the physical grid
  // corresponds to Bin(0,0).

  // x = x_min + x_index*delta_x + rx,   rx < delta_x
  // So
  // x_index = (x-x_min)/delta_x - rx_delta_x
  // So
  // x_index = floor((x-x_min)/delta_x)
  //
  // Same for y_index

  // First we update the min and max x and y values encountered in the
  // samples

   
   if(NumSamplesTotal == 0)
    {
      SamplesMinX = x;
      SamplesMaxX = x;
      SamplesMinY = y;
      SamplesMaxY = y;
    }
   else
    {
      if(x < SamplesMinX)
	SamplesMinX = x;
      if(x > SamplesMaxX)
	SamplesMaxX = x;

      if(y < SamplesMinY)
	SamplesMinY = y;
      if(y > SamplesMaxY)
	SamplesMaxY = y;

    } // end else


   // Next we compute the x and y bin-indicies of the sample and check
   //  whether the sample is in range

  NumSamplesTotal++;

  int SampleInRange = 1;

  int x_index = int(floor((x-x_min)/delta_x));

  //*LogFile << "x = " << x << " x_index = " << x_index << endl;

  if(x_index < 0 || x_index >= NumBinsX)
    {
     NumSamplesWithOutOfRangeX++;
     SampleInRange = 0;
    }
 
  int y_index = int(floor((y-y_min)/delta_y));

  //*LogFile << "y = " << y << " y_index = " << y_index << endl;

  if(y_index < 0 || y_index >= NumBinsY)
    {
     NumSamplesWithOutOfRangeY++;
     SampleInRange = 0;
    }
  
 
  //*LogFile << "SampleInRange = " << SampleInRange << endl << endl;

  // If sample is in range we increment the bin it belongs to

  if(SampleInRange)
    {
      MeanX += x;
      MeanY += y;

      IncrementBin(x_index,y_index,LikelihoodRatio);
      NumSamplesInRange++;
    }

} // ## end AddSample

// ############################################



void Histogram2D::WriteEyeDiagramHistogram(string OutFileName)
{
 
  LogFileSeparator();
  *LogFile << "2D Histogram Range Information" << endl << endl;
  *LogFile << "NumSamplesTotal = " << NumSamplesTotal << endl;
  *LogFile << "NumSamplesInRange = " << NumSamplesInRange << endl;
  *LogFile << "NumSamplesWithOutOfRangeX = " 
          <<  NumSamplesWithOutOfRangeX << endl; 
  *LogFile << "NumSamplesWithOutOfRangeY = " 
	  << NumSamplesWithOutOfRangeY << endl; 

  *LogFile << endl;
  *LogFile << "SamplesMinX = " << SamplesMinX << endl;
  *LogFile << "SamplesMaxX = " << SamplesMaxX << endl;
  *LogFile << "SamplesMinY = " << SamplesMinY << endl;
  *LogFile << "SamplesMaxY = " << SamplesMaxY << endl;

  ofstream OutFile(OutFileName.c_str(),ios::out);

  if (!OutFile) 
   {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }


  // Print out x_window (Time Span for Eye Diagram application) 
  // and y_max

  OutFile << x_window_size << "\t" << y_min << "\t" << y_max << "\t";

  // Next print out the histogram to a file so that first line of file
  // is top row of Histogram, with y_index = NumBinsY - 1. %% WRONG DOC!! JZ

  // Since Ron's matlab Eye Diagram routine assumes that the first three
  // elements of the first row are x_window_size, y_min and y_max we only
  // output the entries 3 to  NumBinsX-1 of the first row.

  int x_index,y_index;

  // Write out the row with x_index = NumBinsX-1

  for(y_index = 3; y_index < NumBinsY; y_index++)
    //OutFile << GetBin(NumBinsX-1,y_index) << "\t";
      OutFile << GetBin(0,y_index) << "\t";

  OutFile << endl;

  // Write out the remaining rows

  for(x_index=NumBinsX-2;x_index >=0; x_index--)
    {
       for(y_index = 0; y_index < NumBinsY; y_index++)
	 // OutFile << GetBin(x_index,y_index) << "\t";
	    OutFile << GetBin(NumBinsX - 1 - x_index,y_index) << "\t";

       OutFile << endl;

    }// end for y loop

  OutFile.close();

} // ## end WriteHistogram

// ############################################

void Histogram2D::UpdateEyeDiagram(OptSignal *oOptSignal,
                                 Photodetector *oPhotodetector)
{

  // Given an Electrical Power Signal obtained by passing the 
  // Optical Field through a Photodetector we update the 
  // eye diagram histogram as follows.
  // The x_window_size variable of the histogram is equal
  // to the bit period of the signal. 
  // For each bit in the signal we loop over the x-bins 
  // of the histogram. We compute the time at the center 
  // of each of these bins, and the corresponding discrete
  // time-index. We then compute the power at this time
  // and add the double pair (time,power) to the histogram
  // using the AddSample method.

  int kk; // Index of bit slot

  int nx; // x-index of Bin

  double StartTimeOfBitSlot, CenterTimeOfBin;
  int CenterTimeIndexOfBin;
  double Time,Power;



  for(kk=0;kk<oOptSignal->GetStringLength();kk++)
    {
      StartTimeOfBitSlot = kk*oOptSignal->GetBitLength();

      for(nx=0;nx<NumBinsX;nx++)
       {
        CenterTimeOfBin = StartTimeOfBitSlot + (nx+0.5)*delta_x;

        CenterTimeIndexOfBin = int(CenterTimeOfBin/oOptSignal->GetDeltaTime());

        
        if(CenterTimeIndexOfBin < 0)
         {
          CenterTimeIndexOfBin = 0;
          *LogFile << "WARNING: Potential problem with UpDateEyeDiagram  (1)"
		  << endl;
         }

        if(CenterTimeIndexOfBin >= oOptSignal->GetQtPoints())
         {
           CenterTimeIndexOfBin = oOptSignal->GetQtPoints() - 1;

           *LogFile << "WARNING: Potential problem with UpDateEyeDiagram  (2)"
		   << endl;
         }

       Time = CenterTimeOfBin - StartTimeOfBitSlot;

       Power = real(oPhotodetector->sfftPM.tPowerSignal[CenterTimeIndexOfBin]);

       AddSample(Time,Power);            
       
       } // ## end for nx loop

    } // ## end for kk loop



} // ## end UpdateEyeDiagram

//########################################################
//##### Imports the data from files created by the histogram class
//##    Convert2dB_Scale also determines the sign of the conversion to dB
//##    By: Ivan T. Lima Jr.

void Histogram2D::MultipleSelectMaxRelNumHitsImport(string InFileName,
                                                  int Convert2dB_Scale)
{
   //## Structure of InFileName
   //##  SampDataFileName    
         
   //## Clear the histogram, just in case...
   Clear();
   
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }     

   //## It can merge up to 100 biased files
   string SampFileName0[100], SampFileName1[100];
   Histogram2D* oHist_pdf2D[100];
      
   *LogFile << "List of file names to be merged:" << endl;
   int FileCounter = 0;
   while (InFile >> SampFileName0[FileCounter]) {
      InFile >> SampFileName1[FileCounter];      
    
      ifstream SampFile0(SampFileName0[FileCounter].c_str(),ios::in);       
      ifstream SampFile1(SampFileName1[FileCounter].c_str(),ios::in);          
      if ( ! SampFile0 || ! SampFile1) {
         cout << endl;
	 if (! SampFile0) {
            cout << "Warning: " << SampFileName0[FileCounter] 
	         << " could not be open." << endl;
	 }           
         if ( ! SampFile1) {
            cout << endl;
            cout << "Warning: " << SampFileName1[FileCounter] 
      	      << " could not be open." << endl;     
	 }     	 
      }     
      else { //## They do exist

         SampFile0.close();
         SampFile1.close();

         *LogFile << "   "<< SampFileName0[FileCounter] 
                 << "   "<< SampFileName1[FileCounter] << endl;  

         FileCounter++;
      }	 
      SampFile0.close();
      SampFile1.close();
   }
   InFile.close();

   //## First, get the pdf for each case
   double Sample0, Sample1;
   double LikelihoodRatio0, LikelihoodRatio1;      
   for (int ii = 0; ii < FileCounter; ii++) {
      oHist_pdf2D[ii] = new Histogram2D(HistInFileName.c_str()); 
                                      //  x_min,x_max,y_min,y_max);
            
      cout << "File Name: "<< ii <<" = " <<SampFileName0[ii]  
                                 << "  " <<SampFileName1[ii] << endl;
      ifstream SampFile0(SampFileName0[ii].c_str(),ios::in);      
      if ( ! SampFile0) {
         cout << endl;
         cout << "Warning: " << SampFileName0[ii] << " could not be open."
              << endl;
	  continue;       
      }             
      ifstream SampFile1(SampFileName1[ii].c_str(),ios::in);        
      if ( ! SampFile1) {
         cout << endl;
         cout << "Warning: " << SampFileName1[ii] << " could not be open."
              << endl;
	  continue;      
      }                
      while ( SampFile0 >> Sample0) {
         SampFile0 >> LikelihoodRatio0;
	 SampFile1 >> Sample1;
	 SampFile1 >> LikelihoodRatio1;
	 if (LikelihoodRatio0 != LikelihoodRatio1) {
	    cout << "ERROR! From: Histogram2D::MultipleSelectMaxRelNumHitsImport(.)"
	         << endl << "The likelihoodRatio of the files "
		 <<SampFileName0[ii] << " and " << SampFileName1[ii] << endl
		 << "are different. They must be equal."  << endl;
            exit(1);		   
	 }
         if (LikelihoodRatio0 > 0) {
            //## Samples can be filtered by the user 
            //#  doing the LikelihoodRatio < 0
            if (Convert2dB_Scale) {
               oHist_pdf2D[ii]->AddSample( sgn(Convert2dB_Scale)
                                                 *Linear2dB(Sample0),
                                           sgn(Convert2dB_Scale)
                                                 *Linear2dB(Sample1),						 
						 LikelihoodRatio0);
            }
            else {
              oHist_pdf2D[ii]->AddSample(Sample0,Sample1,LikelihoodRatio0);
            }   
         }
      }               
      SampFile0.close();
      SampFile1.close();
   } 

/*******   
   //## Second, get the total number of samples per bin
   long *qtSamplesBinTmp = new long[NumBinsX*NumBinsY];
   for (int x_index = 0; x_index < NumBinsX; x_index++) {
      for (int y_index = 0; y_index < NumBinsY; y_index++) {      
         qtSamplesBinTmp[y_index + NumBinsY*x_index] = 0;
      }
   }   	 
   for (int ii = 0; ii < FileCounter; ii++) {
      for (int x_index = 0; x_index < NumBinsX; x_index++) {
         for (int y_index = 0; y_index < NumBinsY; y_index++) {      
            qtSamplesBinTmp[y_index + NumBinsY*x_index] += 
   	       oHist_pdf2D[ii]->GetQtSamplesBin(x_index,y_index);
         }
      }   	 
   }   
********/

   Clear();

   //## Third, select the bin that had the maximum relative
   //#  number of hits from all the distributions
   //#  I have to do it in such a way that Obtain_pdf
   //#  won't mess things up.
   double MaxRelNumHits;
   for (int x_index = 0; x_index < NumBinsX; x_index++) {
      for (int y_index = 0; y_index < NumBinsY; y_index++) {      
         MaxRelNumHits = 0;
         for (int ii = 0; ii < FileCounter; ii++) {
            if ( (double) oHist_pdf2D[ii]->GetQtSamplesBin(x_index,y_index)
                         /oHist_pdf2D[ii]->GetQtSamples() 
             > MaxRelNumHits ) {
            MaxRelNumHits =  (double) 
                  oHist_pdf2D[ii]->GetQtSamplesBin(x_index,y_index)
                  /oHist_pdf2D[ii]->GetQtSamples();              
            SetBin(x_index,y_index, (double)
             	    oHist_pdf2D[ii]->GetBin(x_index,y_index)
	            /oHist_pdf2D[ii]->GetQtSamples() );
            }
         }  
      }   
      
   }   

   //## Fourth, and finally, force the number of samples to be 
   //#  unitary to get the right statistics, since all the 
   //#  pdfs used were already normalized.
   //#  Obviously, I need to denormilize the Bins, since the
   //#  Obtain_pdf method will do it
   NumSamplesTotal = 1;
   NumSamplesInRange = 1;    
     

   *LogFile << "The pdf from "<< InFileName.c_str()<<" was obtained "
           << "from a merging process." << endl;
   *LogFile << "Thus, don't expect fully functionality from the other methods."
           << endl;
   
} //## End MultipleSelectMaxRelNumHitsImport(.)




//########################################################
//##### Imports the data from files created by the histogram class
//##    Convert2dB_Scale also determines the sign of the conversion to dB
//##    Actually, what you get from this is the joint pdf
//##    By: Ivan T. Lima Jr. and Aurenice O. Lima (7/28/02)
void Histogram2D::MultipleHeuristicBalanceImport(string InFileName,
                                                  int Convert2dB_Scale)
{
   //## Structure of InFileName
   //##  SampDataFileName    
         
   //## Clear the histogram, just in case...
   Clear();
   
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }     


   //## It can merge up to 100 biased files
   #define MAX_FILES_HEURISTIC_BALANCE 100
   string SampFileName0[MAX_FILES_HEURISTIC_BALANCE];
   string SampFileName1[MAX_FILES_HEURISTIC_BALANCE];
   string SampHeuristicBalanceFileName[MAX_FILES_HEURISTIC_BALANCE];
   //Histogram2D* oHist_pdf2D[100];
   Histogram2D* oHistMeanParam2D[MAX_FILES_HEURISTIC_BALANCE];   

      
   *LogFile << "List of file names to be merged:" << endl;
   int FileCounter = 0;
   while (InFile >> SampFileName0[FileCounter]) {
      InFile >> SampFileName1[FileCounter];      
      InFile >> SampHeuristicBalanceFileName[FileCounter];   
          
      ifstream SampFile0(SampFileName0[FileCounter].c_str(),ios::in);       
      ifstream SampFile1(SampFileName1[FileCounter].c_str(),ios::in);          
      ifstream SampHeuristicBalanceFile(
            SampHeuristicBalanceFileName[FileCounter].c_str(),ios::in); 
      if ( ! SampFile0 || ! SampFile1 || ! SampHeuristicBalanceFile) {
         cout << endl;
	 if (! SampFile0) {
            cout << "Warning: " << SampFileName0[FileCounter] 
	         << " could not be open." << endl;
	 }           
         if ( ! SampFile1) {
            cout << endl;
            cout << "Warning: " << SampFileName1[FileCounter] 
      	      << " could not be open." << endl;     
	 }     	 
         if ( ! SampHeuristicBalanceFile) {
            cout << endl;
            cout << "Warning: " << SampHeuristicBalanceFileName[FileCounter] 
      	      << " could not be open." << endl;     
	 }  
      }     
      else { //## They do exist

         //SampFile0.close();
         //SampFile1.close();
	 //SampHeuristicBalanceFile.close();

         *LogFile << "   "<< SampFileName0[FileCounter] 
                 << "   "<< SampFileName1[FileCounter] 
		 << "   "<< SampHeuristicBalanceFileName[FileCounter]<< endl;  
		 
         oHistMeanParam2D[FileCounter] = new Histogram2D(HistInFileName);
         FileCounter++;
      }	 
      SampFile0.close();
      SampFile1.close();
      SampHeuristicBalanceFile.close();
   }
   InFile.close();


   //## First, get the pdf for each case
   double Sample0, Sample1;
   double LikelihoodRatio0, LikelihoodRatio1;      
   double LikelihoodRatioHeuristicBalance;
   double HeuristicBalanceWeight;
   double TrueCombinedWeight;
   for (int ii = 0; ii < FileCounter; ii++) {
      //oHist_pdf2D[ii] = new Histogram2D(HistInFileName.c_str()); 
                                      //  x_min,x_max,y_min,y_max);
            
      cout << "File Name: "<< ii <<" = " <<SampFileName0[ii]  
                                 << "  " <<SampFileName1[ii]
				 << "  " <<SampHeuristicBalanceFileName[ii] 
				 << endl;
      ifstream SampFile0(SampFileName0[ii].c_str(),ios::in);      
      if ( ! SampFile0) {
         cout << endl;
         cout << "Warning: " << SampFileName0[ii] << " could not be open."
              << endl;
	  continue;       
      }             
      
      ifstream SampFile1(SampFileName1[ii].c_str(),ios::in);        
      if ( ! SampFile1) {
         cout << endl;
         cout << "Warning: " << SampFileName1[ii] << " could not be open."
              << endl;
	  continue;      
      }                
      
      ifstream SampHeuristicBalanceFile(
            SampHeuristicBalanceFileName[ii].c_str(),ios::in);          
      if ( ! SampHeuristicBalanceFile) {
         cout << endl;
         cout << "Warning: " << SampHeuristicBalanceFileName[ii] 
	      << " could not be open." << endl;
	  continue;      
      }    
      while ( SampFile0 >> Sample0) {
         SampFile0 >> LikelihoodRatio0;
	 SampFile1 >> Sample1;
	 SampFile1 >> LikelihoodRatio1;
	 SampHeuristicBalanceFile >> LikelihoodRatioHeuristicBalance;
	 SampHeuristicBalanceFile >> HeuristicBalanceWeight;
	 SampHeuristicBalanceFile >> TrueCombinedWeight;	 

	 if (  LikelihoodRatio0 != LikelihoodRatio1
	     || abs( (LikelihoodRatio0 - LikelihoodRatioHeuristicBalance)
	      /LikelihoodRatio0) > 1e-4 ) {
	    cout << "ERROR! From: Histogram2D::MultipleSelectMaxRelNumHitsImport(.)"
	         << endl << "The likelihoodRatio of the files "
		 <<SampFileName0[ii] << ", " << SampFileName1[ii] 
		 << " and " << SampHeuristicBalanceFileName[ii]<< endl
		 << "are different. They must be equal."  << endl;
            exit(1);		   
	 }
         if (LikelihoodRatio0 > 0) {
            //## Samples can be filtered by the user 
            //#  doing the LikelihoodRatio < 0
            if (Convert2dB_Scale) {
               AddSample( sgn(Convert2dB_Scale)
                                                 *Linear2dB(Sample0),
                                           sgn(Convert2dB_Scale)
                                                 *Linear2dB(Sample1),						 
						 TrueCombinedWeight);
						 //LikelihoodRatio0);
               oHistMeanParam2D[ii]->AddSample( sgn(Convert2dB_Scale)
                                                 *Linear2dB(Sample0),
                                           sgn(Convert2dB_Scale)
                                                 *Linear2dB(Sample1),						 
                              HeuristicBalanceWeight * LikelihoodRatio0);
						 //TrueCombinedWeight);						 
            }
            else {
              AddSample(Sample0,Sample1, TrueCombinedWeight);
              oHistMeanParam2D[ii]->AddSample(Sample0,Sample1, 
                          HeuristicBalanceWeight * LikelihoodRatio0);	      
	         //   TrueCombinedWeight);	      
	        //LikelihoodRatio0);
            }   
         }
      }               
      SampFile0.close();
      SampFile1.close();
      SampHeuristicBalanceFile.close();
      oHistMeanParam2D[ii]->ObtainConfidence();      
   } 
   
   for (int x_index = 0; x_index < NumBinsX; x_index++) {
      for (int y_index = 0; y_index < NumBinsY; y_index++) {      
         double VarianceBin = 0;      
         for (int jj = 0; jj < FileCounter; jj++) {  	   	   
	    //cout << x_index << "  " << y_index << "  "
	    //     << jj << "  " << sq(oHistMeanParam2D[jj]->Get_pdfConfidence(
	    //      x_index, y_index) ) << endl;
            VarianceBin += sq(oHistMeanParam2D[jj]->Get_pdfConfidence(
	          x_index, y_index) );
	 }
	 //## The confidence interval already came normalized from
	 //## the method: ObtainConfidence().
         pdfConfidence[GetIndex(x_index, y_index)] = sqrt(VarianceBin);
      } 
   }
   //## Prevent an incorrect computation of the variance in Obtain_pdf()
   PreviouslyComputed_pdfConfidenceFlag = 1;

   for (int ii = 0; ii < FileCounter; ii++) {
      delete oHistMeanParam2D[ii];
   }

   *LogFile << "The pdf from "<< InFileName.c_str()<<" was obtained "
           << "from a heuristically balanced merging process." << endl;
   
} //## End 

//########################################################

/*! Computes the 2D probability density function 
f(X=x,Y=y| z_min < Z < z_max), 
the joint probability density for X and Y given that z_min < Z < z_max. 
It is assumed
that samples of the random variables X,Y,Z have been obtained from an 
importance sampling simulation, and that the names of the files containing
the samples are stored in the file InFileName. 

The input parameter Convert2dB_Scale converts the pdfs of X and Y to a 
plus/minus dB scale before adding them to this 2D histogram.

We only add a (X,Y) sample to this 2D histogram if z_min < Z < z_max. 

Note: I'm not sure whether or not the Confidence Interval
is correctly computed for the case that Prob(z_min < Z < z_max) != 1

By JZ, 15/11/2004, modified 23/2/2005 and
adapted from code by Ivan T. Lima Jr. and Aurenice O. Lima (7/28/02)

*/

void Histogram2D::MergeMultipleConditionalHeuristicBalance(string InFileName,
						  double z_min, double z_max,
                                                        int Convert2dB_Scale)
{

   Clear();

   ifstream InFile(InFileName.c_str(),ios::in);

   if ( ! InFile) 
    {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
    }


   //## It can merge up to 100 biased files
   #define MAX_FILES_HEURISTIC_BALANCE 100
   
   string SampFileNameX[MAX_FILES_HEURISTIC_BALANCE];
   string SampFileNameY[MAX_FILES_HEURISTIC_BALANCE];
   string SampFileNameZ[MAX_FILES_HEURISTIC_BALANCE];
  
   string SampHeuristicBalanceFileName[MAX_FILES_HEURISTIC_BALANCE];
   
   Histogram2D* oHistMeanParam2D[MAX_FILES_HEURISTIC_BALANCE];

   *LogFile << "List of file names to be merged:" << endl;
   int FileCounter = 0;
   while (InFile >> SampFileNameX[FileCounter]) {
      InFile >> SampFileNameY[FileCounter];
      InFile >> SampFileNameZ[FileCounter];
      InFile >> SampHeuristicBalanceFileName[FileCounter];

      ifstream SampFileX(SampFileNameX[FileCounter].c_str(),ios::in);
      ifstream SampFileY(SampFileNameY[FileCounter].c_str(),ios::in);
      ifstream SampFileZ(SampFileNameZ[FileCounter].c_str(),ios::in);

      ifstream SampHeuristicBalanceFile(
            SampHeuristicBalanceFileName[FileCounter].c_str(),ios::in);

      if ( ! SampFileX || ! SampFileY || 
	   ! SampFileZ || ! SampHeuristicBalanceFile) 
       {
         cout << endl; 
         if (! SampFileX) 
          {
            cout << "Warning: " << SampFileNameX[FileCounter]
                 << " could not be opened." << endl;
          }
        
         if ( ! SampFileY) 
          {
            cout << endl;
            cout << "Warning: " << SampFileNameY[FileCounter]
              << " could not be opened." << endl;
          }

         if ( ! SampFileZ) 
          {
            cout << endl;
            cout << "Warning: " << SampFileNameZ[FileCounter]
              << " could not be opened." << endl;
          }

         if ( ! SampHeuristicBalanceFile) 
          {
            cout << endl;
            cout << "Warning: " << SampHeuristicBalanceFileName[FileCounter]
              << " could not be open." << endl;
          }
      }
      else { //## They do exist

        *LogFile << "   "<< SampFileNameX[FileCounter]
                 << "   "<< SampFileNameY[FileCounter]
                 << "   "<< SampFileNameZ[FileCounter]
                 << "   "<< SampHeuristicBalanceFileName[FileCounter]<< endl;

         oHistMeanParam2D[FileCounter] = new Histogram2D(HistInFileName);
         FileCounter++;
      }

      SampFileX.close();
      SampFileY.close();
      SampFileZ.close();
      SampHeuristicBalanceFile.close();
   }

   InFile.close();


   //## First, get the pdf for each case

   double SampleX, SampleY, SampleZ;
   double LikelihoodRatioX, LikelihoodRatioY, LikelihoodRatioZ;
   double LikelihoodRatioHeuristicBalance;
   double HeuristicBalanceWeight;
   double TrueCombinedWeight;

   for (int ii = 0; ii < FileCounter; ii++) 
    {
      cout << "File Name: "<< ii <<" = " <<SampFileNameX[ii]
                                 << "  " <<SampFileNameY[ii]
                                 << "  " <<SampFileNameZ[ii]
                                 << "  " <<SampHeuristicBalanceFileName[ii]
                                 << endl;

      ifstream SampFileX(SampFileNameX[ii].c_str(),ios::in);

      if ( ! SampFileX) 
       {
         cout << endl;
         cout << "Warning: " << SampFileNameX[ii] << " could not be open."
              << endl;
          continue;
       }

      ifstream SampFileY(SampFileNameY[ii].c_str(),ios::in);
    
      if ( ! SampFileY) 
       {
         cout << endl;
         cout << "Warning: " << SampFileNameY[ii] << " could not be open."
              << endl;
          continue;
      }


      ifstream SampFileZ(SampFileNameZ[ii].c_str(),ios::in);
    
      if ( ! SampFileZ) 
       {
         cout << endl;
         cout << "Warning: " << SampFileNameZ[ii] << " could not be open."
              << endl;
          continue;
      }


      ifstream SampHeuristicBalanceFile(
            SampHeuristicBalanceFileName[ii].c_str(),ios::in);

      if ( ! SampHeuristicBalanceFile) 
       {
         cout << endl;
         cout << "Warning: " << SampHeuristicBalanceFileName[ii]
              << " could not be open." << endl;
          continue;
      }

      while ( SampFileX >> SampleX) 
       {
         SampFileX >> LikelihoodRatioX;
         SampFileY >> SampleY;
         SampFileY >> LikelihoodRatioY;
         SampFileZ >> SampleZ;
         SampFileZ >> LikelihoodRatioZ;

         SampHeuristicBalanceFile >> LikelihoodRatioHeuristicBalance;
         SampHeuristicBalanceFile >> HeuristicBalanceWeight;
         SampHeuristicBalanceFile >> TrueCombinedWeight;

         if (  LikelihoodRatioX != LikelihoodRatioY || 
               LikelihoodRatioX != LikelihoodRatioZ ||
               LikelihoodRatioY != LikelihoodRatioZ ||
               abs( (LikelihoodRatioX - LikelihoodRatioHeuristicBalance)
              /LikelihoodRatioX) > 1e-4 ) 
            {
         cout << "ERROR! Histogram2D::MergeMultipleConditionalHeuristicBalance"
              << endl << "The LikelihoodRatio of the files "
              <<SampFileNameX[ii] << ", " << SampFileNameY[ii]
	      << ", " << SampFileNameZ[ii]
              << " and " << SampHeuristicBalanceFileName[ii]<< endl
              << "are different. They must be equal."  << endl;
            exit(1);
         }
  
       if(LikelihoodRatioX > 0 && SampleZ > z_min && SampleZ < z_max) 
        {
            //## Samples can be filtered by the user 
            //#  doing the LikelihoodRatio < 0

            if (Convert2dB_Scale) 
             {
               AddSample( sgn(Convert2dB_Scale)*Linear2dB(SampleX),
                          sgn(Convert2dB_Scale)*Linear2dB(SampleY),
                                               TrueCombinedWeight);
                                            
               oHistMeanParam2D[ii]->
                AddSample( sgn(Convert2dB_Scale)*Linear2dB(SampleX),
                           sgn(Convert2dB_Scale)*Linear2dB(SampleY),
                           HeuristicBalanceWeight * LikelihoodRatioX);
                                                
             }
            else 
             {
              AddSample(SampleX,SampleY, TrueCombinedWeight);
              oHistMeanParam2D[ii]->AddSample(SampleX,SampleY,
                          HeuristicBalanceWeight * LikelihoodRatioX);
                 
             }
         } // end if LikelihoodRatioX > 0 && SampleZ in (z_min,z_max)
      } // end while

      SampFileX.close();
      SampFileY.close();
      SampFileZ.close();
      SampHeuristicBalanceFile.close();
      oHistMeanParam2D[ii]->ObtainConfidence();
   }

   for (int x_index = 0; x_index < NumBinsX; x_index++) {
      for (int y_index = 0; y_index < NumBinsY; y_index++) {
         double VarianceBin = 0;
         for (int jj = 0; jj < FileCounter; jj++) {
           VarianceBin += sq(oHistMeanParam2D[jj]->Get_pdfConfidence(
                             x_index, y_index) );
         }

         //## The confidence interval already came normalized from
         //## the method: ObtainConfidence().
         pdfConfidence[GetIndex(x_index, y_index)] = sqrt(VarianceBin);
      }
   }
   //## Prevent an incorrect computation of the variance in Obtain_pdf()
   PreviouslyComputed_pdfConfidenceFlag = 1;

   for (int ii = 0; ii < FileCounter; ii++) {
      delete oHistMeanParam2D[ii];
   }

   *LogFile << "The conditional pdf from "<< InFileName.c_str()
	    <<" was obtained "
            << "from a heuristically balanced merging process." << endl;

} //## end MergeMultipleConditionalHeuristicBalance

// #####################################################

//########################################################

/*! Computes the 2D probability density function 
     f(X=x,Y=y| z_min < Z < z_max, w_min < W < w_max), 
     Samples of the random variables X,Y,Z,W have been obtained from an 
     importance sampling simulation, and that the names of the files containing
     the samples are stored in the file InFileName. 

The input parameter Convert2dB_Scale converts the pdfs of X and Y to a 
plus/minus dB scale before adding them to this 2D histogram.

We only add a (X,Y) sample to this 2D histogram if z_min < Z < z_max
and w_min < W < w_max

Note: I'm not sure whether or not the Confidence Interval
is correctly computed.

By JZ, 23/2/2005 and
adapted from code by Ivan T. Lima Jr. and Aurenice O. Lima (7/28/02)

*/

void Histogram2D::MergeMultipleConditionalHeuristicBalance(string InFileName,
						  double z_min, double z_max,
						  double w_min, double w_max,
	                                                int Convert2dB_Scale)
{

   Clear();

   ifstream InFile(InFileName.c_str(),ios::in);

   if ( ! InFile) 
    {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
    }


   //## It can merge up to 100 biased files
   #define MAX_FILES_HEURISTIC_BALANCE 100
   
   string SampFileNameX[MAX_FILES_HEURISTIC_BALANCE];
   string SampFileNameY[MAX_FILES_HEURISTIC_BALANCE];
   string SampFileNameZ[MAX_FILES_HEURISTIC_BALANCE];
   string SampFileNameW[MAX_FILES_HEURISTIC_BALANCE];

   string SampHeuristicBalanceFileName[MAX_FILES_HEURISTIC_BALANCE];
   
   Histogram2D* oHistMeanParam2D[MAX_FILES_HEURISTIC_BALANCE];

   *LogFile << "List of file names to be merged:" << endl;
   int FileCounter = 0;
   while (InFile >> SampFileNameX[FileCounter]) {
      InFile >> SampFileNameY[FileCounter];
      InFile >> SampFileNameZ[FileCounter];
      InFile >> SampFileNameW[FileCounter];
      InFile >> SampHeuristicBalanceFileName[FileCounter];

      ifstream SampFileX(SampFileNameX[FileCounter].c_str(),ios::in);
      ifstream SampFileY(SampFileNameY[FileCounter].c_str(),ios::in);
      ifstream SampFileZ(SampFileNameZ[FileCounter].c_str(),ios::in);
      ifstream SampFileW(SampFileNameZ[FileCounter].c_str(),ios::in);

      ifstream SampHeuristicBalanceFile(
            SampHeuristicBalanceFileName[FileCounter].c_str(),ios::in);

      if ( ! SampFileX || ! SampFileY || 
	   ! SampFileZ || ! SampFileW || ! SampHeuristicBalanceFile) 
       {
         cout << endl; 
         if (! SampFileX) 
          {
            cout << "Warning: " << SampFileNameX[FileCounter]
                 << " could not be opened." << endl;
          }
        
         if ( ! SampFileY) 
          {
            cout << endl;
            cout << "Warning: " << SampFileNameY[FileCounter]
              << " could not be opened." << endl;
          }

         if ( ! SampFileZ) 
          {
            cout << endl;
            cout << "Warning: " << SampFileNameZ[FileCounter]
              << " could not be opened." << endl;
          }

         if ( ! SampFileW) 
          {
            cout << endl;
            cout << "Warning: " << SampFileNameW[FileCounter]
              << " could not be opened." << endl;
          }


         if ( ! SampHeuristicBalanceFile) 
          {
            cout << endl;
            cout << "Warning: " << SampHeuristicBalanceFileName[FileCounter]
              << " could not be open." << endl;
          }
      }
      else { //## They do exist

        *LogFile << "   "<< SampFileNameX[FileCounter]
                 << "   "<< SampFileNameY[FileCounter]
                 << "   "<< SampFileNameZ[FileCounter]
                 << "   "<< SampFileNameW[FileCounter]
                 << "   "<< SampHeuristicBalanceFileName[FileCounter]<< endl;

         oHistMeanParam2D[FileCounter] = new Histogram2D(HistInFileName);
         FileCounter++;
      }

      SampFileX.close();
      SampFileY.close();
      SampFileZ.close();
      SampFileW.close();
      SampHeuristicBalanceFile.close();
   }

   InFile.close();


   //## First, get the pdf for each case

   double SampleX, SampleY, SampleZ, SampleW;
   double LikelihoodRatioX, LikelihoodRatioY; 
   double LikelihoodRatioZ, LikelihoodRatioW;
   double LikelihoodRatioHeuristicBalance;
   double HeuristicBalanceWeight;
   double TrueCombinedWeight;

   for (int ii = 0; ii < FileCounter; ii++) 
    {
      cout << "File Name: "<< ii <<" = " <<SampFileNameX[ii]
                                 << "  " <<SampFileNameY[ii]
                                 << "  " <<SampFileNameZ[ii]
                                 << "  " <<SampFileNameW[ii]
                                 << "  " <<SampHeuristicBalanceFileName[ii]
                                 << endl;

      ifstream SampFileX(SampFileNameX[ii].c_str(),ios::in);

      if ( ! SampFileX) 
       {
         cout << endl;
         cout << "Warning: " << SampFileNameX[ii] << " could not be open."
              << endl;
          continue;
       }

      ifstream SampFileY(SampFileNameY[ii].c_str(),ios::in);
    
      if ( ! SampFileY) 
       {
         cout << endl;
         cout << "Warning: " << SampFileNameY[ii] << " could not be open."
              << endl;
          continue;
      }


      ifstream SampFileZ(SampFileNameZ[ii].c_str(),ios::in);
    
      if ( ! SampFileZ) 
       {
         cout << endl;
         cout << "Warning: " << SampFileNameZ[ii] << " could not be open."
              << endl;
          continue;
      }

      ifstream SampFileW(SampFileNameW[ii].c_str(),ios::in);
    
      if ( ! SampFileW) 
       {
         cout << endl;
         cout << "Warning: " << SampFileNameW[ii] << " could not be open."
              << endl;
          continue;
      }

      ifstream SampHeuristicBalanceFile(
            SampHeuristicBalanceFileName[ii].c_str(),ios::in);

      if ( ! SampHeuristicBalanceFile) 
       {
         cout << endl;
         cout << "Warning: " << SampHeuristicBalanceFileName[ii]
              << " could not be open." << endl;
          continue;
      }

      while ( SampFileX >> SampleX) 
       {
         SampFileX >> LikelihoodRatioX;
         SampFileY >> SampleY;
         SampFileY >> LikelihoodRatioY;
         SampFileZ >> SampleZ;
         SampFileZ >> LikelihoodRatioZ;
         SampFileW >> SampleW;
         SampFileW >> LikelihoodRatioW;

         SampHeuristicBalanceFile >> LikelihoodRatioHeuristicBalance;
         SampHeuristicBalanceFile >> HeuristicBalanceWeight;
         SampHeuristicBalanceFile >> TrueCombinedWeight;

         if (  LikelihoodRatioX != LikelihoodRatioY || 
               LikelihoodRatioX != LikelihoodRatioZ ||
               LikelihoodRatioX != LikelihoodRatioW ||
               abs( (LikelihoodRatioX - LikelihoodRatioHeuristicBalance)
              /LikelihoodRatioX) > 1e-4 ) 
            {
         cout << "ERROR! Histogram2D::MergeMultipleConditionalHeuristicBalance"
              << endl << "The LikelihoodRatio of the files "
              <<SampFileNameX[ii] << ", " << SampFileNameY[ii]
	      << ", " << SampFileNameZ[ii]
 	      << ", " << SampFileNameW[ii]
             << " and " << SampHeuristicBalanceFileName[ii]<< endl
              << "are different. They must be equal."  << endl;
            exit(1);
         }
  
       if(LikelihoodRatioX > 0 
	  && SampleZ > z_min && SampleZ < z_max
	  && SampleW > w_min && SampleW < w_max) 
        {
            //## Samples can be filtered by the user 
            //#  doing the LikelihoodRatio < 0

            if (Convert2dB_Scale) 
             {
               AddSample( sgn(Convert2dB_Scale)*Linear2dB(SampleX),
                          sgn(Convert2dB_Scale)*Linear2dB(SampleY),
                                               TrueCombinedWeight);
                                            
               oHistMeanParam2D[ii]->
                AddSample( sgn(Convert2dB_Scale)*Linear2dB(SampleX),
                           sgn(Convert2dB_Scale)*Linear2dB(SampleY),
                           HeuristicBalanceWeight * LikelihoodRatioX);
                                                
             }
            else 
             {
              AddSample(SampleX,SampleY, TrueCombinedWeight);
              oHistMeanParam2D[ii]->AddSample(SampleX,SampleY,
                          HeuristicBalanceWeight * LikelihoodRatioX);
                 
             }
         } // end if LikelihoodRatioX > 0 && SampleZ in (z_min,z_max)
      } // end while

      SampFileX.close();
      SampFileY.close();
      SampFileZ.close();
      SampFileW.close();
      SampHeuristicBalanceFile.close();
      oHistMeanParam2D[ii]->ObtainConfidence();
   }

   for (int x_index = 0; x_index < NumBinsX; x_index++) {
      for (int y_index = 0; y_index < NumBinsY; y_index++) {
         double VarianceBin = 0;
         for (int jj = 0; jj < FileCounter; jj++) {
           VarianceBin += sq(oHistMeanParam2D[jj]->Get_pdfConfidence(
                             x_index, y_index) );
         }

         //## The confidence interval already came normalized from
         //## the method: ObtainConfidence().
         pdfConfidence[GetIndex(x_index, y_index)] = sqrt(VarianceBin);
      }
   }
   //## Prevent an incorrect computation of the variance in Obtain_pdf()
   PreviouslyComputed_pdfConfidenceFlag = 1;

   for (int ii = 0; ii < FileCounter; ii++) {
      delete oHistMeanParam2D[ii];
   }

   *LogFile << "The conditional pdf from "<< InFileName.c_str()
	    <<" was obtained "
            << "from a heuristically balanced merging process." << endl;

} //## end MergeMultipleConditionalHeuristicBalance


//########################################################
//##### Imports the data from files created by the histogram class
//##    Convert2dB_Scale also determines the sign of the conversion to dB
//##    Actually, what you get from this is the conditional expectation pdf
//##    YOU SHOULD NOT USE THIS. USE THE HEURISTIC BALANCE WHENEVER YOU CAN
//##    By: Ivan T. Lima Jr. and Aurenice O. Lima (7/28/02)
void Histogram2D::GetLevelCurvesMultipleSelectMaxRelNumHitsImport(string InFileName,
                                                  int Convert2dB_Scale,
						  int ConvertParam2dB_Scale)
{
   //## Structure of InFileName
   //##  SampDataFileName    
         
   //## Clear the histogram, just in case...
   Clear();
   
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }     

   //## It can merge up to 100 biased files
   string SampFileName0[100], SampFileName1[100];
   string SampFileName2[100];
   Histogram2D* oHist_pdf2D[100];
   Histogram2D* oHistMeanParam2D[100];   
      
   *LogFile << "List of file names to be merged:" << endl;
   int FileCounter = 0;
   while (InFile >> SampFileName0[FileCounter]) {
      InFile >> SampFileName1[FileCounter];      
      InFile >> SampFileName2[FileCounter];       
      ifstream SampFile0(SampFileName0[FileCounter].c_str(),ios::in);       
      ifstream SampFile1(SampFileName1[FileCounter].c_str(),ios::in);          
      ifstream SampFile2(SampFileName2[FileCounter].c_str(),ios::in);     
      if ( ! SampFile0 || ! SampFile1 || ! SampFile2) {
         cout << endl;
	 if (! SampFile0) {
            cout << "Warning: " << SampFileName0[FileCounter] 
	         << " could not be open." << endl;
	 }           
         if ( ! SampFile1) {
            cout << endl;
            cout << "Warning: " << SampFileName1[FileCounter] 
      	      << " could not be open." << endl;     
	 }     	 
        if ( ! SampFile2) {
            cout << endl;
            cout << "Warning: " << SampFileName2[FileCounter] 
      	      << " could not be open." << endl;     
	 }     	 	 
      }     
      else { //## They do exist

         SampFile0.close();
         SampFile1.close();
         SampFile2.close();	 

         *LogFile << "   "<< SampFileName0[FileCounter] 
                 << "   "<< SampFileName1[FileCounter] 
		 << "   "<< SampFileName2[FileCounter] << endl;  

         FileCounter++;
      }	 
   }
   InFile.close();

   //## First, get the pdf for each case
   double Sample0, Sample1;
   double Sample2;
   double LikelihoodRatio0, LikelihoodRatio1;      
   double LikelihoodRatio2;
   for (int ii = 0; ii < FileCounter; ii++) {
      oHist_pdf2D[ii] = new Histogram2D(HistInFileName.c_str()); 
      oHistMeanParam2D[ii] = new Histogram2D(HistInFileName.c_str()); 
            
      cout << "File Name: "<< ii <<" = " <<SampFileName0[ii]  
                                 << "  " <<SampFileName1[ii] << endl;
      ifstream SampFile0(SampFileName0[ii].c_str(),ios::in);      
      if ( ! SampFile0) {
         cout << endl;
         cout << "Warning: " << SampFileName0[ii] << " could not be open."
              << endl;
	  continue;       
      }             
      ifstream SampFile1(SampFileName1[ii].c_str(),ios::in);        
      if ( ! SampFile1) {
         cout << endl;
         cout << "Warning: " << SampFileName1[ii] << " could not be open."
              << endl;
	  continue;      
      }                
      ifstream SampFile2(SampFileName2[ii].c_str(),ios::in);            
      if ( ! SampFile2) {
         cout << endl;
         cout << "Warning: " << SampFileName2[ii] << " could not be open."
              << endl;
	  continue;      
      }    
      while ( SampFile0 >> Sample0) {
         SampFile0 >> LikelihoodRatio0;
	 SampFile1 >> Sample1;
	 SampFile1 >> LikelihoodRatio1;
	 SampFile2 >> Sample2;
	 SampFile2 >> LikelihoodRatio2;	 
	 if (   LikelihoodRatio0 != LikelihoodRatio1 
	     || LikelihoodRatio0 != LikelihoodRatio2) {
	    cout << "ERROR! From: Histogram2D::MultipleSelectMaxRelNumHitsImport(.)"
	         << endl << "The likelihoodRatio of the files "
		 <<SampFileName0[ii] << " and " << SampFileName1[ii] 
		 << " and " << SampFileName2[ii] << endl
		 << "are different. They must be equal."  << endl;
            exit(1);		   
	 }
         if (LikelihoodRatio0 > 0) {
            //## Samples can be filtered by the user 
            //#  doing the LikelihoodRatio < 0
            if (Convert2dB_Scale) {
               oHist_pdf2D[ii]->AddSample( sgn(Convert2dB_Scale)
                                                 *Linear2dB(Sample0),
                                           sgn(Convert2dB_Scale)
                                                 *Linear2dB(Sample1),						 
						 LikelihoodRatio0);
               if (ConvertParam2dB_Scale) {
                  oHistMeanParam2D[ii]->AddSample( sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample0),
                                              sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample1),						 
   			   sgn(ConvertParam2dB_Scale)
			   *( - Linear2dB_Limited(1./Sample2,20.) )
   			   *1 +0*LikelihoodRatio0);
               } 
	       else {
                  oHistMeanParam2D[ii]->AddSample( sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample0),
                                              sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample1),						 
   			   Sample2*1 +0*LikelihoodRatio0);	       
	       }  
            }
            else {
              oHist_pdf2D[ii]->AddSample(Sample0,Sample1,LikelihoodRatio0);
              if (ConvertParam2dB_Scale) {
                  oHistMeanParam2D[ii]->AddSample(Sample0,Sample1,						 
   			   sgn(ConvertParam2dB_Scale)
			   *( - Linear2dB_Limited(1./Sample2,20.) )
   			   *1+0*LikelihoodRatio0);
               } 	      
	       else {
                  oHistMeanParam2D[ii]->AddSample(Sample0,Sample1,						 
   			   Sample2
   			   *1+0*LikelihoodRatio0);	       
	       }
            }   
 	    	    
         }
      }               
      SampFile0.close();
      SampFile1.close();
      SampFile2.close();      
   } 
   
   /******
   //## Second, get the total number of samples per bin
   long *qtSamplesBinTmp = new long[NumBinsX*NumBinsY];
   for (int x_index = 0; x_index < NumBinsX; x_index++) {
      for (int y_index = 0; y_index < NumBinsY; y_index++) {      
         qtSamplesBinTmp[y_index + NumBinsY*x_index] = 0;
      }
   }   	 
   for (int ii = 0; ii < FileCounter; ii++) {
      for (int x_index = 0; x_index < NumBinsX; x_index++) {
         for (int y_index = 0; y_index < NumBinsY; y_index++) {      
            qtSamplesBinTmp[y_index + NumBinsY*x_index] += 
   	       oHist_pdf2D[ii]->GetQtSamplesBin(x_index,y_index);
         }
      }   	 
   }   
   *******/

   //oHist_pdf2D[0]->WriteFile_pdf2D("MergedLevelCurves2D_2.dat");
   //oHistMeanParam2D[0]->WriteFile_pdf2D("MergedLevelCurves2DParam_2.dat");
   //exit(1);

   Clear();

   //## Third, select the bin that had the maximum relative
   //#  number of hits from all the distributions
   //#  I have to do it in such a way that Obtain_pdf
   //#  won't mess things up.
   double MaxRelNumHits;
   for (int x_index = 0; x_index < NumBinsX; x_index++) {
      for (int y_index = 0; y_index < NumBinsY; y_index++) {      
         MaxRelNumHits = 0;
         for (int ii = 0; ii < FileCounter; ii++) {
            if ( (double) oHist_pdf2D[ii]->GetQtSamplesBin(x_index,y_index)
                         /oHist_pdf2D[ii]->GetQtSamples() 
             > MaxRelNumHits ) {
            MaxRelNumHits =  (double) 
                  oHist_pdf2D[ii]->GetQtSamplesBin(x_index,y_index)
                  /oHist_pdf2D[ii]->GetQtSamples();              
	    if (oHistMeanParam2D[ii]->GetQtSamplesBin(x_index,y_index) != 0) { 	  
  	       //## I need to eliminate the normalization in the plot of the
	       //## pdf and of the standard deviation. The pdf will be 
	       //## normalized during the plot,
               SetBin(x_index,y_index, (double)
               	    oHistMeanParam2D[ii]->GetBin(x_index,y_index)
	            /oHistMeanParam2D[ii]->GetQtSamplesBin(x_index,y_index)
                   /( NumBinsX/(x_max - x_min)*NumBinsY/(y_max - y_min) )  );		    
               pdfConfidence[GetIndex(x_index, y_index)] = 
	          oHistMeanParam2D[ii]->Get_pdfConfidence(x_index, y_index)
                 /( NumBinsX/(x_max - x_min)*NumBinsY/(y_max - y_min) );
	    }	    
//            SetBin(x_index,y_index, (double)
//             	    oHistMeanParam2D[ii]->GetBin(x_index,y_index)
//		    / oHist_pdf2D[ii]->GetBin(x_index,y_index) ); 
////	            /oHistMeanParam2D[ii]->GetQtSamples() );
////            SetBin(x_index,y_index, (double)
////             	    oHist_pdf2D[ii]->GetBin(x_index,y_index)
////	            /oHist_pdf2D[ii]->GetQtSamples() );
            }
         }  
      }   
      
   }   

   //## Fourth, and finally, force the number of samples to be 
   //#  unitary to get the right statistics, since all the 
   //#  pdfs used were already normalized.
   //#  Obviously, I need to denormilize the Bins, since the
   //#  Obtain_pdf method will do it
   NumSamplesTotal = 1;
   NumSamplesInRange = 1;    
     

   *LogFile << "The level curves of the parameter selected in "
           << InFileName.c_str()<<" was obtained "
           << "from a merging process." << endl;
//   *LogFile << "Thus, don't expect fully functionality from the other methods."
//           << endl;
   
} //## End 





//########################################################
//##### Imports the data from files created by the histogram class
//##    Convert2dB_Scale also determines the sign of the conversion to dB
//##    Actually, what you get from this is the conditional joint pdf
//##    By: Ivan T. Lima Jr. and Aurenice O. Lima (7/29/02)
//##    Computation of standard deviation of the bins by Aurenice O. Lima in 10/15/02.
void Histogram2D::GetLevelCurvesMultipleHeuristicBalanceImport(string InFileName,
                                                  int Convert2dB_Scale,
						  int ConvertParam2dB_Scale)
{
   //## Structure of InFileName
   //##  SampDataFileName    
         
   //## Clear the histogram, just in case...
   Clear();
   
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }     

   //## It can merge up to 100 biased files
   #define MAX_FILES_HEURISTIC_BALANCE 100   
   string SampFileName0[MAX_FILES_HEURISTIC_BALANCE];
   string SampFileName1[MAX_FILES_HEURISTIC_BALANCE];
   string SampFileName2[MAX_FILES_HEURISTIC_BALANCE];
   string SampHeuristicBalanceFileName[MAX_FILES_HEURISTIC_BALANCE];
      
   //Histogram2D* oHist_pdf2D[MAX_FILES_HEURISTIC_BALANCE];
   Histogram2D* oHistMeanParam2D[MAX_FILES_HEURISTIC_BALANCE];   
      
   *LogFile << "List of file names to be merged:" << endl;
   int FileCounter = 0;
   while (InFile >> SampFileName0[FileCounter]) {
      InFile >> SampFileName1[FileCounter];      
      InFile >> SampFileName2[FileCounter];       
      InFile >> SampHeuristicBalanceFileName[FileCounter];   
      
      ifstream SampFile0(SampFileName0[FileCounter].c_str(),ios::in);       
      ifstream SampFile1(SampFileName1[FileCounter].c_str(),ios::in);          
      ifstream SampFile2(SampFileName2[FileCounter].c_str(),ios::in);     
      ifstream SampHeuristicBalanceFile(
            SampHeuristicBalanceFileName[FileCounter].c_str(),ios::in);       
      if (   ! SampFile0 || ! SampFile1 || ! SampFile2 
          || ! SampHeuristicBalanceFile) {
         cout << endl;
	 if (! SampFile0) {
            cout << "Warning: " << SampFileName0[FileCounter] 
	         << " could not be open." << endl;
	 }           
         if ( ! SampFile1) {
            cout << endl;
            cout << "Warning: " << SampFileName1[FileCounter] 
      	      << " could not be open." << endl;     
	 }     	 
        if ( ! SampFile2) {
            cout << endl;
            cout << "Warning: " << SampFileName2[FileCounter] 
      	      << " could not be open." << endl;     
	 }     	 	 
         if ( ! SampHeuristicBalanceFile) {
            cout << endl;
            cout << "Warning: " << SampHeuristicBalanceFileName[FileCounter] 
      	      << " could not be open." << endl;     
	 } 	 
      }     
      else { //## They do exist	 
         *LogFile << "   "<< SampFileName0[FileCounter] 
                 << "   "<< SampFileName1[FileCounter] 
		 << "   "<< SampFileName2[FileCounter]  
		 << "   "<< SampHeuristicBalanceFileName[FileCounter]<< endl;  

         oHistMeanParam2D[FileCounter] = new Histogram2D(HistInFileName);
         FileCounter++;
      }	 
      SampFile0.close();
      SampFile1.close();
      SampFile2.close();	 
      SampHeuristicBalanceFile.close();      
   }
   InFile.close();

   //## First, get the pdf for each case
   double Sample0, Sample1;
   double Sample2;
   double LikelihoodRatio0, LikelihoodRatio1;      
   double LikelihoodRatio2;
   double LikelihoodRatioHeuristicBalance;
   double HeuristicBalanceWeight;
   double TrueCombinedWeight;
   for (int ii = 0; ii < FileCounter; ii++) {
      //oHist_pdf2D[ii] = new Histogram2D(HistInFileName.c_str()); 
      //oHistMeanParam2D[ii] = new Histogram2D(HistInFileName.c_str()); 
            
      cout << "File Name: "<< ii <<" = " <<SampFileName0[ii]  
                                 << "  " <<SampFileName1[ii] 
                                 << "  " <<SampFileName2[ii] 
				 << "  " <<SampHeuristicBalanceFileName[ii] 
				 << endl;				 

      ifstream SampFile0(SampFileName0[ii].c_str(),ios::in);      
      if ( ! SampFile0) {
         cout << endl;
         cout << "Warning: " << SampFileName0[ii] << " could not be open."
              << endl;
	  continue;       
      }             
      ifstream SampFile1(SampFileName1[ii].c_str(),ios::in);        
      if ( ! SampFile1) {
         cout << endl;
         cout << "Warning: " << SampFileName1[ii] << " could not be open."
              << endl;
	  continue;      
      }                
      ifstream SampFile2(SampFileName2[ii].c_str(),ios::in);            
      if ( ! SampFile2) {
         cout << endl;
         cout << "Warning: " << SampFileName2[ii] << " could not be open."
              << endl;
	  continue;      
      }    
      ifstream SampHeuristicBalanceFile(
            SampHeuristicBalanceFileName[ii].c_str(),ios::in);          
      if ( ! SampHeuristicBalanceFile) {
         cout << endl;
         cout << "Warning: " << SampHeuristicBalanceFileName[ii] 
	      << " could not be open." << endl;
	  continue;      
      }    
      
      while ( SampFile0 >> Sample0) {
         SampFile0 >> LikelihoodRatio0;
	 SampFile1 >> Sample1;
	 SampFile1 >> LikelihoodRatio1;
	 SampFile2 >> Sample2;
	 SampFile2 >> LikelihoodRatio2;	 
	 SampHeuristicBalanceFile >> LikelihoodRatioHeuristicBalance;
	 SampHeuristicBalanceFile >> HeuristicBalanceWeight;
	 SampHeuristicBalanceFile >> TrueCombinedWeight;	 
	 if (   LikelihoodRatio0 != LikelihoodRatio1 
	     || LikelihoodRatio0 != LikelihoodRatio2
	     || abs( (LikelihoodRatio0 - LikelihoodRatioHeuristicBalance)
	      /LikelihoodRatio0 ) > 1e-4 ) {
	    cout << "ERROR! From: Histogram2D::MultipleSelectMaxRelNumHitsImport(.)"
	         << endl << "The likelihoodRatio of the files "
		 <<SampFileName0[ii] << ", " << SampFileName1[ii] 
		 << ", " << SampFileName2[ii] 
                 << " and " << SampHeuristicBalanceFileName[ii]<< endl		 
		 << "are different. They must be equal."  << endl;
            exit(1);		   
	 }
         if (LikelihoodRatio0 > 0) {
            //## Samples can be filtered by the user 
            //#  doing the LikelihoodRatio < 0
            if (Convert2dB_Scale) {
	       
               //oHist_pdf2D[ii]->AddSample( sgn(Convert2dB_Scale)
               //                                  *Linear2dB(Sample0),
               //                            sgn(Convert2dB_Scale)
               //                                  *Linear2dB(Sample1),						 
	       //				 LikelihoodRatio0);
               if (ConvertParam2dB_Scale) {
                  AddSample( sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample0),
                                              sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample1),						 
   			   sgn(ConvertParam2dB_Scale)
			   *( - Linear2dB_Limited(1./Sample2,20.) )
			   *1 +0*TrueCombinedWeight);
   			   //*1 +0*LikelihoodRatio0);
                  oHistMeanParam2D[ii]->AddSample( sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample0),
                                              sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample1),						 
   			   sgn(ConvertParam2dB_Scale)
			   *( - Linear2dB_Limited(1./Sample2,20.) )
			   *1 +0*TrueCombinedWeight);
               } 
	       else {
                  AddSample( sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample0),
                                              sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample1),						 
   			   Sample2*1 +0*TrueCombinedWeight);							    
   			  // Sample2*1 +0*LikelihoodRatio0);	       
                  oHistMeanParam2D[ii]->AddSample( sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample0),
                                              sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample1),						 
   			   Sample2*1 +0*TrueCombinedWeight);				  
	       }  
            }
            else {
              //oHist_pdf2D[ii]->AddSample(Sample0,Sample1,LikelihoodRatio0);
              if (ConvertParam2dB_Scale) {
                  AddSample(Sample0,Sample1,						 
   			   sgn(ConvertParam2dB_Scale)
			   *( - Linear2dB_Limited(1./Sample2,20.) )
			   *1+0*TrueCombinedWeight);
   			   //*1+0*LikelihoodRatio0);
                  oHistMeanParam2D[ii]->AddSample(Sample0,Sample1,						 
   			   sgn(ConvertParam2dB_Scale)
			   *( - Linear2dB_Limited(1./Sample2,20.) )
			   *1+0*TrueCombinedWeight);			   
               } 	      
	       else {
                  AddSample(Sample0,Sample1,						 
   			   Sample2
			   *1+0*TrueCombinedWeight);
   			   //*1+0*LikelihoodRatio0);
                  oHistMeanParam2D[ii]->AddSample(Sample0,Sample1,						 
   			   Sample2
			   *1+0*TrueCombinedWeight);			   	       
	       }
            } 
	     	    	    
         }
      }               
      SampFile0.close();
      SampFile1.close();
      SampFile2.close();      
      SampHeuristicBalanceFile.close();      
      //oHistMeanParam2D[ii]->ObtainConfidence();
   } 


  
   //## Computation of standard deviation of the bins by Aurenice O. Lima in 10/15/02.
   //## Here we want the variance of the bin. The computation of the confidence does not apply.
   //## Hence, we have to compute the variance
   //## considering each bin as an "individual probability space".
   //## Aurenice O. Lima in 10/15/02.
   for (int x_index = 0; x_index < NumBinsX; x_index++) {
      for (int y_index = 0; y_index < NumBinsY; y_index++) {      
         double MeanBin = 0;
	 double VarianceBin = 0;
         if ( GetQtSamplesBin(x_index, y_index) > 0) {
            MeanBin =  GetBin(x_index, y_index)
	                     /GetQtSamplesBin(x_index, y_index);
            if ( GetQtSamplesBin(x_index, y_index) > 1
	        &&  GetBinSquare(x_index, y_index)
		   /GetQtSamplesBin(x_index, y_index) 
	           -sq( GetBin(x_index, y_index)
		       /GetQtSamplesBin(x_index, y_index) ) > 0  ) {         			     
               VarianceBin =  ( GetBinSquare(x_index, y_index)
	                       /GetQtSamplesBin(x_index, y_index) 
	                   -sq( GetBin(x_index, y_index)
			       /GetQtSamplesBin(x_index, y_index) ) )
    		             *GetQtSamplesBin(x_index, y_index)
			     /(GetQtSamplesBin(x_index, y_index) - 1);
            }			       			       
         } else {			       
 	     if (x_index > 0 && y_index > 0) { //## Adjust the bin
               MeanBin = GetBin(x_index - 1,y_index - 1);	
             }		
         }	     
		
	 //## I need to eliminate the normalization in the plot of the
	 //## pdf and of the standard deviation. The pdf will be 
	 //## normalized during the plot,
         SetBin(x_index,y_index, MeanBin 
	        /( NumBinsX/(x_max - x_min)*NumBinsY/(y_max - y_min) ) );

	 //if ( GetQtSamplesBin(x_index, y_index) > 5  ) {
         //   pdfConfidence[GetIndex(x_index, y_index)] = sqrt(VarianceBin
	 //               /GetQtSamplesBin(x_index, y_index) );        
         //} else {
             pdfConfidence[GetIndex(x_index, y_index)] = 0; 	 
	 //} 
	 			
      } 
   }   
/*
   for (int x_index = 0; x_index < NumBinsX; x_index++) {
      for (int y_index = 0; y_index < NumBinsY; y_index++) {      
         double VarianceBin = 0;      
	 double qtSamplesBinTotalTmp = 0;
         for (int jj = 0; jj < FileCounter; jj++) {    
	    qtSamplesBinTotalTmp += 
	       oHistMeanParam2D[jj]->GetQtSamplesBin(x_index, y_index);
	 }	  
         for (int jj = 0; jj < FileCounter; jj++) {    
	    double qtSamplesBinTmp =
	       oHistMeanParam2D[jj]->GetQtSamplesBin(x_index, y_index);
	    if ( oHistMeanParam2D[jj]->GetBinSquare(x_index, y_index)
   	          /qtSamplesBinTmp 
		  > sq(oHistMeanParam2D[jj]->GetBin(x_index, y_index)
				     /qtSamplesBinTmp )
               && qtSamplesBinTmp > 2	) {   
	       //## Get the variance weighting by a weighted average
	       //## of the several biased simulations
   	       VarianceBin += qtSamplesBinTmp/qtSamplesBinTotalTmp      
	          *qtSamplesBinTmp*( 
	          oHistMeanParam2D[jj]->GetBinSquare(x_index, y_index)
   	          /qtSamplesBinTmp 
                  - sq( oHistMeanParam2D[jj]->GetBin(x_index, y_index)
				     /qtSamplesBinTmp) )
                  /(qtSamplesBinTmp - 1.)
		  /qtSamplesBinTmp;	   
            }		
	    //cout << jj << " qtSamplesBinTmp = " << qtSamplesBinTmp 
	    //     << "  VarianceBin = " << VarianceBin << endl;
            //VarianceBin += sq(oHistMeanParam2D[jj]->Get_pdfConfidence(
	    //      x_index, y_index) );
	 }
         pdfConfidence[GetIndex(x_index, y_index)] = sqrt(VarianceBin);
	 //cout << x_index<<" , "<< y_index << " #$#$ pdfConfidence = " 
	 //     << pdfConfidence[GetIndex(x_index, y_index)] <<endl;
      } 
   }
*/
   
   //## The confidence is actually the standard deviation, which is
   //## the confidence of a single sample
   //## Prevent an incorrect computation of the variance in Obtain_pdf()
   PreviouslyComputed_pdfConfidenceFlag = 1;

/********
   //## Third, I need to divide each bin by the number of samples that
   //## are in the bin.
   //double MaxRelNumHits;
   for (int x_index = 0; x_index < NumBinsX; x_index++) {
      for (int y_index = 0; y_index < NumBinsY; y_index++) {      
	 //cout << x_index<<" , "<< y_index << " #$#$**#$#$# pdfConfidence = " 
	 //     << pdfConfidence[GetIndex(x_index, y_index)] <<endl;      
         if ( GetQtSamplesBin(x_index,y_index) != 0 ) {
               SetBin(x_index,y_index, (double)
               	    GetBin(x_index,y_index)
	            /GetQtSamplesBin(x_index,y_index) );		    
         }		
	 else {   //## There is no sample in the bin   
 	     if (x_index > 0 && y_index > 0) { //## Adjust the bin
               SetBin(x_index,y_index, (double)
                	   GetBin(x_index - 1,y_index - 1) );	
             }			   
	 }	    	 
      }   
      
   }   
***********/   

   //## Fourth, and finally, force the number of samples to be 
   //#  unitary to get the right statistics, since all the 
   //#  pdfs used were already normalized.
   //#  Obviously, I need to denormilize the Bins, since the
   //#  Obtain_pdf method will do it
   NumSamplesTotal = 1;
   NumSamplesInRange = 1;    
   
   *LogFile << "The level curves of the parameter selected in "
           << InFileName.c_str()<<" was obtained "
           << "from a heuristic balance merging process." << endl;
//   *LogFile << "Thus, don't expect fully functionality from the other methods."
//           << endl;

   for (int ii = 0; ii < FileCounter; ii++) {
      delete oHistMeanParam2D[ii];
   }

   //WriteFile_pdf2D("test_pdf2D.dat");
   
} //## End 



//########################################################
//##### Imports the data from files created by the histogram class
//##    Convert2dB_Scale also determines the sign of the conversion to dB
//##    Actually, what you get from this is the conditional expectation
//##    
//##    By Aurenice O. Lima in Feb/2003.
void Histogram2D::GetLevelCurvesMultipleHeuristicBalanceImportVarianceWeighted(string InFileName,
                                                  int Convert2dB_Scale,
						  int ConvertParam2dB_Scale)
{
   //## Structure of InFileName
   //##  SampDataFileName    
         
   //## Clear the histogram, just in case...
   Clear();
   
   ifstream InFile(InFileName.c_str(),ios::in);
   if ( ! InFile) {
      cout << endl;
      cout << "Error: " << InFileName << " could not be open."
           << endl;
      exit(1);
   }     

   //## It can merge up to 100 biased files
   #define MAX_FILES_HEURISTIC_BALANCE 100   
   string SampFileName0[MAX_FILES_HEURISTIC_BALANCE];
   string SampFileName1[MAX_FILES_HEURISTIC_BALANCE];
   string SampFileName2[MAX_FILES_HEURISTIC_BALANCE];
   string SampHeuristicBalanceFileName[MAX_FILES_HEURISTIC_BALANCE];
      
   //Histogram2D* oHist_pdf2D[MAX_FILES_HEURISTIC_BALANCE];
   Histogram2D* oHistMeanParam2D[MAX_FILES_HEURISTIC_BALANCE]; 
   
   //Histogram2D* oHistJoint_pdf = new Histogram2D(HistInFileName);  
   Histogram2D* oHistJoint_pdf2D[MAX_FILES_HEURISTIC_BALANCE]; 

   *LogFile << "List of file names to be merged:" << endl;
   int FileCounter = 0;
   while (InFile >> SampFileName0[FileCounter]) {
      InFile >> SampFileName1[FileCounter];      
      InFile >> SampFileName2[FileCounter];       
      InFile >> SampHeuristicBalanceFileName[FileCounter];   
      
      ifstream SampFile0(SampFileName0[FileCounter].c_str(),ios::in);       
      ifstream SampFile1(SampFileName1[FileCounter].c_str(),ios::in);          
      ifstream SampFile2(SampFileName2[FileCounter].c_str(),ios::in);     
      ifstream SampHeuristicBalanceFile(
            SampHeuristicBalanceFileName[FileCounter].c_str(),ios::in);       
      if (   ! SampFile0 || ! SampFile1 || ! SampFile2 
          || ! SampHeuristicBalanceFile) {
         cout << endl;
	 if (! SampFile0) {
            cout << "Warning: " << SampFileName0[FileCounter] 
	         << " could not be open." << endl;
	 }           
         if ( ! SampFile1) {
            cout << endl;
            cout << "Warning: " << SampFileName1[FileCounter] 
      	      << " could not be open." << endl;     
	 }     	 
        if ( ! SampFile2) {
            cout << endl;
            cout << "Warning: " << SampFileName2[FileCounter] 
      	      << " could not be open." << endl;     
	 }     	 	 
         if ( ! SampHeuristicBalanceFile) {
            cout << endl;
            cout << "Warning: " << SampHeuristicBalanceFileName[FileCounter] 
      	      << " could not be open." << endl;     
	 } 	 
      }     
      else { //## They do exist	 
         *LogFile << "   "<< SampFileName0[FileCounter] 
                 << "   "<< SampFileName1[FileCounter] 
		 << "   "<< SampFileName2[FileCounter]  
		 << "   "<< SampHeuristicBalanceFileName[FileCounter]<< endl;  

         oHistJoint_pdf2D[FileCounter] = new Histogram2D(HistInFileName); 
         oHistMeanParam2D[FileCounter] = new Histogram2D(HistInFileName);
         FileCounter++;
      }	 
      SampFile0.close();
      SampFile1.close();
      SampFile2.close();	 
      SampHeuristicBalanceFile.close();      
   }
   InFile.close();

   //## First, get the pdf for each case
   double Sample0, Sample1;
   double Sample2;
   double LikelihoodRatio0, LikelihoodRatio1;      
   double LikelihoodRatio2;
   double LikelihoodRatioHeuristicBalance;
   double HeuristicBalanceWeight;
   double TrueCombinedWeight;
   for (int ii = 0; ii < FileCounter; ii++) {
      //oHist_pdf2D[ii] = new Histogram2D(HistInFileName.c_str()); 
      //oHistMeanParam2D[ii] = new Histogram2D(HistInFileName.c_str()); 
            
      cout << "File Name: "<< ii <<" = " <<SampFileName0[ii]  
                                 << "  " <<SampFileName1[ii] 
                                 << "  " <<SampFileName2[ii] 
				 << "  " <<SampHeuristicBalanceFileName[ii] 
				 << endl;				 

      ifstream SampFile0(SampFileName0[ii].c_str(),ios::in);      
      if ( ! SampFile0) {
         cout << endl;
         cout << "Warning: " << SampFileName0[ii] << " could not be open."
              << endl;
	  continue;       
      }             
      ifstream SampFile1(SampFileName1[ii].c_str(),ios::in);        
      if ( ! SampFile1) {
         cout << endl;
         cout << "Warning: " << SampFileName1[ii] << " could not be open."
              << endl;
	  continue;      
      }                
      ifstream SampFile2(SampFileName2[ii].c_str(),ios::in);            
      if ( ! SampFile2) {
         cout << endl;
         cout << "Warning: " << SampFileName2[ii] << " could not be open."
              << endl;
	  continue;      
      }    
      ifstream SampHeuristicBalanceFile(
            SampHeuristicBalanceFileName[ii].c_str(),ios::in);          
      if ( ! SampHeuristicBalanceFile) {
         cout << endl;
         cout << "Warning: " << SampHeuristicBalanceFileName[ii] 
	      << " could not be open." << endl;
	  continue;      
      }    
      
      while ( SampFile0 >> Sample0) {
         SampFile0 >> LikelihoodRatio0;
	 SampFile1 >> Sample1;
	 SampFile1 >> LikelihoodRatio1;
	 SampFile2 >> Sample2;
	 SampFile2 >> LikelihoodRatio2;	 
	 SampHeuristicBalanceFile >> LikelihoodRatioHeuristicBalance;
	 SampHeuristicBalanceFile >> HeuristicBalanceWeight;
	 SampHeuristicBalanceFile >> TrueCombinedWeight;	 
	 if (   LikelihoodRatio0 != LikelihoodRatio1 
	     || LikelihoodRatio0 != LikelihoodRatio2
	     || abs( (LikelihoodRatio0 - LikelihoodRatioHeuristicBalance)
	      /LikelihoodRatio0 ) > 1e-4 ) {
	    cout << "ERROR! From: Histogram2D::"
	         <<"GetLevelCurvesMultipleHeuristicBalanceImportVarianceWeighted(.)"
	         << endl << "The likelihoodRatio of the files "
		 <<SampFileName0[ii] << ", " << SampFileName1[ii] 
		 << ", " << SampFileName2[ii] 
                 << " and " << SampHeuristicBalanceFileName[ii]<< endl		 
		 << "are different. They must be equal."  << endl;
            exit(1);		   
	 }
         if (LikelihoodRatio0 > 0) {
            //## Samples can be filtered by the user 
            //#  doing the LikelihoodRatio < 0
            if (Convert2dB_Scale) {
               //oHistJoint_pdf->AddSample( sgn(Convert2dB_Scale)
	       //                          *Linear2dB(Sample0),
               //                           sgn(Convert2dB_Scale)
	       //			  *Linear2dB(Sample1),						 
		//	                  TrueCombinedWeight);	    
              oHistJoint_pdf2D[ii]->AddSample( sgn(Convert2dB_Scale)
	                                 *Linear2dB(Sample0),
                                          sgn(Convert2dB_Scale)
					  *Linear2dB(Sample1),						 
                  HeuristicBalanceWeight * LikelihoodRatio0);						  
	       
               if (ConvertParam2dB_Scale) {
                  //AddSample( sgn(Convert2dB_Scale)
                  //                                  *Linear2dB(Sample0),
                  //                            sgn(Convert2dB_Scale)
                  //                                  *Linear2dB(Sample1),						 
   		  //	   sgn(ConvertParam2dB_Scale)
		  //	   *( - Linear2dB_Limited(1./Sample2,20.) )
		  //	   *TrueCombinedWeight);			   
                  oHistMeanParam2D[ii]->AddSample( sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample0),
                                              sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample1),						 
   			   sgn(ConvertParam2dB_Scale)
			   *( - Linear2dB_Limited(1./Sample2,20.) )
                           *HeuristicBalanceWeight*LikelihoodRatio0);
               } else {
                  //AddSample( sgn(Convert2dB_Scale)
                  //                                  *Linear2dB(Sample0),
                  //                            sgn(Convert2dB_Scale)
                  //                                  *Linear2dB(Sample1),						 
   		  //	   Sample2*TrueCombinedWeight);			       			  
                  oHistMeanParam2D[ii]->AddSample( sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample0),
                                              sgn(Convert2dB_Scale)
                                                    *Linear2dB(Sample1),						 
 			Sample2 * HeuristicBalanceWeight * LikelihoodRatio0);				  
	       }  
            } else {
               //oHistJoint_pdf->AddSample( Sample0,Sample1,
	       //                           TrueCombinedWeight);	 	    
					  
              oHistJoint_pdf2D[ii]->AddSample( Sample0,Sample1,						 
                     HeuristicBalanceWeight*LikelihoodRatio0);					  
					  
              if (ConvertParam2dB_Scale) {
                  //AddSample(Sample0,Sample1,						 
   		  //	   sgn(ConvertParam2dB_Scale)
		  //	   *( - Linear2dB_Limited(1./Sample2,20.) )
		  //	   *TrueCombinedWeight);
			   
                  oHistMeanParam2D[ii]->AddSample(Sample0,Sample1,						 
   			   sgn(ConvertParam2dB_Scale)
			   *( - Linear2dB_Limited(1./Sample2,20.) )
			   *HeuristicBalanceWeight * LikelihoodRatio0);			   
               } else {
                  //AddSample(Sample0,Sample1,						 
   		  //	   Sample2
		  //	   *TrueCombinedWeight);
			   
                  oHistMeanParam2D[ii]->AddSample(Sample0,Sample1,						 
   			   Sample2
			   *HeuristicBalanceWeight*LikelihoodRatio0);			   	       
	       }
            } 
	     	    	    
         }
      }               
      SampFile0.close();
      SampFile1.close();
      SampFile2.close();      
      SampHeuristicBalanceFile.close();      
      //oHistMeanParam2D[ii]->ObtainConfidence();
   } 

   for (int x_index = 0; x_index < NumBinsX; x_index++) {
      for (int y_index = 0; y_index < NumBinsY; y_index++) {      
         //double MeanParameterBinUnorm = 0;
	 //double MeanIndicatorFunctionBin = 0;
         //double VarianceBinCond = 0;      
	 //double qtSamplesBinTotalTmp = 0;
         //for (int jj = 0; jj < FileCounter; jj++) {    
	 //   qtSamplesBinTotalTmp += 
	 //      oHistMeanParam2D[jj]->GetQtSamplesBin(x_index, y_index);
	 //}	 
	 double MeanBinParamTmp = 0;
	 double MeanBinTmp = 0;
	 double VarianceBinParamTmp = 0;
	 double VarianceBinTmp = 0;
         //double MeanBinParamSquareTmp = 0;
	 //double MeanBinSquareTmp = 0;
	 //double MeanCondForVarianceTmp = 0;
         for (int jj = 0; jj < FileCounter; jj++) {    	   
	    double qtSamplesDistrib = 
	          oHistMeanParam2D[jj]->GetQtSamples();
	    MeanBinParamTmp += oHistMeanParam2D[jj]->GetBin(x_index, y_index)
	                       /qtSamplesDistrib;	    
            MeanBinTmp += oHistJoint_pdf2D[jj]->GetBin(x_index, y_index)
	                 /qtSamplesDistrib;	    
	    VarianceBinParamTmp += sq(oHistMeanParam2D[jj]->Get_pdfConfidence(
	          x_index, y_index) );
	    VarianceBinTmp += sq(oHistJoint_pdf2D[jj]->Get_pdfConfidence(
	          x_index, y_index) );		  
	    //MeanBinParamSquareTmp = 	       
            //             oHistMeanParam2D[jj]->GetBinSquare(x_index, y_index);
            //MeanBinSquareTmp += oHistJoint_pdf->GetBinSquare(x_index, y_index);
	 }	 
	 //## Now, compute the conditional probability and the 
	 //## conditional variance of the bin.	 
	 double MeanBinParamCondTmp = 0;
	 double VarianceBinParamCondTmp = 99;
	 
	 if (MeanBinTmp != 0) {
	    MeanBinParamCondTmp = MeanBinParamTmp/MeanBinTmp;	 	 
            if (MeanBinParamTmp != 0) {	    
   	       VarianceBinParamCondTmp = sq(MeanBinParamCondTmp) 
	             *( VarianceBinTmp/sq(MeanBinTmp)
		       +VarianceBinParamTmp/sq(MeanBinParamTmp) );
            }		       
	    //## Note that the StdDev was normalized
            cout << x_index << " , " << y_index << "  ->  "		 
	        << sqrt(VarianceBinParamCondTmp)/MeanBinParamCondTmp
		 / ( NumBinsX/(x_max - x_min)*NumBinsY/(y_max - y_min) ) 
		<< " " 
		<< sqrt(VarianceBinParamTmp)/MeanBinParamTmp
		  /( NumBinsX/(x_max - x_min)*NumBinsY/(y_max - y_min) )
		<< "  "
                << sqrt(VarianceBinTmp)/MeanBinTmp 
		  /( NumBinsX/(x_max - x_min)*NumBinsY/(y_max - y_min) )
  	        << endl;
	 }				
	 //## I need to eliminate the normalization in the plot of the
	 //## pdf and of the standard deviation. The pdf and
	 //## the confidence normalized during the plot
         SetBin(x_index,y_index, MeanBinParamCondTmp 
	       / ( NumBinsX/(x_max - x_min)*NumBinsY/(y_max - y_min) ) );
         pdfConfidence[GetIndex(x_index, y_index)] 
	       = sqrt(VarianceBinParamCondTmp)	 
	 	 / ( NumBinsX/(x_max - x_min)*NumBinsY/(y_max - y_min) ); 	 
      } 
   }
   //## The confidence is actually the standard deviation of the mean, 
   //##  which determines the confidence interval of the mean
   
   //## Prevent an incorrect computation of the variance in Obtain_pdf()
   PreviouslyComputed_pdfConfidenceFlag = 1;  

   //## Fourth, and finally, force the number of samples to be 
   //#  unitary to get the right statistics, since all the 
   //#  pdfs used were already normalized.
   //#  Obviously, I need to denormilize the Bins, since the
   //#  Obtain_pdf method will do it
   NumSamplesTotal = 1;
   NumSamplesInRange = 1;    
   
   *LogFile << "The level curves of the parameter selected in "
           << InFileName.c_str()<<" was obtained "
           << "from a heuristic balance merging process." << endl;
//   *LogFile << "Thus, don't expect fully functionality from the other methods."
//           << endl;

   for (int ii = 0; ii < FileCounter; ii++) {
      delete oHistMeanParam2D[ii];
      delete oHistJoint_pdf2D[ii];      
   }

   //WriteFile_pdf2D("test_pdf2D.dat");
   
} //## End 






//#######################################
//## By: Ivan T. Lima Jr.

void Histogram2D::WriteFile_pdf2D(string OutFileName)
{

    LogFileSeparator();

   *LogFile << "2D Histogram Range Information for " << OutFileName << endl;
   *LogFile << "NumSamplesTotal = " << NumSamplesTotal << endl; 
   *LogFile << "NumSamplesInRange = " << NumSamplesInRange << endl;
   *LogFile << "NumSamplesWithOutOfRangeX = "
          <<  NumSamplesWithOutOfRangeX << endl;
   *LogFile << "NumSamplesWithOutOfRangeY = "
          << NumSamplesWithOutOfRangeY << endl;

  *LogFile << endl;
  *LogFile << "SamplesMinX = " << SamplesMinX << endl;
  *LogFile << "SamplesMaxX = " << SamplesMaxX << endl;
  *LogFile << "SamplesMinY = " << SamplesMinY << endl;
  *LogFile << "SamplesMaxY = " << SamplesMaxY << endl;


  ofstream OutFile(OutFileName.c_str(),ios::out);

  if (!OutFile) 
   {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  int x_index,y_index; 

   for (int x_index = 0; x_index < NumBinsX; x_index++) { 
      for (int y_index = 0; y_index < NumBinsY; y_index++) {

         //## The confidence interval is already normalized
         double pdfMeanTmp = GetBin(x_index,y_index)/NumSamplesTotal
   	          *NumBinsX/(x_max - x_min)*NumBinsY/(y_max - y_min);
         double pdfStdDevTmp = pdfConfidence[GetIndex(x_index, y_index)]
   	          *NumBinsX/(x_max - x_min)*NumBinsY/(y_max - y_min);	  		  
         double pdfRelativeStdDevTmp = 99;       //## Complete uncertainty	 
         if (pdfMeanTmp != 0) {
	    pdfRelativeStdDevTmp = pdfStdDevTmp/abs(pdfMeanTmp);
	    
         }     
         
         OutFile <<  x_index*delta_x + x_min + delta_x/2 << "  "
	         <<  y_index*delta_y + y_min + delta_y/2 << "  "
		 << pdfMeanTmp << "  "
		 << pdfStdDevTmp << "  "		 
		 << pdfRelativeStdDevTmp
		 << endl; 
      }
      OutFile << endl;
   }
    
  OutFile.close();

} // ## end WriteFile_pdf2D

// ####################################################

double Histogram2D::GetMeanX(void)
{

 return MeanX/NumSamplesInRange;

}

// ########################################################

double Histogram2D::GetMeanY(void)
{

 return MeanY/NumSamplesInRange;

}

// ####################################################

void Histogram2D::ComputePDFfromHistogram(void)
{
  for(int x_index = 0; x_index < NumBinsX; x_index++) 
   for(int y_index = 0; y_index < NumBinsY; y_index++) 
    {
      SetPDF(x_index,y_index,GetBin(x_index,y_index));
    }

NormalizePDF();

}

// ##################################################

void Histogram2D::NormalizePDF(void)
{
  double Integral = ComputeIntegralPDF();

  for(int x_index = 0; x_index < NumBinsX; x_index++)
   for(int y_index = 0; y_index < NumBinsY; y_index++)
    {
      SetPDF(x_index,y_index,GetPDF(x_index,y_index)/Integral);
    }

}

// ####################################################

double Histogram2D::GetMaxPDF(void)
{
   double MaxPDF = GetPDF(0,0);

   for (int x_index = 0; x_index < NumBinsX; x_index++) 
      for (int y_index = 0; y_index < NumBinsY; y_index++) 
        { 
          double Value = GetPDF(x_index,y_index);

          if(Value > MaxPDF)
             MaxPDF = Value;
        }

   return MaxPDF;

}

// ##############################################

double Histogram2D::SetPDFToZeroBelowThreshold(double Threshold)
{

    for (int x_index = 0; x_index < NumBinsX; x_index++) 
      for (int y_index = 0; y_index < NumBinsY; y_index++)
        { 
          if(GetPDF(x_index,y_index) < Threshold)
            {
             SetPDF(x_index,y_index,0);
            }
        }

    NormalizePDF();

}

// ##################################################


void Histogram2D::SetPDF(int x_index, int y_index, double value)
{

  // We use this means of indexing as typically want to access
  // Bins in a double loop like
  // for x
  // for y
  // Bin(x,y) = value

  PDF[y_index + NumBinsY*x_index] = value;
}

// #######################################

double Histogram2D::GetPDF(int x_index, int y_index)
{
  double Value = 0.0;

  if(x_index >= 0 && x_index < NumBinsX &&
     y_index >= 0 && y_index < NumBinsY)
     Value = PDF[y_index + NumBinsY*x_index];

  return Value;

}

// ###############################################

void Histogram2D::ComputeXMarginalPDF(void)
{

for(int x_index = 0; x_index < NumBinsX; x_index++)
{
  for(int y_index = 0; y_index < NumBinsY; y_index++)
   YData[y_index] = GetPDF(x_index,y_index); 

  XMarginal[x_index] = Quadrature(YData,NumBinsY,
                                  y_min+0.5*delta_y,y_max-0.5*delta_y);
}


}

// #############################################

void Histogram2D::ComputeYMarginalPDF(void)
{

for(int y_index = 0; y_index < NumBinsY; y_index++)
{
  for(int x_index = 0; x_index < NumBinsX; x_index++)
   XData[x_index] = GetPDF(x_index,y_index);    

  YMarginal[y_index] = Quadrature(XData,NumBinsX,
                                  x_min+0.5*delta_x,x_max-0.5*delta_x);
}

}

// #################################################

double Histogram2D::ComputeIntegralPDF(void)
{

  // Got same answer when switched order of integration

 ComputeXMarginalPDF();

 double Integral = Quadrature(XMarginal,NumBinsX,
                              x_min+0.5*delta_x,x_max-0.5*delta_x);

 return Integral;
}

// ##################################################

double Histogram2D::ComputeMeanX_PDF(void)
{

 ComputeXMarginalPDF();

 double MeanX_PDF;

 for(int x_index = 0; x_index < NumBinsX; x_index++)
   XData[x_index] = (x_min+x_index*delta_x + 0.5*delta_x)*XMarginal[x_index];

 MeanX_PDF=Quadrature(XData,NumBinsX,x_min+0.5*delta_x,x_max-0.5*delta_x);

 return MeanX_PDF;

}

// ##################################################

double Histogram2D::ComputeMeanY_PDF(void)
{

 ComputeYMarginalPDF();

 double MeanY_PDF;

 for(int y_index = 0; y_index < NumBinsY; y_index++)
   YData[y_index] = (y_min+y_index*delta_y + 0.5*delta_y)*YMarginal[y_index];

 MeanY_PDF=Quadrature(YData,NumBinsY,y_min+0.5*delta_y,y_max-0.5*delta_y);

 return MeanY_PDF;

}

// ################################################

void Histogram2D::GaussianPDF(double mean_x, double mean_y, double std)
{

 for(int x_index = 0; x_index < NumBinsX; x_index++)
  for(int y_index = 0; y_index < NumBinsY; y_index++)
   {
    double x = x_min + x_index*delta_x + 0.5*delta_x;
    double y = y_min + y_index*delta_y + 0.5*delta_y;

    SetPDF(x_index,y_index,(1.0/(2*pi*std*std))*
           exp(-((x-mean_x)*(x-mean_x) + (y-mean_y)*(y-mean_y))/(2*std*std)));

   }
}

// ################################################

void Histogram2D::WritePDF(string OutFileName)
{

 ofstream OutFile(OutFileName.c_str(),ios::out);

  if (!OutFile)
   {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

 

 for(int x_index = 0; x_index < NumBinsX; x_index++)
  for(int y_index = 0; y_index < NumBinsY; y_index++)
   {
    OutFile << x_index*delta_x + x_min + delta_x/2 << "  "
            << y_index*delta_y + y_min + delta_y/2 << "  "
            << GetPDF(x_index,y_index)
            << endl;
   }

 OutFile.close();

/*
 string OutFileName2 = OutFileName + "X.dat" ;

 ofstream OutFile2(OutFileName2.c_str(),ios::out);

 for(int x_index = 0; x_index < NumBinsX; x_index++)
  {
   OutFile2 << x_index*delta_x + x_min + delta_x/2 << "  "
           << XMarginal[x_index]
           << endl;

  }

 OutFile2.close();

 string OutFileName3 = OutFileName + "Y.dat" ;

 ofstream OutFile3(OutFileName3.c_str(),ios::out);

 for(int y_index = 0; y_index < NumBinsY; y_index++)
  {
   OutFile3 << y_index*delta_y + y_min + delta_y/2 << "  "
           << YMarginal[y_index]
           << endl;

  }

 OutFile3.close();
*/

}

// ############################################################

/*! If the data of one of the grid points is off the grid then
    the value at that grid point is set to zero. 
*/

double Histogram2D::BilinearInterpolationPDF(double x, double y)
{

 int x_index = int(floor((x-x_min)/delta_x));
 int y_index = int(floor((y-y_min)/delta_y));

 double t = (x-(x_min+x_index*delta_x))/delta_x;
 double s = (y-(y_min+y_index*delta_y))/delta_y;

 double Value = s*t*GetPDF(x_index+1,y_index+1)   +
                s*(1-t)*GetPDF(x_index,y_index+1) +
		(1-s)*t*GetPDF(x_index+1,y_index) +
		(1-s)*(1-t)*GetPDF(x_index,y_index);

 return Value;

}

// ############################################################

void Histogram2D::RotatePDF(double CenterX, double CenterY, double Angle, 
                            double *OutData)
{

for(int x_index=0;x_index<NumBinsX;x_index++)
for(int y_index=0;y_index<NumBinsY;y_index++)
 { 
   double x = x_min+x_index*delta_x;
   double y = y_min+y_index*delta_y;

   double u = CenterX + cos(Angle)*(x-CenterX) + sin(Angle)*(y-CenterY);
   double v = CenterY - sin(Angle)*(x-CenterX) + cos(Angle)*(y-CenterY);

   OutData[y_index + NumBinsY*x_index] = BilinearInterpolationPDF(u,v);

 }

}

// ############################################################


