// $Id: ocsOptFiber.cc,v 1.2 2006/05/11 22:31:30 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsOptFiber.hh"
extern ofstream *LogFile;

// ################### Methods for the class OptFiber

// ##### Constructor: sets the flag for memory allocation
// ## By: Ivan Lima
OptFiber::OptFiber()
{
   MemAllocated = 0;
}

// ##### Destructor: releases the allocated memory
// ## By: Ivan Lima
OptFiber::~OptFiber()
{
   if (MemAllocated)
      ReleaseMemory();
}

// ##### Specialized function to release memory
// ## By: Ivan Lima
void OptFiber::ReleaseMemory()
{
   delete [] phi;
   delete [] psi;
   delete [] theta;
   delete [] StepLength;
   
   //## For biasing the 2nd Order PDV
   delete [] delta;
   delete [] ksi;
   delete [] chi;
   
   for (int ii = 0; ii < 2; ii++)
      for (int jj = 0; jj < 2; jj++)
         delete [] TransM[ii][jj];   
}

// ######################################################

OptFiber::OptFiber(string InFileName, OptSignal *oOptSignal2, RanNumGen *RNG2) 
{

  // Written by John Zweck Oct 25, 2000
  // With contributions by Ivan Lima

   oOptSignal = oOptSignal2;

  ifstream InFile(InFileName.c_str(),ios::in);    

 if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }

   

DebugLevel = ReadInt("DebugLevel"," ",0,1,
                      LOWER_AND_UPPER,OPTIONAL_NO_WARNING,1,
		     &InFile,0);

 if(DebugLevel)
   {
     LogFileSeparator();
     *LogFile << InFileName << endl << endl;
     *LogFile << "DebugLevel = " << DebugLevel << endl;
   }

LengthFiber = ReadDouble("LengthFiber","m",0,-1,
                                    LOWER_ONLY,&InFile,DebugLevel);

qtStepsFiber = ReadInt("qtStepsFiber"," ",0,-1,
                              LOWER_ONLY,&InFile,DebugLevel);

// QUESTION: Are bounds on PMD correct? ie, Can it be < 0
// Yes, the bounds are correct. The PMD must be >=0.

PMD_Fiber = 0;
MeanDGD_SqrtLength= 0;

PMD_Fiber =  ReadDouble("PMD_Fiber","ps/sqrt(km)",0,-1,
			LOWER_ONLY,&InFile,DebugLevel); 

MeanDGD_SqrtLength =  ReadDouble("MeanDGD_SqrtLength","ps/sqrt(km)",0,-1,
			LOWER_ONLY,&InFile,DebugLevel); 
                        

if (PMD_Fiber != 0 && MeanDGD_SqrtLength != 0) {
   *LogFile << "OptFiber constructor:" << endl
           << "Either include MeanDGD_SqrtLength or "
           << "PMD_Fiber in the input file."<< endl
           << "Be aware that: " << endl
           << " MeanDGD_SqrtLength = sqrt(8/(3*pi))*PMD_Fiber" << endl
           << " It is justified by the fact that DGD is "<< endl
           << " a  Maxwellian distributed random variable"
           << endl;
   cout << endl<< "Error: ABORTING NOW! "
        << " See ocs.log for details" << endl;
   exit(1);
}

if (MeanDGD_SqrtLength != 0)
   SetMeanDGD_SqrtLength(MeanDGD_SqrtLength);
else //## PMD_Fiber might have been != 0 in the input
   SetPMD_Fiber(PMD_Fiber);

double ReferenceFreq = 0;
double ReferenceWavelength = 0;

ReferenceFreq = ReadDouble("ReferenceFreq","Hz",0,-1,
			LOWER_ONLY,&InFile,DebugLevel);

ReferenceWavelength = ReadDouble("ReferenceWavelength","m",0,-1,
			LOWER_ONLY,&InFile,DebugLevel);


if(ReferenceFreq == 0)
   {
     ReferenceFreq = Wavelength2Frequency(ReferenceWavelength);
  
     if(DebugLevel)
      *LogFile << endl 
              << "Since ReferenceFreq is zero we use ReferenceWavelength:"
              << endl
              << endl
              << "ReferenceWavelength [nm] = " << ReferenceWavelength*1e9 
              << endl
              << "ReferenceFreq [GHz] = " << ReferenceFreq*1e-9 << endl 
	      << endl;
   }
  else
   if(ReferenceWavelength == 0)
    {
      ReferenceWavelength = Frequency2Wavelength(ReferenceFreq);

      if(DebugLevel)
       *LogFile << endl 
              << "Since ReferenceWavelength is zero we use ReferenceFreq:" 
              << endl << endl 
              << "ReferenceWavelength [nm] = " << ReferenceWavelength*1e9 
              << endl
              << "ReferenceFreq [GHz] = " << ReferenceFreq*1e-9 << endl 
              << endl;
    }
   else
     {
       *LogFile << endl 
	       << "ERROR: "<< endl
               << "Make one of ReferenceFreq and ReferenceWavelength zero!" 
	       << endl << "The code will then use the other value." 
	       << endl <<  endl << "ABORTING NOW" << endl;

       exit(1);

     }


double DispersionFiberRefWavelength = 1e-12/(1e-9*1e3)*
                                        ReadDouble("DispersionFiber",
                                                  "ps/(nm-km)",0,0,
		                                   NO_BOUNDS,&InFile,
						   DebugLevel); 

// DispersionFiberRefWavelength now in SI units s/m^2

double DispSlopeFiberRefWavelength =  1e-12/(1e-18*1e3)*
                                       ReadDouble("DispSlopeFiber",
                                                 "ps/(nm^2-km)",0,0,
		                                NO_BOUNDS,&InFile,DebugLevel); 

// DispSlopeFiberRefWavelength now in SI units s/m^3

NonLinIndexFiber  =  ReadDouble("NonLinIndexFiber","m^2/W",0,-1,
			LOWER_ONLY,&InFile,DebugLevel); 

EffectAreaFiber  =  ReadDouble("EffectAreaFiber","m^2",0,-1,
			LOWER_ONLY,&InFile,DebugLevel); 

AttenuationFiber = ReadDouble("AttenuationFiber","dB/m",0,-1,
			LOWER_ONLY,&InFile,DebugLevel);

InFile.close();


fft = oOptSignal->fft;
sfftM = oOptSignal->sfftM;
CenterFreq = oOptSignal->GetCenterFreq();

RNG = RNG2;      

// Compute FstOrDispFiber and SndOrDispFiber
 
 if (ReferenceFreq == 0)
   {
     *LogFile << "ERROR: ReferenceFreq = 0" << endl
              << "ABORTING NOW"
              << endl; 
      exit(1);   
   }   


 // ######### Compute FstOrDispFiber and SndOrDispFiber #######

 // The  FstOrDispFiber and SndOrDispFiber are to be computed
 // at the CenterWavelength of the optical signal which is passing through
 // the fiber. 
 // For this we need to convert the values of DisperionFiber and 
 // DisperionSlopeFiber given in the input file at the ReferenceWavelength
 // to values at the CenterWavelength. 
 // This is done using a linear approximation.


double CenterWavelength = Frequency2Wavelength(CenterFreq);

double DispersionFiberCenterWavelength = DispersionFiberRefWavelength + 
         DispSlopeFiberRefWavelength*(CenterWavelength - ReferenceWavelength);

double DispSlopeFiberCenterWavelength = DispSlopeFiberRefWavelength;

FstOrDispFiber = -DispersionFiberCenterWavelength 
                  *LightSpeed/(2*pi*sq(CenterFreq) );

SndOrDispFiber = LightSpeed/(sq(2*pi)*cube(CenterFreq) )*
                (LightSpeed/CenterFreq*DispSlopeFiberCenterWavelength
                +  2*DispersionFiberCenterWavelength);

  
qtPoints         = fft->give_N();        
DeltaFreq        = fft->delta_f;         

WithNonLin       = (NonLinIndexFiber != 0);   
         
// ## Converts from base 10 to the Nepperian base: e = 2.718...  
      
NepperAttenuationFiber = AttenuationFiber*log(10.0)/10.0; 

   
if (PMD_Fiber == 0) {   // ## Two indep NLS
      NonLinCoupling[0][0] = 1;
      NonLinCoupling[0][1] = 0;
      NonLinCoupling[1][0] = 0;            
      NonLinCoupling[1][1] = 1;      
} else {
   if (qtStepsFiber > 1) {
      // ## CNLS -> Manakov-PMD
      NonLinCoupling[0][0] = 8./9.;
      NonLinCoupling[0][1] = 8./9.;
      NonLinCoupling[1][0] = 8./9.;            
      NonLinCoupling[1][1] = 8./9.;         
   } else { //## A single  section with no random mode coupling: Std CNLS
      NonLinCoupling[0][0] = 1.;
      NonLinCoupling[0][1] = 2/3.;
      NonLinCoupling[1][0] = 2/3.;            
      NonLinCoupling[1][1] = 1.;        
   }      
}   
   
   if (WithNonLin) 
    {
      gamma = NonLinIndexFiber*2*pi*CenterFreq/(EffectAreaFiber*LightSpeed);
    }
    else {
       gamma = 0;
   }          
 
   TypeScattering = 1; // ## Euler Matrix (Default)

   FlagImportSamplingSndOrderPDV_VariableAngleExp = 0;
    
   AllocateMemory_SetLengthSteps();    

if(DebugLevel)
 {
  LogFileSeparator();

  *LogFile << "Processing Fiber Parameters" << endl << endl;

  *LogFile <<  "LengthFiber [km] = " << LengthFiber*1e-3 << endl;
  *LogFile <<  "qtStepsFiber = " << qtStepsFiber << endl;
  *LogFile << "PMD_Fiber [s/sqrt(m)] = "<< PMD_Fiber << endl;  
  *LogFile << " -> PMD_Fiber [ps/sqrt(km)] = " << 1e12*sqrt(1e3)*PMD_Fiber << endl;
  *LogFile << "MeanDGD_SqrtLength [s/sqrt(m)] = " << MeanDGD_SqrtLength << endl;
  *LogFile << " -> MeanDGD [ps] = " << 1e12*MeanDGD_SqrtLength*sqrt(LengthFiber) << endl;
  *LogFile << "   Note that:  MeanDGD_SqrtLength = sqrt(8/(3*pi))*PMD_Fiber" << endl;  
  *LogFile <<  "NonLinIndexFiber [m^2/W] = " << NonLinIndexFiber << endl;
  *LogFile <<  "EffectAreaFiber [m^2] = " << EffectAreaFiber << endl;
  *LogFile <<  "AttenuationFiber [dB/km] = " <<  AttenuationFiber*1e3 << endl;
  *LogFile <<  "Center Freq of Signal [GHz] = " << CenterFreq*1e-9 << endl;
  *LogFile <<  "Center Wavelength of Signal [nm] = " << CenterWavelength*1e9 
          << endl;

  *LogFile <<  "PMD_Fiber [ps/sqrt(km)] = " << PMD_Fiber*1e12*sqrt(1e3) << endl;
  
  *LogFile <<  "qtPoints = " << qtPoints << endl; 
  *LogFile <<  "DeltaFreq [GHz] = " <<   DeltaFreq*1e-9 << endl; 
  *LogFile << endl;
  *LogFile <<  "DispersionFiber at Reference Wavelength [ps/(nm-km)] = " 
          << DispersionFiberRefWavelength*1e12/(1e9*1e-3)  << endl; 
  *LogFile <<  "DispSlopeFiber at Reference Wavelength [ps/(nm^2-km)] = " 
          << DispSlopeFiberRefWavelength*1e12/(1e18*1e-3) << endl; 

  *LogFile <<  "FstOrDispFiber at Center Wavelength [ps^2/km] = " 
          << FstOrDispFiber*1e27 << endl; 
  *LogFile <<  "SndOrDispFiber  at Center Wavelength [ps^3/km] = " 
          << SndOrDispFiber*1e39 << endl; 
               

  *LogFile << "Gamma [m^(-1)W^(-1)] = " << gamma << endl;

  *LogFile << "Step Size DeltaZ Fiber [m] = " 
           << LengthFiber/qtStepsFiber << endl;         

 } // end if

}// ## end OptFiber constructor


// ## By: Ivan Lima
void OptFiber::SetMeanDGD_SqrtLength(double MeanDGD_SqrtLength2)
{
   // ## Attention: it is expect to be in [ps/sqrt(km)]

   DeltaZ = LengthFiber/qtStepsFiber; // JZAug02

   MeanDGD_SqrtLength = 1e-12/sqrt(1e3)*MeanDGD_SqrtLength2;
   //PMD_Fiber = sqrt(3.*pi/8.)*MeanDGD_SqrtLength;   
   PMD_Fiber = 1./rms2meanDGD()*MeanDGD_SqrtLength; 
   dBeta_2dOmega = PMD_Fiber/sqrt(4*DeltaZ);  

 
}

// ## By: Ivan Lima
void OptFiber::SetPMD_Fiber(double PMD_Fiber2)
{
   // ## Attention: it is expect to be in [ps/sqrt(km)]

   DeltaZ = LengthFiber/qtStepsFiber; // JZAug02

   PMD_Fiber = 1e-12/sqrt(1e3)*PMD_Fiber2;
   //MeanDGD_SqrtLength = sqrt(8./(3.*pi))*PMD_Fiber;
   MeanDGD_SqrtLength = rms2meanDGD()*PMD_Fiber;
   dBeta_2dOmega = PMD_Fiber/sqrt(4*DeltaZ);    
}

//void OptFiber::SetQtStepsFiber(int qtStepsFiber2)
//{
//   int qtStepsFiber = qtStepsFiber2;
//   ReleaseMemory();
//   AllocateMemory_SetLengthSteps();   
//}

//## Allows for a correct first-order emulation with 
//#  a single section of birefringent fiber
//#  This is an internal function
//#  By: Aurenice Lima
double OptFiber::rms2meanDGD(void)
{
   double rms2meanDGD;
   if (qtStepsFiber != 1) {
      rms2meanDGD = sqrt(8./(3.*pi));
   }   
   else {
      rms2meanDGD = 1;         
   }   

   
   return rms2meanDGD;      
}


// ## Allocates the memory
// # By: Ivan Lima
void OptFiber::AllocateMemory_SetLengthSteps(void)
{
    
   MemAllocated = 1;  
   DeltaZ = LengthFiber/qtStepsFiber;
 
   //JZAug02   dBeta_2dOmega = PMD_Fiber/sqrt(4*DeltaZ); 
   //JZFeb03: For my applications it looks like its OK to have next line too

   dBeta_2dOmega = PMD_Fiber/sqrt(4*DeltaZ);   //## This has to be here

   
   phi     = new double[qtStepsFiber];
   psi     = new double[qtStepsFiber];   
   theta   = new double[qtStepsFiber]; 
   StepLength = new double[qtStepsFiber];  
  
   //## For biasing the 2nd Order PDV
   delta = new double[qtStepsFiber]; 
   ksi = new double[qtStepsFiber];    
   chi = new double[qtStepsFiber];    
   
   for (int ii = 0; ii < 2; ii++)
      for (int jj = 0; jj < 2; jj++)
         TransM[ii][jj] = new cplx[qtStepsFiber];
      
   GotFiberSample = 0;

   for (int ii = 0; ii < qtStepsFiber; ii++)
      StepLength[ii] = DeltaZ;
}

// #####################################################
/*
// ##### Initializes an optical fiber
// ## By: Ivan Lima
void OptFiber::InitFiberParam( cfftw *fft2,fftMatrices sfftM2
                                  ,FiberParam soOptFiber2)
{  

  // THIS METHOD IS NOW OUT OF DATE!!!!!!!


   fft            = fft2;
   sfftM          = sfftM2;
   soOptFiber     = soOptFiber2;
      
   LengthFiber      = soOptFiber.LengthFiber;
   qtStepsFiber     = soOptFiber.qtStepsFiber;
   PMD_Fiber        = soOptFiber.PMD_Fiber;
   FstOrDispFiber   = soOptFiber.FstOrDispFiber;
   SndOrDispFiber   = soOptFiber.SndOrDispFiber;
   NonLinIndexFiber = soOptFiber.NonLinIndexFiber;
   EffectAreaFiber  = soOptFiber.EffectAreaFiber;   
   AttenuationFiber = soOptFiber.AttenuationFiber;                 
   CenterFreq        = soOptFiber.CenterFreq;
   //fSeedPMD         = soOptFiber.fSeedPMD;   
   qtPoints         = fft->give_N();        //soOptFiber.qtPoints;
   DeltaFreq        = fft->delta_f;         //soOptFiber.DeltaFreq; 

   WithNonLin       = (NonLinIndexFiber != 0);   
         
   // ## Converts from base 10 to the Nepperian base: e = 2.718...        
   NepperAttenuationFiber = soOptFiber.AttenuationFiber*log(10.)/10.; 
   //NepperAttenuationFiber = soOptFiber.AttenuationFiber*log(10.)/10.;
   
   if (PMD_Fiber == 0) {   // ## Two coupled NLS - Not fully tested
      NonLinCoupling[0][0] = 1;
      NonLinCoupling[0][1] = 0.;
      NonLinCoupling[1][0] = 0.;            
      NonLinCoupling[1][1] = 1;      
   }
   else {   // ## CNLS -> Manakov-PMD
      NonLinCoupling[0][0] = 8./9.;
      NonLinCoupling[0][1] = 8./9.;
      NonLinCoupling[1][0] = 8./9.;            
      NonLinCoupling[1][1] = 8./9.;         
   }   
   
   if (WithNonLin) {
      gamma = NonLinIndexFiber*2*pi*CenterFreq/(EffectAreaFiber*LightSpeed);
   }      
   
   if (MemAllocated && qtStepsFiber != qtStepsFiberLast) {
      ReleaseMemory();
      MemAllocated = 0;
   }
   if ( ! MemAllocated) { 
      MemAllocated = 1;
      phi     = new double[qtStepsFiber];
      psi     = new double[qtStepsFiber];   
      theta   = new double[qtStepsFiber]; 
      StepLength = new double[qtStepsFiber];  

      for (int ii = 0; ii < 2; ii++)
         for (int jj = 0; jj < 2; jj++)
            TransM[ii][jj] = new cplx[qtStepsFiber];
      qtStepsFiberLast = qtStepsFiber;        
   }   
   GotFiberSample = 0;
   
   DeltaZ = LengthFiber/qtStepsFiber;
   dBeta_2dOmega = PMD_Fiber/sqrt(4*DeltaZ); 
   //ExpectedDGD = sqrt(8/(3*pi)*LengthFiber)*PMD_Fiber;
   //ExpectedSndOrderDGD = sq(ExpectedDGD)/sqrt(3.);
   TypeScattering = 1; // ## Euler Matrix (Default)
   for (int ii = 0; ii < qtStepsFiber; ii++)
      StepLength[ii] = DeltaZ; 
   
} // ## end constructor
*/


//####################################################
//## This function is used to change the optical signal
//#  on the fly.
//#  Check the constructor before using it, since 
//#  it might be out of date.
//##  It only cares about the data, while leaving the
//#   other parameters of the original OptSignal.
//##   Thus, the OptSignal has to have the same number of 
//#   FFT points and everything else.
//## By: Ivan Lima (7/24/01)
void OptFiber::SwitchOptSignalData(OptSignal *oOptSignal2)
{
   oOptSignal = oOptSignal2;
   fft = oOptSignal->fft;
   sfftM = oOptSignal->sfftM;
   qtPoints = oOptSignal->GetQtPoints();
   //CenterFreq = oOptSignal->GetCenterFreq();
}


// #####################################################
// ## Set the lengths of the steps
// ## By: Ivan Lima
void OptFiber::SetStepLengths(double *usrStepLength)
{  // ## usrStepLength[] should have the size == qtStepsFiber
   double TotLengthSteps = 0;
   for (int ii = 0; ii < qtStepsFiber; ii++) {
      StepLength[ii]  = usrStepLength[ii];
      TotLengthSteps += StepLength[ii];   
   }
   if (TotLengthSteps != LengthFiber) {
      cerr << "Warning: the sum of the steps: " << TotLengthSteps << endl
           << " is different from the Length of the Fiber: " 
           << LengthFiber << "." << endl
           << "Thus, the length of the fiber will be redefined" << endl;
      LengthFiber = TotLengthSteps;
   }       
}


// #####################################################
// ## Set the length of a step of the fiber
// ## ATTENTION: avoid using this function, since it 
// ## does not adjust for the PMD automatically.
// ## To change the DGD of a first order PMD emulator 
// ## was the main purpose of this method.
// ## For other purposes use the method SetStepLengths(double *).
// ## By: Aurenice Lima
void OptFiber::SetStepLength(double usrStepLength, int StepNumber)
{
   if (StepNumber > qtStepsFiber || StepNumber < 1) {
      cout << "From OptFiber::SetSpteLength(.)" <<endl
           << "The StepNumber is inconsistent." <<endl
           << "It should be a number between 1 and "
           << qtStepsFiber<< endl;
      exit(1);     
   } 
   
   StepLength[StepNumber - 1] = usrStepLength;
   
//   if (TotLengthSteps != LengthFiber) {
//      cerr << "Warning: the sum of the steps: " << TotLengthSteps << endl
//           << " is different from the Length of the Fiber: " 
//           << LengthFiber << "." << endl
//           << "Thus, the length of the fiber will be redefined" << endl;
//      LengthFiber = TotLengthSteps;
//   }       
}


// #####################################################
// ## Choose random lengths for the step
// ## By: Ivan Lima
void OptFiber::GetRandomStepLengths(double RelativeRangeLengths)
{  // ## The range of the lengths is relative to the mean length: DeltaZ
   // ## I am using fSeedPMD, but I should think about creating an extra
   // ## seed just for this task, since it is a feature of the object.
   double *SplicingNode = new double[qtStepsFiber+1];  
   SplicingNode[qtStepsFiber] = LengthFiber;
   for (int ii = 0; ii < qtStepsFiber; ii++) 
      SplicingNode[ii] = DeltaZ
                     *(ii+(RNG->GetRanNum() -0.5)*RelativeRangeLengths);   
   for (int ii = 0; ii < qtStepsFiber; ii++)
      StepLength[ii] = SplicingNode[ii+1] - SplicingNode[ii];                
   delete [] SplicingNode;  
   
   //cout << 2*pi*DeltaZ*dBeta_2dOmega << endl;    
   //exit(1);
}

// ## Create a file with the lengths
// ## By: Ivan Lima
void OptFiber::WriteFileStepLengths(char *outFile)
{
   ofstream OutFile(outFile,ios::out);
   if (!OutFile) {
      cerr << "Error: " << outFile << "could not be created."<<endl;
      exit(1);
   }
   int ii;
   for (ii = 0; ii < qtStepsFiber; ii++) 
      OutFile << ii << "   " << StepLength[ii]   << endl;   
   OutFile.close();
}


//############################################################
// ## Create a file with the dispersion profile
// ## By: Ivan Lima
void OptFiber::WriteFileDispersionProfile(char *outFile)
{
   ofstream OutFile(outFile,ios::out);
   if (!OutFile) {
      cerr << "Error: " << outFile << "could not be created."<<endl;
      exit(1);
   }
   double FreqWindow = oOptSignal->GetFreqWindow();   
   for (int ii = 0; ii < qtPoints; ii++)  
    {    
      double Frequency = ii*DeltaFreq - FreqWindow/2;
      OutFile <<  Frequency << "   " 
              << FstOrDispFiber + SndOrDispFiber*2*pi*Frequency
               << endl;   
    }
}   


/*****************
// ####################################################
// ## Propagates the optical signal through the fiber
// ## By: Ivan Lima

void OptFiber::PropagateFiberCoupledNLS(double *PropagatedLengthFiber)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }   
   int ii;   // Propagates along the fiber
   int jj;   // Sweep over frequency
   cplx EnvelopeX,EnvelopeY;
   double PowerEnvelopeX, PowerEnvelopeY;        
   cplx PropM[2];
   cplx PhaseBirref_2; //  = jc*2*pi*DeltaZ*dBeta_2dOmega;
   cplx PhaseFstOrDispFiber; // = jc*sq(2*pi)*DeltaZ*FstOrDispFiber/2;
   cplx PhaseSndOrDispFiber; // =-jc*cube(2*pi)*DeltaZ*SndOrDispFiber/6;   
     
  // 12 Feb, 2001. (JZ) I fixed a bug in the sign of the term for 
  // the higher order dispersion SndOrDisp. 
  // With the sign as it is now we have shift-invariance 
  // in frequency: If we change the center frequency then 
  // the only change in the output pulse train is that it
  // it is shifted by the appropriate amount left or right
  // in time. The shape of the pulses will not change.
  // The fact that we have the combination
  // FstOrDisp/2 + omega*SndOrDisp/6
  // comes from the fact that this is the start of the Taylor's series
  // for beta(CenterFreq + freq) - beta(CenterFreq).
  // See Curtis Menyuk's Multiple Length Scales paper in J. Engr Math
  // equation (25). 

  // Caution: If we ever  change the Fourier Transform we may need to
  // change the signs in this method. To check this you must check that
  // the code is shift-invariant in frequency in the sense described above.

   double FreqSignal;   
   for ( ii = 0; ii < qtStepsFiber; ii++) { 
      PhaseBirref_2  = jc*2*pi*StepLength[ii]*dBeta_2dOmega;
      PhaseFstOrDispFiber = jc*sq(2*pi)*StepLength[ii]*FstOrDispFiber/2.;
      PhaseSndOrDispFiber = jc*cube(2*pi)*StepLength[ii]*SndOrDispFiber/6.;  
    
      //cout << TransM[0][0][ii] << " " << TransM[0][1][ii] << " " 
      //	   << TransM[1][0][ii] << " " << TransM[1][1][ii] << endl;

      for ( jj = 0; jj < qtPoints; jj++) {  
         FreqSignal = fft->freq(jj);        
         // ## The term CenterFreq in PropM[.] is not part of the Manakov-PMD,
         // ## but I left it here to apply to other scattering techniques.
	 
	 //## Attention: I changed the signd of the birefringence phase (IL)
         PropM[0] = exp(  -(CenterFreq+FreqSignal)*PhaseBirref_2 
                        +    sq(FreqSignal)*(PhaseFstOrDispFiber
                        +       FreqSignal* PhaseSndOrDispFiber) 
                        - StepLength[ii]*NepperAttenuationFiber/2); 
         PropM[1] = exp( +(CenterFreq+FreqSignal)*PhaseBirref_2 
                        +    sq(FreqSignal)*(PhaseFstOrDispFiber
                        +       FreqSignal* PhaseSndOrDispFiber) 
                        - StepLength[ii]*NepperAttenuationFiber/2); 
      
         EnvelopeX = sfftM.fOptSignalX[jj];
         EnvelopeY = sfftM.fOptSignalY[jj];
         // ## Scatters first, and then propagates though the fiber step
         sfftM.fOptSignalX[jj] = PropM[0]
              *(TransM[0][0][ii]*EnvelopeX + TransM[0][1][ii]*EnvelopeY);
         sfftM.fOptSignalY[jj] = PropM[1]
              *(TransM[1][0][ii]*EnvelopeX + TransM[1][1][ii]*EnvelopeY);
                        
      }
      // ## Non-linear effects: according to the Manakov-PMD equation
      if (WithNonLin) {
         fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
         fft->IFFT(sfftM.fOptSignalY,sfftM.tOptSignalY);           
         for ( jj = 0; jj < qtPoints; jj++) {
            PowerEnvelopeX = sq(abs(sfftM.tOptSignalX[jj]));
            PowerEnvelopeY = sq(abs(sfftM.tOptSignalY[jj]));             
            sfftM.tOptSignalX[jj] = sfftM.tOptSignalX[jj]
                 *exp(jc*gamma*StepLength[ii]
                 *( NonLinCoupling[0][0]*PowerEnvelopeX
                   +NonLinCoupling[0][1]*PowerEnvelopeY) );            
            sfftM.tOptSignalY[jj] = sfftM.tOptSignalY[jj]
                 *exp(jc*gamma*StepLength[ii]
                 *( NonLinCoupling[1][0]*PowerEnvelopeX
                   +NonLinCoupling[1][1]*PowerEnvelopeY) ) ;                                                
         }         
         fft->FFT(sfftM.tOptSignalX,sfftM.fOptSignalX);
         fft->FFT(sfftM.tOptSignalY,sfftM.fOptSignalY);           
      }   
   }   
   if ( ! WithNonLin) {   // generates the time dependence
      fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
      fft->IFFT(sfftM.fOptSignalY,sfftM.tOptSignalY);       
   }   
   *PropagatedLengthFiber += LengthFiber;
}
*************/


//####################################################
//## Propagates the optical signal through the fiber
//## This method is implementing the symmetric CNLS, 
//## which has second-order accuracy.
//## By: Ivan Lima
void OptFiber::PropagateFiberCoupledNLS(double *PropagatedLengthFiber)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }   
   int ii;   // Propagates along the fiber
   int jj;   // Sweep over frequency
   cplx EnvelopeX,EnvelopeY;
   double PowerEnvelopeX, PowerEnvelopeY;        
   cplx PropM[2];
   cplx PhaseBirref_2; //  = jc*2*pi*DeltaZ*dBeta_2dOmega;
   cplx PhaseFstOrDispFiber; // = jc*sq(2*pi)*DeltaZ*FstOrDispFiber/2;
   cplx PhaseSndOrDispFiber; // =-jc*cube(2*pi)*DeltaZ*SndOrDispFiber/6;   
     
  // 12 Feb, 2001. (JZ) I fixed a bug in the sign of the term for 
  // the higher order dispersion SndOrDisp. 
  // With the sign as it is now we have shift-invariance 
  // in frequency: If we change the center frequency then 
  // the only change in the output pulse train is that it
  // it is shifted by the appropriate amount left or right
  // in time. The shape of the pulses will not change.
  // The fact that we have the combination
  // FstOrDisp/2 + omega*SndOrDisp/6
  // comes from the fact that this is the start of the Taylor's series
  // for beta(CenterFreq + freq) - beta(CenterFreq).
  // See Curtis Menyuk's Multiple Length Scales paper in J. Engr Math
  // equation (25). 

  // Caution: If we ever  change the Fourier Transform we may need to
  // change the signs in this method. To check this you must check that
  // the code is shift-invariant in frequency in the sense described above.

   double FreqSignal;   
   double CurrentStepLength;
   for ( ii = 0; ii < qtStepsFiber; ii++) { 
      if (! WithNonLin) { //## Only one linear step per section
         CurrentStepLength = StepLength[ii];
      } 
      else { //## half linear step before and after a non-linear step
         CurrentStepLength = StepLength[ii]/2.;      
      }
  

      PhaseBirref_2  = jc*2*pi*CurrentStepLength*dBeta_2dOmega;
      PhaseFstOrDispFiber = jc*sq(2*pi)*CurrentStepLength*FstOrDispFiber/2.;
      PhaseSndOrDispFiber = jc*cube(2*pi)*CurrentStepLength*SndOrDispFiber/6.;  
   
//      PhaseBirref_2  = jc*2*pi*StepLength[ii]*dBeta_2dOmega;
//      PhaseFstOrDispFiber = jc*sq(2*pi)*StepLength[ii]*FstOrDispFiber/2.;
//      PhaseSndOrDispFiber = jc*cube(2*pi)*StepLength[ii]*SndOrDispFiber/6.;  
    
      //cout << TransM[0][0][ii] << " " << TransM[0][1][ii] << " " 
      //	   << TransM[1][0][ii] << " " << TransM[1][1][ii] << endl;

      for ( jj = 0; jj < qtPoints; jj++) {  
         FreqSignal = fft->freq(jj);        
         // ## The term CenterFreq in PropM[.] is not part of the Manakov-PMD,
         // ## but I left it here to apply to other scattering techniques.
	 
	 //## Attention: I changed the signd of the birefringence phase (IL)


         PropM[0] = exp(-(CenterFreq+FreqSignal)*PhaseBirref_2 
                        +    sq(FreqSignal)*(PhaseFstOrDispFiber
                        +       FreqSignal* PhaseSndOrDispFiber) 
                        - CurrentStepLength*NepperAttenuationFiber/2); 
         PropM[1] = exp(+(CenterFreq+FreqSignal)*PhaseBirref_2 
                        +    sq(FreqSignal)*(PhaseFstOrDispFiber
                        +       FreqSignal* PhaseSndOrDispFiber) 
                        - CurrentStepLength*NepperAttenuationFiber/2); 

//         PropM[0] = exp(  -(CenterFreq+FreqSignal)*PhaseBirref_2 
//                        +    sq(FreqSignal)*(PhaseFstOrDispFiber
//                        +       FreqSignal* PhaseSndOrDispFiber) 
//                        - StepLength[ii]*NepperAttenuationFiber/2); 
//         PropM[1] = exp( +(CenterFreq+FreqSignal)*PhaseBirref_2 
//                        +    sq(FreqSignal)*(PhaseFstOrDispFiber
//                        +       FreqSignal* PhaseSndOrDispFiber) 
//                        - StepLength[ii]*NepperAttenuationFiber/2); 
     

         EnvelopeX = sfftM.fOptSignalX[jj];
         EnvelopeY = sfftM.fOptSignalY[jj];
         // ## Scatters first, and then propagates trough the fiber step
         sfftM.fOptSignalX[jj] = PropM[0]
              *(TransM[0][0][ii]*EnvelopeX + TransM[0][1][ii]*EnvelopeY);
         sfftM.fOptSignalY[jj] = PropM[1]
              *(TransM[1][0][ii]*EnvelopeX + TransM[1][1][ii]*EnvelopeY);         
	 //## Test line. Comment this line later
	 //if (jj == 0) cout << ii << " "<< TransM[0][0][ii] << "  " << TransM[0][1][ii] << endl;	                              
      }
      //## Non-linear effects: according to the Manakov-PMD equation
      //## It propagates the nonlinear effect through the whole step
      if (WithNonLin) {
         oOptSignal->Freq2Time();
         //fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
         //fft->IFFT(sfftM.fOptSignalY,sfftM.tOptSignalY);           
         for ( jj = 0; jj < qtPoints; jj++) {
            PowerEnvelopeX = sq(abs(sfftM.tOptSignalX[jj]));
            PowerEnvelopeY = sq(abs(sfftM.tOptSignalY[jj]));             
            sfftM.tOptSignalX[jj] = sfftM.tOptSignalX[jj]
                 *exp(jc*gamma*StepLength[ii]
                 *( NonLinCoupling[0][0]*PowerEnvelopeX
                   +NonLinCoupling[0][1]*PowerEnvelopeY) );            
            sfftM.tOptSignalY[jj] = sfftM.tOptSignalY[jj]
                 *exp(jc*gamma*StepLength[ii]
                 *( NonLinCoupling[1][0]*PowerEnvelopeX
                   +NonLinCoupling[1][1]*PowerEnvelopeY) ) ;                                      	   
         }         
         //fft->FFT(sfftM.tOptSignalX,sfftM.fOptSignalX);
         //fft->FFT(sfftM.tOptSignalY,sfftM.fOptSignalY);    
	 oOptSignal->Time2Freq();

	 //## Now, the second half of the linear step, if WithNonLin
         for ( jj = 0; jj < qtPoints; jj++) {  
            FreqSignal = fft->freq(jj);        
            // ## The term CenterFreq in PropM[.] is not part of the Manakov-PMD,
            // ## but I left it here to apply to other scattering techniques.
	 
   	    //## Attention: I changed the signd of the birefringence phase (IL)

            //## This thing MUST be here, because is frequency dependent
            PropM[0] = exp(  -(CenterFreq+FreqSignal)*PhaseBirref_2 
                           +    sq(FreqSignal)*(PhaseFstOrDispFiber
                           +       FreqSignal* PhaseSndOrDispFiber) 
                           - CurrentStepLength*NepperAttenuationFiber/2); 
            PropM[1] = exp( +(CenterFreq+FreqSignal)*PhaseBirref_2 
                           +    sq(FreqSignal)*(PhaseFstOrDispFiber
                           +       FreqSignal* PhaseSndOrDispFiber) 
                           - CurrentStepLength*NepperAttenuationFiber/2); 
      
            //EnvelopeX = sfftM.fOptSignalX[jj];
            //EnvelopeY = sfftM.fOptSignalY[jj];
            // ## Already scattered, just propagates though the fiber step
            sfftM.fOptSignalX[jj] *= PropM[0];
            //     *(TransM[0][0][ii]*EnvelopeX + TransM[0][1][ii]*EnvelopeY);
            sfftM.fOptSignalY[jj] *= PropM[1];
            //     *(TransM[1][0][ii]*EnvelopeX + TransM[1][1][ii]*EnvelopeY);                           
         }        
      } //## Processes the nonlinear effect and the second half of the linear
   }   //## Finished last step
   //   if ( ! WithNonLin) {   // generates the time dependence
   //      fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
   //      fft->IFFT(sfftM.fOptSignalY,sfftM.tOptSignalY);       
   //   }   
   oOptSignal->Freq2Time();
   *PropagatedLengthFiber += LengthFiber;
}



//####################################################
//## Propagates the optical signal through the fiber
//## This method is implementing the symmetric CNLS, 
//## which has second-order accuracy.
//## This method incorporates the full nonlinear term 
//## of a standard coupled NLS
//## In addition, it allows to damp the signal that goes through the
//## edges of the time window, which is periodic.
//## By: Ivan Lima   6/18/03
void OptFiber::PropagateFullNonlinearFiberCoupledNLS(double *PropagatedLengthFiber,
                                                     double DampTime)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }   
  
   //## Make sure that the cross-coupling is correct 
   SetNonLinCoupling(1.,2/3.,2/3.,1);  

   //cout << NonLinCoupling[0][0] << endl
   //	<< NonLinCoupling[0][1] << endl
   //	<< NonLinCoupling[1][0] << endl
   //	<< NonLinCoupling[1][1] <<endl;

   
   int ii;   // Propagates along the fiber
   int jj;   // Sweep over frequency
   cplx EnvelopeX,EnvelopeY;
   double PowerEnvelopeX, PowerEnvelopeY;        
   cplx PropM[2];
   cplx PhaseBirref_2; //  = jc*2*pi*DeltaZ*dBeta_2dOmega;
   cplx PhaseFstOrDispFiber; // = jc*sq(2*pi)*DeltaZ*FstOrDispFiber/2;
   cplx PhaseSndOrDispFiber; // =-jc*cube(2*pi)*DeltaZ*SndOrDispFiber/6;   
     
  // 12 Feb, 2001. (JZ) I fixed a bug in the sign of the term for 
  // the higher order dispersion SndOrDisp. 
  // With the sign as it is now we have shift-invariance 
  // in frequency: If we change the center frequency then 
  // the only change in the output pulse train is that it
  // it is shifted by the appropriate amount left or right
  // in time. The shape of the pulses will not change.
  // The fact that we have the combination
  // FstOrDisp/2 + omega*SndOrDisp/6
  // comes from the fact that this is the start of the Taylor's series
  // for beta(CenterFreq + freq) - beta(CenterFreq).
  // See Curtis Menyuk's Multiple Length Scales paper in J. Engr Math
  // equation (25). 

  // Caution: If we ever  change the Fourier Transform we may need to
  // change the signs in this method. To check this you must check that
  // the code is shift-invariant in frequency in the sense described above.

   double FreqSignal;   
   double CurrentStepLength;
   for ( ii = 0; ii < qtStepsFiber; ii++) { 
      if (! WithNonLin) { //## Only one linear step per section
         CurrentStepLength = StepLength[ii];
      } 
      else { //## half linear step before and after a non-linear step
         CurrentStepLength = StepLength[ii]/2.;      
      }
  

      PhaseBirref_2  = jc*2*pi*CurrentStepLength*dBeta_2dOmega;
      PhaseFstOrDispFiber = jc*sq(2*pi)*CurrentStepLength*FstOrDispFiber/2.;
      PhaseSndOrDispFiber = jc*cube(2*pi)*CurrentStepLength*SndOrDispFiber/6.;  
   
//      PhaseBirref_2  = jc*2*pi*StepLength[ii]*dBeta_2dOmega;
//      PhaseFstOrDispFiber = jc*sq(2*pi)*StepLength[ii]*FstOrDispFiber/2.;
//      PhaseSndOrDispFiber = jc*cube(2*pi)*StepLength[ii]*SndOrDispFiber/6.;  
    
      //cout << TransM[0][0][ii] << " " << TransM[0][1][ii] << " " 
      //	   << TransM[1][0][ii] << " " << TransM[1][1][ii] << endl;

      for ( jj = 0; jj < qtPoints; jj++) {  
         FreqSignal = fft->freq(jj);        
         // ## The term CenterFreq in PropM[.] is not part of the Manakov-PMD,
         // ## but I left it here to apply to other scattering techniques.
	 
	 //## Attention: I changed the signd of the birefringence phase (IL)

         //####### This is for a particular application: ################
	 //#### I should comment afterwards
	 //if ( Frequency2Wavelength(FreqSignal) > 800e-9 ) {
	 //   //## The dispersion should be zero above the given lavelength
         //   PhaseFstOrDispFiber = 0.5* jc*sq(2*pi)*CurrentStepLength*FstOrDispFiber/2.;
         //   PhaseSndOrDispFiber = 0.5* jc*cube(2*pi)*CurrentStepLength*SndOrDispFiber/6.; 	
         //}	   


         PropM[0] = exp(  -(CenterFreq+FreqSignal)*PhaseBirref_2 
                        +    sq(FreqSignal)*(PhaseFstOrDispFiber
                        +       FreqSignal* PhaseSndOrDispFiber) 
                        - CurrentStepLength*NepperAttenuationFiber/2); 
         PropM[1] = exp( +(CenterFreq+FreqSignal)*PhaseBirref_2 
                        +    sq(FreqSignal)*(PhaseFstOrDispFiber
                        +       FreqSignal* PhaseSndOrDispFiber) 
                        - CurrentStepLength*NepperAttenuationFiber/2); 

//         PropM[0] = exp(  -(CenterFreq+FreqSignal)*PhaseBirref_2 
//                        +    sq(FreqSignal)*(PhaseFstOrDispFiber
//                        +       FreqSignal* PhaseSndOrDispFiber) 
//                        - StepLength[ii]*NepperAttenuationFiber/2); 
//         PropM[1] = exp( +(CenterFreq+FreqSignal)*PhaseBirref_2 
//                        +    sq(FreqSignal)*(PhaseFstOrDispFiber
//                        +       FreqSignal* PhaseSndOrDispFiber) 
//                        - StepLength[ii]*NepperAttenuationFiber/2); 
     

         EnvelopeX = sfftM.fOptSignalX[jj];
         EnvelopeY = sfftM.fOptSignalY[jj];
         // ## Scatters first, and then propagates trough the fiber step
         sfftM.fOptSignalX[jj] = PropM[0]
              *(TransM[0][0][ii]*EnvelopeX + TransM[0][1][ii]*EnvelopeY);
         sfftM.fOptSignalY[jj] = PropM[1]
              *(TransM[1][0][ii]*EnvelopeX + TransM[1][1][ii]*EnvelopeY);         
	 //## Test line. Comment this line later
	 //if (jj == 0) cout << ii << " "<< TransM[0][0][ii] << "  " << TransM[0][1][ii] << endl;	                              
      }
      //## Non-linear effects: according to the Manakov-PMD equation
      //## It propagates the nonlinear effect through the whole step
      if (WithNonLin) {
         oOptSignal->Freq2Time();
         //fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
         //fft->IFFT(sfftM.fOptSignalY,sfftM.tOptSignalY);           
         for ( jj = 0; jj < qtPoints; jj++) {
            PowerEnvelopeX = sq(abs(sfftM.tOptSignalX[jj]));
            PowerEnvelopeY = sq(abs(sfftM.tOptSignalY[jj]));             
            sfftM.tOptSignalX[jj] = sfftM.tOptSignalX[jj]+
	         (sfftM.tOptSignalX[jj]
                 *jc*gamma
                 *( NonLinCoupling[0][0]*PowerEnvelopeX
                   +NonLinCoupling[0][1]*PowerEnvelopeY)
                + jc*gamma/3.*conj(sfftM.tOptSignalX[jj])
		  *sq(sfftM.tOptSignalY[jj])
		  *exp(-jc*2*ii*StepLength[ii]*2*dBeta_2dOmega*2*pi*CenterFreq)		   
		   )*StepLength[ii];           
            sfftM.tOptSignalY[jj] = sfftM.tOptSignalY[jj]+
	         (sfftM.tOptSignalY[jj]
                 *jc*gamma
                 *( NonLinCoupling[1][0]*PowerEnvelopeX
                   +NonLinCoupling[1][1]*PowerEnvelopeY) 
                + jc*gamma/3.*conj(sfftM.tOptSignalY[jj])
		  *sq(sfftM.tOptSignalX[jj])
		  *exp(jc*2*ii*StepLength[ii]*2*dBeta_2dOmega*2*pi*CenterFreq)			   
		   )*StepLength[ii];   
/**********		                                             	   
            sfftM.tOptSignalX[jj] = sfftM.tOptSignalX[jj]
                 *exp(jc*gamma*StepLength[ii]
                 *( NonLinCoupling[0][0]*PowerEnvelopeX
                   +NonLinCoupling[0][1]*PowerEnvelopeY) );            
            sfftM.tOptSignalY[jj] = sfftM.tOptSignalY[jj]
                 *exp(jc*gamma*StepLength[ii]
                 *( NonLinCoupling[1][0]*PowerEnvelopeX
                   +NonLinCoupling[1][1]*PowerEnvelopeY) ) ;   		   
***********/		   
         }         
         //fft->FFT(sfftM.tOptSignalX,sfftM.fOptSignalX);
         //fft->FFT(sfftM.tOptSignalY,sfftM.fOptSignalY);    
	 
	 if (DampTime != 0) {	 
	    //## Dump the bounds within the specified time
	    oOptSignal->DampBounds(DampTime);
	 }
	 
	 oOptSignal->Time2Freq();

	 //## Now, the second half of the linear step, if WithNonLin
         for ( jj = 0; jj < qtPoints; jj++) {  
            FreqSignal = fft->freq(jj);        
            // ## The term CenterFreq in PropM[.] is not part of the Manakov-PMD,
            // ## but I left it here to apply to other scattering techniques.
	 
   	    //## Attention: I changed the signd of the birefringence phase (IL)

            //## This thing MUST be here, because is frequency dependent
            PropM[0] = exp(  -(CenterFreq+FreqSignal)*PhaseBirref_2 
                           +    sq(FreqSignal)*(PhaseFstOrDispFiber
                           +       FreqSignal* PhaseSndOrDispFiber) 
                           - CurrentStepLength*NepperAttenuationFiber/2); 
            PropM[1] = exp( +(CenterFreq+FreqSignal)*PhaseBirref_2 
                           +    sq(FreqSignal)*(PhaseFstOrDispFiber
                           +       FreqSignal* PhaseSndOrDispFiber) 
                           - CurrentStepLength*NepperAttenuationFiber/2); 
      
            //EnvelopeX = sfftM.fOptSignalX[jj];
            //EnvelopeY = sfftM.fOptSignalY[jj];
            // ## Already scattered, just propagates though the fiber step
            sfftM.fOptSignalX[jj] *= PropM[0];
            //     *(TransM[0][0][ii]*EnvelopeX + TransM[0][1][ii]*EnvelopeY);
            sfftM.fOptSignalY[jj] *= PropM[1];
            //     *(TransM[1][0][ii]*EnvelopeX + TransM[1][1][ii]*EnvelopeY);                           
         }        
      } //## Processes the nonlinear effect and the second half of the linear
   }   //## Finished last step
   //   if ( ! WithNonLin) {   // generates the time dependence
   //      fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
   //      fft->IFFT(sfftM.fOptSignalY,sfftM.tOptSignalY);       
   //   }   
   oOptSignal->Freq2Time();
   *PropagatedLengthFiber += LengthFiber;
}





/***********
// #################################################

//! This propagate method is PropagateFiberCoupledNLS without nonlinearity
//! and done entriely in the frequency domain. No Fourier transforms
//! are taken at all. JZ Sept 6th 2002.

// If you only have linear propagation and 
//    are only analyzing the data in the frequency domain
//    this method will give you god speed. Otherwise don't use it!
//   JZ Sept 6th 2002.

void OptFiber::PropagateFiberCoupledNLSFreqDomainOnly
               (double *PropagatedLength)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }   

   int ii;   // Propagates along the fiber
   int jj;   // Sweep over frequency
   cplx EnvelopeX,EnvelopeY;
   double PowerEnvelopeX, PowerEnvelopeY;        
   cplx PropM[2];
   cplx PhaseBirref_2; //  = jc*2*pi*DeltaZ*dBeta_2dOmega;
   cplx PhaseFstOrDispFiber; // = jc*sq(2*pi)*DeltaZ*FstOrDispFiber/2;
   cplx PhaseSndOrDispFiber; // =-jc*cube(2*pi)*DeltaZ*SndOrDispFiber/6;   
     
   double FreqSignal;   
   double CurrentStepLength;

   for ( ii = 0; ii < qtStepsFiber; ii++) 
     { 
      CurrentStepLength = StepLength[ii];
      
      PhaseBirref_2  = jc*2*pi*CurrentStepLength*dBeta_2dOmega;
      PhaseFstOrDispFiber = jc*sq(2*pi)*CurrentStepLength*FstOrDispFiber/2.;
      PhaseSndOrDispFiber = jc*cube(2*pi)*CurrentStepLength*SndOrDispFiber/6.;
 
      for ( jj = 0; jj < qtPoints; jj++) 
       {  
         FreqSignal = fft->freq(jj);        
         // ## The term CenterFreq in PropM[.] is not part of the Manakov-PMD,
         // ## but I left it here to apply to other scattering techniques.
	 
	 //## Attention: I changed the sign of the birefringence phase (IL)

         PropM[0] = exp(  -(CenterFreq+FreqSignal)*PhaseBirref_2 
                        +    sq(FreqSignal)*(PhaseFstOrDispFiber
                        +       FreqSignal* PhaseSndOrDispFiber) 
                        - CurrentStepLength*NepperAttenuationFiber/2); 
         PropM[1] = exp( +(CenterFreq+FreqSignal)*PhaseBirref_2 
                        +    sq(FreqSignal)*(PhaseFstOrDispFiber
                        +       FreqSignal* PhaseSndOrDispFiber) 
                        - CurrentStepLength*NepperAttenuationFiber/2); 

         EnvelopeX = sfftM.fOptSignalX[jj];
         EnvelopeY = sfftM.fOptSignalY[jj];

         // ## Scatters first, and then propagates trough the fiber step

         sfftM.fOptSignalX[jj] = PropM[0]
              *(TransM[0][0][ii]*EnvelopeX + TransM[0][1][ii]*EnvelopeY);
         sfftM.fOptSignalY[jj] = PropM[1]
              *(TransM[1][0][ii]*EnvelopeX + TransM[1][1][ii]*EnvelopeY);
                        
       } // ## end for qtPoints
     } // ## end for qtStepsFiber

   *PropagatedLength += LengthFiber;

} // ## end PropagateFiberCoupledNLSFreqDomainOnly
************/





// #################################################

//! This propagate method is PropagateFiberCoupledNLS without nonlinearity
//! and done entriely in the frequency domain. No Fourier transforms
//! are taken at all. JZ Sept 6th 2002.

/*! If you only have linear propagation and 
    are only analyzing the data in the frequency domain
    this method will give you god speed. Otherwise don't use it!
   JZ Sept 6th 2002.
*/

void OptFiber::PropagateFiberCoupledNLSFreqDomainOnly
               (double *PropagatedLength)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }   

   int ii;   // Propagates along the fiber
   int jj;   // Sweep over frequency
   cplx EnvelopeX,EnvelopeY;
   double PowerEnvelopeX, PowerEnvelopeY;        
   cplx PropM[2];
   cplx PhaseBirref_2; //  = jc*2*pi*DeltaZ*dBeta_2dOmega;
   cplx PhaseFstOrDispFiber; // = jc*sq(2*pi)*DeltaZ*FstOrDispFiber/2;
   cplx PhaseSndOrDispFiber; // =-jc*cube(2*pi)*DeltaZ*SndOrDispFiber/6;   
     
   double FreqSignal;   
   double CurrentStepLength;

   for ( ii = 0; ii < qtStepsFiber; ii++) 
     { 
      CurrentStepLength = StepLength[ii];
      
      PhaseBirref_2  = jc*2*pi*CurrentStepLength*dBeta_2dOmega;
      PhaseFstOrDispFiber = jc*sq(2*pi)*CurrentStepLength*FstOrDispFiber/2.;
      PhaseSndOrDispFiber = jc*cube(2*pi)*CurrentStepLength*SndOrDispFiber/6.;
 
      for ( jj = 0; jj < qtPoints; jj++) 
       {  
         FreqSignal = fft->freq(jj);        
         // ## The term CenterFreq in PropM[.] is not part of the Manakov-PMD,
         // ## but I left it here to apply to other scattering techniques.
	 
	 //## Attention: I changed the sign of the birefringence phase (IL)

         PropM[0] = exp(  -(CenterFreq+FreqSignal)*PhaseBirref_2 
                        +    sq(FreqSignal)*(PhaseFstOrDispFiber
                        +       FreqSignal* PhaseSndOrDispFiber) 
                        - CurrentStepLength*NepperAttenuationFiber/2); 
         PropM[1] = exp( +(CenterFreq+FreqSignal)*PhaseBirref_2 
                        +    sq(FreqSignal)*(PhaseFstOrDispFiber
                        +       FreqSignal* PhaseSndOrDispFiber) 
                        - CurrentStepLength*NepperAttenuationFiber/2); 

         EnvelopeX = sfftM.fOptSignalX[jj];
         EnvelopeY = sfftM.fOptSignalY[jj];

         // ## Scatters first, and then propagates trough the fiber step

         sfftM.fOptSignalX[jj] = PropM[0]
              *(TransM[0][0][ii]*EnvelopeX + TransM[0][1][ii]*EnvelopeY);
         sfftM.fOptSignalY[jj] = PropM[1]
              *(TransM[1][0][ii]*EnvelopeX + TransM[1][1][ii]*EnvelopeY);
                        
       } // ## end for qtPoints
     } // ## end for qtStepsFiber

   *PropagatedLength += LengthFiber;

} // ## end PropagateFiberCoupledNLSFreqDomainOnly

// ###################################################

void OptFiber::PropagateFiberDecoupledNLS(double *PropagatedLengthFiber)
{

  // Written by John Zweck 6 Oct. 2000
  // This method solves the scalar NLS equation.
  // Portions of this code were borrowed from PropageteFiberCompledNLS() (IL)
  
  // Because of Noise in Amps I figured it's best to 
  // solve for both X and Y parts of signal here,
  // otherwise likely to have problems hard to track down.

  // ## NEED TO UPDATE TO 2ND ORDER ACCURACY
 

  int ii;   // Propagates along the fiber
  int jj;   // Sweep over frequency

   if (! MemAllocated)
    {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      exit(1);
    }   

   cplx PhaseFstOrDispFiber; 
   cplx PhaseSndOrDispFiber; 
     
   double FreqSignal; 
  

  // 12 Feb, 2001. (JZ) I fixed a bug in the sign of the term for 
  // the higher order dispersion SndOrDispDivSix. 
  // With the sign as it is now we have shift-invariance 
  // in frequency: If we change the center frequency then 
  // the only change in the output pulse train is that it
  // it is shifted by the appropriate amount left or right
  // in time. The shape of the pulses will not change.
  // The fact that we have the combination
  // FstOrDispDivTwo + omega*SndOrDispDivSix
  // comes from the fact that this is the start of the Taylor's series
  // for beta(CenterFreq + freq) - beta(CenterFreq).
  // See Curtis Menyuk's Multiple Length Scales paper in J. Engr Math
  // equation (25). 

  // Caution: If we ever  change the Fourier Transform we may need to
  // change the signs in this method. To check this you must check that
  // the code is shift-invariant in frequency in the sense described above.


   for ( ii = 0; ii < qtStepsFiber; ii++) 
    { 
      PhaseFstOrDispFiber = jc*sq(2*pi)*StepLength[ii]*FstOrDispFiber/2.;
      PhaseSndOrDispFiber = jc*cube(2*pi)*StepLength[ii]*SndOrDispFiber/6.;   

      // ## Linear Frequency Domain Evolution

      for ( jj = 0; jj < qtPoints; jj++) 
       {  
         FreqSignal = fft->freq(jj); 
                      // ## The frequency in Hz given by index jj
 
	sfftM.fOptSignalX[jj] = sfftM.fOptSignalX[jj]*
                                exp( sq(FreqSignal)*
                       (PhaseFstOrDispFiber + FreqSignal*PhaseSndOrDispFiber) 
			            - StepLength[ii]*NepperAttenuationFiber/2);

         sfftM.fOptSignalY[jj] = sfftM.fOptSignalY[jj]*
                                exp( sq(FreqSignal)*
                       (PhaseFstOrDispFiber + FreqSignal*PhaseSndOrDispFiber) 
			            - StepLength[ii]*NepperAttenuationFiber/2);



       }// ## end for jj-loop (Frequency domain evolution)


      // ## Non-linear Time Domain Evolution

      if(WithNonLin) 
       {
         fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
         fft->IFFT(sfftM.fOptSignalY,sfftM.tOptSignalY);

         for ( jj = 0; jj < qtPoints; jj++) 
          {
                        
            sfftM.tOptSignalX[jj] = sfftM.tOptSignalX[jj]*
                                    exp(jc*gamma*StepLength[ii]*
				        sq(abs(sfftM.tOptSignalX[jj])));
	
            sfftM.tOptSignalY[jj] = sfftM.tOptSignalY[jj]*
                                    exp(jc*gamma*StepLength[ii]*
				        sq(abs(sfftM.tOptSignalY[jj])));

	  } // end for jj-loop (Time domain evolution) 
  
         fft->FFT(sfftM.tOptSignalX,sfftM.fOptSignalX);
         fft->FFT(sfftM.tOptSignalY,sfftM.fOptSignalY);

       }// end  if (WithNonLin)
       
 
  *PropagatedLengthFiber += StepLength[ii];

 } //## end for ii-loop 

 if (!WithNonLin ) // generates the time dependence
   {
      fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
      fft->IFFT(sfftM.fOptSignalY,sfftM.tOptSignalY);
   }

} // ## end PropagateFiberScalarDecoupledNLS

// #############################################

void OptFiber::PropagateFiberScalarNLS(double *PropagatedLengthFiber)
{

  // Written by John Zweck 6 Oct. 2000
  // This method solves the scalar NLS equation.
  
  // Use in conjunction with AmplifySignalScalar!!!!!!!!

  // ## NEED TO UPDATE TO 2ND ORDER ACCURACY
 

  int ii;   // Propagates along the fiber
  int jj;   // Sweep over frequency

   if (! MemAllocated)
    {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      exit(1);
    }   


   for( jj = 0; jj < qtPoints; jj++) 
     {     
      sfftM.fOptSignalY[jj] = 0.0;
      sfftM.tOptSignalY[jj] = 0.0;
     }

   cplx PhaseFstOrDispFiber; 
   cplx PhaseSndOrDispFiber; 
     
   double FreqSignal; 


  // 12 Feb, 2001. (JZ) I fixed a bug in the sign of the term for 
  // the higher order dispersion SndOrDispDivSix. 
  // With the sign as it is now we have shift-invariance 
  // in frequency: If we change the center frequency then 
  // the only change in the output pulse train is that it
  // it is shifted by the appropriate amount left or right
  // in time. The shape of the pulses will not change.
  // The fact that we have the combination
  // FstOrDispDivTwo + omega*SndOrDispDivSix
  // comes from the fact that this is the start of the Taylor's series
  // for beta(CenterFreq + freq) - beta(CenterFreq).
  // See Curtis Menyuk's Multiple Length Scales paper in J. Engr Math
  // equation (25). 

  // Caution: If we ever  change the Fourier Transform we may need to
  // change the signs in this method. To check this you must check that
  // the code is shift-invariant in frequency in the sense described above.



  
   for ( ii = 0; ii < qtStepsFiber; ii++) 
    { 
      PhaseFstOrDispFiber = jc*sq(2*pi)*StepLength[ii]*FstOrDispFiber/2.;
      PhaseSndOrDispFiber = jc*cube(2*pi)*StepLength[ii]*SndOrDispFiber/6.;  


      // ## Linear Frequency Domain Evolution

      for ( jj = 0; jj < qtPoints; jj++) 
       {  
         FreqSignal = fft->freq(jj); 
                      // ## The frequency in Hz given by index jj
 
	sfftM.fOptSignalX[jj] = sfftM.fOptSignalX[jj]*
                                exp( sq(FreqSignal)*
                       (PhaseFstOrDispFiber + FreqSignal*PhaseSndOrDispFiber) 
			            - StepLength[ii]*NepperAttenuationFiber/2);


       }// ## end for jj-loop (Frequency domain evolution)


      // ## Non-linear Time Domain Evolution

      if(WithNonLin) 
       {
         fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);

         for ( jj = 0; jj < qtPoints; jj++) 
          {
                        
            sfftM.tOptSignalX[jj] = sfftM.tOptSignalX[jj]*
                                    exp(jc*gamma*StepLength[ii]*
				        sq(abs(sfftM.tOptSignalX[jj])));

	  } // end for jj-loop (Time domain evolution) 
  
         fft->FFT(sfftM.tOptSignalX,sfftM.fOptSignalX);

       }// end  if (WithNonLin)
       
 
  *PropagatedLengthFiber += StepLength[ii];

 } //## end for ii-loop 

 if (!WithNonLin ) // generates the time dependence
   {
      fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
   }

} // ## end PropagateFiberScalarNLS

// #############################################

 void OptFiber::PropagateFiberScalarNLSWriteData(double *PropagatedLengthFiber,
                              int WriteFlag,
                              char *DistTimePowerPhaseFileName,
                              int TimeIndexIncrement,
			      double FirstWriteLength,
                              double DeltaWriteLength, 
                              int *NumSlicesZ,
                              int NumSlicesFirstWriteLength)
{

  // Written by John Zweck 6 Oct. 2000
  // This method solves the scalar NLS equation.
  // At z-intervals of DeltaWrite  (z,t,Power) data is written
  // to the file  whose name is given by the char * DistTimePowerFileName
  // The number of slices written out so far is stored in the variable
  // NumSlicesZ


  // ## The PropagateFiberScalarNLS method outputs dist-time-power
  // ## data at  FirstWriteLength + k*DeltaWriteLength for k=0,1,2...
  // ## if WriteFlag is set.
  // ## The int *NumSlicesZ keeps track of how many times we have written
  // ## out the data along the whole transmission line.
  // ## Note that it you can initially have 
  // ## PropogatedLengthFiber > FirstWriteLength
  // ## in which case you output at the first k for which 
  // ## PropogatedLengthFiber >= FirstWriteLength + k*DeltaWriteLength
  // ## To make sure this is done correctly we need to also pass 
  // ## NumSlicesFirstWriteLength. 


  // ## NEED TO UPDATE TO 2ND ORDER ACCURACY

  // ## FIX: Later make out file names to be data in optical signal class
 

  int ii;   // Propagates along the fiber
  int jj;   // Sweep over frequency

   if (! MemAllocated)
    {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      exit(1);
    }   


   cplx PhaseFstOrDispFiber; // = jc*sq(2*pi)*DeltaZ*FstOrDispFiber/2;
   cplx PhaseSndOrDispFiber; // =-jc*cube(2*pi)*DeltaZ*SndOrDispFiber/6;   
     
   double FreqSignal; 
  
   for ( ii = 0; ii < qtStepsFiber; ii++) 
    { 
      PhaseFstOrDispFiber = jc*sq(2*pi)*StepLength[ii]*FstOrDispFiber/2.;
      PhaseSndOrDispFiber = jc*cube(2*pi)*StepLength[ii]*SndOrDispFiber/6.;   

      // ## Linear Frequency Domain Evolution

      for ( jj = 0; jj < qtPoints; jj++) 
       {  
         FreqSignal = fft->freq(jj); 
                      // ## The frequency in Hz given by index jj
 
	sfftM.fOptSignalX[jj] = sfftM.fOptSignalX[jj]*
                                exp( sq(FreqSignal)*
                       (PhaseFstOrDispFiber + FreqSignal*PhaseSndOrDispFiber) 
			            - StepLength[ii]*NepperAttenuationFiber/2);



       }// ## end for jj-loop (Frequency domain evolution)


      // ## Non-linear Time Domain Evolution

      if(WithNonLin) 
       {
         fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
         

         for ( jj = 0; jj < qtPoints; jj++) 
          {
                        
            sfftM.tOptSignalX[jj] = sfftM.tOptSignalX[jj]*
                                    exp(jc*gamma*StepLength[ii]*
				        sq(abs(sfftM.tOptSignalX[jj])));
	

         

	  } // end for jj-loop (Time domain evolution) 
  
         fft->FFT(sfftM.tOptSignalX,sfftM.fOptSignalX);
         

       }// end  if (WithNonLin)
       
 
  *PropagatedLengthFiber += StepLength[ii];

  if(WriteFlag && 
     *PropagatedLengthFiber > FirstWriteLength + 
             (*NumSlicesZ - NumSlicesFirstWriteLength)*DeltaWriteLength)
    {

      if (!WithNonLin) // generates the time dependence
          fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
       
      // ## FIX ME!!!!

      //oOptSignal.WriteDistTimePowerPhase(DistTimePowerPhaseFileName,
      //                         TimeIndexIncrement,*PropagatedLengthFiber,1);

      // The 1 in third argument means append file

      (*NumSlicesZ)++;

      // cout << "NumSlices in Z  " << *NumSlicesZ << endl;


    }// end if

 } //## end for ii-loop 

 if (!WithNonLin ) // generates the time dependence
   {
      fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
      

   }

} // ## end PropagateFiberScalarNLSWriteData

// ###########################################

// ## Change the length and the number of steps
// #  It has improved modularity compared to the old one
// ## By: John Zweck and Ivan Lima 
void OptFiber::SetLengthFiberAndQtStepsFiber(double Length,int numSteps)
{
   LengthFiber = Length;
   qtStepsFiber = numSteps;
   ReleaseMemory();
   AllocateMemory_SetLengthSteps();  
}

/**********
  void OptFiber::SetLengthFiberAndQtStepsFiber(double Length,int numSteps)
{

  // ## Written by John Zweck 6 Oct 2000.
  // ## Adapted from Ivan Lima's method OptFiber::InitFiberParam

  LengthFiber = Length; 
  qtStepsFiber = numSteps;

  if(StepLength != NULL)
    delete[] StepLength;

  StepLength = NULL;

  StepLength = new double[qtStepsFiber];  
   
 if (MemAllocated && qtStepsFiber != qtStepsFiberLast) 
   {
      ReleaseMemory();
      MemAllocated = 0;
   }
   if ( ! MemAllocated) 
    { 
      MemAllocated = 1;
      phi     = new double[qtStepsFiber];
      psi     = new double[qtStepsFiber];   
      theta   = new double[qtStepsFiber]; 
      StepLength = new double[qtStepsFiber];  

      for (int ii = 0; ii < 2; ii++)
         for (int jj = 0; jj < 2; jj++)
            TransM[ii][jj] = new cplx[qtStepsFiber];
      qtStepsFiberLast = qtStepsFiber;        
    } 
  
   GotFiberSample = 0;

   DeltaZ = LengthFiber/qtStepsFiber;
   dBeta_2dOmega = PMD_Fiber/sqrt(4*DeltaZ); 
   //ExpectedDGD = sqrt(8/(3*pi)*LengthFiber)*PMD_Fiber;
   //ExpectedSndOrderDGD = sq(ExpectedDGD)/sqrt(3.);
   TypeScattering = 1; // ## Euler Matrix (Default)

   for (int ii = 0; ii < qtStepsFiber; ii++)
      StepLength[ii] = DeltaZ;    

} // ## end SetLengthFiberAndQtStepsFiber
**************/

// #############################################

void OptFiber::SetSndOrDispFiber(double SndOrDispFiber2)
{

 SndOrDispFiber = SndOrDispFiber2;

 if(DebugLevel)
   {
 LogFileSeparator();

 *LogFile << "WARNING: Second order dispersion reset to " 
         <<  SndOrDispFiber << endl << endl;

   }
}

// #######################################

void OptFiber::SetFstOrDispFiber(double FstOrDispFiber2)
{

 FstOrDispFiber = FstOrDispFiber2;

 if(DebugLevel)
   {
 LogFileSeparator();

 *LogFile << "WARNING: First order dispersion reset to " 
         <<  FstOrDispFiber << endl << endl;
   }
}

// ###################################

void OptFiber::SetGammaFiber(double gamma2)
{

  // Written by John Zweck, Nov 2nd, 2000

 gamma = gamma2;
 WithNonLin  = (gamma != 0); 

 if(DebugLevel)
   {
 LogFileSeparator();

 *LogFile << "WARNING: Nonlinearity coefficient  reset to Gamma = " 
         <<  gamma << endl << endl;
   }  

} // end SetGammaFiber

// ######################################################

void OptFiber::SetAttenuation(double AttenuationFiber2)
{

  // Written by John Zweck, Nov 2nd, 2000

AttenuationFiber = AttenuationFiber2;
NepperAttenuationFiber = AttenuationFiber*log(10.0)/10.0; 

 if(DebugLevel)
 *LogFile << "WARNING: Reset AttenuationFiber to " 
         << AttenuationFiber << endl;

}
//############################################################

double OptFiber::GetLocalBeatLength(void)
   {return 2*pi/abs(2*dBeta_2dOmega*2*pi*CenterFreq);}

// ##########################################################

//! Propagates a given Jones vector through an optical fiber.
/*! The PMD at relative frequency, RelFreq, is measured relative to the
    center frequency of the channel. Compatible with 
    OptFiber::PropagateFiberCoupledNLS provided no dispersion or nonlinearity.
    Written by JZ, Dec 2nd 2005
*/

void OptFiber::TransformJonesVectorThroughFiber(double RelFreq,cplx *JonesVec)
{

  if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }   

double CurrentStepLength; 

cplx PropM[2];
cplx PhaseBirref_2;
cplx TmpJonesVec[2];

for (int ii = 0; ii < qtStepsFiber; ii++)
{

 CurrentStepLength = StepLength[ii];
 PhaseBirref_2  = jc*2*pi*CurrentStepLength*dBeta_2dOmega;

 PropM[0] = exp(-(CenterFreq+RelFreq)*PhaseBirref_2);
 PropM[1] = exp(+(CenterFreq+RelFreq)*PhaseBirref_2);

 TmpJonesVec[0] = JonesVec[0];
 TmpJonesVec[1] = JonesVec[1];

JonesVec[0] = PropM[0]
          *(TransM[0][0][ii]*TmpJonesVec[0] + TransM[0][1][ii]*TmpJonesVec[1]);

JonesVec[1] = PropM[1]
          *(TransM[1][0][ii]*TmpJonesVec[0] + TransM[1][1][ii]*TmpJonesVec[1]);

  } // end for

} // ## end TransformJonesVectorThroughFiber

