// $Id:
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsOptFiber.hh"
extern ofstream *LogFile;


//## Methods of OptFiber class related to PMD


/*! 

   This method calls the random number generator as many times
   as required to generate the number of fibers specified 
   in the input parameter qtOptFibersSkiped. 
   
   This feature allows to accumulate parallel PMD statistics, since
   all the fibers used in all the parallel processes are guaranteed
   to be statistically uncorrelated.   For exmple, if one needs to 
   make a statistics that requires 100,000 fibers, one could run
   10 independent processes with 10,000 fibers each. However one
   needs to skip the fibers that are going to be generated by other 
   proesses using this method. 
   The first process could generate the first 10,000 fibers.
   The second process skips the first 10,000 fibers, then generates the
   following 10,000 fibers. The third process skips the first 20,000
   fiber, then generate the following 10,000 fibers, and so on.

*/

//################################################################
//## Skip a number of optical fibers to allow parallel processing 
//## of statistically independent samples.
//## By: Ivan Lima
void OptFiber::SkipOptFibers(int qtOptFibersSkipped)
{
   for (int ii = 0; ii < qtOptFibersSkipped; ii++) {
      GetFiberSampleCompleteScatt();
   }
   string OutFileName = "qtOptFibersSkipped.dat";
   ofstream OutFile(OutFileName.c_str(),ios::out);   
   if (! OutFile ) {
      cout << endl;
      cout << "From OptFiber::SkipOptFibers(.)" << endl;
      cout << "Error: " << OutFileName
           << " could not be created."
           << endl;
      exit(1);
   }	    
   OutFile << qtOptFibersSkipped << endl;
   OutFile.close();  
}



/*!

   This methods generates random mode couplings between the birefringent
   sections by a random polarization rotation. This polarization rotation 
   transforms any input polarization state to a random output polarization 
   state whose probability density function is uniformly distributed on 
   the Poincar\'e sphere. 

   This rotation is obtained in this method using the three Euler angles. 

*/

// ################################################################
// ## Scatters the polarization uniformly on the poincare sphere
// ## By: Ivan Lima
void OptFiber::GetFiberSampleCompleteScatt()
{
   if (! MemAllocated) {
      cerr << "Error: Object of class OptFiber not initialized." 
           << endl;
      exit(1);
   }   
   TypeScattering = 1; // ## Euler Matrix   
   GotFiberSample = 1;   
   for (int ii = 0; ii < qtStepsFiber; ii++) {
      if (PMD_Fiber != 0) { // ## CNLS -> Manakov-PMD
         // ## Generate the Euler angles            
         cvtheta   = 2*(RNG->GetRanNum() -0.5);                        
         theta[ii] = acos(cvtheta);                     
         phi[ii]   = 2*pi*RNG->GetRanNum();
         psi[ii]   = 2*pi*RNG->GetRanNum();
         
         //cout << cvtheta/2. + 0.5 << endl;
         //cout << phi[ii]/(2.*pi) << endl;         
         //cout << psi[ii]/(2.*pi) << endl;
         //GetEulerMatrix(ii); 
      }        
      else {  // ## No PMD_Fiber -> no mode coupling due to birefringence
         theta[ii] = 0.;                     
         phi[ii]   = 0.;          
         psi[ii]   = 0.; 
         //TransM[0][0][ii] = 1;
         //TransM[0][1][ii] = 0;
         //TransM[1][0][ii] = 0;
         //TransM[1][1][ii] = 1;     
      }      
      GetEulerMatrix(ii); 

      //cout << theta[ii] << "\t" << phi[ii] << "\t" << psi[ii] << endl;

   }        
}


/*!

   This method generates an Euler matrix for the mode coupling
   of the section specified in the input parameter IndexStep. 
   The Euler matrix corresponds to the rotation T = Rx * Ry * Rx.
   
   The Euler angles used to
   generate the matrix are stored in the vectors phi[], theta[],
   and psi[], which are variables of the OptFiber class.   

   This is an internal method.

*/

//###############################################################
// ## Generates the Euler matrix for the section IndexStep
// ## By: Ivan Lima
void OptFiber::GetEulerMatrix(int IndexStep) // ## ii -> IndexStep
{ 
   psi[IndexStep] *=-1;
   theta[IndexStep] *=-1;
   phi[IndexStep] *=-1;   
      
   TransM[0][0][IndexStep] = exp( jc*(psi[IndexStep]+phi[IndexStep])/2.)
                            *cos(theta[IndexStep]/2.);
   TransM[0][1][IndexStep] = jc*exp( jc*(psi[IndexStep]-phi[IndexStep])/2.)
                            *sin(theta[IndexStep]/2.);
   TransM[1][0][IndexStep] = jc*exp(-jc*(psi[IndexStep]-phi[IndexStep])/2.)
                            *sin(theta[IndexStep]/2.);
   TransM[1][1][IndexStep] = exp(-jc*(psi[IndexStep]+phi[IndexStep])/2.)
                            *cos(theta[IndexStep]/2.); 			    	    
			    
   psi[IndexStep] *=-1;
   theta[IndexStep] *=-1;
   phi[IndexStep] *=-1;    			    
}


/*!

   This method generates a rotation T = Ry * Rz * Ry
   for the mode coupling of the section specified in the input 
   parameter IndexStep. This matrix can produces the same effect of
   the Euler matrix. The angles used to generate the matrix are 
   stored in the vectors phi[], theta[],
   and psi[], which are variables of the OptFiber class.

   This method is applied in the use of importance sampling
   to bias first-order PMD. However, it was replaced by the method
   GetRxRzRyMatrix(). Therefore, it is OUT OF DATE.  

   This is an internal method.

*/

//###############################################################
// ## Generates the Matrix with rotation RyRzRy
// #  in the Jones representation.
// #  It is essencial for the Importance Sampling
// ## By: Ivan Lima
void OptFiber::GetRyRzRyMatrix(int IndexStep) // ## ii -> IndexStep
{
//   TransM[0][0][IndexStep] = cos(psi[IndexStep]/2)*cos(theta[IndexStep]/2.)
//                         +jc*sin(psi[IndexStep]/2)*sin(theta[IndexStep]/2.);
//   TransM[0][1][IndexStep] = cos(psi[IndexStep]/2)*sin(theta[IndexStep]/2.)
//                         +jc*sin(psi[IndexStep]/2)*cos(theta[IndexStep]/2.); 
//   TransM[1][0][IndexStep] =-cos(psi[IndexStep]/2)*sin(theta[IndexStep]/2.)
//                         +jc*sin(psi[IndexStep]/2)*cos(theta[IndexStep]/2.); 
//   TransM[1][1][IndexStep] = cos(psi[IndexStep]/2)*cos(theta[IndexStep]/2.)
//                         -jc*sin(psi[IndexStep]/2)*sin(theta[IndexStep]/2.); 

  
   psi[IndexStep] *=-1;
//   theta[IndexStep] *=-1;
   phi[IndexStep] *=-1;    
                         
                         
   TransM[0][0][IndexStep] =     cos(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2)
                             -jc*cos(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                             +jc*sin(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                                -sin(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2); 
   TransM[0][1][IndexStep] =  jc*sin(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2) 
                             +   sin(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                             +   cos(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                             +jc*cos(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2);
   TransM[1][0][IndexStep] = jc*cos(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                     *cos(theta[IndexStep]/2)
                            -   cos(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                     *sin(theta[IndexStep]/2)
                            -   sin(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                     *sin(theta[IndexStep]/2)
                            +jc*sin(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                     *cos(theta[IndexStep]/2); 
   TransM[1][1][IndexStep] = -   sin(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2)
                            - jc*sin(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                            + jc*cos(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                            +    cos(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2);                         

   psi[IndexStep] *=-1;
//   theta[IndexStep] *=-1;
   phi[IndexStep] *=-1;  
}



/*!

   This method generates a rotation T = Rx * Rz * Ry * Rz * Ry
   for the mode coupling of the section specified in the input 
   parameter IndexStep. This matrix can produces the same effect of
   the Euler matrix. The angles used to generate the matrix are 
   stored in the vectors phi[], theta[],
   psi[], ksi[] and delta[], which are variables of the OptFiber class.

   This method is applied in the use of one importance sampling method
   to bias the first- and second-order PMD. However, this method
   is not the most efficient one. Therefore, this method SHOULD
   NOT BE USED for importance sampling applied to PMD.  One 
   shoud use the method GetRzRxRzRyRzRyMatrix(). 
   
   This is an internal method that is OUT OF DATE.

*/

//###############################################################
//## Generates the Matrix with rotation RyRzRy
//#  in the Jones representation.
//#  It is essencial for the Importance Sampling
//#  applied to the 2nd order PDV
//## By: Ivan Lima
void OptFiber::GetRxRzRyRzRyMatrix(int IndexStep) // ## ii -> IndexStep
{
   psi[IndexStep] *=-1;
   //theta[IndexStep] *=-1;
   phi[IndexStep] *=-1;  
   //delta[IndexStep] *=-1;
   ksi[IndexStep] *= -1;

   cplx TransMa[2][2];  //## RyRzRy                         
   TransMa[0][0] =     cos(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2)
                             -jc*cos(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                             +jc*sin(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                                -sin(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2); 
   TransMa[0][1] =  jc*sin(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2) 
                             +   sin(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                             +   cos(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                             +jc*cos(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2);
   TransMa[1][0] = jc*cos(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                     *cos(theta[IndexStep]/2)
                            -   cos(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                     *sin(theta[IndexStep]/2)
                            -   sin(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                     *sin(theta[IndexStep]/2)
                            +jc*sin(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                     *cos(theta[IndexStep]/2); 
   TransMa[1][1] = -   sin(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2)
                            - jc*sin(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                            + jc*cos(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                            +    cos(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2);     
   cplx TransMb[2][2]; //## RxRz
   TransMb[0][0] = exp( jc*ksi[IndexStep]/2.)
                *(  cos(00./2)*cos(delta[IndexStep]/2.)
                +jc*sin(00./2)*sin(delta[IndexStep]/2.));
   TransMb[0][1] = exp( jc*ksi[IndexStep]/2.) 
                *(  cos(00./2)*sin(delta[IndexStep]/2.)
                +jc*sin(00./2)*cos(delta[IndexStep]/2.)); 
   TransMb[1][0] = exp(-jc*ksi[IndexStep]/2.)
                *( -cos(00./2)*sin(delta[IndexStep]/2.)
               +jc*sin(00./2)*cos(delta[IndexStep]/2.)); 
   TransMb[1][1] = exp(-jc*ksi[IndexStep]/2.)
                *(  cos(00./2)*cos(delta[IndexStep]/2.)
                -jc*sin(00./2)*sin(delta[IndexStep]/2.));         

   //## Putting all together: RxRzRyRzRy
   TransM[0][0][IndexStep] = TransMb[0][0]*TransMa[0][0]
                            +TransMb[0][1]*TransMa[1][0];   
   TransM[0][1][IndexStep] = TransMb[0][0]*TransMa[0][1]
                            +TransMb[0][1]*TransMa[1][1];
   TransM[1][0][IndexStep] = TransMb[1][0]*TransMa[0][0]
                            +TransMb[1][1]*TransMa[1][0];                               
   TransM[1][1][IndexStep] = TransMb[1][0]*TransMa[0][1]
                            +TransMb[1][1]*TransMa[1][1];                               

   psi[IndexStep] *=-1;
   //theta[IndexStep] *=-1;
   phi[IndexStep] *=-1;  
   //delta[IndexStep] *=-1;
   ksi[IndexStep] *= -1;

}



/*!

   This method generates a rotation T = Rz * Rx * Rz * Ry * Rz
   for the mode coupling of the section specified in the input 
   parameter IndexStep. This matrix can produces the same effect of
   the Euler matrix. The angles used to generate the matrix are 
   stored in the vectors phi[], theta[],
   psi[], ksi[] and delta[], which are variables of the OptFiber class.

   This method is applied in the use of importance sampling method
   to bias first- and second-order PMD. This is the most efficient
   method in the OptFiber class for this purpose.

   This is an internal method.

*/

//###############################################################
//## Generates the Matrix with rotation RyRzRy
//#  in the Jones representation.
//#  It is essencial for the Importance Sampling
//#  applied to the 2nd order PDV as proposed by Bill Kath
//## By: Aurenice Lima
void OptFiber::GetRzRxRzRyRzRyMatrix(int IndexStep) // ## ii -> IndexStep
{
   psi[IndexStep] *=-1;
   phi[IndexStep] *=-1;  
   ksi[IndexStep] *= -1;

   cplx TransMa[2][2];  //## RyRzRy                         
   TransMa[0][0] =     cos(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2)
                             -jc*cos(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                             +jc*sin(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                                -sin(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2); 
   TransMa[0][1] =  jc*sin(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2) 
                             +   sin(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                             +   cos(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                             +jc*cos(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2);
   TransMa[1][0] = jc*cos(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                     *cos(theta[IndexStep]/2)
                            -   cos(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                     *sin(theta[IndexStep]/2)
                            -   sin(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                     *sin(theta[IndexStep]/2)
                            +jc*sin(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                     *cos(theta[IndexStep]/2); 
   TransMa[1][1] = -   sin(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2)
                            - jc*sin(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                            + jc*cos(psi[IndexStep]/2)*sin(phi[IndexStep]/2)
                                                      *sin(theta[IndexStep]/2)
                            +    cos(psi[IndexStep]/2)*cos(phi[IndexStep]/2)
                                                      *cos(theta[IndexStep]/2);     
   cplx TransMb[2][2]; //## RxRz
   TransMb[0][0] = exp( jc*ksi[IndexStep]/2.)
                *(  cos(00./2)*cos(delta[IndexStep]/2.)
                +jc*sin(00./2)*sin(delta[IndexStep]/2.));
   TransMb[0][1] = exp( jc*ksi[IndexStep]/2.) 
                *(  cos(00./2)*sin(delta[IndexStep]/2.)
                +jc*sin(00./2)*cos(delta[IndexStep]/2.)); 
   TransMb[1][0] = exp(-jc*ksi[IndexStep]/2.)
                *( -cos(00./2)*sin(delta[IndexStep]/2.)
                +jc*sin(00./2)*cos(delta[IndexStep]/2.)); 
   TransMb[1][1] = exp(-jc*ksi[IndexStep]/2.)
                *(  cos(00./2)*cos(delta[IndexStep]/2.)
                -jc*sin(00./2)*sin(delta[IndexStep]/2.));         

   //## Putting together: RxRzRyRzRy
   cplx TransMc[2][2];
   TransMc[0][0] = TransMb[0][0]*TransMa[0][0]
                            +TransMb[0][1]*TransMa[1][0];   
   TransMc[0][1] = TransMb[0][0]*TransMa[0][1]
                            +TransMb[0][1]*TransMa[1][1];
   TransMc[1][0] = TransMb[1][0]*TransMa[0][0]
                            +TransMb[1][1]*TransMa[1][0];                               
   TransMc[1][1] = TransMb[1][0]*TransMa[0][1]
                            +TransMb[1][1]*TransMa[1][1];                               




   cplx TransMd[2][2]; //## Rz
   TransMd[0][0] = exp( jc*00/2.)
                *(  cos(00./2)*cos(chi[IndexStep]/2.)
                +jc*sin(00./2)*sin(chi[IndexStep]/2.));
   TransMd[0][1] = exp( jc*00/2.) 
                *(  cos(00./2)*sin(chi[IndexStep]/2.)
                +jc*sin(00./2)*cos(chi[IndexStep]/2.)); 
   TransMd[1][0] = exp(-jc*00/2.)
                *( -cos(00./2)*sin(chi[IndexStep]/2.)
               +jc*sin(00./2)*cos(chi[IndexStep]/2.)); 
   TransMd[1][1] = exp(-jc*00/2.)
                *(  cos(00./2)*cos(chi[IndexStep]/2.)
                -jc*sin(00./2)*sin(chi[IndexStep]/2.));         

   //## Putting all together: RzRxRzRyRzRy
   TransM[0][0][IndexStep] = TransMd[0][0]*TransMc[0][0]
                            +TransMd[0][1]*TransMc[1][0];   
   TransM[0][1][IndexStep] = TransMd[0][0]*TransMc[0][1]
                            +TransMd[0][1]*TransMc[1][1];
   TransM[1][0][IndexStep] = TransMd[1][0]*TransMc[0][0]
                            +TransMd[1][1]*TransMc[1][0];                               
   TransM[1][1][IndexStep] = TransMd[1][0]*TransMc[0][1]
                            +TransMd[1][1]*TransMc[1][1];                               


   psi[IndexStep] *=-1;
   phi[IndexStep] *=-1;  
   ksi[IndexStep] *= -1;

}


/*!

   This method generates a rotation T = Rx * Rz * Ry
   for the mode coupling of the section specified in the input 
   parameter IndexStep. The angles used to generate the matrix are 
   stored in the vectors phi[], theta[],
   and psi[], which are variables of the OptFiber class.

   This method is applied in the use of importance sampling applied
   to first-order PMD.  
   
   This is an internal method that is OUT OF DATE, since importance 
   sampling with first- and second-order PMD bias should be used 
   even when one wants to bias only the first-order PMD.

*/


//###############################################################
// ## Generates the Matrix with rotation RxRzRy
// #  in the Jones representation.
// #  It is essencial for the Importance Sampling
// ## By: Ivan Lima
void OptFiber::GetRxRzRyMatrix(int IndexStep) // ## ii -> IndexStep
{

   psi[IndexStep] *=-1;
   //theta[IndexStep] *=-1;
   phi[IndexStep] *=-1;  

   TransM[0][0][IndexStep] = exp( jc*phi[IndexStep]/2.)
                *(  cos(psi[IndexStep]/2)*cos(theta[IndexStep]/2.)
                +jc*sin(psi[IndexStep]/2)*sin(theta[IndexStep]/2.));
   TransM[0][1][IndexStep] = exp( jc*phi[IndexStep]/2.) 
                *(  cos(psi[IndexStep]/2)*sin(theta[IndexStep]/2.)
                +jc*sin(psi[IndexStep]/2)*cos(theta[IndexStep]/2.)); 
   TransM[1][0][IndexStep] = exp(-jc*phi[IndexStep]/2.)
                *( -cos(psi[IndexStep]/2)*sin(theta[IndexStep]/2.)
               +jc*sin(psi[IndexStep]/2)*cos(theta[IndexStep]/2.)); 
   TransM[1][1][IndexStep] = exp(-jc*phi[IndexStep]/2.)
                *(  cos(psi[IndexStep]/2)*cos(theta[IndexStep]/2.)
                -jc*sin(psi[IndexStep]/2)*sin(theta[IndexStep]/2.));                                          

   psi[IndexStep] *=-1;
   //theta[IndexStep] *=-1;
   phi[IndexStep] *=-1;  
}


/*!

   This method generates a fiber sample where the random mode
   coupling between each section is produced by a rotation Rz 
   that is produced by a combination of three rotations, 
   T = Rx * Ry * Rx.

   This method can be used when one wants to model fibers with
   PMD using rotators.
   For most applications, it is preferred to use a model that
   produces a uniform rotation on the Poincare shpere, such as
   that one that can be obtained with the Euler angles.

*/


//###############################################################
// ## Rotates the birefringent axis from uniform dist from 0 to 2*pi
// ## By: Ivan Lima
void OptFiber::GetFiberSampleUnifRz()
{
   if (! MemAllocated) {
      cerr << "Error: Object of class OptFiber not initialized." 
           << endl;
      exit(1);
   }   
   TypeScattering = 1; // ## Euler Matrix
   GotFiberSample = 1;   
   for (int ii = 0; ii < qtStepsFiber; ii++) {
      if (PMD_Fiber != 0) { 
         // ## Generate the Euler angles             
         theta[ii] =  2*pi*RNG->GetRanNum();  // ## RotY        
         phi[ii]   =  pi/2;                 // ## RotX
         psi[ii]   = -pi/2;                 // ## RotX                     
         //GetEulerMatrix(ii); 
      }        
      else {  // ## No PMD_Fiber -> no mode coupling: Two independent NLS
         psi[ii]   =  0.;
         theta[ii] =  pi/2.;
         phi[ii]   = -pi/2.;  
         //TransM[0][0][ii] = 1;
         //TransM[0][1][ii] = 0;
         //TransM[1][0][ii] = 0;
         //TransM[1][1][ii] = 1;     
      }      
      GetEulerMatrix(ii); 
   }         
}


/*!

   This method generates a fiber sample where the random mode
   coupling between each section is produced by the combination
   of three random rotations: Ry, Rz, and Ry, whose angles of rotation
   are uniformly distributed between 0 and 2 $\pi$. The polarization
   transformation in each section is equal to T = Ry * Ry * Rx.

   This method does not produce a uniform distribution of the 
   output Stokes vector on the Poinca\'e sphere.
   For most applications, it is preferred to use a model that
   produces a uniform distribution of the Stokes vector on 
   the Poincar\'e shpere, such as
   that one that can be obtained using the Euler angles.

   This method is OUT OF DATE.

*/

//###############################################################
// ## Rotates RyRzRy
// ## By: Ivan Lima
void OptFiber::GetFiberSampleUnifRyRzRy()
{  // ## It doesn't provide a uniform scattering on the poincare sphere
   // #  One would need to set cos(theta) uniformly on the Poincare 
   // #  Sphere. But we already have the Euler matrix for that purpose
   if (! MemAllocated) {
      cerr << "Error: Object of class OptFiber not initialized." 
           << endl;
      exit(1);
   }   
   TypeScattering = 2; // ## RyRzRy
   GotFiberSample = 1;   
   for (int ii = 0; ii < qtStepsFiber; ii++) {
      if (PMD_Fiber != 0) { 
         psi[ii]   = 2*pi*RNG->GetRanNum();  // ## RotY
         cvtheta   = 2*(RNG->GetRanNum()-0.5);                        
         theta[ii] = acos(cvtheta);  // ## RotZ                                           
         phi[ii]   = 2*pi*RNG->GetRanNum();  // ## RotY 
         //GetRyRzRyMatrix(ii); 
      }        
      else {  // ## No PMD_Fiber -> no mode coupling: Two independent NLS
         psi[ii]   = 0.;
         theta[ii] = 0.;
         phi[ii] = 0.;  
         //TransM[0][0][ii] = 1;
         //TransM[0][1][ii] = 0;
         //TransM[1][0][ii] = 0;
         //TransM[1][1][ii] = 1;     
      }      
      GetRyRzRyMatrix(ii); 
   }         
}


/*!

   This method generates a fiber sample where the random mode
   coupling between each section is produced by the combination
   of three random rotations: Rx, Rz, and Ry, whose angles of rotation
   are uniformly distributed between 0 and 2 $\pi$. The polarization
   transformation in each section is equal to T = Rx * Rz * Ry.

   This method does not produce a uniform distribution of the 
   output Stokes vector on the Poincare sphere.
   For most applications, it is preferred to use a model that
   produces a uniform rotation on the Poincare shpere, such as
   that one that can be obtained using the Euler angles.

   This method is OUT OF DATE.

*/

//###############################################################
// ## Rotates RxRzRy uniformly
// ## By: Ivan Lima
void OptFiber::GetFiberSampleUnifRxRzRy()
{  // ## It doesn't provide a uniform scattering on the poincare sphere
   // #  One would need to set cos(theta) uniformly on the Poincare 
   // #  Sphere. But we already have the Euler matrix for that purpose
   if (! MemAllocated) {
      cerr << "Error: Object of class OptFiber not initialized." 
           << endl;
      exit(1);
   }   
   TypeScattering = 3; // ## RxRzRy   
   GotFiberSample = 1;   
   for (int ii = 0; ii < qtStepsFiber; ii++) {
      if (PMD_Fiber != 0) { 
         psi[ii]   = 2*pi*RNG->GetRanNum();  // ## RotY
         cvtheta   = 2*(RNG->GetRanNum()-0.5);                        
         theta[ii] = acos(cvtheta);        // ## ROtZ        
         //theta[ii] = 2*pi*RNG->GetRanNum();  // ## RorZ                              
         phi[ii]   = 2*pi*RNG->GetRanNum();  // ## RotX 
         //GetRxRzRyMatrix(ii); 
      }        
      else {  // ## No PMD_Fiber -> no mode coupling: Two independent NLS
         psi[ii]   = 0.;
         theta[ii] = 0.;
         phi[ii] =   0.;  
         //TransM[0][0][ii] = 1;
         //TransM[0][1][ii] = 0;
         //TransM[1][0][ii] = 0;
         //TransM[1][1][ii] = 1;     
      }      
      GetRxRzRyMatrix(ii); 
   }         
}



/*!

   This method generates a fiber sample with importance sampling
   applied to first-order PMD.  The random mode
   coupling between each section is produced by the combination
   of three random rotations: Rx, Rz, and Ry, that produce an uniform 
   distribution of the Stokes vector on the Poincare sphere. 
   The polarization transformation in each section is equal 
   to T = Rx * Rz * Ry.

   The biasing function used for importance sampling is based on
   the cosine functiion.  It is recommendable to use the method
   GetFiberSampleImportSamplingExp() if you want to use importance
   sampling to bias first-order PMD. 
   Therefore, this method is OUT OF DATE.

*/

//###############################################################
//## Get the low probable DGD values using Important Sampling
//#  using the biasing proposed by Bill Kath.
//#  This function has been used until 10/17/01.
//#  YOU SHOULD NOT BE USING THIS FUNCTION (though it works)
//#  I think that one exponentially decreasing is smoother
//#  You should now use GetFiberSampleImportSamplingExp(.)
//## By: Ivan Lima
double OptFiber::GetFiberSampleImportSamplingCosRx(double BiasingParam)
{
   if (! MemAllocated) {
      cerr << "Error: Object of class OptFiber not initialized." 
           << endl;
      exit(1);
   }   
   if (PMD_Fiber == 0) {
      cerr << "Error: The procedure GetFiberSampleImportSamplingCosRx(...)"
           << "Is being called for a fiber with PMD == 0." 
           << endl;
      exit(1);
   }   
   //if (BiasingParam =< 0) {
   //   cerr << "From GetFiberSampleImportSamplingCosRx(.)"
   //        << "Error: Biasing param "<< BiasingParam 
   //        << " Cannot be =< 0." 
   //        << endl;
   //   exit(1);   
   //}            
   double SourceRandomVar;
   double LikelihoodRatio = 1;
   TypeScattering = 3; // ## RxRzRy 
   GotFiberSample = 1;   
   //int qtStepsFiberTmp = qtStepsFiber;   
   int LastStep;
   double v2Delta;
   double OutPD_Vector[] = {0,0,0}, OutPD_VectorTmp[3];
   double thetaZthetaY[2];     
   //double Rotation[3][3]; // ## it is necessary for TransV3to_Xhat 
   // ## Generate the Euler angles with complete scattering
   // #  for the first section.                      
   cvtheta   = 2*(RNG->GetRanNum()-0.5);                        
   theta[0] = acos(cvtheta);                     
   phi[0]   = 2*pi*RNG->GetRanNum();          
   psi[0]   = 2*pi*RNG->GetRanNum();     
   GetRxRzRyMatrix(0);    
   for (int ii = 1; ii < qtStepsFiber; ii++) {
      // ## Generate the Euler angles

      SourceRandomVar = RNG->GetRanNum();
      theta[ii] = acos(2.*pow(SourceRandomVar,1./BiasingParam) - 1. );
      LikelihoodRatio *= 1./BiasingParam
                *pow(SourceRandomVar,-(1.-1./BiasingParam));
      
      //## This do not belong here
      //SourceRandomVar = RNG->GetRanCosineBiased(
      //                                     BiasingParam,
      //                                     &LikelihoodRatio);     
      //theta[ii] = sgn(RNG->GetRanNum() - 0.5)
      //           *acos(SourceRandomVar);
      //## End
      
      phi[ii] =  2*pi*RNG->GetRanNum();                    
        
      //if (ii > 0) {
         LastStep = ii-1;   
         // ## Calculates the cumulative Pol. Disp. Vector
         v2Delta = 2*dBeta_2dOmega*StepLength[LastStep]*2*pi*CenterFreq;                    
         RotatesAboutY( -psi[LastStep],   OutPD_Vector); //, Rotation);                    
         RotatesAboutZ(  theta[LastStep], OutPD_Vector); //, Rotation);  
         RotatesAboutX(  phi[LastStep],   OutPD_Vector); //, Rotation);                      
         RotatesAboutX(  v2Delta,  OutPD_Vector); //, Rotation);   // ## Propagation    
         OutPD_Vector[0] += -2*dBeta_2dOmega*StepLength[LastStep];  

	       
         for (int jj = 0; jj < 3; jj++)
            OutPD_VectorTmp[jj] = OutPD_Vector[jj];         
         TransRzRyV3to_Xhat(OutPD_VectorTmp,thetaZthetaY); //,Rotation);                   
         theta[ii] +=  thetaZthetaY[0];
         psi[ii]    = -thetaZthetaY[1];          
      //}                      
      GetRxRzRyMatrix(ii);                   
   }   
   return(LikelihoodRatio);
}



/*!

   This method generates a fiber sample with importance sampling
   to bias first-order PMD.  The random mode
   coupling between each section is produced by the combination
   of three random rotations: Rx, Rz, and Ry, that produce a Stokes 
   vector that is uniformly distributed on the Poincare sphere. 
   The polarization transformation in each section is equal 
   to T = Rx * Rz * Ry.

   The biasing function used for importance sampling is based on
   the exponential functiion.  

   To cover the differential group delay (DGD), which is the first-order
   PMD, from 0 to 4 times
   the mean DGD of a fiber with 80 sections one may  
   use 0.0, 0.5, and 1.0 for the bias parameter BiasParam.
   All but $10^{-8}$ of the cumulative distribution function (cdf)
   of the DGD is in the region where the DGD is less than 4 times
   the mean DGD of a fiber.

   The use of 80 section to model a fiber should be sufficient for most
   applications. For a fiber with 200 sections one may 
   use 0.0, 0.3, and 0.6 for the bias parameter BiasParam, if it is necessary.

   One should use importance sampling method to bias first- and second-order 
   PMD even when one wants to bias only first-order PMD if one wants
   to combine samples from multiple distributions using the balanced
   heuristic method.
   
*/

//###############################################################
//## Get the low probable DGD values with Important Sampling
//#  using the biased pdf proposed by Ivan Lima.
//#  Note that the importance sampling applied to PMD was
//#  proposed by Biondini and Kath. I just happend to be
//#  the first one to apply it to PMD-induced distortions 
//#  and PMD compensators.
//#  This is the right function to use.
//#  To cover the normalized DGD from 0 to 4 with 80 sections, 
//#  use bias 0.5 and 1.0, besides the unbiased.
//#  To cover the normalized DGD from 0 to 4 with 200 sections, 
//#  which provides a more accurate emulation of the tail, use 
//#  bias 0.3 and 0.6, besides the unbiased.
//#  I would like to remind that all but $10^{-8}$ of the DGD 
//#  are in the range from 0 to 4.
//## By: Ivan Lima (10/18/01)
double OptFiber::GetFiberSampleImportSamplingExp(double BiasingParam)
{
   if (! MemAllocated) {
      cerr << "Error: Object of class OptFiber not initialized." 
           << endl;
      exit(1);
   }   
   if (PMD_Fiber == 0) {
      cerr << "Error: The procedure GetFiberSampleImportSamplingCosRx(...)"
           << "Is being called for a fiber with PMD == 0." 
           << endl;
      exit(1);
   }   
        
   double SourceRandomVar;
   double LikelihoodRatio = 1;
   TypeScattering = 3; //## RxRzRy 
   GotFiberSample = 1;   
   //int qtStepsFiberTmp = qtStepsFiber;   
   int LastStep;
   double v2Delta;
   double OutPD_Vector[] = {0,0,0}, OutPD_VectorTmp[3];
   double thetaZthetaY[2];     

   double SectionDGD, SectionDGD_Last;

   //## Generate a random rotation for the first section.                      
   //## It can be later changed if one wants to bias the 
   //#  first order PMD distortion
   cvtheta   = 2*(RNG->GetRanNum()-0.5);                        
   theta[0] = acos(cvtheta);                     
   phi[0]   = 2*pi*RNG->GetRanNum();          
   psi[0]   = 2*pi*RNG->GetRanNum();     
   GetRxRzRyMatrix(0);    
   
   SectionDGD = 2*dBeta_2dOmega*StepLength[0];  
   for (int ii = 1; ii < qtStepsFiber; ii++) {
   
      SectionDGD_Last = SectionDGD;
      SectionDGD = 2*dBeta_2dOmega*StepLength[ii];
              
      LastStep = ii-1;   
      
      // ## Calculates the cumulative Pol. Disp. Vector
      v2Delta = SectionDGD_Last*2*pi*CenterFreq;                    
      RotatesAboutY( -psi[LastStep],   OutPD_Vector); //, Rotation);                    
      RotatesAboutZ(  theta[LastStep], OutPD_Vector); //, Rotation);  
      RotatesAboutX(  phi[LastStep],   OutPD_Vector); //, Rotation);                      
      RotatesAboutX(  v2Delta,  OutPD_Vector); //, Rotation);   // ## Propagation    
      OutPD_Vector[0] += -SectionDGD_Last;  
            
      for (int jj = 0; jj < 3; jj++) {
         OutPD_VectorTmp[jj] = OutPD_Vector[jj];         
      }   
      TransRzRyV3to_Xhat(OutPD_VectorTmp,thetaZthetaY); //,Rotation);                   
      //## Generate biased theta and psi. The angle psi aids
      //#  the biasing process.

      //## Generate the rotation angles
      SourceRandomVar = RNG->GetRanCosineBiased(BiasingParam,
                                              &LikelihoodRatio);                                                 
      theta[ii] = acos(SourceRandomVar)*sgn(RNG->GetRanNum() - 0.5);                                        
      phi[ii] =  2*pi*RNG->GetRanNum();               
      theta[ii] +=  thetaZthetaY[0];
      psi[ii]    = -thetaZthetaY[1];          
      GetRxRzRyMatrix(ii);                   
   }   
   return(LikelihoodRatio);
}



/*!

   This method generates a fiber sample with importance sampling
   to bias both first- and second-order PMD.  The random mode
   coupling between each section is produced by the combination
   of five rotations: Rx, Rz, Ry, Rz, and Ry, that produce an uniform 
   Stokes vector on the Poincar\'e sphere. The polarization
   transformation in each section is equal to T = Rx * Rz * Ry * Rz * Ry.

   The biasing function used for importance sampling is based on
   the exponential functiion, where first- and second-order PMD are
   biased independently. Since firs- and second-order PMD are not
   independent quantities, it is not efficient to bias these two
   quantities independently.  
   
   The input parameter BiasingParamFstOrder is the bias parameter of 
   the first-order PMD, and the input parameter BiasingParamSndOrder
   is the bias parameter of the second-order PMD. The parameter
   BiasedAngleFstOrder determines the angle between the polarization
   disperison vector of the previous section and the direction to which
   the current section will be biased. For pure first-order PMD bias
   this angle must be equal to zero. For pure second-order PMD bias,
   this angle must be equal to pi/2, where BiasingParamFstOrder must
   be equal to zero (unbiased). Different combinations of BiasedAngleFstOrder
   between 0 and pi/2 and and values of BiasingParamFstOrder and
   BiasingParamSndOrder must be used to cover the first- and second-order
   PMD plane.

   To cover the differential group delay (DGD), which is 
   the first-order PMD, from 0 to 4 times
   the mean DGD of a fiber with 80 sections one may  
   use 0.0, 0.5, and 1.0 for the bias parameter.
   All but $10^{-8}$ of the cumulative distribution function (cdf)
   of the DGD is in the region where the DGD is less than 4 times
   the mean DGD of a fiber.

   This function is OUT OF DATE, since it does not take into account
   the correlation of the second-order PMD with the first-order PMD.
 
*/

//###############################################################
//## Get the low probable 2nd order PDV values with
//#  Important Sampling as Ivan and Aurenice Lima
//#  THIS METHOD IS OKAY, but you should use
//#  GetFiberSampleImportSamplingSndOrderPDV_VariableAngleExp()
//#  as Curtis has made it mandatory.
//## By: Ivan and Aurenice Lima. Adapted from the codes contained
//#  in GetFiberSampleImportSamplingSndOrderPDV_Exp() by Aurenice. (4/3/2002)
double OptFiber::GetFiberSampleImportSamplingFstSndOrderPDV_Exp(
                                    double BiasedAngleFstOrder,
                                    double BiasingParamFstOrder,
				    double BiasingParamSndOrder)
{
   if (! MemAllocated) {
      cerr << "Error: Object of class OptFiber not initialized." 
           << endl;
      exit(1);
   }   
   if (PMD_Fiber == 0) {
      cerr << "Error: The method OptFiber::GetFiberSampleImportSamplingCosRx(...)"
           << "Is being called for a fiber with PMD == 0." 
           << endl;
      exit(1);
   }   
        
   double RandomAngleFstOrder;
   double RandomAngleSndOrder;
   double LikelihoodRatio = 1;
   TypeScattering = 4; //## RxRzRyRzRy 
   
   GotFiberSample = 1;   
   int LastStep;
   double v2Delta;
   double OutPD_Vector[] = {0,0,0};
   double OutPD_VectorTmp[3];   
   double SndOrderOutPD_Vector[] = {0,0,0};
   double SndOrderOutPD_VectorTmp[3];
     
   double thetaZthetaY[2];     
   double thetaY;

   //## Generate a random rotation for the first section.                                          
   //#  It has no effect on the output PDV.   
   cvtheta   = 2*(RNG->GetRanNum()-0.5);                        
   theta[0] = acos(cvtheta);  //## Ry                   
   phi[0]   = 2*pi*RNG->GetRanNum();  //## Rz            
   psi[0]   = 2*pi*RNG->GetRanNum();  //## Ry
   delta[0] = 0;  //## Rz  
   ksi[0]   = 0; //2*pi*RNG->GetRanNum();   //## Rx

   GetRxRzRyRzRyMatrix(0);    
   
   double SectionDGD, SectionDGD_Last;   
   SectionDGD = 2*dBeta_2dOmega*StepLength[0];
   for (int ii = 1; ii < qtStepsFiber; ii++) {
      SectionDGD_Last = SectionDGD;
      SectionDGD = 2*dBeta_2dOmega*StepLength[ii];
 
      LastStep = ii-1;   
      
      // ## Calculates the cumulative Pol. Disp. Vector
      v2Delta = SectionDGD_Last*2*pi*CenterFreq;                    
      RotatesAboutY( -psi[LastStep],   OutPD_Vector); //, Rotation);                    
      RotatesAboutZ(  theta[LastStep], OutPD_Vector); //, Rotation);  
      RotatesAboutY( -phi[LastStep],   OutPD_Vector); //, Rotation);                    
      RotatesAboutZ(  delta[LastStep], OutPD_Vector); //, Rotation);  
      RotatesAboutX(  ksi[LastStep],   OutPD_Vector); //, Rotation);                      
      RotatesAboutX(  v2Delta,         OutPD_Vector); //, Rotation);
      OutPD_Vector[0] += -SectionDGD_Last;  
      
      RotatesAboutY( -psi[LastStep],   SndOrderOutPD_Vector); //, Rotation);                    
      RotatesAboutZ(  theta[LastStep], SndOrderOutPD_Vector); //, Rotation);  
      RotatesAboutY( -phi[LastStep],   SndOrderOutPD_Vector); //, Rotation);                    
      RotatesAboutZ(  delta[LastStep], SndOrderOutPD_Vector); //, Rotation);  
      RotatesAboutX(  ksi[LastStep],   SndOrderOutPD_Vector); //, Rotation);                      
      RotatesAboutX(  v2Delta,         SndOrderOutPD_Vector); //, Rotation);

      //## Add the cross product between the OutPD_Vectors
      SndOrderOutPD_Vector[1] +=  SectionDGD_Last*OutPD_Vector[2];
      SndOrderOutPD_Vector[2] += -SectionDGD_Last*OutPD_Vector[1];
      
      for (int jj = 0; jj < 3; jj++) {
         OutPD_VectorTmp[jj] = OutPD_Vector[jj];
         SndOrderOutPD_VectorTmp[jj] = SndOrderOutPD_Vector[jj];         
      }   
   
      TransRzRyV3to_Xhat(OutPD_VectorTmp,thetaZthetaY); 
      //## But I really need to rotate to -Yhat      
      RotatesAboutZ( -pi/2., OutPD_VectorTmp);

      //## Apply the same rotations to SndOrderOutPD_VectorTmp
      RotatesAboutY( thetaZthetaY[1], SndOrderOutPD_VectorTmp);    
      RotatesAboutZ( thetaZthetaY[0] - pi/2., SndOrderOutPD_VectorTmp);         

      //## Now I should eliminate the X component of SndOrderOutPD_Vector
      //#  while leaving the Z component positive: 
      TransRyV3to_YZ( SndOrderOutPD_VectorTmp, &thetaY);

      //## Apply the same rotations to OutPD_VectorTmp
      RotatesAboutY( thetaY,   OutPD_VectorTmp);
  
      psi[ii]   = -thetaZthetaY[1];       
      theta[ii] =  thetaZthetaY[0] - pi/2.;                                                
      
      //## Generate the rotation angles for the Second Order PDV  
      //## The unbiased case is uniform from 0 to 2*pi
      RandomAngleSndOrder = pi*(1 - (1+ RNG->GetRanCosineBiased(
            BiasingParamSndOrder,
            &LikelihoodRatio) )/2 );
	    //*sgn(RNG->GetRanNum() - 0.5);

      //## Generate the rotation angles for the First Order PDV           
      //## The unbiased case has the cosine uniform from -1 to 1
      RandomAngleFstOrder = acos(RNG->GetRanCosineBiased( BiasingParamFstOrder,
                                              cos(BiasedAngleFstOrder),
                                              &LikelihoodRatio) );  

      phi[ii] = -thetaY + RandomAngleSndOrder*sgn(RNG->GetRanNum() - 0.5);
      delta[ii] = -RandomAngleFstOrder + pi/2.;  
		         
      ksi[ii] =  2*pi*RNG->GetRanNum();                    
              
      GetRxRzRyRzRyMatrix(ii);                   
   }   
   return(LikelihoodRatio);
}



/*!

   This method generates a fiber sample with importance sampling
   to bias both first- and second-order PMD.  
   It is the same as GetFiberSampleImportSamplingFstSndOrderPDV_Exp(),
   except that the actual biased angle for the first-order PMD bias
   in each section varies linearly from zero to BiasedAngle as the 
   section number varies from the first to the last section in the fiber.
 
   The choice of this linearly varying angle for the first-order PMD
   bias is because large second-order PMD bias also require large
   first-order PMD bias.This phenomenon was demonstrated by G. Biondini 
   and W. L. Kath. 
   
   The input parameter BiasingParamFstOrder is the bias parameter of 
   the first-order PMD, and the input parameter BiasingParamSndOrder
   is the bias parameter of the second-order PMD. The parameter
   BiasedAngleFstOrder determines the angle between the polarization
   disperison vector of the previous section and the direction to which
   the current section will be biased, which varies linearly along 
   the fiber sections as described
   in the previous paragraph. For pure first-order PMD bias
   this angle must be equal to zero. For pure second-order PMD biase,
   this angle must be equal to pi/2, where BiasingParamFstOrder must
   be equal to zero (unbiased). Different combinations of BiasedAngleFstOrder
   between 0 and pi/2 and and values of BiasingParamFstOrder and
   BiasingParamSndOrder must be used to cover the first- and second-order
   PMD plane.

   To cover the differential group delay (DGD), which is 
   the first-order PMD, from 0 to 4 times
   the mean DGD of a fiber with 80 sections one may  
   use 0.0, 0.5, and 1.0 for the bias parameter.
   All but $10^{-8}$ of the cumulative distribution function (cdf)
   of the DGD is in the region where the DGD is less than 4 times
   the mean DGD of a fiber.

   This function is OUT OF DATE, since it does not take into account
   the correlation of the second-order PMD with the first-order PMD.
 
*/

//###############################################################
//## Get the low probable 2nd order PDV values with
//#  Important Sampling as Ivan and Aurenice Lima
//#  THIS METHOD IS OKAY
//#  GetFiberSampleImportSamplingSndOrderPDV_VariableAngleExp()
//#  as Curtis has made it mandatory.
//## This is the adaptive version, where the actual BiasedAngle
//#  varies linearly from 0 to the received value of BiasedAngle.
//#  The goal of this method is to be able to resolve rare values
//#  of the second order component of the polarization dispersion vector
//#  that are associated with small values of the first order component.
//## The linear rotation of the method was mentioned by Bill Kath.
//## By: Ivan and Aurenice Lima. Adapted from the codes contained
//#  in GetFiberSampleImportSamplingSndOrderPDV_Exp() by Aurenice. (4/3/2002)
double OptFiber::GetFiberSampleImportSamplingFstSndOrderPDV_VariableAngleExp(
                                    double BiasedAngleFstOrder,
                                    double BiasingParamFstOrder,
				    double BiasingParamSndOrder)
{
   if (! MemAllocated) {
      cerr << "Error: Object of class OptFiber not initialized." 
           << endl;
      exit(1);
   }   
   if (PMD_Fiber == 0) {
      cerr << "Error: The method OptFiber::GetFiberSampleImportSamplingCosRx(...)"
           << "Is being called for a fiber with PMD == 0." 
           << endl;
      exit(1);
   }   
        
   double RandomAngleFstOrder;
   double RandomAngleSndOrder;
   double LikelihoodRatio = 1;
   TypeScattering = 4; //## RxRzRyRzRy 
   
   GotFiberSample = 1;   
   int LastStep;
   double v2Delta;
   double OutPD_Vector[] = {0,0,0};
   double OutPD_VectorTmp[3];   
   double SndOrderOutPD_Vector[] = {0,0,0};
   double SndOrderOutPD_VectorTmp[3];
     
   double thetaZthetaY[2];     
   double thetaY;

   //## Generate a random rotation for the first section.                                          
   //#  It has no effect on the output PDV.   
   cvtheta   = 2*(RNG->GetRanNum()-0.5);                        
   theta[0] = acos(cvtheta);  //## Ry                   
   phi[0]   = 2*pi*RNG->GetRanNum();  //## Rz            
   psi[0]   = 2*pi*RNG->GetRanNum();  //## Ry
   delta[0] = 0;  //## Rz  
   ksi[0]   = 0; //2*pi*RNG->GetRanNum();   //## Rx

   GetRxRzRyRzRyMatrix(0);    
   
   double SectionDGD, SectionDGD_Last;   
   SectionDGD = 2*dBeta_2dOmega*StepLength[0];
   for (int ii = 1; ii < qtStepsFiber; ii++) {
      SectionDGD_Last = SectionDGD;
      SectionDGD = 2*dBeta_2dOmega*StepLength[ii];
 
      LastStep = ii-1;   
      
      // ## Calculates the cumulative Pol. Disp. Vector
      v2Delta = SectionDGD_Last*2*pi*CenterFreq;                    
      RotatesAboutY( -psi[LastStep],   OutPD_Vector); //, Rotation);                    
      RotatesAboutZ(  theta[LastStep], OutPD_Vector); //, Rotation);  
      RotatesAboutY( -phi[LastStep],   OutPD_Vector); //, Rotation);                    
      RotatesAboutZ(  delta[LastStep], OutPD_Vector); //, Rotation);  
      RotatesAboutX(  ksi[LastStep],   OutPD_Vector); //, Rotation);                      
      RotatesAboutX(  v2Delta,         OutPD_Vector); //, Rotation);
      OutPD_Vector[0] += -SectionDGD_Last;  
      
      RotatesAboutY( -psi[LastStep],   SndOrderOutPD_Vector); //, Rotation);                    
      RotatesAboutZ(  theta[LastStep], SndOrderOutPD_Vector); //, Rotation);  
      RotatesAboutY( -phi[LastStep],   SndOrderOutPD_Vector); //, Rotation);                    
      RotatesAboutZ(  delta[LastStep], SndOrderOutPD_Vector); //, Rotation);  
      RotatesAboutX(  ksi[LastStep],   SndOrderOutPD_Vector); //, Rotation);                      
      RotatesAboutX(  v2Delta,         SndOrderOutPD_Vector); //, Rotation);

      //## Add the cross product between the OutPD_Vectors
      SndOrderOutPD_Vector[1] +=  SectionDGD_Last*OutPD_Vector[2];
      SndOrderOutPD_Vector[2] += -SectionDGD_Last*OutPD_Vector[1];
      
      for (int jj = 0; jj < 3; jj++) {
         OutPD_VectorTmp[jj] = OutPD_Vector[jj];
         SndOrderOutPD_VectorTmp[jj] = SndOrderOutPD_Vector[jj];         
      }   
   
      TransRzRyV3to_Xhat(OutPD_VectorTmp,thetaZthetaY); 
      //## But I really need to rotate to -Yhat      
      RotatesAboutZ( -pi/2., OutPD_VectorTmp);

      //## Apply the same rotations to SndOrderOutPD_VectorTmp
      RotatesAboutY( thetaZthetaY[1], SndOrderOutPD_VectorTmp);    
      RotatesAboutZ( thetaZthetaY[0] - pi/2., SndOrderOutPD_VectorTmp);         

      //## Now I should eliminate the X component of SndOrderOutPD_Vector
      //#  while leaving the Z component positive: 
      TransRyV3to_YZ( SndOrderOutPD_VectorTmp, &thetaY);

      //## Apply the same rotations to OutPD_VectorTmp
      RotatesAboutY( thetaY,   OutPD_VectorTmp);
  
      psi[ii]   = -thetaZthetaY[1];       
      theta[ii] =  thetaZthetaY[0] - pi/2.;                                                
      
      //## Generate the rotation angles for the Second Order PDV  
      //## The unbiased case is uniform from 0 to 2*pi
      RandomAngleSndOrder = pi*(1 - (1+ RNG->GetRanCosineBiased(
            BiasingParamSndOrder,
            &LikelihoodRatio) )/2 );
	    //*sgn(RNG->GetRanNum() - 0.5);

      //## Generate the rotation angles for the First Order PDV           
      //## The unbiased case has the cosine uniform from -1 to 1
      RandomAngleFstOrder = acos(RNG->GetRanCosineBiased( BiasingParamFstOrder,
                                              cos(BiasedAngleFstOrder),
                                              &LikelihoodRatio) );  

      phi[ii] = -thetaY + RandomAngleSndOrder*sgn(RNG->GetRanNum() - 0.5);

      //## Here comes the linear bias of the biased angle
      delta[ii] = -RandomAngleFstOrder*ii/(qtStepsFiber-1) + pi/2.;              
      //delta[ii] = -RandomAngleFstOrder + pi/2.;  		         
      
      ksi[ii] =  2*pi*RNG->GetRanNum();                    
              
      GetRxRzRyRzRyMatrix(ii);                   
   }   
   return(LikelihoodRatio);
}



/*!

   This method generates a fiber sample with importance sampling
   to bias both first- and second-order PMD.  

   The input parameter BiasingParam is the bias parameter for 
   both the first- and the second-order PMD. The parameter
   BiasedAngle determines the angle between the polarization
   disperison vector of the previous section and the direction to which
   the current section will be biased. The second-order PMD
   is indirectly biased when BiasedAngle is different from zero and $\pi$.

   This method is OUT OF DATE.
   
*/

//###############################################################
//## Get the low probable 2nd order PDV values with
//#  Important Sampling as proposed by "Kath et al."
//## By: Aurenice Lima, adapted from the first order bias by Ivan Lima
double OptFiber::GetFiberSampleImportSamplingSndOrderPDV_Exp(
                                     double BiasedAngle,
                                     double BiasingParam) 
{
   if (! MemAllocated) {
      cerr << "Error: Object of class OptFiber not initialized." 
           << endl;
      exit(1);
   }   
   if (PMD_Fiber == 0) {
      cerr << "Error: The method OptFiber::GetFiberSampleImportSamplingCosRx(...)"
           << "Is being called for a fiber with PMD == 0." 
           << endl;
      exit(1);
   }   
        
   double SourceRandomVar;
   double LikelihoodRatio = 1;
   TypeScattering = 5; //## RzRxRzRyRzRy 
   
   GotFiberSample = 1;   
   int LastStep;
   double v2Delta;
   double OutPD_Vector[] = {0,0,0};
   double OutPD_VectorTmp[3];   
   double SndOrderOutPD_Vector[] = {0,0,0};
   double SndOrderOutPD_VectorTmp[3];
     
   double thetaZthetaY[2];     
   double thetaY;

   //## Generate a random rotation for the first section.                      
   //#  It has no effect on the output PDV.
   cvtheta   = 2*(RNG->GetRanNum()-0.5);                        
   theta[0] = acos(cvtheta);  //## Ry                   
   phi[0]   = 2*pi*RNG->GetRanNum();  //## Rz            
   psi[0]   = 2*pi*RNG->GetRanNum();  //## Ry
   delta[0] = 0;  //## Rz  
   ksi[0]   = 0; //2*pi*RNG->GetRanNum();   //## Rx
   chi[0]   = 0; //## Rz

   GetRzRxRzRyRzRyMatrix(0);    
   
   double SectionDGD, SectionDGD_Last;   
   SectionDGD = 2*dBeta_2dOmega*StepLength[0];
   for (int ii = 1; ii < qtStepsFiber; ii++) {
      SectionDGD_Last = SectionDGD;
      SectionDGD = 2*dBeta_2dOmega*StepLength[ii];
 
      LastStep = ii-1;   
      
      // ## Calculates the cumulative Pol. Disp. Vector
      v2Delta = SectionDGD_Last*2*pi*CenterFreq;                    
      RotatesAboutY( -psi[LastStep],   OutPD_Vector);
      RotatesAboutZ(  theta[LastStep], OutPD_Vector);  
      RotatesAboutY( -phi[LastStep],   OutPD_Vector); 
      RotatesAboutZ(  delta[LastStep], OutPD_Vector); 
      RotatesAboutX(  ksi[LastStep],   OutPD_Vector); 
      RotatesAboutZ(  chi[LastStep], OutPD_Vector);
      RotatesAboutX(  v2Delta,         OutPD_Vector); 
      OutPD_Vector[0] += -SectionDGD_Last;  
      
      RotatesAboutY( -psi[LastStep],   SndOrderOutPD_Vector); 
      RotatesAboutZ(  theta[LastStep], SndOrderOutPD_Vector); 
      RotatesAboutY( -phi[LastStep],   SndOrderOutPD_Vector);                   
      RotatesAboutZ(  delta[LastStep], SndOrderOutPD_Vector); 
      RotatesAboutX(  ksi[LastStep],   SndOrderOutPD_Vector);
      RotatesAboutZ(  chi[LastStep], SndOrderOutPD_Vector); 
      RotatesAboutX(  v2Delta,         SndOrderOutPD_Vector);

      //## Add the cross product between the OutPD_Vectors
      SndOrderOutPD_Vector[1] +=  SectionDGD_Last*OutPD_Vector[2];
      SndOrderOutPD_Vector[2] += -SectionDGD_Last*OutPD_Vector[1];
      
      for (int jj = 0; jj < 3; jj++) {
         OutPD_VectorTmp[jj] = OutPD_Vector[jj];
         SndOrderOutPD_VectorTmp[jj] = SndOrderOutPD_Vector[jj];         
      }      

      TransRzRyV3to_Xhat(OutPD_VectorTmp,thetaZthetaY); //,Rotation);                   
      //## But I really need to rotate to -Yhat      
      RotatesAboutZ( -pi/2., OutPD_VectorTmp);

      //## Apply the same rotations to SndOrderOutPD_VectorTmp
      RotatesAboutY( thetaZthetaY[1], SndOrderOutPD_VectorTmp);    
      RotatesAboutZ( thetaZthetaY[0] - pi/2., SndOrderOutPD_VectorTmp);         

      //## Now I should eliminate the X component of SndOrderOutPD_Vector
      //#  while leaving the Z component positive: 
      TransRyV3to_YZ( SndOrderOutPD_VectorTmp, &thetaY);

      //## Apply the same rotations to OutPD_VectorTmp
      RotatesAboutY( thetaY,   OutPD_VectorTmp);
  
      psi[ii]   = -thetaZthetaY[1];       
      theta[ii] =  thetaZthetaY[0] - pi/2.;                                                
      phi[ii]   = -thetaY; // *0000000000000.;

      //## Generate the rotation angles      
      SourceRandomVar = RNG->GetRanCosineBiased(BiasingParam,
                                              &LikelihoodRatio); 
      delta[ii] = pi/2. - BiasedAngle;  
      
      chi[ii] =  acos(SourceRandomVar)*sgn(RNG->GetRanNum() - 0.5);

      ksi[ii] =  2*pi*RNG->GetRanNum();                    
              
      GetRzRxRzRyRzRyMatrix(ii);                   
   }   
   return(LikelihoodRatio);
}



/*!

   This method generates a fiber sample with importance sampling
   to bias both first- and second-order PMD. This biasing technique
   was proposed by G. Biondini and W. L. Kath. 

   The input parameter BiasingParam is the bias parameter for 
   both the first- and the second-order PMD. The parameter
   BiasedAngle determines the angle between the polarization
   disperison vector of the previous section and the direction to which
   the current section will be biased, which varies linearly along 
   the fiber sections from zero to the maximum value defined
   in BiasedAngle.
   
   
   The second-order PMD
   is indirectly biased when BiasedAngle is different from zero and $\pi$.   
   For pure first-order PMD bias
   this angle must be equal to zero. For pure second-order PMD biase,
   this angle must be equal to pi/2. 
   Different combinations of BiasedAngle
   between 0 and pi/2 and and values of BiasingParam and
   must be used to cover the first- and second-order
   PMD plane of interest.

   To cover the first- and second-order PMD plane of a fiber with 80 
   sections one may use the parameters (BiaingParam,BiasedAngle) =
   (0,0), (0.5, 0), (0.5, pi/3) (0.5,2 pi/3), (0.5, pi)
   (0.7,0), (1, 0), (1, pi/3) (1,2 pi/3), (1, pi).
   These distributions are sufficient to accurately compute outage 
   probabilities of 10^{-7}.
   These samples must be combined using the balanced heuristics method
   implemented in GetHeuristicBalanceForSndOrderPDV_VariableAngleExp().
   
   This is the method of choice to generate fibers using importance
   sampling to bias first- and second-order PMD.
   
*/

//###############################################################
//## Get the low probable 2nd order PDV values with
//#  Important Sampling as proposed by "Kath et al."
//## This is the adaptive version, where the actual BiasedAngle
//#  varies linearly from 0 to the received value of BiasedAngle.
//#  The goal of this method is to be able to resolve rare values
//#  of the second order component of the polarization dispersion vector
//#  that are associated with small values of the first order component.
//## By: Aurenice Lima, adapted from the first order bias by Ivan Lima

double OptFiber::GetFiberSampleImportSamplingSndOrderPDV_VariableAngleExp(
                                     double BiasedAngle,
                                     double BiasingParam) 
{
   if (! MemAllocated) {
      cerr << "Error: Object of class OptFiber not initialized." 
           << endl;
      exit(1);
   }   
   if (PMD_Fiber == 0) {
      cerr << "Error: The method OptFiber::GetFiberSampleImportSamplingCosRx(...)"
           << "Is being called for a fiber with PMD == 0." 
           << endl;
      exit(1);
   }   

   //## Create a file that makes possible the use of Heuristic balance
   if (! FlagImportSamplingSndOrderPDV_VariableAngleExp ) {
      FlagImportSamplingSndOrderPDV_VariableAngleExp = 1;
      string OutFileName = "BiasSndOrderPDV_VariableAngleExp.dat";
      ofstream OutFile(OutFileName.c_str(),ios::out);   
      if (! OutFile ) {
         cout << endl;
         cout << "Error: " << OutFileName
              << " could not be created."
              << endl;
         exit(1);
      }	   
      OutFile << BiasingParam << "\t" << BiasedAngle 
              << "\t" << BiasedAngle*180/pi <<  endl;
      OutFile.close();
   }
        
   double SourceRandomVar;
   double LikelihoodRatio = 1;
   TypeScattering = 5; //## RzRxRzRyRzRy 
   
   GotFiberSample = 1;   
   int LastStep;
   double v2Delta;
   double OutPD_Vector[] = {0,0,0};
   double OutPD_VectorTmp[3];   
   double SndOrderOutPD_Vector[] = {0,0,0};
   double SndOrderOutPD_VectorTmp[3];
     
   double thetaZthetaY[2];     
   double thetaY;

   //## Generate a random rotation for the first section.                      
   //#  It has no effect on the output PDV.
   cvtheta   = 2*(RNG->GetRanNum()-0.5);                        
   theta[0] = acos(cvtheta);  //## Ry                   
   phi[0]   = 2*pi*RNG->GetRanNum();  //## Rz            
   psi[0]   = 2*pi*RNG->GetRanNum();  //## Ry
   delta[0] = 0;  //## Rz  
   ksi[0]   = 0; //2*pi*RNG->GetRanNum();   //## Rx
   chi[0]   = 0; //## Rz

   GetRzRxRzRyRzRyMatrix(0);    
   
   double SectionDGD, SectionDGD_Last;   
   SectionDGD = 2*dBeta_2dOmega*StepLength[0];
   for (int ii = 1; ii < qtStepsFiber; ii++) {
      SectionDGD_Last = SectionDGD;
      SectionDGD = 2*dBeta_2dOmega*StepLength[ii];
 
      LastStep = ii-1;   
      
      // ## Calculates the cumulative Pol. Disp. Vector
      v2Delta = SectionDGD_Last*2*pi*CenterFreq;                    
      RotatesAboutY( -psi[LastStep],   OutPD_Vector);
      RotatesAboutZ(  theta[LastStep], OutPD_Vector);  
      RotatesAboutY( -phi[LastStep],   OutPD_Vector); 
      RotatesAboutZ(  delta[LastStep], OutPD_Vector); 
      RotatesAboutX(  ksi[LastStep],   OutPD_Vector); 
      RotatesAboutZ(  chi[LastStep], OutPD_Vector);
      RotatesAboutX(  v2Delta,         OutPD_Vector); 
      OutPD_Vector[0] += -SectionDGD_Last;  
      
      RotatesAboutY( -psi[LastStep],   SndOrderOutPD_Vector); 
      RotatesAboutZ(  theta[LastStep], SndOrderOutPD_Vector); 
      RotatesAboutY( -phi[LastStep],   SndOrderOutPD_Vector);                   
      RotatesAboutZ(  delta[LastStep], SndOrderOutPD_Vector); 
      RotatesAboutX(  ksi[LastStep],   SndOrderOutPD_Vector);
      RotatesAboutZ(  chi[LastStep], SndOrderOutPD_Vector); 
      RotatesAboutX(  v2Delta,         SndOrderOutPD_Vector);

      //## Add the cross product between the OutPD_Vectors
      SndOrderOutPD_Vector[1] +=  SectionDGD_Last*OutPD_Vector[2];
      SndOrderOutPD_Vector[2] += -SectionDGD_Last*OutPD_Vector[1];
      
      for (int jj = 0; jj < 3; jj++) {
         OutPD_VectorTmp[jj] = OutPD_Vector[jj];
         SndOrderOutPD_VectorTmp[jj] = SndOrderOutPD_Vector[jj];         
      }      

      TransRzRyV3to_Xhat(OutPD_VectorTmp,thetaZthetaY); //,Rotation);                   
      //## But I really need to rotate to -Yhat      
      RotatesAboutZ( -pi/2., OutPD_VectorTmp);

      //## Apply the same rotations to SndOrderOutPD_VectorTmp
      RotatesAboutY( thetaZthetaY[1], SndOrderOutPD_VectorTmp);    
      RotatesAboutZ( thetaZthetaY[0] - pi/2., SndOrderOutPD_VectorTmp);         

      //## Now I should eliminate the X component of SndOrderOutPD_Vector
      //#  while leaving the Z component positive: 
      TransRyV3to_YZ( SndOrderOutPD_VectorTmp, &thetaY);

      //## Apply the same rotations to OutPD_VectorTmp
      RotatesAboutY( thetaY,   OutPD_VectorTmp);
  
      psi[ii]   = -thetaZthetaY[1];       
      theta[ii] =  thetaZthetaY[0] - pi/2.;                                                
      phi[ii]   = -thetaY; // *0000000000000.;

      //## Generate the rotation angles      
      SourceRandomVar = RNG->GetRanCosineBiased(BiasingParam,
                                              &LikelihoodRatio); 
      
      //## Here comes the linear bias of the biased angle
      delta[ii] = pi/2. - BiasedAngle*ii/(qtStepsFiber-1.);  
      /////////delta[ii] = pi/2. - BiasedAngle; 

      chi[ii] =  acos(SourceRandomVar)*sgn(RNG->GetRanNum() - 0.5);  

      ksi[ii] =  2*pi*RNG->GetRanNum();                              
        
      /*            
      //## Here comes the linear bias of the biased angle
      delta[ii] = pi/2. + acos(SourceRandomVar)*sgn(RNG->GetRanNum() - 0.5);   
      //       - BiasedAngle*ii/(qtStepsFiber-1.);  
      /////////delta[ii] = pi/2. - BiasedAngle; 

      chi[ii] =  - BiasedAngle*ii/(qtStepsFiber-1.); 

      ksi[ii] =  2*pi*RNG->GetRanNum();  
           
      */
      GetRzRxRzRyRzRyMatrix(ii);                   
   }   
   return(LikelihoodRatio);
}



/*!

   This method generates the combined weights for all the samples 
   generated from multiple importance sampling distributions 
   with firsr- and second-order PMD bias using 
   the balanced heuristic method. All the samples can be later 
   combined in a single distribution using the balanced heuristic
   weight files that are generated by this method.
   Thus method was developed by E. Veach.

   The balanced heuristic weights are only computed for
   the samples that are generated using the method
   GetFiberSampleImportSamplingSndOrderPDV_VariableAngleExp().
       
   The input parameter InFileNameWithDirectories contains a list of 
   all the directories that contais the samples to be combined. 
   The directory names must separated by a new line character. 

   The input parameterInFileNameWithImpSampWeights contains the name
   of the file that contains the importance sampling weights. The 
   name of this file is unique. Therefore, all the files 
   whose samples are being weighted must have the same name
   in their respective directories.

   The input parameter OutFileNameWithHeuristicBalance contains the name
   of the files with the balanced heuristic weights that will be created 
   in each directory.  The name of this file is unique. 
   Therefore, all the directories will have the same file name.
    
   This is the method of choice to generate weights for the combination
   of multiple importance sampling distributions. Then, the samples can
   be combined with these weights using the method
   Histogram::MultipleHeuristicBalanceImport() of the Histogram class.
   
*/

//###############################################################
//## Get the balance heuristic for the samples obtained using
//## GetFiberSampleImportSamplingSndOrderPDV_VariableAngleExp(.)
//## It will read a file containing the directories where
//## the samples were were originated.   
//## By: Ivan Lima (7/16/02), adapted from a code by Aurenice Lima 
double OptFiber::GetHeuristicBalanceForSndOrderPDV_VariableAngleExp(
                            string InFileNameWithDirectories, 
			    string InFileNameWithImpSampWeights,
			    string OutFileNameWithHeuristicBalance) 
{
   if (! MemAllocated) {
      cerr << "Error: Object of class OptFiber not initialized." 
           << endl;
      exit(1);
   }   
   if (PMD_Fiber == 0) {
      cerr << "Error: The method GetHeuristicBalanceForSndOrderPDV_VariableAngleExp(...)"
           << "Is being called for a fiber with PMD == 0." 
           << endl;
      exit(1);
   }   

   #define MAX_FILES_HEURISTIC_BALANCE 100
   string DirectoryName[MAX_FILES_HEURISTIC_BALANCE];
   double vBiasedAngle[MAX_FILES_HEURISTIC_BALANCE];
   double vBiasingParam[MAX_FILES_HEURISTIC_BALANCE];
   double qtSamples[MAX_FILES_HEURISTIC_BALANCE];
   double vLikelihoodRatio[MAX_FILES_HEURISTIC_BALANCE];   

   ifstream InFileWithDirectories(InFileNameWithDirectories.c_str(),ios::in);
   if ( ! InFileWithDirectories) {
      cout << endl;
      cout << "Error: " << InFileNameWithDirectories << " could not be open."
           << endl;
      exit(1);
   }     
   
   int DirectoryCounter = 0;
   while (InFileWithDirectories >> DirectoryName[DirectoryCounter]) { 
      string BiasFileName = DirectoryName[DirectoryCounter]
                           + "/BiasSndOrderPDV_VariableAngleExp.dat";
      ifstream BiasFile(BiasFileName.c_str(),ios::in);      
      if ( ! BiasFile) {
         cout << endl;
         cout << "Warning: " << BiasFileName << " could not be open."
              << endl;
      }   
      else {
         BiasFile >> vBiasingParam[DirectoryCounter];       
         BiasFile >> vBiasedAngle[DirectoryCounter];     
         //cout << DirectoryName[DirectoryCounter] 
         //     << "\t"<< vBiasingParam[DirectoryCounter]
         //     << "\t"<< vBiasedAngle[DirectoryCounter]		 
       	 //     << endl;       
         DirectoryCounter++;
	 
      }	 
      BiasFile.close();
   } // end while    

   //cout << "Aborting now" << endl; exit(1);

   //## Just to create the object

   ifstream * InFileWithImpSampWeights;
   ifstream * FileQtOptFibersSkipped;
   ofstream * OutFileWithHeuristicBalance;

   InFileWithImpSampWeights = new ifstream; //("",ios::in); 
   OutFileWithHeuristicBalance = new ofstream; //("",ios::out); 
   FileQtOptFibersSkipped = new ifstream; //("",ios::in);

   //cout << endl<<"List directories with Balance Weights:" << endl; 
   cout << endl<<"DirectoryName      BiasingParam   BiasedAngle  qtSamples"
        << " qtSamplesSkippped" << endl;    

   //## Counting the number of samples per bias
   double qtSamplesTotal = 0;
   for  (int kk = 0; kk < DirectoryCounter; kk++) {   
      string PathInFileNameWithImpSampWeights = DirectoryName[kk];
      PathInFileNameWithImpSampWeights += "/";
      PathInFileNameWithImpSampWeights += InFileNameWithImpSampWeights;    

      InFileWithImpSampWeights->open(PathInFileNameWithImpSampWeights.c_str(),
                                     ios::in);   				     

      if (! InFileWithImpSampWeights ) {
         cout << endl;
         cout << "Error: " << PathInFileNameWithImpSampWeights 
              << " could not be opened JZ."
              << endl;
         exit(1);
      }	

      double UselessValue, LikelihoodRatioFile;
      qtSamples[kk] = 0;
   
      InFileWithImpSampWeights->clear(); // Resets ifstream state to good
      InFileWithImpSampWeights->seekg(0); // Resets the file pointer to the start of the file.

      while (*InFileWithImpSampWeights >> UselessValue ) {         
      *InFileWithImpSampWeights >> LikelihoodRatioFile; 
         qtSamples[kk]++;   
      }	 
      qtSamplesTotal += qtSamples[kk];
      
      string FileNameQtOptFibersSkipped = DirectoryName[kk]
                                         +"/qtOptFibersSkipped.dat";

       FileQtOptFibersSkipped->open(FileNameQtOptFibersSkipped.c_str(),ios::in);
   
      int qtOptFibersSkipped = 0;  
    
      if ( *FileQtOptFibersSkipped ) {
         *FileQtOptFibersSkipped >> qtOptFibersSkipped;
         FileQtOptFibersSkipped->close();
      }	      
      
      cout << DirectoryName[kk] 
           << "\t"<< vBiasingParam[kk]
           << "\t"<< vBiasedAngle[kk]		 
	   << "\t"<< qtSamples[kk]
	   << "\t"<< qtOptFibersSkipped
           << endl;  
      InFileWithImpSampWeights->close();      
   } // end for kk loop   

   
   //## Compute the weights using the heuristic balance
   for  (int kk = 0; kk < DirectoryCounter; kk++) {

   //## I should load the RNG with values from the given directory
   RNG->ResetRanNumGen();
     
   string FileNameQtOptFibersSkipped = DirectoryName[kk]
                                      +"/qtOptFibersSkipped.dat";

   FileQtOptFibersSkipped->open(FileNameQtOptFibersSkipped.c_str(),ios::in);   

   FileQtOptFibersSkipped->clear(); // Resets ifstream state to good
   FileQtOptFibersSkipped->seekg(0); // Resets the file pointer to the start of the file.

   if ( FileQtOptFibersSkipped ) {
      int qtOptFibersSkipped;
      *FileQtOptFibersSkipped >> qtOptFibersSkipped;
      for (int nn = 0; nn < qtOptFibersSkipped; nn++) {
         GetFiberSampleCompleteScatt();   
      }
      FileQtOptFibersSkipped->close();
   }	    

   ////#$#$#$### Synchronize the RNG   
   //RNG->GetRanNum();   RNG->GetRanNum(); RNG->GetRanNum();
   //GetFiberSampleCompleteScatt();
   
   double BiasedAngle = vBiasedAngle[kk];
   double BiasingParam = vBiasingParam[kk];

   string PathInFileNameWithImpSampWeights = DirectoryName[kk];
   PathInFileNameWithImpSampWeights += "/";
   PathInFileNameWithImpSampWeights += InFileNameWithImpSampWeights;    

   InFileWithImpSampWeights->open(PathInFileNameWithImpSampWeights.c_str(),
                                     ios::in);   

   if (! InFileWithImpSampWeights ) {
      cout << endl;
      cout << "Error: " << PathInFileNameWithImpSampWeights 
           << " could not be open."
           << endl;
      exit(1);
   }	


   string PathOutFileNameWithHeuristicBalance = DirectoryName[kk];
   PathOutFileNameWithHeuristicBalance += "/";
   PathOutFileNameWithHeuristicBalance += OutFileNameWithHeuristicBalance;  

   OutFileWithHeuristicBalance->open(PathOutFileNameWithHeuristicBalance.c_str(),
                                     ios::out);   


   if (! OutFileWithHeuristicBalance ) {
      cout << endl;
      cout << "Error: " << PathOutFileNameWithHeuristicBalance
           << " could not be open."
           << endl;
      exit(1);
   }	



   double HeuristicBalanceWeight = 1;

   double UselessValue, LikelihoodRatioFile;
   double qtSamplesCounter = 0;   
  
   InFileWithImpSampWeights->clear(); // Resets ifstream state to good
   InFileWithImpSampWeights->seekg(0); // Resets the file pointer to the start of the file.

   while (*InFileWithImpSampWeights >> UselessValue ) {         
   
   *InFileWithImpSampWeights >> LikelihoodRatioFile; 
   qtSamplesCounter++;
    
   double SourceRandomVar;
   
   double LikelihoodRatio = 1;
   for  (int mm = 0; mm < DirectoryCounter; mm++) {
      vLikelihoodRatio[mm] = 1.;
   }     


   TypeScattering = 5; //## RzRxRzRyRzRy 
   
   GotFiberSample = 1;   
   int LastStep;
   double v2Delta;
   double OutPD_Vector[] = {0,0,0};
   double OutPD_VectorTmp[3];   
   double SndOrderOutPD_Vector[] = {0,0,0};
   double SndOrderOutPD_VectorTmp[3];
     
   double thetaZthetaY[2];     
   double thetaY;

   //## Generate a random rotation for the first section.                      
   //#  It has no effect on the output PDV.
   cvtheta   = 2*(RNG->GetRanNum()-0.5);                        
   theta[0] = acos(cvtheta);  //## Ry                   
   phi[0]   = 2*pi*RNG->GetRanNum();  //## Rz            
   psi[0]   = 2*pi*RNG->GetRanNum();  //## Ry
   delta[0] = 0;  //## Rz  
   ksi[0]   = 0; //2*pi*RNG->GetRanNum();   //## Rx
   chi[0]   = 0; //## Rz

   ////GetRzRxRzRyRzRyMatrix(0);    
   
   double SectionDGD, SectionDGD_Last;   
   SectionDGD = 2*dBeta_2dOmega*StepLength[0];
   for (int ii = 1; ii < qtStepsFiber; ii++) {
      SectionDGD_Last = SectionDGD;
      SectionDGD = 2*dBeta_2dOmega*StepLength[ii];
 
      LastStep = ii-1;   
      
      // ## Calculates the cumulative Pol. Disp. Vector
      v2Delta = SectionDGD_Last*2*pi*CenterFreq;                    
      RotatesAboutY( -psi[LastStep],   OutPD_Vector);
      RotatesAboutZ(  theta[LastStep], OutPD_Vector);  
      RotatesAboutY( -phi[LastStep],   OutPD_Vector); 
      RotatesAboutZ(  delta[LastStep], OutPD_Vector); 
      RotatesAboutX(  ksi[LastStep],   OutPD_Vector); 
      RotatesAboutZ(  chi[LastStep], OutPD_Vector);
      RotatesAboutX(  v2Delta,         OutPD_Vector); 
      OutPD_Vector[0] += -SectionDGD_Last;  
      
      RotatesAboutY( -psi[LastStep],   SndOrderOutPD_Vector); 
      RotatesAboutZ(  theta[LastStep], SndOrderOutPD_Vector); 
      RotatesAboutY( -phi[LastStep],   SndOrderOutPD_Vector);                   
      RotatesAboutZ(  delta[LastStep], SndOrderOutPD_Vector); 
      RotatesAboutX(  ksi[LastStep],   SndOrderOutPD_Vector);
      RotatesAboutZ(  chi[LastStep], SndOrderOutPD_Vector); 
      RotatesAboutX(  v2Delta,         SndOrderOutPD_Vector);

      //## Add the cross product between the OutPD_Vectors
      SndOrderOutPD_Vector[1] +=  SectionDGD_Last*OutPD_Vector[2];
      SndOrderOutPD_Vector[2] += -SectionDGD_Last*OutPD_Vector[1];
      
      for (int jj = 0; jj < 3; jj++) {
         OutPD_VectorTmp[jj] = OutPD_Vector[jj];
         SndOrderOutPD_VectorTmp[jj] = SndOrderOutPD_Vector[jj];         
      }      

      TransRzRyV3to_Xhat(OutPD_VectorTmp,thetaZthetaY); //,Rotation);                   
      //## But I really need to rotate to -Yhat      
      RotatesAboutZ( -pi/2., OutPD_VectorTmp);

      //## Apply the same rotations to SndOrderOutPD_VectorTmp
      RotatesAboutY( thetaZthetaY[1], SndOrderOutPD_VectorTmp);    
      RotatesAboutZ( thetaZthetaY[0] - pi/2., SndOrderOutPD_VectorTmp);         

      //## Now I should eliminate the X component of SndOrderOutPD_Vector
      //#  while leaving the Z component positive: 
      TransRyV3to_YZ( SndOrderOutPD_VectorTmp, &thetaY);

      //## Apply the same rotations to OutPD_VectorTmp
      //## (It should not make any difference, since it is in -Yhat direction)
      RotatesAboutY( thetaY,   OutPD_VectorTmp);
  
      psi[ii]   = -thetaZthetaY[1];       
      theta[ii] =  thetaZthetaY[0] - pi/2.;                                                
      phi[ii]   = -thetaY; // *0000000000000.;

      //## Generate the rotation angles      
      SourceRandomVar = RNG->GetRanCosineBiased(BiasingParam,
                                              &LikelihoodRatio); 
 
      //## Here comes the linear bias of the biased angle
      delta[ii] = pi/2. - BiasedAngle*ii/(qtStepsFiber-1.);  
      /////////delta[ii] = pi/2. - BiasedAngle; 
            
      chi[ii] =  acos(SourceRandomVar)*sgn(RNG->GetRanNum() - 0.5);

      ksi[ii] =  2*pi*RNG->GetRanNum();                    
     
      for  (int mm = 0; mm < DirectoryCounter; mm++) {
         //## Now I will try to apply it to beta != 0
	 double bVector_kk[] = {-1,0,0};
         RotatesAboutZ(  -chi[ii],   bVector_kk);	 
         RotatesAboutX(  -ksi[ii],   bVector_kk);	 
         RotatesAboutZ( BiasedAngle*ii/(qtStepsFiber-1.),bVector_kk); 	 
    
	 double bVector_mm[] = {-1,0,0};
         RotatesAboutZ( vBiasedAngle[mm]*ii/(qtStepsFiber-1.),bVector_mm); 	 

	 /*
	 double bVector_kk[] = {0,-1,0};
         RotatesAboutZ(  delta[ii], bVector_kk); 
         //RotatesAboutX(  ksi[ii],   bVector_kk);
         RotatesAboutZ(  chi[ii],   bVector_kk);
	 double bVector_mm[] = {0,-1,0};
         RotatesAboutZ( pi/2.-vBiasedAngle[mm]*ii/(qtStepsFiber-1.),bVector_mm); 
         //RotatesAboutX(  ksi[ii],   bVector_mm);
         */
	 double cosine_biased = bVector_kk[0]*bVector_mm[0]
	                       +bVector_kk[1]*bVector_mm[1]
			       +bVector_kk[2]*bVector_mm[2];   
      
         //## For now, only first order PMD bias is accounted for correctly
         vLikelihoodRatio[mm] 
	       *= RNG->GetLikelihoodRatioForRanCosineBiased(vBiasingParam[mm],
	                cosine_biased );
                        //SourceRandomVar);
			
         //if (kk == 3 && ii == qtStepsFiber - 1 ) {			
         //   cout << "#$#$# -> " << kk << "  "<< mm << "  "
	 //        << cosine_biased <<"   "<< SourceRandomVar<<endl;
         //}		 
	 
         //cout << " * " << vLikelihoodRatio[mm]  
	 //     << "   " << vBiasingParam[mm] << endl;			
      }         
            
              
      ////GetRzRxRzRyRzRyMatrix(ii);                   
   }
   ////return(LikelihoodRatio);
       
   double HeuristicBalanceWeightDenominatorTmp = 0;
   for  (int mm = 0; mm < DirectoryCounter; mm++) {
      HeuristicBalanceWeightDenominatorTmp += qtSamples[mm]/
            vLikelihoodRatio[mm];	      
   }         
   HeuristicBalanceWeight = qtSamples[kk]/vLikelihoodRatio[kk]         
                            /HeuristicBalanceWeightDenominatorTmp; 
   double TrueCombinedWeight = vLikelihoodRatio[kk]*HeuristicBalanceWeight
                              *qtSamplesTotal/qtSamples[kk];
			    
   //cout << "***" << qtSamples[kk]
   //     << "  " << vLikelihoodRatio[kk] 
   //	<< "  " << HeuristicBalanceWeightDenominatorTmp << endl;

   /*
   //## Temporary lines for debugging
   cout << endl;  
   for  (int mm = 0; mm < DirectoryCounter; mm++) {    
      cout << mm << "  "<< qtSamples[mm]/vLikelihoodRatio[mm]         
                            /HeuristicBalanceWeightDenominatorTmp; 
      if (mm == kk) { cout << " *** ";}
      cout << endl;      
   } 
   */    

   //## First field:LikelihoodRatio, second field: HeuristicBalanceWeight
   //OutFileWithHeuristicBalance << vLikelihoodRatio[kk]   << "   "
   *OutFileWithHeuristicBalance << LikelihoodRatio   << "   "
                               << HeuristicBalanceWeight << "   "
			       << TrueCombinedWeight << endl;

   if ( abs( (LikelihoodRatio - LikelihoodRatioFile)/LikelihoodRatio)
        > 1e-4 
        && qtSamplesCounter < 10) {   
      cout <<    "LikelihoodRatio = " << LikelihoodRatio 
           << "  LikelihoodRatioFile = " << LikelihoodRatioFile
           << "  Diff = " << LikelihoodRatio - LikelihoodRatioFile 
	   << endl;	   
   }	
   if ( abs( (LikelihoodRatio - LikelihoodRatioFile)/LikelihoodRatio)
        > 1e-4 
        && qtSamplesCounter >= 10) {   
      cout <<    "LikelihoodRatio inconsistent" << endl
           << "Aborting now..." << endl;
      exit(1);	   
   }	   

   } //## End While (
   InFileWithImpSampWeights->close(); 
   OutFileWithHeuristicBalance->close();
   } //## End for (int kk
}

// #########################################################################

/*!

   This method generates a random fiber realization using several 
   methods to produce random mode coupling prior to each section.
   
   The input parameter TypeDistrib specifies the time of random 
   mode coupling:  
      1) Random rotation around the z-axis with a uniformly distributed
         angle of rotation;
      2) Random rotation that produces a Stokes vector that is
         univormly distributed on the Poincar\'e sphere;
      3) Importance sampling to bias first-order PMD;

   The input parameter BiasingParam specifies the amount of bias
   if importance sampling to bias first-order PMD was 
   selected (TypeDistrib == 3).
   
   This method does not have the option of importance sampling to
   bias both first- and second-order PMD. 
   Therefore, this method is OUT OF DATE.


*/

//###############################################################
//## Get a random fiber from the selected distribution
//#  It returns the likelihood ratio
//## By: Ivan Lima
double OptFiber::GetFiberSample(int TypeDistrib, double BiasingParam)
{
   double LikelihoodRatio;
   switch (TypeDistrib) {
      case 0: //## Uniform rotation of the birefringence axis
         LikelihoodRatio = 1.;
         GetFiberSampleUnifRz(); 
         break;   
      case 1: //## Coarse Step Method
         LikelihoodRatio = 1.; 
         GetFiberSampleCompleteScatt();           
         break;                 
      case 2: //## Importance sampling technique
         LikelihoodRatio 
            = GetFiberSampleImportSamplingExp(BiasingParam); 
         break;   
      case 3: //## Old Importance sampling technique. Do not use it
         LikelihoodRatio 
            = GetFiberSampleImportSamplingCosRx(BiasingParam);                                                      
         break;   
      case 4: //## Importance sampling applied to the 2nd order PDV
         LikelihoodRatio 
            = GetFiberSampleImportSamplingSndOrderPDV_Exp (BiasingParam,
                                                   pi/4.); 
         break;                                      
      /********
      case 3: //## Importance sampling on the average DGD
         LikelihoodRatio =  
            GetFiberSampleImportSamplingEffectDGD(BiasingParam,0);   
         break;
      case 4: //## Importance sampling on the penalty  
         LikelihoodRatio =  
            GetFiberSampleImportSamplingEffectDGD(BiasingParam,1);   
         break;
      *********/   
      default: 
         cout << endl << "TypeDistrib: " << TypeDistrib
                      << " inexistent." << endl << endl;                                          
         exit(1);                     
   }        
   return LikelihoodRatio;
}



/*!

   This method imposes a given random mode coupling prior to each section.
   The random mode coupling is defined by the three Euler angles
   of the scattering matrix.
   
   The input parameter IndexStep specifies the number of the section
   where the random mode coupling matrix will be modified. The parameters
   Phi, Theta, and Psi are the Euler angles.
   
   This method should not be called directly to generate random fiber 
   realizations. 

*/

//###############################################################
//## Forces the scattering angles for special user purposes
//## By: Ivan Lima
void OptFiber::SetScatteringMatrix( int IndexStep, double Theta 
                                   ,double Phi, double Psi)
{
   GotFiberSample = 1; 

   TypeScattering = 1;

   //int ii = IndexStep;
   theta[IndexStep] = Theta;
   phi[IndexStep]   = Phi;
   psi[IndexStep]   = Psi;
   GetEulerMatrix(IndexStep);   
   // ## the rest of this function should be the same 
   // #  as GetFiberSampleCompleteScatt() and equivalents
   //cvtheta = cos(theta[ii]);
   //cvtheta2  = sqrt((1.+cvtheta)/2.);      
   // ## Get the Uniform Scattering Matrix
   //TransM[0][0][ii] = exp( jc*(psi[ii]+phi[ii])/2)*cvtheta2;
   //TransM[0][1][ii] = jc*exp( jc*(psi[ii]-phi[ii])/2)*sqrt(1-sq(cvtheta2));
   //TransM[1][0][ii] = jc*exp(-jc*(psi[ii]-phi[ii])/2)*sqrt(1-sq(cvtheta2));
   //TransM[1][1][ii] = exp(-jc*(psi[ii]+phi[ii])/2)*cvtheta2;           
}


//###############################################################
//## Forces the scattering angles for special user purposes
//## By: Ivan Lima
void OptFiber::SetScatteringMatrixRxRzRy( int IndexStep, double Theta 
                                   ,double Phi, double Psi)
{
   GotFiberSample = 1; 
   
   TypeScattering = 3;

   //int ii = IndexStep;
   theta[IndexStep] = Theta; //## Rz
   phi[IndexStep]   = Phi;   //## Rx
   psi[IndexStep]   = Psi;   //## Ry
   GetRxRzRyMatrix(IndexStep);    
}


//###############################################################
//## Forces the scattering angles for special user purposes
//## By: Ivan Lima
void OptFiber::SetScatteringMatrixRzRxRzRyRzRy( int IndexStep, 
      double Chi, double Ksi,double Delta,double Phi,double Theta, double Psi)
{
   GotFiberSample = 1; 

   TypeScattering = 5;

   //int ii = IndexStep;
   chi[IndexStep] = Chi;     //## Rz
   ksi[IndexStep] = Ksi;     //## Rx
   delta[IndexStep] = Delta; //## Rz
   phi[IndexStep]   = Phi;   //## Ry   
   theta[IndexStep] = Theta; //## Rz
   psi[IndexStep]   = Psi;   //## Ry
   
   GetRzRxRzRyRzRyMatrix(IndexStep);    
}

/*!

   This method returns the differential-group delay (DGD) of the 
   current fiber at the angular frequency Freq. The DGD is equal to the 
   length of the polarization-dispersion vector.

*/

//###############################################################
//## Gets the DGD = length of the PD vector
//## By: Ivan Lima
double OptFiber::GetDGD(double Freq)
{
   double OutPD_Vector[] = {0,0,0};
   GetOutPD_Vector(Freq, OutPD_Vector);
   return(sqrt(sq(OutPD_Vector[0])+sq(OutPD_Vector[1])
                +sq(OutPD_Vector[2])) ); 
}



/*!

   This method returns the derivative of the differential-group
   delay (DGD) with respect to the angular frequency $\omega$. 
   The frequency derivative is evaluated at at the angular 
   frequency Freq 

*/

//########################################################
//#### Gets the derivative of the DGD with respect to $\omega$
//## By: Aurenice Lima
double OptFiber::Get_dDGD_domega(double Freq)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: From Get_dDGD_domega(.)" << endl
           << "       Object of class OptFiber not initialized"<< endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }   
   double OutPD_Vector[] = {0,0,0};
   double SndOrdOutPD_Vector[] = {0,0,0};
   GetOutPD_Vector(Freq,  OutPD_Vector);   
   GetSndOrderOutPD_Vector(Freq,  SndOrdOutPD_Vector);
   
   double dDGD_domega = (  OutPD_Vector[0]*SndOrdOutPD_Vector[0]
                          +OutPD_Vector[1]*SndOrdOutPD_Vector[1]
			  +OutPD_Vector[2]*SndOrdOutPD_Vector[2]
			 )/GetDGD(Freq);
   
   //## I should continue from here
   //cout << " Freq = " << Freq << "   Freq2 = " << Freq2 << endl;   
   //cout <<"OutPD_Vector @ Freq = " << OutPD_Vector[0] <<"  "
   //                      << OutPD_Vector[1] <<"  "
   //                     << OutPD_Vector[2]<< endl;                             
   //cout <<"OutPD_Vector @ Freq2 = " << OutPD_Vector2[0] <<"  "
   //                      << OutPD_Vector2[1] <<"  "
   //                     << OutPD_Vector2[2]<< endl;
                          
   //for (int ii = 0; ii < 3; ii++)
   //   SndOrderOutPD_Vector[ii] = (OutPD_Vector2[ii]
   //                              -OutPD_Vector[ii])
   //                              /(2*pi*(Freq2-Freq));
   
   return dDGD_domega;
}                                  

 
 
/*!

   This method returns a numerical derivative of the differential-group
   delay (DGD) with respect to the angular frequency $\omega$. 
   The frequency derivative is evaluated at at the angular 
   frequency Freq 

   This method was designed to test the analytical expression, which
   is the method of choice. Therefore, this method should be not used 
   in other applications.

*/

//########################################################
//#### Gets the derivative of the DGD with respect to $\omega$
//##   numerically.
//## By: Aurenice Lima
double OptFiber::GetNum_dDGD_domega(double Freq)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: From GetNum_dDGD_domega(.)" << endl
           << "       Object of class OptFiber not initialized"<< endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }   
   double Freq2 = Freq + 1e-4/GetExpectedDGD();

   return ( GetDGD(Freq2) - GetDGD(Freq) )/(2*pi*(Freq2-Freq));
}                                  



/*!

   This method returns the numerical second-order derivative of 
   the differential-group delay (DGD) with respect to the angular 
   frequency $\omega$. 
   The second-order frequency derivative is evaluated at at the angular 
   frequency Freq    

*/

//########################################################
//#### Gets the second derivative of the DGD with respect to $\omega$
//##   numerically.
//## By: Aurenice Lima
double OptFiber::GetNum_d2DGD_domega2(double Freq)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: From GetNum_dDGD_domega(.)" << endl
           << "       Object of class OptFiber not initialized"<< endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }   
   double Freq2 = Freq + 1e-4/GetExpectedDGD();

   return ( Get_dDGD_domega(Freq2) - Get_dDGD_domega(Freq) )
         /(2*pi*(Freq2-Freq));
}                                  



/*!

   This method returns the differential-group delay (DGD) of the 
   current fiber at the frequency Freq. The DGD is computed 
   using an analytical method described 
   in the JLT paper "Manakov-PMD equation ...", which is based 
   on the Jones matrices.

   Despite being an analytical method, the coputational time of 
   this method scales with the square of the number of fiber steps.
   Therefore, the method GetDGD(), which scales with the number
   of fiber steps, should be the method of choice.

*/  

//###############################################################
//## Gets the DGD = length of the PD vector
//## Freq should be the actual frequency -> CenterFreq + FreqSignal    
//#  This function is slow. 
//#    It is based on the JLT paper "Manakov-PMD equation ..."
//#    The processing time is proportional to sq(qtStepsFiber).
//#    All the others have proc. time proportional to qtStepsFiber, but
//#    weren't published yet.
//## It only works for equal lengths, for now.
//## By: Ivan Lima
double OptFiber::GetDGD_Manakov(double Freq)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized " 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }        
   double DGD;
   double vDelta = dBeta_2dOmega*DeltaZ*2*pi*Freq;     
   cplx Fps[2][2],Vm[2][2],Vm_tmp[2][2]; 
   cplx Pw00 = exp(jc*vDelta), Pw11 = exp(-jc*vDelta);
   int mm;
   int ii;
   Fps[0][0] = 1; // Should be Sig3*DeltaZ(qtStepsFiber-1) -> the last step.  
   Fps[0][1] = 0;   
   Fps[1][0] = 0;   
   Fps[1][1] = -1;      
   if (qtStepsFiber >1 ) {
      for( mm = qtStepsFiber - 2; mm >= 0; mm--) {
         Vm[0][0] = 1;   
         Vm[0][1] = 0;   
         Vm[1][0] = 0;   
         Vm[1][1] = 1;    
         for( ii = mm + 1; ii < qtStepsFiber; ii++) {
            Vm_tmp[0][0] = Pw00*(TransM[0][0][ii]*Vm[0][0]
                                +TransM[0][1][ii]*Vm[1][0] );
            Vm_tmp[0][1] = Pw00*(TransM[0][0][ii]*Vm[0][1]
                                +TransM[0][1][ii]*Vm[1][1] );
            Vm_tmp[1][0] = Pw11*(TransM[1][0][ii]*Vm[0][0]
                                +TransM[1][1][ii]*Vm[1][0] );
            Vm_tmp[1][1] = Pw11*(TransM[1][0][ii]*Vm[0][1]
                                +TransM[1][1][ii]*Vm[1][1] );              
            Vm[0][0] = Vm_tmp[0][0];   
            Vm[0][1] = Vm_tmp[0][1];   
            Vm[1][0] = Vm_tmp[1][0];   
            Vm[1][1] = Vm_tmp[1][1];                              
         }
         Fps[0][0] = Fps[0][0] + (Vm[0][0]*Vm[1][1]+Vm[0][1]*Vm[1][0]);
         Fps[1][1] = Fps[1][1] - (Vm[0][0]*Vm[1][1]+Vm[0][1]*Vm[1][0]);
         Fps[0][1] = Fps[0][1] - 2*Vm[0][0]*Vm[0][1];
         Fps[1][0] = Fps[1][0] + 2*Vm[1][1]*Vm[1][0];      
      }   
   }   
   DGD =  2*dBeta_2dOmega*DeltaZ
          *sqrt(-real( Fps[0][0]*Fps[1][1]-Fps[1][0]*Fps[0][1] ) ) ;        
   return(DGD);
}



/*!

   This method returns the input polarization-dispersion vector of the 
   current fiber at the frequency Freq in the input parameter InPD_Vector. 
   This method is cumulative. Therefore, the parameter InPD_Vector
   must be initialized before computing for the first fiber.

   The polarization-dispersion vector is defined by the Stokes
   vector at the receiver and and by the frequency derivative of the 
   Stokes vector.  
   
   The length of the polarization-dispersion vector is equal to the
   differential-group delay (DGD), and the direction of the 
   polarization-dispersion vector is parallel to the direction of
   the principal states of polarization (PSPs).

   The polarization-dispersion vector is commonly referred to as 
   the polarization mode dispersion (PMD) vector. 
   
*/ 

//###############################################################
//#### Gets the input Pol. Disp. Vector of the fiber
//## By: Ivan Lima
void OptFiber::GetInPD_Vector(double Freq, 
                                  double *InPD_Vector) 
// ## Freq should be the actual frequency -> CenterFreq + FreqSignal                               
// ## InPD_Vector should have dimension 3: from 0 to 2                            
// ## The initial value of InPD_Vector will be used as the 
//    value of the Input PDV.  "pdispv4.mws"
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }        
   //double Tm[3][3];
   double v2Delta;
   
   for( int ii = qtStepsFiber - 1; ii >= 0; ii--) {  
      v2Delta = 2*dBeta_2dOmega*StepLength[ii]*2*pi*Freq;   
      InPD_Vector[0] += -2*dBeta_2dOmega*StepLength[ii];    
      RotatesAboutX( -v2Delta,  InPD_Vector); //, Tm);   // ## Propagation                                  
      switch (TypeScattering) {      
         case 1: // ## == 1 (default): Euler Matrix: RxRyRx
            RotatesAboutX( -psi[ii],  InPD_Vector); //, Tm);
            RotatesAboutY(  theta[ii],InPD_Vector); //, Tm);                    
            RotatesAboutX( -phi[ii],  InPD_Vector); //, Tm);  
            break;      
         case 2:  // ## RyRzRy
            RotatesAboutY(   phi[ii],   InPD_Vector); //, Tm);       
            RotatesAboutZ( -theta[ii],  InPD_Vector); //, Tm);  
            RotatesAboutY(   psi[ii],   InPD_Vector); //, Tm);         
            break;
         case 3: // ## RxRzRy      
            RotatesAboutX( -phi[ii],    InPD_Vector); //, Tm);       
            RotatesAboutZ( -theta[ii],  InPD_Vector); //, Tm);  
            RotatesAboutY(  psi[ii],    InPD_Vector); //, Tm);         
            break;
         case 4: //## RxRzRyRzRy  
            RotatesAboutX( -ksi[ii],   InPD_Vector); //, Rotation);                      
            RotatesAboutZ( -delta[ii], InPD_Vector); //, Rotation);  
            RotatesAboutY( phi[ii],   InPD_Vector); //, Rotation);                    
            RotatesAboutZ( -theta[ii], InPD_Vector); //, Rotation);  
            RotatesAboutY( psi[ii],   InPD_Vector); //, Rotation);                             
            break;
         case 5: //## RzRxRzRyRzRy  
            RotatesAboutZ( -chi[ii], InPD_Vector); //, Rotation); 	 
            RotatesAboutX( -ksi[ii],   InPD_Vector); //, Rotation);                      
            RotatesAboutZ( -delta[ii], InPD_Vector); //, Rotation);  
            RotatesAboutY( phi[ii],   InPD_Vector); //, Rotation);                    
            RotatesAboutZ( -theta[ii], InPD_Vector); //, Rotation);  
            RotatesAboutY( psi[ii],   InPD_Vector); //, Rotation);                             
            break;
         default: //## Rotation matrix inexistent
            cout << "From OptFiber->GetInPD_Vector(.)" 
                 << endl << "TypeScattering = " << TypeScattering
                 << " inexistent." << endl;              
      }         
   }   
   // ## Test lines. It is just OK.
   //double OutPD_Vector[3];
   //for (int ii = 0; ii < 3; ii++) 
   //   OutPD_Vector[ii] = InPD_Vector[ii];
   //for (int ii = 0; ii < qtStepsFiber; ii++) {
   //   if (TypeScattering == 2) { // ## RzRy      
   //      RotatesAboutY( -psi[ii],   OutPD_Vector); //, Tm);                    
   //      RotatesAboutZ( theta[ii],  OutPD_Vector); //, Tm);  
   //      RotatesAboutY( -phi[ii],   OutPD_Vector); //, Tm);                     
   //   }
   //   else {  // ## == 1 (default): Euler Matrix
   //      RotatesAboutX( phi[ii],  OutPD_Vector); //, Tm);
   //      RotatesAboutY(-theta[ii],OutPD_Vector); //, Tm);
   //      RotatesAboutX( psi[ii],  OutPD_Vector); //, Tm);       
   //   }               
   //   RotatesAboutX( v2Delta,  OutPD_Vector); //, Tm);    
   //}
   //   cout << "Propagating the Input Pol.Dist.Vector forward:" << endl;
   //   cout << OutPD_Vector[0] << "   " 
   //        << OutPD_Vector[1] << "   "
   //        << OutPD_Vector[2] << endl;             
}



/*!

   This method returns the input polarization-dispersion vector of the 
   current fiber at the frequency Freq in the input parameter InPD_Vector.
   This method is cumulative. Therefore, the parameter InPD_Vector
   must be initialized before computing for the first fiber.

   
   This method is OUT OF DATE.
   
*/

//###############################################################
//#### Gets the input Pol. Disp. Vector of the fiber
//## This version should be removed later. 
//#  It only works for the Euler Matrix.
//## Freq should be the actual frequency -> CenterFreq + FreqSignal                               
//#  InPD_Vector should have dimension 3: from 0 to 2                            
//#  The initial value of InPD_Vector will be used as the 
//#    value of the Input PDV.  "pdispv4.mws"
void OptFiber::GetInPD_VectorOld(double Freq, 
                                  double *InPD_Vector) 
//## By: Ivan Lima                                  
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }        
   double InPD_Vector_tmp[3];
   double vTheta, vPhi, vPsi, Tm[3][3];
   //double v2Delta = -2*dBeta_2dOmega*DeltaZ*2*pi*Freq; // -
   double v2Delta;
   
   for( int ii = qtStepsFiber - 1; ii >= 0; ii--) {
      v2Delta = -2*dBeta_2dOmega*StepLength[ii]*2*pi*Freq; // -
      
      InPD_Vector_tmp[0] = InPD_Vector[0];
      InPD_Vector_tmp[1] = InPD_Vector[1];
      InPD_Vector_tmp[2] = InPD_Vector[2];
      vTheta =  theta[ii];         
      vPhi   = -phi[ii];   // -
      vPsi   = -psi[ii];   // -           
      Tm[0][0] = cos(vTheta); 
      Tm[0][1] =-sin(vTheta)*( cos(v2Delta)*sin(vPsi)
                              +sin(v2Delta)*cos(vPsi) );       
      Tm[0][2] = sin(vTheta)*(-sin(v2Delta)*sin(vPsi)
                              +cos(v2Delta)*cos(vPsi) );                      
      Tm[1][0] =-sin(vTheta)*sin(vPhi);
      Tm[1][1] =-cos(vTheta)*sin(vPhi)*( sin(v2Delta)*cos(vPsi)
                                        +cos(v2Delta)*sin(vPsi) )
                +cos(vPhi)*(-sin(v2Delta)*sin(vPsi)
                            +cos(v2Delta)*cos(vPsi) );                         ;       
      Tm[1][2] = cos(vTheta)*sin(vPhi)*( cos(v2Delta)*cos(vPsi)
                                        -sin(v2Delta)*sin(vPsi) )
                +cos(vPhi)*( cos(v2Delta)*sin(vPsi)
                            +sin(v2Delta)*cos(vPsi) );                               
      Tm[2][0] =-sin(vTheta)*cos(vPhi);
       
      Tm[2][1] =-cos(vTheta)*cos(vPhi)*( sin(v2Delta)*cos(vPsi)
                                        +cos(v2Delta)*sin(vPsi) ) 
                +sin(vPhi)*( sin(v2Delta)*sin(vPsi)                        
                            -cos(v2Delta)*cos(vPsi) );
      Tm[2][2] = cos(vTheta)*cos(vPhi)*( cos(v2Delta)*cos(vPsi)
                                        -sin(v2Delta)*sin(vPsi) )
                -sin(vPhi)*( cos(v2Delta)*sin(vPsi)
                            +sin(v2Delta)*cos(vPsi) );    
    
      InPD_Vector[0] =-v2Delta/(2*pi*Freq)*cos(vTheta) // v2Delta < 0
                      +Tm[0][0]*InPD_Vector_tmp[0]
                      +Tm[0][1]*InPD_Vector_tmp[1]
                      +Tm[0][2]*InPD_Vector_tmp[2];
      InPD_Vector[1] = v2Delta/(2*pi*Freq)*sin(vTheta)*sin(vPhi) 
                      +Tm[1][0]*InPD_Vector_tmp[0]
                      +Tm[1][1]*InPD_Vector_tmp[1]
                      +Tm[1][2]*InPD_Vector_tmp[2];
      InPD_Vector[2] = v2Delta/(2*pi*Freq)*sin(vTheta)*cos(vPhi)  
                      +Tm[2][0]*InPD_Vector_tmp[0]
                      +Tm[2][1]*InPD_Vector_tmp[1]
                      +Tm[2][2]*InPD_Vector_tmp[2];
   }                 
}







//##############################################################
//#### Rotate any vector quantity (Stokes parameters, PDV, etc)
//##   Through a fiber.
//## By: Ivan Lima
void OptFiber::RotateThroughFiber(double Freq, double *OutVector)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }        
   //double Tm[3][3];
   double v2Delta;   
   for( int ii = 0; ii < qtStepsFiber; ii++) {
      v2Delta = 2*dBeta_2dOmega*StepLength[ii]*2*pi*Freq;                  
      switch (TypeScattering) {
         case 1: // ## == 1 (default): Euler Matrix
            RotatesAboutX( phi[ii],  OutVector); //, Tm);
            RotatesAboutY(-theta[ii],OutVector); //, Tm);
            RotatesAboutX( psi[ii],  OutVector); //, Tm);
            break;
         case 2: // ## RyRzRy      
            RotatesAboutY( -psi[ii],   OutVector); //, Tm);  
            RotatesAboutZ(  theta[ii], OutVector); //, Tm);  
            RotatesAboutY( -phi[ii],   OutVector); //, Tm);  
            break;
         case 3: // ## RxRzRy      
            RotatesAboutY( -psi[ii],   OutVector); //, Tm);  
            RotatesAboutZ(  theta[ii], OutVector); //, Tm);  
            RotatesAboutX(  phi[ii],   OutVector); //, Tm);  
            break;
         case 4: //## RxRzRyRzRy  
            RotatesAboutY( -psi[ii],   OutVector); //, Rotation);  
            RotatesAboutZ(  theta[ii], OutVector); //, Rotation);  
            RotatesAboutY( -phi[ii],   OutVector); //, Rotation);  
            RotatesAboutZ(  delta[ii], OutVector); //, Rotation);  
            RotatesAboutX(  ksi[ii],   OutVector); //, Rotation);  
            break;
         case 5: //## RzRxRzRyRzRy  
            RotatesAboutY( -psi[ii],   OutVector); //, Rotation);  
            RotatesAboutZ(  theta[ii], OutVector); //, Rotation);  
            RotatesAboutY( -phi[ii],   OutVector); //, Rotation);  
            RotatesAboutZ(  delta[ii], OutVector); //, Rotation);  
            RotatesAboutX(  ksi[ii],   OutVector); //, Rotation);  
            RotatesAboutZ(  chi[ii], OutVector); //, Rotation);  
            break;
         default: //## Rotation matrix inexistent
            cout << "From OptFiber->GetOrderPD_Vector(.)" 
                 << endl << "TypeScattering = " << TypeScattering
                 << " inexistent." << endl;              
      }         
      RotatesAboutX( v2Delta,  OutVector); //, Tm);   // ## Propagation    
      /////OutPD_Vector[0] += -2*dBeta_2dOmega*StepLength[ii];           
      //cout << ii<<" "<< dBeta_2dOmega <<"  " << StepLength[ii] << endl;
   }                 
}



/*!

   This method returns the output polarization-dispersion vector of the 
   current fiber at the frequency Freq in the input parameter OutPD_Vector. 
   This method is cumulative. Therefore, the parameter OutPD_Vector
   must be initialized before computing for the first fiber.

   The polarization-dispersion vector is defined by the Stokes
   vector at the receiver and and by the frequency derivative of the 
   Stokes vector.  
   
   The length of the polarization-dispersion vector is equal to the
   differential-group delay (DGD), and the direction of the 
   polarization-dispersion vector is parallel to the direction of
   the principal states of polarization (PSPs).

   The polarization-dispersion vector is commonly referred to as 
   the polarization mode dispersion (PMD) vector. 
   
*/

//##############################################################
//#### Gets the output Pol. Disp. Vector of the fiber
//## Freq should be the actual frequency -> CenterFreq + FreqSignal                               
//#  OutPD_Vector should have dimension 3: from 0 to 2                            
//#  The initial value of OutPD_Vector will be used as the 
//#  initial value of the Output PDV. "pdispv4.mws"
//## By: Ivan Lima
void OptFiber::GetOutPD_Vector(double Freq, double *OutPD_Vector)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }        
   //double Tm[3][3];
   double v2Delta;   
   for( int ii = 0; ii < qtStepsFiber; ii++) {
      v2Delta = 2*dBeta_2dOmega*StepLength[ii]*2*pi*Freq;                  
      switch (TypeScattering) {
         case 1: // ## == 1 (default): Euler Matrix
            RotatesAboutX( phi[ii],  OutPD_Vector); //, Tm);
            RotatesAboutY(-theta[ii],OutPD_Vector); //, Tm);
            RotatesAboutX( psi[ii],  OutPD_Vector); //, Tm);
            break;
         case 2: // ## RyRzRy      
            RotatesAboutY( -psi[ii],   OutPD_Vector); //, Tm);  
            RotatesAboutZ(  theta[ii], OutPD_Vector); //, Tm);  
            RotatesAboutY( -phi[ii],   OutPD_Vector); //, Tm);  
            break;
         case 3: // ## RxRzRy      
            RotatesAboutY( -psi[ii],   OutPD_Vector); //, Tm);  
            RotatesAboutZ(  theta[ii], OutPD_Vector); //, Tm);  
            RotatesAboutX(  phi[ii],   OutPD_Vector); //, Tm);  
            break;
         case 4: //## RxRzRyRzRy  
            RotatesAboutY( -psi[ii],   OutPD_Vector); //, Rotation);  
            RotatesAboutZ(  theta[ii], OutPD_Vector); //, Rotation);  
            RotatesAboutY( -phi[ii],   OutPD_Vector); //, Rotation);  
            RotatesAboutZ(  delta[ii], OutPD_Vector); //, Rotation);  
            RotatesAboutX(  ksi[ii],   OutPD_Vector); //, Rotation);  
            break;
         case 5: //## RzRxRzRyRzRy  
            RotatesAboutY( -psi[ii],   OutPD_Vector); //, Rotation);  
            RotatesAboutZ(  theta[ii], OutPD_Vector); //, Rotation);  
            RotatesAboutY( -phi[ii],   OutPD_Vector); //, Rotation);  
            RotatesAboutZ(  delta[ii], OutPD_Vector); //, Rotation);  
            RotatesAboutX(  ksi[ii],   OutPD_Vector); //, Rotation);  
            RotatesAboutZ(  chi[ii], OutPD_Vector); //, Rotation);  
            break;
         default: //## Rotation matrix inexistent
            cout << "From OptFiber->GetOrderPD_Vector(.)" 
                 << endl << "TypeScattering = " << TypeScattering
                 << " inexistent." << endl;              
      }         
      RotatesAboutX( v2Delta,  OutPD_Vector); //, Tm);   // ## Propagation    
      OutPD_Vector[0] += -2*dBeta_2dOmega*StepLength[ii];           
      //cout << ii<<" "<< dBeta_2dOmega <<"  " << StepLength[ii] << endl;
   }                 
}



/*!

   This method returns the output polarization-dispersion vector of the 
   current fiber at the frequency Freq in the parameter OutPD_Vector.
   This method is cumulative. Therefore, the parameter OutPD_Vector
   must be initialized before computing for the first fiber.
      
   This method is OUT OF DATE.
   
*/

//############################################################
//#### Gets the output Pol. Disp. Vector of the fiber
//## This version should be removed later. 
//#  It only works for the Euler Matrix
//## Freq should be the actual frequency -> CenterFreq + FreqSignal                               
//#  OutPD_Vector should have dimension 3: from 0 to 2                            
//#  The initial value of OutPD_Vector will be used as the 
//#   initial value of the Output PDV. "pdispv4.mws"
//## By: Ivan Lima
void OptFiber::GetOutPD_VectorOld(double Freq, double *OutPD_Vector)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }        
   double OutPD_Vector_tmp[3];
   double vTheta, vPhi, vPsi, Tm[3][3];
   //double v2Delta = -2*dBeta_2dOmega*DeltaZ*2*pi*Freq;  // -
   double v2Delta;
   
   for( int ii = 0; ii < qtStepsFiber; ii++) { // The first was included above
      v2Delta = -2*dBeta_2dOmega*StepLength[ii]*2*pi*Freq;  // -   
      OutPD_Vector_tmp[0] = OutPD_Vector[0];
      OutPD_Vector_tmp[1] = OutPD_Vector[1];
      OutPD_Vector_tmp[2] = OutPD_Vector[2];
      vTheta = theta[ii];
      vPhi   = -phi[ii];  // -
      vPsi   = -psi[ii];  // -             
      Tm[0][0] = cos(vTheta); 
      Tm[0][1] =-sin(vTheta)*sin(vPhi); 
      Tm[0][2] =-sin(vTheta)*cos(vPhi);
      Tm[1][0] =-cos(v2Delta)*sin(vPsi)*sin(vTheta)
                -sin(v2Delta)*cos(vPsi)*sin(vTheta); 
      Tm[1][1] = cos(v2Delta)*
               (-sin(vPsi)*cos(vTheta)*sin(vPhi)+cos(vPsi)*cos(vPhi)) 
                -sin(v2Delta)*
               (sin(vPsi)*cos(vPhi)+cos(vPsi)*cos(vTheta)*sin(vPhi)); 
      Tm[1][2] = cos(v2Delta)*
               (-cos(vPsi)*sin(vPhi)-sin(vPsi)*cos(vTheta)*cos(vPhi)) 
                -sin(v2Delta)*
               (cos(vPsi)*cos(vTheta)*cos(vPhi)-sin(vPsi)*sin(vPhi));
      Tm[2][0] =-sin(v2Delta)*sin(vPsi)*sin(vTheta)
                +cos(v2Delta)*cos(vPsi)*sin(vTheta); 
      Tm[2][1] = sin(v2Delta)*
               (-sin(vPsi)*cos(vTheta)*sin(vPhi)+cos(vPsi)*cos(vPhi)) 
                +cos(v2Delta)*
                (sin(vPsi)*cos(vPhi)+cos(vPsi)*cos(vTheta)*sin(vPhi)); 
      Tm[2][2] = sin(v2Delta)*
               (-cos(vPsi)*sin(vPhi)-sin(vPsi)*cos(vTheta)*cos(vPhi)) 
                +cos(v2Delta)*
                (cos(vPsi)*cos(vTheta)*cos(vPhi)-sin(vPsi)*sin(vPhi));   
      OutPD_Vector[0] = -2*dBeta_2dOmega*StepLength[ii]         // -
                    +Tm[0][0]*OutPD_Vector_tmp[0]
                    +Tm[0][1]*OutPD_Vector_tmp[1]
                    +Tm[0][2]*OutPD_Vector_tmp[2]; 
      OutPD_Vector[1] = Tm[1][0]*OutPD_Vector_tmp[0]
                    +Tm[1][1]*OutPD_Vector_tmp[1]
                    +Tm[1][2]*OutPD_Vector_tmp[2];
      OutPD_Vector[2] = Tm[2][0]*OutPD_Vector_tmp[0]
                    +Tm[2][1]*OutPD_Vector_tmp[1]
                    +Tm[2][2]*OutPD_Vector_tmp[2];       
   }                 
}



/*!

   This method returns the output second-order polarization-dispersion 
   vector of the current fiber at the frequency Freq
   in the input parameter SndOrderOutPD_Vector. 
   This method is not cumulative. Therefore, it does not require
   initialization.
   
   The second-order polarization-dispersion vector is the
   derivative of the polarization-dispersion vector with respect
   to the angular frequency. 
   
*/

//##############################################################
//#### Get the second order pol. disp. vector of the fiber
//## Freq should be the actual frequency -> CenterFreq + FreqSignal
//## The incoming vector will be initialized to zero.
//## By: Ivan Lima
void OptFiber::GetSndOrderOutPD_Vector(double Freq
                                  ,double *SndOrderOutPD_Vector)
{
   //## The SndOrderOutPD_Vector has to be initialized
   for (int ii = 0; ii < 3; ii++) {
      SndOrderOutPD_Vector[ii] = 0;
   }
   
   //## It uses the local PD_Vector in the computation
   double OutPD_Vector[] = {0,0,0};
   
   GetSndOrderOutPD_Vector( Freq ,SndOrderOutPD_Vector
                                 ,OutPD_Vector);
}



/*!

   This method returns both the output polarization-dispersion vector
   and the output second-order polarization-dispersion 
   vector of the current fiber at the frequency Freq in the input
   paremeters OutPD_Vector and SndOrderOutPD_Vector. This method is 
   cumulative. Therefore, the input parameters must be initialized
   before computing for the first fiber.

   The second-order polarization-dispersion vector is the
   derivative of the polarization-dispersion vector with respect
   to the angular frequency.    
   
*/

//##############################################################
//#### Get the second order pol. disp. vector of the fiber
//## Freq should be the actual frequency -> CenterFreq + FreqSignal                        
//## Warning: this module allows incremental computation 
//## by using the previous values that were stored in the 
//## vector. The user has to initialize them in the firs run
//## By: Ivan Lima
void OptFiber::GetSndOrderOutPD_Vector(double Freq
                                  ,double *SndOrderOutPD_Vector
                                  ,double *OutPD_Vector)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }        
   

   double SectionDGD;
   
   //double Tm[3][3];
   double v2Delta;   
   for( int ii = 0; ii < qtStepsFiber; ii++) {
      SectionDGD = 2*dBeta_2dOmega*StepLength[ii];
      //v2Delta = 2*dBeta_2dOmega*StepLength[ii]*2*pi*Freq;
      v2Delta = SectionDGD*2*pi*Freq;  

      //## First, obtain the local OutPD_Vector
      switch (TypeScattering) {
         case 1: // ## == 1 (default): Euler Matrix
            RotatesAboutX( phi[ii],  OutPD_Vector); //, Tm);
            RotatesAboutY(-theta[ii],OutPD_Vector); //, Tm);
            RotatesAboutX( psi[ii],  OutPD_Vector); //, Tm);       
            break;
         case 2: // ## RyRzRy      
            RotatesAboutY( -psi[ii],   OutPD_Vector); //, Tm);                    
            RotatesAboutZ(  theta[ii], OutPD_Vector); //, Tm);  
            RotatesAboutY( -phi[ii],   OutPD_Vector); //, Tm);                     
            break;
         case 3: // ## RxRzRy      
            RotatesAboutY( -psi[ii],   OutPD_Vector); //, Tm);                    
            RotatesAboutZ(  theta[ii], OutPD_Vector); //, Tm);  
            RotatesAboutX(  phi[ii],   OutPD_Vector); //, Tm);                     
            break;
         case 4: //## RxRzRyRzRy  
            RotatesAboutY( -psi[ii],   OutPD_Vector); //, Rotation);  
            RotatesAboutZ(  theta[ii], OutPD_Vector); //, Rotation);  
            RotatesAboutY( -phi[ii],   OutPD_Vector); //, Rotation);  
            RotatesAboutZ(  delta[ii], OutPD_Vector); //, Rotation);  
            RotatesAboutX(  ksi[ii],   OutPD_Vector); //, Rotation);  
            break;
         case 5: //## RzRxRzRyRzRy  
            RotatesAboutY( -psi[ii],   OutPD_Vector); //, Rotation);  
            RotatesAboutZ(  theta[ii], OutPD_Vector); //, Rotation);  
            RotatesAboutY( -phi[ii],   OutPD_Vector); //, Rotation);  
            RotatesAboutZ(  delta[ii], OutPD_Vector); //, Rotation);  
            RotatesAboutX(  ksi[ii],   OutPD_Vector); //, Rotation);  
            RotatesAboutZ(  chi[ii], OutPD_Vector); //, Rotation);  
            break;
         default: //## Rotation matrix inexistent
            cout << "From OptFiber->GetSndOrderPD_Vector(.)" 
                 << endl << "TypeScattering = " << TypeScattering
                 << " inexistent." << endl;              
      }         
      RotatesAboutX( v2Delta,  OutPD_Vector); //, Tm);   // ## Propagation    
      OutPD_Vector[0] += -SectionDGD;     

      //## Start rotating the last SndOrderOutPD_Vector
      switch (TypeScattering) {
         case 1: // ## == 1 (default): Euler Matrix
            RotatesAboutX( phi[ii],  SndOrderOutPD_Vector); //, Tm);
            RotatesAboutY(-theta[ii],SndOrderOutPD_Vector); //, Tm);
            RotatesAboutX( psi[ii],  SndOrderOutPD_Vector); //, Tm);       
            break;
         case 2: // ## RyRzRy      
            RotatesAboutY( -psi[ii],   SndOrderOutPD_Vector); //, Tm);                    
            RotatesAboutZ(  theta[ii], SndOrderOutPD_Vector); //, Tm);  
            RotatesAboutY( -phi[ii],   SndOrderOutPD_Vector); //, Tm);                     
            break;
         case 3: // ## RxRzRy      
            RotatesAboutY( -psi[ii],   SndOrderOutPD_Vector); //, Tm);                    
            RotatesAboutZ(  theta[ii], SndOrderOutPD_Vector); //, Tm);  
            RotatesAboutX(  phi[ii],   SndOrderOutPD_Vector); //, Tm);                     
            break;
         case 4: //## RxRzRyRzRy  
            RotatesAboutY( -psi[ii],   SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  theta[ii], SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutY( -phi[ii],   SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  delta[ii], SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutX(  ksi[ii],   SndOrderOutPD_Vector); //, Rotation);  
            break;            
         case 5: //## RzRxRzRyRzRy  
            RotatesAboutY( -psi[ii],   SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  theta[ii], SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutY( -phi[ii],   SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  delta[ii], SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutX(  ksi[ii],   SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  chi[ii], SndOrderOutPD_Vector); //, Rotation);
	    //cout << "**** chi["<< ii<<"] = " << chi[ii] << endl;
            break; 
         default: //## Rotation matrix inexistent
            cout << "From OptFiber->GetSndOrderPD_Vector(.)" 
                 << endl << "TypeScattering = " << TypeScattering
                 << " inexistent." << endl;   
      }         
      RotatesAboutX( v2Delta,  SndOrderOutPD_Vector); //, Tm);  

      //## Add the cross product between the OutPD_Vectors
      SndOrderOutPD_Vector[1] +=  SectionDGD*OutPD_Vector[2];
      SndOrderOutPD_Vector[2] += -SectionDGD*OutPD_Vector[1];


      //## At this stage the ii-th SndOrderOutPD_Vector 
      //## and OutPD_Vector were computed
      
   }                 
}



/*!

   This method returns the output third-order polarization-dispersion 
   vector of the current fiber at the frequency Freq
   in the input parameter ThirdOrderOutPD_Vector. 
   This method is not cumulative. Therefore, it does not require
   initialization of the input parameter.
   
   The third-order polarization-dispersion vector is the
   second derivative of the polarization-dispersion vector
   with respect to the angular frequency. 
   
*/

//########################################################
//#### Gets the Third order output Pol. Disp. Vector of the fiber
//## Freq should be the actual frequency -> CenterFreq + FreqSignal
//## The vector is initialized
//## By: Ivan Lima
void OptFiber::GetThirdOrderOutPD_Vector(double Freq
                                  ,double *ThirdOrderOutPD_Vector)
{
   //## The SndOrderOutPD_Vector has to be initialized
   for (int ii = 0; ii < 3; ii++) {
      ThirdOrderOutPD_Vector[ii] = 0;
   }
   
   //## It uses the local PD_Vector in the computation
   double OutPD_Vector[] = {0,0,0};
   double SndOrderOutPD_Vector[] = {0,0,0};
   GetThirdOrderOutPD_Vector(Freq,ThirdOrderOutPD_Vector
                                 ,SndOrderOutPD_Vector
                                 ,OutPD_Vector);
}



/*!

   This method returns the output polarization-dispersion vector,
   the output second-order polarization-dispersion vector, and 
   the third-order polarization-dispersion vector of the current fiber 
   at the frequency Freq in the input
   paremeters OutPD_Vector, SndOrderOutPD_Vector, and ThirdOrderOutPD_Vector. 
   This method is cumulative. Therefore, the input parameters must be 
   initialized before computing for the first fiber.

   The third-order polarization-dispersion vector is the
   second derivative of the polarization-dispersion vector
   with respect to the angular frequency.    
   
*/

//##############################################################
//#### Get the third order pol. disp. vector of the fiber
//## Freq should be the actual frequency -> CenterFreq + FreqSignal
//## Warning: this module allows incremental computation 
//## by using the previous values that were stored in the 
//## vector. The user has to initialize them in the firs run
//## By: Ivan Lima
void OptFiber::GetThirdOrderOutPD_Vector(double Freq
                                  ,double *ThirdOrderOutPD_Vector
                                  ,double *SndOrderOutPD_Vector
                                  ,double *OutPD_Vector)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }        
   

   double SectionDGD;
   
   //double Tm[3][3];
   double v2Delta;   
   for( int ii = 0; ii < qtStepsFiber; ii++) {
      SectionDGD = 2*dBeta_2dOmega*StepLength[ii];
      //v2Delta = 2*dBeta_2dOmega*StepLength[ii]*2*pi*Freq;
      v2Delta = SectionDGD*2*pi*Freq;  

      //## First, obtain the local OutPD_Vector
      switch (TypeScattering) {
         case 1: // ## == 1 (default): Euler Matrix
            RotatesAboutX( phi[ii],  OutPD_Vector); //, Tm);
            RotatesAboutY(-theta[ii],OutPD_Vector); //, Tm);
            RotatesAboutX( psi[ii],  OutPD_Vector); //, Tm);
            break;
         case 2: // ## RyRzRy      
            RotatesAboutY( -psi[ii],   OutPD_Vector); //, Tm);
            RotatesAboutZ(  theta[ii], OutPD_Vector); //, Tm);
            RotatesAboutY( -phi[ii],   OutPD_Vector); //, Tm);
            break;
         case 3: // ## RxRzRy      
            RotatesAboutY( -psi[ii],   OutPD_Vector); //, Tm);
            RotatesAboutZ(  theta[ii], OutPD_Vector); //, Tm);
            RotatesAboutX(  phi[ii],   OutPD_Vector); //, Tm);
            break;
         case 4: //## RxRzRyRzRy  
            RotatesAboutY( -psi[ii],   OutPD_Vector); //, Rotation);  
            RotatesAboutZ(  theta[ii], OutPD_Vector); //, Rotation);  
            RotatesAboutY( -phi[ii],   OutPD_Vector); //, Rotation);  
            RotatesAboutZ(  delta[ii], OutPD_Vector); //, Rotation);  
            RotatesAboutX(  ksi[ii],   OutPD_Vector); //, Rotation);              
            break;
         case 5: //## RzRxRzRyRzRy  
            RotatesAboutY( -psi[ii],   OutPD_Vector); //, Rotation);  
            RotatesAboutZ(  theta[ii], OutPD_Vector); //, Rotation);  
            RotatesAboutY( -phi[ii],   OutPD_Vector); //, Rotation);  
            RotatesAboutZ(  delta[ii], OutPD_Vector); //, Rotation);  
            RotatesAboutX(  ksi[ii],   OutPD_Vector); //, Rotation);              
            RotatesAboutZ(  chi[ii], OutPD_Vector); //, Rotation);  
            break;
         default: //## Rotation matrix inexistent
            cout << "From OptFiber->GetThirdOrderPD_Vector(.)" 
                 << endl << "TypeScattering = " << TypeScattering
                 << " inexistent." << endl;              
      }         
      RotatesAboutX( v2Delta,  OutPD_Vector); //, Tm);   // ## Propagation    
   

      //## At this stage the local OutPD_Vector is available

      //## Start rotating the last SndOrderOutPD_Vector
      switch (TypeScattering) {
         case 1: // ## == 1 (default): Euler Matrix
            RotatesAboutX( phi[ii],  SndOrderOutPD_Vector); //, Tm);
            RotatesAboutY(-theta[ii],SndOrderOutPD_Vector); //, Tm);
            RotatesAboutX( psi[ii],  SndOrderOutPD_Vector); //, Tm);       
            break;      
         case 2: // ## RyRzRy      
            RotatesAboutY( -psi[ii],   SndOrderOutPD_Vector); //, Tm);                    
            RotatesAboutZ(  theta[ii], SndOrderOutPD_Vector); //, Tm);  
            RotatesAboutY( -phi[ii],   SndOrderOutPD_Vector); //, Tm);                     
            break;
         case 3: // ## RxRzRy      
            RotatesAboutY( -psi[ii],   SndOrderOutPD_Vector); //, Tm);                    
            RotatesAboutZ(  theta[ii], SndOrderOutPD_Vector); //, Tm);  
            RotatesAboutX(  phi[ii],   SndOrderOutPD_Vector); //, Tm);                     
            break;
         case 4: //## RxRzRyRzRy  
            RotatesAboutY( -psi[ii],   SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  theta[ii], SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutY( -phi[ii],   SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  delta[ii], SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutX(  ksi[ii],   SndOrderOutPD_Vector); //, Rotation);  
            break;                        
         case 5: //## RzRxRzRyRzRy  
            RotatesAboutY( -psi[ii],   SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  theta[ii], SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutY( -phi[ii],   SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  delta[ii], SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutX(  ksi[ii],   SndOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  chi[ii], SndOrderOutPD_Vector); //, Rotation);  
            break;   
         default: //## Rotation matrix inexistent
            cout << "From OptFiber->GetThirdOrderPD_Vector(.)" 
                 << endl << "TypeScattering = " << TypeScattering
                 << " inexistent." << endl;              
      }         
      RotatesAboutX( v2Delta,  SndOrderOutPD_Vector); //, Tm);  


      //## At this stage the ii-th SndOrderOutPD_Vector 
      //## and OutPD_Vector were rotated by R_{ii}
      //## It is time to wrap up the third order PDV
      
      //## Start rotating the last ThirdOrderOutPD_Vector
      switch (TypeScattering) {
         case 1: //## == 1 (default): Euler Matrix
            RotatesAboutX( phi[ii],  ThirdOrderOutPD_Vector); //, Tm);
            RotatesAboutY(-theta[ii],ThirdOrderOutPD_Vector); //, Tm);
            RotatesAboutX( psi[ii],  ThirdOrderOutPD_Vector); //, Tm);       
            break;
         case 2: //## RyRzRy      
            RotatesAboutY( -psi[ii],   ThirdOrderOutPD_Vector); //, Tm);                    
            RotatesAboutZ(  theta[ii], ThirdOrderOutPD_Vector); //, Tm);  
            RotatesAboutY( -phi[ii],   ThirdOrderOutPD_Vector); //, Tm);                     
            break;
         case 3: //## RxRzRy      
            RotatesAboutY( -psi[ii],   ThirdOrderOutPD_Vector); //, Tm);                    
            RotatesAboutZ(  theta[ii], ThirdOrderOutPD_Vector); //, Tm);  
            RotatesAboutX(  phi[ii],   ThirdOrderOutPD_Vector); //, Tm);                     
            break;
         case 4: //## RxRzRyRzRy  
            RotatesAboutY( -psi[ii],   ThirdOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  theta[ii], ThirdOrderOutPD_Vector); //, Rotation);  
            RotatesAboutY( -phi[ii],   ThirdOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  delta[ii], ThirdOrderOutPD_Vector); //, Rotation);  
            RotatesAboutX(  ksi[ii],   ThirdOrderOutPD_Vector); //, Rotation);  
            break;                                    
         case 5: //## RxRzRyRzRy  
            RotatesAboutY( -psi[ii],   ThirdOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  theta[ii], ThirdOrderOutPD_Vector); //, Rotation);  
            RotatesAboutY( -phi[ii],   ThirdOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  delta[ii], ThirdOrderOutPD_Vector); //, Rotation);  
            RotatesAboutX(  ksi[ii],   ThirdOrderOutPD_Vector); //, Rotation);  
            RotatesAboutZ(  chi[ii], ThirdOrderOutPD_Vector); //, Rotation);  
            break;  
         default: //## Rotation matrix inexistent
            cout << "From OptFiber->GetThirdOrderPD_Vector(.)" 
                 << endl << "TypeScattering = " << TypeScattering
                 << " inexistent." << endl;   
      }         
      RotatesAboutX( v2Delta,  ThirdOrderOutPD_Vector); //, Tm);        
      
      //## Add the contribution of the 3rd order PDV
      ThirdOrderOutPD_Vector[1] +=  2*SectionDGD*SndOrderOutPD_Vector[2]
                                   -  sq(SectionDGD)*OutPD_Vector[1];
      ThirdOrderOutPD_Vector[2] += -2*SectionDGD*SndOrderOutPD_Vector[1]
                                   -  sq(SectionDGD)*OutPD_Vector[2];      
            
      
      //## Add the new contribution to the OutPD_Vector
      OutPD_Vector[0] += -SectionDGD;  
      
      //## Add the cross product between the OutPD_Vectors
      //## This is the new contribution to the 2nd Order PDV
      SndOrderOutPD_Vector[1] +=  SectionDGD*OutPD_Vector[2];
      SndOrderOutPD_Vector[2] += -SectionDGD*OutPD_Vector[1];
      
      
   }
}



/*!

   This method returns the numerical output second-order 
   polarization-dispersion vector of the current fiber at the frequency Freq
   in the input parameter SndOrderOutPD_Vector. This method is not cumulative.
   Therefore, it does not require initialization.
   
   The second-order polarization-dispersion vector is the
   derivative of the polarization-dispersion vector with respect
   to the angular frequency. 

   Since this method is numerical, one should use a method that returns
   the second-order polarization-mode dispersion from an analytical 
   method: GetSndOrderOutPD_Vector(). 
   This method was developed only to test the analytical
   method for the computation of the second-order PMD.
   
*/

//########################################################
//#### Gets the second order output Pol. Disp. Vector of the fiber
//## It calculates the derivative numerically
//## The original value of the vector is ignored
//## By: Ivan Lima
void OptFiber::GetNumSndOrderOutPD_Vector(double Freq
                                  ,double *SndOrderOutPD_Vector)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }   
   double Freq2 = Freq + 1e-4/GetExpectedDGD();
   double OutPD_Vector[] = {0,0,0};
   double OutPD_Vector2[] = {0,0,0};
   GetOutPD_Vector(Freq,  OutPD_Vector);
   GetOutPD_Vector(Freq2, OutPD_Vector2);
   
   //cout << " Freq = " << Freq << "   Freq2 = " << Freq2 << endl;   
   //cout <<"OutPD_Vector @ Freq = " << OutPD_Vector[0] <<"  "
   //                      << OutPD_Vector[1] <<"  "
   //                     << OutPD_Vector[2]<< endl;                             
   //cout <<"OutPD_Vector @ Freq2 = " << OutPD_Vector2[0] <<"  "
   //                      << OutPD_Vector2[1] <<"  "
   //                     << OutPD_Vector2[2]<< endl;
                          
   for (int ii = 0; ii < 3; ii++)
      SndOrderOutPD_Vector[ii] = (OutPD_Vector2[ii]
                                 -OutPD_Vector[ii])
                                 /(2*pi*(Freq2-Freq));
}                                  



/*!

   This method returns the length of the second-order polarization
   dispersion vector, which is the derivative of the polarization-dispersion
   vector with respect to the angular frequency.
   
   Note that the length of the second-order polarization-dispersion vector
   is not the same as the derivative of the DGD with respect to the 
   angular frequency. 

*/
  
//###############################################################
//## Gets the Snd Order DGD = length of the Snd Order PD vector
//## By: Ivan Lima
double OptFiber::GetSndOrderDGD(double Freq)
{
   double SndOrderOutPD_Vector[] = {0,0,0};
   GetSndOrderOutPD_Vector(Freq, SndOrderOutPD_Vector);
   return(sqrt(sq(SndOrderOutPD_Vector[0])
              +sq(SndOrderOutPD_Vector[1])
              +sq(SndOrderOutPD_Vector[2])) ); 
}



/*!

   This method returns the numerical output third-order 
   polarization-dispersion vector of the current fiber at the frequency Freq
   in the input parameter SndOrderOutPD_Vector. This method is not cumulative.
   Therefore, it does not require initialization of the input parameters.
   
   The third-order polarization-dispersion vector is the
   second derivative of the polarization-dispersion vector with 
   respect to the angular frequency. 

   Since this method is numerical, one should use a method that returns
   the third-order polarization-mode dispersion from an analytical 
   expression. This method was developed only to test the analytical
   method for the computation of the second-order PMD:
   GetThirdOrderOutPD_Vector().
   
   
*/

//########################################################
//#### Gets the Third order output Pol. Disp. Vector of the fiber
//## It calculates the derivative numerically
//## The original value of the vector is ignored
//## By: Ivan Lima
void OptFiber::GetNumThirdOrderOutPD_Vector(double Freq
                                  ,double *ThirdOrderOutPD_Vector)
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }   

   double Freq2 = Freq + 1e-4/GetExpectedDGD();
   double SndOrderOutPD_Vector[] = {0,0,0};
   double SndOrderOutPD_Vector2[] = {0,0,0};
   GetSndOrderOutPD_Vector(Freq,  SndOrderOutPD_Vector);
   GetSndOrderOutPD_Vector(Freq2, SndOrderOutPD_Vector2);
   
   //cout << " Freq = " << Freq << "   Freq2 = " << Freq2 << endl;   
   //cout <<"OutPD_Vector @ Freq = " << OutPD_Vector[0] <<"  "
   //                      << OutPD_Vector[1] <<"  "
   //                     << OutPD_Vector[2]<< endl;                             
   //cout <<"OutPD_Vector @ Freq2 = " << OutPD_Vector2[0] <<"  "
   //                      << OutPD_Vector2[1] <<"  "
   //                     << OutPD_Vector2[2]<< endl;
                          
   for (int ii = 0; ii < 3; ii++)
      ThirdOrderOutPD_Vector[ii] = (SndOrderOutPD_Vector2[ii]
                                 -SndOrderOutPD_Vector[ii])
                                 /(2*pi*(Freq2-Freq));
}     


/*!

   This method returns the length of the third-order polarization
   dispersion vector, which is the second derivative of the polarization-
   dispersion vector with respect to the angular frequency.
   
   Note that the length of the third-order polarization-dispersion vector
   is not the same as the second derivative of the DGD with respect to the 
   angular frequency. 

*/

//###############################################################
//## Gets the 3rd Order DGD = length of the 3rd Order PD vector
//## By: Ivan Lima
double OptFiber::GetThirdOrderDGD(double Freq)
{
   double ThirdOrderOutPD_Vector[] = {0,0,0};
   GetThirdOrderOutPD_Vector(Freq, ThirdOrderOutPD_Vector);
   return(sqrt(sq(ThirdOrderOutPD_Vector[0])
              +sq(ThirdOrderOutPD_Vector[1])
              +sq(ThirdOrderOutPD_Vector[2])) ); 
}



/*!

   This method returns the numericl output polarization-dispersion vector 
   of the current fiber for all the frequencies in the FFT window.
   The polarization-dispersion vector at each frequency of the FFT 
   window is stored in the pointer *NumOutPD_Vector[3], which 
   has three dimensions. the pointer numer correspond to the frequency
   in the FFT array. The memory must have been alocated to the pointer
   array before calling this method.
   
   This method is numerical and is very slow, since the numerical
   value of the polarization dispersion vector is obtained by
   using a numerical derivative of the Stokes vector with respect
   to the frequency. Therefore, this function should not be used.
   This function was written only to validate
   the analytical method GetNumOutPD_Vector().
   
*/

//#################################################################
//## Gets numerically the output PSP of the fiber for the whole
//#  Frequency band.
//## OutPD_Vector should have dimension 3(0 to 2)xqtPoints.
//#  It should be initialize as zero before.
//#  This function is based on the Poicareh Sphere Method        
//#  It computes the PD vector with the effects caused
//#  by the non-linearity and the dispersion.
//#  It is very slow, because the PD vector is calculated for all
//#  the qtPoints in the FFT vector.
//## By: Ivan Lima
void OptFiber::GetNumOutPD_Vector(double *NumOutPD_Vector[3])
{
   if (! MemAllocated || ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }        
   if (PMD_Fiber == 0) {
      cerr << "Error: ProcessNumericOptPD_vector cannot run with PMD_Fiber=0." 
           << endl;   
      exit(1);   
   } 
   sfftNumPD_Vector.tOptSignalX = NewCplx(qtPoints);
   sfftNumPD_Vector.tOptSignalY = NewCplx(qtPoints);
   
   for (int jj = 0; jj < 3; jj++) {                      
      Tmp1PD_Vector[jj]    = new double[qtPoints];
      Tmp2PD_Vector[jj]    = new double[qtPoints];
      Dev1PD_Vector[jj]    = new double[qtPoints];                
      Dev2PD_Vector[jj]    = new double[qtPoints];        
   }
   double PropagatedLength = 0;
   // ## Initialize fixed paramters
   double FreqDifferential = 1e-4/(sqrt(LengthFiber)*PMD_Fiber); // ACF PDV ~ 1/PMD_Fiber
   //cout << "FreqDifferential [GHz] = " << FreqDifferential*1e-9 << endl;
   
   double OptAmplitude;
   double CenterFreq_Original = CenterFreq;
   double StokesA[] = {0, 1, 1/2., 1/3.};  // The 0-th component is blank
   double StokesB[] = {0,-1, 1/2.,-1/3.};
   double TmpStokes[4];
   cplx JonesA[2], JonesB[2], TmpJones[2];

   Stokes2Jones(StokesA, JonesA);    
   Stokes2Jones(StokesB, JonesB);          
   //cout << JonesA[0]<< "  " << JonesA[1] << endl;   
   //cout << JonesB[0]<< "  " << JonesB[1] << endl;
   
   // ## Back up the incoming optical pulse 
   // and initialize the 1st signal
   for (int ii = 0; ii < qtPoints; ii++) { 
      sfftNumPD_Vector.tOptSignalX[ii] = sfftM.tOptSignalX[ii];
      sfftNumPD_Vector.tOptSignalY[ii] = sfftM.tOptSignalY[ii];      

      OptAmplitude = sqrt( sq(abs(sfftNumPD_Vector.tOptSignalX[ii])) 
                              +sq(abs(sfftNumPD_Vector.tOptSignalY[ii])) );      
      if (OptAmplitude < NEAR_ZERO )
         OptAmplitude = NEAR_ZERO;  // from ocsTools.hh                          
      sfftM.tOptSignalX[ii] = JonesA[0]*OptAmplitude;
      sfftM.tOptSignalY[ii] = JonesA[1]*OptAmplitude;
   }   
   fft->FFT(sfftM.tOptSignalX ,sfftM.fOptSignalX);
   fft->FFT(sfftM.tOptSignalY ,sfftM.fOptSignalY);   
   PropagateFiberCoupledNLS(&PropagatedLength);    
   
   CenterFreq = CenterFreq + FreqDifferential;
   // ## Process 1st signal and propagate the signal with the 2nd freq.
   //fft->fftshift(sfftM.fOptSignalX);
   //fft->fftshift(sfftM.fOptSignalY);    
   for (int ii = 0; ii < qtPoints; ii++) {  
      TmpJones[0] = sfftM.fOptSignalX[ii];
      TmpJones[1] = sfftM.fOptSignalY[ii];      
      Jones2Stokes(TmpJones, TmpStokes);         
      for (int jj = 0; jj < 3; jj++)
         Tmp1PD_Vector[jj][ii] = TmpStokes[jj+1];                  

      OptAmplitude = sqrt( sq(abs(sfftNumPD_Vector.tOptSignalX[ii])) 
                              +sq(abs(sfftNumPD_Vector.tOptSignalY[ii])) );
      if (OptAmplitude < NEAR_ZERO )
         OptAmplitude = NEAR_ZERO; // fro ocsTools.hh                         
      sfftM.tOptSignalX[ii] = JonesA[0]*OptAmplitude;
      sfftM.tOptSignalY[ii] = JonesA[1]*OptAmplitude;
   }      
   fft->FFT(sfftM.tOptSignalX ,sfftM.fOptSignalX);
   fft->FFT(sfftM.tOptSignalY ,sfftM.fOptSignalY);      
   PropagateFiberCoupledNLS(&PropagatedLength);    

   CenterFreq = CenterFreq_Original;
   // ## Process the 1st differential and propagate the 2nd signal 
   // with the 1st freq.
   //fft->fftshift(sfftM.fOptSignalX);
   //fft->fftshift(sfftM.fOptSignalY);    
   for (int ii = 0; ii < qtPoints; ii++) {  
      TmpJones[0] = sfftM.fOptSignalX[ii];
      TmpJones[1] = sfftM.fOptSignalY[ii];      
      Jones2Stokes(TmpJones, TmpStokes);         
      for (int jj = 0; jj < 3; jj++)      
         Dev1PD_Vector[jj][ii] = (TmpStokes[jj+1] - Tmp1PD_Vector[jj][ii])
                              /(2*pi*FreqDifferential);

      OptAmplitude = sqrt( sq(abs(sfftNumPD_Vector.tOptSignalX[ii])) 
                              +sq(abs(sfftNumPD_Vector.tOptSignalY[ii])) );
      if (OptAmplitude < NEAR_ZERO )
         OptAmplitude = NEAR_ZERO; // fro ocsTools.hh                           
      sfftM.tOptSignalX[ii] = JonesB[0]*OptAmplitude;
      sfftM.tOptSignalY[ii] = JonesB[1]*OptAmplitude;
   }     
   fft->FFT(sfftM.tOptSignalX ,sfftM.fOptSignalX);
   fft->FFT(sfftM.tOptSignalY ,sfftM.fOptSignalY);      
   PropagateFiberCoupledNLS(&PropagatedLength);    
 
   CenterFreq = CenterFreq + FreqDifferential;
   // ## Process the 1st frep of the 2nd signal, and propagate
   // the 2nd freq of the 2nd signal
   //fft->fftshift(sfftM.fOptSignalX);
   //fft->fftshift(sfftM.fOptSignalY);    
   for (int ii = 0; ii < qtPoints; ii++) { 
      TmpJones[0] = sfftM.fOptSignalX[ii];
      TmpJones[1] = sfftM.fOptSignalY[ii];      
      Jones2Stokes(TmpJones, TmpStokes);   
      for (int jj = 0; jj < 3; jj++)            
         Tmp2PD_Vector[jj][ii] = TmpStokes[jj+1]; 

      OptAmplitude = sqrt( sq(abs(sfftNumPD_Vector.tOptSignalX[ii])) 
                              +sq(abs(sfftNumPD_Vector.tOptSignalY[ii])) );
      if (OptAmplitude < NEAR_ZERO )
         OptAmplitude = NEAR_ZERO; // fro ocsTools.hh                           
      sfftM.tOptSignalX[ii] = JonesB[0]*OptAmplitude;
      sfftM.tOptSignalY[ii] = JonesB[1]*OptAmplitude;
   }    
   fft->FFT(sfftM.tOptSignalX ,sfftM.fOptSignalX);
   fft->FFT(sfftM.tOptSignalY ,sfftM.fOptSignalY);  
   PropagateFiberCoupledNLS(&PropagatedLength);    
   
   
   // ## Get the PD vector and restore the original signal.  
   double Denominator20, Denominator01, Denominator12;   
   //fft->fftshift(sfftM.fOptSignalX);
   //fft->fftshift(sfftM.fOptSignalY);    
   for (int ii = 0; ii < qtPoints; ii++) { 
      TmpJones[0] = sfftM.fOptSignalX[ii];
      TmpJones[1] = sfftM.fOptSignalY[ii];      
      Jones2Stokes(TmpJones, TmpStokes);
      for (int jj = 0; jj < 3; jj++)                       
         Dev2PD_Vector[jj][ii] = (TmpStokes[jj+1] - Tmp2PD_Vector[jj][ii])
                              /(2*pi*FreqDifferential);
                            
      Denominator20 = Tmp1PD_Vector[2][ii]*Tmp2PD_Vector[0][ii]
                     -Tmp2PD_Vector[2][ii]*Tmp1PD_Vector[0][ii];  //den31=s3*s1c-s3c*s1
      Denominator01 = Tmp1PD_Vector[0][ii]*Tmp2PD_Vector[1][ii]
                     -Tmp1PD_Vector[1][ii]*Tmp2PD_Vector[0][ii]; //den12=s1*s2c-s1c*s2
      Denominator12 = Tmp1PD_Vector[1][ii]*Tmp2PD_Vector[2][ii]
                     -Tmp1PD_Vector[2][ii]*Tmp2PD_Vector[1][ii]; //den23=s2*s3c-s2c*s3
                           
      if (abs(Denominator20) > abs(Denominator01) ) 
          NumOutPD_Vector[0][ii] 
                      += ( Tmp1PD_Vector[0][ii]*Dev2PD_Vector[1][ii]
                          -Tmp2PD_Vector[0][ii]*Dev1PD_Vector[1][ii])
                          /Denominator20; //(ds2c*s1-ds2*s1c)/den31
      else
          NumOutPD_Vector[0][ii] 
                      += ( Tmp2PD_Vector[0][ii]*Dev1PD_Vector[2][ii]
                          -Tmp1PD_Vector[0][ii]*Dev2PD_Vector[2][ii])
                          /(-Denominator01); //dispv1=(ds3c*s1-ds3*s1c)/den12                

      if (abs(Denominator12) > abs(Denominator01) ) 
          NumOutPD_Vector[1][ii] 
                      += ( Tmp2PD_Vector[1][ii]*Dev1PD_Vector[0][ii]
                          -Tmp1PD_Vector[1][ii]*Dev2PD_Vector[0][ii])
                          /(-Denominator12); //dispv2=(ds1c*s2-ds1*s2c)/den23
      else
          NumOutPD_Vector[1][ii] 
                      +=( Tmp1PD_Vector[1][ii]*Dev2PD_Vector[2][ii]
                         -Tmp2PD_Vector[1][ii]*Dev1PD_Vector[2][ii])
                         /Denominator01; //dispv2=(ds3c*s2-ds3*s2c)/den12

      if (abs(Denominator12) > abs(Denominator20) ) 
          NumOutPD_Vector[2][ii] 
                      +=( Tmp1PD_Vector[2][ii]*Dev2PD_Vector[0][ii]
                         -Tmp2PD_Vector[2][ii]*Dev1PD_Vector[0][ii])
                         /Denominator12; //dispv3=(ds1c*s3-ds1*s3c)/den23
      else
          NumOutPD_Vector[2][ii] 
                      +=( Tmp2PD_Vector[2][ii]*Dev1PD_Vector[1][ii]
                         -Tmp1PD_Vector[2][ii]*Dev2PD_Vector[1][ii])
                         /(-Denominator20); //dispv3=(ds2c*s3-ds2*s3c)/den31

   
      sfftM.tOptSignalX[ii] = sfftNumPD_Vector.tOptSignalX[ii];
      sfftM.tOptSignalY[ii] = sfftNumPD_Vector.tOptSignalY[ii];      
   }       
   fft->FFT(sfftM.tOptSignalX ,sfftM.fOptSignalX);
   fft->FFT(sfftM.tOptSignalY ,sfftM.fOptSignalY);     
      
   // ## Disalocate the memory
   FreeCplx(sfftNumPD_Vector.tOptSignalX);
   FreeCplx(sfftNumPD_Vector.tOptSignalY);     
   for (int jj = 0; jj < 3; jj++) {   
      delete [] Tmp1PD_Vector[jj];
      delete [] Tmp2PD_Vector[jj];  
      delete [] Dev1PD_Vector[jj]; 
      delete [] Dev2PD_Vector[jj];                              
   }   
   CenterFreq = CenterFreq_Original;  // ## Restore the central frequency
}



/*************
//#############################################################
//## Methods, local functions and static variables used by 
//#  the importance sampling applied to PMD: Effected DGD.
//#  Unfortunately, it did not worked as expected.
//#  Thus, for now you only have the option of biasing 
//#  the DGD at a selected frequency, 
//#  such as the central frequency.

//## Function prototypes and static variables used by the 
//#  optimization with importance sampling applied to PMD.

double GetImportSampEffectDGD( double thetaZthetaX[]);      
void GradientImportSampEffectDGD( double thetaZthetaX[],
                                  double Gradient_thetaZthetaX[]);                                            
enum TypeBiasEffectDGD { MEAN_SQ_DGD = 0,
                         PENALTY     = 1,
                         MEAN_DGD    = 2};
TypeBiasEffectDGD geTypeBiasEffectDGD;
double WeightFreqImportSamp; 
double SectionDGD;
double NormExpectedDGD;
double RetMinimum;
double OutStokesVectorCenterFreq[3];
double OutStokesVectorLeftFreq[3];   
double OutStokesVectorRightFreq[3]; 
double OutPD_VectorCenterFreq[3];
double OutPD_VectorLeftFreq[3];   
double OutPD_VectorRightFreq[3]; 
double v2DeltaCenterFreq;
double v2DeltaLeftFreq;
double v2DeltaRightFreq;   


// ## Get the low probable average DGD values 
// #  weighted by the spectrum of the signal 
// #  using Important Sampling.
// #  This variation was proposed by Ivan and Aurenice Lima
// ## By: Ivan Lima
double OptFiber::GetFiberSampleImportSamplingEffectDGD(double BiasingParam, 
                                                       int TypeBiasEffectDGDTmp)
{
   geTypeBiasEffectDGD = TypeBiasEffectDGD(TypeBiasEffectDGDTmp);
   if (! MemAllocated) {
      cerr << "Error: Object of class OptFiber not initialized." 
           << endl;
      exit(1);
   }  
   if (PMD_Fiber == 0) {
      cerr << "Error: The procedure GetFiberSampleImportSampling(...)"
           << "Is being called for a fiber with PMD == 0." 
           << endl;
      exit(1);
   }   
   if (BiasingParam < 1) {
      cerr << "From GetFiberSampleImportSamplingCos(double)"
           << "Error: Biasing param "<< BiasingParam 
           << " Cannot be < 1." 
           << endl;
      exit(1);   
   }            
   
//   if ( DeltaFreqImportSamp == 0 || WeightFreqImportSamp == 0) {
//     cout << "From GetFiberSampleImportSamplingEffectDGD(double)"
//           << "Error: DeltaFreqImpSampling or WeightFreqImpSampling "
//           << " are not defined for this format." << endl
//           << "The definition should be done by the method"
//           << "InitImpSamplingEffectDGD()." << endl
//           << "Thus, check the constructor of the OptFiber class "
//           << endl;
//      exit(1);      
//   
//   }

   //## Initialize the parameters for the importance sampling
   //#  using the Effective DGD implementation
   //#  It should be set according to the format
   double DeltaFreqImportSamp;
   DeltaFreqImportSamp = oOptSignal->GetBitRateChann();
   WeightFreqImportSamp = 0.75; //0.3;   
   
   //double SectionDGD;
   double SourceRandomVar;
   double LikelihoodRatio = 1;
   TypeScattering = 3; // ## RxRzRy 
   GotFiberSample = 1;   
   //int qtStepsFiberTmp = qtStepsFiber;   
   int LastStep;
   //double v2DeltaCenterFreq;
   //double v2DeltaLeftFreq;
   //double v2DeltaRightFreq;      

   double OutPD_VectorTmp[3];
   double thetaZthetaY[2];  
   //## For the optimization
   double Theta_Psi[3];        //_WeightFreq_SectionDGD_NomrDGD[6]; 
       
   //double Rotation[3][3]; // ## it is necessary for TransV3to_Xhat 
   //## Definition of the the rotation angles as RxRzRy 
   //## for the fiber.          
   GetFiberSampleUnifRxRzRy();               
   //cvtheta   = 2*(RNG->GetRanNum()-0.5);                        
   //theta[0] = acos(cvtheta);                     
   //phi[0]   = 2*pi*RNG->GetRanNum();          
   //psi[0]   = 2*pi*RNG->GetRanNum();        
   //GetRxRzRyMatrix(0);
   for (int ii = 0; ii < 3; ii++) {
      //## Get the Stokes Vector of the Channel 0
      OutStokesVectorCenterFreq[ii]
         = OutStokesVectorLeftFreq[ii]    
            = OutStokesVectorRightFreq[ii]  
               = oOptSignal->GetInStokes(0,ii+1);      
      //cout << "OutStokesVector = "<< OutStokesVectorCenterFreq[ii] << endl;
      OutPD_VectorCenterFreq[ii] = 0;
      OutPD_VectorLeftFreq[ii]   = 0;   
      OutPD_VectorRightFreq[ii]  = 0;   
   }      

   SectionDGD =  -2*dBeta_2dOmega*StepLength[0];             
   //OutPD_VectorCenterFreq[0] += SectionDGD; //-2*dBeta_2dOmega*StepLength[0];  
   //OutPD_VectorLeftFreq[0]   += SectionDGD; //-2*dBeta_2dOmega*StepLength[0];  
   //OutPD_VectorRightFreq[0]  += SectionDGD; //-2*dBeta_2dOmega*StepLength[0];               
   //cout <<"OutPD_VectorCenterFreq[0] = " << OutPD_VectorCenterFreq[0]<< endl;
   //double RetMinimum, 
   double Tolerance = 1e-3;
   int qtIterations;
   for (int ii = 0; ii < qtStepsFiber; ii++) {        
      //if (ii > 0) {
         //LastStep = ii-1;   
         
         //## The DGD of this section
         SectionDGD =  -2*dBeta_2dOmega*StepLength[ii];           
         //## The expected DGD thus far
         NormExpectedDGD = GetExpectedDGD()*sqrt(ii+1)/sqrt(qtStepsFiber);
                           
         // ## Calculates the frequency dependent phase variation
         v2DeltaCenterFreq = 2*dBeta_2dOmega*StepLength[ii]*2*pi*CenterFreq;                    
         v2DeltaLeftFreq   = 2*dBeta_2dOmega*StepLength[ii]*2*pi
                                *(CenterFreq - DeltaFreqImportSamp);                           
         v2DeltaRightFreq  = 2*dBeta_2dOmega*StepLength[ii]*2*pi                             
                                *(CenterFreq + DeltaFreqImportSamp);          

         //## Here comes the optimization algorithm
         //#  I should scan for theta and psi that maximizes 
         //#  the average DGD weighted by the spectral 
         //#  shape of the signal. Then, I will introduce 
         //#  an extra random rotation in phi.
         Theta_Psi[1] = 0;
         Theta_Psi[2] = 0;
         //cout << endl <<ii<<" "<< Theta_Psi[1] <<" "
         //     << Theta_Psi[2] <<" "
         //     << WeightFreqImportSamp <<" "        
         //     << SectionDGD <<" "        
         //     << NormExpectedDGD <<endl;               
         frprmnd( Theta_Psi, (int) 2
                ,Tolerance,&qtIterations
                ,&RetMinimum,&GetImportSampEffectDGD
                ,&GradientImportSampEffectDGD);                         
                
         psi[ii] = Theta_Psi[2];
         theta[ii] = Theta_Psi[1];    

         //## Generates the Euler angles, biasing towards the 
         //#  optimized value of theta
         SourceRandomVar = RNG->GetRanNum();
         theta[ii] += acos(2.*pow(SourceRandomVar,1./BiasingParam) - 1. );
         LikelihoodRatio *= 1./BiasingParam
                   *pow(SourceRandomVar,-(1.-1./BiasingParam));
         //## This will cause a rotation about X. It adds the 
         //#  third degree of freedom, which does not mess up 
         //#  the biasing because the added polarization dispersion
         //#  vector section is always aligned to the X axis
                  
         //## Extra rotation
         if (geTypeBiasEffectDGD) {
            phi[ii] = 0;           
         }   
         else {
            //## Add an extra rotation to avoid 2-D random walk
            phi[ii] = 2*pi*RNG->GetRanNum();          
         }         
         
         //## Computes the cumulative polarization dispersion vector 
         RotatesAboutY( -psi[ii],    OutPD_VectorCenterFreq); //, Rotation);                    
         RotatesAboutZ(  theta[ii],  OutPD_VectorCenterFreq); //, Rotation);  
         RotatesAboutX(  phi[ii],    OutPD_VectorCenterFreq); //, Rotation);                      
         RotatesAboutX(  v2DeltaCenterFreq,OutPD_VectorCenterFreq); //, Rotation);   // ## Propagation    
         OutPD_VectorCenterFreq[0] += SectionDGD; //-2*dBeta_2dOmega*StepLength[ii];  

         RotatesAboutY( -psi[ii],   OutPD_VectorLeftFreq); //, Rotation);                    
         RotatesAboutZ(  theta[ii], OutPD_VectorLeftFreq); //, Rotation);  
         RotatesAboutX(  phi[ii],   OutPD_VectorLeftFreq); //, Rotation);                      
         RotatesAboutX(  v2DeltaLeftFreq, OutPD_VectorLeftFreq); //, Rotation);   // ## Propagation    
         OutPD_VectorLeftFreq[0] += SectionDGD; //-2*dBeta_2dOmega*StepLength[ii];  

         RotatesAboutY( -psi[ii],   OutPD_VectorRightFreq); //, Rotation);                    
         RotatesAboutZ(  theta[ii], OutPD_VectorRightFreq); //, Rotation);  
         RotatesAboutX(  phi[ii],   OutPD_VectorRightFreq); //, Rotation);                      
         RotatesAboutX(  v2DeltaRightFreq,OutPD_VectorRightFreq); //, Rotation);   // ## Propagation    
         OutPD_VectorRightFreq[0] += SectionDGD;//-2*dBeta_2dOmega*StepLength[ii];  


         //## Rotates the polarization vectors of the signal
         RotatesAboutY( -psi[ii],    OutStokesVectorCenterFreq); //, Rotation);                    
         RotatesAboutZ(  theta[ii],  OutStokesVectorCenterFreq); //, Rotation);  
         RotatesAboutX(  phi[ii],    OutStokesVectorCenterFreq); //, Rotation);                               
         RotatesAboutX(  v2DeltaCenterFreq,OutStokesVectorCenterFreq); //, Rotation);   // ## Propagation    

         RotatesAboutY( -psi[ii],    OutStokesVectorLeftFreq); //, Rotation);                    
         RotatesAboutZ(  theta[ii],  OutStokesVectorLeftFreq); //, Rotation);  
         RotatesAboutX(  phi[ii],    OutStokesVectorLeftFreq); //, Rotation);                               
         RotatesAboutX(  v2DeltaLeftFreq,OutStokesVectorLeftFreq); //, Rotation);   // ## Propagation    

         RotatesAboutY( -psi[ii],    OutStokesVectorRightFreq); //, Rotation);                    
         RotatesAboutZ(  theta[ii],  OutStokesVectorRightFreq); //, Rotation);  
         RotatesAboutX(  phi[ii],    OutStokesVectorRightFreq); //, Rotation);                            
         RotatesAboutX(  v2DeltaRightFreq,OutStokesVectorRightFreq); //, Rotation);   // ## Propagation    

         //for (int jj = 0; jj < 3; jj++)
         //   OutPD_VectorTmp[jj] = OutPD_Vector[jj];         
         //TransRzRyV3to_Xhat(OutPD_VectorTmp,thetaZthetaY); //,Rotation);                   
         //theta[ii] +=  thetaZthetaY[0];
         //psi[ii]    = -thetaZthetaY[1];          
      //}                      
      GetRxRzRyMatrix(ii);                   
   }   
   return(LikelihoodRatio);
}


//##############################################################
//## Returns the parameter that was optimized (maximized)
//## By: Ivan Lima
double OptFiber::GetOptimParamImportSamplingEffectDGD(void) 
{
   return -RetMinimum;
}


//#############################################################
double GetImportSampEffectDGD( double Theta_Psi[])
{
   double OptimParam, OptimParamScalarProd;
   double OutPD_VectorTmp[3];
   double OutStokesVectorTmp[3];   
   

   double psi = Theta_Psi[2];  
   double theta = Theta_Psi[1];     

   for (int ii = 0; ii < 3; ii++) {
      OutPD_VectorTmp[ii] = OutPD_VectorCenterFreq[ii];
      OutStokesVectorTmp[ii] = OutStokesVectorCenterFreq[ii];
   }         
   //## Computes the cumulative polarization dispersion vector 
   RotatesAboutY( -psi,    OutPD_VectorTmp); //, Rotation);                    
   RotatesAboutZ(  theta,  OutPD_VectorTmp); //, Rotation);                       
   RotatesAboutX(  v2DeltaCenterFreq,OutPD_VectorTmp); //, Rotation);   // ## Propagation    
   OutPD_VectorTmp[0] +=  SectionDGD;  

   if (geTypeBiasEffectDGD == MEAN_DGD) {   
      OptimParam = GetVectorLength(OutPD_VectorTmp,3);    
   }
   else {
      OptimParam = sq(GetVectorLength(OutPD_VectorTmp,3) );    
   }   
   if (geTypeBiasEffectDGD == PENALTY) {
      //## Rotates the Stokes vector at this frequency
      RotatesAboutY( -psi,    OutStokesVectorTmp); //, Rotation);                    
      RotatesAboutZ(  theta,  OutStokesVectorTmp); //, Rotation);                       
      RotatesAboutX(  v2DeltaCenterFreq,OutStokesVectorTmp); //, Rotation);   // ## Propagation       

      OptimParamScalarProd = ScalarProduct(OutPD_VectorTmp,
                                           OutStokesVectorTmp);                         
   }               
 
   for (int ii = 0; ii < 3; ii++) {
      OutPD_VectorTmp[ii] = OutPD_VectorLeftFreq[ii];
      OutStokesVectorTmp[ii] = OutStokesVectorLeftFreq[ii];
   }    
   //## Computes the cumulative polarization dispersion vector 
   RotatesAboutY( -psi,    OutPD_VectorTmp); //, Rotation);                    
   RotatesAboutZ(  theta,  OutPD_VectorTmp); //, Rotation);  
   RotatesAboutX(  v2DeltaLeftFreq,OutPD_VectorTmp); //, Rotation);   // ## Propagation    
   OutPD_VectorTmp[0] +=  SectionDGD;  

   //## The contribution of the left side band
   if (geTypeBiasEffectDGD == MEAN_DGD) {      
      OptimParam += WeightFreqImportSamp
                   *GetVectorLength(OutPD_VectorTmp,3);   
   }                
   else {
      OptimParam += WeightFreqImportSamp
                   *sq(GetVectorLength(OutPD_VectorTmp,3) );   
   }                
   if (geTypeBiasEffectDGD == PENALTY) {                 
      //## Rotates the Stokes vector at this frequency
      RotatesAboutY( -psi,    OutStokesVectorTmp); //, Rotation);                    
      RotatesAboutZ(  theta,  OutStokesVectorTmp); //, Rotation);                       
      RotatesAboutX(  v2DeltaLeftFreq,OutStokesVectorTmp); //, Rotation);   // ## Propagation       

      OptimParamScalarProd += WeightFreqImportSamp 
                             *ScalarProduct(OutPD_VectorTmp,
                                            OutStokesVectorTmp);                                              
   }

   for (int ii = 0; ii < 3; ii++) {
      OutPD_VectorTmp[ii] = OutPD_VectorRightFreq[ii];
      OutStokesVectorTmp[ii] = OutStokesVectorRightFreq[ii];
   }  
   //## Computes the cumulative polarization dispersion vector 
   RotatesAboutY( -psi,    OutPD_VectorTmp); //, Rotation);                    
   RotatesAboutZ(  theta,  OutPD_VectorTmp); //, Rotation);                       
   RotatesAboutX(  v2DeltaRightFreq,OutPD_VectorTmp); //, Rotation);   // ## Propagation    
   OutPD_VectorTmp[0] +=  SectionDGD;  

   //## The contribution of the right side band
   if (geTypeBiasEffectDGD == MEAN_DGD) {         
      OptimParam += WeightFreqImportSamp
                   *GetVectorLength(OutPD_VectorTmp,3);
   }
   else {                
      OptimParam += WeightFreqImportSamp
                   *sq(GetVectorLength(OutPD_VectorTmp,3) );   
   }                
   OptimParam /= (1+2*WeightFreqImportSamp);             
   if (geTypeBiasEffectDGD == PENALTY) {                 
      //## Rotates the Stokes vector at this frequency
      RotatesAboutY( -psi,    OutStokesVectorTmp); //, Rotation);                    
      RotatesAboutZ(  theta,  OutStokesVectorTmp); //, Rotation);                       
      RotatesAboutX(  v2DeltaRightFreq,OutStokesVectorTmp); //, Rotation);   // ## Propagation       

      OptimParamScalarProd += WeightFreqImportSamp 
                             *ScalarProduct(OutPD_VectorTmp,
                                            OutStokesVectorTmp);                                              

      //cout << "geTypeBiasEffectDGD = "<< geTypeBiasEffectDGD << endl;
      //cout << "OptimParamScalarProd = " << sq(OptimParamScalarProd)
      //                                    /(1+2*WeightFreqImportSamp) << endl;                
      //cout << "LengthOutStokesVectorTmp = "<< GetVectorLength(OutStokesVectorTmp,3)<< endl;                                    
      //cout << "OptimParam = " << OptimParam << endl;                

      //## Subtract the scalar product squared                                      
      OptimParam -= sq( OptimParamScalarProd
                       /(1+2*WeightFreqImportSamp) );                                            

   }                  
   
   //## Produces the normalization according 
   //#  to the weights and the expected DGD. 
   //#  It also changes the sign, since the 
   //#  function looks for a minimum, while 
   //#  we want to maximize the penalty.
   if (geTypeBiasEffectDGD != MEAN_DGD) {  
      OptimParam = sqrt(OptimParam);
   }   
   OptimParam /= -NormExpectedDGD;                 
   

   //cout <<"& "<< Theta_Psi[1] <<" "
   //     << Theta_Psi[2] <<" "
   //     << WeightFreqImportSamp <<" "        
   //     << SectionDGD <<" "        
   //     << NormExpectedDGD <<" " 
   //     << OptimParam <<endl;    

   return( OptimParam );
}


//#############################################################
// ## Get the gradient of PolDispVectorVariation
void GradientImportSampEffectDGD( double Theta_Psi[]
                                 ,double Gradient_Theta_Psi[])
{
   //cout << "I am in the gradient"<< endl;
   double LocalValue = GetImportSampEffectDGD(Theta_Psi);

   double GradientStep = 1e-4;

   Theta_Psi[1] += GradientStep;
   Gradient_Theta_Psi[1] = (GetImportSampEffectDGD(
       Theta_Psi) - LocalValue)
                              /GradientStep;
   Theta_Psi[1] -= GradientStep;  

   Theta_Psi[2] += GradientStep;
   Gradient_Theta_Psi[2] = (GetImportSampEffectDGD(
           Theta_Psi) - LocalValue)
                              /GradientStep;
   Theta_Psi[2] -= GradientStep;                     
   
   
   //cout <<"   Gradient " << Gradient_Theta_Psi[1] << " "
   //     << Gradient_Theta_Psi[2] << endl; 
   //cout <<"Gradient [1]" << Gradient_Theta_Psi[1] << endl;
   //cout <<"Gradient [2]" << Gradient_Theta_Psi[2] << endl;      
}

**************/




/*****************
//#############################################################
// ## Get the low probable DGD values using Important Sampling
// #  using the bias proposed by Biondini and Kath.
// #  It should not be used because it lacks an extra rotation. 
// ## By: Ivan Lima
double OptFiber::GetFiberSampleImportSamplingCos(double BiasingParam)
{
   if (! MemAllocated) {
      cerr << "Error: Object of class OptFiber not initialized." 
           << endl;
      exit(1);
   }   
   if (PMD_Fiber == 0) {
      cerr << "Error: The procedure GetFiberSampleImportSampling(...)"
           << "Is being called for a fiber with PMD == 0." 
           << endl;
      exit(1);
   }   
   if (BiasingParam < 1) {
      cerr << "From GetFiberSampleImportSamplingCos(double)"
           << "Error: Biasing param "<< BiasingParam 
           << " Cannot be < 1." 
           << endl;
      exit(1);   
   }            
   double SourceRandomVar;
   double LikelihoodRatio = 1;
   TypeScattering = 1; // ## Euler Matrix
   GotFiberSample = 1;   
   //int qtStepsFiberTmp = qtStepsFiber;   
   int LastStep;
   double v2Delta;
   double OutPD_Vector[] = {0,0,0}, OutPD_VectorTmp[3];
   double thetaZthetaX[2];     
   //double Rotation[3][3]; // ## it is necessary for TransV3to_Xhat 
   // ## Generate the Euler angles with complete scattering
   // #  for the first section.                      
   cvtheta   = 2*(RNG->GetRanNum()-0.5);                        
   theta[0] = acos(cvtheta);                     
   phi[0]   = 2*pi*RNG->GetRanNum();          
   psi[0]   = 2*pi*RNG->GetRanNum();     
   GetEulerMatrix(0);    
   for (int ii = 1; ii < qtStepsFiber; ii++) {
      // ## Generate the Euler angles

      SourceRandomVar = RNG->GetRanNum();
      theta[ii] = acos(2.*pow(SourceRandomVar,1./BiasingParam) - 1. );
      LikelihoodRatio *= 1./BiasingParam
                *pow(SourceRandomVar,-(1.-1./BiasingParam));
                    
      // ## This module the Euler Matrix
      psi[ii] = -pi/2.;          
      //if (ii > 0) {
         LastStep = ii-1;   
         // ## Calculates the cumulative Pol. Disp. Vector
         v2Delta = 2*dBeta_2dOmega*StepLength[LastStep]*2*pi*CenterFreq;                    
         RotatesAboutX( phi[LastStep],  OutPD_Vector); //, Rotation);
         RotatesAboutY(-theta[LastStep],OutPD_Vector); //, Rotation);
         RotatesAboutX( psi[LastStep],  OutPD_Vector); //, Rotation);           
         RotatesAboutX( v2Delta,  OutPD_Vector); //, Rotation);   // ## Propagation    
         OutPD_Vector[0] += -2*dBeta_2dOmega*StepLength[LastStep];  
      
         for (int jj = 0; jj < 3; jj++)
            OutPD_VectorTmp[jj] = OutPD_Vector[jj];         
         TransRzRxV3to_Xhat(OutPD_VectorTmp,thetaZthetaX); //,Rotation);                                     
         theta[ii] += thetaZthetaX[0];
         phi[ii]    = thetaZthetaX[1]+pi/2.;          
      //}                      
      GetEulerMatrix(ii);                   
   }   
   return(LikelihoodRatio);
}
************/



/*
//###############################################################
//## Get the low probable DGD values with Important Sampling
//#  using the biased pdf proposed by Ivan Lima.
//#  Note that the importance sampling applied to PMD was
//#  proposed by Biondini and Kath. I just happend to be
//#  the first one to apply it to PMD-induced distortions 
//#  and PMD compensators.
//#  This is the right function to use.
//#  To cover the normalized DGD from 0 to 4 with 80 sections, 
//#  use bias 0.5 and 1.0, besides the unbiased.
//#  To cover the normalized DGD from 0 to 4 with 200 sections, 
//#  which provides a more accurate emulation of the tail, use 
//#  bias 0.3 and 0.6, besides the unbiased.
//#  I would like to remind that all but $10^{-8}$ of the DGD 
//#  are in the range from 0 to 4.
//## By: Ivan Lima (10/18/01)
double OptFiber::GetFiberSampleImportSamplingExp(double BiasingParam)
{
   if (! MemAllocated) {
      cerr << "Error: Object of class OptFiber not initialized." 
           << endl;
      exit(1);
   }   
   if (PMD_Fiber == 0) {
      cerr << "Error: The procedure GetFiberSampleImportSamplingCosRx(...)"
           << "Is being called for a fiber with PMD == 0." 
           << endl;
      exit(1);
   }   
        
   double SourceRandomVar;
   double LikelihoodRatio = 1;
   TypeScattering = 3; //## RxRzRy 
   GotFiberSample = 1;   
   //int qtStepsFiberTmp = qtStepsFiber;   
   int LastStep;
   double v2Delta;
   double OutPD_Vector[] = {0,0,0}, OutPD_VectorTmp[3];
   double thetaZthetaY[2];     

   //## Generate a random rotation for the first section.                      
   //## It can be later changed if one wants to bias the 
   //#  first order PMD distortion
   cvtheta   = 2*(RNG->GetRanNum()-0.5);                        
   theta[0] = acos(cvtheta);                     
   phi[0]   = 2*pi*RNG->GetRanNum();          
   psi[0]   = 2*pi*RNG->GetRanNum();     
   GetRxRzRyMatrix(0);    
   for (int ii = 1; ii < qtStepsFiber; ii++) {
      //## Generate the rotation angles
      SourceRandomVar = RNG->GetRanCosineBiased(BiasingParam,
                                              &LikelihoodRatio);                                                 
      theta[ii] = acos(SourceRandomVar);                                        
      phi[ii] =  2*pi*RNG->GetRanNum();                    
        
      LastStep = ii-1;   
      
      // ## Calculates the cumulative Pol. Disp. Vector
      v2Delta = 2*dBeta_2dOmega*StepLength[LastStep]*2*pi*CenterFreq;                    
      RotatesAboutY( -psi[LastStep],   OutPD_Vector); //, Rotation);                    
      RotatesAboutZ(  theta[LastStep], OutPD_Vector); //, Rotation);  
      RotatesAboutX(  phi[LastStep],   OutPD_Vector); //, Rotation);                      
      RotatesAboutX(  v2Delta,  OutPD_Vector); //, Rotation);   // ## Propagation    
      OutPD_Vector[0] += -2*dBeta_2dOmega*StepLength[LastStep];  
      
      for (int jj = 0; jj < 3; jj++) {
         OutPD_VectorTmp[jj] = OutPD_Vector[jj];         
      }   
      TransRzRyV3to_Xhat(OutPD_VectorTmp,thetaZthetaY); //,Rotation);                   
      //## Generate biased theta and psi. The angle psi aids
      //#  the biasing process.
      theta[ii] +=  thetaZthetaY[0];
      psi[ii]    = -thetaZthetaY[1];          
      GetRxRzRyMatrix(ii);                   
   }   
   return(LikelihoodRatio);
}
*/




