// $Id:
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsPolarizBeamSplitter.hh"

// ################### Methods for the class PolarizBeamSplitter      
// #### Initializes the object PolarizBeamSplitter
// ## By: Ivan Lima
PolarizBeamSplitter::PolarizBeamSplitter( OptSignal *oOptSignalATmp
                                         ,OptSignal *oOptSignalBTmp)
// ## It is assumed that the original signal is in oOptSignalATmp. 
// #  It is important that oOptSignalBTmp have the same # of points
// #  and the same # of bits of oOptSignalATmp.
{
   oOptSignalA = oOptSignalATmp;
   oOptSignalB = oOptSignalBTmp; 
   if (oOptSignalA->GetQtPoints() 
       != oOptSignalB->GetQtPoints() ) {
      cerr << "Error: The two OptSignal objects" 
           << "should have the same size."
           << endl;
      exit(1);
   }         
   qtPoints = oOptSignalA->GetQtPoints();        
   
   // ## Sets the default splitting polarization as the horizontal
   //SetSplittingPolarization(1.,0,0);
   
   SetPolarizBeamSplitterAngle(pi/4);  //## Default == 0
   SetPolarizBeamRetarderAngle(pi/2.); //## Default == 0  
}


// #### Defines the polarizer to split horizontal and vertical polarization
// ## By: Ivan Lima
void PolarizBeamSplitter::SetPolarizBeamSplitterHorizontal_Vertical(void)
{  
   SetPolarizBeamSplitterAngle(0.);  //## Default == 0
   SetPolarizBeamRetarderAngle(0.); //## Default == 0  
}


// #### Defines the polarizer to split circular right and left polarizations
// ## By: Ivan Lima
void PolarizBeamSplitter::SetPolarizBeamSplitterCircular(void)
{  
   SetPolarizBeamSplitterAngle(pi/4.);  //## Default == 0
   SetPolarizBeamRetarderAngle(pi/2.); //## Default == 0  
}


// #### Defines the polarizer to split diagonal +45 and -45 polarizations
// ## By: Ivan Lima
void PolarizBeamSplitter::SetPolarizBeamSplitterDiagonal(void)
{  
   SetPolarizBeamSplitterAngle(pi/4.);  //## Default == 0
   SetPolarizBeamRetarderAngle(0.); //## Default == 0  
}



// #### Defines the new splitting polarization of the polarizer
// ## By: Ivan Lima
void PolarizBeamSplitter::SetPolarizBeamSplitterAngle(double theta2)
{
   //## Attention: theta == pi/4. will give +45 and -45 polarizations.
   //## theta == 0 corresponds to horizontal and vertical
   theta = theta2;
}

void PolarizBeamSplitter::SetPolarizBeamRetarderAngle(double phi2)
{
   //## Attention: phi == pi/2. will convert right and left 
   //#  circular polarizations into liner +45 and -45 polarizations
   //#  It corresponds to a quarter wave-plate
   //#  It should be set to zero for linear polarizer
   phi = phi2;
}


/*********
// #### Defines the new splitting polarization of the polarizer
// ## By: Ivan Lima
void PolarizBeamSplitter::SetSplittingPolarization(double S1PolarizerTmp
                        ,double S2PolarizerTmp
                        ,double S3PolarizerTmp)
{
   S1Polarizer = S1PolarizerTmp;
   S2Polarizer = S2PolarizerTmp;
   S3Polarizer = S3PolarizerTmp;      
   double S0Polarizer = sqrt( sq(S1Polarizer)
                             +sq(S2Polarizer)
                             +sq(S3Polarizer) );
   if (S0Polarizer == 0) {
      cerr << "Error: The polarization components of the " 
           << "Polarization beam splitter cannot be zero."
           << endl;
      exit(1);
   }                      
   // ## Normalizing the polarization    
   S1Polarizer =  S1Polarizer/S0Polarizer; 
   S2Polarizer =  S2Polarizer/S0Polarizer; 
   S3Polarizer =  S3Polarizer/S0Polarizer;      
}                       
*********/

// #### Split the optical signal
// ## By: Ivan Lima
void PolarizBeamSplitter::PolarizSplitOptSignal()
// ## The object oOptSignalA contains the input signal.
// ## The object oOptSignalA will end up with the component of 
// #  the signal in the linear polarization received in 
// #  this method. The object oOptSignalB will end up with 
// #  the component of the 
// #  signal in the orthogonal polarization. 
{   
   //ObjectInitialized = 1;
   //qtPoints = fft->give_N();      

   //int qtPoints = oOptSignal->GetQtPoints();
   //double Stokes[4];
   //cplx TransmissionA, TransmissionB;
   //double StokesShifted[3];
   //double SPolarizer[] = {S1Polarizer, S2Polarizer, S3Polarizer};          
   cplx Jones[2];
   for (int ii = 0; ii < qtPoints; ii++) {
      Jones[0] = cos(theta)*oOptSignalA->sfftM.fOptSignalX[ii]
                -sin(theta)*exp(-jc*phi)*oOptSignalA->sfftM.fOptSignalY[ii];
      Jones[1] = sin(theta)*oOptSignalA->sfftM.fOptSignalX[ii]
                +cos(theta)*exp(-jc*phi)*oOptSignalA->sfftM.fOptSignalY[ii];      
                
      //Jones2Stokes(Jones,Stokes);
      //for (int kk = 1; kk < 4; kk++)
         //StokesShifted[kk-1] = Stokes[kk];        
      // ## Gets the transmission coefficient  
      // ## The term exp(.) guarantees the maintenance of the chirp
      //TransmissionA = sqrt(0.5
      //       *( ScalarProduct(SPolarizer,StokesShifted)+1.)
      //       *( sq(abs(oOptSignalA->sfftM.tOptSignalX[ii]))
      //         +sq(abs(oOptSignalA->sfftM.tOptSignalY[ii])) ) )
      //        *exp(jc*GetPhase(oOptSignalA->sfftM.tOptSignalX[ii]));         
      //TransmissionB = sqrt(0.5
      //       *(-ScalarProduct(SPolarizer,StokesShifted)+1.)
      //       *( sq(abs(oOptSignalA->sfftM.tOptSignalX[ii]))
      //         +sq(abs(oOptSignalA->sfftM.tOptSignalY[ii])) ) )
      //        *exp(jc*GetPhase(oOptSignalA->sfftM.tOptSignalX[ii]));                 
      // ## Leaves the signal polarized                     
      //for (int kk = 1; kk < 4; kk++)
      //   Stokes[kk] = SPolarizer[kk-1];        
      //Stokes2Jones(Stokes,Jones);
      oOptSignalA->sfftM.fOptSignalX[ii] = Jones[0];
      //                  TransmissionA*Jones[0];   
      oOptSignalA->sfftM.fOptSignalY[ii] = 0.;
      //                  TransmissionA*Jones[1];
      // ## Leaves the orthogonal polarization in oOptSignalB                  
      //for (int kk = 1; kk < 4; kk++)
      //   Stokes[kk] = -SPolarizer[kk-1];        
      //Stokes2Jones(Stokes,Jones);                        
      oOptSignalB->sfftM.fOptSignalX[ii] = 0.;
      //                  TransmissionB*Jones[0];   
      oOptSignalB->sfftM.fOptSignalY[ii] = Jones[1];
      //                  TransmissionB*Jones[1];

   }
   oOptSignalA->Freq2Time();
   oOptSignalB->Freq2Time();   
   //oOptSignalA->Time2Freq();
   //oOptSignalB->Time2Freq();
   
   int StringLength = oOptSignalA->GetStringLength();
   for (int ii = 0; ii < StringLength; ii++) {
      oOptSignalB->oBitString->SetBit(0,ii,
         oOptSignalA->oBitString->GetBit(0,ii) );
   }   
   
}


/*********
// #### Split the optical signal
// ## By: Ivan Lima
void PolarizBeamSplitter::PolarizSplitOptSignal_OLD()
// ## The object oOptSignalA contains the input signal.
// ## The object oOptSignalA will end up with the component of 
// #  the signal in the polarization received in this method. The 
// #  object oOptSignalB will end up with the component of the 
// #  signal in the orthogonal polarization. 
{   
   //ObjectInitialized = 1;
   //qtPoints = fft->give_N();      

   //int qtPoints = oOptSignal->GetQtPoints();
   double Stokes[4];
   cplx TransmissionA, TransmissionB;
   double StokesShifted[3];
   double SPolarizer[] = {S1Polarizer, S2Polarizer, S3Polarizer};          
   cplx Jones[2];
   for (int ii = 0; ii < qtPoints; ii++) {
      Jones[0] = oOptSignalA->sfftM.tOptSignalX[ii];
      Jones[1] = oOptSignalA->sfftM.tOptSignalY[ii];      
      Jones2Stokes(Jones,Stokes);
      for (int kk = 1; kk < 4; kk++)
         StokesShifted[kk-1] = Stokes[kk];        
      // ## Gets the transmission coefficient  
      // ## The term exp(.) guarantees the maintenance of the chirp
      TransmissionA = sqrt(0.5
             *( ScalarProduct(SPolarizer,StokesShifted)+1.)
             *( sq(abs(oOptSignalA->sfftM.tOptSignalX[ii]))
               +sq(abs(oOptSignalA->sfftM.tOptSignalY[ii])) ) )
              *exp(jc*GetPhase(oOptSignalA->sfftM.tOptSignalX[ii]));         
      TransmissionB = sqrt(0.5
             *(-ScalarProduct(SPolarizer,StokesShifted)+1.)
             *( sq(abs(oOptSignalA->sfftM.tOptSignalX[ii]))
               +sq(abs(oOptSignalA->sfftM.tOptSignalY[ii])) ) )
              *exp(jc*GetPhase(oOptSignalA->sfftM.tOptSignalX[ii]));                 
      // ## Leaves the signal polarized                     
      for (int kk = 1; kk < 4; kk++)
         Stokes[kk] = SPolarizer[kk-1];        
      Stokes2Jones(Stokes,Jones);
      oOptSignalA->sfftM.tOptSignalX[ii] = 
                        TransmissionA*Jones[0];   
      oOptSignalA->sfftM.tOptSignalY[ii] = 
                        TransmissionA*Jones[1];
      // ## Leaves the orthogonal polarization in oOptSignalB                  
      for (int kk = 1; kk < 4; kk++)
         Stokes[kk] = -SPolarizer[kk-1];        
      Stokes2Jones(Stokes,Jones);                        
      oOptSignalB->sfftM.tOptSignalX[ii] = 
                        TransmissionB*Jones[0];   
      oOptSignalB->sfftM.tOptSignalY[ii] = 
                        TransmissionB*Jones[1];

   }
   oOptSignalA->Time2Freq();
   oOptSignalB->Time2Freq();
   
   int StringLength = oOptSignalA->GetStringLength();
   for (int ii = 0; ii < StringLength; ii++) {
      oOptSignalB->oBitString->SetBit(0,ii,
         oOptSignalA->oBitString->GetBit(0,ii) );
   }   
   
}
********/               
               
               
/**********
// #### This method shall not be used, yet
// #### Split the optical signal in the frequency domain
// ## By: Ivan Lima
void PolarizBeamSplitter::PolarizSplitOptSignalFreq()
// ## The object oOptSignalA contains the input signal.
// ## The object oOptSignalA will end up with the component of 
// #  the signal in the polarization received in this method. The 
// #  object oOptSignalB will end up with the component of the 
// #  signal in the orthogonal polarization. 
{   
   //ObjectInitialized = 1;
   //qtPoints = fft->give_N();      

   //int qtPoints = oOptSignal->GetQtPoints();
   double Stokes[4];
   cplx TransmissionA, TransmissionB;
   double StokesShifted[3];
   double SPolarizer[] = {S1Polarizer, S2Polarizer, S3Polarizer};          
   cplx Jones[2];
   for (int ii = 0; ii < qtPoints; ii++) {
      Jones[0] = oOptSignalA->sfftM.fOptSignalX[ii];
      Jones[1] = oOptSignalA->sfftM.fOptSignalY[ii];      
      Jones2Stokes(Jones,Stokes); //## The Stokes comes normalized
      //cout << " J 0 -> " << Jones[0] 
      //     << " J 1 -> " << Jones[1] << endl;
      for (int kk = 1; kk < 4; kk++) {
         //cout << " " << Stokes[kk];
         StokesShifted[kk-1] = Stokes[kk];        
      }   
      //cout << endl;
      // ## Gets the transmission coefficient  
      // ## The term exp(.) guarantees the maintenance of the chirp      
      TransmissionA = sqrt(0.5
             *( ScalarProduct(SPolarizer,StokesShifted)+1.)
             *( sq(abs(oOptSignalA->sfftM.fOptSignalX[ii]))
               +sq(abs(oOptSignalA->sfftM.fOptSignalY[ii])) ) )
               *exp(jc*GetPhase(oOptSignalA->sfftM.fOptSignalX[ii])); 
      TransmissionB = sqrt(0.5
             *(-ScalarProduct(SPolarizer,StokesShifted)+1.)
             *( sq(abs(oOptSignalA->sfftM.fOptSignalX[ii]))
               +sq(abs(oOptSignalA->sfftM.fOptSignalY[ii])) ) )
               *exp(jc*GetPhase(oOptSignalA->sfftM.fOptSignalX[ii]));                 
      //cout << "T-A " <<   sqrt(0.5
      //       *( ScalarProduct(SPolarizer,StokesShifted)+1.) )<< endl;
      //cout << "T-B " <<   sqrt(0.5
      //       *(-ScalarProduct(SPolarizer,StokesShifted)+1.) )<< endl;             
                         
      //cout << "T-A " << TransmissionA/
      //        ( sq(abs(oOptSignalA->sfftM.fOptSignalX[ii]))
      //         +sq(abs(oOptSignalA->sfftM.fOptSignalY[ii])) ) 
      //    << " T-B "<< TransmissionB/ 
      //        ( sq(abs(oOptSignalA->sfftM.fOptSignalX[ii]))
      //         +sq(abs(oOptSignalA->sfftM.fOptSignalY[ii])) )          
      //    << endl;                
      // ## Leaves the signal polarized                     
      for (int kk = 1; kk < 4; kk++)
         Stokes[kk] = SPolarizer[kk-1];        
      Stokes2Jones(Stokes,Jones); //## The Jones comes normalized
      //cout << "A J 0 -> " << Jones[0] 
      //     << "A J 1 -> " << Jones[1] << endl;      
      oOptSignalA->sfftM.fOptSignalX[ii] = 
                        TransmissionA*Jones[0];   
      oOptSignalA->sfftM.fOptSignalY[ii] = 
                        TransmissionA*Jones[1];
      //cout <<"A -> " << oOptSignalB->sfftM.fOptSignalX[ii] <<" "
      //     << oOptSignalB->sfftM.fOptSignalY[ii] << endl;                                  
      for (int kk = 1; kk < 4; kk++)
         Stokes[kk] = -SPolarizer[kk-1];        
      Stokes2Jones(Stokes,Jones); //## The Jones comes normalized                        
      //cout << "B J 0 -> " << Jones[0] 
      //     << "B J 1 -> " << Jones[1] << endl;   
      oOptSignalB->sfftM.fOptSignalX[ii] = 
                        TransmissionB*Jones[0];   
      oOptSignalB->sfftM.fOptSignalY[ii] = 
                        TransmissionB*Jones[1];
      //cout <<"B -> " << oOptSignalB->sfftM.fOptSignalX[ii] <<" "
      //     << oOptSignalB->sfftM.fOptSignalY[ii] << endl;                 
   }
   oOptSignalA->Freq2Time();
   oOptSignalB->Freq2Time();
   
   int StringLength = oOptSignalA->GetStringLength();
   for (int ii = 0; ii < StringLength; ii++) {
      oOptSignalB->oBitString->SetBit(0,ii,
         oOptSignalA->oBitString->GetBit(0,ii) );
   }   
   
}
               
***********/


               
