// $Id: ocsPolarizer.cc,v 1.2 2006/05/11 22:34:49 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsPolarizer.hh"

// ################### Methods for the class Polarizer         

//! Constructor. Sets default polarization splitting vector in Stokes space
//! to be the horizontal state (1,0,0).

Polarizer::Polarizer(OptSignal *oOptSignalTmp)
{
   oOptSignal = oOptSignalTmp;
   SPolarizer = new double[3];   

   oReduPolarizer = new ReduPolarizer(oOptSignal->oReduOptSignal);

  // ## Sets the default polarization as the horizontal   
   SetPolarization(1.0,0,0);



}

// ########################################################

//! Sets a new polarization splitting vector in Stokes space

void Polarizer::SetPolarization(double S1PolarizerTmp,
				double S2PolarizerTmp,
				double S3PolarizerTmp)
{
   SPolarizer[0] = S1PolarizerTmp;
   SPolarizer[1] = S2PolarizerTmp;
   SPolarizer[2] = S3PolarizerTmp; 
     
   double S0Polarizer = sqrt( sq(SPolarizer[0])
                             +sq(SPolarizer[1])
                             +sq(SPolarizer[2]) );

   if (S0Polarizer == 0) 
   {
      cerr << "Error: The polarization components of the " 
           << "Polarization cannot be zero."
           << endl;
      exit(1);
   }                      

   // ## Normalizing the polarization    

   SPolarizer[0] =  SPolarizer[0]/S0Polarizer; 
   SPolarizer[1] =  SPolarizer[1]/S0Polarizer; 
   SPolarizer[2] =  SPolarizer[2]/S0Polarizer;      

   
   oReduPolarizer->SetPolarization(S1PolarizerTmp,
				   S2PolarizerTmp,S3PolarizerTmp);
}                       

// #########################################################

//! Polarizes optical signal
/*!

Suppose that the input Jones signal to the polarizer is

U_{in}(t) = u_{parallel}(t)V_{parallel} + u_{perp}(t)V_{perp},

where V_{parallel} is a Jones vector corresponding to the Stokes vector
SPolarizer, and {V_{parallel}, V_{perp}} for an orthonormal basis
for Jones space. Then we define the output signal after the polarizer to be

U_{out}(t) = u_{parallel}(t)V_{parallel}.

Use the theory of projection operators from linear algebra we have that

If Jones vectors are column vectors then

U_{out}(t) = [conjugatetranspose(V_{parallel})U_in(t)]V_{parallel}.

You can check that U_{out} is independent of the choice of phase for
V_{parallel} that arises when V_{parallel} is computed from SPolarizer.

*/

void Polarizer::PolarizeOptSignal()    
{

  cplx V_Parallel[2];
  double StokesParameters[4];

  StokesParameters[0] = 1;

  for(int kk = 1; kk < 4; kk++)
    StokesParameters[kk] = SPolarizer[kk-1];        

  Stokes2Jones(StokesParameters,V_Parallel);

  for(int ii=0;ii<oOptSignal->GetQtPoints();ii++)
   {
    cplx InnerProduct = conj(V_Parallel[0])*oOptSignal->sfftM.tOptSignalX[ii]
                       +conj(V_Parallel[1])*oOptSignal->sfftM.tOptSignalY[ii];

    oOptSignal->sfftM.tOptSignalX[ii] = InnerProduct*V_Parallel[0];
    oOptSignal->sfftM.tOptSignalY[ii] = InnerProduct*V_Parallel[1];

   }
 
  oOptSignal->Time2Freq();
 
  oReduPolarizer->PolarizeOptSignal();

} // ## end PolarizeOptSignal

// #######################################################

/*****

//! Polarize the signal
//! By: Ivan Lima

void Polarizer::PolarizeOptSignal()                
{
   double Stokes[4];
   cplx Transmission;
   double StokesShifted[3];
   double SPolarizer[] = {S1Polarizer, S2Polarizer, S3Polarizer};
        
   cplx Jones[2];
   for (int ii = 0; ii < qtPoints; ii++) 
    {
      Jones[0] = oOptSignal->sfftM.tOptSignalX[ii];
      Jones[1] = oOptSignal->sfftM.tOptSignalY[ii];      
    
      Jones2Stokes(Jones,Stokes);
    
      for (int kk = 1; kk < 4; kk++)
         StokesShifted[kk-1] = Stokes[kk];        

      // ## Gets the transmission coefficient  

      Transmission = sqrt(0.5
             *(ScalarProduct(SPolarizer,StokesShifted)+1.)
             *( sq(abs(oOptSignal->sfftM.tOptSignalX[ii]))
		+sq(abs(oOptSignal->sfftM.tOptSignalY[ii])) ) )
             *exp(jc*GetPhase(oOptSignal->sfftM.tOptSignalX[ii]));    
            
      // ## Leaves the signal polarized                     

      for (int kk = 1; kk < 4; kk++)
         Stokes[kk] = SPolarizer[kk-1];        

      Stokes2Jones(Stokes,Jones);

      oOptSignal->sfftM.tOptSignalX[ii] = 
                        Transmission*Jones[0];   
      oOptSignal->sfftM.tOptSignalY[ii] = 
                        Transmission*Jones[1];
   }

   oOptSignal->Time2Freq();   
                       
} // ## end PolarizeOptSignal

***/

// ########################################################
               
