// $Id:
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsReduOptAmplifier.hh" 

extern ofstream *LogFile;

// ################### Methods for the class ReduOptAmplifier

// #### Constructor of ReduOptAmplifier
// ## By: Ivan Lima

/*!

\param string InFileName: The name of the input parameter file
\param ReduOptSignal *oReduOptSignalTmp: Pointer to the reduced 
       optical signal object that is to be amplified.


The constructor does the following:

-# Initializes class parameters that it gets from the ReduOptSignal object
-# Reads input parameters from file and initializes internal class parameters.
-# Initializes a ReduPolDepAttenuator object that is used internally to do the
   PDG. 
-# Calls void AllocateNoiseAmplitudeChannels(void)

The parameters that need to be set in the input parameter file are as
follows. 

(Note that in some of the example input files we also include
many other parameters. These are for our full time and frequency domain
amplifier and are NOT required by the reduced amplifier.
Consequently this constructor could be cleaned up quite a bit...)


-# TypeAmplifier: Should be set to 6; 
   See ReduOptAmplifier::TypeAmplifier
-# TypeAmplifierNoise: Should be set to 4; 
    See ReduOptAmplifier::TypeAmplifierNoise
-# NoiseFigOptAmplifier: The noise figure of the amplifier in dB
-# SpontEmissionFactor: The spontaneous emission factor for the noise, n_sp, 
                        on a linear scale.
-# PolDepGainOptAmplifier: The PDG in dB.


We only need to set one of 
NoiseFigOptAmplifier and SpontEmissionFactor since 
SpontaneousEmissionFactor = LinearNoiseFigOptAmplif/2.0
where LinearNoiseFigOptAmplif = dB2Linear(NoiseFigOptAmplifier)

The code knows which  of these two parameters 
to use provided that the user sets
the other one to zero in the input file.

So exactly one of  NoiseFigOptAmplifier and SpontEmissionFactor 
must be zero in the input file.

 */


ReduOptAmplifier::ReduOptAmplifier(string InFileName,
                                   ReduOptSignal *oReduOptSignalTmp)
{    
   oReduOptSignal = oReduOptSignalTmp;
   CenterFreq = oReduOptSignal->GetCenterFreq();
   NumChannels = oReduOptSignal->GetNumChannels();
   
   //PolarizNoiseBandwidth = oReduOptSignal->GetPolarizNoiseBandwidth();
   ChannSpacing = oReduOptSignal->GetChannSpacing();

   ifstream InFile(InFileName.c_str(),ios::in);    
   
   if (!InFile) 
   {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
   }

   LogFileSeparator();
   *LogFile << InFileName << endl;

   DebugLevel  = ReadInt("DebugLevel"," ",1,3,LOWER_AND_UPPER,&InFile);



   TypeAmplifier  = typeAmplifier(ReadInt("TypeAmplifier"," ",0,6,
                                        LOWER_AND_UPPER,&InFile));

   *LogFile << "AmplifierType = " << WriteTypeAmplifier() << endl;

   TypeAmplifierNoise  = typeAmplifierNoise(ReadInt("TypeAmplifierNoise"," ",
                                             0,5,
                                             LOWER_AND_UPPER,&InFile));

   *LogFile << "AmplifierTypeNoise = " << WriteTypeAmplifierNoise() << endl; 
      

   dBGain = ReadDouble("GainOptAmplifier","dB",0,0,NO_BOUNDS,
                        OPTIONAL_NO_WARNING,
                        &InFile);

   // ## Converts from dB to the linear scale                       

   LinearGainOptAmplif = dB2Linear(dBGain);
                       
   if(   TypeAmplifier == VECTOR_FIXED_OUTPUT_POWER
      || TypeAmplifier == VECTOR_FIXED_TOTAL_OUTPUT_POWER)  
    {
      double OutputPowerPerChanneldBm = ReadDouble("OutputPowerPerChanneldBm",
                                              "dBm",0,0,NO_BOUNDS,
                                               OPTIONAL_WARNING,&InFile);
      //OutputPowerLinear = oOptSignal->GetNumChannels()*
      //                     dBm2Linear(OutputPowerPerChanneldBm);
    
      OutputPowerLinear = dBm2Linear(OutputPowerPerChanneldBm)
                          *oReduOptSignal->GetNumChannels();
    
      //if ( OutputPowerLinear == 0.) {
      //   // ## Calculates automatically from the launched power
      //   *LogFile << "The fixed output power is set as"
      //           << " the total original power" << endl;
      //   OutputPowerLinear = oReduOptSignal->GetOptPowerChann();
      //   *LogFile << "OutputPowerLinear = " << OutputPowerLinear << endl;                                                    // 
      //} 
     
   } // ## end if(TypeAmplifier == VECTOR_FIXED_OUTPUT_POWER) 

   NoiseFigOptAmplifier = 0;
   SpontaneousEmissionFactor = 0;
 

   NoiseFigOptAmplifier 
               = ReadDouble("NoiseFigOptAmplifier","dB",0,0,
				  NO_BOUNDS,&InFile);

   SpontaneousEmissionFactor 
                    = ReadDouble("SpontEmissionFactor"," ",
                                        1.0,-1.0,
				        LOWER_ONLY,&InFile);

   if      (NoiseFigOptAmplifier != 0
          && SpontaneousEmissionFactor == 0) 
   {
      LinearNoiseFigOptAmplif 
           = dB2Linear(NoiseFigOptAmplifier);
      SpontaneousEmissionFactor = LinearNoiseFigOptAmplif/2.;     
   }   
   else if (NoiseFigOptAmplifier == 0
          && SpontaneousEmissionFactor != 0) 
   {
      LinearNoiseFigOptAmplif = 2*SpontaneousEmissionFactor;
      //     = dB2Linear(NoiseFigOptAmplifier);
      //SpontaneousEmissionFactor = LinearNoiseFigOptAmplif/2.;     
   }
      
   else if (NoiseFigOptAmplifier != 0
       && SpontaneousEmissionFactor != 0) 
    {    
      cerr << "Error: Object of class ReduOptAmplifier" 
           << endl << "initialized with both "
           << "SpontaneousEmissionFactor and "
           << "NoiseFigOptAmplif non-zero." << endl
           << "Only one of them should remain non-zero"
           << endl;
      exit(1);  
   }    
   else if (NoiseFigOptAmplifier == 0
       && SpontaneousEmissionFactor == 0 
       && TypeAmplifierNoise == NOISE_ON_SEMI_ANALYTICAL) {    
      cout << "Error: Object of class ReduOptAmplifier" 
           << endl << "initialized with both "
           << "SpontaneousEmissionFactor and "
           << "NoiseFigOptAmplif equal to zero" 
           << endl
           << "If you don't want noise, set TypeAmplifierNoise OFF."
           << endl;
      exit(1);      
   }          
       

PolDepGainOptAmplifier = ReadDouble("PolDepGainOptAmplifier","dB"
                               ,0,-1,LOWER_ONLY, MANDATORY,&InFile);   
 
AmpLength = 1.0;

 if(DebugLevel >=3)
   {

    *LogFile << "AmpLength = 1.0." 
	 << endl 
	 << "This value doesn't have any physical meaning at moment."
	 << endl
	 << "The Saturated Amplifier gives same output regardless of"
	 << "value of AmpLength" << endl;
   }

NumZSteps = ReadInt("NumZSteps"," ",1,-1,LOWER_ONLY,OPTIONAL_NO_WARNING,
                     &InFile);

SaturatingPowerdBm = ReadDouble("SaturatingPowerdBm","dBm",
                     1,-1,NO_BOUNDS,OPTIONAL_NO_WARNING,&InFile);

SaturatingPowerLinear = dBm2Linear(SaturatingPowerdBm);

UnsaturatedGaindB = ReadDouble("UnsaturatedGaindB","dB",
                     1,-1,NO_BOUNDS,OPTIONAL_NO_WARNING,&InFile);

UnsaturatedGainLinear = dB2Linear(UnsaturatedGaindB);

UnsaturatedDifferentialGainLinear = log(UnsaturatedGainLinear)/AmpLength;
   
   
  
   LogFileSeparator();

   *LogFile << "Processing Opt Amplifier Input Parameters" << endl << endl;

   *LogFile << "LinearGainOptAmplif = " << LinearGainOptAmplif << endl;
   *LogFile << "LinearNoiseFigOptAmplif = " << LinearNoiseFigOptAmplif << endl;

   *LogFile << "CenterFreq = " << CenterFreq << endl;

   oReduPolDepAttenuator =  new ReduPolDepAttenuator(oReduOptSignal);

   oReduPolDepAttenuator->SetPolDepAttenuation(0.,0.);
   
   AllocateNoiseAmplitudeChannels();
   
   InFile.close();
}             


// #############################################################

// ## Syncronize the number of channels with OptSignal.
// #  By: Ivan Lima
void ReduOptAmplifier::SyncNumChannels(void)
{
   NumChannels = oReduOptSignal->GetNumChannels();
   oReduPolDepAttenuator->SyncNumChannels();
   ReleaseMemory();
   AllocateNoiseAmplitudeChannels();
}

// ##########################################################

// #### Initializes the noise power per channel
// ## By: Ivan Lima

void ReduOptAmplifier::AllocateNoiseAmplitudeChannels(void)
{
   NormalizedNoiseAmplitude = new double[NumChannels];

   for (int jj = 0; jj < NumChannels; jj++) 
   {
      // ## In order to get the noise one needs to 
      // #  multiply the factor by (LinearGain - 1).    
      // #  The factor of two comes from the two 
      // #  orthogonal polarizations.
 
      NormalizedNoiseAmplitude[jj] =  hPlanck
                      *oReduOptSignal->GetFrequency(jj)
                      *2*SpontaneousEmissionFactor;
                      

      //cout << jj << "  " 
      //     << oReduOptSignal->Frequency[jj] << endl;
        //cout << jj <<" NormalizedNoiseAmplitude -> " 
        //     << NormalizedNoiseAmplitude[jj] << endl;         
   }
}

// #######################################################

// #### Destructor of ReduOptAmplifier
// ## By: Ivan Lima
ReduOptAmplifier::~ReduOptAmplifier()
{
   ReleaseMemory();
}

// ####################################################

// #### Destructor of ReduOptAmplifier
// ## By: Ivan Lima

void ReduOptAmplifier::ReleaseMemory()
{
   delete [] NormalizedNoiseAmplitude;
   delete oReduPolDepAttenuator;
}


// ################################################

// #### Amplify the optical signal 
// ## By: Ivan Lima

/*! In the reduced model we use 
    ReduOptAmplifier::TypeAmplifier = 6 = VECTOR_FIXED_TOTAL_OUTPUT_POWER

    So you should ignore most the the options in the switch statement
    in this method.

  The algorithm is as follows:

-# Compute LinearGainOptAmplifier = 
   GetGainFixedOutputPower(oReduOptSignal->GetTotalPower()) \n
   The idea here is to chose the gain that will ensure that 
   the total output power is given by the pre-specified value 
   OutputPowerLinear
-# Amplify with gain given by LinearGainOptAmplifier. The amplification
   involves multiplying the Stokes parameters of the signal and noise
   in each channel as well as the oReduOptSignal->StokesExtraUnpolChannNoise
   by LinearGainOptAmplifier. 
-# Add the noise in the oReduOptSignal->StokesExtraUnpolChannNoise
   where the noise figure is computed using the frequency of 
  the central channel, which is assumed have index NumChannels/2 which it will
  if you use set the flag EvenlySpacedChannels = 1 in the reduced optical
  signal input parameter file. 
-# Add the noise to the Noise Power in each of the channels.
-# Now that we have added noise the total power will be higher than
   the value we wnted it to be. So we decrease the power of the signal
   in each channel to end up with the required total output power.
   Note that we don't change the power of the noise in this step.
-# The consequence of the last step is that after many amplifications
   the noise power grows and the signal power drops. This is what happens
   in realistic gain saturated amplifiers.
 
\sa void SetFixedOutputPowerFromCurrentOptSignal(void),\n
    double GetGainFixedOutputPower(double InputPower),\n
    double OutputPowerLinear,\n
    void AddPolDepGainEffect(void)
*/

void ReduOptAmplifier::AmplifyOptSignal(void)
{

  // The enum types typeAmplifier and typeAmplifierNoise are 
  // defined in ocsReduOptAmplifier.hh

  double InPower;

  if(DebugLevel >=3) 
   {
      LogFileSeparator();
      InPower = Linear2dBm(oReduOptSignal->GetTotalPower());
   }

   //## Add the PDG effect, if it exist.
   //if (PolDepGainOptAmplifier != 0 ) 
   if (  TypeAmplifier != NO_AMPLIFICATION
       && PolDepGainOptAmplifier != 0 )    
    { 
      AddPolDepGainEffect();
    }   


// JZ June 2006 
//  if(   TypeAmplifier == SCALAR_SIMPLE 
//     || TypeAmplifier == SCALAR_SATURATED ) 
 
if(TypeAmplifier == SCALAR_SATURATED )
   {
     *LogFile << endl
            << "WARNING: Scalar Amps will work just" 
            << " as vector amps for the reduced model."
            << endl << "Please, use the vector model" 
            << " when you are using the reduced Stokes"
            << " model." << endl << endl; 
     *LogFile << "The SCALAR_SATURATED amplifier was not tested."
             << endl;
     // JZ cout << "Program aborted. See *.log for details."<< endl;                 // JZ exit(1);             
  }

   switch(TypeAmplifier) 
    {
      case NO_AMPLIFICATION:
         return; //## Do nothing
         break;
      case SCALAR_SIMPLE: 
      case VECTOR_SIMPLE:
         //ComputeNoiseAmplitudeFactor();
         LinearGainOptAmplif = dB2Linear(dBGain);          
         break;
      case SCALAR_SATURATED: 
      case VECTOR_SATURATED:
         LinearGainOptAmplif = SteadyStateSaturatedGain
                             (oReduOptSignal->GetTotalPower());      
         break;
      case VECTOR_FIXED_OUTPUT_POWER:
         LinearGainOptAmplif = GetGainFixedOutputPower(
                               oReduOptSignal->GetSignalPower());  
                       
         break;                    
      case VECTOR_FIXED_TOTAL_OUTPUT_POWER:      
         LinearGainOptAmplif = GetGainFixedOutputPower(
                               oReduOptSignal->GetTotalPower()); 
	 //LinearGainOptAmplif = GetGainFixedOutputPowerDiffNoise(
         //                      oReduOptSignal->GetTotalPower());
                                                     
         //cout << " ---> LinearGainOptAmplif = " 
	 //<< LinearGainOptAmplif << endl;
                                
         break;                      
      default: 
         *LogFile << " WARNING: Error in TypeAmplifier" << endl;

   }//## end switch


   // ## Multiply by the gain and add noise
   
   //cout << "LinearGainOptAmplif -1 = " << LinearGainOptAmplif-1 << endl;   
   
   //## Extra noise inside the channel bandwidth
   //#  The normalized noise amplitude is take 
   //#  from the center of the optical amplifier

   //oReduOptSignal->StokesExtraUnpolChannNoise
   //    *= LinearGainOptAmplif;
       
   /*       
   if (   TypeAmplifierNoise == NOISE_ON
       && LinearGainOptAmplif > 1.) 
    {                
      oReduOptSignal->StokesExtraUnpolChannNoise   
         += NumChannels 
           *NormalizedNoiseAmplitude[NumChannels/2]
           *(ChannSpacing - PolarizNoiseBandwidth)
           *(LinearGainOptAmplif-1);         
   }
   */
   

   //cout << "ChannSpacing = " << ChannSpacing  
   //     << "  PolarizNoiseBandwidth = " << PolarizNoiseBandwidth << endl;
      
   for (int jj = 0; jj < NumChannels; jj++) 
    {
      for (int ii = 0; ii < 4; ii++) 
      {
         //cout << ii << " " << jj << " StokesSignal *-> "
         //     << oReduOptSignal->StokesSignal[ii][jj]<<endl;
         //cout << ii << " " << jj << " StokesNoise *-> "
         //     << oReduOptSignal->StokesNoise[ii][jj]<<endl;  
      
         oReduOptSignal->StokesSignal[ii][jj] 
                    *= LinearGainOptAmplif;
    
         oReduOptSignal->StokesNoise[ii][jj]  
                    *= LinearGainOptAmplif;   
      
         //cout << ii << " " << jj << " StokesSignal *-> "
         //     << oReduOptSignal->StokesSignal[ii][jj]<<endl;
         //cout << ii << " " << jj << " StokesNoise *-> "
         //     << oReduOptSignal->StokesNoise[ii][jj]<<endl; 
                    
      }          

      if (    TypeAmplifierNoise == NOISE_ON_SEMI_ANALYTICAL
           && LinearGainOptAmplif > 1.) 
       {           
         // ## Adds unpolarized noise from spontaneous emission

         oReduOptSignal->StokesNoise[0][jj] 
            += NormalizedNoiseAmplitude[jj]
               *ChannSpacing*(LinearGainOptAmplif-1);     	    
               //*PolarizNoiseBandwidth*(LinearGainOptAmplif-1);            

            //cout << "Noise added ("<<jj <<") = " 
	    // << NormalizedNoiseAmplitude[jj]
            //   *PolarizNoiseBandwidth
            //   *(LinearGainOptAmplif-1) << endl;
            //cout << "  ->StokesNoise = " 
	    // << oReduOptSignal->StokesNoise[0][jj] << endl;   

        //cout << "Local SNR = "<< oReduOptSignal->StokesSignal[0][jj]
        //                        /oReduOptSignal->StokesNoise[0][jj] << endl;           
        //cout << jj <<" New Unpolariz StokesNoise *-> "
        //     << oReduOptSignal->StokesNoise[0][jj]<<endl;              
        //cout << jj <<" NormalizedNoiseAmplitude -> " 
        //     << NormalizedNoiseAmplitude[jj] << endl;            
      }    
   }   

   //cout << "NoiseDensity = "<<  NormalizedNoiseAmplitude[jj]
   //       *(LinearGainOptAmplif-1)/PolarizNoiseBandwidth << endl;  

   /********* This should be deactivate   
   if (TypeAmplifier ==  VECTOR_FIXED_TOTAL_OUTPUT_POWER) 
   {      
      //## Keep the total output power fixed
      //#  This module should be commented in case 
      //#  the GetGainFixedOutputPowerDiffNoise() is 
      //#  used to calculate LinearGainOptAmplif above.
      //#  This is the saturation in Ding's thesis.
   
     double LinearGainAdjustOptAmplif = GetGainFixedOutputPower(
                             oReduOptSignal->GetTotalPower());
                                                     
      // ## Multiply by the gain adjust

      //## Extra noise inside the channel bandwidth

      //oReduOptSignal->StokesExtraUnpolChannNoise      
      //            *= LinearGainAdjustOptAmplif;    
   
      for (int jj = 0; jj < NumChannels; jj++) 
       {
         for (int ii = 0; ii < 4; ii++) {
            oReduOptSignal->StokesSignal[ii][jj] 
                       *= LinearGainAdjustOptAmplif;    
       
            oReduOptSignal->StokesNoise[ii][jj]  
                       *= LinearGainAdjustOptAmplif;   
                    
         }          
      }                                
   } 
   **********/                        
   
   //cout << "LinearGainOptAmplif = " << LinearGainOptAmplif<<endl; 
                            
      //for (int ii = 0; ii < 4; ii++)
      //   cout << oReduOptSignal->StokesSignal[ii][0] << "  " ;
      //cout << endl;                

   if(DebugLevel >=3)    
    {         
      double OutPower = Linear2dBm(oReduOptSignal->GetTotalPower());    
      *LogFile << "InPower [dBm] = " << InPower << endl;
      *LogFile << "OutPower [dBm] = " << OutPower << endl;
      *LogFile << "Gain [dB] = " << OutPower - InPower << endl;
   }      

} // ## end Amplify...

// ################################################################# 

// #### Add the PDG effect
// ## By: Ivan Lima

/*!

The PDG is implemented somewhat differently from the formulae
given on page 489 of [WM 2001] but the final result will be identical.

In this method the idea is that PDG can be modeled by
a negative-loss (i.e. gain!) PDL element, provided that the signal is rotated
so that the high loss axis of the (negative loss)
PDL element is lined up with the 
high gain direction for the PDG. The high gain direction for the PDG 
is the direction that is orthogonal to the direction determined by the 
total Stokes parameters.

The amount of gain in the high gain direction is given by\n 
PolDepGainOptAmplifier*GetDegreeOfPolarization(StokesTotal) in dB.

 */


void ReduOptAmplifier::AddPolDepGainEffect(void)
{  
   double StokesTotal[4];
   double StokesTotalShifted[3]; 
   double StokesSignalShifted[3];
   double StokesNoiseShifted[3];    
   double thetaZthetaY[2];   

   for (int ii = 0; ii < 4; ii++) 
      StokesTotal[ii] = 0;    

   //StokesTotal[0] += 
   //      oReduOptSignal->StokesExtraUnpolChannNoise;   

   for (int jj = 0; jj < NumChannels; jj++) 
    {
      for (int ii = 0; ii < 4; ii++) 
       {
         //cout << ii << " " << jj << " StokesSignal -> "
         //     << oReduOptSignal->StokesSignal[ii][jj]<<endl;
         //cout << ii << " " << jj << " StokesNoise -> "
         //     << oReduOptSignal->StokesNoise[ii][jj]<<endl;              

         StokesTotal[ii] += 
                  oReduOptSignal->StokesSignal[ii][jj] 
                 +oReduOptSignal->StokesNoise[ii][jj];      

         //cout << ii << " " << jj << " StokesTotal-> "
         //     << StokesTotal[ii]<<endl;                    
      }                 
   }   

   // ## PolDepAttenuationY was set to 0 => Linear == 1

   //## It will produce a differential gain (that's -loss)                              
   oReduPolDepAttenuator->SetPolDepAttenuationX(
                  -PolDepGainOptAmplifier
                 *GetDegreeOfPolarization(StokesTotal) );  

//   oReduPolDepAttenuator.SetPolDepAttenuationY(
//                  PolDepGainOptAmplifier
//                 *GetDegreeOfPolarization(StokesTotal) );  

   //cout << "DOP = " << GetDegreeOfPolarization(StokesTotal) << endl;
   //cout << "PDL = " << PolDepGainOptAmplifier
   //              *GetDegreeOfPolarization(StokesTotal) <<endl;

   for (int ii = 1; ii < 4; ii++) 
      StokesTotalShifted[ii-1] = StokesTotal[ii]; 
      
   // ## Get the direction of the Polarization Dependent Gain. 
   // ## The total polarized power defines the direction.

   TransRzRyV3to_Xhat( StokesTotalShifted, thetaZthetaY);
                        
   // ## Rotate all the signal and noise components                        

   for (int jj = 0; jj < NumChannels; jj++) 
    {
      for (int ii = 1; ii < 4; ii++) 
       {
         StokesSignalShifted[ii-1] 
               = oReduOptSignal->StokesSignal[ii][jj];    
         StokesNoiseShifted[ii-1] 
               = oReduOptSignal->StokesNoise[ii][jj];          
      }

      RotatesAboutY( thetaZthetaY[1],StokesSignalShifted);
      RotatesAboutZ( thetaZthetaY[0],StokesSignalShifted);  
      RotatesAboutY( thetaZthetaY[1],StokesNoiseShifted);
      RotatesAboutZ( thetaZthetaY[0],StokesNoiseShifted);     
    
      // ## Put together the whole rotated Stokes parameters
      for (int ii = 1; ii < 4; ii++) 
       {
         oReduOptSignal->StokesSignal[ii][jj] 
                      = StokesSignalShifted[ii-1];    
       
         oReduOptSignal->StokesNoise[ii][jj] 
                      = StokesNoiseShifted[ii-1];          
      }
   }   

   // ## The total polarization is now (S0,-1,0,0)
   // ## Here comes the PDG acting as a PDAttenuator

   oReduPolDepAttenuator->PolDepAttenuate();

   // ## Now rotate the polarized part back to the original coordinates
   for (int jj = 0; jj < NumChannels; jj++) 
    {
      for (int ii = 1; ii < 4; ii++) 
       {
         StokesSignalShifted[ii-1] 
              = oReduOptSignal->StokesSignal[ii][jj];    
         StokesNoiseShifted[ii-1]  
              = oReduOptSignal->StokesNoise[ii][jj];          
      } 
  
      RotatesAboutZ(-thetaZthetaY[0],StokesSignalShifted);
      RotatesAboutY(-thetaZthetaY[1],StokesSignalShifted);
      RotatesAboutZ(-thetaZthetaY[0],StokesNoiseShifted);
      RotatesAboutY(-thetaZthetaY[1],StokesNoiseShifted);
     
      for (int ii = 1; ii < 4; ii++) 
       {
         oReduOptSignal->StokesSignal[ii][jj] 
              = StokesSignalShifted[ii-1];    
         oReduOptSignal->StokesNoise[ii][jj] 
              = StokesNoiseShifted[ii-1];                        
      }      
   }                        
   
} //## End of AddPolDepGainEffect()


// ###############################################
// ## This method is a straight copy from OptAmplifier
// ## By: John Zweck

void ReduOptAmplifier::SetTypeAmplifier(typeAmplifier TypeAmplifier2)
{

  TypeAmplifier = TypeAmplifier2;

  if(DebugLevel >=2)
    {
     *LogFile << "WARNING: Setting type of amplifier to " 
	  << WriteTypeAmplifier()
	  << endl; 
    }

}

// #############################################
// ## This method is a straight copy from OptAmplifier
// ## By: John Zweck

void ReduOptAmplifier::SetTypeAmplifierNoise(
                                typeAmplifierNoise TypeAmplifierNoise2)
{

 TypeAmplifierNoise = TypeAmplifierNoise2;

  if(DebugLevel >=2)
    {
      *LogFile << "WARNING: Setting type of amplifier noise to " 
	  << WriteTypeAmplifierNoise()
	  << endl;
    } 
}


// ###########################################
// ## This method was adapted from the method 
// #  of OptAmplifier by John Zweck
// ## Ivan Lima added the VECTOR_FIXED_..._POWER in 2/16/01

string ReduOptAmplifier::WriteTypeAmplifier(void)
{

 switch(TypeAmplifier)
   {

    case NO_AMPLIFICATION:

           return "NO_AMPLIFICATION";
           break;

    case SCALAR_SIMPLE: 

            return "SCALAR_SIMPLE";
            break;

    case  VECTOR_SIMPLE:
            return "VECTOR_SIMPLE";
            break;

    case SCALAR_SATURATED: 
           return "SCALAR_SATURATED";
           break;


   case VECTOR_SATURATED:
           return "VECTOR_SATURATED";
           break;
           
   case VECTOR_FIXED_OUTPUT_POWER:
           return "VECTOR_FIXED_OUTPUT_POWER";
           break;         
             
   case VECTOR_FIXED_TOTAL_OUTPUT_POWER:
           return "VECTOR_FIXED_TOTAL_OUTPUT_POWER";
           break;         
    }

}

// ######################################
// ## This method is a straight copy from OptAmplifier
// ## By: John Zweck

string ReduOptAmplifier::WriteTypeAmplifierNoise(void)
{

 switch(TypeAmplifierNoise)
   {

    case NOISE_OFF:

           return "NOISE_OFF";
           break;

    case NOISE_ON_CONST_POWER_RANDOM_PHASE:

           return "NOISE_ON_CONST_POWER_RANDOM_PHASE";
           break;

    case NOISE_ON_GAUSSIAN_WHITE:

           return "NOISE_ON_GAUSSIAN_WHITE";
           break;

    case NOISE_ON_NO_RNG:

           return "NOISE_ON_NO_RNG";
           break;

    case NOISE_ON_SEMI_ANALYTICAL:

           return "NOISE_ON_SEMI_ANALYTICAL";
           break;

    case NOISE_ON_BIASED:

           return "NOISE_ON_BIASED";
           break;



        default:
                cerr << "With TypeAmplifierNoise must between 0 and 5" << endl
                << "Aborting now! " << endl;
                exit(1);

    }

}


void  ReduOptAmplifier::SetGain(double dBGain2)
{

 dBGain = dBGain2;
 LinearGainOptAmplif = dB2Linear(dBGain2); 
 ////## The following line is required for the noiseless simulation
 ////LinearGainFixedTotalOutputPower = LinearGainOptAmplif; 
 if(DebugLevel >=3)
   {

     *LogFile << "WARNING: Gain reset to dBGain [dB] = "
         << dBGain << endl;
   }

}


//#########################################################
//## This method should be executed when the optical 
//#  signal object contains the intensity that one wants 
//#  to have after the amplification.
//## This method should be executed only once at 
//#  the beggining
//## By: Ivan Lima, 5/14/01

void ReduOptAmplifier::SetFixedOutputPowerFromCurrentOptSignal(void)
{
   OutputPowerLinear = oReduOptSignal->GetTotalPower();
   //OutputPowerLinear = oReduOptSignal->GetTotalPower()
   //                   /oReduOptSignal->GetNumChannels();

   //cout << "Setting OutputPowerLinar*NumChannels = "
   //<< OutputPowerLinear*NumChannels << endl; 
                     
}


/*
// ######################################################
// ## Gets the gain such that the total output power
// #  is a constant. It includes the added power from the 
// #  noise of the amplifier, as well.
// #  By: Ivan Lima

double ReduOptAmplifier::GetGainFixedOutputPower( double InputPower)
{
   // ## We want only the total power to be fixed.
         
   //cout << "OutputPowerLinear*NumChannels = "
  // << NumChannels*OutputPowerLinear << endl;                        
   //cout << "InputPower = " << InputPower << endl;                         
   //cout << " -> LinearGainOptAmplif = " << LinearGainOptAmplif << endl;
   //return NumChannels*OutputPowerLinear/InputPower;  

   //cout << "InputPower = " << InputPower << endl
   //     << "OutputPowerLinear = " << OutputPowerLinear << endl
   //     << "OutputPowerLinear/InputPower" 
   //	  << OutputPowerLinear/InputPower << endl;
   
   return OutputPowerLinear/InputPower;      
}
*/


// ###########################################################
// ## Gets the gain such that the total output power
// #  is a constant. It includes the added power from the 
// #  noise of the amplifier, as well.
// #  By: Ivan Lima
double ReduOptAmplifier::GetGainFixedOutputPower( double InputPower)
{    

   //## The total power is normalized, including the differential noise.
   //## Note: the noise is normalized with respect to the gain
   double NewNormalizedNoiseAmplitude = 0;
   for (int jj = 0; jj < NumChannels; jj++) {         
      NewNormalizedNoiseAmplitude 
                     += NormalizedNoiseAmplitude[jj];                  
   }
   ///NewNormalizedNoiseAmplitudePerChannel /= oReduOptSignal->GetNumChannels();
                   
   LinearGainOptAmplif =  ( OutputPowerLinear
                           +ChannSpacing * NewNormalizedNoiseAmplitude)
                         /( InputPower
			   +ChannSpacing * NewNormalizedNoiseAmplitude);         

   //cout << "InputPower = " << InputPower << endl
   //     << "DifferentialNoise = " 
   //	  << ChannSpacing*NewNormalizedNoiseAmplitude << endl
   //     << "InputPower/DifferentialNoise = " 
   //     << InputPower/(ChannSpacing*NewNormalizedNoiseAmplitude) << endl;

   
   //## I am assuming that the differential noise contribution of this
   //#  amplifier is negligible
   //   LinearGainOptAmplif =  OutputPowerLinear/InputPower;
   
   return LinearGainOptAmplif;
}


// ###############################################
// ## This method is a straight copy from OptAmplifier
// ## By: John Zweck

double ReduOptAmplifier::SteadyStateSaturatedGain(double InputPower)
{

  // Dec 14th, 2000: This method still needs more testing

  // Written by John Zweck, Nov 30th, 2000

  // For theory see Agrawal [NLO,p476ff] and Ruo Mei's 
  // "Comparison of theory and experiment" paper.

  // This routine returns the linear saturated gain G for which
  // P_out = G*P_in, where P_in is the input power to
  // the amplifier and P_out is the output power.
  // The power is computed by taking the average power
  // of the pulse train. 

  // G = LinearGainOptAmplif

  // The ODE which describes the evolution of the power in the 
  // amplifier is dP/dz = g(z,P(z))*P
  // where z is distance along the amplifier.

  // Ignoring frequency dependence of the gain we choose 
  //
  //  g(z,P) = g_0/( 1+ tau*P(z)/P_sat )

  // We solve the ODE using Euler's method.

  // Variables set in the constructor are:

  // AmpLength
  // NumZSteps
  // SaturatingPower (P_sat)

   //cout << "UnsaturatedDifferentialGainLinear = " 
   //     << UnsaturatedDifferentialGainLinear << endl;  
   //cout << "SaturatingPowerLinear = " << SaturatingPowerLinear << endl;  

  // LinearUnsaturatedDifferentialGain , g_0 = g(z,0)
  // Note that g_0 = (1+tau)*g(z,P_sat) also
  // This could be a way to compute tau I guess.....

  // CHECK: denominator in Eulers method eqn below is correct,
  // and work out how to get SaturatingPower.

  double DeltaZ =  AmpLength/float(NumZSteps);
  
  double OutputPower = InputPower;

  for(int ii=0;ii<NumZSteps;ii++)
    {
      OutputPower += UnsaturatedDifferentialGainLinear*DeltaZ*OutputPower/
                     (1+ (OutputPower/SaturatingPowerLinear));
    }
  
  return OutputPower/InputPower; 

}// ## end SteadyStateSaturatedGain


// ####################################################
// ## This method is a straight copy from OptAmplifier
// ## By: John Zweck

void ReduOptAmplifier::WriteSteadyStateGainVersusPower(string OutFileName,
                   double InputPowerMaxLinear, int NumPowerSteps)
{

  // Dec 14th, 2000: This method still needs more testing

  // Written by John Zweck, Nov 30th, 2000

  // This method writes out (10*log10(G) , 10*log10(P_in/1mW)) pairs
  // in units (dB, dBm)
  // where G_0 = exp(g_0 AmpLength) is the UnsaturatedGain
  // and G is the SteadyStateSaturatedGain

 ofstream OutFile(OutFileName.c_str(),ios::out);

 if (!OutFile) 
   {
      *LogFile << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }



 double DeltaPower = InputPowerMaxLinear/NumPowerSteps;
 
 LogFileSeparator();

 *LogFile << "WriteSteadyStateGainVersusPower" << endl << endl;


 *LogFile << "InputPowerMaxLinear [W] = " << InputPowerMaxLinear << endl;
 *LogFile << "NumPowerSteps = " << NumPowerSteps << endl;
 *LogFile << "DeltaPower = " <<  DeltaPower << endl;


 double SaturatedGain;
 double InputPower;

 for(int ii = 1; ii <= NumPowerSteps; ii++) 
     // Start at ii==1 as need InputPower > 0
   {
     InputPower = ii*DeltaPower;
     SaturatedGain =  SteadyStateSaturatedGain(InputPower);

     OutFile << 10*log10(InputPower/1e-3)  << " " 
	     << 10*log10(SaturatedGain) << endl;

   } // end for loop


 OutFile.close();


}// ## end WriteSteadyStateGainVersusPower

//####################################################
//## Plot Input versus output power in linear scale
//## By: Ivan Lima (9/3/2002)
void ReduOptAmplifier::WriteSteadyInputPowerVersusOutputPower(string OutFileName,
                   double InputPowerMaxLinear, int NumPowerSteps)
{
   //## This method was adapted from another function by John Zweck
   // This method writes out (P_in (mW), P_out (mW), G) pairs
   // where G_0 = exp(g_0 AmpLength) is the UnsaturatedGain
   // and G is the SteadyStateSaturatedGain

   ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) {
      *LogFile << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }
   double DeltaPower = InputPowerMaxLinear/NumPowerSteps;
   LogFileSeparator();

   *LogFile << "WriteSteadyStateInputPowerVersusOutputPower" << endl << endl;

   *LogFile << "InputPowerMaxLinear [W] = " << InputPowerMaxLinear << endl;
   *LogFile << "NumPowerSteps = " << NumPowerSteps << endl;
   *LogFile << "DeltaPower = " <<  DeltaPower << endl;

   double SaturatedGain;
   double InputPower;

   for(int ii = 1; ii <= NumPowerSteps; ii++) {
      InputPower = ii*DeltaPower;
      SaturatedGain =  SteadyStateSaturatedGain(InputPower);

      OutFile << InputPower  << "  " 
 	      << SaturatedGain*InputPower << "  "
	      << SaturatedGain << endl;

   } // end for loop
   OutFile.close();
}// ## end WriteSteadyStateGainVersusPower






//####################################################
//## Set the parameters for a saturated amplifier operating
//## in gain compression
//## By: Ivan Lima (9/3/2002)
void ReduOptAmplifier::SetGainCompressionSaturatedAmplifier(
                   double GainTarget_dB, double GainCompression_dB, 
		   double OutputPowerTarget)
{
   double GainTargetLinear = dB2Linear(GainTarget_dB);
   double GainCompressionLinear = dB2Linear(GainCompression_dB);

   //## If the amplifier was not saturated, set it now
   SetTypeAmplifier(VECTOR_SATURATED);
   UnsaturatedGainLinear =  GainTargetLinear * GainCompressionLinear;  
   UnsaturatedDifferentialGainLinear = log(UnsaturatedGainLinear)/AmpLength;   
   //////SaturatingPowerLinear = OutputPowerTarget/GainTargetLinear
   //////                       *(UnsaturatedGainLinear - 2)
   //////			  /(2*log(2));   
   SaturatingPowerLinear = (GainTargetLinear - 1.)/GainTargetLinear
                           *OutputPowerTarget
			   /log(UnsaturatedGainLinear/GainTargetLinear);
			   
   //cout << "GainTargetLinear = " << GainTargetLinear << endl;
   //cout << "GainCompressionLinear = " << GainCompressionLinear << endl;
   //cout << "   exp(UnsaturatedDifferentialGainLinear) = " 
   //     << exp(UnsaturatedDifferentialGainLinear) << endl;  
   //cout << "   SaturatingPowerLinear = " << SaturatingPowerLinear << endl;
}		   
