// $Id:
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsReduOptFiber.hh" // ## In order to access the friend methods

extern ofstream *LogFile;

// ################### Methods for the class ReduOptFiber                    

// #####################################################################

// ##### Initializes ReduOptFiber
// ## By: Ivan Lima

/*!

\param string InFileName: The name of the input parameter file.
\param ReduOptSignal *oReduOptSignalTmp: Pointer to the reduced optical
        signal object that we are going to propagate through the fiber.
\param RanNumGen *oRanNumGenTmp: Random number generator object used
        to obtain the random realizations of the fiber PMD


The important parameters read in from the input parameter file are

-# LengthFiber: The length of the fiber in meters
-# qtStepsFiber: The number of sections in the fiber. 
-# PMD_Fiber: The PMD of the fiber in ps/sqrt(km). 
-# MeanDGD_SqrtLength: The mean differential group delay per square root
   length in ps/sqrt(km). 
-# AttenuationFiber: The fiber loss in dB/meter.


You should set either PMD_Fiber or MeanDGD_SqrtLength. The code
will use whichever of these is non-zero in the input parameter file.

The constructor also allocated memory for the Euler angles and 
computes derived parameters.

\sa LengthFiber, qtStepsFiber, PMD_Fiber, MeanDGD_SqrtLength, oRanNumGen,
    oReduOptSignal.

 */


ReduOptFiber::ReduOptFiber( string InFileName,
                            ReduOptSignal *oReduOptSignalTmp,
                            RanNumGen *oRanNumGenTmp) 
                                                
{
   oReduOptSignal = oReduOptSignalTmp;
   CenterFreq = oReduOptSignal->GetCenterFreq();
   oRanNumGen = oRanNumGenTmp;    
   NumChannels = oReduOptSignal->GetNumChannels();
   
  ifstream InFile(InFileName.c_str(),ios::in);    
  if (!InFile) 
  {
     cerr << "Error: Input file "<< InFileName 
          << " could not be opened."<< endl;
     exit(1);
  }
     
  LogFileSeparator();
  *LogFile << InFileName << endl << endl;


   LengthFiber = ReadDouble("LengthFiber","m",0,-1,
                                    LOWER_ONLY,&InFile);

   qtStepsFiber = ReadInt("qtStepsFiber"," ",0,-1,
                              LOWER_ONLY,&InFile);

   PMD_Fiber = 0;
   MeanDGD_SqrtLength= 0;

   PMD_Fiber =  ReadDouble("PMD_Fiber","ps/sqrt(km)",0,-1,
   			LOWER_ONLY,&InFile); 

   MeanDGD_SqrtLength =  ReadDouble("MeanDGD_SqrtLength","ps/sqrt(km)",0,-1,
   			LOWER_ONLY,&InFile);                         

   if (PMD_Fiber != 0 && MeanDGD_SqrtLength != 0) {
      *LogFile << "OptFiber constructor:" << endl
              << "Either include MeanDGD_SqrtLength or "
              << "PMD_Fiber in the input file."<< endl
              << "Be aware that: " << endl
              << " MeanDGD_SqrtLength = sqrt(8/(3*pi))*PMD_Fiber" << endl
              << " It is justified by the fact that DGD is "<< endl
              << " a  Maxwellian distributed random variable"
              << endl;
      cout << endl<< "Error: abortingng now. "
           << " See ocs.log for details" << endl;
      exit(1);
   }


   AttenuationFiber = ReadDouble("AttenuationFiber","dB/m",0,-1,
			LOWER_ONLY,&InFile);

   // ## Converts from base 10 to the
   // #  Nepperian base: e = 2.718...  
   NepperAttenuationFiber = AttenuationFiber*log(10.0)/10.0; 

   InFile.close();
      
   //MemAllocated = 1;

   phi     = new double[qtStepsFiber];
   psi     = new double[qtStepsFiber];   
   theta   = new double[qtStepsFiber]; 
   StepLength = new double[qtStepsFiber];  

   GotFiberSample = 0;
   
   DeltaZ = LengthFiber/qtStepsFiber;

   // JZ: BUG FIXED on August 18th 2002.
   // DeltaZ needs to be initialized before you call 
   // SetMeanDGD_SqrtLength and  SetPMD_Fiber.
   // This was not the case before. 

   if (MeanDGD_SqrtLength != 0)
       SetMeanDGD_SqrtLength(MeanDGD_SqrtLength);
   else //## PMD_Fiber might have been != 0 in the input
       SetPMD_Fiber(PMD_Fiber);

   TypeScattering = 1; // ## Euler Matrix (Default)

   // ## Sets equal section lengths as default

   // JZ 18 August 2002. NOTE: For the PMD rotations to be correct
   // we need constant step sizes. 

   for (int ii = 0; ii < qtStepsFiber; ii++)
      StepLength[ii] = DeltaZ;

   LogFileSeparator();

   *LogFile << "Processing Fiber Parameters" << endl << endl;

   // ## Set derived parameters 

   *LogFile <<  "LengthFiber [km] = " << LengthFiber*1e-3 << endl;
   *LogFile <<  "qtStepsFiber = " << qtStepsFiber << endl;
   *LogFile <<  "AttenuationFiber [dB/km] = " <<  AttenuationFiber*1e3 << endl;
   *LogFile <<  "Center Freq of Signal [GHz] = " << CenterFreq*1e-9 << endl;

   *LogFile <<  "PMD_Fiber [ps/sqrt(km)] = " << PMD_Fiber*1e12*sqrt(1e3) 
           << endl;

   *LogFile << "PMD_Fiber [s/sqrt(m)] = "<< PMD_Fiber << endl;  
   *LogFile << "MeanDGD_SqrtLength = " << MeanDGD_SqrtLength << endl;
   *LogFile << " -> MeanDGD = " << MeanDGD_SqrtLength*sqrt(LengthFiber) << endl;
   *LogFile << "   Note that:  MeanDGD_SqrtLength = sqrt(8/(3*pi))*PMD_Fiber" 
           << endl;   

   *LogFile << "Default fiber step Size DeltaZ [m] = " 
            << DeltaZ << endl;         

}// ## end constructor
 
// ################################################################

// ## Copy constructor
ReduOptFiber::ReduOptFiber
           (const ReduOptFiber &oReduOptFiber)
{
   CopyReduOptFiber(oReduOptFiber);
}

// ###############################################################

// ## Deep copy constructor
void ReduOptFiber::CopyReduOptFiber
           (const ReduOptFiber &oReduOptFiber)
{
   oReduOptSignal = oReduOptFiber.oReduOptSignal;
   CenterFreq = oReduOptFiber.CenterFreq;
   oRanNumGen = oReduOptFiber.oRanNumGen;    
   NumChannels = oReduOptFiber.NumChannels;

   LengthFiber = oReduOptFiber.LengthFiber;

   qtStepsFiber = oReduOptFiber.qtStepsFiber;
   PMD_Fiber = oReduOptFiber.PMD_Fiber;

   AttenuationFiber = oReduOptFiber.AttenuationFiber;
  
   // ## Converts from base 10 to the
   // #  Nepperian base: e = 2.718...  
   NepperAttenuationFiber = AttenuationFiber*log(10.0)/10.0; 
    
   AllocateMemory_SetLengthSteps();
    
   //MemAllocated = 1;

   //phi     = new double[qtStepsFiber];
   //psi     = new double[qtStepsFiber];   
   //theta   = new double[qtStepsFiber]; 
   //StepLength = new double[qtStepsFiber];  

   //GotFiberSample = 0;
   
   //DeltaZ = LengthFiber/qtStepsFiber;
   //dBeta_2dOmega = PMD_Fiber/sqrt(4*DeltaZ); 
   TypeScattering = 1; // ## Euler Matrix (Default)

   // ## Sets equal section lengths as default
   //for (int ii = 0; ii < qtStepsFiber; ii++)
   //   StepLength[ii] = DeltaZ;

   LogFileSeparator();

   *LogFile << "OptFiber object initialized using copy "
           << "constructor" << endl;
        
   *LogFile << "Processing Fiber Parameters" << endl << endl;

   *LogFile <<  "LengthFiber [km] = " << LengthFiber*1e-3 << endl;
   *LogFile <<  "qtStepsFiber = " << qtStepsFiber << endl;
   *LogFile <<  "AttenuationFiber [dB/km] = " <<  AttenuationFiber*1e3 << endl;
   *LogFile <<  "Center Freq of Signal [GHz] = " << CenterFreq*1e-9 << endl;

   *LogFile <<  "PMD_Fiber [ps/sqrt(km)] = " << PMD_Fiber*1e12*sqrt(1e3) << endl;

   *LogFile << "Default fiber step Size DeltaZ [m] = " 
            << DeltaZ << endl;         
   
} // ## End of copy constructor

// ##################################################################

// ## Allocates the memory
// # By: Ivan Lima

void ReduOptFiber::AllocateMemory_SetLengthSteps(void)
{
    
   DeltaZ = LengthFiber/qtStepsFiber;
   dBeta_2dOmega = PMD_Fiber/sqrt(4*DeltaZ); 

   phi     = new double[qtStepsFiber];
   psi     = new double[qtStepsFiber];   
   theta   = new double[qtStepsFiber]; 
   StepLength = new double[qtStepsFiber];  
         
   GotFiberSample = 0;

   for (int ii = 0; ii < qtStepsFiber; ii++)
      StepLength[ii] = DeltaZ;
}

// ##################################################################

// ## Change the length and the number of steps
// #  It has improved modularity compared to the old one
// ## By: John Zweck and Ivan Lima 

void ReduOptFiber::SetLengthFiberAndQtStepsFiber(double Length,int numSteps)
{
   LengthFiber = Length;
   qtStepsFiber = numSteps;
   ReleaseMemory();
   AllocateMemory_SetLengthSteps();  
}

// ################################################################

// ##### Destructor: releases the allocated memory
// ## By: Ivan Lima
ReduOptFiber::~ReduOptFiber()
{
   ReleaseMemory();
}

// ##############################################################

// ##### Specialized function to release memory
// ## By: Ivan Lima
void ReduOptFiber::ReleaseMemory()
{
   delete [] phi;
   delete [] psi;
   delete [] theta;
   delete [] StepLength;  
  
}

// ###############################################################

// ## By: Ivan Lima. Adapted by JZ on August 19th 2002 to get agreement
// ## with Full Model when only have one section in a fiber

void ReduOptFiber::SetMeanDGD_SqrtLength(double MeanDGD_SqrtLength2)
{
   // ## Attention: it is expect to be in [ps/sqrt(km)]

   MeanDGD_SqrtLength = 1e-12/sqrt(1e3)*MeanDGD_SqrtLength2;
   //PMD_Fiber = sqrt(3.*pi/8.)*MeanDGD_SqrtLength;
   PMD_Fiber = (1.0/RMS2meanDGD())*MeanDGD_SqrtLength; 
   dBeta_2dOmega = PMD_Fiber/sqrt(4*DeltaZ);    
}

// ###############################################################

// ## By: Ivan Lima. Adapted by JZ on August 19th 2002 to get agreement
// ## with Full Model when only have one section in a fiber

void ReduOptFiber::SetPMD_Fiber(double PMD_Fiber2)
{
   // ## Attention: it is expect to be in [ps/sqrt(km)]
   PMD_Fiber = 1e-12/sqrt(1e3)*PMD_Fiber2;
   //MeanDGD_SqrtLength = sqrt(8./(3.*pi))*PMD_Fiber;
   MeanDGD_SqrtLength = RMS2meanDGD()*PMD_Fiber;
   dBeta_2dOmega = PMD_Fiber/sqrt(4*DeltaZ);    
}

// ################################################################

// ## Copied from OptFiber::rms2meanDGD by JZ on August 19th 2002 
// ## to get agreement
// ## with Full Model when only have one section in a fiber

double ReduOptFiber::RMS2meanDGD(void)
{
   double RMS2meanDGD;
   if (qtStepsFiber != 1) {
      RMS2meanDGD = sqrt(8./(3.*pi));
   }   
   else {
      RMS2meanDGD = 1;         
   }   

   return RMS2meanDGD;      
}

// ##########################################################

void ReduOptFiber::SetAttenuation(double AttenuationFiber2)
{

  // Written by John Zweck, Nov 2nd, 2000

AttenuationFiber = AttenuationFiber2;
NepperAttenuationFiber = AttenuationFiber*log(10.0)/10.0; 

 *LogFile << "WARNING: Reset AttenuationFiber to " 
         << AttenuationFiber << endl;

}

// ################################################################

// ## Scatters the polarization uniformly on the poincare sphere
// ## By: Ivan Lima

/*! This method must be called by the user in an application program.
    It is called after the fiber constructor and before the propagation
    method. It should be called each time the user wants a new random
    realization of the fiber.
 
    In the coarse-step method we divide the fiber into sections
    each of about 1 km in length. Before each section we randomly
    rotate the Stokes vectors on the Poincare sphere. This rotation
    is equivalent to a random rotation of the axes of birefringence
    of the fiber. The rotations are defined in terms of the
    three Euler angles theta,phi,psi. The purpose of this method
    is to randomly select the Euler angles for the rotations.

    As explained in [WM 2001] page 488 the Euler angles are randomly
    chosen so that we get a random rotation chosen from 
    a uniform probability distribution on the Poincare sphere. 

    The number of fiber sections is given by the class parameter
    qtStepsFiber. The Euler angles are stored in the arrays 
    theta* phi* and psi*

   \sa qtStepsFiber, theta,phi,psi \n
       
 */

void ReduOptFiber::GetFiberSampleCompleteScatt()
{

   TypeScattering = 1; // ## Euler Matrix   
   GotFiberSample = 1;   
   for (int ii = 0; ii < qtStepsFiber; ii++) 
     {
      if (PMD_Fiber != 0) 
       { // ## CNLS -> Manakov-PMD
         // ## Generate the Euler angles            
         cvtheta   = 2*(oRanNumGen->GetRanNum()-0.5);                        
         theta[ii] = acos(cvtheta);                     
         phi[ii]   = 2*pi*oRanNumGen->GetRanNum();          
         psi[ii]   = 2*pi*oRanNumGen->GetRanNum();            
     
       }        
      else 
       {  // ## No PMD_Fiber -> no mode coupling due to birefringence
         theta[ii] = 0.;                     
         phi[ii]   = 0.;          
         psi[ii]   = 0.; 
       }      
      
     }   // end for      

} // ## end GetFiberSampleCompleteScatt()

// ################################################################

// ## Forces the scattering angles for special user purposes
// ## By: Ivan Lima

void ReduOptFiber::SetScatteringMatrix( int IndexStep, double Theta 
                                   ,double Phi, double Psi)
{
  if(qtStepsFiber == 1 && IndexStep == 0)
    {
     GotFiberSample = 1; 
     TypeScattering = 1;
    }

   theta[IndexStep] = Theta;
   phi[IndexStep]   = Phi;
   psi[IndexStep]   = Psi;
}


// ## Propagates the optical signal through the fiber
// ## By: Ivan Lima

/*!
\param double *PropagatedLengthFiber: Pointer to a double which keeps
       track of the length the signal has propagated.

You should call the constructor and  void GetFiberSampleCompleteScatt(void)   
before clling this method. 

Uses the coarse step method to model a fiber with PMD with random
rotations given by the Euler angles.

Acts on the Stokes vectors of the signal and noise in each 
channel as in Equations (3,4a,4b) of [WM 2001] page 488.

Also includes fiber loss which is lumped at the end of the 
fiber. Depending on how the amplification is done you may wish
to set the loss parameter to zero.

\sa void GetFiberSampleCompleteScatt(void)   

 */


void ReduOptFiber::PropagateFiberCoupledNLS(double *PropagatedLengthFiber)
{

   if ( ! GotFiberSample) 
    {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
    }   
   
   double v2Delta;
   double Freq;  
   
   double StokesSignalShifted[3];
   double StokesNoiseShifted[3];

   // ## Start the propagation
   
   //## The extra unpolarized noise component
   //#  just suffers the attenuation
   // ## Adds lumped losses, since there isn't any 
   // #  non-linear effect   

   //oReduOptSignal->StokesExtraUnpolChannNoise
   //   *= exp( -LengthFiber*NepperAttenuationFiber);      
      
 
   for (int jj = 0; jj < NumChannels; jj++) 
    {
      for (int ii = 1; ii < 4; ii++) 
       {
         StokesSignalShifted[ii-1] 
                = oReduOptSignal->StokesSignal[ii][jj];
         StokesNoiseShifted[ii-1]  
                = oReduOptSignal->StokesNoise[ii][jj];
      }


        // The following frequency is the absolute frequency.

	 Freq = oReduOptSignal->GetFrequency(jj); 

      for ( int ii = 0; ii < qtStepsFiber; ii++) 
      { 
        
         //PhaseBirref_2  = 2*pi*StepLength[ii]*dBeta_2dOmega;

	 // The following formula was checked by JZ on
	 // August 17 2002 to be in exact agreement with the
	 // Full Model PMD rotation. If we choose a CW signal
	 // for the full model and the dBeta_2dOmega of the
	 // full and reduced models agree exactly and the
	 // step sizes are the same and we only have the 
	 // PMD rotation in the full model, then we get
	 // exact agreement with the full model.

         v2Delta = 2*dBeta_2dOmega*StepLength[ii]*2*pi*Freq; 

	 // Note: The rotations do not all have the same handedness.
	 // See ocsTools.cc for more information.
	 
         switch (TypeScattering) 
          {
            case 1: // ## RxRyRx (Euler Matrix)
               RotatesAboutX( phi[ii],   StokesSignalShifted); 
               RotatesAboutY(-theta[ii], StokesSignalShifted); 
               RotatesAboutX( psi[ii],   StokesSignalShifted); 

               RotatesAboutX( phi[ii],   StokesNoiseShifted); 
               RotatesAboutY(-theta[ii], StokesNoiseShifted); 
               RotatesAboutX( psi[ii],   StokesNoiseShifted); 
	       break;
            case 2: // ## RyRzRy      
               RotatesAboutY( -psi[ii],   StokesSignalShifted); 
               RotatesAboutZ(  theta[ii], StokesSignalShifted); 
               RotatesAboutY( -phi[ii],   StokesSignalShifted); 

               RotatesAboutY( -psi[ii],   StokesNoiseShifted); 
               RotatesAboutZ(  theta[ii], StokesNoiseShifted); 
               RotatesAboutY( -phi[ii],   StokesNoiseShifted); 
               break;
            case 3: // ## RxRzRy      
               RotatesAboutY( -psi[ii],   StokesSignalShifted); 
               RotatesAboutZ(  theta[ii], StokesSignalShifted); 
               RotatesAboutX(  phi[ii],   StokesSignalShifted); 

               RotatesAboutY( -psi[ii],   StokesNoiseShifted); 
               RotatesAboutZ(  theta[ii], StokesNoiseShifted); 
               RotatesAboutX(  phi[ii],   StokesNoiseShifted); 
               break;
           default: //## Rotation matrix inexistent
               cout << "From ReduOptFiber->PropagateFiberCoulpledNLS(.)" 
                    << endl << "TypeScattering = " << TypeScattering
                    << " inexistent." << endl;    
         }         

         RotatesAboutX( v2Delta,  StokesSignalShifted); // ## Propagation   
         RotatesAboutX( v2Delta,  StokesNoiseShifted); // ## Propagation
               
      }

      for (int ii = 1; ii < 4; ii++) 
       {
         oReduOptSignal->StokesSignal[ii][jj] 
                     = StokesSignalShifted[ii-1];
         oReduOptSignal->StokesNoise[ii][jj]  
                     = StokesNoiseShifted[ii-1];
      }

      // ## Adds lumped losses, since there isn't any 
      // #  non-linear effect

      for (int ii = 0; ii < 4; ii++) 
       {
         //cout << StokesSignal[ii] << "  ";
         oReduOptSignal->StokesSignal[ii][jj] 
             *= exp( -LengthFiber*NepperAttenuationFiber);
         oReduOptSignal->StokesNoise[ii][jj]  
             *= exp( -LengthFiber*NepperAttenuationFiber);
         //cout << StokesSignal[ii] << endl;  
      }
   }   
 
  *PropagatedLengthFiber += LengthFiber;

} // ## end Propagate...

// ############################################################

// ## Syncronize the number of channels with OptSignal.
// #  By: Ivan Lima
void ReduOptFiber::SyncNumChannels(void)
{
   NumChannels = oReduOptSignal->GetNumChannels();
}

// ########################################################

// ## Set the lengths of the steps
// ## By: Ivan Lima

void ReduOptFiber::SetStepLengths(double *usrStepLength)
{  
   // ## usrStepLength[] should have the size == qtStepsFiber

   double TotLengthSteps = 0;

   for (int ii = 0; ii < qtStepsFiber; ii++) 
    {
      StepLength[ii]  = usrStepLength[ii];
      TotLengthSteps += StepLength[ii];   
    }

   if(TotLengthSteps != LengthFiber) 
    {
      cerr << "Warning: the sum of the steps: " << TotLengthSteps << endl
           << " is different from the Length of the Fiber: " 
           << LengthFiber << "." << endl
           << "Thus, the length of the fiber will be redefined" << endl;
      LengthFiber = TotLengthSteps;
   }       
}

// ##########################################################

// ## Choose random lengths for the step
// ## By: Ivan Lima

void ReduOptFiber::GetRandomStepLengths(double RelativeRangeLengths)
{  
    // ## The range of the lengths is relative to the mean length: DeltaZ

   double *SplicingNode = new double[qtStepsFiber+1];  

   SplicingNode[qtStepsFiber] = LengthFiber;

   for (int ii = 0; ii < qtStepsFiber; ii++) 
     {
      SplicingNode[ii] = DeltaZ*( ii+(oRanNumGen->GetRanNum()-0.5)
                         *RelativeRangeLengths);
     }
   
   for (int ii = 0; ii < qtStepsFiber; ii++)
      { 
        StepLength[ii] = SplicingNode[ii+1] - SplicingNode[ii];                
      }

   delete [] SplicingNode;  
   
}

// ##########################################################

// ## Create a file with the lengths
// ## By: Ivan Lima

void ReduOptFiber::WriteFileSteps(char *outFile)
{
   ofstream OutFile(outFile,ios::out);
   if (!OutFile) {
      cerr << "Error: " << outFile << "could not be created."<<endl;
      exit(1);
   }
   int ii;
   for (ii = 0; ii < qtStepsFiber; ii++) 
      OutFile << ii << "   " << StepLength[ii]   << endl;   
   OutFile.close();
}

// ##########################################################

// ## Gets the DGD = length of the PD vector
// ## By: Ivan Lima

double ReduOptFiber::GetDGD(double Freq)
{
   double OutPD_Vector[] = {0,0,0};
   GetOutPD_Vector(Freq, OutPD_Vector);
   return(sqrt(sq(OutPD_Vector[0])+sq(OutPD_Vector[1])
                +sq(OutPD_Vector[2])) ); 
}

// #######################################################

// #### Gets the input Pol. Disp. Vector of the fiber
// ## By: Ivan Lima

/*! The polarization dispersion vector (PDV) at the input to this
    fiber is the Stokes vector that when transformed by the random realization
    of the fiber maps to the output PDV.
    
    \sa void GetOutPD_Vector((double Freq, double *OutPD_Vector)
*/


void ReduOptFiber::GetInPD_Vector(double Freq, 
                                  double *InPD_Vector) 

// ## Freq should be the actual frequency -> CenterFreq + FreqSignal
                               
// ## InPD_Vector should have dimension 3: from 0 to 2                      
      
// ## The initial value of InPD_Vector will be used as the 
//    value of the Input PDV.  "pdispv4.mws"

{

   if ( ! GotFiberSample) {
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }        

   double v2Delta;
   
   for( int ii = qtStepsFiber - 1; ii >= 0; ii--) 
    {  
      v2Delta = 2*dBeta_2dOmega*StepLength[ii]*2*pi*Freq;   
      InPD_Vector[0] += -2*dBeta_2dOmega*StepLength[ii];    
      RotatesAboutX( -v2Delta,  InPD_Vector); 
                                  
      switch (TypeScattering) {      
         case 2:  // ## RyRzRy
            RotatesAboutY(   phi[ii],   InPD_Vector); 
            RotatesAboutZ( -theta[ii],  InPD_Vector); 
            RotatesAboutY(   psi[ii],   InPD_Vector); 
            break;
         case 3: // ## RxRzRy      
            RotatesAboutX( -phi[ii],    InPD_Vector); 
            RotatesAboutZ( -theta[ii],  InPD_Vector); 
            RotatesAboutY(  psi[ii],    InPD_Vector); 
            break;
         default: // ## == 1 (default): Euler Matrix: RxRyRx
            RotatesAboutX( -psi[ii],  InPD_Vector); 
            RotatesAboutY(  theta[ii],InPD_Vector); 
            RotatesAboutX( -phi[ii],  InPD_Vector); 
      }         
   }   

   // ## Test lines. It is just OK.
   //double OutPD_Vector[3];
   //for (int ii = 0; ii < 3; ii++) 
   //   OutPD_Vector[ii] = InPD_Vector[ii];
   //for (int ii = 0; ii < qtStepsFiber; ii++) {
   //   if (TypeScattering == 2) { // ## RzRy      
   //      RotatesAboutY( -psi[ii],   OutPD_Vector); 
   //      RotatesAboutZ( theta[ii],  OutPD_Vector); 
   //      RotatesAboutY( -phi[ii],   OutPD_Vector); 
   //   }
   //   else {  // ## == 1 (default): Euler Matrix
   //      RotatesAboutX( phi[ii],  OutPD_Vector); 
   //      RotatesAboutY(-theta[ii],OutPD_Vector); 
   //      RotatesAboutX( psi[ii],  OutPD_Vector); 
   //   }               
   //   RotatesAboutX( v2Delta,  OutPD_Vector); 
   //}
   //   cout << "Propagating the Input Pol.Dist.Vector forward:" << endl;
   //   cout << OutPD_Vector[0] << "   " 
   //        << OutPD_Vector[1] << "   "
   //        << OutPD_Vector[2] << endl;             
}


// ########################################################


// #### Gets the output Pol. Disp. Vector of the fiber
// ## By: Ivan Lima

/*!

\param Freq: The absolute frequency at which we wish to compute
             the polarization dispersion vector (PDV). 
\param OutPD_Vector: A pointer to an array of length 3 which stores
                     the PDV at output. We assume that memory for this array
                     has been allocated by the user or by another 
                     method in this class.


To get the PDV vector for a system (or single fiber) 
you should set OutPD_Vector = {0,0,0} before calling this function
for the first fiber in a system. Then you need to call the function
once for each time the signal is passed through the fiber. 
The answer you get will of course depend on the random realization
of the fiber that you choose.

This method is called by \sa double GetDGD(double Freq) 
      
The method we use to compute the PDV is described in Section II of
the paper
"Comparison of Polarization  Mode Dispersion
    Emulators" by Ivan Lima et al. which I think appeared in JLT
 in about 2001.

 */


void ReduOptFiber::GetOutPD_Vector(double Freq, double *OutPD_Vector)
{

// ## Freq should be the actual frequency -> CenterFreq + FreqSignal
                               
// ## OutPD_Vector should have dimension 3: from 0 to 2
                            
// ## The initial value of OutPD_Vector will be used as the 
//    initial value of the Output PDV. "pdispv4.mws"


 if ( ! GotFiberSample) 
  {   
      cerr << "Error: Object of class OptFiber not initialized" 
           << endl;
      cerr << "       or optical fiber not generated." << endl;      
      exit(1);
   }        

   double v2Delta; 
  
   for( int ii = 0; ii < qtStepsFiber; ii++) 
    {
      v2Delta = 2*dBeta_2dOmega*StepLength[ii]*2*pi*Freq;                  

      switch (TypeScattering) 
       {
         case 2: // ## RyRzRy      
            RotatesAboutY( -psi[ii],   OutPD_Vector); 
            RotatesAboutZ(  theta[ii], OutPD_Vector); 
            RotatesAboutY( -phi[ii],   OutPD_Vector); 
            break;

         case 3: // ## RxRzRy      
            RotatesAboutY( -psi[ii],   OutPD_Vector); 
            RotatesAboutZ(  theta[ii], OutPD_Vector); 
            RotatesAboutX(  phi[ii],   OutPD_Vector); 
            break;

         default: // ## == 1 (default): Euler Matrix
            RotatesAboutX( phi[ii],  OutPD_Vector); 
            RotatesAboutY(-theta[ii],OutPD_Vector); 
            RotatesAboutX( psi[ii],  OutPD_Vector); 
      }         

      RotatesAboutX( v2Delta,  OutPD_Vector);    // ## Propagation    

      OutPD_Vector[0] += -2*dBeta_2dOmega*StepLength[ii];           

      //cout << ii<<" "<< dBeta_2dOmega <<"  " << StepLength[ii] << endl;      
   }                 
}

// ################################################################

// #### Gets the second order output Pol. Disp. Vector of the fiber
// ## It calculates the derivative numerically#

// ## Warning: You should set the vector to zero before 
// #  obtaining the value for the first fiber
// ## By: Ivan Lima

/*! Obtained by taking a numerical derivative of the output PDV 
    with respect to frequency where the frequency step is given
    by the inverse of the expected DGD.

    \sa double GetExpectedDGD(void),\n  
        void GetOutPD_Vector(double Freq,double *OutPD_Vector)

    \param Freq: The absolute frequency at which we wish to compute
             the second order polarization dispersion vector (PDV2) at output.
 
    \param OutPD_Vector: A pointer to an array of length 3 which stores
                     the PDV2 at output. We assume that memory for this array
                     has been allocated by the user or by another 
                     method in this class.\n
                     You should set this vector to zero before  
                     obtaining the value for the first fiber

*/

void ReduOptFiber::GetSndOrderOutPD_Vector(double Freq,
                                           double *SndOrderOutPD_Vector)
{
   double Freq2 = Freq + 1e-3/GetExpectedDGD();
   double OutPD_Vector[] = {0,0,0};
   double OutPD_Vector2[] = {0,0,0};
   GetOutPD_Vector(Freq,  OutPD_Vector);
   GetOutPD_Vector(Freq2, OutPD_Vector2);
   
   //cout << " Freq = " << Freq << "   Freq2 = " << Freq2 << endl;   
   //cout <<"OutPD_Vector @ Freq = " << OutPD_Vector[0] <<"  "
   //                      << OutPD_Vector[1] <<"  "
   //                     << OutPD_Vector[2]<< endl;                             
   //cout <<"OutPD_Vector @ Freq2 = " << OutPD_Vector2[0] <<"  "
   //                      << OutPD_Vector2[1] <<"  "
   //                     << OutPD_Vector2[2]<< endl;
                          
   for (int ii = 0; ii < 3; ii++)
      SndOrderOutPD_Vector[ii] += (OutPD_Vector2[ii]
                                 -OutPD_Vector[ii])
                                 /(2*pi*(Freq2-Freq));
}                                  

// ############################################################


// ## Gets the Snd Order DGD = length of the Snd Order PD vector
// ## By: Ivan Lima
double ReduOptFiber::GetSndOrderDGD(double Freq)
{
   double SndOrderOutPD_Vector[] = {0,0,0};
   GetSndOrderOutPD_Vector(Freq, SndOrderOutPD_Vector);
   return(sqrt(sq(SndOrderOutPD_Vector[0])
              +sq(SndOrderOutPD_Vector[1])
              +sq(SndOrderOutPD_Vector[2])) ); 
}

// ##############################################################





