// $Id: 
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsReduOptSignal.hh" 

extern ofstream *LogFile;

// ################### Methods for the class ReduOptSignal
// # This class initializes the optical signal and the noise 
// # for the reduced model, only in its Stokes representation,
// # the optical signal will be used throughout the other
// # objects of the reduced model.

// ##### Constructor of the reduced optical signal
// ## By: Ivan Lima
// #  Part of this code was adapted from 
// #  John Zweck's constructor of OptSignal.

/*!
    The constructor reads in parameters from a file and initializes:
    -# The central frequency of the simulation frequency window
    -# The frequencies of each WDM channel
    -# The power and Stokes parameters of the signal and noise in each
       channel
    -# The type of pulse and efficiency factor of the receiver that is
       used to compute the electrical SNR from the optical SNR.
    -# The parameters SignalNoiseBandwidth and PolarizNoiseBandwidth.

   
       \param InFileName2: The name of the input parameter file (string)
       \param SignalNoiseBandwidth2: Used to initialize the class parameter
              SignalNoiseBandwidth
       \param PolarizNoiseBandwidth2: Used to initialize the class parameter
              PolarizNoiseBandwidth

        The code requires that

        SignalNoiseBandwidth <= PolarizNoiseBandwidth <= ChannSpacing

        If PolarizNoiseBandwidth2 == 0
        then PolarizNoiseBandwidth = SignalNoiseBandwidth

        If PolarizNoiseBandwidth2 < 0
        then PolarizNoiseBandwidth = ChannSpacing.
        
       \sa ReduOptSignal(string InFileName,double SignalNoiseBandwidth2),
           RegenerateReduOptSignal,
           ~ReduOptSignal ProcessReduOptSignalParameters
 */

ReduOptSignal::ReduOptSignal(string InFileName2, 
                             double SignalNoiseBandwidth2)
{
    SignalNoiseBandwidth = SignalNoiseBandwidth2;      

    InFileName = InFileName2;
    ifstream InFile(InFileName.c_str(),ios::in);    
   
    if (!InFile) 
    {
       cout << "Error: Input file "<< InFileName 
            << " could not be opened."<< endl;
       exit(1);
    }

    TypePulse =  typePulse(ReadInt("TypePulse"," ",1,5,
                           LOWER_AND_UPPER,&InFile));

   CenterFreq = 0;
   CenterWavelength = 0;

   CenterFreq =  ReadDouble("CenterFreq","Hz"
                     ,0,-1,LOWER_ONLY,&InFile);
   CenterWavelength = ReadDouble("CenterWavelength","m",0
                     ,-1,LOWER_ONLY,&InFile);

   if(CenterFreq == 0)
   {
      CenterFreq = Wavelength2Frequency(CenterWavelength);
  
      *LogFile << endl 
              << "Since CenterFreq is zero we use CenterWavelength:" << endl
              << endl
              << "CenterWavelength [nm] = " << CenterWavelength*1e9 << endl
              << "CenterFreq [GHz] = " << CenterFreq*1e-9 << endl << endl;
   }
   else
   if(CenterWavelength == 0)
   {
      CenterWavelength = Frequency2Wavelength(CenterFreq);
  
      *LogFile << endl << "Since CenterWavelength is zero we use CenterFreq:" 
              << endl << endl 
              << "CenterWavelength [nm] = " << CenterWavelength*1e9 << endl
              << "CenterFreq [GHz] = " << CenterFreq*1e-9 << endl << endl;
   }
   else
   {
      *LogFile << endl 
              << "ERROR: Make one of CenterFreq and CenterWavelength zero!" 
	      << endl << "The code will then use the other value." 
	      << endl <<  endl << "ABORTING NOW" << endl;

      exit(1);

   }
       
   ChannSpacing =  ReadDouble("ChannSpacing","Hz",0,-1,LOWER_ONLY,&InFile);
   
   if(SignalNoiseBandwidth > ChannSpacing) 
    {
      *LogFile << "Error: The SignalNoiseBandwidth is lagrer " 
              << "than the ChannSpacing. Please, fix it." <<endl;
      cout << "Error: See *.log for details." << endl;        
      exit(1);
   }

   if(SignalNoiseBandwidth < 0) 
   {
      *LogFile << "Error: The SignalNoiseBandwidth " 
              << "cannot be negative. Please, fix it." <<endl;
      cout << "Error: See *.log for details." << endl;        
      exit(1);
   }   
 
   BitRateChann =  ReadDouble("BitRateChann","bit/s",0,-1,LOWER_ONLY,&InFile);
   
   OptPowerChann = ReadDouble("OptPowerChann","W",0,-1,LOWER_ONLY,&InFile);
   NumChannels =   ReadInt("NumChannels"," ",0,-1,LOWER_ONLY,&InFile);
   

   *LogFile << endl;
   *LogFile << "SIGNAL PARAMETERS" << endl << endl;
   *LogFile << "Pulse Type = " << WriteTypePulse() << endl << endl;  

   // Really out of date now

   EnhancementFactor =  ReadDouble("EnhancementFactor","",0,-1,
                                   LOWER_ONLY,OPTIONAL_NO_WARNING,-1,
				   &InFile);   
          
   *LogFile << "BitRateChann [Gb/s] = " << BitRateChann*1e-9 << endl;
   *LogFile << "OptPowerChann [mW] = " << OptPowerChann*1e3 << endl;
   *LogFile << "Warning : OptPowerChann represents the average" << endl;
   *LogFile << "Power for the reduced model."<< endl << endl;
   *LogFile << "NumChannels = " << NumChannels << endl;
   *LogFile << "Center Frequency [GHz] = " << CenterFreq*1e-9  << endl;
   *LogFile << "Center Wavelength [nm] = " << CenterWavelength*1e9  << endl;
   *LogFile << "Channel Spacing [GHz] = " << ChannSpacing*1e-9  << endl;
 
   AllocateMemory();

   ProcessReduOptSignalParameters(&InFile);
   RegenerateReduOptSignal(); 
   // ## These variables are used to calculate the Q-factor
   OptFilterBandwidth = 0;
   ElecFilterBandwidth = 0;
   
   InFile.close();    

} //## End of ReduOptSignal()

// #####################################################

//! JZAug02

ReduOptSignal::ReduOptSignal(string InFileName2, string JobName)
{
    InFileName = InFileName2;

    ifstream InFile(InFileName.c_str(),ios::in);    
   
    if (!InFile) 
    {
       cout << "Error: Input file "<< InFileName 
            << " could not be opened."<< endl;
       exit(1);
    }

    TypePulse =  typePulse(ReadInt("TypePulse"," ",1,5,
                           LOWER_AND_UPPER,&InFile));

   CenterFreq = 0;
   CenterWavelength = 0;

   CenterFreq =  ReadDouble("CenterFreq","Hz"
                     ,0,-1,LOWER_ONLY,&InFile);
   CenterWavelength = ReadDouble("CenterWavelength","m",0
                     ,-1,LOWER_ONLY,&InFile);

   if(CenterFreq == 0)
   {
      CenterFreq = Wavelength2Frequency(CenterWavelength);
  
      *LogFile << endl 
              << "Since CenterFreq is zero we use CenterWavelength:" << endl
              << endl
              << "CenterWavelength [nm] = " << CenterWavelength*1e9 << endl
              << "CenterFreq [GHz] = " << CenterFreq*1e-9 << endl << endl;
   }
   else
   if(CenterWavelength == 0)
   {
      CenterWavelength = Frequency2Wavelength(CenterFreq);
  
      *LogFile << endl << "Since CenterWavelength is zero we use CenterFreq:" 
              << endl << endl 
              << "CenterWavelength [nm] = " << CenterWavelength*1e9 << endl
              << "CenterFreq [GHz] = " << CenterFreq*1e-9 << endl << endl;
   }
   else
   {
      *LogFile << endl 
              << "ERROR: Make one of CenterFreq and CenterWavelength zero!" 
	      << endl << "The code will then use the other value." 
	      << endl <<  endl << "ABORTING NOW" << endl;

      exit(1);

   }
       

  BitRateChann =  ReadDouble("BitRateChann","bit/s",0,-1,LOWER_ONLY,&InFile);

  EnhancementFactor =  ReadDouble("EnhancementFactor","",0,-1,
                                         LOWER_ONLY,&InFile);   

   NumChannels =   ReadInt("NumChannels"," ",0,-1,LOWER_ONLY,&InFile);

   AllocateMemory();

   string FrequencyDataInFileName = ReadString("FrequencyDataInFileName",
					       &InFile);  

   InFile.close();

   ReadFileFreqPower(FrequencyDataInFileName);


   SignalNoiseBandwidth = ChannSpacing; // TEMP

   // JZAug02: Temporary Stokes stuff

   for( int jj = 0; jj < NumChannels; jj++) 
    {
     SetInStokes(jj,0,1.0);
     SetInStokes(jj,1,1.0);
     SetInStokes(jj,2,0.0);
     SetInStokes(jj,3,0.0);
    }

 

   RegenerateReduOptSignalVaryingPower(); 

   WriteMatlabFileOptSignal(JobName+"fInOpt.dat");

   // ## These variables are used to calculate the Q-factor

   OptFilterBandwidth = 0;
   ElecFilterBandwidth = 0;
          
   *LogFile << endl;
   *LogFile << "SIGNAL PARAMETERS" << endl << endl;
   *LogFile << "Pulse Type = " << WriteTypePulse() << endl << endl;  
 
   *LogFile << "BitRateChann [Gb/s] = " << BitRateChann*1e-9 << endl;
  
   *LogFile << "NumChannels = " << NumChannels << endl;
   *LogFile << "Center Frequency [GHz] = " << CenterFreq*1e-9  << endl;
   *LogFile << "Center Wavelength [nm] = " << CenterWavelength*1e9  << endl;
   *LogFile << "Channel Spacing [GHz] = " << ChannSpacing*1e-9  << endl;


} //## End of ReduOptSignal()

// #######################################################

/*!

In the input parameter file the channels are labelled 1,2,3,.., NumChannels.
Internally in the code the channels are indexed by ChannelArrayIndex
which takes values 0,1,2,...,NumChannels-1. \n

In the input parameter file if the flag
$EvenlySpacedFreqsFlag is set to 1 then the channels are evenly
spaced in frequency, with the channel spacing determined by the parameter
$ChannSpacing. As I understand it code is only guarenteed to work
when EvenlySpacedFreqsFlag=1.\n

Example A: NumChannels = 5\n

If ChannSpacing is 100GHz and CenterFreq = 193THz
then channels are\n

InputFileIndex     1      2     3      4      5\n
ChannelArrayIndex  0      1     2      3       4\n
Frequencies        192.8, 192.9 193.0, 193.1, 193.2\n

In general for an odd number of channels:\n

CenterFreq has index (NumChanns+1)/2.\n

Example B: NumChannels = 6\n

If ChannSpacing is 100GHz and CenterFreq = 193THz
then channels are\n

InputFileIndex     1      2      3     4      5      6\n
ChannelArrayIndex  0      1     2      3      4      5\n
Frequencies        192.7  192.8, 192.9 193.0, 193.1, 193.2\n

In general for an even number of channels:\n

CenterFreq has also index (NumChanns+1)/2.\n

If EvenlySpacedChannsFlag is set to 0 then the 
wavelengths of the channels are read in using\n

$WavelengthChann1       1550e-9         # m \n
$WavelengthChann2       1552e-9             \n
$WavelengthChann3       1553e-9             \n

etc.

*/

void ReduOptSignal::ProcessReduOptSignalParameters(ifstream *InFile)
{

   int EvenlySpacedFreqsFlag =  ReadInt("EvenlySpacedFreqsFlag","",0,1,
                                   LOWER_AND_UPPER,InFile);

   if ( ! EvenlySpacedFreqsFlag) 
    {
  
      char C_String[40];
      string WavelengthChannName;

      double minFreq;
      double maxFreq;      

      for(int ch=0;ch<NumChannels;ch++) 
       {
         sprintf(C_String,"WavelengthChann%d",ch+1);
         WavelengthChannName = C_String;

	 // These are absolute frequencies

         Frequency[ch] =  Wavelength2Frequency(
                                   ReadDouble(WavelengthChannName,"m",0,-1,
                                   LOWER_ONLY,InFile));

	 // Compute the min and max frequencies

         if (ch==0)  
          {
            minFreq = Frequency[ch];
            maxFreq = Frequency[ch];
	  }
         else 
          {
            if(minFreq > Frequency[ch])
               minFreq = Frequency[ch];

            if(maxFreq < Frequency[ch])
               maxFreq = Frequency[ch];
         }
         

      } // end for ch loop

      LogFileSeparator();

      for(int ch=0;ch<NumChannels;ch++)
         *LogFile << "Frequency[" << ch << "] = " <<  Frequency[ch] << endl;


      *LogFile << endl << endl;

      *LogFile << "Min frequency of channels [GHz] = " 
              << minFreq*1e-9 << endl;
      *LogFile << "Max frequency of channels [GHz] = " 
                    << maxFreq*1e-9 << endl;

   } // ## end if(!EvenlySpacedFreqFlag)      
   

   if(EvenlySpacedFreqsFlag) {

      LogFileSeparator();

      *LogFile << "Evenly spaced channels" << endl << endl;

      for(int ch = 0;ch<NumChannels;ch++)
      {
         Frequency[ch] = CenterFreq 
               + GetChannelPhysicalIndex(ch)*ChannSpacing;

         *LogFile << "Frequency in channel "
                 << GetChannelPhysicalIndex(ch) << " = "
                 <<  Frequency[ch]*1e-9 << " GHz" << endl;
  
      }


      *LogFile << endl << endl;

      *LogFile << "Min frequency of channels [GHz] = " 
              << Frequency[0]*1e-9 << endl;
      *LogFile << "Max frequency of channels [GHz] = " 
              << Frequency[NumChannels-1]*1e-9 << endl;
   
   } // ## end if EvenlySpacedFreqsFlag   


   // ## Read and Process the stokes parameters

   InitializeInStokes(InFile);   
   
   // ## The Stokes Parameters are normalized by S0
   // #  The procedure check for problems in input.

   for( int jj = 0; jj < NumChannels; jj++) 
    {
      if ( GetInStokes(jj,0)*(1+ 2*NEAR_ZERO) < 
          sqrt( sq(GetInStokes(jj,1)) 
               +sq(GetInStokes(jj,2))
               +sq(GetInStokes(jj,3)) ) ) 
       {
        cout << "From: ReduOptSignal::ReduOptSignal()" << endl 
             << "Error: The input Stokes para. "
             << "of the channel # " << jj 
             << " has DOP > 1." << endl
             << "Check  the file " << InFileName
             << endl;
        exit(1);
       }
      
      for( int ii = 1; ii < 4; ii++) 
       {
         SetInStokes(jj,ii, GetInStokes(jj,ii) 
                           /GetInStokes(jj,0) );
       }

        SetInStokes(jj,0,1.);

   }
} // ## End of ProcessInputStokesParameters();

// #####################################################

// ## Allocate the memory
// #  By: Ivan Lima
void ReduOptSignal::AllocateMemory(void)
{

   Frequency = new double[NumChannels];
   InStokes = new double[NumChannels*4];

      for (int ii = 0; ii < 4; ii++) 
       {
         StokesSignal[ii] = new double[NumChannels];
         StokesNoise[ii]  = new double[NumChannels];         
       }      

 OptPowerChannArray = new double[NumChannels];
}



// #####################################################

//! Copy Constructor
//! By John Zweck, Jan 2006

ReduOptSignal::ReduOptSignal(ReduOptSignal & OrigReduOptSignal)
{
   CenterFreq      = OrigReduOptSignal.CenterFreq;
   CenterWavelength = OrigReduOptSignal.CenterWavelength;
   BitRateChann    = OrigReduOptSignal.BitRateChann;
   NumChannels = OrigReduOptSignal.NumChannels;
   ChannSpacing = OrigReduOptSignal.ChannSpacing;   
   SignalNoiseBandwidth = OrigReduOptSignal.SignalNoiseBandwidth;

   OptFilterBandwidth = OrigReduOptSignal.OptFilterBandwidth;   
   ElecFilterBandwidth = OrigReduOptSignal.ElecFilterBandwidth;   
   EnhancementFactor = OrigReduOptSignal.EnhancementFactor;   

   TypePulse = OrigReduOptSignal.TypePulse;
   OptPowerChann = OrigReduOptSignal.OptPowerChann;
   InFileName = OrigReduOptSignal.InFileName;

   AllocateMemory();

 //## Set the Stokes vector for each channel
   for(int ch=0;ch<NumChannels;ch++) {
      for(int ss = 0; ss < 4; ss++) {
        SetInStokes(ch,ss,OrigReduOptSignal.GetInStokes(ch,ss));
      }
   }      	

   //## Get the central frequency of each channel     
   for(int ch=0;ch<NumChannels;ch++) {
     Frequency[ch] = OrigReduOptSignal.Frequency[ch];   
   }     
   
   for(int ch=0;ch<NumChannels;ch++) 
    OptPowerChannArray[ch] = OrigReduOptSignal.OptPowerChannArray[ch];   
      
   //## Get the data
   for(int ch=0;ch<NumChannels;ch++) 
     {
      for(int ss = 0; ss < 4; ss++) 
	{
         StokesSignal[ss][ch] = OrigReduOptSignal.StokesSignal[ss][ch];
	 StokesNoise[ss][ch]  = OrigReduOptSignal.StokesNoise[ss][ch];
	}
     }   

}

// #####################################################

//! Assignment operator
//!  By: Ivan Lima (9/24/02)

ReduOptSignal & ReduOptSignal::operator=(ReduOptSignal & RHS_ReduOptSignal)
{


   if( NumChannels != RHS_ReduOptSignal.NumChannels) {
       cout <<  NumChannels << " " << RHS_ReduOptSignal.NumChannels << endl;

       *LogFile << "Invalid use of ReduOptSignal::operator=" << endl
                << "Aborting Now." << endl << flush;

       cout << "Invalid use of ReduOptSignal::operator=" << endl
                << "Aborting Now." << endl;

       exit(1);

   }

   CenterFreq      = RHS_ReduOptSignal.CenterFreq;
   CenterWavelength = RHS_ReduOptSignal.CenterWavelength;
   BitRateChann    = RHS_ReduOptSignal.BitRateChann;
   NumChannels = RHS_ReduOptSignal.NumChannels;
   ChannSpacing = RHS_ReduOptSignal.ChannSpacing;   
   SignalNoiseBandwidth = RHS_ReduOptSignal.SignalNoiseBandwidth;

   OptFilterBandwidth = RHS_ReduOptSignal.OptFilterBandwidth;   
   ElecFilterBandwidth = RHS_ReduOptSignal.ElecFilterBandwidth;   
   EnhancementFactor = RHS_ReduOptSignal.EnhancementFactor;   

   //## Set the Stokes vector for each channel
   for(int ch=0;ch<NumChannels;ch++) {
      for(int ss = 0; ss < 4; ss++) {
        SetInStokes(ch,ss,RHS_ReduOptSignal.GetInStokes(ch,ss));
      }
   }      	

   //## Get the central frequency of each channel     
   for(int ch=0;ch<NumChannels;ch++) {
     Frequency[ch] = RHS_ReduOptSignal.Frequency[ch];   
   }     
   
   
   //## Get the data
   for(int ch=0;ch<NumChannels;ch++) {
      for(int ss = 0; ss < 4; ss++) {
	//cout << "StokesNoise Before = " << StokesNoise[ss][ch] << endl;
	//cout << "StokesSignal Before = " << StokesSignal[ss][ch] << endl;
         StokesSignal[ss][ch] = RHS_ReduOptSignal.StokesSignal[ss][ch];
	 StokesNoise[ss][ch]  = RHS_ReduOptSignal.StokesNoise[ss][ch];
         //cout << "StokesNoise After = " << StokesNoise[ss][ch] << endl;
         //cout << "StokesSignal After = " << StokesSignal[ss][ch] << endl;

      }
   }   

   return *this;

}// end assignment operator (=)

// #############################################################

  //! The noise in ReduOptSignal1 is added incoherently to the
  //! current ReduOptSignal. For each channel, the Stokes vectors of the noise 
  //! in ReduOptSignal1 are added to those of the current ReduOptSignal


void ReduOptSignal::AddReducedNoise(ReduOptSignal * ReduOptSignal1)
{

if( NumChannels != ReduOptSignal1->GetNumChannels()) 
   {
       cerr << "Invalid use of ReduOptSignal::AddReducedNoise" 
	    << endl
            << "Aborting." << endl;
       exit(1);

   }

  for(int ch=0;ch<NumChannels;ch++)
   for(int ss = 0; ss < 4; ss++)
    StokesNoise[ss][ch] += ReduOptSignal1->StokesNoise[ss][ch]; 

}

// ##############################################################

//! JZAug02

void ReduOptSignal::ReadFileFreqPower(string FrequencyDataInFileName)
{
  
  ifstream DataInFile(FrequencyDataInFileName.c_str(),ios::in);    
   
  if(!DataInFile) 
    {
       cout << "Error: Input file "<< FrequencyDataInFileName 
            << " could not be opened."<< endl;
       exit(1);
    }

  for(int jj = 0;jj< NumChannels;jj++)
    {
      DataInFile >> Frequency[jj] 
	         >> OptPowerChannArray[jj];

      Frequency[jj] += CenterFreq;

    }
   
  ChannSpacing = abs(Frequency[0]-Frequency[1]); 

  DataInFile.close();

} // ## end ReduOptSignal::ReadFileFreqPower

//#########################################################

// ## Change the number of channels on the fly
// #  This function was created to allow the analyzis 
// #  of the outage as a function of the NumChannels
// #  Warning: All the other objects have to be 
// #  readjusted in their NumChannel
// ## Attention: You should not be using this unless 
// ## you really want to make a channel sweep.
// #  By: Ivan Lima

/*! This function was created to compute outage probability
    as a function of the NumChannels. You should not be using this method
    unless you really want to perform a sweep across different values
    of the parameter NumChannels.
*/

void ReduOptSignal::SetNumChannels(int NumChannels2)
{
   if (NumChannels2 < 1) {
      cerr << "From: ReduOptSignal::SetNumChannels(.)" << endl
           << "The number of channels must be >=1." << endl
           << "The value received was = " << NumChannels2 << endl;
      exit(1);
   }        
   NumChannels = NumChannels2;
   ifstream InFile(InFileName.c_str(),ios::in);    
   if (!InFile) 
   {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
   }
   ReleaseMemory();
   AllocateMemory();
   ProcessReduOptSignalParameters(&InFile);
   RegenerateReduOptSignal();
   InFile.close();
}

//#########################################

// ## Obtain the degree of polarization
// #  By: Ivan Lima
double ReduOptSignal::GetDegreeOfPolarization()
{
   double StokesTotal[4];
/****   
   for (int ii = 0; ii < 4; ii++) {
      StokesTotal[ii] = 0;    
   }
   StokesTotal[0] += StokesExtraUnpolChannNoise;  //## Extra Noise
   for (int jj = 0; jj < NumChannels; jj++) {
      for (int ii = 0; ii < 4; ii++) {
         //cout << ii << " " << jj << " -> "
         //     << oReduOptSignal->StokesSignal[ii][jj]<<endl;
         StokesTotal[ii] += StokesSignal[ii][jj] 
                           +StokesNoise[ii][jj];                     
      }           
   }   
****/
   GetTotalStokes(StokesTotal);   
   return sqrt( sq(StokesTotal[1]) + sq(StokesTotal[2])
               +sq(StokesTotal[3]) )/StokesTotal[0] ;
}

// ##############################################################

// ## Get the total Stokes parameters of the light
// #  By: Ivan Lima

/*!
    \param StokesTotal: The total Stokes parameters, computed by this method. 
                        (A double array of length 4)

    The unnormalized total Stokes parameters are defined as follows.\n
    StokesTotal[0] = Sum of Powers of the signal and Noise in each
                     channel as well as the power in the Unpolarized noise.\n

    The total Stokes vector is\n
    (StokesTotal[1],StokesTotal[2],StokesTotal[3]) = 
      Sum of the Stokes vectors of the signal and noise in each channel.\n

    These unnormalized total Stokes parameters are then normalized
    by dividing by StokesTotal[0]. 

    \sa GetTotalPower 

 */

void ReduOptSignal::GetTotalStokes( double *StokesTotal)
{

   for (int ii = 0; ii < 4; ii++) 
    {
      StokesTotal[ii] = 0;    
    }

   //StokesTotal[0] += StokesExtraUnpolChannNoise;  //## Extra Noise

   for (int jj = 0; jj < NumChannels; jj++) 
    {
      for (int ii = 0; ii < 4; ii++) 
        {
         
         StokesTotal[ii] += StokesSignal[ii][jj] 
                           +StokesNoise[ii][jj];                     
        }           
    } 

   if (StokesTotal[0] != 0) 
     {  
      for (int ii = 1; ii < 4; ii++) {
         StokesTotal[ii] /= StokesTotal[0];
      }

      StokesTotal[0] = 1;

   }   

} // ## end GetTotalStokes

// ############################################################

// ## Gets the total power in system
// #  By: Ivan Lima

/*!
TotalPower = Sum of Powers of the signal and Noise in each
             channel as well as the power in the Unpolarized noise.\n

\sa GetTotalStokes
*/

double ReduOptSignal::GetTotalPower(void)
{
   double TotalPower = 0;     

   //TotalPower += StokesExtraUnpolChannNoise; //## Extra Noise

   for (int jj = 0; jj < NumChannels; jj++) 
      TotalPower += StokesSignal[0][jj] 
                   +StokesNoise[0][jj];           

   return TotalPower;
}

// ########################################################

// ## Gets the power of the signal in the system
// #  It means that the noise is not included
// #  By: Ivan Lima

/*!
SignalPower = Sum of Powers of the signal in each channel.\n

So noise power is not included!

\sa GetTotalPower
   
 */

double ReduOptSignal::GetSignalPower(void)
{
   double SignalPower = 0;     
   for (int jj = 0; jj < NumChannels; jj++) 
      SignalPower += StokesSignal[0][jj];           
   return SignalPower;
}

//######################################################

/*!

   Called by the constructors and also before each Monte Carlo
   experiment. 
   -# Sets the Stokes 4-vector of the signal in each channel
      to be the input Stokes 4-vector multiplied by OptPowerChann.
   -# Sets the Stokes vector of the noise in each channel to be zero.
   -# Sets the extra Unpolarized noise outside the channels to zero.

  \sa  InitializeInStokes, ReduOptSignal(string InFileName, 
                    double SignalNoiseBandwidth2,
                    double PolarizNoiseBandwidth2).

 */

// ##### Regenerate the reduced optical signal
// ## By: Ivan Lima

void ReduOptSignal::RegenerateReduOptSignal(void)
{
   // ## Initializes the reduced signal parameters

   for( int jj = 0; jj < NumChannels; jj++) 
    {
      for( int ii = 0; ii < 4; ii++) 
       {
         StokesSignal[ii][jj] = GetInStokes(jj,ii)*OptPowerChann;
         StokesNoise[ii][jj] = 0.0;                              
 
      }
   }
} // ## end RegenerateReduOptSignal

// ###################################################

/*!

   Special purpose method to allow for different optical power
   levels in the different channels

   \param OptPowerChannArray: double array of size NumChanns used
          to store the power for each channel

   -# Sets the Stokes 4-vector of the signal in each channel
      to be the input Stokes 4-vector multiplied by the optical power
      of the channel
   -# Sets the Stokes vector of the noise in each channel to be zero.
   -# Sets the extra unpolarized noise outside the channels to zero.

  \sa  InitializeInStokes, ReduOptSignal(string InFileName, 
                    double SignalNoiseBandwidth2).

 */

void ReduOptSignal::RegenerateReduOptSignal(double * OptPowerChannArray2)
{
   // ## Initializes the reduced signal parameters

   for( int jj = 0; jj < NumChannels; jj++) 
    {
      for( int ii = 0; ii < 4; ii++) 
       {
         StokesSignal[ii][jj] = GetInStokes(jj,ii)*OptPowerChannArray2[jj];
         StokesNoise[ii][jj] = 0.0;                              
 
      }
   }
} // ## end RegenerateReduOptSignal

// #########################################################

// JZAug02

void ReduOptSignal::RegenerateReduOptSignalVaryingPower(void)
{
   // ## Initializes the reduced signal parameters


   for( int jj = 0; jj < NumChannels; jj++) 
    {
      for( int ii = 0; ii < 4; ii++) 
       {
         StokesSignal[ii][jj] = GetInStokes(jj,ii)*OptPowerChannArray[jj];
         StokesNoise[ii][jj] = 0.0;                              
 
      }
   }
} // ## end RegenerateReduOptSignalVaryingPower

//###########################################################

/*! Sets the three Stokes parameters of the noise in each channel
    to zero, thus depolarizing the noise. 
    The power of the noise and the StokesExtraUnpolChannNoise power
    are not changed by this method.
*/

// ##### Depolarize the Noise
// ## By: Ivan Lima (01/06/19)

void ReduOptSignal::DepolarizeReduNoise(void)
{
   //## Only the S0 component of the noise 
   //#  and the UnpolExtraNoise will remain   
   for( int jj = 0; jj < NumChannels; jj++) {
      for( int ii = 1; ii < 4; ii++) {
         StokesNoise[ii][jj] = 0.;                                      
      }
   }
}

//##########################################################

// ##### Destructor of the reduced optical signal
// ## By: Ivan Lima

/*!
\sa ReleaseMemory()
 */

ReduOptSignal::~ReduOptSignal()
{
   ReleaseMemory();
}

// ###################################################

// ##### Release the memory
// ## By: Ivan Lima

/*!
   Frees memory previosuly allocated for the Stokes parameters
   of the signal and noise
   in each channel and for the array of central frequencies for the
   channels.
    \sa ~ReduOptSignal()
   */

void ReduOptSignal::ReleaseMemory(void)
{
   delete [] Frequency;
   delete [] InStokes;
   for (int ii = 0; ii < 4; ii++) {
      delete [] StokesSignal[ii];
      delete [] StokesNoise[ii];         
   }   
}

// ##################################################

// ##### Initialize the filters in order to be able to 
// #     calculate the Q-factor for the reduced model
// ## By: Ivan Lima

 /*! These bandwidths are required to compute the Q-factor
      in the reduced model from the electrical signal-to-noise ratio
      using the formula of Marcuse.

\sa  
void SetOptElecFilterBandwidth( string FileNameOptFilter,
                                string FileNameElecFilter)

 */

void ReduOptSignal::SetOptElecFilterBandwidth
                   ( double OptFilterBandwidthTmp
                    ,double ElecFilterBandwidthTmp)
{
   OptFilterBandwidth = OptFilterBandwidthTmp;
   ElecFilterBandwidth = ElecFilterBandwidthTmp;
   *LogFile << "From: ReduOptSignal::SetOptElecFilterBandwidth(double,double)"
           << endl
           << "OptFilterBandwidth  = "<< OptFilterBandwidth << endl
           << "ElecFilterBandwidth = "<< ElecFilterBandwidth << endl;
              
}                   

// ##############################################################

// ##### Initialize the filters in order to be able to 
// #     calculate the Q-factor for the reduced model
// #  The data are extracted from the files
// ## By: Ivan Lima

 /*! These bandwidths are required to compute the Q-factor
      in the reduced model from the electrical signal-to-noise ratio
      using the formula of Marcuse.\n
      The bandwidths are read in from input parameter files.\n

      The bandwidth of the optical filter is set using the 
      input file parameter FreqFWHM_OptFilter.\n

      The bandwidth of the electrical filter is set using the 
      input file parameter f_3dB.\n

      The electrical filter is a low-pass filter.
      The input value, f_3dB [GHz] =
      0.5*FreqFWHM_ElecFilter*1e-9 
      is the HALF width at half maximum of the filter.\n
	   
      This means that the amplitude of the 
      filter transfer function has height 1/sqrt(2) at f_3dB.\n
	     
  */

void ReduOptSignal::SetOptElecFilterBandwidth
                   ( string FileNameOptFilter
                    ,string FileNameElecFilter)
{
   ifstream InFileOptFilter(FileNameOptFilter.c_str(),ios::in);    
   
   if (!InFileOptFilter) 
    {
      cerr << "Error: Input file "<< InFileOptFilter 
           << " could not be opened."<< endl;
      exit(1);
    }

   ifstream InFileElecFilter(FileNameElecFilter.c_str(),ios::in);    
   
   if (!InFileElecFilter) 
    {
      cerr << "Error: Input file "<< InFileElecFilter 
           << " could not be opened."<< endl;
      exit(1);
    }
    
   OptFilterBandwidth = ReadDouble("FreqFWHM_OptFilter","Hz",0,-1,
                                    LOWER_ONLY,&InFileOptFilter);

   ElecFilterBandwidth = ReadDouble("f_3dB","Hz",0,-1,
                             LOWER_ONLY,MANDATORY,&InFileElecFilter);

   *LogFile << "From: ReduOptSignal::SetOptElecFilterBandwidth(string,string)"
           << endl
           << "OptFilterBandwidth  = "<< OptFilterBandwidth << endl
           << "ElecFilterBandwidth = "<< ElecFilterBandwidth << endl;                    
   InFileOptFilter.close();
   InFileElecFilter.close();

} // ## end SetOptElecFilterBandwidth

// #########################################################

//##### Calculate the Q-factor for the reduced model
//## By: Ivan Lima

/*!
    The Q factor of a channel is computed from the electrical
    signal-to-noise ratio using the formula of Marcuse.\n

    See Marcuse paper which is referenced in UMBC's report to
    SAIC (June 2001, page 9, equation (2)).

\sa GetElecTimeDomainSNR(int ChannelArrayIndex)

 */

double ReduOptSignal::GetQ_Factor(int ChannelArrayIndex)
 {
   double SNR = GetElecTimeDomainSNR(ChannelArrayIndex);
           
   return( SNR/(sqrt(2*SNR +1 ) +1 )    
          *sqrt( 2*OptFilterBandwidth
                 /ElecFilterBandwidth)   );
} 
 

// ###########################################################
 
//## Get the electrical time domain SNR
//#  It is an approximation, since the reduced model
//#  only contains the reduced frequency domain information
//#  on the signal and the noise.
//## By: Ivan Lima (7/8/01)

/*!
    Computes the  electrical signal-to-noise ratio (ESNR) of the channel
    with index ChannelArrayIndex. The ESNR is usd to compute the 
    Q-factor of the channel using the formula of Marcuse.\n
    
    The ESNR is defined to be the efficiency factor of the 
    receiver multiplied by the OSNR. The OSNR is the ratio of the
    signal power in the channel to the noise power in the channel.
    The noise power in the channel is defined to be the total noise
    in a band of size OptFilterBandwidth about the channel.\n

    The efficiency factor is the ratio of the peak power in a pulse
    to the average power of the pulse and is given by the value
    of  EnhancementFactor.\n

    An explanation of why we use this definition can be found in 
    UMBC's report to SAIC (June 2001, pages 9-10). A more up to date
    explanation can be found in Curtis Menyuk's book chapter on 
    polarization effects in WDM systems, which UMBC will provide
    to SAIC upon request.

    \sa OptFilterBandwidth, EnhancementFactor,
     GetNoiseInsideBandwidth(int ChannelArrayIndex,double OptFilterBandwidth),
     GetQ_Factor(int ChannelArrayIndex)

*/

double ReduOptSignal::GetElecTimeDomainSNR(int ChannelArrayIndex)
{
   if (   ElecFilterBandwidth == 0 || OptFilterBandwidth == 0) 
    {
      cout << "From: ReduOptSignal::ElecTimeDomainSNR()"<<endl 
           << "Error: the bandwidth of the filters were"
           << "not initialized." << endl
           << "Please, check the source code."
           << endl;
      exit(1);   
    }
   
   if ( GetNoiseInsideChannel(ChannelArrayIndex) == 0) 
    {
      cout << "From: ReduOptSignal::GetElecTimeDomainSNR()"<<endl 
           << "Error: There is no noise in the channel: "
           << ChannelArrayIndex<< endl
           << "The total number of channels is "
           <<  NumChannels << endl
           << "Please, check the source code."
           << endl;
      exit(1);   
   }   


   //# EnhancementFactor is the enhancement factor of the receiver.
 
  return EnhancementFactor
               *GetSignal(ChannelArrayIndex)
               /GetNoiseInsideBandwidth(ChannelArrayIndex,OptFilterBandwidth);

}// ## end GetElecTimeDomainSNR

// #############################################################

//## Get the optical SNR
//## By: Ivan Lima (7/8/01)

/*
The OSNR computed by this method is the ratio of the
    signal power in the channel to the noise power in the channel.\n
    The noise power in the channel is computed using 
    double GetNoiseInsideChannel(int ChannelArrayIndex).\n
    
QUESTION: Why are the ESNR and OSNR computed using different
          bandwidths for the noise? \n

\sa double GetNoiseInsideChannel(int ChannelArrayIndex)
\sa double GetElecTimeDomainSNR(int ChannelArrayIndex)
*/

double ReduOptSignal::GetOptSNR(int ChannelArrayIndex)
{
   if ( GetNoiseInsideChannel(ChannelArrayIndex) == 0) {     
      cout << "From: ReduOptSignal::GetQ_Factor(...)"<<endl 
           << "Error: There is no noise in the channel: "
           << ChannelArrayIndex<< endl
           //<< "The channel physical index associated to it is "
           //<< ChannelPhysicalIndex << endl
           << "The total number of channels is "
           <<  NumChannels << endl
           << "Please, check the source code."
           << endl;
      exit(1);   
   }   

   return GetSignal(ChannelArrayIndex)
         /GetNoiseInsideChannel(ChannelArrayIndex);

} // ## end GetOptSNR

// ##################################################################

// ##### Get the noise within the bandwidth of the signal
// ## By: Ivan Lima (01/06/19)

/*!
   In this method the noise is the total noise in a band of size
   SignalNoiseBandwidth about the channel

\sa SignalNoiseBandwidth, 
    double GetNoiseInsideBandwidth(int ChannelArrayIndex,
                                   double SignalNoiseBandwidth)
   
 */

double ReduOptSignal::GetNoiseInsideChannel(int ChannelArrayIndex)
{

   GetNoiseInsideBandwidth(ChannelArrayIndex,
                           SignalNoiseBandwidth);
        
}


double ReduOptSignal::GetCoPolarizedNoiseInsideChannel(int ChannelArrayIndex)
{

   GetCoPolarizedNoiseInsideBandwidth(ChannelArrayIndex,
                           SignalNoiseBandwidth);
        
}


double ReduOptSignal::GetOrthogPolarizedNoiseInsideChannel(int ChannelArrayIndex)
{

   GetOrthogPolarizedNoiseInsideBandwidth(ChannelArrayIndex,
                           SignalNoiseBandwidth);
        
}

// ########################################################



//##### Get the noise within any arbitrary bandwidth
//##    This method gets the total noise in the given bandwidth
//## By: Ivan Lima (7/8/19)
double ReduOptSignal::GetNoiseInsideBandwidth(int ChannelArrayIndex,
                                              double BandwidthNoise )
{   
   double Noise;
/*   
   if (BandwidthNoise <= PolarizNoiseBandwidth) {
      Noise = StokesNoise[0][ChannelArrayIndex]
             *BandwidthNoise/PolarizNoiseBandwidth; 
   }
   else if (   BandwidthNoise > PolarizNoiseBandwidth 
            && BandwidthNoise <= ChannSpacing) {
      Noise = StokesNoise[0][ChannelArrayIndex]          
             +StokesExtraUnpolChannNoise/NumChannels
             *(BandwidthNoise - PolarizNoiseBandwidth)
             /(ChannSpacing - PolarizNoiseBandwidth);
   }          
*/   
   if (BandwidthNoise <= ChannSpacing) {
      Noise = StokesNoise[0][ChannelArrayIndex]
             *BandwidthNoise/ChannSpacing; 
   }
   else {
      cout << "From: ReduOptSignal::GetNoiseInsideBandwidth"<< endl
           << "The bandwith: " << BandwidthNoise << " is larger"
           << "than the channel spacing."<< endl;
      exit(1);   
   }

cout << Noise  << endl;
   
   return Noise;    
           
}




//##### Get the copolarized noise within any arbitrary bandwidth
//## It is copolarized with respect to the polarized part of the noise
//## By: Ivan Lima (7/8/19)
double ReduOptSignal::GetCoPolarizedNoiseInsideBandwidth(int ChannelArrayIndex,
                                              double BandwidthNoise )
{ 
   double PolarizedNoise = sqrt( sq(StokesNoise[1][ChannelArrayIndex])
                                +sq(StokesNoise[2][ChannelArrayIndex])
                                +sq(StokesNoise[3][ChannelArrayIndex]) );   
		
   double UnpolarizedNoise = StokesNoise[0][ChannelArrayIndex]
                            -PolarizedNoise; 					   
   double CoPolarizedNoise;   
   /*
   if (BandwidthNoise <= PolarizNoiseBandwidth) {
      CoPolarizedNoise = (PolarizedNoise + UnpolarizedNoise/2.)
             *BandwidthNoise/PolarizNoiseBandwidth; 
   }
   else if (   BandwidthNoise > PolarizNoiseBandwidth 
            && BandwidthNoise <= ChannSpacing) {
      CoPolarizedNoise = PolarizedNoise + UnpolarizedNoise/2.
             +StokesExtraUnpolChannNoise/NumChannels/2.
             *(BandwidthNoise - PolarizNoiseBandwidth)
             /(ChannSpacing - PolarizNoiseBandwidth);
   } 
   */
   if (  BandwidthNoise <= ChannSpacing) {
      CoPolarizedNoise = (PolarizedNoise + UnpolarizedNoise/2.)
             *BandwidthNoise/ChannSpacing; 
   }             
   else {
      cout << "From: ReduOptSignal::GetNoiseInsideBandwidth"<< endl
           << "The bandwith: " << BandwidthNoise << " is larger"
           << "than the channel spacing."<< endl;
      exit(1);   
   }
   
   return CoPolarizedNoise;               
}




/*
// ##### Get the copolarized noise within any arbitrary bandwidth
// ## By: Ivan Lima (7/8/19)
double ReduOptSignal::GetCoPolarizedNoiseInsideBandwidth(int ChannelArrayIndex,
                                              double BandwidthNoise )
{
   //## It is being assumed in the beggining that
   //#  SignalNoiseBandwidth <= PolarizNoiseBandwidth
   
   double PolarizedNoise = sqrt( sq(StokesNoise[1][ChannelArrayIndex])
                                +sq(StokesNoise[2][ChannelArrayIndex])
                                +sq(StokesNoise[3][ChannelArrayIndex]) );   
   double NoisePolarization[] = {0,0,0,0};
   if (PolarizedNoise != 0) {
      NoisePolarization[1] = StokesNoise[1][ChannelArrayIndex]
                            /PolarizedNoise;
      NoisePolarization[2] = StokesNoise[2][ChannelArrayIndex]
                            /PolarizedNoise;
      NoisePolarization[3] = StokesNoise[3][ChannelArrayIndex]
                            /PolarizedNoise;			       
   }				
   double UnpolarizedNoise = StokesNoise[0][ChannelArrayIndex]
                            -PolarizedNoise; 
					    
   double PolarizedSignal = sqrt( sq(StokesSignal[1][ChannelArrayIndex])
                                +sq(StokesSignal[2][ChannelArrayIndex])
                                +sq(StokesSignal[3][ChannelArrayIndex]) );   
   double SignalPolarization[] = {0,0,0,0};
   if (PolarizedSignal != 0) {
      SignalPolarization[1] = StokesSignal[1][ChannelArrayIndex]
                            /PolarizedSignal;
      SignalPolarization[2] = StokesSignal[2][ChannelArrayIndex]
                            /PolarizedSignal;
      SignalPolarization[3] = StokesSignal[3][ChannelArrayIndex]
                            /PolarizedSignal;			       
   }				
   double ScalarProdSignalNoise =  
                   SignalPolarization[1] * NoisePolarization[1]
		  +SignalPolarization[2] * NoisePolarization[2]
		  +SignalPolarization[3] * NoisePolarization[3];
   //cout << "ScalarProdSignalNoise = " << ScalarProdSignalNoise << endl;
					    
   double CoPolarizedNoiseCoeff = 0.5 
                *( ScalarProdSignalNoise + 1)
		*PolarizedNoise/(PolarizedNoise + UnpolarizedNoise) 
	      +0.5*( UnpolarizedNoise/(PolarizedNoise + UnpolarizedNoise) );   
   
   double CoPolarizedNoise;   
   if (BandwidthNoise <= PolarizNoiseBandwidth) {
      CoPolarizedNoise = StokesNoise[0][ChannelArrayIndex]
             *CoPolarizedNoiseCoeff
             *BandwidthNoise/PolarizNoiseBandwidth; 
   }
   else if (   BandwidthNoise > PolarizNoiseBandwidth 
            && BandwidthNoise <= ChannSpacing) {
      CoPolarizedNoise = StokesNoise[0][ChannelArrayIndex]          
             *CoPolarizedNoiseCoeff
             +StokesExtraUnpolChannNoise/NumChannels
             *(BandwidthNoise - PolarizNoiseBandwidth)
             /(ChannSpacing - PolarizNoiseBandwidth);
   }          
   else {
      cout << "From: ReduOptSignal::GetNoiseInsideBandwidth"<< endl
           << "The bandwith: " << BandwidthNoise << " is larger"
           << "than the channel spacing."<< endl;
      exit(1);   
   }
   
   return CoPolarizedNoise;               
}
*/



// ##### Get the OrthogPolarized noise within any arbitrary bandwidth
// ## By: Ivan Lima (7/8/19)
double ReduOptSignal::GetOrthogPolarizedNoiseInsideBandwidth(int ChannelArrayIndex,
                                              double BandwidthNoise )
{
   //## It is being assumed in the beggining that
   //#  SignalNoiseBandwidth <= PolarizNoiseBandwidth
   
   double PolarizedNoise = sqrt( sq(StokesNoise[1][ChannelArrayIndex])
                                +sq(StokesNoise[2][ChannelArrayIndex])
                                +sq(StokesNoise[3][ChannelArrayIndex]) );   		
   double UnpolarizedNoise = StokesNoise[0][ChannelArrayIndex]
                            -PolarizedNoise; 

   double OrthogPolarizedNoise; 
   /*  
   if (BandwidthNoise <= PolarizNoiseBandwidth) {
      OrthogPolarizedNoise = UnpolarizedNoise/2.
             *BandwidthNoise/PolarizNoiseBandwidth; 
   }
   else if (   BandwidthNoise > PolarizNoiseBandwidth 
            && BandwidthNoise <= ChannSpacing) {
      OrthogPolarizedNoise = UnpolarizedNoise/2.
             +StokesExtraUnpolChannNoise/NumChannels/2
             *(BandwidthNoise - PolarizNoiseBandwidth)
             /(ChannSpacing - PolarizNoiseBandwidth);
   } 
   */         
   if (BandwidthNoise <= ChannSpacing) {
      OrthogPolarizedNoise = UnpolarizedNoise/2.
             *BandwidthNoise/ChannSpacing; 
   }   
   else {
      cout << "From: ReduOptSignal::GetNoiseInsideBandwidth"<< endl
           << "The bandwith: " << BandwidthNoise << " is larger"
           << "than the channel spacing."<< endl;
      exit(1);   
   }
   
   return OrthogPolarizedNoise;               
}


/*
// ##### Get the OrthogPolarized noise within any arbitrary bandwidth
// ## By: Ivan Lima (7/8/19)
double ReduOptSignal::GetOrthogPolarizedNoiseInsideBandwidth(int ChannelArrayIndex,
                                              double BandwidthNoise )
{
   //## It is being assumed in the beggining that
   //#  SignalNoiseBandwidth <= PolarizNoiseBandwidth
   
   double PolarizedNoise = sqrt( sq(StokesNoise[1][ChannelArrayIndex])
                                +sq(StokesNoise[2][ChannelArrayIndex])
                                +sq(StokesNoise[3][ChannelArrayIndex]) );   
   double NoisePolarization[] = {0,0,0,0};
   if (PolarizedNoise != 0) {
      NoisePolarization[1] = StokesNoise[1][ChannelArrayIndex]
                            /PolarizedNoise;
      NoisePolarization[2] = StokesNoise[2][ChannelArrayIndex]
                            /PolarizedNoise;
      NoisePolarization[3] = StokesNoise[3][ChannelArrayIndex]
                            /PolarizedNoise;			       
   }				
   double UnpolarizedNoise = StokesNoise[0][ChannelArrayIndex]
                            -PolarizedNoise; 
					    
   double PolarizedSignal = sqrt( sq(StokesSignal[1][ChannelArrayIndex])
                                +sq(StokesSignal[2][ChannelArrayIndex])
                                +sq(StokesSignal[3][ChannelArrayIndex]) );   
   double SignalPolarization[] = {0,0,0,0};
   if (PolarizedSignal != 0) {
      SignalPolarization[1] = StokesSignal[1][ChannelArrayIndex]
                            /PolarizedSignal;
      SignalPolarization[2] = StokesSignal[2][ChannelArrayIndex]
                            /PolarizedSignal;
      SignalPolarization[3] = StokesSignal[3][ChannelArrayIndex]
                            /PolarizedSignal;			       
   }				

   double ScalarProdSignalNoise =  
                   SignalPolarization[1] * NoisePolarization[1]
		  +SignalPolarization[2] * NoisePolarization[2]
		  +SignalPolarization[3] * NoisePolarization[3];
					    
   double OrthogPolarizedNoiseCoeff = 0.5 
                *(-ScalarProdSignalNoise + 1)
		*PolarizedNoise/(PolarizedNoise + UnpolarizedNoise) 
	      +0.5*( UnpolarizedNoise/(PolarizedNoise + UnpolarizedNoise) );  
  
   double OrthogPolarizedNoise;   
   if (BandwidthNoise <= PolarizNoiseBandwidth) {
      OrthogPolarizedNoise = StokesNoise[0][ChannelArrayIndex]
             *OrthogPolarizedNoiseCoeff
             *BandwidthNoise/PolarizNoiseBandwidth; 
   }
   else if (   BandwidthNoise > PolarizNoiseBandwidth 
            && BandwidthNoise <= ChannSpacing) {
      OrthogPolarizedNoise = StokesNoise[0][ChannelArrayIndex]          
             *OrthogPolarizedNoiseCoeff
             +StokesExtraUnpolChannNoise/NumChannels
             *(BandwidthNoise - PolarizNoiseBandwidth)
             /(ChannSpacing - PolarizNoiseBandwidth);
   }          
   else {
      cout << "From: ReduOptSignal::GetNoiseInsideBandwidth"<< endl
           << "The bandwith: " << BandwidthNoise << " is larger"
           << "than the channel spacing."<< endl;
      exit(1);   
   }
   
   return OrthogPolarizedNoise;               
}
<<<<<<< ocsReduOptSignal.cc

=======
*/



//####################################################################3
//## By: Ivan Lima (7/31/02)
double ReduOptSignal::GetScalarProdSignalNoise(
                            int ChannelArrayIndex)
{
   double PolarizedNoise = sqrt( sq(StokesNoise[1][ChannelArrayIndex])
                                +sq(StokesNoise[2][ChannelArrayIndex])
                                +sq(StokesNoise[3][ChannelArrayIndex]) );   
   double NoisePolarization[] = {0,0,0,0};
   if (PolarizedNoise != 0) {
      NoisePolarization[1] = StokesNoise[1][ChannelArrayIndex]
                            /PolarizedNoise;
      NoisePolarization[2] = StokesNoise[2][ChannelArrayIndex]
                            /PolarizedNoise;
      NoisePolarization[3] = StokesNoise[3][ChannelArrayIndex]
                            /PolarizedNoise;			       
   }	   

   double PolarizedSignal = sqrt( sq(StokesSignal[1][ChannelArrayIndex])
                                +sq(StokesSignal[2][ChannelArrayIndex])
                                +sq(StokesSignal[3][ChannelArrayIndex]) ); 
   double SignalPolarization[] = {0,0,0,0};
   if (PolarizedSignal != 0) {
      SignalPolarization[1] = StokesSignal[1][ChannelArrayIndex]
                            /PolarizedSignal;
      SignalPolarization[2] = StokesSignal[2][ChannelArrayIndex]
                            /PolarizedSignal;
      SignalPolarization[3] = StokesSignal[3][ChannelArrayIndex]
                            /PolarizedSignal;			       
   }				

   double ScalarProdSignalNoise =  
                   SignalPolarization[1] * NoisePolarization[1]
		  +SignalPolarization[2] * NoisePolarization[2]
		  +SignalPolarization[3] * NoisePolarization[3];
   return ScalarProdSignalNoise;		  
}			    






//####################################################################3
//## By: Ivan Lima (7/31/02)
double ReduOptSignal::GetNormSignalNoiseBeatingCopolarized(
                            int ChannelArrayIndex)
{
   double ScalarProdSignalNoise = GetScalarProdSignalNoise(ChannelArrayIndex);
   return 0.5*(1.+ ScalarProdSignalNoise);   		  					       
}




//####################################################################3
//## By: Ivan Lima (7/31/02)
double ReduOptSignal::GetNormSignalNoiseBeatingOrthogPolarized(
                            int ChannelArrayIndex)
{
   /***
   double PolarizedNoise = sqrt( sq(StokesNoise[1][ChannelArrayIndex])
                                +sq(StokesNoise[2][ChannelArrayIndex])
                                +sq(StokesNoise[3][ChannelArrayIndex]) );   
   double NoisePolarization[] = {0,0,0,0};
   if (PolarizedNoise != 0) {
      NoisePolarization[1] = StokesNoise[1][ChannelArrayIndex]
                            /PolarizedNoise;
      NoisePolarization[2] = StokesNoise[2][ChannelArrayIndex]
                            /PolarizedNoise;
      NoisePolarization[3] = StokesNoise[3][ChannelArrayIndex]
                            /PolarizedNoise;			       
   }	   

   double PolarizedSignal = sqrt( sq(StokesSignal[1][ChannelArrayIndex])
                                +sq(StokesSignal[2][ChannelArrayIndex])
                                +sq(StokesSignal[3][ChannelArrayIndex]) ); 
   double SignalPolarization[] = {0,0,0,0};
   if (PolarizedSignal != 0) {
      SignalPolarization[1] = StokesSignal[1][ChannelArrayIndex]
                            /PolarizedSignal;
      SignalPolarization[2] = StokesSignal[2][ChannelArrayIndex]
                            /PolarizedSignal;
      SignalPolarization[3] = StokesSignal[3][ChannelArrayIndex]
                            /PolarizedSignal;			       
   }				

   double ScalarProdSignalNoise =  
                   SignalPolarization[1] * NoisePolarization[1]
		  +SignalPolarization[2] * NoisePolarization[2]
		  +SignalPolarization[3] * NoisePolarization[3];
   ******/		  
   double ScalarProdSignalNoise = GetScalarProdSignalNoise(ChannelArrayIndex);
   return 0.5*(1.- ScalarProdSignalNoise);   		  					       
}



// ##############################################################



// ##### Calculates the Outage probability 
// ##    using Gaussian approximation

/*!
   In an application this method is used to compute the probability 
   that $\Delta Q$ is larger than MaxAcceptablePenalty. It assumes
   that $\Delta Q$ is Gaussian distributed with mean MeanPenalty and 
   standard deviation  StdDevPenalty.

 */

double ReduOptSignal::GetOutageProbGaussian(double MeanPenalty, 
                   double StdDevPenalty, double MaxAcceptablePenalty)
{
   return 0.5*erffc( (MeanPenalty - MaxAcceptablePenalty)
                    /StdDevPenalty/sqrt(2) );
} 
                    
// ###################################################

/*!

   \param outFile: The name of the output file (a char array)

   The following data is written to the output file in order.
   In this discussion $S_0$ denotes the power, $(S_1,S_2,S_3)$ 
   is the Stokes vector, $s_i = S_i/S_0$ for $i=1,2,3$ are the
   components of the normalized Stokes vector, and the normalized power
   is $s_0 = 1$.

   For each channel:
   - First for the signal in that channel
    -# The power in the given channel.
    -# The fraction of the signal or noise that is polarized.
       This is  defined to be the
       length of the normalized Stokes vector $(s_1,s_2,2_3)$.
    -# The fraction of the signal (or noise) that is Unpolarized, which
       is by definition 1 - polarized part of signal (or noise).
    -# $s_0$, $(s_1,s_2,s_3)$.
   - Next the noise power in the channel which is defined to be 
     the sum of $S_0$ for the noise in the channel and the 
     fraction of the additional noise that lies in the channel bandwidth.
   - The total Stokes parameters are also output.

 */

// ## Creates a file with the total polarization
// ## By: Ivan Lima
void ReduOptSignal::WriteFileOptSignal(char *outFile)
{
   ofstream OutFile(outFile,ios::out);
   if (!OutFile) {
      cerr << "Error: " << outFile << "could not be created."<<endl;
      exit(1);
   }
   double StokesSignalTotal[4];
   double StokesNoiseTotal[4];
   
   for(int ii = 0; ii < 4; ii++)  {
      StokesSignalTotal[ii] = 0;
      StokesNoiseTotal[ii]  = 0;   
   }   

   double PolarizedSignal;
   double PolarizedNoise;
   
   for( int jj = 0; jj < NumChannels; jj++) {
      PolarizedSignal = sqrt( sq(StokesSignal[1][jj])
                             +sq(StokesSignal[2][jj])
                             +sq(StokesSignal[3][jj]) );

      PolarizedNoise = sqrt( sq(StokesNoise[1][jj])
                            +sq(StokesNoise[2][jj])
                            +sq(StokesNoise[3][jj]) );   

      //if (StokesNoise[0][jj] == 0 )
      //   StokesNoise[0][jj] = NEAR_ZERO;

      OutFile << endl;                              
      OutFile << "Components of the Channel: " << GetChannelPhysicalIndex(jj) 
              << " Relative Freq. [GHz] = " 
              << (GetFrequency(jj) - CenterFreq)*1e-9 <<endl;

      OutFile << " Signal Power: " <<  StokesSignal[0][jj] << endl;
      OutFile << " Polarized    " << PolarizedSignal/StokesSignal[0][jj] 
              << endl;
      OutFile << " Unpolarized  " << 1 - PolarizedSignal/StokesSignal[0][jj]
              << endl;   
      OutFile << " S0           " << StokesSignal[0][jj]/StokesSignal[0][jj] 
              << endl;   
      OutFile << " S1           " << StokesSignal[1][jj]/StokesSignal[0][jj] 
	      << endl;
      OutFile << " S2           " << StokesSignal[2][jj]/StokesSignal[0][jj] 
	      << endl;   
      OutFile << " S3           " << StokesSignal[3][jj]/StokesSignal[0][jj] 
	      << endl;           
      OutFile << endl;
   
      //OutFile << "Components of the optical Noise: " << endl;

      OutFile << " Noise Power: " <<  StokesNoise[0][jj] << endl;
      //        +StokesExtraUnpolChannNoise/NumChannels << endl;

      if (StokesNoise[0][jj] != 0) 
       {
         OutFile << "Noise inside the channel bandwidth "<< endl;
         OutFile << " Polarized    " << PolarizedNoise/StokesNoise[0][jj] 
		 << endl;
         OutFile << " Unpolarized  " << 1. - PolarizedNoise/StokesNoise[0][jj]
		 << endl;   
         OutFile << " S0           " << StokesNoise[0][jj]/StokesNoise[0][jj] 
		 << endl;   
         OutFile << " S1           " << StokesNoise[1][jj]/StokesNoise[0][jj] 
		 << endl;
         OutFile << " S2           " << StokesNoise[2][jj]/StokesNoise[0][jj] 
		 << endl;   
         OutFile << " S3           " << StokesNoise[3][jj]/StokesNoise[0][jj] 
		 << endl;            

      }   

      for(int ii = 0; ii < 4; ii++)  
       {
         //cout << ii << " " << jj <<" " << StokesSignal[ii][jj] << endl;
         StokesSignalTotal[ii] += StokesSignal[ii][jj];
         StokesNoiseTotal[ii]  += StokesNoise[ii][jj];   
       }   
   } // ## end for jj loop ocer channels   

   PolarizedSignal = sqrt( sq(StokesSignalTotal[1])
                          +sq(StokesSignalTotal[2])
                          +sq(StokesSignalTotal[3]) );

   PolarizedNoise = sqrt( sq(StokesNoiseTotal[1])
                         +sq(StokesNoiseTotal[2])
                         +sq(StokesNoiseTotal[3]) );

   //if (StokesNoiseTotal[0] == 0 )
   //   StokesNoiseTotal[0] = NEAR_ZERO;

   OutFile << "==================================="<< endl;                
              
   OutFile << "Components of the optical Signal: " << endl;
   OutFile << " Total Signal Power: " <<  StokesSignalTotal[0] << endl;
   OutFile << " Polarized    " << PolarizedSignal/StokesSignalTotal[0] << endl;
   OutFile << " Unpolarized  " << 1 - PolarizedSignal/StokesSignalTotal[0]
	   << endl;   
   OutFile << " S0           " << StokesSignalTotal[0]/StokesSignalTotal[0] 
	   << endl;   
   OutFile << " S1           " << StokesSignalTotal[1]/StokesSignalTotal[0] 
	   << endl;
   OutFile << " S2           " << StokesSignalTotal[2]/StokesSignalTotal[0] 
	   << endl;   
   OutFile << " S3           " << StokesSignalTotal[3]/StokesSignalTotal[0] 
	   << endl;           
   OutFile << endl;   

//   OutFile << "Components of the optical Noise: " << endl;
   OutFile << " Total Noise Power: " <<  StokesNoiseTotal[0] << endl;
   //                                     +StokesExtraUnpolChannNoise << endl;
   //OutFile << " Noise Signal BW  : " <<  StokesNoiseTotal[0] << endl;
   //OutFile << " Unpol Extra Noise: " << StokesExtraUnpolChannNoise<< endl;  
    
   if (StokesNoiseTotal[0] != 0) 
    { 
      OutFile << "Noise inside the channel bandwidth "<< endl;   
      OutFile << " Polarized    " << PolarizedNoise/StokesNoiseTotal[0] 
	      << endl;
      OutFile << " Unpolarized  " << 1 - PolarizedNoise/StokesNoiseTotal[0] 
	      << endl;   
      OutFile << " S0           " << StokesNoiseTotal[0]/StokesNoiseTotal[0] 
	      << endl;   
      OutFile << " S1           " << StokesNoiseTotal[1]/StokesNoiseTotal[0] 
	      << endl;
      OutFile << " S2           " << StokesNoiseTotal[2]/StokesNoiseTotal[0] 
	      << endl;   
      OutFile << " S3           " << StokesNoiseTotal[3]/StokesNoiseTotal[0] 
	      << endl;           
   }   

   OutFile.close();   
}

// ############################################################

  //! JZAug02

void ReduOptSignal::WriteMatlabFileOptSignal(string OutFileName)
{

   ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
    }


  OutFile << "%(1) RelFreq [Hz] (2) S0 [W] (3-5) S1 to S3, Normalized" << endl;

  for( int jj = 0; jj < NumChannels; jj++) 
   {
     
     OutFile << GetFrequency(jj) - CenterFreq << "\t"
	     << StokesSignal[0][jj] << "\t"
	     << StokesSignal[1][jj]/StokesSignal[0][jj]  << "\t"
	     << StokesSignal[2][jj]/StokesSignal[0][jj]  << "\t"
	     << StokesSignal[3][jj]/StokesSignal[0][jj]  
	     << endl;
     
   }

} // ## end WriteMatlabFileOptSignal

// #######################################################
// ## This method is a straight copy from OptSignal class

/*! The input Stokes parameters, InStokes, for the signal in each channel
    are read from the input file stream InFile.
    If the InStokes parameters for channels 1,2 are not specified in the 
    input parameter file then we abort.
    If the InStokes parameters for a channel whose index is larger 
    than two are not specified in the input parameter file then we proceed
    as follows.

    -# If the channel has an odd index then we use the InStokes parameters
       specified for Channel 1.
    -# If the channel has an even index then we use the InStokes parameters
       specified for Channel 2.

*/

void ReduOptSignal::InitializeInStokes(ifstream * InFile)
{

char C_String[20]; 
string InStokesName;
int ParameterReadFlag;

  for(int ch = 0; ch < NumChannels;ch++)
    {

     // To create a string object from a string and an int
     // it seems simplest to use the sprintf function for
     // C Strings and then initialize the string using the
     // C String.

     sprintf(C_String,"InStokesChann%d",ch+1);
     InStokesName = C_String;

     ParameterReadFlag = ReadInStokes(InStokesName,ch,InFile);


     // If the InStokes parameters for channels 1,2 are not present
     // we abort.
     // If the InStokes parameters for a channel > 2 are not present
     // we use channel 1 parameters for odd channels and channel 2 
     // parameters for even channels

     if(ch < 2 && !ParameterReadFlag)
       {
         *LogFile << "ERROR initializing Optical Signal." << endl
		 << "InStokesChann1 and InStokesChann2 must be specified" 
                 << endl
		 << "ABORTING NOW"
		 << endl;

         exit(1);

       }

     if(ch>=2 && !ParameterReadFlag)
       {
	 SetInStokes(ch,GetInStokes(ch%2));

         *LogFile << endl << "WARNING: "
		 << endl
                 << "Since InStokesChann" << ch+1 << " was not specified"
	         << endl
	         << "set InStokesChann" << ch+1 << " = " << "InStokesChann" 
                 << (ch)%2 +1
                 << endl;
       }	
     
     
    } // end for ch loop

} // ## end InitializeInStokes

// ######################################################
// ## This method is a straight copy from OptSignal class

int ReduOptSignal::ReadInStokes(string InStokesName, int ch, ifstream * InFile)
{

  int ParameterRead = 0;

  string NameBuffer;
  string ParameterNameDollar = "$"+InStokesName;

  double Stokes[4];

   InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
      if (ParameterNameDollar==NameBuffer) 
          {

            *InFile >> Stokes[0] >> Stokes[1] >> Stokes[2] >> Stokes[3];

	    SetInStokes(ch,&Stokes[0]);

	    ParameterRead = 1;

      }  // end if   
    } // end while

  return ParameterRead;

}

// ###################################################
// ## This method is a straight copy from OptSignal class
void ReduOptSignal::SetInStokes(const int ChannelArrayIndex, 
                            const int StokesIndex, 
                            double StokesValue)                            
{
  //## Written by John Zweck Oct 11, 2000

  // ## FIX: Add bounds checking!!

  InStokes[StokesIndex + ChannelArrayIndex*4]  = StokesValue;
}

// #######################################################
// ## This method is a straight copy from OptSignal class
void ReduOptSignal::SetInStokes(const int ChannelArrayIndex,
                               double * StokesVector)
{

  for(int StokesIndex = 0; StokesIndex < 4; StokesIndex ++)
    InStokes[StokesIndex + ChannelArrayIndex*4] = StokesVector[StokesIndex];

}

// #######################################################
// ## This method is a straight copy from OptSignal class
double ReduOptSignal::GetInStokes(const int ChannelArrayIndex,
                                  const int StokesIndex)
{
  //## Written by John Zweck Oct 11, 2000

  // ## FIX: Add bounds checking!!

   return InStokes[StokesIndex + ChannelArrayIndex*4];
}


// #####################################################
// ## This method is a straight copy from OptSignal class
double * ReduOptSignal::GetInStokes(const int ChannelArrayIndex)
{
  return &InStokes[ChannelArrayIndex*4];
}

// #########################################################
// ## This method is a straight copy from BitString class

/*!
    ChannelPhysicalIndex runs:\n
    -N/2,..... N/2 - 1 for N = NumChannels even and\n
    -(N-1)/2, ... (N-1)/2 for N odd\n
  
    ChannelArrayIndex runs 0,1,...,N-1\n

\sa ProcessReduOptSignalParameters (ifstream * InFile) 
    GetChannelPhysicalIndex(int ChannelArrayIndex)
*/
                                                                      
int ReduOptSignal::GetChannelArrayIndex(int ChannelPhysicalIndex)
{

  //## Written by John Zweck Oct 11, 2000
  //## 
  //## ChannelPhysicalIndex runs 
  //## -N/2,..... N/2 - 1 for N = NumChannels even and
  //## - (N-1)/2, ... (N-1)/2 for N odd
  //##
  //## ChannelArrayIndex runs 0,1,...,N-1

  // FIX Add bounds checker!!

  if(NumChannels %2 == 0) // NumChannels even 
    return ChannelPhysicalIndex + NumChannels/2;
  else // NumChannels odd
    return ChannelPhysicalIndex + (NumChannels-1)/2;

} // ## end GetChannelArrayIndex

// ##############################################
// ## This method is a straight copy from BitString class

/*!
    ChannelPhysicalIndex runs:\n
    -N/2,..... N/2 - 1 for N = NumChannels even and\n
    -(N-1)/2, ... (N-1)/2 for N odd\n
  
    ChannelArrayIndex runs 0,1,...,N-1\n

\sa ProcessReduOptSignalParameters (ifstream * InFile) 
    GetChannelArrayIndex(int ChannelPhysicalIndex)

*/

int ReduOptSignal::GetChannelPhysicalIndex(int ChannelArrayIndex)
{

  //## Written by John Zweck Oct 11, 2000
  //## 
  //## ChannelPhysicalIndex runs 
  //## -N/2,..... N/2 - 1 for N = NumChannels even and
  //## - (N-1)/2, ... (N-1)/2 for N odd
  //##
  //## ChannelArrayIndex runs 0,1,...,N-1

  // FIX Add bounds checker!

  if(NumChannels %2 == 0) // NumChannels even 
    return ChannelArrayIndex - NumChannels/2;
  else // NumChannels odd
    return ChannelArrayIndex - (NumChannels-1)/2;
}

// #####################################################
// ## This method is a straight copy from OptSignal class

const char *ReduOptSignal::WriteTypePulse(void)
{

  char * name;

  switch(TypePulse)
    {
      case NRZ: name = "NRZ"; break;
      case RZ_GAUSS: name = "RZ Gaussian"; break;
      case RZ_SECH: name = "RZ Sech"; break;
      case CRZ_GAUSS: name = "CRZ Gaussian"; break;
      case CRZ_RCOS: name = "CRZ Raised Cosine"; break;
      default : name = "Invalid Pulse Type";
    }

  return name;  

}

// ################### End of the class ReduOptSignal


