// $Id: 
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsReduPolDepAttenuator.hh" 


extern ofstream *LogFile;  


// ################### Methods for the class ReduPolDepAttenuator

// ################################################################

// #### Constructor of ReduPolDepAttenuator
// ## By: Ivan Lima

ReduPolDepAttenuator::ReduPolDepAttenuator
                      ( ReduOptSignal *oReduOptSignalTmp)
{
   InitPolDepAttenuator( oReduOptSignalTmp);
}

// ##################################################################

// #### Constructor of ReduPolDepAttenuator w/ initialization
// ## By: Ivan Lima



ReduPolDepAttenuator::ReduPolDepAttenuator( 
                           ReduOptSignal *oReduOptSignalTmp
                          ,double PolDepAttenuationXTmp
                          ,double PolDepAttenuationYTmp)  
{
   InitPolDepAttenuator( oReduOptSignalTmp);
   SetPolDepAttenuation( PolDepAttenuationXTmp
                        ,PolDepAttenuationYTmp);
}

// #################################################################

// ## Initializes the parameters

void ReduPolDepAttenuator::InitPolDepAttenuator
                      (ReduOptSignal *oReduOptSignalTmp)
{
   oReduOptSignal = oReduOptSignalTmp;
   NumChannels = oReduOptSignal->GetNumChannels();
   ObjectInitialized = 0;
}               

// ############################################################

// ## Allows for a change in the loss at both polarizations

void ReduPolDepAttenuator::SetPolDepAttenuation(
                             double PolDepAttenuationXTmp
                            ,double PolDepAttenuationYTmp)
{
   SetPolDepAttenuationX(PolDepAttenuationXTmp);
   SetPolDepAttenuationY(PolDepAttenuationYTmp);              
   ObjectInitialized = 1;   
}                            

// #########################################################
           
// ## Allows for a change in the loss at the X polarization

void ReduPolDepAttenuator::SetPolDepAttenuationX(
                              double PolDepAttenuationXTmp)
{
   // ## The factor /2. comes from the fact that 
   // #  the LinearPolDepAttenuationX is actually
   // #  from the field, not from the power.
   // #  But the attenuation is done correctly in
   // #  the method PolDepAttenuate.

   LinearPolDepAttenuationX 
           = dB2Linear(-PolDepAttenuationXTmp/2.);
   GetTransmMatrixCoefficients();
}

// ###############################################################

// ## Allows for a change in the loss at the Y polarization
void ReduPolDepAttenuator::SetPolDepAttenuationY(
                              double PolDepAttenuationYTmp)
{
   // ## The factor /2. comes from the fact that 
   // #  the LinarPolDepAttenuationX is actually
   // #  from the field, not from the power.
   // #  But the attenuation is done correctly in
   // #  the method PolDepAttenuator.
   LinearPolDepAttenuationY 
           = dB2Linear(-PolDepAttenuationYTmp/2.);
   GetTransmMatrixCoefficients();   
}


// ##########################################################

// ## Syncronize the number of channels with OptSignal.
// #  By: Ivan Lima

void ReduPolDepAttenuator::SyncNumChannels(void)
{
   NumChannels = oReduOptSignal->GetNumChannels();
}

// ##########################################################

//## Update the Attenuation matrix

void ReduPolDepAttenuator::GetTransmMatrixCoefficients(void)
{
   // ## Transmission coefficients of the sparse 4x4 matrix
   Tm00 = ( sq(LinearPolDepAttenuationX) 
           +sq(LinearPolDepAttenuationY))/2.;
   Tm01 = ( sq(LinearPolDepAttenuationX) 
           -sq(LinearPolDepAttenuationY))/2.; 
   Tm22 =  LinearPolDepAttenuationX 
          *LinearPolDepAttenuationY;  
   //   cout << Tm00 << endl << Tm01 << endl << Tm22 << endl;               
}

// ############################################################

// #### Propagates through the PolDepAttenuator
// ## By: Ivan Lima

void ReduPolDepAttenuator::PolDepAttenuate(void)
{
   if (! ObjectInitialized) 
   {
      cerr << "Error: Object of class ReduPolDepAttenuator"
           << " not initialized" 
           << endl;
      exit(1);
   }                      

   //cout << "LinearPolDepAttenuationX = " << LinearPolDepAttenuationX << endl;
   //cout << "LinearPolDepAttenuationY = " << LinearPolDepAttenuationY << endl;

   double StokesSignalTmp[4];
   double StokesNoiseTmp[4];     
   
   //## The extra unpolarized noise component

   //oReduOptSignal->StokesExtraUnpolChannNoise *= Tm00;       

   for (int jj = 0; jj < NumChannels; jj++) 
    {
      for (int ii = 0; ii < 4; ii++) 
       {
         StokesSignalTmp[ii] = 
                  oReduOptSignal->StokesSignal[ii][jj];
         StokesNoiseTmp[ii]  =  
                  oReduOptSignal->StokesNoise[ii][jj];
      }                    
      
      // ## Actual propagation through the PolDepAttenuator

      oReduOptSignal->StokesSignal[0][jj] = 
                        Tm00 * StokesSignalTmp[0]
                       +Tm01 * StokesSignalTmp[1]; 
      oReduOptSignal->StokesSignal[1][jj] = 
                        Tm01 * StokesSignalTmp[0]
                       +Tm00 * StokesSignalTmp[1];             
      oReduOptSignal->StokesSignal[2][jj] = 
                        Tm22 * StokesSignalTmp[2];
      oReduOptSignal->StokesSignal[3][jj] = 
                       Tm22 * StokesSignalTmp[3];  
                       
      oReduOptSignal->StokesNoise[0][jj] = 
                       Tm00 * StokesNoiseTmp[0]
                      +Tm01 * StokesNoiseTmp[1]; 
      oReduOptSignal->StokesNoise[1][jj] = 
                       Tm01 * StokesNoiseTmp[0]
                      +Tm00 * StokesNoiseTmp[1];             
      oReduOptSignal->StokesNoise[2][jj] = 
                       Tm22 * StokesNoiseTmp[2];
      oReduOptSignal->StokesNoise[3][jj] = 
                       Tm22 * StokesNoiseTmp[3];         
                       
                     
   } 
}


// #################################################################




