// $Id: ocsReduPolarizTransformer.cc,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $
//###################################################################
//
//       Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsReduPolarizTransformer.hh"

extern ofstream *LogFile;

// ########### Methods for the class ReduPolarizTransformer         

// #### Initializes the object ReduPolarizTransformer
// ## By: Ivan Lima

ReduPolarizTransformer::ReduPolarizTransformer
                  (int TypeScattering2, 
                   ReduOptSignal *oReduOptSignalTmp)
{
   oReduOptSignal = oReduOptSignalTmp;
   FlagImportSamplingInitialized = 0;   

   TypeScattering = typeScattering(TypeScattering2);
   
   LogFileSeparator();   
   *LogFile << "Initializing ReduPolarizTransformer" << endl;
   *LogFile << "Type Scattering = " << WriteTypeScattering() << endl;

   //## Default values -> Do nothing
   psiLast = thetaLast = phiLast = 0;
}



// #### Returns the TypeScattering
// ## By: Ivan Lima

string ReduPolarizTransformer::WriteTypeScattering(void)
{
   string sTypeScattering;

   switch(TypeScattering) {
      case RX_RY_RX:
         sTypeScattering = "RX_RY_RX == 1: The Euler Matrix";
         break;
      case RY_RZ_RY: 
         sTypeScattering = "RY_RZ_RY == 2";
         break;
      case  RX_RZ_RY:
         sTypeScattering = "RX_RZ_RY == 3";  
         break;
      default:
         cout << "From: ReduPolarizTransformer"<< endl;
         cout << "ERROR: Invalid choice of TypeScattering" << endl;
         cout << "TypeScattering = " << TypeScattering << endl;
         cout << "ABORTING NOW!!" << endl;
         exit(1);
   }
   return sTypeScattering;
}



//#### Polarization Transformation
//## It causes the following rotations based on the 
//#  selected sequence of rotation
//## By: Ivan Lima

void ReduPolarizTransformer::TransformPolarization(
                 double psi
                ,double theta
                ,double phi )
{
   double StokesSignalShifted[3];
   double StokesNoiseShifted[3];

   psiLast   = psi;
   thetaLast = theta;
   phiLast   = phi;

   double NumChannels = oReduOptSignal->GetNumChannels();

   for (int jj = 0; jj < NumChannels; jj++) 
    {
      for (int ii = 1; ii < 4; ii++) 
       {
         StokesSignalShifted[ii-1] 
               = oReduOptSignal->StokesSignal[ii][jj];    
         StokesNoiseShifted[ii-1] 
               = oReduOptSignal->StokesNoise[ii][jj];          
      }   

      switch (TypeScattering) 
       {
         case RX_RY_RX: // ## : Euler Matrix

               RotatesAboutX( phi,   StokesSignalShifted); 
               RotatesAboutY(-theta, StokesSignalShifted); 
               RotatesAboutX( psi,   StokesSignalShifted); 

               RotatesAboutX( phi,   StokesNoiseShifted); 
               RotatesAboutY(-theta, StokesNoiseShifted); 
               RotatesAboutX( psi,   StokesNoiseShifted); 

          case RY_RZ_RY: // ## RyRzRy      
               RotatesAboutY( -psi,   StokesSignalShifted); 
               RotatesAboutZ(  theta, StokesSignalShifted); 
               RotatesAboutY( -phi,   StokesSignalShifted); 

               RotatesAboutY( -psi,   StokesNoiseShifted); 
               RotatesAboutZ(  theta, StokesNoiseShifted); 
               RotatesAboutY( -phi,   StokesNoiseShifted); 
               break;
          case RX_RZ_RY: // ## RxRzRy      
               RotatesAboutY( -psi,   StokesSignalShifted); 
               RotatesAboutZ(  theta, StokesSignalShifted); 
               RotatesAboutX(  phi,   StokesSignalShifted); 

               RotatesAboutY( -psi,   StokesNoiseShifted); 
               RotatesAboutZ(  theta, StokesNoiseShifted); 
               RotatesAboutX(  phi,   StokesNoiseShifted); 

               break;
      }

      // ## Put together the whole rotated Stokes parameters

      for (int ii = 1; ii < 4; ii++) 
       {
         oReduOptSignal->StokesSignal[ii][jj] 
                      = StokesSignalShifted[ii-1];    
         oReduOptSignal->StokesNoise[ii][jj] 
                      = StokesNoiseShifted[ii-1];          
      }                                        
   }                                                   
} 

// ########################################################


 
// #### Polarization transformation
// ## It causes the following rotations based on the Euler angles:
// ## Rx(phi = 0)  Rz(theta) Ry(-psi)
// #  It cannot provide a uniform rotation on the 
// #  Poincare sphere, but is useful for dithering, 
// #  since it requires the control of only two parameters
// ## By: Ivan Lima

void ReduPolarizTransformer::TransformPolarization(
                 double psi
                ,double theta)
{ 
   TransformPolarization(psi,theta,0.);    
}

// #########################################


//## Produce the polarization transformation selected 
//#  in the angles psiLast, thetaLast and phiLast
//#  This function is particularly useful for the 
//#  application of the importance sampling.

//## By: Ivan Lima 

void ReduPolarizTransformer::TransformPolarization(void)
{ 
   TransformPolarization(psiLast,thetaLast,phiLast);
}  

// ##################################################### 
 
//## Undo the last polarization transformation
//## By: Ivan Lima 

void ReduPolarizTransformer::UndoLastPolarizTransformation(void)
{
   double StokesSignalShifted[3];
   double StokesNoiseShifted[3];

   double psi, theta, phi, delta, gamma;

   psi   = -psiLast;
   theta = -thetaLast;
   phi   = -phiLast;

   double NumChannels = oReduOptSignal->GetNumChannels();

   for (int jj = 0; jj < NumChannels; jj++) 
    {
      for (int ii = 1; ii < 4; ii++) 
      {
         StokesSignalShifted[ii-1] 
               = oReduOptSignal->StokesSignal[ii][jj];    
         StokesNoiseShifted[ii-1] 
               = oReduOptSignal->StokesNoise[ii][jj];          
      }   
      switch (TypeScattering) {
         case RX_RY_RX: // ## : Euler Matrix
               RotatesAboutX( psi,   StokesSignalShifted); 
               RotatesAboutY(-theta, StokesSignalShifted); 
               RotatesAboutX( phi,   StokesSignalShifted); 

               RotatesAboutX( psi,   StokesNoiseShifted); 
               RotatesAboutY(-theta, StokesNoiseShifted); 
               RotatesAboutX( phi,   StokesNoiseShifted); 
               break;
          case RY_RZ_RY: // ## RyRzRy      
               RotatesAboutY( -phi,   StokesSignalShifted); 
               RotatesAboutZ(  theta, StokesSignalShifted); 
               RotatesAboutY( -psi,   StokesSignalShifted); 

               RotatesAboutY( -phi,   StokesNoiseShifted); 
               RotatesAboutZ(  theta, StokesNoiseShifted); 
               RotatesAboutY( -psi,   StokesNoiseShifted); 
               break;            
          case RX_RZ_RY: // ## RxRzRy      
               RotatesAboutX(  phi,   StokesSignalShifted); 
               RotatesAboutZ(  theta, StokesSignalShifted); 
               RotatesAboutY( -psi,   StokesSignalShifted); 

               RotatesAboutX(  phi,   StokesNoiseShifted); 
               RotatesAboutZ(  theta, StokesNoiseShifted); 
               RotatesAboutY( -psi,   StokesNoiseShifted); 
               break;

      }           

      // ## Put together the whole rotated Stokes parameters

      for (int ii = 1; ii < 4; ii++) 
      {
         oReduOptSignal->StokesSignal[ii][jj] 
                      = StokesSignalShifted[ii-1];    
         oReduOptSignal->StokesNoise[ii][jj] 
                      = StokesNoiseShifted[ii-1];          
      }                                        
   }                             
}


// #################################################

//## Initialization of the importance sampling applied to 
//#  the penalty induced by the combination of the 
//#  polarization effects
//## By: Ivan Lima

void ReduPolarizTransformer::InitImportSampPolarizTransformation(
                             double BiasingParam2,
                             RanNumGen *oRanImportSampPolarizTransform2) 
{
   BiasingParam = BiasingParam2;
   oRanImportSampPolarizTransform = oRanImportSampPolarizTransform2;
   FlagImportSamplingInitialized = 1;
}                             


// ####################################

//## Application of importance sampling to the polarization
//#  transformation for PDL. The rotation will bias the 
//#  Stokes vector of the ChannelArrayIndex toward where 
//#  the -X (-1,0,0), the vertical axis 
//## where the PDL element should be placed, 
//#  with a shift in the angle that is originated by
//#  the PDG effect.
//#  The actual rotations will be available in psiLast, 
//#  thetaLast and phiLast, which can be used to undo 
//#  the rotation after the PDL element using the method
//#  UndoLastPolarizTransformation().
//#  The Stokes vector should contain the Stokes parameters
//#  of the channel that is going to be biased.
//#  ATTENTION: The likelihood ratio should be initialized 
//#  to 1 at the beggining of each sample (propagation 
//#  through the system). The likelihood ratio should be
//#  passed to the Histogram class with the sample.
//#  If it is not done, the statistical
//#  results would be completelly wrong.

//## By: Ivan Lima  

void ReduPolarizTransformer::GetImportSampPolarizTransformation(
                             int ChannelArrayIndex,
                             double *LikelihoodRatio)

{
   if ( !FlagImportSamplingInitialized) {
      cout << "FlagImportSampling not initialized"<< endl;
      exit(1);   
   }


   double psi2, theta2, phi2, delta2, gamma2;
   double thetaZthetaY[2];
   double StokesSignalShifted[3];
   for (int ii = 1; ii < 4; ii++) 
    {
      StokesSignalShifted[ii-1] 
            = oReduOptSignal->GetStokesSignal(ii,ChannelArrayIndex);

   }

   //double PolarizedStokesSignal = sqrt( sq(StokesSignalShifted[0])
   //                                    +sq(StokesSignalShifted[1])
   //                                    +sq(StokesSignalShifted[2]));
   //## Normalized the polarized component of the signal
   //#  It is particularly important when PDG != 0                                   
   //for (int ii = 1; ii < 4; ii++) {
   //   StokesSignalShifted[ii-1] *= PolarizedStokesSignal; 
   //}   

/*****   
      double StokesTotal[4];   
      double StokesTotalShifted[3];
      oReduOptSignal->GetTotalStokes(StokesTotal);
      double PolarizedTotalStokes = sqrt( sq(StokesTotal[1])
                                         +sq(StokesTotal[2])
                                         +sq(StokesTotal[3]) );      
      for (int ii = 1; ii < 4; ii++) {
         StokesTotalShifted[ii-1] 
               = StokesTotal[ii]/PolarizedTotalStokes;
      }      
      cout <<"StokesTotalShifted = "
           << StokesTotalShifted[0] << " "
           << StokesTotalShifted[1] << " "
           << StokesTotalShifted[2] << endl;     
      cout <<"StokesSignalShifted = "
           << StokesSignalShifted[0] << " "
           << StokesSignalShifted[1] << " "
           << StokesSignalShifted[2] << endl;                                 
      cout << "ScalarProd = " 
           << StokesSignalShifted[0]*StokesTotalShifted[0]
             +StokesSignalShifted[1]*StokesTotalShifted[1]         
             +StokesSignalShifted[2]*StokesTotalShifted[2] << endl;         
********/   
   
      
   TransRzRyV3to_Xhat( StokesSignalShifted, thetaZthetaY);


   //## I should be able to tune the angle to where the cosine
   //#  should be biased with respect to the PDL element due to
   //#  the PDG effect.
   
   double SourceRandomVar;                 

      TypeScattering = RX_RZ_RY;      
      SourceRandomVar = oRanImportSampPolarizTransform->GetRanCosineBiased(
                                           BiasingParam,
                                           LikelihoodRatio);   
                                              
      //theta2  =  acos(SourceRandomVar);    
      //theta2 = sgn(oRanImportSampPolarizTransform->GetRanNum() - 0.5)
      //        *acos(SourceRandomVar);

      psi2    = -thetaZthetaY[1];
      theta2 =  thetaZthetaY[0]
               +acos(SourceRandomVar)
               *sgn(oRanImportSampPolarizTransform->GetRanNum() - 0.5);

      phi2    =  2*pi*oRanImportSampPolarizTransform->GetRanNum();   


   
   //## The transformation has to be called by the user
   //## Call TransformPolarization without arguments.
   //## For example: TransformPolarization()

   psiLast   = psi2;
   thetaLast = theta2;
   phiLast   = phi2;
}                                   


/************* 
//## Application of importance sampling to the polarization
//#  transformation for PDL. The rotation will bias the 
//#  Stokes vector of the ChannelArrayIndex toward the -X 
//#  (-1,0,0), the vertical axis,
//#  where the PDL element should be placed. 
//#  The actual rotations will be available in psiLast, 
//#  thetaLast and phiLast, which can be used to undo 
//#  the rotation after the PDL element using the method
//#  UndoLastPolarizTransformation().
//#  The Stokes vector should contain the Stokes parameters
//#  of the channel that is going to be biased.
//#  ATTENTION: The likelihood ratio should be initialized 
//#  to 1 at the beggining of each sample (propagation 
//#  through the system). The likelihood ratio should be
//#  passed to the Histogram class with the sample.
//#  If it is not done, the statistical
//#  results would be completelly wrong.
//## By: Ivan Lima  
void ReduPolarizTransformer::GetImportSampPolarizTransformation(
                             int ChannelArrayIndex,
                             double *LikelihoodRatio)
{
   if ( !FlagImportSamplingInitialized) {
      cout << "FlagImportSampling not initialized"<< endl;
      exit(1);   
   }
   TypeScattering = RX_RZ_RY;
   double psi2, theta2, phi2;
   double thetaZthetaY[2];
   double StokesSignalShifted[3];
   for (int ii = 1; ii < 4; ii++) {
      StokesSignalShifted[ii-1] 
            = oReduOptSignal->StokesSignal[ii][ChannelArrayIndex];
   }

   double SourceRandomVar = oRanImportSampPolarizTransform->GetRanNum();
   theta2 = acos(2.*pow(SourceRandomVar,1./BiasingParam) - 1. );     
   *LikelihoodRatio *= 1./BiasingParam
             *pow(SourceRandomVar,-(1.-1./BiasingParam));     
   
   TransRzRyV3to_Xhat( StokesSignalShifted, thetaZthetaY);
   psi2    = -thetaZthetaY[1];
   theta2 += thetaZthetaY[0];
   phi2    = 2*pi*oRanImportSampPolarizTransform->GetRanNum();
   
   //## The transformation has to be called by the user
   //## Call TransformPolarization();
   psiLast   = psi2;
   thetaLast = theta2;
   phiLast   = phi2;
      
   //////TransformPolarization(psi2, theta2, phi2);
}                                            
******/




