// $Id: 
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsReduPolarizer.hh"

// ################### Methods for the class ReduPolarizer         

//! Constructor. Sets default polarization splitting vector in Stokes space
//! to be the horizontal state (1,0,0).

ReduPolarizer::ReduPolarizer(ReduOptSignal *oReduOptSignalTmp)
{
   oReduOptSignal = oReduOptSignalTmp;
   SPolarizer = new double[4]; //Stokes parameters in reduced model!!   

  // ## Sets the default polarization as the horizontal   
   SetPolarization(1.0,0,0);
}

// ########################################################

//! Sets a new polarization splitting vector in Stokes space

void ReduPolarizer::SetPolarization(double S1PolarizerTmp,
				    double S2PolarizerTmp,
				    double S3PolarizerTmp)
{
   SPolarizer[1] = S1PolarizerTmp;
   SPolarizer[2] = S2PolarizerTmp;
   SPolarizer[3] = S3PolarizerTmp; 
     
   double S0Polarizer = sqrt( sq(SPolarizer[1])
                             +sq(SPolarizer[2])
                             +sq(SPolarizer[3]) );

   if (S0Polarizer == 0) 
   {
      cerr << "Error: The polarization components of the " 
           << "Polarization cannot be zero."
           << endl;
      exit(1);
   }                      

   // ## Normalizing the polarization    

   SPolarizer[0] =  1.0;
   SPolarizer[1] =  SPolarizer[1]/S0Polarizer; 
   SPolarizer[2] =  SPolarizer[2]/S0Polarizer; 
   SPolarizer[3] =  SPolarizer[3]/S0Polarizer;      
}                       

// #########################################################

//! Polarizes the reduced optical signal

/*!

We derive an approximate formula in Stokes space for the action
of a polarizer on a reduced signal, based on the theory documented
in Polarizer::PolarizeOptSignal(). The approximation relies on the
assumption that the time integral of product of u(t) and v(t) is zero,
whenever u,v are a pair of OptSignals that respresent 
the signal in two different channels, noise in two different channels,
or signal and noise in a channel (or two different channels). This
assumption underlies the entire Wang_Menyuk reduced model (see for example
eqs (6) and (15) of the paper JLT19p487.

The formula is derived from Polarizer::PolarizeOptSignal() using
Gordon and Kogelnick Eq 3.11.

For the signal and noise in each channel, we separately set 

S_out = 0.5InnerProduct(SPolarizer,S_in)SPolarizer

where all Stokes parameters in the above formula are 4x1 vectors
and SPolarizer = (1,UnitThreeVector).

*/

void ReduPolarizer::PolarizeOptSignal()    
{

   for (int jj = 0; jj < oReduOptSignal->GetNumChannels(); jj++)
     {
      double InnerProduct = 0;

      for(int ii = 0; ii < 4; ii++)
       InnerProduct += oReduOptSignal->StokesSignal[ii][jj]*SPolarizer[ii];

      for(int ii = 0; ii < 4; ii++)
       oReduOptSignal->StokesSignal[ii][jj] = 0.5*InnerProduct*SPolarizer[ii];

      InnerProduct = 0;

      for(int ii = 0; ii < 4; ii++)
       InnerProduct += oReduOptSignal->StokesNoise[ii][jj]*SPolarizer[ii];

      for(int ii = 0; ii < 4; ii++)
       oReduOptSignal->StokesNoise[ii][jj] = 0.5*InnerProduct*SPolarizer[ii];

     } 

}
